/* Generated by Cython 0.29.33 */

/* BEGIN: Cython Metadata
{
    "distutils": {
        "depends": [],
        "include_dirs": [
            "src/glycopeptidepy/_c"
        ],
        "name": "glycopeptidepy._c.structure.glycan",
        "sources": [
            "src/glycopeptidepy/_c/structure/glycan.pyx"
        ]
    },
    "module_name": "glycopeptidepy._c.structure.glycan"
}
END: Cython Metadata */

#ifndef PY_SSIZE_T_CLEAN
#define PY_SSIZE_T_CLEAN
#endif /* PY_SSIZE_T_CLEAN */
#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02060000 || (0x03000000 <= PY_VERSION_HEX && PY_VERSION_HEX < 0x03030000)
    #error Cython requires Python 2.6+ or Python 3.3+.
#else
#define CYTHON_ABI "0_29_33"
#define CYTHON_HEX_VERSION 0x001D21F0
#define CYTHON_FUTURE_DIVISION 0
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#define __PYX_COMMA ,
#ifndef HAVE_LONG_LONG
  #if PY_VERSION_HEX >= 0x02070000
    #define HAVE_LONG_LONG
  #endif
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#ifdef PYPY_VERSION
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_NOGIL 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
#elif defined(PYSTON_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_NOGIL 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_ASYNC_SLOTS
  #define CYTHON_USE_ASYNC_SLOTS 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
#elif defined(PY_NOGIL)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_NOGIL 1
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #ifndef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE 1
  #endif
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #define CYTHON_COMPILING_IN_NOGIL 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYTYPE_LOOKUP
    #define CYTHON_USE_PYTYPE_LOOKUP 0
  #elif !defined(CYTHON_USE_PYTYPE_LOOKUP)
    #define CYTHON_USE_PYTYPE_LOOKUP 1
  #endif
  #if PY_MAJOR_VERSION < 3
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 0
  #elif !defined(CYTHON_USE_PYLONG_INTERNALS)
    #define CYTHON_USE_PYLONG_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if PY_VERSION_HEX < 0x030300F0 || PY_VERSION_HEX >= 0x030B00A2
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #if PY_VERSION_HEX >= 0x030B00A4
    #undef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 0
  #elif !defined(CYTHON_FAST_THREAD_STATE)
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL (PY_VERSION_HEX < 0x030A0000)
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT (PY_VERSION_HEX >= 0x03050000)
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE (PY_VERSION_HEX >= 0x030400a1)
  #endif
  #ifndef CYTHON_USE_DICT_VERSIONS
    #define CYTHON_USE_DICT_VERSIONS (PY_VERSION_HEX >= 0x030600B1)
  #endif
  #if PY_VERSION_HEX >= 0x030B00A4
    #undef CYTHON_USE_EXC_INFO_STACK
    #define CYTHON_USE_EXC_INFO_STACK 0
  #elif !defined(CYTHON_USE_EXC_INFO_STACK)
    #define CYTHON_USE_EXC_INFO_STACK (PY_VERSION_HEX >= 0x030700A3)
  #endif
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 1
  #endif
#endif
#if !defined(CYTHON_FAST_PYCCALL)
#define CYTHON_FAST_PYCCALL  (CYTHON_FAST_PYCALL && PY_VERSION_HEX >= 0x030600B1)
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #if PY_MAJOR_VERSION < 3
    #include "longintrepr.h"
  #endif
  #undef SHIFT
  #undef BASE
  #undef MASK
  #ifdef SIZEOF_VOID_P
    enum { __pyx_check_sizeof_voidp = 1 / (int)(SIZEOF_VOID_P == sizeof(void*)) };
  #endif
#endif
#ifndef __has_attribute
  #define __has_attribute(x) 0
#endif
#ifndef __has_cpp_attribute
  #define __has_cpp_attribute(x) 0
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_MAYBE_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_MAYBE_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)
#ifdef _MSC_VER
    #ifndef _MSC_STDINT_H_
        #if _MSC_VER < 1300
           typedef unsigned char     uint8_t;
           typedef unsigned int      uint32_t;
        #else
           typedef unsigned __int8   uint8_t;
           typedef unsigned __int32  uint32_t;
        #endif
    #endif
#else
   #include <stdint.h>
#endif
#ifndef CYTHON_FALLTHROUGH
  #if defined(__cplusplus) && __cplusplus >= 201103L
    #if __has_cpp_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH [[fallthrough]]
    #elif __has_cpp_attribute(clang::fallthrough)
      #define CYTHON_FALLTHROUGH [[clang::fallthrough]]
    #elif __has_cpp_attribute(gnu::fallthrough)
      #define CYTHON_FALLTHROUGH [[gnu::fallthrough]]
    #endif
  #endif
  #ifndef CYTHON_FALLTHROUGH
    #if __has_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH __attribute__((fallthrough))
    #else
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
  #if defined(__clang__ ) && defined(__apple_build_version__)
    #if __apple_build_version__ < 7000000
      #undef  CYTHON_FALLTHROUGH
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
#endif

#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #elif defined(__GNUC__)
    #define CYTHON_INLINE __inline__
  #elif defined(_MSC_VER)
    #define CYTHON_INLINE __inline
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_INLINE inline
  #else
    #define CYTHON_INLINE
  #endif
#endif

#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x02070600 && !defined(Py_OptimizeFlag)
  #define Py_OptimizeFlag 0
#endif
#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a+k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyClass_Type
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
  #define __Pyx_DefaultClassType PyType_Type
#if PY_VERSION_HEX >= 0x030B00A1
    static CYTHON_INLINE PyCodeObject* __Pyx_PyCode_New(int a, int k, int l, int s, int f,
                                                    PyObject *code, PyObject *c, PyObject* n, PyObject *v,
                                                    PyObject *fv, PyObject *cell, PyObject* fn,
                                                    PyObject *name, int fline, PyObject *lnos) {
        PyObject *kwds=NULL, *argcount=NULL, *posonlyargcount=NULL, *kwonlyargcount=NULL;
        PyObject *nlocals=NULL, *stacksize=NULL, *flags=NULL, *replace=NULL, *call_result=NULL, *empty=NULL;
        const char *fn_cstr=NULL;
        const char *name_cstr=NULL;
        PyCodeObject* co=NULL;
        PyObject *type, *value, *traceback;
        PyErr_Fetch(&type, &value, &traceback);
        if (!(kwds=PyDict_New())) goto end;
        if (!(argcount=PyLong_FromLong(a))) goto end;
        if (PyDict_SetItemString(kwds, "co_argcount", argcount) != 0) goto end;
        if (!(posonlyargcount=PyLong_FromLong(0))) goto end;
        if (PyDict_SetItemString(kwds, "co_posonlyargcount", posonlyargcount) != 0) goto end;
        if (!(kwonlyargcount=PyLong_FromLong(k))) goto end;
        if (PyDict_SetItemString(kwds, "co_kwonlyargcount", kwonlyargcount) != 0) goto end;
        if (!(nlocals=PyLong_FromLong(l))) goto end;
        if (PyDict_SetItemString(kwds, "co_nlocals", nlocals) != 0) goto end;
        if (!(stacksize=PyLong_FromLong(s))) goto end;
        if (PyDict_SetItemString(kwds, "co_stacksize", stacksize) != 0) goto end;
        if (!(flags=PyLong_FromLong(f))) goto end;
        if (PyDict_SetItemString(kwds, "co_flags", flags) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_code", code) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_consts", c) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_names", n) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_varnames", v) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_freevars", fv) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_cellvars", cell) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_linetable", lnos) != 0) goto end;
        if (!(fn_cstr=PyUnicode_AsUTF8AndSize(fn, NULL))) goto end;
        if (!(name_cstr=PyUnicode_AsUTF8AndSize(name, NULL))) goto end;
        if (!(co = PyCode_NewEmpty(fn_cstr, name_cstr, fline))) goto end;
        if (!(replace = PyObject_GetAttrString((PyObject*)co, "replace"))) goto cleanup_code_too;
        if (!(empty = PyTuple_New(0))) goto cleanup_code_too; // unfortunately __pyx_empty_tuple isn't available here
        if (!(call_result = PyObject_Call(replace, empty, kwds))) goto cleanup_code_too;
        Py_XDECREF((PyObject*)co);
        co = (PyCodeObject*)call_result;
        call_result = NULL;
        if (0) {
            cleanup_code_too:
            Py_XDECREF((PyObject*)co);
            co = NULL;
        }
        end:
        Py_XDECREF(kwds);
        Py_XDECREF(argcount);
        Py_XDECREF(posonlyargcount);
        Py_XDECREF(kwonlyargcount);
        Py_XDECREF(nlocals);
        Py_XDECREF(stacksize);
        Py_XDECREF(replace);
        Py_XDECREF(call_result);
        Py_XDECREF(empty);
        if (type) {
            PyErr_Restore(type, value, traceback);
        }
        return co;
    }
#else
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#endif
  #define __Pyx_DefaultClassType PyType_Type
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#ifndef METH_STACKLESS
  #define METH_STACKLESS 0
#endif
#if PY_VERSION_HEX <= 0x030700A3 || !defined(METH_FASTCALL)
  #ifndef METH_FASTCALL
     #define METH_FASTCALL 0x80
  #endif
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject *const *args, Py_ssize_t nargs);
  typedef PyObject *(*__Pyx_PyCFunctionFastWithKeywords) (PyObject *self, PyObject *const *args,
                                                          Py_ssize_t nargs, PyObject *kwnames);
#else
  #define __Pyx_PyCFunctionFast _PyCFunctionFast
  #define __Pyx_PyCFunctionFastWithKeywords _PyCFunctionFastWithKeywords
#endif
#if CYTHON_FAST_PYCCALL
#define __Pyx_PyFastCFunction_Check(func)\
    ((PyCFunction_Check(func) && (METH_FASTCALL == (PyCFunction_GET_FLAGS(func) & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)))))
#else
#define __Pyx_PyFastCFunction_Check(func) 0
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030400A1
  #define PyMem_RawMalloc(n)           PyMem_Malloc(n)
  #define PyMem_RawRealloc(p, n)       PyMem_Realloc(p, n)
  #define PyMem_RawFree(p)             PyMem_Free(p)
#endif
#if CYTHON_COMPILING_IN_PYSTON
  #define __Pyx_PyCode_HasFreeVars(co)  PyCode_HasFreeVars(co)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) PyFrame_SetLineNumber(frame, lineno)
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#if !CYTHON_FAST_THREAD_STATE || PY_VERSION_HEX < 0x02070000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#elif PY_VERSION_HEX >= 0x03060000
  #define __Pyx_PyThreadState_Current _PyThreadState_UncheckedGet()
#elif PY_VERSION_HEX >= 0x03000000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#else
  #define __Pyx_PyThreadState_Current _PyThreadState_Current
#endif
#if PY_VERSION_HEX < 0x030700A2 && !defined(PyThread_tss_create) && !defined(Py_tss_NEEDS_INIT)
#include "pythread.h"
#define Py_tss_NEEDS_INIT 0
typedef int Py_tss_t;
static CYTHON_INLINE int PyThread_tss_create(Py_tss_t *key) {
  *key = PyThread_create_key();
  return 0;
}
static CYTHON_INLINE Py_tss_t * PyThread_tss_alloc(void) {
  Py_tss_t *key = (Py_tss_t *)PyObject_Malloc(sizeof(Py_tss_t));
  *key = Py_tss_NEEDS_INIT;
  return key;
}
static CYTHON_INLINE void PyThread_tss_free(Py_tss_t *key) {
  PyObject_Free(key);
}
static CYTHON_INLINE int PyThread_tss_is_created(Py_tss_t *key) {
  return *key != Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE void PyThread_tss_delete(Py_tss_t *key) {
  PyThread_delete_key(*key);
  *key = Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE int PyThread_tss_set(Py_tss_t *key, void *value) {
  return PyThread_set_key_value(*key, value);
}
static CYTHON_INLINE void * PyThread_tss_get(Py_tss_t *key) {
  return PyThread_get_key_value(*key);
}
#endif
#if CYTHON_COMPILING_IN_CPYTHON || defined(_PyDict_NewPresized)
#define __Pyx_PyDict_NewPresized(n)  ((n <= 8) ? PyDict_New() : _PyDict_NewPresized(n))
#else
#define __Pyx_PyDict_NewPresized(n)  PyDict_New()
#endif
#if PY_MAJOR_VERSION >= 3 || CYTHON_FUTURE_DIVISION
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_Divide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceDivide(x,y)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1 && CYTHON_USE_UNICODE_INTERNALS
#define __Pyx_PyDict_GetItemStr(dict, name)  _PyDict_GetItem_KnownHash(dict, name, ((PyASCIIObject *) name)->hash)
#else
#define __Pyx_PyDict_GetItemStr(dict, name)  PyDict_GetItem(dict, name)
#endif
#if PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #if PY_VERSION_HEX >= 0x030C0000
    #define __Pyx_PyUnicode_READY(op)       (0)
  #else
    #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                                0 : _PyUnicode_Ready((PyObject *)(op)))
  #endif
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         PyUnicode_KIND(u)
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, ch)
  #if PY_VERSION_HEX >= 0x030C0000
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_LENGTH(u))
  #else
    #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03090000
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : ((PyCompactUnicodeObject *)(u))->wstr_length))
    #else
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
    #endif
  #endif
#else
  #define CYTHON_PEP393_ENABLED 0
  #define PyUnicode_1BYTE_KIND  1
  #define PyUnicode_2BYTE_KIND  2
  #define PyUnicode_4BYTE_KIND  4
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((sizeof(Py_UNICODE) == 2) ? 65535 : 1114111)
  #define __Pyx_PyUnicode_KIND(u)         (sizeof(Py_UNICODE))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)PyUnicode_AS_UNICODE(u))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)(k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  (((void)(k)), ((Py_UNICODE*)d)[i] = ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_SIZE(u))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyUnicode_Contains)
  #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyByteArray_Check)
  #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Format)
  #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
#endif
#define __Pyx_PyString_FormatSafe(a, b)   ((unlikely((a) == Py_None || (PyString_Check(b) && !PyString_CheckExact(b)))) ? PyNumber_Remainder(a, b) : __Pyx_PyString_Format(a, b))
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None || (PyUnicode_Check(b) && !PyUnicode_CheckExact(b)))) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyString_Format(a, b)  PyUnicode_Format(a, b)
#else
  #define __Pyx_PyString_Format(a, b)  PyString_Format(a, b)
#endif
#if PY_MAJOR_VERSION < 3 && !defined(PyObject_ASCII)
  #define PyObject_ASCII(o)            PyObject_Repr(o)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
#ifndef PyObject_Unicode
  #define PyObject_Unicode             PyObject_Str
#endif
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyBaseString_Check(obj) PyUnicode_Check(obj)
  #define __Pyx_PyBaseString_CheckExact(obj) PyUnicode_CheckExact(obj)
#else
  #define __Pyx_PyBaseString_Check(obj) (PyString_Check(obj) || PyUnicode_Check(obj))
  #define __Pyx_PyBaseString_CheckExact(obj) (PyString_CheckExact(obj) || PyUnicode_CheckExact(obj))
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        (Py_TYPE(obj) == &PySet_Type)
#endif
#if PY_VERSION_HEX >= 0x030900A4
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_SET_REFCNT(obj, refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SET_SIZE(obj, size)
#else
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_REFCNT(obj) = (refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SIZE(obj) = (size)
#endif
#if CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PySequence_SIZE(seq)  Py_SIZE(seq)
#else
  #define __Pyx_PySequence_SIZE(seq)  PySequence_Size(seq)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
  #define PyNumber_Int                 PyNumber_Long
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_MAJOR_VERSION >= 3 && CYTHON_COMPILING_IN_PYPY
  #ifndef PyUnicode_InternFromString
    #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
  #endif
#endif
#if PY_VERSION_HEX < 0x030200A4
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   __Pyx_PyIndex_AsHash_t
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   __Pyx_PyIndex_AsSsize_t
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyMethod_New(func, self, klass) ((self) ? ((void)(klass), PyMethod_New(func, self)) : __Pyx_NewRef(func))
#else
  #define __Pyx_PyMethod_New(func, self, klass) PyMethod_New(func, self, klass)
#endif
#if CYTHON_USE_ASYNC_SLOTS
  #if PY_VERSION_HEX >= 0x030500B1
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
    #define __Pyx_PyType_AsAsync(obj) (Py_TYPE(obj)->tp_as_async)
  #else
    #define __Pyx_PyType_AsAsync(obj) ((__Pyx_PyAsyncMethodsStruct*) (Py_TYPE(obj)->tp_reserved))
  #endif
#else
  #define __Pyx_PyType_AsAsync(obj) NULL
#endif
#ifndef __Pyx_PyAsyncMethodsStruct
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
    } __Pyx_PyAsyncMethodsStruct;
#endif

#if defined(_WIN32) || defined(WIN32) || defined(MS_WINDOWS)
  #if !defined(_USE_MATH_DEFINES)
    #define _USE_MATH_DEFINES
  #endif
#endif
#include <math.h>
#ifdef NAN
#define __PYX_NAN() ((float) NAN)
#else
static CYTHON_INLINE float __PYX_NAN() {
  float value;
  memset(&value, 0xFF, sizeof(value));
  return value;
}
#endif
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif

#define __PYX_MARK_ERR_POS(f_index, lineno) \
    { __pyx_filename = __pyx_f[f_index]; (void)__pyx_filename; __pyx_lineno = lineno; (void)__pyx_lineno; __pyx_clineno = __LINE__; (void)__pyx_clineno; }
#define __PYX_ERR(f_index, lineno, Ln_error) \
    { __PYX_MARK_ERR_POS(f_index, lineno) goto Ln_error; }

#ifndef __PYX_EXTERN_C
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE__glycopeptidepy___c__structure__glycan
#define __PYX_HAVE_API__glycopeptidepy___c__structure__glycan
/* Early includes */
#include <string.h>
#include <stdio.h>
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#if defined(PYREX_WITHOUT_ASSERTIONS) && !defined(CYTHON_WITHOUT_ASSERTIONS)
#define CYTHON_WITHOUT_ASSERTIONS
#endif

typedef struct {PyObject **p; const char *s; const Py_ssize_t n; const char* encoding;
                const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry;

#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_UTF8 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT (PY_MAJOR_VERSION >= 3 && __PYX_DEFAULT_STRING_ENCODING_IS_UTF8)
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
static CYTHON_INLINE int __Pyx_is_valid_index(Py_ssize_t i, Py_ssize_t limit) {
    return (size_t) i < (size_t) limit;
}
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER)
    #define __Pyx_sst_abs(value) ((Py_ssize_t)_abs64(value))
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
#define __Pyx_PyByteArray_FromString(s) PyByteArray_FromStringAndSize((const char*)s, strlen((const char*)s))
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if PY_MAJOR_VERSION < 3
    #define __Pyx_PyStr_FromString        __Pyx_PyBytes_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#else
    #define __Pyx_PyStr_FromString        __Pyx_PyUnicode_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyUnicode_FromStringAndSize
#endif
#define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyObject_AsWritableString(s)    ((char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableSString(s)    ((signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableUString(s)    ((unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsSString(s)    ((const signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((const unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyStr_FromCString(s)     __Pyx_PyStr_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
static CYTHON_INLINE size_t __Pyx_Py_UNICODE_strlen(const Py_UNICODE *u) {
    const Py_UNICODE *u_end = u;
    while (*u_end++) ;
    return (size_t)(u_end - u - 1);
}
#define __Pyx_PyUnicode_FromUnicode(u)       PyUnicode_FromUnicode(u, __Pyx_Py_UNICODE_strlen(u))
#define __Pyx_PyUnicode_FromUnicodeAndLength PyUnicode_FromUnicode
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
#define __Pyx_NewRef(obj) (Py_INCREF(obj), obj)
#define __Pyx_Owned_Py_None(b) __Pyx_NewRef(Py_None)
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b);
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x);
#define __Pyx_PySequence_Tuple(obj)\
    (likely(PyTuple_CheckExact(obj)) ? __Pyx_NewRef(obj) : PySequence_Tuple(obj))
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject*);
#if CYTHON_ASSUME_SAFE_MACROS
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#else
#define __Pyx_PyNumber_Int(x) (PyInt_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Int(x))
#endif
#define __Pyx_PyNumber_Float(x) (PyFloat_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Float(x))
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
static int __Pyx_sys_getdefaultencoding_not_ascii;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    PyObject* ascii_chars_u = NULL;
    PyObject* ascii_chars_b = NULL;
    const char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    if (strcmp(default_encoding_c, "ascii") == 0) {
        __Pyx_sys_getdefaultencoding_not_ascii = 0;
    } else {
        char ascii_chars[128];
        int c;
        for (c = 0; c < 128; c++) {
            ascii_chars[c] = c;
        }
        __Pyx_sys_getdefaultencoding_not_ascii = 1;
        ascii_chars_u = PyUnicode_DecodeASCII(ascii_chars, 128, NULL);
        if (!ascii_chars_u) goto bad;
        ascii_chars_b = PyUnicode_AsEncodedString(ascii_chars_u, default_encoding_c, NULL);
        if (!ascii_chars_b || !PyBytes_Check(ascii_chars_b) || memcmp(ascii_chars, PyBytes_AS_STRING(ascii_chars_b), 128) != 0) {
            PyErr_Format(
                PyExc_ValueError,
                "This module compiled with c_string_encoding=ascii, but default encoding '%.200s' is not a superset of ascii.",
                default_encoding_c);
            goto bad;
        }
        Py_DECREF(ascii_chars_u);
        Py_DECREF(ascii_chars_b);
    }
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    Py_XDECREF(ascii_chars_u);
    Py_XDECREF(ascii_chars_b);
    return -1;
}
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT && PY_MAJOR_VERSION >= 3
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#else
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
static char* __PYX_DEFAULT_STRING_ENCODING;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) (const char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    __PYX_DEFAULT_STRING_ENCODING = (char*) malloc(strlen(default_encoding_c) + 1);
    if (!__PYX_DEFAULT_STRING_ENCODING) goto bad;
    strcpy(__PYX_DEFAULT_STRING_ENCODING, default_encoding_c);
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    return -1;
}
#endif
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */
static CYTHON_INLINE void __Pyx_pretend_to_initialize(void* ptr) { (void)ptr; }

static PyObject *__pyx_m = NULL;
static PyObject *__pyx_d;
static PyObject *__pyx_b;
static PyObject *__pyx_cython_runtime = NULL;
static PyObject *__pyx_empty_tuple;
static PyObject *__pyx_empty_bytes;
static PyObject *__pyx_empty_unicode;
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm= __FILE__;
static const char *__pyx_filename;


static const char *__pyx_f[] = {
  "src/glycopeptidepy/_c/structure/glycan.pyx",
  "src/glycopeptidepy/_c/structure/glycan.pxd",
  "stringsource",
  "ccomposition.pxd",
  "glycan_composition.pxd",
  "src/glycopeptidepy/_c/structure/base.pxd",
  "type.pxd",
  "cenum.pxd",
  "src/glycopeptidepy/_c/structure/modification/rule.pxd",
  "src/glycopeptidepy/_c/structure/modification/modification.pxd",
};

/*--- Type declarations ---*/
struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition;
struct __pyx_obj_5glypy_2_c_9structure_18glycan_composition__CompositionBase;
struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_PeptideSequenceBase;
struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_TerminalGroup;
struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_AminoAcidResidueBase;
struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_ModificationBase;
struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_SequencePosition;
struct __pyx_obj_5glypy_5utils_5cenum_EnumValue;
struct __pyx_obj_5glypy_5utils_5cenum_IntEnumValue;
struct __pyx_obj_5glypy_5utils_5cenum_EnumMeta;
struct __pyx_obj_14glycopeptidepy_2_c_9structure_12modification_4rule_ModificationRuleBase;
struct __pyx_obj_14glycopeptidepy_2_c_9structure_12modification_4rule_NeutralLossBase;
struct __pyx_obj_14glycopeptidepy_2_c_9structure_12modification_4rule_GlycosylationBase;
struct __pyx_obj_14glycopeptidepy_2_c_9structure_12modification_12modification_ModificationInstanceBase;
struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy;
struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionWithOffsetProxy;
struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager;
struct __pyx_opt_args_5glypy_11composition_12ccomposition_12CComposition_calc_mass;
struct __pyx_opt_args_5glypy_11composition_12ccomposition_calculate_mass;

/* "glypy/composition/ccomposition.pxd":6
 *     cdef tuple _mass_args
 *     cpdef CComposition clone(self)
 *     cpdef double calc_mass(self, int average=?, charge=?, dict mass_data=?) except -1             # <<<<<<<<<<<<<<
 *     cpdef _from_formula(self, str formula, dict mass_data)
 *     cpdef _from_dict(self, comp)
 */
struct __pyx_opt_args_5glypy_11composition_12ccomposition_12CComposition_calc_mass {
  int __pyx_n;
  int average;
  PyObject *charge;
  PyObject *mass_data;
};

/* "glypy/composition/ccomposition.pxd":31
 * 
 * 
 * cpdef double calculate_mass(CComposition composition=?, str formula=?, int average=?, charge=?, mass_data=?) except -1             # <<<<<<<<<<<<<<
 */
struct __pyx_opt_args_5glypy_11composition_12ccomposition_calculate_mass {
  int __pyx_n;
  struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *composition;
  PyObject *formula;
  int average;
  PyObject *charge;
  PyObject *mass_data;
};
struct __pyx_opt_args_5glypy_2_c_9structure_18glycan_composition_16_CompositionBase__update_from_typed_map;

/* "glypy/_c/structure/glycan_composition.pxd":12
 *     cpdef object _setitem_fast(self, object key, object value)
 * 
 *     cpdef object _update_from_typed_map(self, _CompositionBase template, bint copy_nodes=*)             # <<<<<<<<<<<<<<
 *     cpdef str serialize(self)
 * 
 */
struct __pyx_opt_args_5glypy_2_c_9structure_18glycan_composition_16_CompositionBase__update_from_typed_map {
  int __pyx_n;
  int copy_nodes;
};
struct __pyx_opt_args_5glypy_5utils_5cenum_9EnumValue_add_name;

/* "glypy/utils/cenum.pxd":10
 *         public object _hash
 * 
 *     cpdef add_name(self, basestring name, bint force=*)             # <<<<<<<<<<<<<<
 *     cpdef int int_value(self) except -1
 * 
 */
struct __pyx_opt_args_5glypy_5utils_5cenum_9EnumValue_add_name {
  int __pyx_n;
  int force;
};
struct __pyx_opt_args_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_pop;

/* "glycopeptidepy/_c/structure/glycan.pxd":48
 *     cpdef values(self)
 *     cpdef items(self)
 *     cpdef pop(self, key, default=*)             # <<<<<<<<<<<<<<
 * 
 *     cdef size_t get_size(self)
 */
struct __pyx_opt_args_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_pop {
  int __pyx_n;
  PyObject *__pyx_default;
};

/* "glypy/composition/ccomposition.pxd":2
 * 
 * cdef class CComposition(dict):             # <<<<<<<<<<<<<<
 *     cdef object _mass
 *     cdef tuple _mass_args
 */
struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition {
  PyDictObject __pyx_base;
  struct __pyx_vtabstruct_5glypy_11composition_12ccomposition_CComposition *__pyx_vtab;
  PyObject *_mass;
  PyObject *_mass_args;
};


/* "glypy/_c/structure/glycan_composition.pxd":3
 * from glypy.composition.ccomposition cimport CComposition
 * 
 * cdef class _CompositionBase(dict):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public object _mass
 */
struct __pyx_obj_5glypy_2_c_9structure_18glycan_composition__CompositionBase {
  PyDictObject __pyx_base;
  struct __pyx_vtabstruct_5glypy_2_c_9structure_18glycan_composition__CompositionBase *__pyx_vtab;
  PyObject *_mass;
  PyObject *_reducing_end;
  struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *_composition_offset;
};


/* "glycopeptidepy/_c/structure/base.pxd":5
 * 
 * 
 * cdef class PeptideSequenceBase(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public list sequence
 */
struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_PeptideSequenceBase {
  PyObject_HEAD
  struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_4base_PeptideSequenceBase *__pyx_vtab;
  PyObject *sequence;
  double _mass;
  PyObject *_glycosylation_manager;
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_TerminalGroup *_n_term;
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_TerminalGroup *_c_term;
  struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *_total_composition;
  struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *_peptide_composition;
};


/* "glycopeptidepy/_c/structure/base.pxd":21
 * 
 * 
 * cdef class TerminalGroup(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public CComposition base_composition
 */
struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_TerminalGroup {
  PyObject_HEAD
  struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_4base_TerminalGroup *__pyx_vtab;
  struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *base_composition;
  double mass;
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_ModificationBase *_modification;
};


/* "glycopeptidepy/_c/structure/base.pxd":35
 * 
 * 
 * cdef class AminoAcidResidueBase(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public basestring name
 */
struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_AminoAcidResidueBase {
  PyObject_HEAD
  struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_4base_AminoAcidResidueBase *__pyx_vtab;
  PyObject *name;
  PyObject *symbol;
  double mass;
  struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *composition;
  PyObject *neutral_loss;
  Py_hash_t _hash;
};


/* "glycopeptidepy/_c/structure/base.pxd":50
 * 
 * 
 * cdef class ModificationBase(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public basestring name
 */
struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_ModificationBase {
  PyObject_HEAD
  struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_4base_ModificationBase *__pyx_vtab;
  PyObject *name;
  double mass;
  struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *composition;
};


/* "glycopeptidepy/_c/structure/base.pxd":82
 * 
 * 
 * cdef class SequencePosition(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public AminoAcidResidueBase amino_acid
 */
struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_SequencePosition {
  PyObject_HEAD
  struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_4base_SequencePosition *__pyx_vtab;
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_AminoAcidResidueBase *amino_acid;
  PyObject *modifications;
};


/* "glypy/utils/cenum.pxd":2
 * 
 * cdef class EnumValue(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public object group
 */
struct __pyx_obj_5glypy_5utils_5cenum_EnumValue {
  PyObject_HEAD
  struct __pyx_vtabstruct_5glypy_5utils_5cenum_EnumValue *__pyx_vtab;
  PyObject *group;
  PyObject *name;
  PyObject *value;
  PyObject *names;
  PyObject *_hash;
};


/* "glypy/utils/cenum.pxd":14
 * 
 * 
 * cdef class IntEnumValue(EnumValue):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public int _int_value
 */
struct __pyx_obj_5glypy_5utils_5cenum_IntEnumValue {
  struct __pyx_obj_5glypy_5utils_5cenum_EnumValue __pyx_base;
  int _int_value;
};


/* "glypy/utils/cenum.pxd":19
 * 
 * 
 * cdef class EnumMeta(type):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public dict name_map
 */
struct __pyx_obj_5glypy_5utils_5cenum_EnumMeta {
  PyHeapTypeObject __pyx_base;
  struct __pyx_vtabstruct_5glypy_5utils_5cenum_EnumMeta *__pyx_vtab;
  PyObject *name_map;
  PyObject *value_map;
};


/* "glycopeptidepy/_c/structure/modification/rule.pxd":5
 * from glypy.utils.cenum cimport EnumValue
 * 
 * cdef class ModificationRuleBase(ModificationBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public set names
 */
struct __pyx_obj_14glycopeptidepy_2_c_9structure_12modification_4rule_ModificationRuleBase {
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_ModificationBase __pyx_base;
  PyObject *names;
  PyObject *categories;
  PyObject *preferred_name;
  PyObject *title;
  PyObject *common_name;
  PyObject *unimod_name;
  Py_hash_t _hash;
};


/* "glycopeptidepy/_c/structure/modification/rule.pxd":22
 * 
 * 
 * cdef class NeutralLossBase(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public CComposition composition
 */
struct __pyx_obj_14glycopeptidepy_2_c_9structure_12modification_4rule_NeutralLossBase {
  PyObject_HEAD
  struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *composition;
  double mass;
  PyObject *label;
};


/* "glycopeptidepy/_c/structure/modification/rule.pxd":32
 * 
 * 
 * cdef class GlycosylationBase(ModificationRuleBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public bint _is_composition
 */
struct __pyx_obj_14glycopeptidepy_2_c_9structure_12modification_4rule_GlycosylationBase {
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_12modification_4rule_ModificationRuleBase __pyx_base;
  int _is_composition;
  int _is_core;
  struct __pyx_obj_5glypy_5utils_5cenum_EnumValue *_glycosylation_type;
};


/* "glycopeptidepy/_c/structure/modification/modification.pxd":5
 * 
 * 
 * cdef class ModificationInstanceBase(ModificationBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public ModificationRuleBase rule
 */
struct __pyx_obj_14glycopeptidepy_2_c_9structure_12modification_12modification_ModificationInstanceBase {
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_ModificationBase __pyx_base;
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_12modification_4rule_ModificationRuleBase *rule;
  Py_hash_t _hash;
};


/* "glycopeptidepy/_c/structure/glycan.pxd":12
 * 
 * 
 * cdef class GlycanCompositionProxy(object):             # <<<<<<<<<<<<<<
 *     '''A mapping-like object that imitates the GlycanComposition interface in
 *     a read-only fashion.
 */
struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy {
  PyObject_HEAD
  struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *__pyx_vtab;
  struct __pyx_obj_5glypy_2_c_9structure_18glycan_composition__CompositionBase *obj;
  PyObject *_serialized;
};


/* "glycopeptidepy/_c/structure/glycan.pxd":29
 * 
 * 
 * cdef class GlycanCompositionWithOffsetProxy(GlycanCompositionProxy):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public CComposition composition_offset
 */
struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionWithOffsetProxy {
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy __pyx_base;
  struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *composition_offset;
};


/* "glycopeptidepy/_c/structure/glycan.pxd":34
 * 
 * 
 * cdef class GlycosylationManager(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public dict mapping
 */
struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager {
  PyObject_HEAD
  struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_vtab;
  PyObject *mapping;
  PyObject *parent;
  PyObject *_aggregate;
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *_proxy;
  PyObject *_type_track;
  int _total_glycosylation_size;
};



/* "glypy/composition/ccomposition.pxd":2
 * 
 * cdef class CComposition(dict):             # <<<<<<<<<<<<<<
 *     cdef object _mass
 *     cdef tuple _mass_args
 */

struct __pyx_vtabstruct_5glypy_11composition_12ccomposition_CComposition {
  struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *(*clone)(struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *, int __pyx_skip_dispatch);
  double (*calc_mass)(struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *, int __pyx_skip_dispatch, struct __pyx_opt_args_5glypy_11composition_12ccomposition_12CComposition_calc_mass *__pyx_optional_args);
  PyObject *(*_from_formula)(struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *, PyObject *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*_from_dict)(struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *, PyObject *, int __pyx_skip_dispatch);
  long (*getitem)(struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *, PyObject *);
  void (*setitem)(struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *, PyObject *, long);
  struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *(*add_from)(struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *, PyObject *);
  struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *(*subtract_from)(struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *, PyObject *);
  struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *(*_create)(struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *);
};
static struct __pyx_vtabstruct_5glypy_11composition_12ccomposition_CComposition *__pyx_vtabptr_5glypy_11composition_12ccomposition_CComposition;


/* "glypy/_c/structure/glycan_composition.pxd":3
 * from glypy.composition.ccomposition cimport CComposition
 * 
 * cdef class _CompositionBase(dict):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public object _mass
 */

struct __pyx_vtabstruct_5glypy_2_c_9structure_18glycan_composition__CompositionBase {
  PyObject *(*_getitem_fast)(struct __pyx_obj_5glypy_2_c_9structure_18glycan_composition__CompositionBase *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*_setitem_fast)(struct __pyx_obj_5glypy_2_c_9structure_18glycan_composition__CompositionBase *, PyObject *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*_update_from_typed_map)(struct __pyx_obj_5glypy_2_c_9structure_18glycan_composition__CompositionBase *, struct __pyx_obj_5glypy_2_c_9structure_18glycan_composition__CompositionBase *, int __pyx_skip_dispatch, struct __pyx_opt_args_5glypy_2_c_9structure_18glycan_composition_16_CompositionBase__update_from_typed_map *__pyx_optional_args);
  PyObject *(*serialize)(struct __pyx_obj_5glypy_2_c_9structure_18glycan_composition__CompositionBase *, int __pyx_skip_dispatch);
  void (*_add_from)(struct __pyx_obj_5glypy_2_c_9structure_18glycan_composition__CompositionBase *, struct __pyx_obj_5glypy_2_c_9structure_18glycan_composition__CompositionBase *);
  void (*_subtract_from)(struct __pyx_obj_5glypy_2_c_9structure_18glycan_composition__CompositionBase *, struct __pyx_obj_5glypy_2_c_9structure_18glycan_composition__CompositionBase *);
};
static struct __pyx_vtabstruct_5glypy_2_c_9structure_18glycan_composition__CompositionBase *__pyx_vtabptr_5glypy_2_c_9structure_18glycan_composition__CompositionBase;


/* "glycopeptidepy/_c/structure/base.pxd":5
 * 
 * 
 * cdef class PeptideSequenceBase(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public list sequence
 */

struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_4base_PeptideSequenceBase {
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_SequencePosition *(*get)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_PeptideSequenceBase *, Py_ssize_t);
  PyObject *(*_invalidate)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_PeptideSequenceBase *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_4base_PeptideSequenceBase *__pyx_vtabptr_14glycopeptidepy_2_c_9structure_4base_PeptideSequenceBase;


/* "glycopeptidepy/_c/structure/base.pxd":21
 * 
 * 
 * cdef class TerminalGroup(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public CComposition base_composition
 */

struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_4base_TerminalGroup {
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_ModificationBase *(*get_modification)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_TerminalGroup *);
  struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *(*get_composition)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_TerminalGroup *);
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_TerminalGroup *(*modify)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_TerminalGroup *, struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_ModificationBase *, int __pyx_skip_dispatch);
  void (*set_modification)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_TerminalGroup *, struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_ModificationBase *);
  int (*equal_to)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_TerminalGroup *, PyObject *);
  PyObject *(*serialize)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_TerminalGroup *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_4base_TerminalGroup *__pyx_vtabptr_14glycopeptidepy_2_c_9structure_4base_TerminalGroup;


/* "glycopeptidepy/_c/structure/base.pxd":35
 * 
 * 
 * cdef class AminoAcidResidueBase(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public basestring name
 */

struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_4base_AminoAcidResidueBase {
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_AminoAcidResidueBase *(*_create)(PyObject *, PyObject *, double, struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *);
  int (*equal_to)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_AminoAcidResidueBase *, struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_AminoAcidResidueBase *);
};
static struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_4base_AminoAcidResidueBase *__pyx_vtabptr_14glycopeptidepy_2_c_9structure_4base_AminoAcidResidueBase;


/* "glycopeptidepy/_c/structure/base.pxd":50
 * 
 * 
 * cdef class ModificationBase(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public basestring name
 */

struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_4base_ModificationBase {
  int (*is_a)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_ModificationBase *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*serialize)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_ModificationBase *, int __pyx_skip_dispatch);
  int (*is_tracked_for)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_ModificationBase *, PyObject *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_4base_ModificationBase *__pyx_vtabptr_14glycopeptidepy_2_c_9structure_4base_ModificationBase;


/* "glycopeptidepy/_c/structure/base.pxd":82
 * 
 * 
 * cdef class SequencePosition(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public AminoAcidResidueBase amino_acid
 */

struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_4base_SequencePosition {
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_SequencePosition *(*_create)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_AminoAcidResidueBase *, PyObject *);
  double (*get_mass)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_SequencePosition *);
  int (*has_modification)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_SequencePosition *, PyObject *, int __pyx_skip_dispatch);
  int (*is_modified)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_SequencePosition *, int __pyx_skip_dispatch);
  PyObject *(*drop_modification)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_SequencePosition *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*add_modification)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_SequencePosition *, PyObject *, int __pyx_skip_dispatch);
  size_t (*get_modification_count)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_SequencePosition *);
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_ModificationBase *(*get_modification)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_SequencePosition *, size_t);
};
static struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_4base_SequencePosition *__pyx_vtabptr_14glycopeptidepy_2_c_9structure_4base_SequencePosition;


/* "glypy/utils/cenum.pxd":2
 * 
 * cdef class EnumValue(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public object group
 */

struct __pyx_vtabstruct_5glypy_5utils_5cenum_EnumValue {
  PyObject *(*add_name)(struct __pyx_obj_5glypy_5utils_5cenum_EnumValue *, PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_5glypy_5utils_5cenum_9EnumValue_add_name *__pyx_optional_args);
  int (*int_value)(struct __pyx_obj_5glypy_5utils_5cenum_EnumValue *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_5glypy_5utils_5cenum_EnumValue *__pyx_vtabptr_5glypy_5utils_5cenum_EnumValue;


/* "glypy/utils/cenum.pxd":14
 * 
 * 
 * cdef class IntEnumValue(EnumValue):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public int _int_value
 */

struct __pyx_vtabstruct_5glypy_5utils_5cenum_IntEnumValue {
  struct __pyx_vtabstruct_5glypy_5utils_5cenum_EnumValue __pyx_base;
};
static struct __pyx_vtabstruct_5glypy_5utils_5cenum_IntEnumValue *__pyx_vtabptr_5glypy_5utils_5cenum_IntEnumValue;


/* "glypy/utils/cenum.pxd":19
 * 
 * 
 * cdef class EnumMeta(type):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public dict name_map
 */

struct __pyx_vtabstruct_5glypy_5utils_5cenum_EnumMeta {
  struct __pyx_obj_5glypy_5utils_5cenum_EnumValue *(*name)(struct __pyx_obj_5glypy_5utils_5cenum_EnumMeta *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*translate)(struct __pyx_obj_5glypy_5utils_5cenum_EnumMeta *, PyObject *, int __pyx_skip_dispatch);
  struct __pyx_obj_5glypy_5utils_5cenum_EnumValue *(*get)(struct __pyx_obj_5glypy_5utils_5cenum_EnumMeta *, PyObject *);
  int (*put)(struct __pyx_obj_5glypy_5utils_5cenum_EnumMeta *, PyObject *, PyObject *);
};
static struct __pyx_vtabstruct_5glypy_5utils_5cenum_EnumMeta *__pyx_vtabptr_5glypy_5utils_5cenum_EnumMeta;


/* "glycopeptidepy/_c/structure/modification/rule.pxd":5
 * from glypy.utils.cenum cimport EnumValue
 * 
 * cdef class ModificationRuleBase(ModificationBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public set names
 */

struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_12modification_4rule_ModificationRuleBase {
  struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_4base_ModificationBase __pyx_base;
};
static struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_12modification_4rule_ModificationRuleBase *__pyx_vtabptr_14glycopeptidepy_2_c_9structure_12modification_4rule_ModificationRuleBase;


/* "glycopeptidepy/_c/structure/modification/rule.pxd":32
 * 
 * 
 * cdef class GlycosylationBase(ModificationRuleBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public bint _is_composition
 */

struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_12modification_4rule_GlycosylationBase {
  struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_12modification_4rule_ModificationRuleBase __pyx_base;
  int (*get_is_composition)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_12modification_4rule_GlycosylationBase *);
  void (*set_is_composition)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_12modification_4rule_GlycosylationBase *, int);
  int (*get_is_core)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_12modification_4rule_GlycosylationBase *);
  void (*set_is_core)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_12modification_4rule_GlycosylationBase *, int);
  struct __pyx_obj_5glypy_5utils_5cenum_EnumValue *(*get_glycosylation_type)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_12modification_4rule_GlycosylationBase *);
  void (*set_glycosylation_type)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_12modification_4rule_GlycosylationBase *, struct __pyx_obj_5glypy_5utils_5cenum_EnumValue *);
};
static struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_12modification_4rule_GlycosylationBase *__pyx_vtabptr_14glycopeptidepy_2_c_9structure_12modification_4rule_GlycosylationBase;


/* "glycopeptidepy/_c/structure/modification/modification.pxd":5
 * 
 * 
 * cdef class ModificationInstanceBase(ModificationBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public ModificationRuleBase rule
 */

struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_12modification_12modification_ModificationInstanceBase {
  struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_4base_ModificationBase __pyx_base;
  PyObject *(*_init_from_rule)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_12modification_12modification_ModificationInstanceBase *, struct __pyx_obj_14glycopeptidepy_2_c_9structure_12modification_4rule_ModificationRuleBase *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_12modification_12modification_ModificationInstanceBase *__pyx_vtabptr_14glycopeptidepy_2_c_9structure_12modification_12modification_ModificationInstanceBase;


/* "glycopeptidepy/_c/structure/glycan.pyx":16
 * 
 * @cython.freelist(10000)
 * cdef class GlycanCompositionProxy(object):             # <<<<<<<<<<<<<<
 *     '''A mapping-like object that imitates the GlycanComposition interface in
 *     a read-only fashion.
 */

struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy {
  PyObject *(*_getitem_fast)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*serialize)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *, int __pyx_skip_dispatch);
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *(*__pyx_fuse_0_create)(struct __pyx_obj_5glypy_2_c_9structure_18glycan_composition__CompositionBase *);
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *(*__pyx_fuse_1_create)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *);
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *(*__pyx_fuse_2_create)(PyObject *);
};
static struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *__pyx_vtabptr_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy;


/* "glycopeptidepy/_c/structure/glycan.pyx":119
 * 
 * 
 * cdef class GlycanCompositionWithOffsetProxy(GlycanCompositionProxy):             # <<<<<<<<<<<<<<
 *     def __init__(self, obj, offset=None):
 *         if offset is None:
 */

struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionWithOffsetProxy {
  struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy __pyx_base;
};
static struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionWithOffsetProxy *__pyx_vtabptr_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionWithOffsetProxy;


/* "glycopeptidepy/_c/structure/glycan.pyx":146
 * 
 * @cython.freelist(10000)
 * cdef class GlycosylationManager(object):             # <<<<<<<<<<<<<<
 *     def __init__(self, parent, aggregate=None):
 *         self._init()
 */

struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager {
  PyObject *(*update)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*clear)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *, int __pyx_skip_dispatch);
  PyObject *(*keys)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *, int __pyx_skip_dispatch);
  PyObject *(*values)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *, int __pyx_skip_dispatch);
  PyObject *(*items)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *, int __pyx_skip_dispatch);
  PyObject *(*pop)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *, PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_pop *__pyx_optional_args);
  size_t (*get_size)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *);
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *(*copy)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *, int __pyx_skip_dispatch);
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *(*clone)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *, int __pyx_skip_dispatch);
  PyObject *(*invalidate)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *, int __pyx_skip_dispatch);
  PyObject *(*get_aggregate)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *);
  int (*set_aggregate)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *, PyObject *);
  PyObject *(*_patch_aggregate)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *, int __pyx_skip_dispatch);
  PyObject *(*get_glycosylation_types)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *);
  PyObject *(*_update_track)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *);
  struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *(*total_composition)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *, int __pyx_skip_dispatch);
  double (*mass)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *, int __pyx_skip_dispatch);
  int (*is_fully_specified_topologies)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *, int __pyx_skip_dispatch);
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *(*get_glycan_composition)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *);
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *(*_make_glycan_composition_proxy)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *);
  int (*get_total_glycosylation_size)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *);
  int (*total_glycosylation_size)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *, int __pyx_skip_dispatch);
  int (*count_glycosylation_type)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *, PyObject *, int __pyx_skip_dispatch);
  void (*_init)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *);
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *(*_create)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_PeptideSequenceBase *, PyObject *);
};
static struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_vtabptr_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager;

/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, int);
    void (*DECREF)(void*, PyObject*, int);
    void (*GOTREF)(void*, PyObject*, int);
    void (*GIVEREF)(void*, PyObject*, int);
    void* (*SetupContext)(const char*, int, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__)
#endif
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_XINCREF(r)  do { if((r) != NULL) {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) != NULL) {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) != NULL) {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) != NULL) {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyObjectGetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* GetBuiltinName.proto */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* RaiseDoubleKeywords.proto */
static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

/* ParseKeywords.proto */
static int __Pyx_ParseOptionalKeywords(PyObject *kwds, PyObject **argnames[],\
    PyObject *kwds2, PyObject *values[], Py_ssize_t num_pos_args,\
    const char* function_name);

/* RaiseArgTupleInvalid.proto */
static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

/* ExtTypeTest.proto */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type);

/* KeywordStringCheck.proto */
static int __Pyx_CheckKeywordStrings(PyObject *kwdict, const char* function_name, int kw_allowed);

/* PyObjectCall.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* GetItemInt.proto */
#define __Pyx_GetItemInt(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Fast(o, (Py_ssize_t)i, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL) :\
               __Pyx_GetItemInt_Generic(o, to_py_func(i))))
#define __Pyx_GetItemInt_List(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_List_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
#define __Pyx_GetItemInt_Tuple(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Tuple_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "tuple index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j);
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i,
                                                     int is_list, int wraparound, int boundscheck);

/* ObjectGetItem.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject* key);
#else
#define __Pyx_PyObject_GetItem(obj, key)  PyObject_GetItem(obj, key)
#endif

/* PyDictVersioning.proto */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
#define __PYX_DICT_VERSION_INIT  ((PY_UINT64_T) -1)
#define __PYX_GET_DICT_VERSION(dict)  (((PyDictObject*)(dict))->ma_version_tag)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)\
    (version_var) = __PYX_GET_DICT_VERSION(dict);\
    (cache_var) = (value);
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP) {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    if (likely(__PYX_GET_DICT_VERSION(DICT) == __pyx_dict_version)) {\
        (VAR) = __pyx_dict_cached_value;\
    } else {\
        (VAR) = __pyx_dict_cached_value = (LOOKUP);\
        __pyx_dict_version = __PYX_GET_DICT_VERSION(DICT);\
    }\
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj);
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj);
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version);
#else
#define __PYX_GET_DICT_VERSION(dict)  (0)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP)  (VAR) = (LOOKUP);
#endif

/* PyCFunctionFastCall.proto */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject *__Pyx_PyCFunction_FastCall(PyObject *func, PyObject **args, Py_ssize_t nargs);
#else
#define __Pyx_PyCFunction_FastCall(func, args, nargs)  (assert(0), NULL)
#endif

/* PyFunctionFastCall.proto */
#if CYTHON_FAST_PYCALL
#define __Pyx_PyFunction_FastCall(func, args, nargs)\
    __Pyx_PyFunction_FastCallDict((func), (args), (nargs), NULL)
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs);
#else
#define __Pyx_PyFunction_FastCallDict(func, args, nargs, kwargs) _PyFunction_FastCallDict(func, args, nargs, kwargs)
#endif
#define __Pyx_BUILD_ASSERT_EXPR(cond)\
    (sizeof(char [1 - 2*!(cond)]) - 1)
#ifndef Py_MEMBER_SIZE
#define Py_MEMBER_SIZE(type, member) sizeof(((type *)0)->member)
#endif
#if CYTHON_FAST_PYCALL
  static size_t __pyx_pyframe_localsplus_offset = 0;
  #include "frameobject.h"
#if PY_VERSION_HEX >= 0x030b00a6
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
  #define __Pxy_PyFrame_Initialize_Offsets()\
    ((void)__Pyx_BUILD_ASSERT_EXPR(sizeof(PyFrameObject) == offsetof(PyFrameObject, f_localsplus) + Py_MEMBER_SIZE(PyFrameObject, f_localsplus)),\
     (void)(__pyx_pyframe_localsplus_offset = ((size_t)PyFrame_Type.tp_basicsize) - Py_MEMBER_SIZE(PyFrameObject, f_localsplus)))
  #define __Pyx_PyFrame_GetLocalsplus(frame)\
    (assert(__pyx_pyframe_localsplus_offset), (PyObject **)(((char *)(frame)) + __pyx_pyframe_localsplus_offset))
#endif // CYTHON_FAST_PYCALL
#endif

/* PyObjectCall2Args.proto */
static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2);

/* PyObjectCallMethO.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectCallOneArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

/* PySequenceContains.proto */
static CYTHON_INLINE int __Pyx_PySequence_ContainsTF(PyObject* item, PyObject* seq, int eq) {
    int result = PySequence_Contains(seq, item);
    return unlikely(result < 0) ? result : (result == (eq == Py_EQ));
}

/* PyObjectCallNoArg.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func);
#else
#define __Pyx_PyObject_CallNoArg(func) __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL)
#endif

/* IterFinish.proto */
static CYTHON_INLINE int __Pyx_IterFinish(void);

/* PyObjectGetMethod.proto */
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method);

/* PyObjectCallMethod0.proto */
static PyObject* __Pyx_PyObject_CallMethod0(PyObject* obj, PyObject* method_name);

/* RaiseNeedMoreValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index);

/* RaiseTooManyValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected);

/* UnpackItemEndCheck.proto */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected);

/* RaiseNoneIterError.proto */
static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void);

/* UnpackTupleError.proto */
static void __Pyx_UnpackTupleError(PyObject *, Py_ssize_t index);

/* UnpackTuple2.proto */
#define __Pyx_unpack_tuple2(tuple, value1, value2, is_tuple, has_known_size, decref_tuple)\
    (likely(is_tuple || PyTuple_Check(tuple)) ?\
        (likely(has_known_size || PyTuple_GET_SIZE(tuple) == 2) ?\
            __Pyx_unpack_tuple2_exact(tuple, value1, value2, decref_tuple) :\
            (__Pyx_UnpackTupleError(tuple, 2), -1)) :\
        __Pyx_unpack_tuple2_generic(tuple, value1, value2, has_known_size, decref_tuple))
static CYTHON_INLINE int __Pyx_unpack_tuple2_exact(
    PyObject* tuple, PyObject** value1, PyObject** value2, int decref_tuple);
static int __Pyx_unpack_tuple2_generic(
    PyObject* tuple, PyObject** value1, PyObject** value2, int has_known_size, int decref_tuple);

/* dict_iter.proto */
static CYTHON_INLINE PyObject* __Pyx_dict_iterator(PyObject* dict, int is_dict, PyObject* method_name,
                                                   Py_ssize_t* p_orig_length, int* p_is_dict);
static CYTHON_INLINE int __Pyx_dict_iter_next(PyObject* dict_or_iter, Py_ssize_t orig_length, Py_ssize_t* ppos,
                                              PyObject** pkey, PyObject** pvalue, PyObject** pitem, int is_dict);

/* MergeKeywords.proto */
static int __Pyx_MergeKeywords(PyObject *kwdict, PyObject *source_mapping);

/* GetTopmostException.proto */
#if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem * __Pyx_PyErr_GetTopmostException(PyThreadState *tstate);
#endif

/* PyThreadStateGet.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = __Pyx_PyThreadState_Current;
#define __Pyx_PyErr_Occurred()  __pyx_tstate->curexc_type
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#define __Pyx_PyErr_Occurred()  PyErr_Occurred()
#endif

/* SaveResetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSave(type, value, tb)  __Pyx__ExceptionSave(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#define __Pyx_ExceptionReset(type, value, tb)  __Pyx__ExceptionReset(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
#else
#define __Pyx_ExceptionSave(type, value, tb)   PyErr_GetExcInfo(type, value, tb)
#define __Pyx_ExceptionReset(type, value, tb)  PyErr_SetExcInfo(type, value, tb)
#endif

/* PyErrExceptionMatches.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_ExceptionMatches(err) __Pyx_PyErr_ExceptionMatchesInState(__pyx_tstate, err)
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err);
#else
#define __Pyx_PyErr_ExceptionMatches(err)  PyErr_ExceptionMatches(err)
#endif

/* PyErrFetchRestore.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_Clear() __Pyx_ErrRestore(NULL, NULL, NULL)
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_PyErr_SetNone(exc) (Py_INCREF(exc), __Pyx_ErrRestore((exc), NULL, NULL))
#else
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#endif
#else
#define __Pyx_PyErr_Clear() PyErr_Clear()
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestoreInState(tstate, type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchInState(tstate, type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* GetAttr.proto */
static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *, PyObject *);

/* GetAttr3.proto */
static CYTHON_INLINE PyObject *__Pyx_GetAttr3(PyObject *, PyObject *, PyObject *);

/* GetModuleGlobalName.proto */
#if CYTHON_USE_DICT_VERSIONS
#define __Pyx_GetModuleGlobalName(var, name)  do {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    (var) = (likely(__pyx_dict_version == __PYX_GET_DICT_VERSION(__pyx_d))) ?\
        (likely(__pyx_dict_cached_value) ? __Pyx_NewRef(__pyx_dict_cached_value) : __Pyx_GetBuiltinName(name)) :\
        __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
} while(0)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  do {\
    PY_UINT64_T __pyx_dict_version;\
    PyObject *__pyx_dict_cached_value;\
    (var) = __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
} while(0)
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value);
#else
#define __Pyx_GetModuleGlobalName(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name);
#endif

/* WriteUnraisableException.proto */
static void __Pyx_WriteUnraisable(const char *name, int clineno,
                                  int lineno, const char *filename,
                                  int full_traceback, int nogil);

/* DictGetItem.proto */
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key);
#define __Pyx_PyObject_Dict_GetItem(obj, name)\
    (likely(PyDict_CheckExact(obj)) ?\
     __Pyx_PyDict_GetItem(obj, name) : PyObject_GetItem(obj, name))
#else
#define __Pyx_PyDict_GetItem(d, key) PyObject_GetItem(d, key)
#define __Pyx_PyObject_Dict_GetItem(obj, name)  PyObject_GetItem(obj, name)
#endif

/* py_dict_keys.proto */
static CYTHON_INLINE PyObject* __Pyx_PyDict_Keys(PyObject* d);

/* UnpackUnboundCMethod.proto */
typedef struct {
    PyObject *type;
    PyObject **method_name;
    PyCFunction func;
    PyObject *method;
    int flag;
} __Pyx_CachedCFunction;

/* CallUnboundCMethod0.proto */
static PyObject* __Pyx__CallUnboundCMethod0(__Pyx_CachedCFunction* cfunc, PyObject* self);
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_CallUnboundCMethod0(cfunc, self)\
    (likely((cfunc)->func) ?\
        (likely((cfunc)->flag == METH_NOARGS) ?  (*((cfunc)->func))(self, NULL) :\
         (PY_VERSION_HEX >= 0x030600B1 && likely((cfunc)->flag == METH_FASTCALL) ?\
            (PY_VERSION_HEX >= 0x030700A0 ?\
                (*(__Pyx_PyCFunctionFast)(void*)(PyCFunction)(cfunc)->func)(self, &__pyx_empty_tuple, 0) :\
                (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)(cfunc)->func)(self, &__pyx_empty_tuple, 0, NULL)) :\
          (PY_VERSION_HEX >= 0x030700A0 && (cfunc)->flag == (METH_FASTCALL | METH_KEYWORDS) ?\
            (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)(cfunc)->func)(self, &__pyx_empty_tuple, 0, NULL) :\
            (likely((cfunc)->flag == (METH_VARARGS | METH_KEYWORDS)) ?  ((*(PyCFunctionWithKeywords)(void*)(PyCFunction)(cfunc)->func)(self, __pyx_empty_tuple, NULL)) :\
               ((cfunc)->flag == METH_VARARGS ?  (*((cfunc)->func))(self, __pyx_empty_tuple) :\
               __Pyx__CallUnboundCMethod0(cfunc, self)))))) :\
        __Pyx__CallUnboundCMethod0(cfunc, self))
#else
#define __Pyx_CallUnboundCMethod0(cfunc, self)  __Pyx__CallUnboundCMethod0(cfunc, self)
#endif

/* py_dict_values.proto */
static CYTHON_INLINE PyObject* __Pyx_PyDict_Values(PyObject* d);

/* py_dict_items.proto */
static CYTHON_INLINE PyObject* __Pyx_PyDict_Items(PyObject* d);

/* py_dict_pop.proto */
static CYTHON_INLINE PyObject *__Pyx_PyDict_Pop(PyObject *d, PyObject *key, PyObject *default_value);

/* CallUnboundCMethod2.proto */
static PyObject* __Pyx__CallUnboundCMethod2(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg1, PyObject* arg2);
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030600B1
static CYTHON_INLINE PyObject *__Pyx_CallUnboundCMethod2(__Pyx_CachedCFunction *cfunc, PyObject *self, PyObject *arg1, PyObject *arg2);
#else
#define __Pyx_CallUnboundCMethod2(cfunc, self, arg1, arg2)  __Pyx__CallUnboundCMethod2(cfunc, self, arg1, arg2)
#endif

/* CallUnboundCMethod1.proto */
static PyObject* __Pyx__CallUnboundCMethod1(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg);
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_CallUnboundCMethod1(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg);
#else
#define __Pyx_CallUnboundCMethod1(cfunc, self, arg)  __Pyx__CallUnboundCMethod1(cfunc, self, arg)
#endif

/* py_dict_clear.proto */
#define __Pyx_PyDict_Clear(d) (PyDict_Clear(d), 0)

/* PyObjectSetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
#define __Pyx_PyObject_DelAttrStr(o,n) __Pyx_PyObject_SetAttrStr(o, n, NULL)
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value);
#else
#define __Pyx_PyObject_DelAttrStr(o,n)   PyObject_DelAttr(o,n)
#define __Pyx_PyObject_SetAttrStr(o,n,v) PyObject_SetAttr(o,n,v)
#endif

/* Import.proto */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level);

/* ImportFrom.proto */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name);

/* RaiseException.proto */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* HasAttr.proto */
static CYTHON_INLINE int __Pyx_HasAttr(PyObject *, PyObject *);

/* IncludeStringH.proto */
#include <string.h>

/* PyObject_GenericGetAttrNoDict.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttrNoDict PyObject_GenericGetAttr
#endif

/* PyObject_GenericGetAttr.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttr PyObject_GenericGetAttr
#endif

/* SetVTable.proto */
static int __Pyx_SetVtable(PyObject *dict, void *vtable);

/* PyObjectGetAttrStrNoError.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name);

/* SetupReduce.proto */
static int __Pyx_setup_reduce(PyObject* type_obj);

/* TypeImport.proto */
#ifndef __PYX_HAVE_RT_ImportType_proto
#define __PYX_HAVE_RT_ImportType_proto
enum __Pyx_ImportType_CheckSize {
   __Pyx_ImportType_CheckSize_Error = 0,
   __Pyx_ImportType_CheckSize_Warn = 1,
   __Pyx_ImportType_CheckSize_Ignore = 2
};
static PyTypeObject *__Pyx_ImportType(PyObject* module, const char *module_name, const char *class_name, size_t size, enum __Pyx_ImportType_CheckSize check_size);
#endif

/* GetVTable.proto */
static void* __Pyx_GetVtable(PyObject *dict);

/* CLineInTraceback.proto */
#ifdef CYTHON_CLINE_IN_TRACEBACK
#define __Pyx_CLineForTraceback(tstate, c_line)  (((CYTHON_CLINE_IN_TRACEBACK)) ? c_line : 0)
#else
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line);
#endif

/* CodeObjectCache.proto */
typedef struct {
    PyCodeObject* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

/* GCCDiagnostics.proto */
#if defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 6))
#define __Pyx_HAS_GCC_DIAGNOSTIC
#endif

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value);

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value);

/* FastTypeChecks.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_TypeCheck(obj, type) __Pyx_IsSubtype(Py_TYPE(obj), (PyTypeObject *)type)
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject *type);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2);
#else
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_PyErr_GivenExceptionMatches(err, type) PyErr_GivenExceptionMatches(err, type)
#define __Pyx_PyErr_GivenExceptionMatches2(err, type1, type2) (PyErr_GivenExceptionMatches(err, type1) || PyErr_GivenExceptionMatches(err, type2))
#endif
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)

/* CheckBinaryVersion.proto */
static int __Pyx_check_binary_version(void);

/* VoidPtrImport.proto */
static int __Pyx_ImportVoidPtr(PyObject *module, const char *name, void **p, const char *sig);

/* InitStrings.proto */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t);

static struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *__pyx_fuse_0__pyx_f_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy__create(struct __pyx_obj_5glypy_2_c_9structure_18glycan_composition__CompositionBase *__pyx_v_obj); /* proto*/
static struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *__pyx_fuse_1__pyx_f_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy__create(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *__pyx_v_obj); /* proto*/
static struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *__pyx_fuse_2__pyx_f_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy__create(CYTHON_UNUSED PyObject *__pyx_v_obj); /* proto*/
static PyObject *__pyx_f_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy__getitem_fast(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *__pyx_v_self, PyObject *__pyx_v_key, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_serialize(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_f_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager__create(struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_PeptideSequenceBase *__pyx_v_parent, PyObject *__pyx_v_aggregate); /* proto*/
static void __pyx_f_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager__init(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_self); /* proto*/
static size_t __pyx_f_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_get_size(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_keys(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_values(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_items(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_pop(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_self, PyObject *__pyx_v_key, int __pyx_skip_dispatch, struct __pyx_opt_args_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_pop *__pyx_optional_args); /* proto*/
static PyObject *__pyx_f_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_clear(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_update(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_self, PyObject *__pyx_v_other, int __pyx_skip_dispatch); /* proto*/
static struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_f_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_copy(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_f_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_clone(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager__patch_aggregate(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_invalidate(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_get_aggregate(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_self); /* proto*/
static int __pyx_f_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_set_aggregate(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_self, PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_f_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_get_glycosylation_types(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager__update_track(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_self); /* proto*/
static struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *__pyx_f_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_total_composition(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *__pyx_f_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_get_glycan_composition(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_self); /* proto*/
static double __pyx_f_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_mass(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static int __pyx_f_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_get_total_glycosylation_size(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_self); /* proto*/
static int __pyx_f_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_total_glycosylation_size(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static int __pyx_f_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_is_fully_specified_topologies(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *__pyx_f_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager__make_glycan_composition_proxy(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_self); /* proto*/
static int __pyx_f_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_count_glycosylation_type(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_self, PyObject *__pyx_v_glycosylation_type, int __pyx_skip_dispatch); /* proto*/

/* Module declarations from 'glypy.composition.ccomposition' */
static PyTypeObject *__pyx_ptype_5glypy_11composition_12ccomposition_CComposition = 0;
static PyObject **__pyx_vp_5glypy_11composition_12ccomposition__atom = 0;
#define __pyx_v_5glypy_11composition_12ccomposition__atom (*__pyx_vp_5glypy_11composition_12ccomposition__atom)
static PyObject **__pyx_vp_5glypy_11composition_12ccomposition__formula = 0;
#define __pyx_v_5glypy_11composition_12ccomposition__formula (*__pyx_vp_5glypy_11composition_12ccomposition__formula)
static PyObject **__pyx_vp_5glypy_11composition_12ccomposition__isotope_string = 0;
#define __pyx_v_5glypy_11composition_12ccomposition__isotope_string (*__pyx_vp_5glypy_11composition_12ccomposition__isotope_string)
static PyObject **__pyx_vp_5glypy_11composition_12ccomposition_isotope_pattern = 0;
#define __pyx_v_5glypy_11composition_12ccomposition_isotope_pattern (*__pyx_vp_5glypy_11composition_12ccomposition_isotope_pattern)
static PyObject **__pyx_vp_5glypy_11composition_12ccomposition_formula_pattern = 0;
#define __pyx_v_5glypy_11composition_12ccomposition_formula_pattern (*__pyx_vp_5glypy_11composition_12ccomposition_formula_pattern)

/* Module declarations from 'glypy._c.structure.glycan_composition' */
static PyTypeObject *__pyx_ptype_5glypy_2_c_9structure_18glycan_composition__CompositionBase = 0;

/* Module declarations from 'cython' */

/* Module declarations from 'glycopeptidepy._c.structure.base' */
static PyTypeObject *__pyx_ptype_14glycopeptidepy_2_c_9structure_4base_PeptideSequenceBase = 0;
static PyTypeObject *__pyx_ptype_14glycopeptidepy_2_c_9structure_4base_TerminalGroup = 0;
static PyTypeObject *__pyx_ptype_14glycopeptidepy_2_c_9structure_4base_AminoAcidResidueBase = 0;
static PyTypeObject *__pyx_ptype_14glycopeptidepy_2_c_9structure_4base_ModificationBase = 0;
static PyTypeObject *__pyx_ptype_14glycopeptidepy_2_c_9structure_4base_SequencePosition = 0;

/* Module declarations from 'libc.string' */

/* Module declarations from 'libc.stdio' */

/* Module declarations from '__builtin__' */

/* Module declarations from 'cpython.type' */
static PyTypeObject *__pyx_ptype_7cpython_4type_type = 0;

/* Module declarations from 'cpython' */

/* Module declarations from 'cpython.object' */

/* Module declarations from 'cpython.list' */

/* Module declarations from 'cpython.dict' */

/* Module declarations from 'glypy.utils.cenum' */
static PyTypeObject *__pyx_ptype_5glypy_5utils_5cenum_EnumValue = 0;
static PyTypeObject *__pyx_ptype_5glypy_5utils_5cenum_IntEnumValue = 0;
static PyTypeObject *__pyx_ptype_5glypy_5utils_5cenum_EnumMeta = 0;
static PyObject **__pyx_vp_5glypy_5utils_5cenum_QMARK = 0;
#define __pyx_v_5glypy_5utils_5cenum_QMARK (*__pyx_vp_5glypy_5utils_5cenum_QMARK)

/* Module declarations from 'glycopeptidepy._c.structure.modification.rule' */
static PyTypeObject *__pyx_ptype_14glycopeptidepy_2_c_9structure_12modification_4rule_ModificationRuleBase = 0;
static PyTypeObject *__pyx_ptype_14glycopeptidepy_2_c_9structure_12modification_4rule_NeutralLossBase = 0;
static PyTypeObject *__pyx_ptype_14glycopeptidepy_2_c_9structure_12modification_4rule_GlycosylationBase = 0;

/* Module declarations from 'glycopeptidepy._c.structure.modification.modification' */
static PyTypeObject *__pyx_ptype_14glycopeptidepy_2_c_9structure_12modification_12modification_ModificationInstanceBase = 0;

/* Module declarations from 'glycopeptidepy._c.structure.glycan' */
static PyTypeObject *__pyx_ptype_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy = 0;
static PyTypeObject *__pyx_ptype_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionWithOffsetProxy = 0;
static PyTypeObject *__pyx_ptype_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager = 0;
static struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *__pyx_v_14glycopeptidepy_2_c_9structure_6glycan_WATER_OFFSET = 0;
static PyObject *__pyx_v_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxyType = 0;
static PyObject *__pyx_f_14glycopeptidepy_2_c_9structure_6glycan___pyx_unpickle_GlycanCompositionProxy__set_state(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *, PyObject *); /*proto*/
static PyObject *__pyx_f_14glycopeptidepy_2_c_9structure_6glycan___pyx_unpickle_GlycanCompositionWithOffsetProxy__set_state(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionWithOffsetProxy *, PyObject *); /*proto*/
static PyObject *__pyx_f_14glycopeptidepy_2_c_9structure_6glycan___pyx_unpickle_GlycosylationManager__set_state(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *, PyObject *); /*proto*/
#define __Pyx_MODULE_NAME "glycopeptidepy._c.structure.glycan"
extern int __pyx_module_is_main_glycopeptidepy___c__structure__glycan;
int __pyx_module_is_main_glycopeptidepy___c__structure__glycan = 0;

/* Implementation of 'glycopeptidepy._c.structure.glycan' */
static PyObject *__pyx_builtin_AttributeError;
static PyObject *__pyx_builtin_sum;
static const char __pyx_k_H[] = "H";
static const char __pyx_k_O[] = "O";
static const char __pyx_k_key[] = "key";
static const char __pyx_k_new[] = "__new__";
static const char __pyx_k_obj[] = "obj";
static const char __pyx_k_pop[] = "pop";
static const char __pyx_k_sum[] = "sum";
static const char __pyx_k_copy[] = "copy";
static const char __pyx_k_dict[] = "__dict__";
static const char __pyx_k_init[] = "__init__";
static const char __pyx_k_keys[] = "keys";
static const char __pyx_k_main[] = "__main__";
static const char __pyx_k_mass[] = "mass";
static const char __pyx_k_name[] = "__name__";
static const char __pyx_k_self[] = "self";
static const char __pyx_k_test[] = "__test__";
static const char __pyx_k_class[] = "__class__";
static const char __pyx_k_clear[] = "clear";
static const char __pyx_k_clone[] = "clone";
static const char __pyx_k_exact[] = "exact";
static const char __pyx_k_items[] = "items";
static const char __pyx_k_query[] = "query";
static const char __pyx_k_format[] = "format";
static const char __pyx_k_glycan[] = "glycan";
static const char __pyx_k_import[] = "__import__";
static const char __pyx_k_offset[] = "offset";
static const char __pyx_k_parent[] = "parent";
static const char __pyx_k_pickle[] = "pickle";
static const char __pyx_k_reduce[] = "__reduce__";
static const char __pyx_k_update[] = "update";
static const char __pyx_k_values[] = "values";
static const char __pyx_k_default[] = "default";
static const char __pyx_k_is_core[] = "is_core";
static const char __pyx_k_mapping[] = "mapping";
static const char __pyx_k_getstate[] = "__getstate__";
static const char __pyx_k_original[] = "_original";
static const char __pyx_k_pyx_type[] = "__pyx_type";
static const char __pyx_k_setstate[] = "__setstate__";
static const char __pyx_k_aggregate[] = "aggregate";
static const char __pyx_k_pyx_state[] = "__pyx_state";
static const char __pyx_k_reduce_ex[] = "__reduce_ex__";
static const char __pyx_k_serialize[] = "serialize";
static const char __pyx_k_invalidate[] = "invalidate";
static const char __pyx_k_pyx_result[] = "__pyx_result";
static const char __pyx_k_pyx_vtable[] = "__pyx_vtable__";
static const char __pyx_k_serialized[] = "_serialized";
static const char __pyx_k_PickleError[] = "PickleError";
static const char __pyx_k_from_glycan[] = "from_glycan";
static const char __pyx_k_getitem_fast[] = "_getitem_fast";
static const char __pyx_k_pyx_checksum[] = "__pyx_checksum";
static const char __pyx_k_stringsource[] = "stringsource";
static const char __pyx_k_reduce_cython[] = "__reduce_cython__";
static const char __pyx_k_AttributeError[] = "AttributeError";
static const char __pyx_k_is_composition[] = "is_composition";
static const char __pyx_k_patch_aggregate[] = "_patch_aggregate";
static const char __pyx_k_pyx_PickleError[] = "__pyx_PickleError";
static const char __pyx_k_setstate_cython[] = "__setstate_cython__";
static const char __pyx_k_total_composition[] = "total_composition";
static const char __pyx_k_cline_in_traceback[] = "cline_in_traceback";
static const char __pyx_k_composition_offset[] = "composition_offset";
static const char __pyx_k_glycan_composition[] = "glycan_composition";
static const char __pyx_k_GlycosylationManager[] = "GlycosylationManager";
static const char __pyx_k_GlycanCompositionProxy[] = "GlycanCompositionProxy";
static const char __pyx_k_count_glycosylation_type[] = "count_glycosylation_type";
static const char __pyx_k_total_glycosylation_size[] = "total_glycosylation_size";
static const char __pyx_k_HashableGlycanComposition[] = "HashableGlycanComposition";
static const char __pyx_k_is_fully_specified_topologies[] = "is_fully_specified_topologies";
static const char __pyx_k_pyx_unpickle_GlycanComposition[] = "__pyx_unpickle_GlycanCompositionProxy";
static const char __pyx_k_pyx_unpickle_GlycosylationMana[] = "__pyx_unpickle_GlycosylationManager";
static const char __pyx_k_self___class_____name___self_ma[] = "{self.__class__.__name__}({self.mapping})";
static const char __pyx_k_GlycanCompositionWithOffsetProxy[] = "GlycanCompositionWithOffsetProxy";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0[] = "Incompatible checksums (0x%x vs (0x4126d5b, 0x40511ef, 0x04f04be) = (_serialized, obj))";
static const char __pyx_k_glycopeptidepy__c_structure_glyc[] = "glycopeptidepy._c.structure.glycan";
static const char __pyx_k_glypy_structure_glycan_compositi[] = "glypy.structure.glycan_composition";
static const char __pyx_k_pyx_unpickle_GlycanComposition_2[] = "__pyx_unpickle_GlycanCompositionWithOffsetProxy";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0_2[] = "Incompatible checksums (0x%x vs (0x32b2f99, 0x6f0dd41, 0x3381ea0) = (_serialized, composition_offset, obj))";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0_3[] = "Incompatible checksums (0x%x vs (0xeaeda1b, 0x38357ea, 0x45d3e33) = (_aggregate, _proxy, _total_glycosylation_size, _type_track, mapping, parent))";
static PyObject *__pyx_n_s_AttributeError;
static PyObject *__pyx_n_s_GlycanCompositionProxy;
static PyObject *__pyx_n_s_GlycanCompositionWithOffsetProxy;
static PyObject *__pyx_n_s_GlycosylationManager;
static PyObject *__pyx_n_s_H;
static PyObject *__pyx_n_s_HashableGlycanComposition;
static PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0;
static PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_2;
static PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_3;
static PyObject *__pyx_n_s_O;
static PyObject *__pyx_n_s_PickleError;
static PyObject *__pyx_n_s_aggregate;
static PyObject *__pyx_n_s_class;
static PyObject *__pyx_n_s_clear;
static PyObject *__pyx_n_s_cline_in_traceback;
static PyObject *__pyx_n_s_clone;
static PyObject *__pyx_n_s_composition_offset;
static PyObject *__pyx_n_s_copy;
static PyObject *__pyx_n_s_count_glycosylation_type;
static PyObject *__pyx_n_s_default;
static PyObject *__pyx_n_s_dict;
static PyObject *__pyx_n_s_exact;
static PyObject *__pyx_n_s_format;
static PyObject *__pyx_n_s_from_glycan;
static PyObject *__pyx_n_s_getitem_fast;
static PyObject *__pyx_n_s_getstate;
static PyObject *__pyx_n_s_glycan;
static PyObject *__pyx_n_s_glycan_composition;
static PyObject *__pyx_n_s_glycopeptidepy__c_structure_glyc;
static PyObject *__pyx_n_s_glypy_structure_glycan_compositi;
static PyObject *__pyx_n_s_import;
static PyObject *__pyx_n_s_init;
static PyObject *__pyx_n_s_invalidate;
static PyObject *__pyx_n_s_is_composition;
static PyObject *__pyx_n_s_is_core;
static PyObject *__pyx_n_s_is_fully_specified_topologies;
static PyObject *__pyx_n_s_items;
static PyObject *__pyx_n_s_key;
static PyObject *__pyx_n_s_keys;
static PyObject *__pyx_n_s_main;
static PyObject *__pyx_n_s_mapping;
static PyObject *__pyx_n_s_mass;
static PyObject *__pyx_n_s_name;
static PyObject *__pyx_n_s_new;
static PyObject *__pyx_n_s_obj;
static PyObject *__pyx_n_s_offset;
static PyObject *__pyx_n_s_original;
static PyObject *__pyx_n_s_parent;
static PyObject *__pyx_n_s_patch_aggregate;
static PyObject *__pyx_n_s_pickle;
static PyObject *__pyx_n_s_pop;
static PyObject *__pyx_n_s_pyx_PickleError;
static PyObject *__pyx_n_s_pyx_checksum;
static PyObject *__pyx_n_s_pyx_result;
static PyObject *__pyx_n_s_pyx_state;
static PyObject *__pyx_n_s_pyx_type;
static PyObject *__pyx_n_s_pyx_unpickle_GlycanComposition;
static PyObject *__pyx_n_s_pyx_unpickle_GlycanComposition_2;
static PyObject *__pyx_n_s_pyx_unpickle_GlycosylationMana;
static PyObject *__pyx_n_s_pyx_vtable;
static PyObject *__pyx_n_s_query;
static PyObject *__pyx_n_s_reduce;
static PyObject *__pyx_n_s_reduce_cython;
static PyObject *__pyx_n_s_reduce_ex;
static PyObject *__pyx_n_s_self;
static PyObject *__pyx_kp_s_self___class_____name___self_ma;
static PyObject *__pyx_n_s_serialize;
static PyObject *__pyx_n_s_serialized;
static PyObject *__pyx_n_s_setstate;
static PyObject *__pyx_n_s_setstate_cython;
static PyObject *__pyx_kp_s_stringsource;
static PyObject *__pyx_n_s_sum;
static PyObject *__pyx_n_s_test;
static PyObject *__pyx_n_s_total_composition;
static PyObject *__pyx_n_s_total_glycosylation_size;
static PyObject *__pyx_n_s_update;
static PyObject *__pyx_n_s_values;
static int __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy___init__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *__pyx_v_self, PyObject *__pyx_v_glycan_composition); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_2mass(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *__pyx_v_self, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_4total_composition(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *__pyx_v_self, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_6__iter__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_8__getitem__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *__pyx_v_self, PyObject *__pyx_v_key); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_10_getitem_fast(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *__pyx_v_self, PyObject *__pyx_v_key); /* proto */
static int __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_12__contains__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *__pyx_v_self, PyObject *__pyx_v_key); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_14keys(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_16values(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_18items(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_20clone(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *__pyx_v_self, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_22query(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *__pyx_v_self, PyObject *__pyx_v_query, PyObject *__pyx_v_exact, PyObject *__pyx_v_kwargs); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_24serialize(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_26__repr__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_28__str__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *__pyx_v_self); /* proto */
static Py_hash_t __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_30__hash__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_32__eq__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *__pyx_v_self, PyObject *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_34__ne__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *__pyx_v_self, PyObject *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_36__add__(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_38__sub__(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_40__mul__(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /* proto */
static Py_ssize_t __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_42__len__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_3obj___get__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *__pyx_v_self); /* proto */
static int __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_3obj_2__set__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_3obj_4__del__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_11_serialized___get__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *__pyx_v_self); /* proto */
static int __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_11_serialized_2__set__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_11_serialized_4__del__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_44__reduce_cython__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_46__setstate_cython__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_32GlycanCompositionWithOffsetProxy___init__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionWithOffsetProxy *__pyx_v_self, PyObject *__pyx_v_obj, PyObject *__pyx_v_offset); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_32GlycanCompositionWithOffsetProxy_2clone(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionWithOffsetProxy *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_args, CYTHON_UNUSED PyObject *__pyx_v_kwargs); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_32GlycanCompositionWithOffsetProxy_4mass(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionWithOffsetProxy *__pyx_v_self, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_32GlycanCompositionWithOffsetProxy_6total_composition(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionWithOffsetProxy *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_32GlycanCompositionWithOffsetProxy_18composition_offset___get__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionWithOffsetProxy *__pyx_v_self); /* proto */
static int __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_32GlycanCompositionWithOffsetProxy_18composition_offset_2__set__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionWithOffsetProxy *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_32GlycanCompositionWithOffsetProxy_18composition_offset_4__del__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionWithOffsetProxy *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_32GlycanCompositionWithOffsetProxy_8__reduce_cython__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionWithOffsetProxy *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_32GlycanCompositionWithOffsetProxy_10__setstate_cython__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionWithOffsetProxy *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager___init__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_self, PyObject *__pyx_v_parent, PyObject *__pyx_v_aggregate); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_2__getitem__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_self, PyObject *__pyx_v_key); /* proto */
static int __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_4__setitem__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_self, PyObject *__pyx_v_key, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_6__delitem__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_self, PyObject *__pyx_v_key); /* proto */
static Py_ssize_t __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_8__len__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_10keys(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_12values(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_14items(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_16pop(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_self, PyObject *__pyx_v_key, PyObject *__pyx_v_default); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_18clear(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_20update(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_self, PyObject *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_22copy(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_24clone(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_26_patch_aggregate(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_28invalidate(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_9aggregate___get__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_self); /* proto */
static int __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_9aggregate_2__set__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_30total_composition(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_18glycan_composition___get__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_32mass(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_34total_glycosylation_size(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_36is_fully_specified_topologies(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_38count_glycosylation_type(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_self, PyObject *__pyx_v_glycosylation_type); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_40__repr__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_7mapping___get__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_self); /* proto */
static int __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_7mapping_2__set__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_7mapping_4__del__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_6parent___get__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_self); /* proto */
static int __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_6parent_2__set__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_6parent_4__del__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_10_aggregate___get__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_self); /* proto */
static int __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_10_aggregate_2__set__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_10_aggregate_4__del__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_6_proxy___get__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_self); /* proto */
static int __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_6_proxy_2__set__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_6_proxy_4__del__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_11_type_track___get__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_self); /* proto */
static int __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_11_type_track_2__set__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_11_type_track_4__del__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_25_total_glycosylation_size___get__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_self); /* proto */
static int __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_25_total_glycosylation_size_2__set__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_42__reduce_cython__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_44__setstate_cython__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_6glycan___pyx_unpickle_GlycanCompositionProxy(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_2__pyx_unpickle_GlycanCompositionWithOffsetProxy(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_4__pyx_unpickle_GlycosylationManager(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_tp_new_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionWithOffsetProxy(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static __Pyx_CachedCFunction __pyx_umethod_PyDict_Type_items = {0, &__pyx_n_s_items, 0, 0, 0};
static __Pyx_CachedCFunction __pyx_umethod_PyDict_Type_keys = {0, &__pyx_n_s_keys, 0, 0, 0};
static __Pyx_CachedCFunction __pyx_umethod_PyDict_Type_pop = {0, &__pyx_n_s_pop, 0, 0, 0};
static __Pyx_CachedCFunction __pyx_umethod_PyDict_Type_values = {0, &__pyx_n_s_values, 0, 0, 0};
static PyObject *__pyx_int_1;
static PyObject *__pyx_int_2;
static PyObject *__pyx_int_5178558;
static PyObject *__pyx_int_53161881;
static PyObject *__pyx_int_54009504;
static PyObject *__pyx_int_58939370;
static PyObject *__pyx_int_67441135;
static PyObject *__pyx_int_68316507;
static PyObject *__pyx_int_73219635;
static PyObject *__pyx_int_116448577;
static PyObject *__pyx_int_246340123;
static PyObject *__pyx_tuple_;
static PyObject *__pyx_tuple__2;
static PyObject *__pyx_tuple__3;
static PyObject *__pyx_tuple__4;
static PyObject *__pyx_tuple__6;
static PyObject *__pyx_tuple__8;
static PyObject *__pyx_codeobj__5;
static PyObject *__pyx_codeobj__7;
static PyObject *__pyx_codeobj__9;
/* Late includes */

/* "glycopeptidepy/_c/structure/glycan.pyx":22
 * 
 *     @staticmethod
 *     cdef GlycanCompositionProxy _create(glycan_composition_type obj):             # <<<<<<<<<<<<<<
 *         cdef:
 *             GlycanCompositionProxy self
 */

static struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *__pyx_fuse_0__pyx_f_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy__create(struct __pyx_obj_5glypy_2_c_9structure_18glycan_composition__CompositionBase *__pyx_v_obj) {
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *__pyx_v_self = 0;
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_fuse_0_create", 0);

  /* "glycopeptidepy/_c/structure/glycan.pyx":25
 *         cdef:
 *             GlycanCompositionProxy self
 *         self = <GlycanCompositionProxy>GlycanCompositionProxy.__new__(GlycanCompositionProxy)             # <<<<<<<<<<<<<<
 *         if glycan_composition_type is _CompositionBase:
 *             self.obj = obj
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy(((PyTypeObject *)__pyx_ptype_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 25, __pyx_L1_error)
  __Pyx_GOTREF(((PyObject *)__pyx_t_1));
  __pyx_t_2 = ((PyObject *)__pyx_t_1);
  __Pyx_INCREF(__pyx_t_2);
  __Pyx_DECREF(((PyObject *)__pyx_t_1)); __pyx_t_1 = 0;
  __pyx_v_self = ((struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "glycopeptidepy/_c/structure/glycan.pyx":27
 *         self = <GlycanCompositionProxy>GlycanCompositionProxy.__new__(GlycanCompositionProxy)
 *         if glycan_composition_type is _CompositionBase:
 *             self.obj = obj             # <<<<<<<<<<<<<<
 *             self._serialized = None
 *         elif glycan_composition_type is GlycanCompositionProxy:
 */
  __Pyx_INCREF(((PyObject *)__pyx_v_obj));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_obj));
  __Pyx_GOTREF(__pyx_v_self->obj);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->obj));
  __pyx_v_self->obj = __pyx_v_obj;

  /* "glycopeptidepy/_c/structure/glycan.pyx":28
 *         if glycan_composition_type is _CompositionBase:
 *             self.obj = obj
 *             self._serialized = None             # <<<<<<<<<<<<<<
 *         elif glycan_composition_type is GlycanCompositionProxy:
 *             self.obj = obj.obj
 */
  if (!(likely(PyString_CheckExact(Py_None))||((Py_None) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "str", Py_TYPE(Py_None)->tp_name), 0))) __PYX_ERR(0, 28, __pyx_L1_error)
  __pyx_t_2 = Py_None;
  __Pyx_INCREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_GOTREF(__pyx_v_self->_serialized);
  __Pyx_DECREF(__pyx_v_self->_serialized);
  __pyx_v_self->_serialized = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "glycopeptidepy/_c/structure/glycan.pyx":32
 *             self.obj = obj.obj
 *             self._serialized = obj._serialized
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, glycan_composition):
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __Pyx_INCREF(((PyObject *)__pyx_v_self));
  __pyx_r = __pyx_v_self;
  goto __pyx_L0;

  /* "glycopeptidepy/_c/structure/glycan.pyx":22
 * 
 *     @staticmethod
 *     cdef GlycanCompositionProxy _create(glycan_composition_type obj):             # <<<<<<<<<<<<<<
 *         cdef:
 *             GlycanCompositionProxy self
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.glycan.GlycanCompositionProxy._create", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_self);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *__pyx_fuse_1__pyx_f_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy__create(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *__pyx_v_obj) {
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *__pyx_v_self = 0;
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_fuse_1_create", 0);

  /* "glycopeptidepy/_c/structure/glycan.pyx":25
 *         cdef:
 *             GlycanCompositionProxy self
 *         self = <GlycanCompositionProxy>GlycanCompositionProxy.__new__(GlycanCompositionProxy)             # <<<<<<<<<<<<<<
 *         if glycan_composition_type is _CompositionBase:
 *             self.obj = obj
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy(((PyTypeObject *)__pyx_ptype_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 25, __pyx_L1_error)
  __Pyx_GOTREF(((PyObject *)__pyx_t_1));
  __pyx_t_2 = ((PyObject *)__pyx_t_1);
  __Pyx_INCREF(__pyx_t_2);
  __Pyx_DECREF(((PyObject *)__pyx_t_1)); __pyx_t_1 = 0;
  __pyx_v_self = ((struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "glycopeptidepy/_c/structure/glycan.pyx":30
 *             self._serialized = None
 *         elif glycan_composition_type is GlycanCompositionProxy:
 *             self.obj = obj.obj             # <<<<<<<<<<<<<<
 *             self._serialized = obj._serialized
 *         return self
 */
  __pyx_t_2 = ((PyObject *)__pyx_v_obj->obj);
  __Pyx_INCREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_GOTREF(__pyx_v_self->obj);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->obj));
  __pyx_v_self->obj = ((struct __pyx_obj_5glypy_2_c_9structure_18glycan_composition__CompositionBase *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "glycopeptidepy/_c/structure/glycan.pyx":31
 *         elif glycan_composition_type is GlycanCompositionProxy:
 *             self.obj = obj.obj
 *             self._serialized = obj._serialized             # <<<<<<<<<<<<<<
 *         return self
 * 
 */
  __pyx_t_2 = __pyx_v_obj->_serialized;
  __Pyx_INCREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_GOTREF(__pyx_v_self->_serialized);
  __Pyx_DECREF(__pyx_v_self->_serialized);
  __pyx_v_self->_serialized = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "glycopeptidepy/_c/structure/glycan.pyx":32
 *             self.obj = obj.obj
 *             self._serialized = obj._serialized
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, glycan_composition):
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __Pyx_INCREF(((PyObject *)__pyx_v_self));
  __pyx_r = __pyx_v_self;
  goto __pyx_L0;

  /* "glycopeptidepy/_c/structure/glycan.pyx":22
 * 
 *     @staticmethod
 *     cdef GlycanCompositionProxy _create(glycan_composition_type obj):             # <<<<<<<<<<<<<<
 *         cdef:
 *             GlycanCompositionProxy self
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.glycan.GlycanCompositionProxy._create", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_self);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *__pyx_fuse_2__pyx_f_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy__create(CYTHON_UNUSED PyObject *__pyx_v_obj) {
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *__pyx_v_self = 0;
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_fuse_2_create", 0);

  /* "glycopeptidepy/_c/structure/glycan.pyx":25
 *         cdef:
 *             GlycanCompositionProxy self
 *         self = <GlycanCompositionProxy>GlycanCompositionProxy.__new__(GlycanCompositionProxy)             # <<<<<<<<<<<<<<
 *         if glycan_composition_type is _CompositionBase:
 *             self.obj = obj
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy(((PyTypeObject *)__pyx_ptype_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 25, __pyx_L1_error)
  __Pyx_GOTREF(((PyObject *)__pyx_t_1));
  __pyx_t_2 = ((PyObject *)__pyx_t_1);
  __Pyx_INCREF(__pyx_t_2);
  __Pyx_DECREF(((PyObject *)__pyx_t_1)); __pyx_t_1 = 0;
  __pyx_v_self = ((struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "glycopeptidepy/_c/structure/glycan.pyx":32
 *             self.obj = obj.obj
 *             self._serialized = obj._serialized
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, glycan_composition):
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __Pyx_INCREF(((PyObject *)__pyx_v_self));
  __pyx_r = __pyx_v_self;
  goto __pyx_L0;

  /* "glycopeptidepy/_c/structure/glycan.pyx":22
 * 
 *     @staticmethod
 *     cdef GlycanCompositionProxy _create(glycan_composition_type obj):             # <<<<<<<<<<<<<<
 *         cdef:
 *             GlycanCompositionProxy self
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.glycan.GlycanCompositionProxy._create", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_self);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/glycan.pyx":34
 *         return self
 * 
 *     def __init__(self, glycan_composition):             # <<<<<<<<<<<<<<
 *         if isinstance(glycan_composition, GlycanCompositionProxy):
 *             self._serialized = glycan_composition._serialized
 */

/* Python wrapper */
static int __pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_glycan_composition = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_glycan_composition,0};
    PyObject* values[1] = {0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_glycan_composition)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 34, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 1) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
    }
    __pyx_v_glycan_composition = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 34, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("glycopeptidepy._c.structure.glycan.GlycanCompositionProxy.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy___init__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *)__pyx_v_self), __pyx_v_glycan_composition);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy___init__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *__pyx_v_self, PyObject *__pyx_v_glycan_composition) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "glycopeptidepy/_c/structure/glycan.pyx":35
 * 
 *     def __init__(self, glycan_composition):
 *         if isinstance(glycan_composition, GlycanCompositionProxy):             # <<<<<<<<<<<<<<
 *             self._serialized = glycan_composition._serialized
 *             self.obj = glycan_composition.obj
 */
  __pyx_t_1 = __Pyx_TypeCheck(__pyx_v_glycan_composition, __pyx_ptype_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy); 
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "glycopeptidepy/_c/structure/glycan.pyx":36
 *     def __init__(self, glycan_composition):
 *         if isinstance(glycan_composition, GlycanCompositionProxy):
 *             self._serialized = glycan_composition._serialized             # <<<<<<<<<<<<<<
 *             self.obj = glycan_composition.obj
 *         else:
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_glycan_composition, __pyx_n_s_serialized); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 36, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (!(likely(PyString_CheckExact(__pyx_t_3))||((__pyx_t_3) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "str", Py_TYPE(__pyx_t_3)->tp_name), 0))) __PYX_ERR(0, 36, __pyx_L1_error)
    __Pyx_GIVEREF(__pyx_t_3);
    __Pyx_GOTREF(__pyx_v_self->_serialized);
    __Pyx_DECREF(__pyx_v_self->_serialized);
    __pyx_v_self->_serialized = ((PyObject*)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "glycopeptidepy/_c/structure/glycan.pyx":37
 *         if isinstance(glycan_composition, GlycanCompositionProxy):
 *             self._serialized = glycan_composition._serialized
 *             self.obj = glycan_composition.obj             # <<<<<<<<<<<<<<
 *         else:
 *             self.obj = glycan_composition
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_glycan_composition, __pyx_n_s_obj); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 37, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_5glypy_2_c_9structure_18glycan_composition__CompositionBase))))) __PYX_ERR(0, 37, __pyx_L1_error)
    __Pyx_GIVEREF(__pyx_t_3);
    __Pyx_GOTREF(__pyx_v_self->obj);
    __Pyx_DECREF(((PyObject *)__pyx_v_self->obj));
    __pyx_v_self->obj = ((struct __pyx_obj_5glypy_2_c_9structure_18glycan_composition__CompositionBase *)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "glycopeptidepy/_c/structure/glycan.pyx":35
 * 
 *     def __init__(self, glycan_composition):
 *         if isinstance(glycan_composition, GlycanCompositionProxy):             # <<<<<<<<<<<<<<
 *             self._serialized = glycan_composition._serialized
 *             self.obj = glycan_composition.obj
 */
    goto __pyx_L3;
  }

  /* "glycopeptidepy/_c/structure/glycan.pyx":39
 *             self.obj = glycan_composition.obj
 *         else:
 *             self.obj = glycan_composition             # <<<<<<<<<<<<<<
 *             self._serialized = None
 * 
 */
  /*else*/ {
    if (!(likely(((__pyx_v_glycan_composition) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_glycan_composition, __pyx_ptype_5glypy_2_c_9structure_18glycan_composition__CompositionBase))))) __PYX_ERR(0, 39, __pyx_L1_error)
    __pyx_t_3 = __pyx_v_glycan_composition;
    __Pyx_INCREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_3);
    __Pyx_GOTREF(__pyx_v_self->obj);
    __Pyx_DECREF(((PyObject *)__pyx_v_self->obj));
    __pyx_v_self->obj = ((struct __pyx_obj_5glypy_2_c_9structure_18glycan_composition__CompositionBase *)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "glycopeptidepy/_c/structure/glycan.pyx":40
 *         else:
 *             self.obj = glycan_composition
 *             self._serialized = None             # <<<<<<<<<<<<<<
 * 
 *     def mass(self, *args, **kwargs):
 */
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    __Pyx_GOTREF(__pyx_v_self->_serialized);
    __Pyx_DECREF(__pyx_v_self->_serialized);
    __pyx_v_self->_serialized = ((PyObject*)Py_None);
  }
  __pyx_L3:;

  /* "glycopeptidepy/_c/structure/glycan.pyx":34
 *         return self
 * 
 *     def __init__(self, glycan_composition):             # <<<<<<<<<<<<<<
 *         if isinstance(glycan_composition, GlycanCompositionProxy):
 *             self._serialized = glycan_composition._serialized
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.glycan.GlycanCompositionProxy.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/glycan.pyx":42
 *             self._serialized = None
 * 
 *     def mass(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         return self.obj.mass(*args, **kwargs)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_3mass(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_2mass[] = "GlycanCompositionProxy.mass(self, *args, **kwargs)";
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_3mass(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_args = 0;
  PyObject *__pyx_v_kwargs = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("mass (wrapper)", 0);
  if (unlikely(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "mass", 1))) return NULL;
  __pyx_v_kwargs = (__pyx_kwds) ? PyDict_Copy(__pyx_kwds) : PyDict_New(); if (unlikely(!__pyx_v_kwargs)) return NULL;
  __Pyx_GOTREF(__pyx_v_kwargs);
  __Pyx_INCREF(__pyx_args);
  __pyx_v_args = __pyx_args;
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_2mass(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *)__pyx_v_self), __pyx_v_args, __pyx_v_kwargs);

  /* function exit code */
  __Pyx_XDECREF(__pyx_v_args);
  __Pyx_XDECREF(__pyx_v_kwargs);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_2mass(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *__pyx_v_self, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("mass", 0);

  /* "glycopeptidepy/_c/structure/glycan.pyx":43
 * 
 *     def mass(self, *args, **kwargs):
 *         return self.obj.mass(*args, **kwargs)             # <<<<<<<<<<<<<<
 * 
 *     def total_composition(self, *args, **kwargs):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->obj), __pyx_n_s_mass); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 43, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyDict_Copy(__pyx_v_kwargs); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 43, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_v_args, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 43, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "glycopeptidepy/_c/structure/glycan.pyx":42
 *             self._serialized = None
 * 
 *     def mass(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         return self.obj.mass(*args, **kwargs)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.glycan.GlycanCompositionProxy.mass", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/glycan.pyx":45
 *         return self.obj.mass(*args, **kwargs)
 * 
 *     def total_composition(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         return self.obj.total_composition(*args, **kwargs)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_5total_composition(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_4total_composition[] = "GlycanCompositionProxy.total_composition(self, *args, **kwargs)";
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_5total_composition(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_args = 0;
  PyObject *__pyx_v_kwargs = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("total_composition (wrapper)", 0);
  if (unlikely(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "total_composition", 1))) return NULL;
  __pyx_v_kwargs = (__pyx_kwds) ? PyDict_Copy(__pyx_kwds) : PyDict_New(); if (unlikely(!__pyx_v_kwargs)) return NULL;
  __Pyx_GOTREF(__pyx_v_kwargs);
  __Pyx_INCREF(__pyx_args);
  __pyx_v_args = __pyx_args;
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_4total_composition(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *)__pyx_v_self), __pyx_v_args, __pyx_v_kwargs);

  /* function exit code */
  __Pyx_XDECREF(__pyx_v_args);
  __Pyx_XDECREF(__pyx_v_kwargs);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_4total_composition(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *__pyx_v_self, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("total_composition", 0);

  /* "glycopeptidepy/_c/structure/glycan.pyx":46
 * 
 *     def total_composition(self, *args, **kwargs):
 *         return self.obj.total_composition(*args, **kwargs)             # <<<<<<<<<<<<<<
 * 
 *     def __iter__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->obj), __pyx_n_s_total_composition); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 46, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyDict_Copy(__pyx_v_kwargs); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 46, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_v_args, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 46, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "glycopeptidepy/_c/structure/glycan.pyx":45
 *         return self.obj.mass(*args, **kwargs)
 * 
 *     def total_composition(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         return self.obj.total_composition(*args, **kwargs)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.glycan.GlycanCompositionProxy.total_composition", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/glycan.pyx":48
 *         return self.obj.total_composition(*args, **kwargs)
 * 
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         return iter(self.obj)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_7__iter__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_7__iter__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__iter__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_6__iter__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_6__iter__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__iter__", 0);

  /* "glycopeptidepy/_c/structure/glycan.pyx":49
 * 
 *     def __iter__(self):
 *         return iter(self.obj)             # <<<<<<<<<<<<<<
 * 
 *     def __getitem__(self, key):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_v_self->obj);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 49, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "glycopeptidepy/_c/structure/glycan.pyx":48
 *         return self.obj.total_composition(*args, **kwargs)
 * 
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         return iter(self.obj)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.glycan.GlycanCompositionProxy.__iter__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/glycan.pyx":51
 *         return iter(self.obj)
 * 
 *     def __getitem__(self, key):             # <<<<<<<<<<<<<<
 *         return self.obj[key]
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_9__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_key); /*proto*/
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_9__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_key) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getitem__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_8__getitem__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *)__pyx_v_self), ((PyObject *)__pyx_v_key));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_8__getitem__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *__pyx_v_self, PyObject *__pyx_v_key) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__getitem__", 0);

  /* "glycopeptidepy/_c/structure/glycan.pyx":52
 * 
 *     def __getitem__(self, key):
 *         return self.obj[key]             # <<<<<<<<<<<<<<
 * 
 *     cpdef object _getitem_fast(self, key):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetItem(((PyObject *)__pyx_v_self->obj), __pyx_v_key); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 52, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "glycopeptidepy/_c/structure/glycan.pyx":51
 *         return iter(self.obj)
 * 
 *     def __getitem__(self, key):             # <<<<<<<<<<<<<<
 *         return self.obj[key]
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.glycan.GlycanCompositionProxy.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/glycan.pyx":54
 *         return self.obj[key]
 * 
 *     cpdef object _getitem_fast(self, key):             # <<<<<<<<<<<<<<
 *         return self.obj._getitem_fast(key)
 * 
 */

static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_11_getitem_fast(PyObject *__pyx_v_self, PyObject *__pyx_v_key); /*proto*/
static PyObject *__pyx_f_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy__getitem_fast(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *__pyx_v_self, PyObject *__pyx_v_key, int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_getitem_fast", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_getitem_fast); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 54, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_11_getitem_fast)) {
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_v_key) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_key);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 54, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "glycopeptidepy/_c/structure/glycan.pyx":55
 * 
 *     cpdef object _getitem_fast(self, key):
 *         return self.obj._getitem_fast(key)             # <<<<<<<<<<<<<<
 * 
 *     def __contains__(self, key):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((struct __pyx_vtabstruct_5glypy_2_c_9structure_18glycan_composition__CompositionBase *)__pyx_v_self->obj->__pyx_vtab)->_getitem_fast(__pyx_v_self->obj, __pyx_v_key, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 55, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "glycopeptidepy/_c/structure/glycan.pyx":54
 *         return self.obj[key]
 * 
 *     cpdef object _getitem_fast(self, key):             # <<<<<<<<<<<<<<
 *         return self.obj._getitem_fast(key)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.glycan.GlycanCompositionProxy._getitem_fast", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_11_getitem_fast(PyObject *__pyx_v_self, PyObject *__pyx_v_key); /*proto*/
static char __pyx_doc_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_10_getitem_fast[] = "GlycanCompositionProxy._getitem_fast(self, key)";
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_11_getitem_fast(PyObject *__pyx_v_self, PyObject *__pyx_v_key) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_getitem_fast (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_10_getitem_fast(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *)__pyx_v_self), ((PyObject *)__pyx_v_key));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_10_getitem_fast(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *__pyx_v_self, PyObject *__pyx_v_key) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_getitem_fast", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy__getitem_fast(__pyx_v_self, __pyx_v_key, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 54, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.glycan.GlycanCompositionProxy._getitem_fast", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/glycan.pyx":57
 *         return self.obj._getitem_fast(key)
 * 
 *     def __contains__(self, key):             # <<<<<<<<<<<<<<
 *         return key in self.obj
 * 
 */

/* Python wrapper */
static int __pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_13__contains__(PyObject *__pyx_v_self, PyObject *__pyx_v_key); /*proto*/
static int __pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_13__contains__(PyObject *__pyx_v_self, PyObject *__pyx_v_key) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__contains__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_12__contains__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *)__pyx_v_self), ((PyObject *)__pyx_v_key));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_12__contains__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *__pyx_v_self, PyObject *__pyx_v_key) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__contains__", 0);

  /* "glycopeptidepy/_c/structure/glycan.pyx":58
 * 
 *     def __contains__(self, key):
 *         return key in self.obj             # <<<<<<<<<<<<<<
 * 
 *     def keys(self):
 */
  __pyx_t_1 = (__Pyx_PySequence_ContainsTF(__pyx_v_key, ((PyObject *)__pyx_v_self->obj), Py_EQ)); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 58, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "glycopeptidepy/_c/structure/glycan.pyx":57
 *         return self.obj._getitem_fast(key)
 * 
 *     def __contains__(self, key):             # <<<<<<<<<<<<<<
 *         return key in self.obj
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("glycopeptidepy._c.structure.glycan.GlycanCompositionProxy.__contains__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/glycan.pyx":60
 *         return key in self.obj
 * 
 *     def keys(self):             # <<<<<<<<<<<<<<
 *         return self.obj.keys()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_15keys(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_14keys[] = "GlycanCompositionProxy.keys(self)";
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_15keys(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("keys (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_14keys(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_14keys(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("keys", 0);

  /* "glycopeptidepy/_c/structure/glycan.pyx":61
 * 
 *     def keys(self):
 *         return self.obj.keys()             # <<<<<<<<<<<<<<
 * 
 *     def values(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->obj), __pyx_n_s_keys); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 61, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 61, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "glycopeptidepy/_c/structure/glycan.pyx":60
 *         return key in self.obj
 * 
 *     def keys(self):             # <<<<<<<<<<<<<<
 *         return self.obj.keys()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.glycan.GlycanCompositionProxy.keys", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/glycan.pyx":63
 *         return self.obj.keys()
 * 
 *     def values(self):             # <<<<<<<<<<<<<<
 *         return self.obj.values()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_17values(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_16values[] = "GlycanCompositionProxy.values(self)";
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_17values(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("values (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_16values(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_16values(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("values", 0);

  /* "glycopeptidepy/_c/structure/glycan.pyx":64
 * 
 *     def values(self):
 *         return self.obj.values()             # <<<<<<<<<<<<<<
 * 
 *     def items(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->obj), __pyx_n_s_values); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 64, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 64, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "glycopeptidepy/_c/structure/glycan.pyx":63
 *         return self.obj.keys()
 * 
 *     def values(self):             # <<<<<<<<<<<<<<
 *         return self.obj.values()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.glycan.GlycanCompositionProxy.values", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/glycan.pyx":66
 *         return self.obj.values()
 * 
 *     def items(self):             # <<<<<<<<<<<<<<
 *         return self.obj.items()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_19items(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_18items[] = "GlycanCompositionProxy.items(self)";
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_19items(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("items (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_18items(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_18items(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("items", 0);

  /* "glycopeptidepy/_c/structure/glycan.pyx":67
 * 
 *     def items(self):
 *         return self.obj.items()             # <<<<<<<<<<<<<<
 * 
 *     def clone(self, *args, **kwargs):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->obj), __pyx_n_s_items); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 67, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 67, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "glycopeptidepy/_c/structure/glycan.pyx":66
 *         return self.obj.values()
 * 
 *     def items(self):             # <<<<<<<<<<<<<<
 *         return self.obj.items()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.glycan.GlycanCompositionProxy.items", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/glycan.pyx":69
 *         return self.obj.items()
 * 
 *     def clone(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         return self.obj.clone(*args, **kwargs)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_21clone(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_20clone[] = "GlycanCompositionProxy.clone(self, *args, **kwargs)";
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_21clone(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_args = 0;
  PyObject *__pyx_v_kwargs = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("clone (wrapper)", 0);
  if (unlikely(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "clone", 1))) return NULL;
  __pyx_v_kwargs = (__pyx_kwds) ? PyDict_Copy(__pyx_kwds) : PyDict_New(); if (unlikely(!__pyx_v_kwargs)) return NULL;
  __Pyx_GOTREF(__pyx_v_kwargs);
  __Pyx_INCREF(__pyx_args);
  __pyx_v_args = __pyx_args;
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_20clone(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *)__pyx_v_self), __pyx_v_args, __pyx_v_kwargs);

  /* function exit code */
  __Pyx_XDECREF(__pyx_v_args);
  __Pyx_XDECREF(__pyx_v_kwargs);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_20clone(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *__pyx_v_self, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("clone", 0);

  /* "glycopeptidepy/_c/structure/glycan.pyx":70
 * 
 *     def clone(self, *args, **kwargs):
 *         return self.obj.clone(*args, **kwargs)             # <<<<<<<<<<<<<<
 * 
 *     def query(self, query, exact=True, **kwargs):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->obj), __pyx_n_s_clone); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 70, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyDict_Copy(__pyx_v_kwargs); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 70, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_v_args, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 70, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "glycopeptidepy/_c/structure/glycan.pyx":69
 *         return self.obj.items()
 * 
 *     def clone(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         return self.obj.clone(*args, **kwargs)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.glycan.GlycanCompositionProxy.clone", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/glycan.pyx":72
 *         return self.obj.clone(*args, **kwargs)
 * 
 *     def query(self, query, exact=True, **kwargs):             # <<<<<<<<<<<<<<
 *         return self.obj.query(query, exact=exact, **kwargs)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_23query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_22query[] = "GlycanCompositionProxy.query(self, query, exact=True, **kwargs)";
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_23query(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_query = 0;
  PyObject *__pyx_v_exact = 0;
  PyObject *__pyx_v_kwargs = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("query (wrapper)", 0);
  __pyx_v_kwargs = PyDict_New(); if (unlikely(!__pyx_v_kwargs)) return NULL;
  __Pyx_GOTREF(__pyx_v_kwargs);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_query,&__pyx_n_s_exact,0};
    PyObject* values[2] = {0,0};
    values[1] = ((PyObject *)Py_True);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_query)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_exact);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, __pyx_v_kwargs, values, pos_args, "query") < 0)) __PYX_ERR(0, 72, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_query = values[0];
    __pyx_v_exact = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("query", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 72, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_DECREF(__pyx_v_kwargs); __pyx_v_kwargs = 0;
  __Pyx_AddTraceback("glycopeptidepy._c.structure.glycan.GlycanCompositionProxy.query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_22query(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *)__pyx_v_self), __pyx_v_query, __pyx_v_exact, __pyx_v_kwargs);

  /* function exit code */
  __Pyx_XDECREF(__pyx_v_kwargs);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_22query(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *__pyx_v_self, PyObject *__pyx_v_query, PyObject *__pyx_v_exact, PyObject *__pyx_v_kwargs) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("query", 0);

  /* "glycopeptidepy/_c/structure/glycan.pyx":73
 * 
 *     def query(self, query, exact=True, **kwargs):
 *         return self.obj.query(query, exact=exact, **kwargs)             # <<<<<<<<<<<<<<
 * 
 *     cpdef str serialize(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->obj), __pyx_n_s_query); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 73, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 73, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_query);
  __Pyx_GIVEREF(__pyx_v_query);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_query);
  __pyx_t_4 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 73, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_exact, __pyx_v_exact) < 0) __PYX_ERR(0, 73, __pyx_L1_error)
  __pyx_t_3 = __pyx_t_4;
  __pyx_t_4 = 0;
  if (__Pyx_MergeKeywords(__pyx_t_3, __pyx_v_kwargs) < 0) __PYX_ERR(0, 73, __pyx_L1_error)
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 73, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "glycopeptidepy/_c/structure/glycan.pyx":72
 *         return self.obj.clone(*args, **kwargs)
 * 
 *     def query(self, query, exact=True, **kwargs):             # <<<<<<<<<<<<<<
 *         return self.obj.query(query, exact=exact, **kwargs)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.glycan.GlycanCompositionProxy.query", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/glycan.pyx":75
 *         return self.obj.query(query, exact=exact, **kwargs)
 * 
 *     cpdef str serialize(self):             # <<<<<<<<<<<<<<
 *         if self._serialized is not None:
 *             return self._serialized
 */

static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_25serialize(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_f_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_serialize(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *__pyx_v_self, int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("serialize", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_serialize); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 75, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_25serialize)) {
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 75, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (!(likely(PyString_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "str", Py_TYPE(__pyx_t_2)->tp_name), 0))) __PYX_ERR(0, 75, __pyx_L1_error)
        __pyx_r = ((PyObject*)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "glycopeptidepy/_c/structure/glycan.pyx":76
 * 
 *     cpdef str serialize(self):
 *         if self._serialized is not None:             # <<<<<<<<<<<<<<
 *             return self._serialized
 *         self._serialized = self.obj.serialize()
 */
  __pyx_t_5 = (__pyx_v_self->_serialized != ((PyObject*)Py_None));
  __pyx_t_6 = (__pyx_t_5 != 0);
  if (__pyx_t_6) {

    /* "glycopeptidepy/_c/structure/glycan.pyx":77
 *     cpdef str serialize(self):
 *         if self._serialized is not None:
 *             return self._serialized             # <<<<<<<<<<<<<<
 *         self._serialized = self.obj.serialize()
 *         return self._serialized
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_self->_serialized);
    __pyx_r = __pyx_v_self->_serialized;
    goto __pyx_L0;

    /* "glycopeptidepy/_c/structure/glycan.pyx":76
 * 
 *     cpdef str serialize(self):
 *         if self._serialized is not None:             # <<<<<<<<<<<<<<
 *             return self._serialized
 *         self._serialized = self.obj.serialize()
 */
  }

  /* "glycopeptidepy/_c/structure/glycan.pyx":78
 *         if self._serialized is not None:
 *             return self._serialized
 *         self._serialized = self.obj.serialize()             # <<<<<<<<<<<<<<
 *         return self._serialized
 * 
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_5glypy_2_c_9structure_18glycan_composition__CompositionBase *)__pyx_v_self->obj->__pyx_vtab)->serialize(__pyx_v_self->obj, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 78, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->_serialized);
  __Pyx_DECREF(__pyx_v_self->_serialized);
  __pyx_v_self->_serialized = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "glycopeptidepy/_c/structure/glycan.pyx":79
 *             return self._serialized
 *         self._serialized = self.obj.serialize()
 *         return self._serialized             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->_serialized);
  __pyx_r = __pyx_v_self->_serialized;
  goto __pyx_L0;

  /* "glycopeptidepy/_c/structure/glycan.pyx":75
 *         return self.obj.query(query, exact=exact, **kwargs)
 * 
 *     cpdef str serialize(self):             # <<<<<<<<<<<<<<
 *         if self._serialized is not None:
 *             return self._serialized
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.glycan.GlycanCompositionProxy.serialize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_25serialize(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_24serialize[] = "GlycanCompositionProxy.serialize(self) -> str";
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_25serialize(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("serialize (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_24serialize(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_24serialize(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("serialize", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_serialize(__pyx_v_self, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 75, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.glycan.GlycanCompositionProxy.serialize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/glycan.pyx":81
 *         return self._serialized
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return repr(self.obj)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_27__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_27__repr__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_26__repr__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_26__repr__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "glycopeptidepy/_c/structure/glycan.pyx":82
 * 
 *     def __repr__(self):
 *         return repr(self.obj)             # <<<<<<<<<<<<<<
 * 
 *     def __str__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_v_self->obj);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = PyObject_Repr(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 82, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "glycopeptidepy/_c/structure/glycan.pyx":81
 *         return self._serialized
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return repr(self.obj)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.glycan.GlycanCompositionProxy.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/glycan.pyx":84
 *         return repr(self.obj)
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         if self._serialized is not None:
 *             return self._serialized
 */

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_29__str__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_29__str__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__str__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_28__str__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_28__str__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__str__", 0);

  /* "glycopeptidepy/_c/structure/glycan.pyx":85
 * 
 *     def __str__(self):
 *         if self._serialized is not None:             # <<<<<<<<<<<<<<
 *             return self._serialized
 *         return self.serialize()
 */
  __pyx_t_1 = (__pyx_v_self->_serialized != ((PyObject*)Py_None));
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "glycopeptidepy/_c/structure/glycan.pyx":86
 *     def __str__(self):
 *         if self._serialized is not None:
 *             return self._serialized             # <<<<<<<<<<<<<<
 *         return self.serialize()
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_self->_serialized);
    __pyx_r = __pyx_v_self->_serialized;
    goto __pyx_L0;

    /* "glycopeptidepy/_c/structure/glycan.pyx":85
 * 
 *     def __str__(self):
 *         if self._serialized is not None:             # <<<<<<<<<<<<<<
 *             return self._serialized
 *         return self.serialize()
 */
  }

  /* "glycopeptidepy/_c/structure/glycan.pyx":87
 *         if self._serialized is not None:
 *             return self._serialized
 *         return self.serialize()             # <<<<<<<<<<<<<<
 * 
 *     def __hash__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_3 = ((struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *)__pyx_v_self->__pyx_vtab)->serialize(__pyx_v_self, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 87, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "glycopeptidepy/_c/structure/glycan.pyx":84
 *         return repr(self.obj)
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         if self._serialized is not None:
 *             return self._serialized
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.glycan.GlycanCompositionProxy.__str__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/glycan.pyx":89
 *         return self.serialize()
 * 
 *     def __hash__(self):             # <<<<<<<<<<<<<<
 *         return hash(self.obj)
 * 
 */

/* Python wrapper */
static Py_hash_t __pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_31__hash__(PyObject *__pyx_v_self); /*proto*/
static Py_hash_t __pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_31__hash__(PyObject *__pyx_v_self) {
  Py_hash_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__hash__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_30__hash__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static Py_hash_t __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_30__hash__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *__pyx_v_self) {
  Py_hash_t __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_hash_t __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__hash__", 0);

  /* "glycopeptidepy/_c/structure/glycan.pyx":90
 * 
 *     def __hash__(self):
 *         return hash(self.obj)             # <<<<<<<<<<<<<<
 * 
 *     def __eq__(self, other):
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_self->obj);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = PyObject_Hash(__pyx_t_1); if (unlikely(__pyx_t_2 == ((Py_hash_t)-1))) __PYX_ERR(0, 90, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  goto __pyx_L0;

  /* "glycopeptidepy/_c/structure/glycan.pyx":89
 *         return self.serialize()
 * 
 *     def __hash__(self):             # <<<<<<<<<<<<<<
 *         return hash(self.obj)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.glycan.GlycanCompositionProxy.__hash__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  if (unlikely(__pyx_r == -1) && !PyErr_Occurred()) __pyx_r = -2;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/glycan.pyx":92
 *         return hash(self.obj)
 * 
 *     def __eq__(self, other):             # <<<<<<<<<<<<<<
 *         try:
 *             other = other.obj
 */

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_33__eq__(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_33__eq__(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__eq__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_32__eq__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *)__pyx_v_self), ((PyObject *)__pyx_v_other));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_32__eq__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__eq__", 0);
  __Pyx_INCREF(__pyx_v_other);

  /* "glycopeptidepy/_c/structure/glycan.pyx":93
 * 
 *     def __eq__(self, other):
 *         try:             # <<<<<<<<<<<<<<
 *             other = other.obj
 *         except AttributeError:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "glycopeptidepy/_c/structure/glycan.pyx":94
 *     def __eq__(self, other):
 *         try:
 *             other = other.obj             # <<<<<<<<<<<<<<
 *         except AttributeError:
 *             pass
 */
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_other, __pyx_n_s_obj); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 94, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF_SET(__pyx_v_other, __pyx_t_4);
      __pyx_t_4 = 0;

      /* "glycopeptidepy/_c/structure/glycan.pyx":93
 * 
 *     def __eq__(self, other):
 *         try:             # <<<<<<<<<<<<<<
 *             other = other.obj
 *         except AttributeError:
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "glycopeptidepy/_c/structure/glycan.pyx":95
 *         try:
 *             other = other.obj
 *         except AttributeError:             # <<<<<<<<<<<<<<
 *             pass
 *         return self.obj == other
 */
    __pyx_t_5 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_AttributeError);
    if (__pyx_t_5) {
      __Pyx_ErrRestore(0,0,0);
      goto __pyx_L4_exception_handled;
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "glycopeptidepy/_c/structure/glycan.pyx":93
 * 
 *     def __eq__(self, other):
 *         try:             # <<<<<<<<<<<<<<
 *             other = other.obj
 *         except AttributeError:
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L4_exception_handled:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    __pyx_L8_try_end:;
  }

  /* "glycopeptidepy/_c/structure/glycan.pyx":97
 *         except AttributeError:
 *             pass
 *         return self.obj == other             # <<<<<<<<<<<<<<
 * 
 *     def __ne__(self, other):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_4 = PyObject_RichCompare(((PyObject *)__pyx_v_self->obj), __pyx_v_other, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 97, __pyx_L1_error)
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "glycopeptidepy/_c/structure/glycan.pyx":92
 *         return hash(self.obj)
 * 
 *     def __eq__(self, other):             # <<<<<<<<<<<<<<
 *         try:
 *             other = other.obj
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.glycan.GlycanCompositionProxy.__eq__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_other);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/glycan.pyx":99
 *         return self.obj == other
 * 
 *     def __ne__(self, other):             # <<<<<<<<<<<<<<
 *         try:
 *             other = other.obj
 */

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_35__ne__(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_35__ne__(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__ne__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_34__ne__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *)__pyx_v_self), ((PyObject *)__pyx_v_other));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_34__ne__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__ne__", 0);
  __Pyx_INCREF(__pyx_v_other);

  /* "glycopeptidepy/_c/structure/glycan.pyx":100
 * 
 *     def __ne__(self, other):
 *         try:             # <<<<<<<<<<<<<<
 *             other = other.obj
 *         except AttributeError:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "glycopeptidepy/_c/structure/glycan.pyx":101
 *     def __ne__(self, other):
 *         try:
 *             other = other.obj             # <<<<<<<<<<<<<<
 *         except AttributeError:
 *             pass
 */
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_other, __pyx_n_s_obj); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 101, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF_SET(__pyx_v_other, __pyx_t_4);
      __pyx_t_4 = 0;

      /* "glycopeptidepy/_c/structure/glycan.pyx":100
 * 
 *     def __ne__(self, other):
 *         try:             # <<<<<<<<<<<<<<
 *             other = other.obj
 *         except AttributeError:
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "glycopeptidepy/_c/structure/glycan.pyx":102
 *         try:
 *             other = other.obj
 *         except AttributeError:             # <<<<<<<<<<<<<<
 *             pass
 *         return (self.obj != other)
 */
    __pyx_t_5 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_AttributeError);
    if (__pyx_t_5) {
      __Pyx_ErrRestore(0,0,0);
      goto __pyx_L4_exception_handled;
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "glycopeptidepy/_c/structure/glycan.pyx":100
 * 
 *     def __ne__(self, other):
 *         try:             # <<<<<<<<<<<<<<
 *             other = other.obj
 *         except AttributeError:
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L4_exception_handled:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    __pyx_L8_try_end:;
  }

  /* "glycopeptidepy/_c/structure/glycan.pyx":104
 *         except AttributeError:
 *             pass
 *         return (self.obj != other)             # <<<<<<<<<<<<<<
 * 
 *     def __add__(self, other):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_4 = PyObject_RichCompare(((PyObject *)__pyx_v_self->obj), __pyx_v_other, Py_NE); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 104, __pyx_L1_error)
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "glycopeptidepy/_c/structure/glycan.pyx":99
 *         return self.obj == other
 * 
 *     def __ne__(self, other):             # <<<<<<<<<<<<<<
 *         try:
 *             other = other.obj
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.glycan.GlycanCompositionProxy.__ne__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_other);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/glycan.pyx":106
 *         return (self.obj != other)
 * 
 *     def __add__(self, other):             # <<<<<<<<<<<<<<
 *         return self.obj + other
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_37__add__(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_37__add__(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__add__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_36__add__(((PyObject *)__pyx_v_self), ((PyObject *)__pyx_v_other));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_36__add__(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__add__", 0);

  /* "glycopeptidepy/_c/structure/glycan.pyx":107
 * 
 *     def __add__(self, other):
 *         return self.obj + other             # <<<<<<<<<<<<<<
 * 
 *     def __sub__(self, other):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_obj); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 107, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyNumber_Add(__pyx_t_1, __pyx_v_other); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 107, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "glycopeptidepy/_c/structure/glycan.pyx":106
 *         return (self.obj != other)
 * 
 *     def __add__(self, other):             # <<<<<<<<<<<<<<
 *         return self.obj + other
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.glycan.GlycanCompositionProxy.__add__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/glycan.pyx":109
 *         return self.obj + other
 * 
 *     def __sub__(self, other):             # <<<<<<<<<<<<<<
 *         return self.obj - other
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_39__sub__(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_39__sub__(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__sub__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_38__sub__(((PyObject *)__pyx_v_self), ((PyObject *)__pyx_v_other));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_38__sub__(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__sub__", 0);

  /* "glycopeptidepy/_c/structure/glycan.pyx":110
 * 
 *     def __sub__(self, other):
 *         return self.obj - other             # <<<<<<<<<<<<<<
 * 
 *     def __mul__(self, other):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_obj); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 110, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyNumber_Subtract(__pyx_t_1, __pyx_v_other); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 110, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "glycopeptidepy/_c/structure/glycan.pyx":109
 *         return self.obj + other
 * 
 *     def __sub__(self, other):             # <<<<<<<<<<<<<<
 *         return self.obj - other
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.glycan.GlycanCompositionProxy.__sub__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/glycan.pyx":112
 *         return self.obj - other
 * 
 *     def __mul__(self, other):             # <<<<<<<<<<<<<<
 *         return self.obj * other
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_41__mul__(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_41__mul__(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__mul__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_40__mul__(((PyObject *)__pyx_v_self), ((PyObject *)__pyx_v_other));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_40__mul__(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__mul__", 0);

  /* "glycopeptidepy/_c/structure/glycan.pyx":113
 * 
 *     def __mul__(self, other):
 *         return self.obj * other             # <<<<<<<<<<<<<<
 * 
 *     def __len__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_obj); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 113, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyNumber_Multiply(__pyx_t_1, __pyx_v_other); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 113, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "glycopeptidepy/_c/structure/glycan.pyx":112
 *         return self.obj - other
 * 
 *     def __mul__(self, other):             # <<<<<<<<<<<<<<
 *         return self.obj * other
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.glycan.GlycanCompositionProxy.__mul__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/glycan.pyx":115
 *         return self.obj * other
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return len(self.obj)
 * 
 */

/* Python wrapper */
static Py_ssize_t __pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_43__len__(PyObject *__pyx_v_self); /*proto*/
static Py_ssize_t __pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_43__len__(PyObject *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_42__len__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static Py_ssize_t __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_42__len__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__len__", 0);

  /* "glycopeptidepy/_c/structure/glycan.pyx":116
 * 
 *     def __len__(self):
 *         return len(self.obj)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_self->obj);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = PyObject_Length(__pyx_t_1); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(0, 116, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  goto __pyx_L0;

  /* "glycopeptidepy/_c/structure/glycan.pyx":115
 *         return self.obj * other
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return len(self.obj)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.glycan.GlycanCompositionProxy.__len__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/glycan.pxd":18
 * 
 *     cdef:
 *         public _CompositionBase obj             # <<<<<<<<<<<<<<
 *         public str _serialized
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_3obj_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_3obj_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_3obj___get__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_3obj___get__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->obj));
  __pyx_r = ((PyObject *)__pyx_v_self->obj);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_3obj_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_3obj_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_3obj_2__set__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_3obj_2__set__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);
  if (!(likely(((__pyx_v_value) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_value, __pyx_ptype_5glypy_2_c_9structure_18glycan_composition__CompositionBase))))) __PYX_ERR(1, 18, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_value;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->obj);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->obj));
  __pyx_v_self->obj = ((struct __pyx_obj_5glypy_2_c_9structure_18glycan_composition__CompositionBase *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.glycan.GlycanCompositionProxy.obj.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_3obj_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_3obj_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_3obj_4__del__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_3obj_4__del__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->obj);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->obj));
  __pyx_v_self->obj = ((struct __pyx_obj_5glypy_2_c_9structure_18glycan_composition__CompositionBase *)Py_None);

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/glycan.pxd":19
 *     cdef:
 *         public _CompositionBase obj
 *         public str _serialized             # <<<<<<<<<<<<<<
 * 
 *     cpdef object _getitem_fast(self, key)
 */

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_11_serialized_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_11_serialized_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_11_serialized___get__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_11_serialized___get__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->_serialized);
  __pyx_r = __pyx_v_self->_serialized;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_11_serialized_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_11_serialized_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_11_serialized_2__set__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_11_serialized_2__set__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);
  if (!(likely(PyString_CheckExact(__pyx_v_value))||((__pyx_v_value) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "str", Py_TYPE(__pyx_v_value)->tp_name), 0))) __PYX_ERR(1, 19, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_value;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->_serialized);
  __Pyx_DECREF(__pyx_v_self->_serialized);
  __pyx_v_self->_serialized = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.glycan.GlycanCompositionProxy._serialized.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_11_serialized_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_11_serialized_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_11_serialized_4__del__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_11_serialized_4__del__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->_serialized);
  __Pyx_DECREF(__pyx_v_self->_serialized);
  __pyx_v_self->_serialized = ((PyObject*)Py_None);

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_45__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_44__reduce_cython__[] = "GlycanCompositionProxy.__reduce_cython__(self)";
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_45__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_44__reduce_cython__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_44__reduce_cython__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self._serialized, self.obj)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_self->_serialized);
  __Pyx_GIVEREF(__pyx_v_self->_serialized);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_self->_serialized);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->obj));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self->obj));
  PyTuple_SET_ITEM(__pyx_t_1, 1, ((PyObject *)__pyx_v_self->obj));
  __pyx_v_state = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self._serialized, self.obj)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = (self._serialized, self.obj)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict);
    __pyx_t_4 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_4));
    __pyx_t_4 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self._serialized is not None or self.obj is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self._serialized, self.obj)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self._serialized is not None or self.obj is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_GlycanCompositionProxy, (type(self), 0x4126d5b, None), state
 */
  /*else*/ {
    __pyx_t_2 = (__pyx_v_self->_serialized != ((PyObject*)Py_None));
    __pyx_t_5 = (__pyx_t_2 != 0);
    if (!__pyx_t_5) {
    } else {
      __pyx_t_3 = __pyx_t_5;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_5 = (((PyObject *)__pyx_v_self->obj) != Py_None);
    __pyx_t_2 = (__pyx_t_5 != 0);
    __pyx_t_3 = __pyx_t_2;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_3;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self._serialized is not None or self.obj is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_GlycanCompositionProxy, (type(self), 0x4126d5b, None), state
 *     else:
 */
  __pyx_t_3 = (__pyx_v_use_setstate != 0);
  if (__pyx_t_3) {

    /* "(tree fragment)":13
 *         use_setstate = self._serialized is not None or self.obj is not None
 *     if use_setstate:
 *         return __pyx_unpickle_GlycanCompositionProxy, (type(self), 0x4126d5b, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_GlycanCompositionProxy, (type(self), 0x4126d5b, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_pyx_unpickle_GlycanComposition); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_68316507);
    __Pyx_GIVEREF(__pyx_int_68316507);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_68316507);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None);
    __pyx_t_6 = PyTuple_New(3); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_1);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_v_state);
    __pyx_t_4 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_6;
    __pyx_t_6 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self._serialized is not None or self.obj is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_GlycanCompositionProxy, (type(self), 0x4126d5b, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_GlycanCompositionProxy, (type(self), 0x4126d5b, None), state
 *     else:
 *         return __pyx_unpickle_GlycanCompositionProxy, (type(self), 0x4126d5b, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_GlycanCompositionProxy__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_pyx_unpickle_GlycanComposition); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_68316507);
    __Pyx_GIVEREF(__pyx_int_68316507);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_68316507);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state);
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1);
    __pyx_t_6 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.glycan.GlycanCompositionProxy.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_GlycanCompositionProxy, (type(self), 0x4126d5b, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_GlycanCompositionProxy__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_47__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static char __pyx_doc_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_46__setstate_cython__[] = "GlycanCompositionProxy.__setstate_cython__(self, __pyx_state)";
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_47__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_46__setstate_cython__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_46__setstate_cython__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_GlycanCompositionProxy, (type(self), 0x4126d5b, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_GlycanCompositionProxy__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(2, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_14glycopeptidepy_2_c_9structure_6glycan___pyx_unpickle_GlycanCompositionProxy__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_GlycanCompositionProxy, (type(self), 0x4126d5b, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_GlycanCompositionProxy__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.glycan.GlycanCompositionProxy.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/glycan.pyx":120
 * 
 * cdef class GlycanCompositionWithOffsetProxy(GlycanCompositionProxy):
 *     def __init__(self, obj, offset=None):             # <<<<<<<<<<<<<<
 *         if offset is None:
 *             offset = CComposition._create(None)
 */

/* Python wrapper */
static int __pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_32GlycanCompositionWithOffsetProxy_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_32GlycanCompositionWithOffsetProxy_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_obj = 0;
  PyObject *__pyx_v_offset = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_obj,&__pyx_n_s_offset,0};
    PyObject* values[2] = {0,0};
    values[1] = ((PyObject *)Py_None);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_obj)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_offset);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 120, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_obj = values[0];
    __pyx_v_offset = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 120, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("glycopeptidepy._c.structure.glycan.GlycanCompositionWithOffsetProxy.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_32GlycanCompositionWithOffsetProxy___init__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionWithOffsetProxy *)__pyx_v_self), __pyx_v_obj, __pyx_v_offset);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_32GlycanCompositionWithOffsetProxy___init__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionWithOffsetProxy *__pyx_v_self, PyObject *__pyx_v_obj, PyObject *__pyx_v_offset) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);
  __Pyx_INCREF(__pyx_v_offset);

  /* "glycopeptidepy/_c/structure/glycan.pyx":121
 * cdef class GlycanCompositionWithOffsetProxy(GlycanCompositionProxy):
 *     def __init__(self, obj, offset=None):
 *         if offset is None:             # <<<<<<<<<<<<<<
 *             offset = CComposition._create(None)
 *         GlycanCompositionProxy.__init__(self, obj)
 */
  __pyx_t_1 = (__pyx_v_offset == Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "glycopeptidepy/_c/structure/glycan.pyx":122
 *     def __init__(self, obj, offset=None):
 *         if offset is None:
 *             offset = CComposition._create(None)             # <<<<<<<<<<<<<<
 *         GlycanCompositionProxy.__init__(self, obj)
 *         self.composition_offset = offset
 */
    __pyx_t_3 = ((PyObject *)__pyx_vtabptr_5glypy_11composition_12ccomposition_CComposition->_create(((struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *)Py_None))); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 122, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF_SET(__pyx_v_offset, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "glycopeptidepy/_c/structure/glycan.pyx":121
 * cdef class GlycanCompositionWithOffsetProxy(GlycanCompositionProxy):
 *     def __init__(self, obj, offset=None):
 *         if offset is None:             # <<<<<<<<<<<<<<
 *             offset = CComposition._create(None)
 *         GlycanCompositionProxy.__init__(self, obj)
 */
  }

  /* "glycopeptidepy/_c/structure/glycan.pyx":123
 *         if offset is None:
 *             offset = CComposition._create(None)
 *         GlycanCompositionProxy.__init__(self, obj)             # <<<<<<<<<<<<<<
 *         self.composition_offset = offset
 * 
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy), __pyx_n_s_init); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 123, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_4)) {
    PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_self), __pyx_v_obj};
    __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 123, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_3);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
    PyObject *__pyx_temp[3] = {__pyx_t_5, ((PyObject *)__pyx_v_self), __pyx_v_obj};
    __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 123, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GOTREF(__pyx_t_3);
  } else
  #endif
  {
    __pyx_t_7 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 123, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (__pyx_t_5) {
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_5); __pyx_t_5 = NULL;
    }
    __Pyx_INCREF(((PyObject *)__pyx_v_self));
    __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
    PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_6, ((PyObject *)__pyx_v_self));
    __Pyx_INCREF(__pyx_v_obj);
    __Pyx_GIVEREF(__pyx_v_obj);
    PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_6, __pyx_v_obj);
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_7, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 123, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  }
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "glycopeptidepy/_c/structure/glycan.pyx":124
 *             offset = CComposition._create(None)
 *         GlycanCompositionProxy.__init__(self, obj)
 *         self.composition_offset = offset             # <<<<<<<<<<<<<<
 * 
 * 
 */
  if (!(likely(((__pyx_v_offset) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_offset, __pyx_ptype_5glypy_11composition_12ccomposition_CComposition))))) __PYX_ERR(0, 124, __pyx_L1_error)
  __pyx_t_3 = __pyx_v_offset;
  __Pyx_INCREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __Pyx_GOTREF(__pyx_v_self->composition_offset);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->composition_offset));
  __pyx_v_self->composition_offset = ((struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "glycopeptidepy/_c/structure/glycan.pyx":120
 * 
 * cdef class GlycanCompositionWithOffsetProxy(GlycanCompositionProxy):
 *     def __init__(self, obj, offset=None):             # <<<<<<<<<<<<<<
 *         if offset is None:
 *             offset = CComposition._create(None)
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.glycan.GlycanCompositionWithOffsetProxy.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_offset);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/glycan.pyx":127
 * 
 * 
 *     def clone(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         return self.__class__(self.obj, self.composition_offset.copy())
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_32GlycanCompositionWithOffsetProxy_3clone(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_14glycopeptidepy_2_c_9structure_6glycan_32GlycanCompositionWithOffsetProxy_2clone[] = "GlycanCompositionWithOffsetProxy.clone(self, *args, **kwargs)";
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_32GlycanCompositionWithOffsetProxy_3clone(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_args = 0;
  CYTHON_UNUSED PyObject *__pyx_v_kwargs = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("clone (wrapper)", 0);
  if (unlikely(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "clone", 1))) return NULL;
  __Pyx_INCREF(__pyx_args);
  __pyx_v_args = __pyx_args;
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_32GlycanCompositionWithOffsetProxy_2clone(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionWithOffsetProxy *)__pyx_v_self), __pyx_v_args, __pyx_v_kwargs);

  /* function exit code */
  __Pyx_XDECREF(__pyx_v_args);
  __Pyx_XDECREF(__pyx_v_kwargs);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_32GlycanCompositionWithOffsetProxy_2clone(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionWithOffsetProxy *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_args, CYTHON_UNUSED PyObject *__pyx_v_kwargs) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("clone", 0);

  /* "glycopeptidepy/_c/structure/glycan.pyx":128
 * 
 *     def clone(self, *args, **kwargs):
 *         return self.__class__(self.obj, self.composition_offset.copy())             # <<<<<<<<<<<<<<
 * 
 *     def mass(self, *args, **kwargs):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_class); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 128, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->composition_offset), __pyx_n_s_copy); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 128, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 128, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = NULL;
  __pyx_t_6 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_6 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_4, ((PyObject *)__pyx_v_self->__pyx_base.obj), __pyx_t_3};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 128, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_4, ((PyObject *)__pyx_v_self->__pyx_base.obj), __pyx_t_3};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_6, 2+__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 128, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else
  #endif
  {
    __pyx_t_5 = PyTuple_New(2+__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 128, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (__pyx_t_4) {
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4); __pyx_t_4 = NULL;
    }
    __Pyx_INCREF(((PyObject *)__pyx_v_self->__pyx_base.obj));
    __Pyx_GIVEREF(((PyObject *)__pyx_v_self->__pyx_base.obj));
    PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_6, ((PyObject *)__pyx_v_self->__pyx_base.obj));
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_6, __pyx_t_3);
    __pyx_t_3 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 128, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "glycopeptidepy/_c/structure/glycan.pyx":127
 * 
 * 
 *     def clone(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         return self.__class__(self.obj, self.composition_offset.copy())
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.glycan.GlycanCompositionWithOffsetProxy.clone", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/glycan.pyx":130
 *         return self.__class__(self.obj, self.composition_offset.copy())
 * 
 *     def mass(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         mass = self.obj.mass(*args, **kwargs)
 *         return mass + self.composition_offset.mass
 */

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_32GlycanCompositionWithOffsetProxy_5mass(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_14glycopeptidepy_2_c_9structure_6glycan_32GlycanCompositionWithOffsetProxy_4mass[] = "GlycanCompositionWithOffsetProxy.mass(self, *args, **kwargs)";
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_32GlycanCompositionWithOffsetProxy_5mass(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_args = 0;
  PyObject *__pyx_v_kwargs = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("mass (wrapper)", 0);
  if (unlikely(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "mass", 1))) return NULL;
  __pyx_v_kwargs = (__pyx_kwds) ? PyDict_Copy(__pyx_kwds) : PyDict_New(); if (unlikely(!__pyx_v_kwargs)) return NULL;
  __Pyx_GOTREF(__pyx_v_kwargs);
  __Pyx_INCREF(__pyx_args);
  __pyx_v_args = __pyx_args;
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_32GlycanCompositionWithOffsetProxy_4mass(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionWithOffsetProxy *)__pyx_v_self), __pyx_v_args, __pyx_v_kwargs);

  /* function exit code */
  __Pyx_XDECREF(__pyx_v_args);
  __Pyx_XDECREF(__pyx_v_kwargs);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_32GlycanCompositionWithOffsetProxy_4mass(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionWithOffsetProxy *__pyx_v_self, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs) {
  PyObject *__pyx_v_mass = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("mass", 0);

  /* "glycopeptidepy/_c/structure/glycan.pyx":131
 * 
 *     def mass(self, *args, **kwargs):
 *         mass = self.obj.mass(*args, **kwargs)             # <<<<<<<<<<<<<<
 *         return mass + self.composition_offset.mass
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->__pyx_base.obj), __pyx_n_s_mass); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 131, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyDict_Copy(__pyx_v_kwargs); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 131, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_v_args, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 131, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_mass = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "glycopeptidepy/_c/structure/glycan.pyx":132
 *     def mass(self, *args, **kwargs):
 *         mass = self.obj.mass(*args, **kwargs)
 *         return mass + self.composition_offset.mass             # <<<<<<<<<<<<<<
 * 
 *     def total_composition(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->composition_offset), __pyx_n_s_mass); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 132, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = PyNumber_Add(__pyx_v_mass, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 132, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "glycopeptidepy/_c/structure/glycan.pyx":130
 *         return self.__class__(self.obj, self.composition_offset.copy())
 * 
 *     def mass(self, *args, **kwargs):             # <<<<<<<<<<<<<<
 *         mass = self.obj.mass(*args, **kwargs)
 *         return mass + self.composition_offset.mass
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.glycan.GlycanCompositionWithOffsetProxy.mass", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_mass);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/glycan.pyx":134
 *         return mass + self.composition_offset.mass
 * 
 *     def total_composition(self):             # <<<<<<<<<<<<<<
 *         comp = self.obj.total_composition()
 *         return comp + self.composition_offset
 */

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_32GlycanCompositionWithOffsetProxy_7total_composition(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_14glycopeptidepy_2_c_9structure_6glycan_32GlycanCompositionWithOffsetProxy_6total_composition[] = "GlycanCompositionWithOffsetProxy.total_composition(self)";
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_32GlycanCompositionWithOffsetProxy_7total_composition(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("total_composition (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_32GlycanCompositionWithOffsetProxy_6total_composition(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionWithOffsetProxy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_32GlycanCompositionWithOffsetProxy_6total_composition(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionWithOffsetProxy *__pyx_v_self) {
  PyObject *__pyx_v_comp = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("total_composition", 0);

  /* "glycopeptidepy/_c/structure/glycan.pyx":135
 * 
 *     def total_composition(self):
 *         comp = self.obj.total_composition()             # <<<<<<<<<<<<<<
 *         return comp + self.composition_offset
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->__pyx_base.obj), __pyx_n_s_total_composition); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 135, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 135, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_comp = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "glycopeptidepy/_c/structure/glycan.pyx":136
 *     def total_composition(self):
 *         comp = self.obj.total_composition()
 *         return comp + self.composition_offset             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyNumber_Add(__pyx_v_comp, ((PyObject *)__pyx_v_self->composition_offset)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 136, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "glycopeptidepy/_c/structure/glycan.pyx":134
 *         return mass + self.composition_offset.mass
 * 
 *     def total_composition(self):             # <<<<<<<<<<<<<<
 *         comp = self.obj.total_composition()
 *         return comp + self.composition_offset
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.glycan.GlycanCompositionWithOffsetProxy.total_composition", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_comp);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/glycan.pxd":31
 * cdef class GlycanCompositionWithOffsetProxy(GlycanCompositionProxy):
 *     cdef:
 *         public CComposition composition_offset             # <<<<<<<<<<<<<<
 * 
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_32GlycanCompositionWithOffsetProxy_18composition_offset_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_32GlycanCompositionWithOffsetProxy_18composition_offset_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_32GlycanCompositionWithOffsetProxy_18composition_offset___get__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionWithOffsetProxy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_32GlycanCompositionWithOffsetProxy_18composition_offset___get__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionWithOffsetProxy *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->composition_offset));
  __pyx_r = ((PyObject *)__pyx_v_self->composition_offset);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_32GlycanCompositionWithOffsetProxy_18composition_offset_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_32GlycanCompositionWithOffsetProxy_18composition_offset_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_32GlycanCompositionWithOffsetProxy_18composition_offset_2__set__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionWithOffsetProxy *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_32GlycanCompositionWithOffsetProxy_18composition_offset_2__set__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionWithOffsetProxy *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);
  if (!(likely(((__pyx_v_value) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_value, __pyx_ptype_5glypy_11composition_12ccomposition_CComposition))))) __PYX_ERR(1, 31, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_value;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->composition_offset);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->composition_offset));
  __pyx_v_self->composition_offset = ((struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.glycan.GlycanCompositionWithOffsetProxy.composition_offset.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_32GlycanCompositionWithOffsetProxy_18composition_offset_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_32GlycanCompositionWithOffsetProxy_18composition_offset_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_32GlycanCompositionWithOffsetProxy_18composition_offset_4__del__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionWithOffsetProxy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_32GlycanCompositionWithOffsetProxy_18composition_offset_4__del__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionWithOffsetProxy *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->composition_offset);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->composition_offset));
  __pyx_v_self->composition_offset = ((struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *)Py_None);

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_32GlycanCompositionWithOffsetProxy_9__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_14glycopeptidepy_2_c_9structure_6glycan_32GlycanCompositionWithOffsetProxy_8__reduce_cython__[] = "GlycanCompositionWithOffsetProxy.__reduce_cython__(self)";
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_32GlycanCompositionWithOffsetProxy_9__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_32GlycanCompositionWithOffsetProxy_8__reduce_cython__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionWithOffsetProxy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_32GlycanCompositionWithOffsetProxy_8__reduce_cython__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionWithOffsetProxy *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self._serialized, self.composition_offset, self.obj)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_self->__pyx_base._serialized);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base._serialized);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_self->__pyx_base._serialized);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->composition_offset));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self->composition_offset));
  PyTuple_SET_ITEM(__pyx_t_1, 1, ((PyObject *)__pyx_v_self->composition_offset));
  __Pyx_INCREF(((PyObject *)__pyx_v_self->__pyx_base.obj));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self->__pyx_base.obj));
  PyTuple_SET_ITEM(__pyx_t_1, 2, ((PyObject *)__pyx_v_self->__pyx_base.obj));
  __pyx_v_state = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self._serialized, self.composition_offset, self.obj)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = (self._serialized, self.composition_offset, self.obj)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict);
    __pyx_t_4 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_4));
    __pyx_t_4 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self._serialized is not None or self.composition_offset is not None or self.obj is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self._serialized, self.composition_offset, self.obj)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self._serialized is not None or self.composition_offset is not None or self.obj is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_GlycanCompositionWithOffsetProxy, (type(self), 0x32b2f99, None), state
 */
  /*else*/ {
    __pyx_t_2 = (__pyx_v_self->__pyx_base._serialized != ((PyObject*)Py_None));
    __pyx_t_5 = (__pyx_t_2 != 0);
    if (!__pyx_t_5) {
    } else {
      __pyx_t_3 = __pyx_t_5;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_5 = (((PyObject *)__pyx_v_self->composition_offset) != Py_None);
    __pyx_t_2 = (__pyx_t_5 != 0);
    if (!__pyx_t_2) {
    } else {
      __pyx_t_3 = __pyx_t_2;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_2 = (((PyObject *)__pyx_v_self->__pyx_base.obj) != Py_None);
    __pyx_t_5 = (__pyx_t_2 != 0);
    __pyx_t_3 = __pyx_t_5;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_3;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self._serialized is not None or self.composition_offset is not None or self.obj is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_GlycanCompositionWithOffsetProxy, (type(self), 0x32b2f99, None), state
 *     else:
 */
  __pyx_t_3 = (__pyx_v_use_setstate != 0);
  if (__pyx_t_3) {

    /* "(tree fragment)":13
 *         use_setstate = self._serialized is not None or self.composition_offset is not None or self.obj is not None
 *     if use_setstate:
 *         return __pyx_unpickle_GlycanCompositionWithOffsetProxy, (type(self), 0x32b2f99, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_GlycanCompositionWithOffsetProxy, (type(self), 0x32b2f99, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_pyx_unpickle_GlycanComposition_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_53161881);
    __Pyx_GIVEREF(__pyx_int_53161881);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_53161881);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None);
    __pyx_t_6 = PyTuple_New(3); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_1);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_v_state);
    __pyx_t_4 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_6;
    __pyx_t_6 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self._serialized is not None or self.composition_offset is not None or self.obj is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_GlycanCompositionWithOffsetProxy, (type(self), 0x32b2f99, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_GlycanCompositionWithOffsetProxy, (type(self), 0x32b2f99, None), state
 *     else:
 *         return __pyx_unpickle_GlycanCompositionWithOffsetProxy, (type(self), 0x32b2f99, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_GlycanCompositionWithOffsetProxy__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_pyx_unpickle_GlycanComposition_2); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_53161881);
    __Pyx_GIVEREF(__pyx_int_53161881);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_53161881);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state);
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1);
    __pyx_t_6 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.glycan.GlycanCompositionWithOffsetProxy.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_GlycanCompositionWithOffsetProxy, (type(self), 0x32b2f99, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_GlycanCompositionWithOffsetProxy__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_32GlycanCompositionWithOffsetProxy_11__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static char __pyx_doc_14glycopeptidepy_2_c_9structure_6glycan_32GlycanCompositionWithOffsetProxy_10__setstate_cython__[] = "GlycanCompositionWithOffsetProxy.__setstate_cython__(self, __pyx_state)";
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_32GlycanCompositionWithOffsetProxy_11__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_32GlycanCompositionWithOffsetProxy_10__setstate_cython__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionWithOffsetProxy *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_32GlycanCompositionWithOffsetProxy_10__setstate_cython__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionWithOffsetProxy *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_GlycanCompositionWithOffsetProxy, (type(self), 0x32b2f99, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_GlycanCompositionWithOffsetProxy__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(2, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_14glycopeptidepy_2_c_9structure_6glycan___pyx_unpickle_GlycanCompositionWithOffsetProxy__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_GlycanCompositionWithOffsetProxy, (type(self), 0x32b2f99, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_GlycanCompositionWithOffsetProxy__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.glycan.GlycanCompositionWithOffsetProxy.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/glycan.pyx":147
 * @cython.freelist(10000)
 * cdef class GlycosylationManager(object):
 *     def __init__(self, parent, aggregate=None):             # <<<<<<<<<<<<<<
 *         self._init()
 *         self.parent = parent
 */

/* Python wrapper */
static int __pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_parent = 0;
  PyObject *__pyx_v_aggregate = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_parent,&__pyx_n_s_aggregate,0};
    PyObject* values[2] = {0,0};
    values[1] = ((PyObject *)Py_None);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_parent)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_aggregate);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 147, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_parent = values[0];
    __pyx_v_aggregate = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 147, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("glycopeptidepy._c.structure.glycan.GlycosylationManager.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager___init__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *)__pyx_v_self), __pyx_v_parent, __pyx_v_aggregate);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager___init__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_self, PyObject *__pyx_v_parent, PyObject *__pyx_v_aggregate) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "glycopeptidepy/_c/structure/glycan.pyx":148
 * cdef class GlycosylationManager(object):
 *     def __init__(self, parent, aggregate=None):
 *         self._init()             # <<<<<<<<<<<<<<
 *         self.parent = parent
 *         if aggregate is not None:
 */
  ((struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *)__pyx_v_self->__pyx_vtab)->_init(__pyx_v_self);

  /* "glycopeptidepy/_c/structure/glycan.pyx":149
 *     def __init__(self, parent, aggregate=None):
 *         self._init()
 *         self.parent = parent             # <<<<<<<<<<<<<<
 *         if aggregate is not None:
 *             self.set_aggregate(aggregate)
 */
  __Pyx_INCREF(__pyx_v_parent);
  __Pyx_GIVEREF(__pyx_v_parent);
  __Pyx_GOTREF(__pyx_v_self->parent);
  __Pyx_DECREF(__pyx_v_self->parent);
  __pyx_v_self->parent = __pyx_v_parent;

  /* "glycopeptidepy/_c/structure/glycan.pyx":150
 *         self._init()
 *         self.parent = parent
 *         if aggregate is not None:             # <<<<<<<<<<<<<<
 *             self.set_aggregate(aggregate)
 * 
 */
  __pyx_t_1 = (__pyx_v_aggregate != Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "glycopeptidepy/_c/structure/glycan.pyx":151
 *         self.parent = parent
 *         if aggregate is not None:
 *             self.set_aggregate(aggregate)             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */
    (void)(((struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *)__pyx_v_self->__pyx_vtab)->set_aggregate(__pyx_v_self, __pyx_v_aggregate));

    /* "glycopeptidepy/_c/structure/glycan.pyx":150
 *         self._init()
 *         self.parent = parent
 *         if aggregate is not None:             # <<<<<<<<<<<<<<
 *             self.set_aggregate(aggregate)
 * 
 */
  }

  /* "glycopeptidepy/_c/structure/glycan.pyx":147
 * @cython.freelist(10000)
 * cdef class GlycosylationManager(object):
 *     def __init__(self, parent, aggregate=None):             # <<<<<<<<<<<<<<
 *         self._init()
 *         self.parent = parent
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/glycan.pyx":154
 * 
 *     @staticmethod
 *     cdef GlycosylationManager _create(PeptideSequenceBase parent, object aggregate):             # <<<<<<<<<<<<<<
 *         cdef:
 *             GlycosylationManager self
 */

static struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_f_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager__create(struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_PeptideSequenceBase *__pyx_v_parent, PyObject *__pyx_v_aggregate) {
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_self = 0;
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_create", 0);

  /* "glycopeptidepy/_c/structure/glycan.pyx":157
 *         cdef:
 *             GlycosylationManager self
 *         self = GlycosylationManager.__new__(GlycosylationManager)             # <<<<<<<<<<<<<<
 *         self._init()
 *         self.parent = parent
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager(((PyTypeObject *)__pyx_ptype_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 157, __pyx_L1_error)
  __Pyx_GOTREF(((PyObject *)__pyx_t_1));
  __pyx_v_self = ((struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "glycopeptidepy/_c/structure/glycan.pyx":158
 *             GlycosylationManager self
 *         self = GlycosylationManager.__new__(GlycosylationManager)
 *         self._init()             # <<<<<<<<<<<<<<
 *         self.parent = parent
 *         if aggregate is not None:
 */
  ((struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *)__pyx_v_self->__pyx_vtab)->_init(__pyx_v_self);

  /* "glycopeptidepy/_c/structure/glycan.pyx":159
 *         self = GlycosylationManager.__new__(GlycosylationManager)
 *         self._init()
 *         self.parent = parent             # <<<<<<<<<<<<<<
 *         if aggregate is not None:
 *             self.set_aggregate(aggregate)
 */
  __Pyx_INCREF(((PyObject *)__pyx_v_parent));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_parent));
  __Pyx_GOTREF(__pyx_v_self->parent);
  __Pyx_DECREF(__pyx_v_self->parent);
  __pyx_v_self->parent = ((PyObject *)__pyx_v_parent);

  /* "glycopeptidepy/_c/structure/glycan.pyx":160
 *         self._init()
 *         self.parent = parent
 *         if aggregate is not None:             # <<<<<<<<<<<<<<
 *             self.set_aggregate(aggregate)
 *         return self
 */
  __pyx_t_2 = (__pyx_v_aggregate != Py_None);
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {

    /* "glycopeptidepy/_c/structure/glycan.pyx":161
 *         self.parent = parent
 *         if aggregate is not None:
 *             self.set_aggregate(aggregate)             # <<<<<<<<<<<<<<
 *         return self
 * 
 */
    (void)(((struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *)__pyx_v_self->__pyx_vtab)->set_aggregate(__pyx_v_self, __pyx_v_aggregate));

    /* "glycopeptidepy/_c/structure/glycan.pyx":160
 *         self._init()
 *         self.parent = parent
 *         if aggregate is not None:             # <<<<<<<<<<<<<<
 *             self.set_aggregate(aggregate)
 *         return self
 */
  }

  /* "glycopeptidepy/_c/structure/glycan.pyx":162
 *         if aggregate is not None:
 *             self.set_aggregate(aggregate)
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     cdef void _init(self):
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __Pyx_INCREF(((PyObject *)__pyx_v_self));
  __pyx_r = __pyx_v_self;
  goto __pyx_L0;

  /* "glycopeptidepy/_c/structure/glycan.pyx":154
 * 
 *     @staticmethod
 *     cdef GlycosylationManager _create(PeptideSequenceBase parent, object aggregate):             # <<<<<<<<<<<<<<
 *         cdef:
 *             GlycosylationManager self
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.glycan.GlycosylationManager._create", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_self);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/glycan.pyx":164
 *         return self
 * 
 *     cdef void _init(self):             # <<<<<<<<<<<<<<
 *         self.mapping = {}
 *         self._aggregate = None
 */

static void __pyx_f_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager__init(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_init", 0);

  /* "glycopeptidepy/_c/structure/glycan.pyx":165
 * 
 *     cdef void _init(self):
 *         self.mapping = {}             # <<<<<<<<<<<<<<
 *         self._aggregate = None
 *         self._proxy = None
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 165, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->mapping);
  __Pyx_DECREF(__pyx_v_self->mapping);
  __pyx_v_self->mapping = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "glycopeptidepy/_c/structure/glycan.pyx":166
 *     cdef void _init(self):
 *         self.mapping = {}
 *         self._aggregate = None             # <<<<<<<<<<<<<<
 *         self._proxy = None
 *         self._type_track = None
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->_aggregate);
  __Pyx_DECREF(__pyx_v_self->_aggregate);
  __pyx_v_self->_aggregate = Py_None;

  /* "glycopeptidepy/_c/structure/glycan.pyx":167
 *         self.mapping = {}
 *         self._aggregate = None
 *         self._proxy = None             # <<<<<<<<<<<<<<
 *         self._type_track = None
 *         self._total_glycosylation_size = -1
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->_proxy);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->_proxy));
  __pyx_v_self->_proxy = ((struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *)Py_None);

  /* "glycopeptidepy/_c/structure/glycan.pyx":168
 *         self._aggregate = None
 *         self._proxy = None
 *         self._type_track = None             # <<<<<<<<<<<<<<
 *         self._total_glycosylation_size = -1
 * 
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->_type_track);
  __Pyx_DECREF(__pyx_v_self->_type_track);
  __pyx_v_self->_type_track = ((PyObject*)Py_None);

  /* "glycopeptidepy/_c/structure/glycan.pyx":169
 *         self._proxy = None
 *         self._type_track = None
 *         self._total_glycosylation_size = -1             # <<<<<<<<<<<<<<
 * 
 *     def __getitem__(self, key):
 */
  __pyx_v_self->_total_glycosylation_size = -1;

  /* "glycopeptidepy/_c/structure/glycan.pyx":164
 *         return self
 * 
 *     cdef void _init(self):             # <<<<<<<<<<<<<<
 *         self.mapping = {}
 *         self._aggregate = None
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_WriteUnraisable("glycopeptidepy._c.structure.glycan.GlycosylationManager._init", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "glycopeptidepy/_c/structure/glycan.pyx":171
 *         self._total_glycosylation_size = -1
 * 
 *     def __getitem__(self, key):             # <<<<<<<<<<<<<<
 *         return self.mapping[key]
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_3__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_key); /*proto*/
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_3__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_key) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getitem__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_2__getitem__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *)__pyx_v_self), ((PyObject *)__pyx_v_key));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_2__getitem__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_self, PyObject *__pyx_v_key) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__getitem__", 0);

  /* "glycopeptidepy/_c/structure/glycan.pyx":172
 * 
 *     def __getitem__(self, key):
 *         return self.mapping[key]             # <<<<<<<<<<<<<<
 * 
 *     def __setitem__(self, key, value):
 */
  __Pyx_XDECREF(__pyx_r);
  if (unlikely(__pyx_v_self->mapping == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 172, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyDict_GetItem(__pyx_v_self->mapping, __pyx_v_key); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 172, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "glycopeptidepy/_c/structure/glycan.pyx":171
 *         self._total_glycosylation_size = -1
 * 
 *     def __getitem__(self, key):             # <<<<<<<<<<<<<<
 *         return self.mapping[key]
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.glycan.GlycosylationManager.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/glycan.pyx":174
 *         return self.mapping[key]
 * 
 *     def __setitem__(self, key, value):             # <<<<<<<<<<<<<<
 *         self.invalidate()
 *         self.mapping[key] = value
 */

/* Python wrapper */
static int __pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_5__setitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_key, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_5__setitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_key, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setitem__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_4__setitem__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *)__pyx_v_self), ((PyObject *)__pyx_v_key), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_4__setitem__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_self, PyObject *__pyx_v_key, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setitem__", 0);

  /* "glycopeptidepy/_c/structure/glycan.pyx":175
 * 
 *     def __setitem__(self, key, value):
 *         self.invalidate()             # <<<<<<<<<<<<<<
 *         self.mapping[key] = value
 * 
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *)__pyx_v_self->__pyx_vtab)->invalidate(__pyx_v_self, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 175, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "glycopeptidepy/_c/structure/glycan.pyx":176
 *     def __setitem__(self, key, value):
 *         self.invalidate()
 *         self.mapping[key] = value             # <<<<<<<<<<<<<<
 * 
 *     def __delitem__(self, key):
 */
  if (unlikely(__pyx_v_self->mapping == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 176, __pyx_L1_error)
  }
  if (unlikely(PyDict_SetItem(__pyx_v_self->mapping, __pyx_v_key, __pyx_v_value) < 0)) __PYX_ERR(0, 176, __pyx_L1_error)

  /* "glycopeptidepy/_c/structure/glycan.pyx":174
 *         return self.mapping[key]
 * 
 *     def __setitem__(self, key, value):             # <<<<<<<<<<<<<<
 *         self.invalidate()
 *         self.mapping[key] = value
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.glycan.GlycosylationManager.__setitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/glycan.pyx":178
 *         self.mapping[key] = value
 * 
 *     def __delitem__(self, key):             # <<<<<<<<<<<<<<
 *         self.invalidate()
 *         del self.mapping[key]
 */

/* Python wrapper */
static int __pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_7__delitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_key); /*proto*/
static int __pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_7__delitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_key) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__delitem__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_6__delitem__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *)__pyx_v_self), ((PyObject *)__pyx_v_key));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_6__delitem__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_self, PyObject *__pyx_v_key) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__delitem__", 0);

  /* "glycopeptidepy/_c/structure/glycan.pyx":179
 * 
 *     def __delitem__(self, key):
 *         self.invalidate()             # <<<<<<<<<<<<<<
 *         del self.mapping[key]
 * 
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *)__pyx_v_self->__pyx_vtab)->invalidate(__pyx_v_self, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 179, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "glycopeptidepy/_c/structure/glycan.pyx":180
 *     def __delitem__(self, key):
 *         self.invalidate()
 *         del self.mapping[key]             # <<<<<<<<<<<<<<
 * 
 *     def __len__(self):
 */
  if (unlikely(__pyx_v_self->mapping == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 180, __pyx_L1_error)
  }
  if (unlikely(PyDict_DelItem(__pyx_v_self->mapping, __pyx_v_key) < 0)) __PYX_ERR(0, 180, __pyx_L1_error)

  /* "glycopeptidepy/_c/structure/glycan.pyx":178
 *         self.mapping[key] = value
 * 
 *     def __delitem__(self, key):             # <<<<<<<<<<<<<<
 *         self.invalidate()
 *         del self.mapping[key]
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.glycan.GlycosylationManager.__delitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/glycan.pyx":182
 *         del self.mapping[key]
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return self.get_size()
 * 
 */

/* Python wrapper */
static Py_ssize_t __pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_9__len__(PyObject *__pyx_v_self); /*proto*/
static Py_ssize_t __pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_9__len__(PyObject *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_8__len__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static Py_ssize_t __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_8__len__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__", 0);

  /* "glycopeptidepy/_c/structure/glycan.pyx":183
 * 
 *     def __len__(self):
 *         return self.get_size()             # <<<<<<<<<<<<<<
 * 
 *     cdef size_t get_size(self):
 */
  __pyx_r = ((struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *)__pyx_v_self->__pyx_vtab)->get_size(__pyx_v_self);
  goto __pyx_L0;

  /* "glycopeptidepy/_c/structure/glycan.pyx":182
 *         del self.mapping[key]
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return self.get_size()
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/glycan.pyx":185
 *         return self.get_size()
 * 
 *     cdef size_t get_size(self):             # <<<<<<<<<<<<<<
 *         return PyDict_Size(self.mapping)
 * 
 */

static size_t __pyx_f_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_get_size(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_self) {
  size_t __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_size", 0);

  /* "glycopeptidepy/_c/structure/glycan.pyx":186
 * 
 *     cdef size_t get_size(self):
 *         return PyDict_Size(self.mapping)             # <<<<<<<<<<<<<<
 * 
 *     cpdef keys(self):
 */
  __pyx_t_1 = __pyx_v_self->mapping;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = PyDict_Size(__pyx_t_1); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1L))) __PYX_ERR(0, 186, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  goto __pyx_L0;

  /* "glycopeptidepy/_c/structure/glycan.pyx":185
 *         return self.get_size()
 * 
 *     cdef size_t get_size(self):             # <<<<<<<<<<<<<<
 *         return PyDict_Size(self.mapping)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_WriteUnraisable("glycopeptidepy._c.structure.glycan.GlycosylationManager.get_size", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/glycan.pyx":188
 *         return PyDict_Size(self.mapping)
 * 
 *     cpdef keys(self):             # <<<<<<<<<<<<<<
 *         return self.mapping.keys()
 * 
 */

static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_11keys(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_f_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_keys(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_self, int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("keys", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_keys); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 188, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_11keys)) {
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 188, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "glycopeptidepy/_c/structure/glycan.pyx":189
 * 
 *     cpdef keys(self):
 *         return self.mapping.keys()             # <<<<<<<<<<<<<<
 * 
 *     cpdef values(self):
 */
  __Pyx_XDECREF(__pyx_r);
  if (unlikely(__pyx_v_self->mapping == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "keys");
    __PYX_ERR(0, 189, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyDict_Keys(__pyx_v_self->mapping); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 189, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "glycopeptidepy/_c/structure/glycan.pyx":188
 *         return PyDict_Size(self.mapping)
 * 
 *     cpdef keys(self):             # <<<<<<<<<<<<<<
 *         return self.mapping.keys()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.glycan.GlycosylationManager.keys", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_11keys(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_10keys[] = "GlycosylationManager.keys(self)";
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_11keys(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("keys (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_10keys(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_10keys(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("keys", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_keys(__pyx_v_self, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 188, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.glycan.GlycosylationManager.keys", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/glycan.pyx":191
 *         return self.mapping.keys()
 * 
 *     cpdef values(self):             # <<<<<<<<<<<<<<
 *         return self.mapping.values()
 * 
 */

static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_13values(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_f_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_values(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_self, int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("values", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_values); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 191, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_13values)) {
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 191, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "glycopeptidepy/_c/structure/glycan.pyx":192
 * 
 *     cpdef values(self):
 *         return self.mapping.values()             # <<<<<<<<<<<<<<
 * 
 *     cpdef items(self):
 */
  __Pyx_XDECREF(__pyx_r);
  if (unlikely(__pyx_v_self->mapping == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "values");
    __PYX_ERR(0, 192, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyDict_Values(__pyx_v_self->mapping); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 192, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "glycopeptidepy/_c/structure/glycan.pyx":191
 *         return self.mapping.keys()
 * 
 *     cpdef values(self):             # <<<<<<<<<<<<<<
 *         return self.mapping.values()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.glycan.GlycosylationManager.values", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_13values(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_12values[] = "GlycosylationManager.values(self)";
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_13values(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("values (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_12values(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_12values(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("values", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_values(__pyx_v_self, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 191, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.glycan.GlycosylationManager.values", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/glycan.pyx":194
 *         return self.mapping.values()
 * 
 *     cpdef items(self):             # <<<<<<<<<<<<<<
 *         return self.mapping.items()
 * 
 */

static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_15items(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_f_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_items(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_self, int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("items", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_items); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 194, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_15items)) {
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 194, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "glycopeptidepy/_c/structure/glycan.pyx":195
 * 
 *     cpdef items(self):
 *         return self.mapping.items()             # <<<<<<<<<<<<<<
 * 
 *     cpdef pop(self, key, default=None):
 */
  __Pyx_XDECREF(__pyx_r);
  if (unlikely(__pyx_v_self->mapping == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "items");
    __PYX_ERR(0, 195, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyDict_Items(__pyx_v_self->mapping); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 195, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "glycopeptidepy/_c/structure/glycan.pyx":194
 *         return self.mapping.values()
 * 
 *     cpdef items(self):             # <<<<<<<<<<<<<<
 *         return self.mapping.items()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.glycan.GlycosylationManager.items", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_15items(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_14items[] = "GlycosylationManager.items(self)";
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_15items(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("items (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_14items(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_14items(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("items", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_items(__pyx_v_self, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 194, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.glycan.GlycosylationManager.items", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/glycan.pyx":197
 *         return self.mapping.items()
 * 
 *     cpdef pop(self, key, default=None):             # <<<<<<<<<<<<<<
 *         self.invalidate()
 *         self.mapping.pop(key, default)
 */

static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_17pop(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_pop(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_self, PyObject *__pyx_v_key, int __pyx_skip_dispatch, struct __pyx_opt_args_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_pop *__pyx_optional_args) {
  PyObject *__pyx_v_default = ((PyObject *)Py_None);
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("pop", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_default = __pyx_optional_args->__pyx_default;
    }
  }
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pop); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 197, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_17pop)) {
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        __pyx_t_5 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_5 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_key, __pyx_v_default};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 197, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
          PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_key, __pyx_v_default};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 197, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_GOTREF(__pyx_t_2);
        } else
        #endif
        {
          __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 197, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          if (__pyx_t_4) {
            __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4); __pyx_t_4 = NULL;
          }
          __Pyx_INCREF(__pyx_v_key);
          __Pyx_GIVEREF(__pyx_v_key);
          PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_v_key);
          __Pyx_INCREF(__pyx_v_default);
          __Pyx_GIVEREF(__pyx_v_default);
          PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_v_default);
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_6, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 197, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "glycopeptidepy/_c/structure/glycan.pyx":198
 * 
 *     cpdef pop(self, key, default=None):
 *         self.invalidate()             # <<<<<<<<<<<<<<
 *         self.mapping.pop(key, default)
 * 
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *)__pyx_v_self->__pyx_vtab)->invalidate(__pyx_v_self, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 198, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "glycopeptidepy/_c/structure/glycan.pyx":199
 *     cpdef pop(self, key, default=None):
 *         self.invalidate()
 *         self.mapping.pop(key, default)             # <<<<<<<<<<<<<<
 * 
 *     cpdef clear(self):
 */
  if (unlikely(__pyx_v_self->mapping == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "pop");
    __PYX_ERR(0, 199, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyDict_Pop(__pyx_v_self->mapping, __pyx_v_key, __pyx_v_default); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 199, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "glycopeptidepy/_c/structure/glycan.pyx":197
 *         return self.mapping.items()
 * 
 *     cpdef pop(self, key, default=None):             # <<<<<<<<<<<<<<
 *         self.invalidate()
 *         self.mapping.pop(key, default)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.glycan.GlycosylationManager.pop", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_17pop(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_16pop[] = "GlycosylationManager.pop(self, key, default=None)";
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_17pop(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_key = 0;
  PyObject *__pyx_v_default = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("pop (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_key,&__pyx_n_s_default,0};
    PyObject* values[2] = {0,0};
    values[1] = ((PyObject *)Py_None);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_key)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_default);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "pop") < 0)) __PYX_ERR(0, 197, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_key = values[0];
    __pyx_v_default = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("pop", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 197, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("glycopeptidepy._c.structure.glycan.GlycosylationManager.pop", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_16pop(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *)__pyx_v_self), __pyx_v_key, __pyx_v_default);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_16pop(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_self, PyObject *__pyx_v_key, PyObject *__pyx_v_default) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  struct __pyx_opt_args_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_pop __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("pop", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2.__pyx_n = 1;
  __pyx_t_2.__pyx_default = __pyx_v_default;
  __pyx_t_1 = __pyx_vtabptr_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager->pop(__pyx_v_self, __pyx_v_key, 1, &__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 197, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.glycan.GlycosylationManager.pop", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/glycan.pyx":201
 *         self.mapping.pop(key, default)
 * 
 *     cpdef clear(self):             # <<<<<<<<<<<<<<
 *         self.invalidate()
 *         self.set_aggregate(None)
 */

static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_19clear(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_f_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_clear(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_self, int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("clear", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_clear); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 201, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_19clear)) {
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 201, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "glycopeptidepy/_c/structure/glycan.pyx":202
 * 
 *     cpdef clear(self):
 *         self.invalidate()             # <<<<<<<<<<<<<<
 *         self.set_aggregate(None)
 *         self.mapping.clear()
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *)__pyx_v_self->__pyx_vtab)->invalidate(__pyx_v_self, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 202, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "glycopeptidepy/_c/structure/glycan.pyx":203
 *     cpdef clear(self):
 *         self.invalidate()
 *         self.set_aggregate(None)             # <<<<<<<<<<<<<<
 *         self.mapping.clear()
 * 
 */
  (void)(((struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *)__pyx_v_self->__pyx_vtab)->set_aggregate(__pyx_v_self, Py_None));

  /* "glycopeptidepy/_c/structure/glycan.pyx":204
 *         self.invalidate()
 *         self.set_aggregate(None)
 *         self.mapping.clear()             # <<<<<<<<<<<<<<
 * 
 *     cpdef update(self, other):
 */
  if (unlikely(__pyx_v_self->mapping == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "clear");
    __PYX_ERR(0, 204, __pyx_L1_error)
  }
  __pyx_t_5 = __Pyx_PyDict_Clear(__pyx_v_self->mapping); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 204, __pyx_L1_error)

  /* "glycopeptidepy/_c/structure/glycan.pyx":201
 *         self.mapping.pop(key, default)
 * 
 *     cpdef clear(self):             # <<<<<<<<<<<<<<
 *         self.invalidate()
 *         self.set_aggregate(None)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.glycan.GlycosylationManager.clear", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_19clear(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_18clear[] = "GlycosylationManager.clear(self)";
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_19clear(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("clear (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_18clear(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_18clear(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("clear", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_clear(__pyx_v_self, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 201, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.glycan.GlycosylationManager.clear", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/glycan.pyx":206
 *         self.mapping.clear()
 * 
 *     cpdef update(self, other):             # <<<<<<<<<<<<<<
 *         self.invalidate()
 *         if isinstance(other, GlycosylationManager):
 */

static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_21update(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static PyObject *__pyx_f_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_update(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_self, PyObject *__pyx_v_other, int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("update", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_update); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 206, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_21update)) {
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_v_other) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_other);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 206, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "glycopeptidepy/_c/structure/glycan.pyx":207
 * 
 *     cpdef update(self, other):
 *         self.invalidate()             # <<<<<<<<<<<<<<
 *         if isinstance(other, GlycosylationManager):
 *             self.mapping.update(other.mapping)
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *)__pyx_v_self->__pyx_vtab)->invalidate(__pyx_v_self, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 207, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "glycopeptidepy/_c/structure/glycan.pyx":208
 *     cpdef update(self, other):
 *         self.invalidate()
 *         if isinstance(other, GlycosylationManager):             # <<<<<<<<<<<<<<
 *             self.mapping.update(other.mapping)
 *         else:
 */
  __pyx_t_5 = __Pyx_TypeCheck(__pyx_v_other, __pyx_ptype_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager); 
  __pyx_t_6 = (__pyx_t_5 != 0);
  if (__pyx_t_6) {

    /* "glycopeptidepy/_c/structure/glycan.pyx":209
 *         self.invalidate()
 *         if isinstance(other, GlycosylationManager):
 *             self.mapping.update(other.mapping)             # <<<<<<<<<<<<<<
 *         else:
 *             self.mapping.update(other)
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->mapping, __pyx_n_s_update); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 209, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_other, __pyx_n_s_mapping); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 209, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 209, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "glycopeptidepy/_c/structure/glycan.pyx":208
 *     cpdef update(self, other):
 *         self.invalidate()
 *         if isinstance(other, GlycosylationManager):             # <<<<<<<<<<<<<<
 *             self.mapping.update(other.mapping)
 *         else:
 */
    goto __pyx_L3;
  }

  /* "glycopeptidepy/_c/structure/glycan.pyx":211
 *             self.mapping.update(other.mapping)
 *         else:
 *             self.mapping.update(other)             # <<<<<<<<<<<<<<
 * 
 *     cpdef GlycosylationManager copy(self):
 */
  /*else*/ {
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->mapping, __pyx_n_s_update); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 211, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_v_other) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_other);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 211, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }
  __pyx_L3:;

  /* "glycopeptidepy/_c/structure/glycan.pyx":206
 *         self.mapping.clear()
 * 
 *     cpdef update(self, other):             # <<<<<<<<<<<<<<
 *         self.invalidate()
 *         if isinstance(other, GlycosylationManager):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.glycan.GlycosylationManager.update", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_21update(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static char __pyx_doc_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_20update[] = "GlycosylationManager.update(self, other)";
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_21update(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("update (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_20update(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *)__pyx_v_self), ((PyObject *)__pyx_v_other));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_20update(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("update", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_update(__pyx_v_self, __pyx_v_other, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 206, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.glycan.GlycosylationManager.update", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/glycan.pyx":213
 *             self.mapping.update(other)
 * 
 *     cpdef GlycosylationManager copy(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             GlycosylationManager inst
 */

static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_23copy(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_f_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_copy(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_self, int __pyx_skip_dispatch) {
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_inst = 0;
  PyObject *__pyx_v_aggregate = 0;
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("copy", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_copy); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 213, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_23copy)) {
        __Pyx_XDECREF(((PyObject *)__pyx_r));
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 213, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager))))) __PYX_ERR(0, 213, __pyx_L1_error)
        __pyx_r = ((struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "glycopeptidepy/_c/structure/glycan.pyx":217
 *             GlycosylationManager inst
 *             object aggregate
 *         aggregate = self.get_aggregate()             # <<<<<<<<<<<<<<
 * 
 *         inst = GlycosylationManager._create(
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *)__pyx_v_self->__pyx_vtab)->get_aggregate(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 217, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_aggregate = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "glycopeptidepy/_c/structure/glycan.pyx":220
 * 
 *         inst = GlycosylationManager._create(
 *             self.parent,             # <<<<<<<<<<<<<<
 *             aggregate.clone() if aggregate is not None else None)
 *         inst.update(self)
 */
  if (!(likely(((__pyx_v_self->parent) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_self->parent, __pyx_ptype_14glycopeptidepy_2_c_9structure_4base_PeptideSequenceBase))))) __PYX_ERR(0, 220, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_self->parent;
  __Pyx_INCREF(__pyx_t_1);

  /* "glycopeptidepy/_c/structure/glycan.pyx":221
 *         inst = GlycosylationManager._create(
 *             self.parent,
 *             aggregate.clone() if aggregate is not None else None)             # <<<<<<<<<<<<<<
 *         inst.update(self)
 *         return inst
 */
  __pyx_t_5 = (__pyx_v_aggregate != Py_None);
  if ((__pyx_t_5 != 0)) {
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_aggregate, __pyx_n_s_clone); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 221, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_3 = (__pyx_t_6) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_6) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 221, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_2 = __pyx_t_3;
    __pyx_t_3 = 0;
  } else {
    __Pyx_INCREF(Py_None);
    __pyx_t_2 = Py_None;
  }

  /* "glycopeptidepy/_c/structure/glycan.pyx":219
 *         aggregate = self.get_aggregate()
 * 
 *         inst = GlycosylationManager._create(             # <<<<<<<<<<<<<<
 *             self.parent,
 *             aggregate.clone() if aggregate is not None else None)
 */
  __pyx_t_3 = ((PyObject *)__pyx_f_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager__create(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_PeptideSequenceBase *)__pyx_t_1), __pyx_t_2)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 219, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_inst = ((struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "glycopeptidepy/_c/structure/glycan.pyx":222
 *             self.parent,
 *             aggregate.clone() if aggregate is not None else None)
 *         inst.update(self)             # <<<<<<<<<<<<<<
 *         return inst
 * 
 */
  __pyx_t_3 = ((struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *)__pyx_v_inst->__pyx_vtab)->update(__pyx_v_inst, ((PyObject *)__pyx_v_self), 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 222, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "glycopeptidepy/_c/structure/glycan.pyx":223
 *             aggregate.clone() if aggregate is not None else None)
 *         inst.update(self)
 *         return inst             # <<<<<<<<<<<<<<
 * 
 *     cpdef GlycosylationManager clone(self):
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __Pyx_INCREF(((PyObject *)__pyx_v_inst));
  __pyx_r = __pyx_v_inst;
  goto __pyx_L0;

  /* "glycopeptidepy/_c/structure/glycan.pyx":213
 *             self.mapping.update(other)
 * 
 *     cpdef GlycosylationManager copy(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             GlycosylationManager inst
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.glycan.GlycosylationManager.copy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_inst);
  __Pyx_XDECREF(__pyx_v_aggregate);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_23copy(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_22copy[] = "GlycosylationManager.copy(self) -> GlycosylationManager";
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_23copy(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("copy (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_22copy(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_22copy(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("copy", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_copy(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 213, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.glycan.GlycosylationManager.copy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/glycan.pyx":225
 *         return inst
 * 
 *     cpdef GlycosylationManager clone(self):             # <<<<<<<<<<<<<<
 *         return self.copy()
 * 
 */

static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_25clone(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_f_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_clone(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_self, int __pyx_skip_dispatch) {
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("clone", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_clone); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 225, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_25clone)) {
        __Pyx_XDECREF(((PyObject *)__pyx_r));
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 225, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager))))) __PYX_ERR(0, 225, __pyx_L1_error)
        __pyx_r = ((struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "glycopeptidepy/_c/structure/glycan.pyx":226
 * 
 *     cpdef GlycosylationManager clone(self):
 *         return self.copy()             # <<<<<<<<<<<<<<
 * 
 *     cpdef object _patch_aggregate(self):
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *)__pyx_v_self->__pyx_vtab)->copy(__pyx_v_self, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 226, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = ((struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "glycopeptidepy/_c/structure/glycan.pyx":225
 *         return inst
 * 
 *     cpdef GlycosylationManager clone(self):             # <<<<<<<<<<<<<<
 *         return self.copy()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.glycan.GlycosylationManager.clone", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_25clone(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_24clone[] = "GlycosylationManager.clone(self) -> GlycosylationManager";
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_25clone(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("clone (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_24clone(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_24clone(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("clone", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_clone(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 225, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.glycan.GlycosylationManager.clone", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/glycan.pyx":228
 *         return self.copy()
 * 
 *     cpdef object _patch_aggregate(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             object aggregate
 */

static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_27_patch_aggregate(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_f_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager__patch_aggregate(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_self, int __pyx_skip_dispatch) {
  PyObject *__pyx_v_aggregate = 0;
  struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *__pyx_v_offset = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_patch_aggregate", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_patch_aggregate); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 228, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_27_patch_aggregate)) {
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 228, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "glycopeptidepy/_c/structure/glycan.pyx":232
 *             object aggregate
 *             CComposition offset
 *         aggregate = self.get_aggregate()             # <<<<<<<<<<<<<<
 *         if isinstance(aggregate, GlycanCompositionWithOffsetProxy):
 *             offset = (<GlycanCompositionWithOffsetProxy>aggregate).composition_offset
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *)__pyx_v_self->__pyx_vtab)->get_aggregate(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 232, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_aggregate = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "glycopeptidepy/_c/structure/glycan.pyx":233
 *             CComposition offset
 *         aggregate = self.get_aggregate()
 *         if isinstance(aggregate, GlycanCompositionWithOffsetProxy):             # <<<<<<<<<<<<<<
 *             offset = (<GlycanCompositionWithOffsetProxy>aggregate).composition_offset
 *         elif isinstance(aggregate, _CompositionBase):
 */
  __pyx_t_5 = __Pyx_TypeCheck(__pyx_v_aggregate, __pyx_ptype_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionWithOffsetProxy); 
  __pyx_t_6 = (__pyx_t_5 != 0);
  if (__pyx_t_6) {

    /* "glycopeptidepy/_c/structure/glycan.pyx":234
 *         aggregate = self.get_aggregate()
 *         if isinstance(aggregate, GlycanCompositionWithOffsetProxy):
 *             offset = (<GlycanCompositionWithOffsetProxy>aggregate).composition_offset             # <<<<<<<<<<<<<<
 *         elif isinstance(aggregate, _CompositionBase):
 *             offset = (<_CompositionBase>aggregate)._composition_offset
 */
    __pyx_t_1 = ((PyObject *)((struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionWithOffsetProxy *)__pyx_v_aggregate)->composition_offset);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v_offset = ((struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *)__pyx_t_1);
    __pyx_t_1 = 0;

    /* "glycopeptidepy/_c/structure/glycan.pyx":233
 *             CComposition offset
 *         aggregate = self.get_aggregate()
 *         if isinstance(aggregate, GlycanCompositionWithOffsetProxy):             # <<<<<<<<<<<<<<
 *             offset = (<GlycanCompositionWithOffsetProxy>aggregate).composition_offset
 *         elif isinstance(aggregate, _CompositionBase):
 */
    goto __pyx_L3;
  }

  /* "glycopeptidepy/_c/structure/glycan.pyx":235
 *         if isinstance(aggregate, GlycanCompositionWithOffsetProxy):
 *             offset = (<GlycanCompositionWithOffsetProxy>aggregate).composition_offset
 *         elif isinstance(aggregate, _CompositionBase):             # <<<<<<<<<<<<<<
 *             offset = (<_CompositionBase>aggregate)._composition_offset
 *         else:
 */
  __pyx_t_6 = __Pyx_TypeCheck(__pyx_v_aggregate, __pyx_ptype_5glypy_2_c_9structure_18glycan_composition__CompositionBase); 
  __pyx_t_5 = (__pyx_t_6 != 0);
  if (__pyx_t_5) {

    /* "glycopeptidepy/_c/structure/glycan.pyx":236
 *             offset = (<GlycanCompositionWithOffsetProxy>aggregate).composition_offset
 *         elif isinstance(aggregate, _CompositionBase):
 *             offset = (<_CompositionBase>aggregate)._composition_offset             # <<<<<<<<<<<<<<
 *         else:
 *             offset = <CComposition>(aggregate).composition_offset
 */
    __pyx_t_1 = ((PyObject *)((struct __pyx_obj_5glypy_2_c_9structure_18glycan_composition__CompositionBase *)__pyx_v_aggregate)->_composition_offset);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v_offset = ((struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *)__pyx_t_1);
    __pyx_t_1 = 0;

    /* "glycopeptidepy/_c/structure/glycan.pyx":235
 *         if isinstance(aggregate, GlycanCompositionWithOffsetProxy):
 *             offset = (<GlycanCompositionWithOffsetProxy>aggregate).composition_offset
 *         elif isinstance(aggregate, _CompositionBase):             # <<<<<<<<<<<<<<
 *             offset = (<_CompositionBase>aggregate)._composition_offset
 *         else:
 */
    goto __pyx_L3;
  }

  /* "glycopeptidepy/_c/structure/glycan.pyx":238
 *             offset = (<_CompositionBase>aggregate)._composition_offset
 *         else:
 *             offset = <CComposition>(aggregate).composition_offset             # <<<<<<<<<<<<<<
 *         offset.subtract_from(WATER_OFFSET)
 * 
 */
  /*else*/ {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_aggregate, __pyx_n_s_composition_offset); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 238, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __pyx_t_1;
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_offset = ((struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *)__pyx_t_2);
    __pyx_t_2 = 0;
  }
  __pyx_L3:;

  /* "glycopeptidepy/_c/structure/glycan.pyx":239
 *         else:
 *             offset = <CComposition>(aggregate).composition_offset
 *         offset.subtract_from(WATER_OFFSET)             # <<<<<<<<<<<<<<
 * 
 *     cpdef object invalidate(self):
 */
  __pyx_t_2 = ((PyObject *)__pyx_v_14glycopeptidepy_2_c_9structure_6glycan_WATER_OFFSET);
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_5glypy_11composition_12ccomposition_CComposition *)__pyx_v_offset->__pyx_vtab)->subtract_from(__pyx_v_offset, __pyx_t_2)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 239, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "glycopeptidepy/_c/structure/glycan.pyx":228
 *         return self.copy()
 * 
 *     cpdef object _patch_aggregate(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             object aggregate
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.glycan.GlycosylationManager._patch_aggregate", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_aggregate);
  __Pyx_XDECREF((PyObject *)__pyx_v_offset);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_27_patch_aggregate(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_26_patch_aggregate[] = "GlycosylationManager._patch_aggregate(self)";
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_27_patch_aggregate(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_patch_aggregate (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_26_patch_aggregate(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_26_patch_aggregate(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_patch_aggregate", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager__patch_aggregate(__pyx_v_self, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 228, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.glycan.GlycosylationManager._patch_aggregate", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/glycan.pyx":241
 *         offset.subtract_from(WATER_OFFSET)
 * 
 *     cpdef object invalidate(self):             # <<<<<<<<<<<<<<
 *         self._proxy = None
 *         self._type_track = None
 */

static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_29invalidate(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_f_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_invalidate(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_self, int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("invalidate", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_invalidate); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 241, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_29invalidate)) {
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 241, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "glycopeptidepy/_c/structure/glycan.pyx":242
 * 
 *     cpdef object invalidate(self):
 *         self._proxy = None             # <<<<<<<<<<<<<<
 *         self._type_track = None
 *         self._total_glycosylation_size = -1
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->_proxy);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->_proxy));
  __pyx_v_self->_proxy = ((struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *)Py_None);

  /* "glycopeptidepy/_c/structure/glycan.pyx":243
 *     cpdef object invalidate(self):
 *         self._proxy = None
 *         self._type_track = None             # <<<<<<<<<<<<<<
 *         self._total_glycosylation_size = -1
 * 
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->_type_track);
  __Pyx_DECREF(__pyx_v_self->_type_track);
  __pyx_v_self->_type_track = ((PyObject*)Py_None);

  /* "glycopeptidepy/_c/structure/glycan.pyx":244
 *         self._proxy = None
 *         self._type_track = None
 *         self._total_glycosylation_size = -1             # <<<<<<<<<<<<<<
 * 
 *     cdef object get_aggregate(self):
 */
  __pyx_v_self->_total_glycosylation_size = -1;

  /* "glycopeptidepy/_c/structure/glycan.pyx":241
 *         offset.subtract_from(WATER_OFFSET)
 * 
 *     cpdef object invalidate(self):             # <<<<<<<<<<<<<<
 *         self._proxy = None
 *         self._type_track = None
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.glycan.GlycosylationManager.invalidate", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_29invalidate(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_28invalidate[] = "GlycosylationManager.invalidate(self)";
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_29invalidate(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("invalidate (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_28invalidate(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_28invalidate(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("invalidate", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_invalidate(__pyx_v_self, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 241, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.glycan.GlycosylationManager.invalidate", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/glycan.pyx":246
 *         self._total_glycosylation_size = -1
 * 
 *     cdef object get_aggregate(self):             # <<<<<<<<<<<<<<
 *         return self._aggregate
 * 
 */

static PyObject *__pyx_f_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_get_aggregate(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_aggregate", 0);

  /* "glycopeptidepy/_c/structure/glycan.pyx":247
 * 
 *     cdef object get_aggregate(self):
 *         return self._aggregate             # <<<<<<<<<<<<<<
 * 
 *     cdef int set_aggregate(self, object value):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->_aggregate);
  __pyx_r = __pyx_v_self->_aggregate;
  goto __pyx_L0;

  /* "glycopeptidepy/_c/structure/glycan.pyx":246
 *         self._total_glycosylation_size = -1
 * 
 *     cdef object get_aggregate(self):             # <<<<<<<<<<<<<<
 *         return self._aggregate
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/glycan.pyx":249
 *         return self._aggregate
 * 
 *     cdef int set_aggregate(self, object value):             # <<<<<<<<<<<<<<
 *         self._aggregate = value
 *         if self._aggregate is not None:
 */

static int __pyx_f_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_set_aggregate(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("set_aggregate", 0);

  /* "glycopeptidepy/_c/structure/glycan.pyx":250
 * 
 *     cdef int set_aggregate(self, object value):
 *         self._aggregate = value             # <<<<<<<<<<<<<<
 *         if self._aggregate is not None:
 *             self._patch_aggregate()
 */
  __Pyx_INCREF(__pyx_v_value);
  __Pyx_GIVEREF(__pyx_v_value);
  __Pyx_GOTREF(__pyx_v_self->_aggregate);
  __Pyx_DECREF(__pyx_v_self->_aggregate);
  __pyx_v_self->_aggregate = __pyx_v_value;

  /* "glycopeptidepy/_c/structure/glycan.pyx":251
 *     cdef int set_aggregate(self, object value):
 *         self._aggregate = value
 *         if self._aggregate is not None:             # <<<<<<<<<<<<<<
 *             self._patch_aggregate()
 *         self.invalidate()
 */
  __pyx_t_1 = (__pyx_v_self->_aggregate != Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "glycopeptidepy/_c/structure/glycan.pyx":252
 *         self._aggregate = value
 *         if self._aggregate is not None:
 *             self._patch_aggregate()             # <<<<<<<<<<<<<<
 *         self.invalidate()
 *         return 0
 */
    __pyx_t_3 = ((struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *)__pyx_v_self->__pyx_vtab)->_patch_aggregate(__pyx_v_self, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 252, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "glycopeptidepy/_c/structure/glycan.pyx":251
 *     cdef int set_aggregate(self, object value):
 *         self._aggregate = value
 *         if self._aggregate is not None:             # <<<<<<<<<<<<<<
 *             self._patch_aggregate()
 *         self.invalidate()
 */
  }

  /* "glycopeptidepy/_c/structure/glycan.pyx":253
 *         if self._aggregate is not None:
 *             self._patch_aggregate()
 *         self.invalidate()             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
  __pyx_t_3 = ((struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *)__pyx_v_self->__pyx_vtab)->invalidate(__pyx_v_self, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 253, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "glycopeptidepy/_c/structure/glycan.pyx":254
 *             self._patch_aggregate()
 *         self.invalidate()
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "glycopeptidepy/_c/structure/glycan.pyx":249
 *         return self._aggregate
 * 
 *     cdef int set_aggregate(self, object value):             # <<<<<<<<<<<<<<
 *         self._aggregate = value
 *         if self._aggregate is not None:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_WriteUnraisable("glycopeptidepy._c.structure.glycan.GlycosylationManager.set_aggregate", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/glycan.pyx":257
 * 
 *     @property
 *     def aggregate(self):             # <<<<<<<<<<<<<<
 *         return self.get_aggregate()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_9aggregate_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_9aggregate_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_9aggregate___get__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_9aggregate___get__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "glycopeptidepy/_c/structure/glycan.pyx":258
 *     @property
 *     def aggregate(self):
 *         return self.get_aggregate()             # <<<<<<<<<<<<<<
 * 
 *     @aggregate.setter
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *)__pyx_v_self->__pyx_vtab)->get_aggregate(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 258, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "glycopeptidepy/_c/structure/glycan.pyx":257
 * 
 *     @property
 *     def aggregate(self):             # <<<<<<<<<<<<<<
 *         return self.get_aggregate()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.glycan.GlycosylationManager.aggregate.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/glycan.pyx":261
 * 
 *     @aggregate.setter
 *     def aggregate(self, value):             # <<<<<<<<<<<<<<
 *         self.set_aggregate(value)
 * 
 */

/* Python wrapper */
static int __pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_9aggregate_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_9aggregate_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_9aggregate_2__set__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_9aggregate_2__set__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 0);

  /* "glycopeptidepy/_c/structure/glycan.pyx":262
 *     @aggregate.setter
 *     def aggregate(self, value):
 *         self.set_aggregate(value)             # <<<<<<<<<<<<<<
 * 
 *     cdef dict get_glycosylation_types(self):
 */
  (void)(((struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *)__pyx_v_self->__pyx_vtab)->set_aggregate(__pyx_v_self, __pyx_v_value));

  /* "glycopeptidepy/_c/structure/glycan.pyx":261
 * 
 *     @aggregate.setter
 *     def aggregate(self, value):             # <<<<<<<<<<<<<<
 *         self.set_aggregate(value)
 * 
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/glycan.pyx":264
 *         self.set_aggregate(value)
 * 
 *     cdef dict get_glycosylation_types(self):             # <<<<<<<<<<<<<<
 *         if self._type_track is None:
 *             self._type_track = self._update_track()
 */

static PyObject *__pyx_f_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_get_glycosylation_types(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_glycosylation_types", 0);

  /* "glycopeptidepy/_c/structure/glycan.pyx":265
 * 
 *     cdef dict get_glycosylation_types(self):
 *         if self._type_track is None:             # <<<<<<<<<<<<<<
 *             self._type_track = self._update_track()
 *         return self._type_track
 */
  __pyx_t_1 = (__pyx_v_self->_type_track == ((PyObject*)Py_None));
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "glycopeptidepy/_c/structure/glycan.pyx":266
 *     cdef dict get_glycosylation_types(self):
 *         if self._type_track is None:
 *             self._type_track = self._update_track()             # <<<<<<<<<<<<<<
 *         return self._type_track
 * 
 */
    __pyx_t_3 = ((struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *)__pyx_v_self->__pyx_vtab)->_update_track(__pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 266, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_3);
    __Pyx_GOTREF(__pyx_v_self->_type_track);
    __Pyx_DECREF(__pyx_v_self->_type_track);
    __pyx_v_self->_type_track = ((PyObject*)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "glycopeptidepy/_c/structure/glycan.pyx":265
 * 
 *     cdef dict get_glycosylation_types(self):
 *         if self._type_track is None:             # <<<<<<<<<<<<<<
 *             self._type_track = self._update_track()
 *         return self._type_track
 */
  }

  /* "glycopeptidepy/_c/structure/glycan.pyx":267
 *         if self._type_track is None:
 *             self._type_track = self._update_track()
 *         return self._type_track             # <<<<<<<<<<<<<<
 * 
 *     cdef dict _update_track(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->_type_track);
  __pyx_r = __pyx_v_self->_type_track;
  goto __pyx_L0;

  /* "glycopeptidepy/_c/structure/glycan.pyx":264
 *         self.set_aggregate(value)
 * 
 *     cdef dict get_glycosylation_types(self):             # <<<<<<<<<<<<<<
 *         if self._type_track is None:
 *             self._type_track = self._update_track()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.glycan.GlycosylationManager.get_glycosylation_types", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/glycan.pyx":269
 *         return self._type_track
 * 
 *     cdef dict _update_track(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             dict track
 */

static PyObject *__pyx_f_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager__update_track(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_self) {
  PyObject *__pyx_v_track = 0;
  Py_ssize_t __pyx_v_pos;
  PyObject *__pyx_v_pkey;
  PyObject *__pyx_v_pval;
  PyObject *__pyx_v_tmp;
  PyObject *__pyx_v_bucket = 0;
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_12modification_4rule_GlycosylationBase *__pyx_v_glycosylation = 0;
  struct __pyx_obj_5glypy_5utils_5cenum_EnumValue *__pyx_v_glycosylation_type = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_update_track", 0);

  /* "glycopeptidepy/_c/structure/glycan.pyx":279
 *             GlycosylationBase glycosylation
 * 
 *         track = dict()             # <<<<<<<<<<<<<<
 *         pos = 0
 *         while PyDict_Next(self.mapping, &pos, &pkey, &pval):
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 279, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_track = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "glycopeptidepy/_c/structure/glycan.pyx":280
 * 
 *         track = dict()
 *         pos = 0             # <<<<<<<<<<<<<<
 *         while PyDict_Next(self.mapping, &pos, &pkey, &pval):
 *             glycosylation = <GlycosylationBase>(<ModificationInstanceBase>pval).rule
 */
  __pyx_v_pos = 0;

  /* "glycopeptidepy/_c/structure/glycan.pyx":281
 *         track = dict()
 *         pos = 0
 *         while PyDict_Next(self.mapping, &pos, &pkey, &pval):             # <<<<<<<<<<<<<<
 *             glycosylation = <GlycosylationBase>(<ModificationInstanceBase>pval).rule
 *             glycosylation_type = glycosylation.get_glycosylation_type()
 */
  while (1) {
    __pyx_t_1 = __pyx_v_self->mapping;
    __Pyx_INCREF(__pyx_t_1);
    __pyx_t_2 = (PyDict_Next(__pyx_t_1, (&__pyx_v_pos), (&__pyx_v_pkey), (&__pyx_v_pval)) != 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (!__pyx_t_2) break;

    /* "glycopeptidepy/_c/structure/glycan.pyx":282
 *         pos = 0
 *         while PyDict_Next(self.mapping, &pos, &pkey, &pval):
 *             glycosylation = <GlycosylationBase>(<ModificationInstanceBase>pval).rule             # <<<<<<<<<<<<<<
 *             glycosylation_type = glycosylation.get_glycosylation_type()
 *             tmp = PyDict_GetItem(track, glycosylation_type)
 */
    __pyx_t_1 = ((PyObject *)((struct __pyx_obj_14glycopeptidepy_2_c_9structure_12modification_12modification_ModificationInstanceBase *)__pyx_v_pval)->rule);
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_glycosylation, ((struct __pyx_obj_14glycopeptidepy_2_c_9structure_12modification_4rule_GlycosylationBase *)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "glycopeptidepy/_c/structure/glycan.pyx":283
 *         while PyDict_Next(self.mapping, &pos, &pkey, &pval):
 *             glycosylation = <GlycosylationBase>(<ModificationInstanceBase>pval).rule
 *             glycosylation_type = glycosylation.get_glycosylation_type()             # <<<<<<<<<<<<<<
 *             tmp = PyDict_GetItem(track, glycosylation_type)
 *             if tmp == NULL:
 */
    __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_12modification_4rule_GlycosylationBase *)__pyx_v_glycosylation->__pyx_base.__pyx_base.__pyx_vtab)->get_glycosylation_type(__pyx_v_glycosylation)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 283, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_glycosylation_type, ((struct __pyx_obj_5glypy_5utils_5cenum_EnumValue *)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "glycopeptidepy/_c/structure/glycan.pyx":284
 *             glycosylation = <GlycosylationBase>(<ModificationInstanceBase>pval).rule
 *             glycosylation_type = glycosylation.get_glycosylation_type()
 *             tmp = PyDict_GetItem(track, glycosylation_type)             # <<<<<<<<<<<<<<
 *             if tmp == NULL:
 *                 bucket = []
 */
    __pyx_v_tmp = PyDict_GetItem(__pyx_v_track, ((PyObject *)__pyx_v_glycosylation_type));

    /* "glycopeptidepy/_c/structure/glycan.pyx":285
 *             glycosylation_type = glycosylation.get_glycosylation_type()
 *             tmp = PyDict_GetItem(track, glycosylation_type)
 *             if tmp == NULL:             # <<<<<<<<<<<<<<
 *                 bucket = []
 *                 PyDict_SetItem(track, glycosylation_type, bucket)
 */
    __pyx_t_2 = ((__pyx_v_tmp == NULL) != 0);
    if (__pyx_t_2) {

      /* "glycopeptidepy/_c/structure/glycan.pyx":286
 *             tmp = PyDict_GetItem(track, glycosylation_type)
 *             if tmp == NULL:
 *                 bucket = []             # <<<<<<<<<<<<<<
 *                 PyDict_SetItem(track, glycosylation_type, bucket)
 *             else:
 */
      __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 286, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_XDECREF_SET(__pyx_v_bucket, ((PyObject*)__pyx_t_1));
      __pyx_t_1 = 0;

      /* "glycopeptidepy/_c/structure/glycan.pyx":287
 *             if tmp == NULL:
 *                 bucket = []
 *                 PyDict_SetItem(track, glycosylation_type, bucket)             # <<<<<<<<<<<<<<
 *             else:
 *                 bucket = <list>tmp
 */
      __pyx_t_3 = PyDict_SetItem(__pyx_v_track, ((PyObject *)__pyx_v_glycosylation_type), __pyx_v_bucket); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(0, 287, __pyx_L1_error)

      /* "glycopeptidepy/_c/structure/glycan.pyx":285
 *             glycosylation_type = glycosylation.get_glycosylation_type()
 *             tmp = PyDict_GetItem(track, glycosylation_type)
 *             if tmp == NULL:             # <<<<<<<<<<<<<<
 *                 bucket = []
 *                 PyDict_SetItem(track, glycosylation_type, bucket)
 */
      goto __pyx_L5;
    }

    /* "glycopeptidepy/_c/structure/glycan.pyx":289
 *                 PyDict_SetItem(track, glycosylation_type, bucket)
 *             else:
 *                 bucket = <list>tmp             # <<<<<<<<<<<<<<
 *             PyList_Append(bucket, (<object>pkey, <object>pval))
 *         return  track
 */
    /*else*/ {
      __pyx_t_1 = ((PyObject *)__pyx_v_tmp);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_XDECREF_SET(__pyx_v_bucket, ((PyObject*)__pyx_t_1));
      __pyx_t_1 = 0;
    }
    __pyx_L5:;

    /* "glycopeptidepy/_c/structure/glycan.pyx":290
 *             else:
 *                 bucket = <list>tmp
 *             PyList_Append(bucket, (<object>pkey, <object>pval))             # <<<<<<<<<<<<<<
 *         return  track
 * 
 */
    __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 290, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)__pyx_v_pkey));
    __Pyx_GIVEREF(((PyObject *)__pyx_v_pkey));
    PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)__pyx_v_pkey));
    __Pyx_INCREF(((PyObject *)__pyx_v_pval));
    __Pyx_GIVEREF(((PyObject *)__pyx_v_pval));
    PyTuple_SET_ITEM(__pyx_t_1, 1, ((PyObject *)__pyx_v_pval));
    __pyx_t_3 = PyList_Append(__pyx_v_bucket, __pyx_t_1); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(0, 290, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "glycopeptidepy/_c/structure/glycan.pyx":291
 *                 bucket = <list>tmp
 *             PyList_Append(bucket, (<object>pkey, <object>pval))
 *         return  track             # <<<<<<<<<<<<<<
 * 
 *     cpdef CComposition total_composition(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_track);
  __pyx_r = __pyx_v_track;
  goto __pyx_L0;

  /* "glycopeptidepy/_c/structure/glycan.pyx":269
 *         return self._type_track
 * 
 *     cdef dict _update_track(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             dict track
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.glycan.GlycosylationManager._update_track", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_track);
  __Pyx_XDECREF(__pyx_v_bucket);
  __Pyx_XDECREF((PyObject *)__pyx_v_glycosylation);
  __Pyx_XDECREF((PyObject *)__pyx_v_glycosylation_type);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/glycan.pyx":293
 *         return  track
 * 
 *     cpdef CComposition total_composition(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             CComposition total
 */

static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_31total_composition(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *__pyx_f_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_total_composition(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_self, int __pyx_skip_dispatch) {
  struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *__pyx_v_total = 0;
  int __pyx_v_has_aggregate;
  PyObject *__pyx_v_pkey;
  PyObject *__pyx_v_pval;
  Py_ssize_t __pyx_v_pos;
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_12modification_12modification_ModificationInstanceBase *__pyx_v_value = 0;
  PyObject *__pyx_v_aggregate = 0;
  struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("total_composition", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_total_composition); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 293, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_31total_composition)) {
        __Pyx_XDECREF(((PyObject *)__pyx_r));
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 293, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_5glypy_11composition_12ccomposition_CComposition))))) __PYX_ERR(0, 293, __pyx_L1_error)
        __pyx_r = ((struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "glycopeptidepy/_c/structure/glycan.pyx":302
 *             ModificationInstanceBase value
 *             object aggregate
 *         aggregate = self.get_aggregate()             # <<<<<<<<<<<<<<
 *         total = CComposition._create(None)
 *         has_aggregate = aggregate is not None
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *)__pyx_v_self->__pyx_vtab)->get_aggregate(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 302, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_aggregate = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "glycopeptidepy/_c/structure/glycan.pyx":303
 *             object aggregate
 *         aggregate = self.get_aggregate()
 *         total = CComposition._create(None)             # <<<<<<<<<<<<<<
 *         has_aggregate = aggregate is not None
 *         pos = 0
 */
  __pyx_t_1 = ((PyObject *)__pyx_vtabptr_5glypy_11composition_12ccomposition_CComposition->_create(((struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *)Py_None))); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 303, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_total = ((struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "glycopeptidepy/_c/structure/glycan.pyx":304
 *         aggregate = self.get_aggregate()
 *         total = CComposition._create(None)
 *         has_aggregate = aggregate is not None             # <<<<<<<<<<<<<<
 *         pos = 0
 *         while PyDict_Next(self.mapping, &pos, &pkey, &pval):
 */
  __pyx_t_5 = (__pyx_v_aggregate != Py_None);
  __pyx_v_has_aggregate = __pyx_t_5;

  /* "glycopeptidepy/_c/structure/glycan.pyx":305
 *         total = CComposition._create(None)
 *         has_aggregate = aggregate is not None
 *         pos = 0             # <<<<<<<<<<<<<<
 *         while PyDict_Next(self.mapping, &pos, &pkey, &pval):
 *             value = <ModificationInstanceBase>pval
 */
  __pyx_v_pos = 0;

  /* "glycopeptidepy/_c/structure/glycan.pyx":306
 *         has_aggregate = aggregate is not None
 *         pos = 0
 *         while PyDict_Next(self.mapping, &pos, &pkey, &pval):             # <<<<<<<<<<<<<<
 *             value = <ModificationInstanceBase>pval
 *             if has_aggregate and (<GlycosylationBase>(value.rule)).is_core:
 */
  while (1) {
    __pyx_t_1 = __pyx_v_self->mapping;
    __Pyx_INCREF(__pyx_t_1);
    __pyx_t_5 = (PyDict_Next(__pyx_t_1, (&__pyx_v_pos), (&__pyx_v_pkey), (&__pyx_v_pval)) != 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (!__pyx_t_5) break;

    /* "glycopeptidepy/_c/structure/glycan.pyx":307
 *         pos = 0
 *         while PyDict_Next(self.mapping, &pos, &pkey, &pval):
 *             value = <ModificationInstanceBase>pval             # <<<<<<<<<<<<<<
 *             if has_aggregate and (<GlycosylationBase>(value.rule)).is_core:
 *                 continue
 */
    __pyx_t_1 = ((PyObject *)__pyx_v_pval);
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_value, ((struct __pyx_obj_14glycopeptidepy_2_c_9structure_12modification_12modification_ModificationInstanceBase *)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "glycopeptidepy/_c/structure/glycan.pyx":308
 *         while PyDict_Next(self.mapping, &pos, &pkey, &pval):
 *             value = <ModificationInstanceBase>pval
 *             if has_aggregate and (<GlycosylationBase>(value.rule)).is_core:             # <<<<<<<<<<<<<<
 *                 continue
 *             total.add_from(value.composition)
 */
    __pyx_t_6 = (__pyx_v_has_aggregate != 0);
    if (__pyx_t_6) {
    } else {
      __pyx_t_5 = __pyx_t_6;
      goto __pyx_L6_bool_binop_done;
    }
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_value->rule), __pyx_n_s_is_core); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 308, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 308, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_5 = __pyx_t_6;
    __pyx_L6_bool_binop_done:;
    if (__pyx_t_5) {

      /* "glycopeptidepy/_c/structure/glycan.pyx":309
 *             value = <ModificationInstanceBase>pval
 *             if has_aggregate and (<GlycosylationBase>(value.rule)).is_core:
 *                 continue             # <<<<<<<<<<<<<<
 *             total.add_from(value.composition)
 *         if has_aggregate:
 */
      goto __pyx_L3_continue;

      /* "glycopeptidepy/_c/structure/glycan.pyx":308
 *         while PyDict_Next(self.mapping, &pos, &pkey, &pval):
 *             value = <ModificationInstanceBase>pval
 *             if has_aggregate and (<GlycosylationBase>(value.rule)).is_core:             # <<<<<<<<<<<<<<
 *                 continue
 *             total.add_from(value.composition)
 */
    }

    /* "glycopeptidepy/_c/structure/glycan.pyx":310
 *             if has_aggregate and (<GlycosylationBase>(value.rule)).is_core:
 *                 continue
 *             total.add_from(value.composition)             # <<<<<<<<<<<<<<
 *         if has_aggregate:
 *             total.add_from(aggregate.total_composition())
 */
    __pyx_t_1 = ((PyObject *)__pyx_v_value->__pyx_base.composition);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_t_2 = ((PyObject *)((struct __pyx_vtabstruct_5glypy_11composition_12ccomposition_CComposition *)__pyx_v_total->__pyx_vtab)->add_from(__pyx_v_total, __pyx_t_1)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 310, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_L3_continue:;
  }

  /* "glycopeptidepy/_c/structure/glycan.pyx":311
 *                 continue
 *             total.add_from(value.composition)
 *         if has_aggregate:             # <<<<<<<<<<<<<<
 *             total.add_from(aggregate.total_composition())
 *         return total
 */
  __pyx_t_5 = (__pyx_v_has_aggregate != 0);
  if (__pyx_t_5) {

    /* "glycopeptidepy/_c/structure/glycan.pyx":312
 *             total.add_from(value.composition)
 *         if has_aggregate:
 *             total.add_from(aggregate.total_composition())             # <<<<<<<<<<<<<<
 *         return total
 * 
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_aggregate, __pyx_n_s_total_composition); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 312, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_2 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_1);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 312, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_5glypy_11composition_12ccomposition_CComposition *)__pyx_v_total->__pyx_vtab)->add_from(__pyx_v_total, __pyx_t_2)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 312, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "glycopeptidepy/_c/structure/glycan.pyx":311
 *                 continue
 *             total.add_from(value.composition)
 *         if has_aggregate:             # <<<<<<<<<<<<<<
 *             total.add_from(aggregate.total_composition())
 *         return total
 */
  }

  /* "glycopeptidepy/_c/structure/glycan.pyx":313
 *         if has_aggregate:
 *             total.add_from(aggregate.total_composition())
 *         return total             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __Pyx_INCREF(((PyObject *)__pyx_v_total));
  __pyx_r = __pyx_v_total;
  goto __pyx_L0;

  /* "glycopeptidepy/_c/structure/glycan.pyx":293
 *         return  track
 * 
 *     cpdef CComposition total_composition(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             CComposition total
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.glycan.GlycosylationManager.total_composition", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_total);
  __Pyx_XDECREF((PyObject *)__pyx_v_value);
  __Pyx_XDECREF(__pyx_v_aggregate);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_31total_composition(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_30total_composition[] = "GlycosylationManager.total_composition(self) -> CComposition";
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_31total_composition(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("total_composition (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_30total_composition(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_30total_composition(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("total_composition", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_total_composition(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 293, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.glycan.GlycosylationManager.total_composition", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/glycan.pyx":316
 * 
 *     @property
 *     def glycan_composition(self):             # <<<<<<<<<<<<<<
 *         return self.get_glycan_composition()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_18glycan_composition_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_18glycan_composition_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_18glycan_composition___get__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_18glycan_composition___get__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "glycopeptidepy/_c/structure/glycan.pyx":317
 *     @property
 *     def glycan_composition(self):
 *         return self.get_glycan_composition()             # <<<<<<<<<<<<<<
 * 
 *     cdef GlycanCompositionProxy get_glycan_composition(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *)__pyx_v_self->__pyx_vtab)->get_glycan_composition(__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 317, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "glycopeptidepy/_c/structure/glycan.pyx":316
 * 
 *     @property
 *     def glycan_composition(self):             # <<<<<<<<<<<<<<
 *         return self.get_glycan_composition()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.glycan.GlycosylationManager.glycan_composition.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/glycan.pyx":319
 *         return self.get_glycan_composition()
 * 
 *     cdef GlycanCompositionProxy get_glycan_composition(self):             # <<<<<<<<<<<<<<
 *         if self._proxy is None:
 *             self._proxy = self._make_glycan_composition_proxy()
 */

static struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *__pyx_f_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_get_glycan_composition(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_self) {
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_glycan_composition", 0);

  /* "glycopeptidepy/_c/structure/glycan.pyx":320
 * 
 *     cdef GlycanCompositionProxy get_glycan_composition(self):
 *         if self._proxy is None:             # <<<<<<<<<<<<<<
 *             self._proxy = self._make_glycan_composition_proxy()
 *         return self._proxy
 */
  __pyx_t_1 = (((PyObject *)__pyx_v_self->_proxy) == Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "glycopeptidepy/_c/structure/glycan.pyx":321
 *     cdef GlycanCompositionProxy get_glycan_composition(self):
 *         if self._proxy is None:
 *             self._proxy = self._make_glycan_composition_proxy()             # <<<<<<<<<<<<<<
 *         return self._proxy
 * 
 */
    __pyx_t_3 = ((PyObject *)((struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *)__pyx_v_self->__pyx_vtab)->_make_glycan_composition_proxy(__pyx_v_self)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 321, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_3);
    __Pyx_GOTREF(__pyx_v_self->_proxy);
    __Pyx_DECREF(((PyObject *)__pyx_v_self->_proxy));
    __pyx_v_self->_proxy = ((struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "glycopeptidepy/_c/structure/glycan.pyx":320
 * 
 *     cdef GlycanCompositionProxy get_glycan_composition(self):
 *         if self._proxy is None:             # <<<<<<<<<<<<<<
 *             self._proxy = self._make_glycan_composition_proxy()
 *         return self._proxy
 */
  }

  /* "glycopeptidepy/_c/structure/glycan.pyx":322
 *         if self._proxy is None:
 *             self._proxy = self._make_glycan_composition_proxy()
 *         return self._proxy             # <<<<<<<<<<<<<<
 * 
 *     cpdef double mass(self):
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __Pyx_INCREF(((PyObject *)__pyx_v_self->_proxy));
  __pyx_r = __pyx_v_self->_proxy;
  goto __pyx_L0;

  /* "glycopeptidepy/_c/structure/glycan.pyx":319
 *         return self.get_glycan_composition()
 * 
 *     cdef GlycanCompositionProxy get_glycan_composition(self):             # <<<<<<<<<<<<<<
 *         if self._proxy is None:
 *             self._proxy = self._make_glycan_composition_proxy()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.glycan.GlycosylationManager.get_glycan_composition", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/glycan.pyx":324
 *         return self._proxy
 * 
 *     cpdef double mass(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             double total
 */

static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_33mass(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static double __pyx_f_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_mass(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_self, int __pyx_skip_dispatch) {
  double __pyx_v_total;
  int __pyx_v_has_aggregate;
  PyObject *__pyx_v_pkey;
  PyObject *__pyx_v_pval;
  Py_ssize_t __pyx_v_pos;
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_12modification_12modification_ModificationInstanceBase *__pyx_v_value = 0;
  PyObject *__pyx_v_aggregate = 0;
  double __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  double __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("mass", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_mass); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 324, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_33mass)) {
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 324, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_5 = __pyx_PyFloat_AsDouble(__pyx_t_2); if (unlikely((__pyx_t_5 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 324, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_r = __pyx_t_5;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "glycopeptidepy/_c/structure/glycan.pyx":333
 *             ModificationInstanceBase value
 *             object aggregate
 *         total = 0.             # <<<<<<<<<<<<<<
 *         aggregate = self.get_aggregate()
 *         has_aggregate = aggregate is not None
 */
  __pyx_v_total = 0.;

  /* "glycopeptidepy/_c/structure/glycan.pyx":334
 *             object aggregate
 *         total = 0.
 *         aggregate = self.get_aggregate()             # <<<<<<<<<<<<<<
 *         has_aggregate = aggregate is not None
 *         pos = 0
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *)__pyx_v_self->__pyx_vtab)->get_aggregate(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 334, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_aggregate = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "glycopeptidepy/_c/structure/glycan.pyx":335
 *         total = 0.
 *         aggregate = self.get_aggregate()
 *         has_aggregate = aggregate is not None             # <<<<<<<<<<<<<<
 *         pos = 0
 *         while PyDict_Next(self.mapping, &pos, &pkey, &pval):
 */
  __pyx_t_6 = (__pyx_v_aggregate != Py_None);
  __pyx_v_has_aggregate = __pyx_t_6;

  /* "glycopeptidepy/_c/structure/glycan.pyx":336
 *         aggregate = self.get_aggregate()
 *         has_aggregate = aggregate is not None
 *         pos = 0             # <<<<<<<<<<<<<<
 *         while PyDict_Next(self.mapping, &pos, &pkey, &pval):
 *             value = <ModificationInstanceBase>pval
 */
  __pyx_v_pos = 0;

  /* "glycopeptidepy/_c/structure/glycan.pyx":337
 *         has_aggregate = aggregate is not None
 *         pos = 0
 *         while PyDict_Next(self.mapping, &pos, &pkey, &pval):             # <<<<<<<<<<<<<<
 *             value = <ModificationInstanceBase>pval
 *             if has_aggregate and (<GlycosylationBase>(value.rule)).is_core:
 */
  while (1) {
    __pyx_t_1 = __pyx_v_self->mapping;
    __Pyx_INCREF(__pyx_t_1);
    __pyx_t_6 = (PyDict_Next(__pyx_t_1, (&__pyx_v_pos), (&__pyx_v_pkey), (&__pyx_v_pval)) != 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (!__pyx_t_6) break;

    /* "glycopeptidepy/_c/structure/glycan.pyx":338
 *         pos = 0
 *         while PyDict_Next(self.mapping, &pos, &pkey, &pval):
 *             value = <ModificationInstanceBase>pval             # <<<<<<<<<<<<<<
 *             if has_aggregate and (<GlycosylationBase>(value.rule)).is_core:
 *                 continue
 */
    __pyx_t_1 = ((PyObject *)__pyx_v_pval);
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_value, ((struct __pyx_obj_14glycopeptidepy_2_c_9structure_12modification_12modification_ModificationInstanceBase *)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "glycopeptidepy/_c/structure/glycan.pyx":339
 *         while PyDict_Next(self.mapping, &pos, &pkey, &pval):
 *             value = <ModificationInstanceBase>pval
 *             if has_aggregate and (<GlycosylationBase>(value.rule)).is_core:             # <<<<<<<<<<<<<<
 *                 continue
 *             total += value.mass
 */
    __pyx_t_7 = (__pyx_v_has_aggregate != 0);
    if (__pyx_t_7) {
    } else {
      __pyx_t_6 = __pyx_t_7;
      goto __pyx_L6_bool_binop_done;
    }
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_value->rule), __pyx_n_s_is_core); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 339, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 339, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_6 = __pyx_t_7;
    __pyx_L6_bool_binop_done:;
    if (__pyx_t_6) {

      /* "glycopeptidepy/_c/structure/glycan.pyx":340
 *             value = <ModificationInstanceBase>pval
 *             if has_aggregate and (<GlycosylationBase>(value.rule)).is_core:
 *                 continue             # <<<<<<<<<<<<<<
 *             total += value.mass
 *         if has_aggregate:
 */
      goto __pyx_L3_continue;

      /* "glycopeptidepy/_c/structure/glycan.pyx":339
 *         while PyDict_Next(self.mapping, &pos, &pkey, &pval):
 *             value = <ModificationInstanceBase>pval
 *             if has_aggregate and (<GlycosylationBase>(value.rule)).is_core:             # <<<<<<<<<<<<<<
 *                 continue
 *             total += value.mass
 */
    }

    /* "glycopeptidepy/_c/structure/glycan.pyx":341
 *             if has_aggregate and (<GlycosylationBase>(value.rule)).is_core:
 *                 continue
 *             total += value.mass             # <<<<<<<<<<<<<<
 *         if has_aggregate:
 *             total += aggregate.mass()
 */
    __pyx_v_total = (__pyx_v_total + __pyx_v_value->__pyx_base.mass);
    __pyx_L3_continue:;
  }

  /* "glycopeptidepy/_c/structure/glycan.pyx":342
 *                 continue
 *             total += value.mass
 *         if has_aggregate:             # <<<<<<<<<<<<<<
 *             total += aggregate.mass()
 *         return total
 */
  __pyx_t_6 = (__pyx_v_has_aggregate != 0);
  if (__pyx_t_6) {

    /* "glycopeptidepy/_c/structure/glycan.pyx":343
 *             total += value.mass
 *         if has_aggregate:
 *             total += aggregate.mass()             # <<<<<<<<<<<<<<
 *         return total
 * 
 */
    __pyx_t_1 = PyFloat_FromDouble(__pyx_v_total); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 343, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_aggregate, __pyx_n_s_mass); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 343, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 343, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_t_1, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 343, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_5 = __pyx_PyFloat_AsDouble(__pyx_t_3); if (unlikely((__pyx_t_5 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 343, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_total = __pyx_t_5;

    /* "glycopeptidepy/_c/structure/glycan.pyx":342
 *                 continue
 *             total += value.mass
 *         if has_aggregate:             # <<<<<<<<<<<<<<
 *             total += aggregate.mass()
 *         return total
 */
  }

  /* "glycopeptidepy/_c/structure/glycan.pyx":344
 *         if has_aggregate:
 *             total += aggregate.mass()
 *         return total             # <<<<<<<<<<<<<<
 * 
 *     cdef int get_total_glycosylation_size(self):
 */
  __pyx_r = __pyx_v_total;
  goto __pyx_L0;

  /* "glycopeptidepy/_c/structure/glycan.pyx":324
 *         return self._proxy
 * 
 *     cpdef double mass(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             double total
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_WriteUnraisable("glycopeptidepy._c.structure.glycan.GlycosylationManager.mass", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_value);
  __Pyx_XDECREF(__pyx_v_aggregate);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_33mass(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_32mass[] = "GlycosylationManager.mass(self) -> double";
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_33mass(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("mass (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_32mass(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_32mass(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("mass", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_f_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_mass(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 324, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.glycan.GlycosylationManager.mass", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/glycan.pyx":346
 *         return total
 * 
 *     cdef int get_total_glycosylation_size(self):             # <<<<<<<<<<<<<<
 *         if self._total_glycosylation_size == -1:
 *             aggregate = self.get_aggregate()
 */

static int __pyx_f_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_get_total_glycosylation_size(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_self) {
  PyObject *__pyx_v_aggregate = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_total_glycosylation_size", 0);

  /* "glycopeptidepy/_c/structure/glycan.pyx":347
 * 
 *     cdef int get_total_glycosylation_size(self):
 *         if self._total_glycosylation_size == -1:             # <<<<<<<<<<<<<<
 *             aggregate = self.get_aggregate()
 *             if aggregate is None:
 */
  __pyx_t_1 = ((__pyx_v_self->_total_glycosylation_size == -1L) != 0);
  if (__pyx_t_1) {

    /* "glycopeptidepy/_c/structure/glycan.pyx":348
 *     cdef int get_total_glycosylation_size(self):
 *         if self._total_glycosylation_size == -1:
 *             aggregate = self.get_aggregate()             # <<<<<<<<<<<<<<
 *             if aggregate is None:
 *                 self._total_glycosylation_size = 0
 */
    __pyx_t_2 = ((struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *)__pyx_v_self->__pyx_vtab)->get_aggregate(__pyx_v_self); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 348, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_v_aggregate = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "glycopeptidepy/_c/structure/glycan.pyx":349
 *         if self._total_glycosylation_size == -1:
 *             aggregate = self.get_aggregate()
 *             if aggregate is None:             # <<<<<<<<<<<<<<
 *                 self._total_glycosylation_size = 0
 *             else:
 */
    __pyx_t_1 = (__pyx_v_aggregate == Py_None);
    __pyx_t_3 = (__pyx_t_1 != 0);
    if (__pyx_t_3) {

      /* "glycopeptidepy/_c/structure/glycan.pyx":350
 *             aggregate = self.get_aggregate()
 *             if aggregate is None:
 *                 self._total_glycosylation_size = 0             # <<<<<<<<<<<<<<
 *             else:
 *                 self._total_glycosylation_size = sum(self.get_aggregate().values())
 */
      __pyx_v_self->_total_glycosylation_size = 0;

      /* "glycopeptidepy/_c/structure/glycan.pyx":349
 *         if self._total_glycosylation_size == -1:
 *             aggregate = self.get_aggregate()
 *             if aggregate is None:             # <<<<<<<<<<<<<<
 *                 self._total_glycosylation_size = 0
 *             else:
 */
      goto __pyx_L4;
    }

    /* "glycopeptidepy/_c/structure/glycan.pyx":352
 *                 self._total_glycosylation_size = 0
 *             else:
 *                 self._total_glycosylation_size = sum(self.get_aggregate().values())             # <<<<<<<<<<<<<<
 *         return self._total_glycosylation_size
 * 
 */
    /*else*/ {
      __pyx_t_4 = ((struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *)__pyx_v_self->__pyx_vtab)->get_aggregate(__pyx_v_self); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 352, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_values); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 352, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
        }
      }
      __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 352, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_5 = __Pyx_PyObject_CallOneArg(__pyx_builtin_sum, __pyx_t_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 352, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_6 = __Pyx_PyInt_As_int(__pyx_t_5); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 352, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_v_self->_total_glycosylation_size = __pyx_t_6;
    }
    __pyx_L4:;

    /* "glycopeptidepy/_c/structure/glycan.pyx":347
 * 
 *     cdef int get_total_glycosylation_size(self):
 *         if self._total_glycosylation_size == -1:             # <<<<<<<<<<<<<<
 *             aggregate = self.get_aggregate()
 *             if aggregate is None:
 */
  }

  /* "glycopeptidepy/_c/structure/glycan.pyx":353
 *             else:
 *                 self._total_glycosylation_size = sum(self.get_aggregate().values())
 *         return self._total_glycosylation_size             # <<<<<<<<<<<<<<
 * 
 *     cpdef int total_glycosylation_size(self):
 */
  __pyx_r = __pyx_v_self->_total_glycosylation_size;
  goto __pyx_L0;

  /* "glycopeptidepy/_c/structure/glycan.pyx":346
 *         return total
 * 
 *     cdef int get_total_glycosylation_size(self):             # <<<<<<<<<<<<<<
 *         if self._total_glycosylation_size == -1:
 *             aggregate = self.get_aggregate()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_WriteUnraisable("glycopeptidepy._c.structure.glycan.GlycosylationManager.get_total_glycosylation_size", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_aggregate);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/glycan.pyx":355
 *         return self._total_glycosylation_size
 * 
 *     cpdef int total_glycosylation_size(self):             # <<<<<<<<<<<<<<
 *         return self.get_total_glycosylation_size()
 * 
 */

static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_35total_glycosylation_size(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static int __pyx_f_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_total_glycosylation_size(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_self, int __pyx_skip_dispatch) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("total_glycosylation_size", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_total_glycosylation_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 355, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_35total_glycosylation_size)) {
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 355, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_5 = __Pyx_PyInt_As_int(__pyx_t_2); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 355, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_r = __pyx_t_5;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "glycopeptidepy/_c/structure/glycan.pyx":356
 * 
 *     cpdef int total_glycosylation_size(self):
 *         return self.get_total_glycosylation_size()             # <<<<<<<<<<<<<<
 * 
 *     cpdef bint is_fully_specified_topologies(self):
 */
  __pyx_r = ((struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *)__pyx_v_self->__pyx_vtab)->get_total_glycosylation_size(__pyx_v_self);
  goto __pyx_L0;

  /* "glycopeptidepy/_c/structure/glycan.pyx":355
 *         return self._total_glycosylation_size
 * 
 *     cpdef int total_glycosylation_size(self):             # <<<<<<<<<<<<<<
 *         return self.get_total_glycosylation_size()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_WriteUnraisable("glycopeptidepy._c.structure.glycan.GlycosylationManager.total_glycosylation_size", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_35total_glycosylation_size(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_34total_glycosylation_size[] = "GlycosylationManager.total_glycosylation_size(self) -> int";
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_35total_glycosylation_size(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("total_glycosylation_size (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_34total_glycosylation_size(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_34total_glycosylation_size(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("total_glycosylation_size", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_f_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_total_glycosylation_size(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 355, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.glycan.GlycosylationManager.total_glycosylation_size", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/glycan.pyx":358
 *         return self.get_total_glycosylation_size()
 * 
 *     cpdef bint is_fully_specified_topologies(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             bint is_fully_specified
 */

static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_37is_fully_specified_topologies(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static int __pyx_f_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_is_fully_specified_topologies(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_self, int __pyx_skip_dispatch) {
  int __pyx_v_is_fully_specified;
  PyObject *__pyx_v_pkey;
  PyObject *__pyx_v_pval;
  Py_ssize_t __pyx_v_pos;
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_12modification_12modification_ModificationInstanceBase *__pyx_v_value = 0;
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_12modification_4rule_GlycosylationBase *__pyx_v_rule = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("is_fully_specified_topologies", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_is_fully_specified_topologies); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 358, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_37is_fully_specified_topologies)) {
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 358, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 358, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_r = __pyx_t_5;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "glycopeptidepy/_c/structure/glycan.pyx":367
 *             GlycosylationBase rule
 * 
 *         is_fully_specified = self.get_size() > 0             # <<<<<<<<<<<<<<
 *         pos = 0
 *         while PyDict_Next(self.mapping, &pos, &pkey, &pval):
 */
  __pyx_v_is_fully_specified = (((struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *)__pyx_v_self->__pyx_vtab)->get_size(__pyx_v_self) > 0);

  /* "glycopeptidepy/_c/structure/glycan.pyx":368
 * 
 *         is_fully_specified = self.get_size() > 0
 *         pos = 0             # <<<<<<<<<<<<<<
 *         while PyDict_Next(self.mapping, &pos, &pkey, &pval):
 *             value = <ModificationInstanceBase>pval
 */
  __pyx_v_pos = 0;

  /* "glycopeptidepy/_c/structure/glycan.pyx":369
 *         is_fully_specified = self.get_size() > 0
 *         pos = 0
 *         while PyDict_Next(self.mapping, &pos, &pkey, &pval):             # <<<<<<<<<<<<<<
 *             value = <ModificationInstanceBase>pval
 *             rule = <GlycosylationBase>value.rule
 */
  while (1) {
    __pyx_t_1 = __pyx_v_self->mapping;
    __Pyx_INCREF(__pyx_t_1);
    __pyx_t_5 = (PyDict_Next(__pyx_t_1, (&__pyx_v_pos), (&__pyx_v_pkey), (&__pyx_v_pval)) != 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (!__pyx_t_5) break;

    /* "glycopeptidepy/_c/structure/glycan.pyx":370
 *         pos = 0
 *         while PyDict_Next(self.mapping, &pos, &pkey, &pval):
 *             value = <ModificationInstanceBase>pval             # <<<<<<<<<<<<<<
 *             rule = <GlycosylationBase>value.rule
 *             if rule.is_core:
 */
    __pyx_t_1 = ((PyObject *)__pyx_v_pval);
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_value, ((struct __pyx_obj_14glycopeptidepy_2_c_9structure_12modification_12modification_ModificationInstanceBase *)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "glycopeptidepy/_c/structure/glycan.pyx":371
 *         while PyDict_Next(self.mapping, &pos, &pkey, &pval):
 *             value = <ModificationInstanceBase>pval
 *             rule = <GlycosylationBase>value.rule             # <<<<<<<<<<<<<<
 *             if rule.is_core:
 *                 is_fully_specified = False
 */
    __pyx_t_1 = ((PyObject *)__pyx_v_value->rule);
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_rule, ((struct __pyx_obj_14glycopeptidepy_2_c_9structure_12modification_4rule_GlycosylationBase *)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "glycopeptidepy/_c/structure/glycan.pyx":372
 *             value = <ModificationInstanceBase>pval
 *             rule = <GlycosylationBase>value.rule
 *             if rule.is_core:             # <<<<<<<<<<<<<<
 *                 is_fully_specified = False
 *                 break
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_rule), __pyx_n_s_is_core); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 372, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 372, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__pyx_t_5) {

      /* "glycopeptidepy/_c/structure/glycan.pyx":373
 *             rule = <GlycosylationBase>value.rule
 *             if rule.is_core:
 *                 is_fully_specified = False             # <<<<<<<<<<<<<<
 *                 break
 *             elif rule.is_composition:
 */
      __pyx_v_is_fully_specified = 0;

      /* "glycopeptidepy/_c/structure/glycan.pyx":374
 *             if rule.is_core:
 *                 is_fully_specified = False
 *                 break             # <<<<<<<<<<<<<<
 *             elif rule.is_composition:
 *                 is_fully_specified = False
 */
      goto __pyx_L4_break;

      /* "glycopeptidepy/_c/structure/glycan.pyx":372
 *             value = <ModificationInstanceBase>pval
 *             rule = <GlycosylationBase>value.rule
 *             if rule.is_core:             # <<<<<<<<<<<<<<
 *                 is_fully_specified = False
 *                 break
 */
    }

    /* "glycopeptidepy/_c/structure/glycan.pyx":375
 *                 is_fully_specified = False
 *                 break
 *             elif rule.is_composition:             # <<<<<<<<<<<<<<
 *                 is_fully_specified = False
 *                 break
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_rule), __pyx_n_s_is_composition); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 375, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 375, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__pyx_t_5) {

      /* "glycopeptidepy/_c/structure/glycan.pyx":376
 *                 break
 *             elif rule.is_composition:
 *                 is_fully_specified = False             # <<<<<<<<<<<<<<
 *                 break
 *         return is_fully_specified
 */
      __pyx_v_is_fully_specified = 0;

      /* "glycopeptidepy/_c/structure/glycan.pyx":377
 *             elif rule.is_composition:
 *                 is_fully_specified = False
 *                 break             # <<<<<<<<<<<<<<
 *         return is_fully_specified
 * 
 */
      goto __pyx_L4_break;

      /* "glycopeptidepy/_c/structure/glycan.pyx":375
 *                 is_fully_specified = False
 *                 break
 *             elif rule.is_composition:             # <<<<<<<<<<<<<<
 *                 is_fully_specified = False
 *                 break
 */
    }
  }
  __pyx_L4_break:;

  /* "glycopeptidepy/_c/structure/glycan.pyx":378
 *                 is_fully_specified = False
 *                 break
 *         return is_fully_specified             # <<<<<<<<<<<<<<
 * 
 *     cdef GlycanCompositionProxy _make_glycan_composition_proxy(self):
 */
  __pyx_r = __pyx_v_is_fully_specified;
  goto __pyx_L0;

  /* "glycopeptidepy/_c/structure/glycan.pyx":358
 *         return self.get_total_glycosylation_size()
 * 
 *     cpdef bint is_fully_specified_topologies(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             bint is_fully_specified
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_WriteUnraisable("glycopeptidepy._c.structure.glycan.GlycosylationManager.is_fully_specified_topologies", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_value);
  __Pyx_XDECREF((PyObject *)__pyx_v_rule);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_37is_fully_specified_topologies(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_36is_fully_specified_topologies[] = "GlycosylationManager.is_fully_specified_topologies(self) -> bool";
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_37is_fully_specified_topologies(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("is_fully_specified_topologies (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_36is_fully_specified_topologies(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_36is_fully_specified_topologies(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("is_fully_specified_topologies", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_f_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_is_fully_specified_topologies(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 358, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.glycan.GlycosylationManager.is_fully_specified_topologies", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/glycan.pyx":380
 *         return is_fully_specified
 * 
 *     cdef GlycanCompositionProxy _make_glycan_composition_proxy(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             object aggregate
 */

static struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *__pyx_f_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager__make_glycan_composition_proxy(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_self) {
  PyObject *__pyx_v_aggregate = 0;
  PyObject *__pyx_v_base = 0;
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_12modification_12modification_ModificationInstanceBase *__pyx_v_value = 0;
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_12modification_4rule_GlycosylationBase *__pyx_v_rule = 0;
  Py_ssize_t __pyx_v_pos;
  PyObject *__pyx_v_pkey;
  PyObject *__pyx_v_pval;
  PyObject *__pyx_v_gc = NULL;
  PyObject *__pyx_v_proxy = NULL;
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_make_glycan_composition_proxy", 0);

  /* "glycopeptidepy/_c/structure/glycan.pyx":390
 *             PyObject* pval
 * 
 *         aggregate = self.get_aggregate()             # <<<<<<<<<<<<<<
 *         if aggregate is not None:
 *             base = aggregate.clone()
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *)__pyx_v_self->__pyx_vtab)->get_aggregate(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 390, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_aggregate = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "glycopeptidepy/_c/structure/glycan.pyx":391
 * 
 *         aggregate = self.get_aggregate()
 *         if aggregate is not None:             # <<<<<<<<<<<<<<
 *             base = aggregate.clone()
 *         else:
 */
  __pyx_t_2 = (__pyx_v_aggregate != Py_None);
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {

    /* "glycopeptidepy/_c/structure/glycan.pyx":392
 *         aggregate = self.get_aggregate()
 *         if aggregate is not None:
 *             base = aggregate.clone()             # <<<<<<<<<<<<<<
 *         else:
 *             base = HashableGlycanComposition()
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_aggregate, __pyx_n_s_clone); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 392, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_1 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 392, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_v_base = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "glycopeptidepy/_c/structure/glycan.pyx":391
 * 
 *         aggregate = self.get_aggregate()
 *         if aggregate is not None:             # <<<<<<<<<<<<<<
 *             base = aggregate.clone()
 *         else:
 */
    goto __pyx_L3;
  }

  /* "glycopeptidepy/_c/structure/glycan.pyx":394
 *             base = aggregate.clone()
 *         else:
 *             base = HashableGlycanComposition()             # <<<<<<<<<<<<<<
 *             # Represent the initial amide bond between the peptide
 *             # and the first glycan. Subsequent glycans do not need
 */
  /*else*/ {
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_HashableGlycanComposition); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 394, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_1 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 394, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_v_base = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "glycopeptidepy/_c/structure/glycan.pyx":399
 *             # further chemical losses because of the dehyration built
 *             # directly into the Residue abstraction.
 *             base.composition_offset -= WATER_OFFSET             # <<<<<<<<<<<<<<
 *         pos = 0
 *         while PyDict_Next(self.mapping, &pos, &pkey, &pval):
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_base, __pyx_n_s_composition_offset); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 399, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = PyNumber_InPlaceSubtract(__pyx_t_1, ((PyObject *)__pyx_v_14glycopeptidepy_2_c_9structure_6glycan_WATER_OFFSET)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 399, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_base, __pyx_n_s_composition_offset, __pyx_t_4) < 0) __PYX_ERR(0, 399, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }
  __pyx_L3:;

  /* "glycopeptidepy/_c/structure/glycan.pyx":400
 *             # directly into the Residue abstraction.
 *             base.composition_offset -= WATER_OFFSET
 *         pos = 0             # <<<<<<<<<<<<<<
 *         while PyDict_Next(self.mapping, &pos, &pkey, &pval):
 *             value = <ModificationInstanceBase>pval
 */
  __pyx_v_pos = 0;

  /* "glycopeptidepy/_c/structure/glycan.pyx":401
 *             base.composition_offset -= WATER_OFFSET
 *         pos = 0
 *         while PyDict_Next(self.mapping, &pos, &pkey, &pval):             # <<<<<<<<<<<<<<
 *             value = <ModificationInstanceBase>pval
 *             rule = <GlycosylationBase>(value.rule)
 */
  while (1) {
    __pyx_t_4 = __pyx_v_self->mapping;
    __Pyx_INCREF(__pyx_t_4);
    __pyx_t_3 = (PyDict_Next(__pyx_t_4, (&__pyx_v_pos), (&__pyx_v_pkey), (&__pyx_v_pval)) != 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (!__pyx_t_3) break;

    /* "glycopeptidepy/_c/structure/glycan.pyx":402
 *         pos = 0
 *         while PyDict_Next(self.mapping, &pos, &pkey, &pval):
 *             value = <ModificationInstanceBase>pval             # <<<<<<<<<<<<<<
 *             rule = <GlycosylationBase>(value.rule)
 *             if rule.is_core:
 */
    __pyx_t_4 = ((PyObject *)__pyx_v_pval);
    __Pyx_INCREF(__pyx_t_4);
    __Pyx_XDECREF_SET(__pyx_v_value, ((struct __pyx_obj_14glycopeptidepy_2_c_9structure_12modification_12modification_ModificationInstanceBase *)__pyx_t_4));
    __pyx_t_4 = 0;

    /* "glycopeptidepy/_c/structure/glycan.pyx":403
 *         while PyDict_Next(self.mapping, &pos, &pkey, &pval):
 *             value = <ModificationInstanceBase>pval
 *             rule = <GlycosylationBase>(value.rule)             # <<<<<<<<<<<<<<
 *             if rule.is_core:
 *                 continue
 */
    __pyx_t_4 = ((PyObject *)__pyx_v_value->rule);
    __Pyx_INCREF(__pyx_t_4);
    __Pyx_XDECREF_SET(__pyx_v_rule, ((struct __pyx_obj_14glycopeptidepy_2_c_9structure_12modification_4rule_GlycosylationBase *)__pyx_t_4));
    __pyx_t_4 = 0;

    /* "glycopeptidepy/_c/structure/glycan.pyx":404
 *             value = <ModificationInstanceBase>pval
 *             rule = <GlycosylationBase>(value.rule)
 *             if rule.is_core:             # <<<<<<<<<<<<<<
 *                 continue
 *             elif rule.is_composition:
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_rule), __pyx_n_s_is_core); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 404, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 404, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (__pyx_t_3) {

      /* "glycopeptidepy/_c/structure/glycan.pyx":405
 *             rule = <GlycosylationBase>(value.rule)
 *             if rule.is_core:
 *                 continue             # <<<<<<<<<<<<<<
 *             elif rule.is_composition:
 *                 base += rule.glycan
 */
      goto __pyx_L4_continue;

      /* "glycopeptidepy/_c/structure/glycan.pyx":404
 *             value = <ModificationInstanceBase>pval
 *             rule = <GlycosylationBase>(value.rule)
 *             if rule.is_core:             # <<<<<<<<<<<<<<
 *                 continue
 *             elif rule.is_composition:
 */
    }

    /* "glycopeptidepy/_c/structure/glycan.pyx":406
 *             if rule.is_core:
 *                 continue
 *             elif rule.is_composition:             # <<<<<<<<<<<<<<
 *                 base += rule.glycan
 *             else:
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_rule), __pyx_n_s_is_composition); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 406, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 406, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (__pyx_t_3) {

      /* "glycopeptidepy/_c/structure/glycan.pyx":407
 *                 continue
 *             elif rule.is_composition:
 *                 base += rule.glycan             # <<<<<<<<<<<<<<
 *             else:
 *                 # Convert Glycan object into a composition, using the original
 */
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_rule), __pyx_n_s_glycan); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 407, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_1 = PyNumber_InPlaceAdd(__pyx_v_base, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 407, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF_SET(__pyx_v_base, __pyx_t_1);
      __pyx_t_1 = 0;

      /* "glycopeptidepy/_c/structure/glycan.pyx":406
 *             if rule.is_core:
 *                 continue
 *             elif rule.is_composition:             # <<<<<<<<<<<<<<
 *                 base += rule.glycan
 *             else:
 */
      goto __pyx_L6;
    }

    /* "glycopeptidepy/_c/structure/glycan.pyx":415
 *                 # but doesn't need to be explicitly included as the loss is tracked
 *                 # when initializing the base above.
 *                 gc = HashableGlycanComposition.from_glycan(rule._original)             # <<<<<<<<<<<<<<
 *                 base += gc
 *         if not isinstance(base, GlycanCompositionProxy):
 */
    /*else*/ {
      __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_HashableGlycanComposition); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 415, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_from_glycan); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 415, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_rule), __pyx_n_s_original); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 415, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_6 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_6)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
        }
      }
      __pyx_t_1 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_6, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_4);
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 415, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF_SET(__pyx_v_gc, __pyx_t_1);
      __pyx_t_1 = 0;

      /* "glycopeptidepy/_c/structure/glycan.pyx":416
 *                 # when initializing the base above.
 *                 gc = HashableGlycanComposition.from_glycan(rule._original)
 *                 base += gc             # <<<<<<<<<<<<<<
 *         if not isinstance(base, GlycanCompositionProxy):
 *             proxy = GlycanCompositionProxy(base)
 */
      __pyx_t_1 = PyNumber_InPlaceAdd(__pyx_v_base, __pyx_v_gc); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 416, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF_SET(__pyx_v_base, __pyx_t_1);
      __pyx_t_1 = 0;
    }
    __pyx_L6:;
    __pyx_L4_continue:;
  }

  /* "glycopeptidepy/_c/structure/glycan.pyx":417
 *                 gc = HashableGlycanComposition.from_glycan(rule._original)
 *                 base += gc
 *         if not isinstance(base, GlycanCompositionProxy):             # <<<<<<<<<<<<<<
 *             proxy = GlycanCompositionProxy(base)
 *         else:
 */
  __pyx_t_3 = __Pyx_TypeCheck(__pyx_v_base, __pyx_ptype_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy); 
  __pyx_t_2 = ((!(__pyx_t_3 != 0)) != 0);
  if (__pyx_t_2) {

    /* "glycopeptidepy/_c/structure/glycan.pyx":418
 *                 base += gc
 *         if not isinstance(base, GlycanCompositionProxy):
 *             proxy = GlycanCompositionProxy(base)             # <<<<<<<<<<<<<<
 *         else:
 *             proxy = base
 */
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy), __pyx_v_base); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 418, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_proxy = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "glycopeptidepy/_c/structure/glycan.pyx":417
 *                 gc = HashableGlycanComposition.from_glycan(rule._original)
 *                 base += gc
 *         if not isinstance(base, GlycanCompositionProxy):             # <<<<<<<<<<<<<<
 *             proxy = GlycanCompositionProxy(base)
 *         else:
 */
    goto __pyx_L7;
  }

  /* "glycopeptidepy/_c/structure/glycan.pyx":420
 *             proxy = GlycanCompositionProxy(base)
 *         else:
 *             proxy = base             # <<<<<<<<<<<<<<
 *         return proxy
 * 
 */
  /*else*/ {
    __Pyx_INCREF(__pyx_v_base);
    __pyx_v_proxy = __pyx_v_base;
  }
  __pyx_L7:;

  /* "glycopeptidepy/_c/structure/glycan.pyx":421
 *         else:
 *             proxy = base
 *         return proxy             # <<<<<<<<<<<<<<
 * 
 *     cpdef int count_glycosylation_type(self, glycosylation_type):
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  if (!(likely(((__pyx_v_proxy) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_proxy, __pyx_ptype_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy))))) __PYX_ERR(0, 421, __pyx_L1_error)
  __Pyx_INCREF(__pyx_v_proxy);
  __pyx_r = ((struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *)__pyx_v_proxy);
  goto __pyx_L0;

  /* "glycopeptidepy/_c/structure/glycan.pyx":380
 *         return is_fully_specified
 * 
 *     cdef GlycanCompositionProxy _make_glycan_composition_proxy(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             object aggregate
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.glycan.GlycosylationManager._make_glycan_composition_proxy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_aggregate);
  __Pyx_XDECREF(__pyx_v_base);
  __Pyx_XDECREF((PyObject *)__pyx_v_value);
  __Pyx_XDECREF((PyObject *)__pyx_v_rule);
  __Pyx_XDECREF(__pyx_v_gc);
  __Pyx_XDECREF(__pyx_v_proxy);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/glycan.pyx":423
 *         return proxy
 * 
 *     cpdef int count_glycosylation_type(self, glycosylation_type):             # <<<<<<<<<<<<<<
 *         cdef:
 *             dict tracks
 */

static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_39count_glycosylation_type(PyObject *__pyx_v_self, PyObject *__pyx_v_glycosylation_type); /*proto*/
static int __pyx_f_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_count_glycosylation_type(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_self, PyObject *__pyx_v_glycosylation_type, int __pyx_skip_dispatch) {
  PyObject *__pyx_v_tracks = 0;
  PyObject *__pyx_v_tmp;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("count_glycosylation_type", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_count_glycosylation_type); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 423, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_39count_glycosylation_type)) {
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_v_glycosylation_type) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_glycosylation_type);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 423, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_5 = __Pyx_PyInt_As_int(__pyx_t_2); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 423, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_r = __pyx_t_5;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "glycopeptidepy/_c/structure/glycan.pyx":427
 *             dict tracks
 *             PyObject* tmp
 *         tracks = self.get_glycosylation_types()             # <<<<<<<<<<<<<<
 *         tmp = PyDict_GetItem(tracks, glycosylation_type)
 *         if tmp == NULL:
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *)__pyx_v_self->__pyx_vtab)->get_glycosylation_types(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 427, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_tracks = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "glycopeptidepy/_c/structure/glycan.pyx":428
 *             PyObject* tmp
 *         tracks = self.get_glycosylation_types()
 *         tmp = PyDict_GetItem(tracks, glycosylation_type)             # <<<<<<<<<<<<<<
 *         if tmp == NULL:
 *             return 0
 */
  __pyx_v_tmp = PyDict_GetItem(__pyx_v_tracks, __pyx_v_glycosylation_type);

  /* "glycopeptidepy/_c/structure/glycan.pyx":429
 *         tracks = self.get_glycosylation_types()
 *         tmp = PyDict_GetItem(tracks, glycosylation_type)
 *         if tmp == NULL:             # <<<<<<<<<<<<<<
 *             return 0
 *         else:
 */
  __pyx_t_6 = ((__pyx_v_tmp == NULL) != 0);
  if (__pyx_t_6) {

    /* "glycopeptidepy/_c/structure/glycan.pyx":430
 *         tmp = PyDict_GetItem(tracks, glycosylation_type)
 *         if tmp == NULL:
 *             return 0             # <<<<<<<<<<<<<<
 *         else:
 *             return PyList_Size(<list>tmp)
 */
    __pyx_r = 0;
    goto __pyx_L0;

    /* "glycopeptidepy/_c/structure/glycan.pyx":429
 *         tracks = self.get_glycosylation_types()
 *         tmp = PyDict_GetItem(tracks, glycosylation_type)
 *         if tmp == NULL:             # <<<<<<<<<<<<<<
 *             return 0
 *         else:
 */
  }

  /* "glycopeptidepy/_c/structure/glycan.pyx":432
 *             return 0
 *         else:
 *             return PyList_Size(<list>tmp)             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
  /*else*/ {
    __pyx_t_7 = PyList_Size(((PyObject *)__pyx_v_tmp)); if (unlikely(__pyx_t_7 == ((Py_ssize_t)-1L))) __PYX_ERR(0, 432, __pyx_L1_error)
    __pyx_r = __pyx_t_7;
    goto __pyx_L0;
  }

  /* "glycopeptidepy/_c/structure/glycan.pyx":423
 *         return proxy
 * 
 *     cpdef int count_glycosylation_type(self, glycosylation_type):             # <<<<<<<<<<<<<<
 *         cdef:
 *             dict tracks
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_WriteUnraisable("glycopeptidepy._c.structure.glycan.GlycosylationManager.count_glycosylation_type", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_tracks);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_39count_glycosylation_type(PyObject *__pyx_v_self, PyObject *__pyx_v_glycosylation_type); /*proto*/
static char __pyx_doc_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_38count_glycosylation_type[] = "GlycosylationManager.count_glycosylation_type(self, glycosylation_type) -> int";
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_39count_glycosylation_type(PyObject *__pyx_v_self, PyObject *__pyx_v_glycosylation_type) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("count_glycosylation_type (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_38count_glycosylation_type(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *)__pyx_v_self), ((PyObject *)__pyx_v_glycosylation_type));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_38count_glycosylation_type(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_self, PyObject *__pyx_v_glycosylation_type) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("count_glycosylation_type", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_f_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_count_glycosylation_type(__pyx_v_self, __pyx_v_glycosylation_type, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 423, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.glycan.GlycosylationManager.count_glycosylation_type", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/glycan.pyx":434
 *             return PyList_Size(<list>tmp)
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return "{self.__class__.__name__}({self.mapping})".format(self=self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_41__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_41__repr__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_40__repr__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_40__repr__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "glycopeptidepy/_c/structure/glycan.pyx":435
 * 
 *     def __repr__(self):
 *         return "{self.__class__.__name__}({self.mapping})".format(self=self)             # <<<<<<<<<<<<<<
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_kp_s_self___class_____name___self_ma, __pyx_n_s_format); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 435, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 435, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_self, ((PyObject *)__pyx_v_self)) < 0) __PYX_ERR(0, 435, __pyx_L1_error)
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_empty_tuple, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 435, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "glycopeptidepy/_c/structure/glycan.pyx":434
 *             return PyList_Size(<list>tmp)
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return "{self.__class__.__name__}({self.mapping})".format(self=self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.glycan.GlycosylationManager.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/glycan.pxd":36
 * cdef class GlycosylationManager(object):
 *     cdef:
 *         public dict mapping             # <<<<<<<<<<<<<<
 *         public object parent
 *         public object _aggregate
 */

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_7mapping_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_7mapping_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_7mapping___get__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_7mapping___get__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->mapping);
  __pyx_r = __pyx_v_self->mapping;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_7mapping_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_7mapping_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_7mapping_2__set__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_7mapping_2__set__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);
  if (!(likely(PyDict_CheckExact(__pyx_v_value))||((__pyx_v_value) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "dict", Py_TYPE(__pyx_v_value)->tp_name), 0))) __PYX_ERR(1, 36, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_value;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->mapping);
  __Pyx_DECREF(__pyx_v_self->mapping);
  __pyx_v_self->mapping = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.glycan.GlycosylationManager.mapping.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_7mapping_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_7mapping_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_7mapping_4__del__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_7mapping_4__del__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->mapping);
  __Pyx_DECREF(__pyx_v_self->mapping);
  __pyx_v_self->mapping = ((PyObject*)Py_None);

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/glycan.pxd":37
 *     cdef:
 *         public dict mapping
 *         public object parent             # <<<<<<<<<<<<<<
 *         public object _aggregate
 *         public GlycanCompositionProxy _proxy
 */

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_6parent_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_6parent_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_6parent___get__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_6parent___get__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->parent);
  __pyx_r = __pyx_v_self->parent;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_6parent_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_6parent_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_6parent_2__set__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_6parent_2__set__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 0);
  __Pyx_INCREF(__pyx_v_value);
  __Pyx_GIVEREF(__pyx_v_value);
  __Pyx_GOTREF(__pyx_v_self->parent);
  __Pyx_DECREF(__pyx_v_self->parent);
  __pyx_v_self->parent = __pyx_v_value;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_6parent_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_6parent_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_6parent_4__del__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_6parent_4__del__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->parent);
  __Pyx_DECREF(__pyx_v_self->parent);
  __pyx_v_self->parent = Py_None;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/glycan.pxd":38
 *         public dict mapping
 *         public object parent
 *         public object _aggregate             # <<<<<<<<<<<<<<
 *         public GlycanCompositionProxy _proxy
 *         public dict _type_track
 */

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_10_aggregate_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_10_aggregate_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_10_aggregate___get__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_10_aggregate___get__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->_aggregate);
  __pyx_r = __pyx_v_self->_aggregate;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_10_aggregate_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_10_aggregate_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_10_aggregate_2__set__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_10_aggregate_2__set__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 0);
  __Pyx_INCREF(__pyx_v_value);
  __Pyx_GIVEREF(__pyx_v_value);
  __Pyx_GOTREF(__pyx_v_self->_aggregate);
  __Pyx_DECREF(__pyx_v_self->_aggregate);
  __pyx_v_self->_aggregate = __pyx_v_value;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_10_aggregate_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_10_aggregate_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_10_aggregate_4__del__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_10_aggregate_4__del__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->_aggregate);
  __Pyx_DECREF(__pyx_v_self->_aggregate);
  __pyx_v_self->_aggregate = Py_None;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/glycan.pxd":39
 *         public object parent
 *         public object _aggregate
 *         public GlycanCompositionProxy _proxy             # <<<<<<<<<<<<<<
 *         public dict _type_track
 *         public int _total_glycosylation_size
 */

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_6_proxy_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_6_proxy_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_6_proxy___get__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_6_proxy___get__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->_proxy));
  __pyx_r = ((PyObject *)__pyx_v_self->_proxy);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_6_proxy_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_6_proxy_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_6_proxy_2__set__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_6_proxy_2__set__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);
  if (!(likely(((__pyx_v_value) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_value, __pyx_ptype_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy))))) __PYX_ERR(1, 39, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_value;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->_proxy);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->_proxy));
  __pyx_v_self->_proxy = ((struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.glycan.GlycosylationManager._proxy.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_6_proxy_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_6_proxy_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_6_proxy_4__del__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_6_proxy_4__del__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->_proxy);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->_proxy));
  __pyx_v_self->_proxy = ((struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *)Py_None);

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/glycan.pxd":40
 *         public object _aggregate
 *         public GlycanCompositionProxy _proxy
 *         public dict _type_track             # <<<<<<<<<<<<<<
 *         public int _total_glycosylation_size
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_11_type_track_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_11_type_track_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_11_type_track___get__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_11_type_track___get__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->_type_track);
  __pyx_r = __pyx_v_self->_type_track;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_11_type_track_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_11_type_track_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_11_type_track_2__set__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_11_type_track_2__set__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);
  if (!(likely(PyDict_CheckExact(__pyx_v_value))||((__pyx_v_value) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "dict", Py_TYPE(__pyx_v_value)->tp_name), 0))) __PYX_ERR(1, 40, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_value;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->_type_track);
  __Pyx_DECREF(__pyx_v_self->_type_track);
  __pyx_v_self->_type_track = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.glycan.GlycosylationManager._type_track.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_11_type_track_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_11_type_track_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_11_type_track_4__del__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_11_type_track_4__del__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->_type_track);
  __Pyx_DECREF(__pyx_v_self->_type_track);
  __pyx_v_self->_type_track = ((PyObject*)Py_None);

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/glycan.pxd":41
 *         public GlycanCompositionProxy _proxy
 *         public dict _type_track
 *         public int _total_glycosylation_size             # <<<<<<<<<<<<<<
 * 
 *     cpdef update(self, other)
 */

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_25_total_glycosylation_size_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_25_total_glycosylation_size_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_25_total_glycosylation_size___get__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_25_total_glycosylation_size___get__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_self->_total_glycosylation_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 41, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.glycan.GlycosylationManager._total_glycosylation_size.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_25_total_glycosylation_size_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_25_total_glycosylation_size_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_25_total_glycosylation_size_2__set__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_25_total_glycosylation_size_2__set__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyInt_As_int(__pyx_v_value); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 41, __pyx_L1_error)
  __pyx_v_self->_total_glycosylation_size = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("glycopeptidepy._c.structure.glycan.GlycosylationManager._total_glycosylation_size.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_43__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_42__reduce_cython__[] = "GlycosylationManager.__reduce_cython__(self)";
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_43__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_42__reduce_cython__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_42__reduce_cython__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self._aggregate, self._proxy, self._total_glycosylation_size, self._type_track, self.mapping, self.parent)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_self->_total_glycosylation_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(6); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_self->_aggregate);
  __Pyx_GIVEREF(__pyx_v_self->_aggregate);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_self->_aggregate);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->_proxy));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self->_proxy));
  PyTuple_SET_ITEM(__pyx_t_2, 1, ((PyObject *)__pyx_v_self->_proxy));
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_t_1);
  __Pyx_INCREF(__pyx_v_self->_type_track);
  __Pyx_GIVEREF(__pyx_v_self->_type_track);
  PyTuple_SET_ITEM(__pyx_t_2, 3, __pyx_v_self->_type_track);
  __Pyx_INCREF(__pyx_v_self->mapping);
  __Pyx_GIVEREF(__pyx_v_self->mapping);
  PyTuple_SET_ITEM(__pyx_t_2, 4, __pyx_v_self->mapping);
  __Pyx_INCREF(__pyx_v_self->parent);
  __Pyx_GIVEREF(__pyx_v_self->parent);
  PyTuple_SET_ITEM(__pyx_t_2, 5, __pyx_v_self->parent);
  __pyx_t_1 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self._aggregate, self._proxy, self._total_glycosylation_size, self._type_track, self.mapping, self.parent)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_2 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v__dict = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "(tree fragment)":7
 *     state = (self._aggregate, self._proxy, self._total_glycosylation_size, self._type_track, self.mapping, self.parent)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_3 = (__pyx_v__dict != Py_None);
  __pyx_t_4 = (__pyx_t_3 != 0);
  if (__pyx_t_4) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v__dict);
    __pyx_t_1 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self._aggregate is not None or self._proxy is not None or self._type_track is not None or self.mapping is not None or self.parent is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self._aggregate, self._proxy, self._total_glycosylation_size, self._type_track, self.mapping, self.parent)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self._aggregate is not None or self._proxy is not None or self._type_track is not None or self.mapping is not None or self.parent is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_GlycosylationManager, (type(self), 0xeaeda1b, None), state
 */
  /*else*/ {
    __pyx_t_3 = (__pyx_v_self->_aggregate != Py_None);
    __pyx_t_5 = (__pyx_t_3 != 0);
    if (!__pyx_t_5) {
    } else {
      __pyx_t_4 = __pyx_t_5;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_5 = (((PyObject *)__pyx_v_self->_proxy) != Py_None);
    __pyx_t_3 = (__pyx_t_5 != 0);
    if (!__pyx_t_3) {
    } else {
      __pyx_t_4 = __pyx_t_3;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_3 = (__pyx_v_self->_type_track != ((PyObject*)Py_None));
    __pyx_t_5 = (__pyx_t_3 != 0);
    if (!__pyx_t_5) {
    } else {
      __pyx_t_4 = __pyx_t_5;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_5 = (__pyx_v_self->mapping != ((PyObject*)Py_None));
    __pyx_t_3 = (__pyx_t_5 != 0);
    if (!__pyx_t_3) {
    } else {
      __pyx_t_4 = __pyx_t_3;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_3 = (__pyx_v_self->parent != Py_None);
    __pyx_t_5 = (__pyx_t_3 != 0);
    __pyx_t_4 = __pyx_t_5;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_4;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self._aggregate is not None or self._proxy is not None or self._type_track is not None or self.mapping is not None or self.parent is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_GlycosylationManager, (type(self), 0xeaeda1b, None), state
 *     else:
 */
  __pyx_t_4 = (__pyx_v_use_setstate != 0);
  if (__pyx_t_4) {

    /* "(tree fragment)":13
 *         use_setstate = self._aggregate is not None or self._proxy is not None or self._type_track is not None or self.mapping is not None or self.parent is not None
 *     if use_setstate:
 *         return __pyx_unpickle_GlycosylationManager, (type(self), 0xeaeda1b, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_GlycosylationManager, (type(self), 0xeaeda1b, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_pyx_unpickle_GlycosylationMana); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_246340123);
    __Pyx_GIVEREF(__pyx_int_246340123);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_int_246340123);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    PyTuple_SET_ITEM(__pyx_t_2, 2, Py_None);
    __pyx_t_6 = PyTuple_New(3); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_2);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_v_state);
    __pyx_t_1 = 0;
    __pyx_t_2 = 0;
    __pyx_r = __pyx_t_6;
    __pyx_t_6 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self._aggregate is not None or self._proxy is not None or self._type_track is not None or self.mapping is not None or self.parent is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_GlycosylationManager, (type(self), 0xeaeda1b, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_GlycosylationManager, (type(self), 0xeaeda1b, None), state
 *     else:
 *         return __pyx_unpickle_GlycosylationManager, (type(self), 0xeaeda1b, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_GlycosylationManager__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_pyx_unpickle_GlycosylationMana); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_246340123);
    __Pyx_GIVEREF(__pyx_int_246340123);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_int_246340123);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_v_state);
    __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_2);
    __pyx_t_6 = 0;
    __pyx_t_2 = 0;
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.glycan.GlycosylationManager.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_GlycosylationManager, (type(self), 0xeaeda1b, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_GlycosylationManager__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_45__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static char __pyx_doc_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_44__setstate_cython__[] = "GlycosylationManager.__setstate_cython__(self, __pyx_state)";
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_45__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_44__setstate_cython__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_44__setstate_cython__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_GlycosylationManager, (type(self), 0xeaeda1b, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_GlycosylationManager__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(2, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_14glycopeptidepy_2_c_9structure_6glycan___pyx_unpickle_GlycosylationManager__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_GlycosylationManager, (type(self), 0xeaeda1b, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_GlycosylationManager__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.glycan.GlycosylationManager.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_GlycanCompositionProxy(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_1__pyx_unpickle_GlycanCompositionProxy(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_14glycopeptidepy_2_c_9structure_6glycan___pyx_unpickle_GlycanCompositionProxy[] = "__pyx_unpickle_GlycanCompositionProxy(__pyx_type, long __pyx_checksum, __pyx_state)";
static PyMethodDef __pyx_mdef_14glycopeptidepy_2_c_9structure_6glycan_1__pyx_unpickle_GlycanCompositionProxy = {"__pyx_unpickle_GlycanCompositionProxy", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_1__pyx_unpickle_GlycanCompositionProxy, METH_VARARGS|METH_KEYWORDS, __pyx_doc_14glycopeptidepy_2_c_9structure_6glycan___pyx_unpickle_GlycanCompositionProxy};
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_1__pyx_unpickle_GlycanCompositionProxy(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_GlycanCompositionProxy (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_type)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_checksum)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_GlycanCompositionProxy", 1, 3, 3, 1); __PYX_ERR(2, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_GlycanCompositionProxy", 1, 3, 3, 2); __PYX_ERR(2, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_unpickle_GlycanCompositionProxy") < 0)) __PYX_ERR(2, 1, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_GlycanCompositionProxy", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(2, 1, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("glycopeptidepy._c.structure.glycan.__pyx_unpickle_GlycanCompositionProxy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan___pyx_unpickle_GlycanCompositionProxy(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_6glycan___pyx_unpickle_GlycanCompositionProxy(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_GlycanCompositionProxy", 0);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x4126d5b, 0x40511ef, 0x04f04be):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0x4126d5b, 0x40511ef, 0x04f04be) = (_serialized, obj))" % __pyx_checksum)
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple_, Py_NE)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x4126d5b, 0x40511ef, 0x04f04be):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0x4126d5b, 0x40511ef, 0x04f04be) = (_serialized, obj))" % __pyx_checksum)
 *     __pyx_result = GlycanCompositionProxy.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError);
    __pyx_t_4 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, -1); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_4, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0x4126d5b, 0x40511ef, 0x04f04be):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0x4126d5b, 0x40511ef, 0x04f04be) = (_serialized, obj))" % __pyx_checksum)             # <<<<<<<<<<<<<<
 *     __pyx_result = GlycanCompositionProxy.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0, __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_INCREF(__pyx_v___pyx_PickleError);
    __pyx_t_1 = __pyx_v___pyx_PickleError; __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_4 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_6, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_5);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __PYX_ERR(2, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x4126d5b, 0x40511ef, 0x04f04be):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0x4126d5b, 0x40511ef, 0x04f04be) = (_serialized, obj))" % __pyx_checksum)
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0x4126d5b, 0x40511ef, 0x04f04be) = (_serialized, obj))" % __pyx_checksum)
 *     __pyx_result = GlycanCompositionProxy.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_GlycanCompositionProxy__set_state(<GlycanCompositionProxy> __pyx_result, __pyx_state)
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy), __pyx_n_s_new); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_4 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_5, __pyx_v___pyx_type) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v___pyx_type);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0x4126d5b, 0x40511ef, 0x04f04be) = (_serialized, obj))" % __pyx_checksum)
 *     __pyx_result = GlycanCompositionProxy.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_GlycanCompositionProxy__set_state(<GlycanCompositionProxy> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_3 = (__pyx_v___pyx_state != Py_None);
  __pyx_t_2 = (__pyx_t_3 != 0);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = GlycanCompositionProxy.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_GlycanCompositionProxy__set_state(<GlycanCompositionProxy> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_GlycanCompositionProxy__set_state(GlycanCompositionProxy __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(2, 9, __pyx_L1_error)
    __pyx_t_4 = __pyx_f_14glycopeptidepy_2_c_9structure_6glycan___pyx_unpickle_GlycanCompositionProxy__set_state(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0x4126d5b, 0x40511ef, 0x04f04be) = (_serialized, obj))" % __pyx_checksum)
 *     __pyx_result = GlycanCompositionProxy.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_GlycanCompositionProxy__set_state(<GlycanCompositionProxy> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_GlycanCompositionProxy__set_state(<GlycanCompositionProxy> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_GlycanCompositionProxy__set_state(GlycanCompositionProxy __pyx_result, tuple __pyx_state):
 *     __pyx_result._serialized = __pyx_state[0]; __pyx_result.obj = __pyx_state[1]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_GlycanCompositionProxy(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.glycan.__pyx_unpickle_GlycanCompositionProxy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_GlycanCompositionProxy__set_state(<GlycanCompositionProxy> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_GlycanCompositionProxy__set_state(GlycanCompositionProxy __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result._serialized = __pyx_state[0]; __pyx_result.obj = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_14glycopeptidepy_2_c_9structure_6glycan___pyx_unpickle_GlycanCompositionProxy__set_state(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_GlycanCompositionProxy__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_GlycanCompositionProxy__set_state(GlycanCompositionProxy __pyx_result, tuple __pyx_state):
 *     __pyx_result._serialized = __pyx_state[0]; __pyx_result.obj = __pyx_state[1]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyString_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "str", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->_serialized);
  __Pyx_DECREF(__pyx_v___pyx_result->_serialized);
  __pyx_v___pyx_result->_serialized = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5glypy_2_c_9structure_18glycan_composition__CompositionBase))))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->obj);
  __Pyx_DECREF(((PyObject *)__pyx_v___pyx_result->obj));
  __pyx_v___pyx_result->obj = ((struct __pyx_obj_5glypy_2_c_9structure_18glycan_composition__CompositionBase *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_GlycanCompositionProxy__set_state(GlycanCompositionProxy __pyx_result, tuple __pyx_state):
 *     __pyx_result._serialized = __pyx_state[0]; __pyx_result.obj = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(2, 13, __pyx_L1_error)
  }
  __pyx_t_3 = PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_4 = ((__pyx_t_3 > 2) != 0);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_5 = (__pyx_t_4 != 0);
  __pyx_t_2 = __pyx_t_5;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result._serialized = __pyx_state[0]; __pyx_result.obj = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[2])             # <<<<<<<<<<<<<<
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_update); if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(2, 14, __pyx_L1_error)
    }
    __pyx_t_6 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_8 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
      }
    }
    __pyx_t_1 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_8, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_6);
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_GlycanCompositionProxy__set_state(GlycanCompositionProxy __pyx_result, tuple __pyx_state):
 *     __pyx_result._serialized = __pyx_state[0]; __pyx_result.obj = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[2])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_GlycanCompositionProxy__set_state(<GlycanCompositionProxy> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_GlycanCompositionProxy__set_state(GlycanCompositionProxy __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result._serialized = __pyx_state[0]; __pyx_result.obj = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.glycan.__pyx_unpickle_GlycanCompositionProxy__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_GlycanCompositionWithOffsetProxy(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_3__pyx_unpickle_GlycanCompositionWithOffsetProxy(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_14glycopeptidepy_2_c_9structure_6glycan_2__pyx_unpickle_GlycanCompositionWithOffsetProxy[] = "__pyx_unpickle_GlycanCompositionWithOffsetProxy(__pyx_type, long __pyx_checksum, __pyx_state)";
static PyMethodDef __pyx_mdef_14glycopeptidepy_2_c_9structure_6glycan_3__pyx_unpickle_GlycanCompositionWithOffsetProxy = {"__pyx_unpickle_GlycanCompositionWithOffsetProxy", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_3__pyx_unpickle_GlycanCompositionWithOffsetProxy, METH_VARARGS|METH_KEYWORDS, __pyx_doc_14glycopeptidepy_2_c_9structure_6glycan_2__pyx_unpickle_GlycanCompositionWithOffsetProxy};
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_3__pyx_unpickle_GlycanCompositionWithOffsetProxy(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_GlycanCompositionWithOffsetProxy (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_type)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_checksum)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_GlycanCompositionWithOffsetProxy", 1, 3, 3, 1); __PYX_ERR(2, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_GlycanCompositionWithOffsetProxy", 1, 3, 3, 2); __PYX_ERR(2, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_unpickle_GlycanCompositionWithOffsetProxy") < 0)) __PYX_ERR(2, 1, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_GlycanCompositionWithOffsetProxy", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(2, 1, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("glycopeptidepy._c.structure.glycan.__pyx_unpickle_GlycanCompositionWithOffsetProxy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_2__pyx_unpickle_GlycanCompositionWithOffsetProxy(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_2__pyx_unpickle_GlycanCompositionWithOffsetProxy(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_GlycanCompositionWithOffsetProxy", 0);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x32b2f99, 0x6f0dd41, 0x3381ea0):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0x32b2f99, 0x6f0dd41, 0x3381ea0) = (_serialized, composition_offset, obj))" % __pyx_checksum)
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__2, Py_NE)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x32b2f99, 0x6f0dd41, 0x3381ea0):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0x32b2f99, 0x6f0dd41, 0x3381ea0) = (_serialized, composition_offset, obj))" % __pyx_checksum)
 *     __pyx_result = GlycanCompositionWithOffsetProxy.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError);
    __pyx_t_4 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, -1); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_4, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0x32b2f99, 0x6f0dd41, 0x3381ea0):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0x32b2f99, 0x6f0dd41, 0x3381ea0) = (_serialized, composition_offset, obj))" % __pyx_checksum)             # <<<<<<<<<<<<<<
 *     __pyx_result = GlycanCompositionWithOffsetProxy.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_2, __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_INCREF(__pyx_v___pyx_PickleError);
    __pyx_t_1 = __pyx_v___pyx_PickleError; __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_4 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_6, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_5);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __PYX_ERR(2, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x32b2f99, 0x6f0dd41, 0x3381ea0):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0x32b2f99, 0x6f0dd41, 0x3381ea0) = (_serialized, composition_offset, obj))" % __pyx_checksum)
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0x32b2f99, 0x6f0dd41, 0x3381ea0) = (_serialized, composition_offset, obj))" % __pyx_checksum)
 *     __pyx_result = GlycanCompositionWithOffsetProxy.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_GlycanCompositionWithOffsetProxy__set_state(<GlycanCompositionWithOffsetProxy> __pyx_result, __pyx_state)
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionWithOffsetProxy), __pyx_n_s_new); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_4 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_5, __pyx_v___pyx_type) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v___pyx_type);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0x32b2f99, 0x6f0dd41, 0x3381ea0) = (_serialized, composition_offset, obj))" % __pyx_checksum)
 *     __pyx_result = GlycanCompositionWithOffsetProxy.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_GlycanCompositionWithOffsetProxy__set_state(<GlycanCompositionWithOffsetProxy> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_3 = (__pyx_v___pyx_state != Py_None);
  __pyx_t_2 = (__pyx_t_3 != 0);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = GlycanCompositionWithOffsetProxy.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_GlycanCompositionWithOffsetProxy__set_state(<GlycanCompositionWithOffsetProxy> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_GlycanCompositionWithOffsetProxy__set_state(GlycanCompositionWithOffsetProxy __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(2, 9, __pyx_L1_error)
    __pyx_t_4 = __pyx_f_14glycopeptidepy_2_c_9structure_6glycan___pyx_unpickle_GlycanCompositionWithOffsetProxy__set_state(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionWithOffsetProxy *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0x32b2f99, 0x6f0dd41, 0x3381ea0) = (_serialized, composition_offset, obj))" % __pyx_checksum)
 *     __pyx_result = GlycanCompositionWithOffsetProxy.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_GlycanCompositionWithOffsetProxy__set_state(<GlycanCompositionWithOffsetProxy> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_GlycanCompositionWithOffsetProxy__set_state(<GlycanCompositionWithOffsetProxy> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_GlycanCompositionWithOffsetProxy__set_state(GlycanCompositionWithOffsetProxy __pyx_result, tuple __pyx_state):
 *     __pyx_result._serialized = __pyx_state[0]; __pyx_result.composition_offset = __pyx_state[1]; __pyx_result.obj = __pyx_state[2]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_GlycanCompositionWithOffsetProxy(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.glycan.__pyx_unpickle_GlycanCompositionWithOffsetProxy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_GlycanCompositionWithOffsetProxy__set_state(<GlycanCompositionWithOffsetProxy> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_GlycanCompositionWithOffsetProxy__set_state(GlycanCompositionWithOffsetProxy __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result._serialized = __pyx_state[0]; __pyx_result.composition_offset = __pyx_state[1]; __pyx_result.obj = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_14glycopeptidepy_2_c_9structure_6glycan___pyx_unpickle_GlycanCompositionWithOffsetProxy__set_state(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionWithOffsetProxy *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_GlycanCompositionWithOffsetProxy__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_GlycanCompositionWithOffsetProxy__set_state(GlycanCompositionWithOffsetProxy __pyx_result, tuple __pyx_state):
 *     __pyx_result._serialized = __pyx_state[0]; __pyx_result.composition_offset = __pyx_state[1]; __pyx_result.obj = __pyx_state[2]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[3])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyString_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "str", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base._serialized);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base._serialized);
  __pyx_v___pyx_result->__pyx_base._serialized = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5glypy_11composition_12ccomposition_CComposition))))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->composition_offset);
  __Pyx_DECREF(((PyObject *)__pyx_v___pyx_result->composition_offset));
  __pyx_v___pyx_result->composition_offset = ((struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5glypy_2_c_9structure_18glycan_composition__CompositionBase))))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base.obj);
  __Pyx_DECREF(((PyObject *)__pyx_v___pyx_result->__pyx_base.obj));
  __pyx_v___pyx_result->__pyx_base.obj = ((struct __pyx_obj_5glypy_2_c_9structure_18glycan_composition__CompositionBase *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_GlycanCompositionWithOffsetProxy__set_state(GlycanCompositionWithOffsetProxy __pyx_result, tuple __pyx_state):
 *     __pyx_result._serialized = __pyx_state[0]; __pyx_result.composition_offset = __pyx_state[1]; __pyx_result.obj = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[3])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(2, 13, __pyx_L1_error)
  }
  __pyx_t_3 = PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_4 = ((__pyx_t_3 > 3) != 0);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_5 = (__pyx_t_4 != 0);
  __pyx_t_2 = __pyx_t_5;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result._serialized = __pyx_state[0]; __pyx_result.composition_offset = __pyx_state[1]; __pyx_result.obj = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[3])             # <<<<<<<<<<<<<<
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_update); if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(2, 14, __pyx_L1_error)
    }
    __pyx_t_6 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_8 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
      }
    }
    __pyx_t_1 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_8, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_6);
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_GlycanCompositionWithOffsetProxy__set_state(GlycanCompositionWithOffsetProxy __pyx_result, tuple __pyx_state):
 *     __pyx_result._serialized = __pyx_state[0]; __pyx_result.composition_offset = __pyx_state[1]; __pyx_result.obj = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[3])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_GlycanCompositionWithOffsetProxy__set_state(<GlycanCompositionWithOffsetProxy> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_GlycanCompositionWithOffsetProxy__set_state(GlycanCompositionWithOffsetProxy __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result._serialized = __pyx_state[0]; __pyx_result.composition_offset = __pyx_state[1]; __pyx_result.obj = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.glycan.__pyx_unpickle_GlycanCompositionWithOffsetProxy__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_GlycosylationManager(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_5__pyx_unpickle_GlycosylationManager(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_14glycopeptidepy_2_c_9structure_6glycan_4__pyx_unpickle_GlycosylationManager[] = "__pyx_unpickle_GlycosylationManager(__pyx_type, long __pyx_checksum, __pyx_state)";
static PyMethodDef __pyx_mdef_14glycopeptidepy_2_c_9structure_6glycan_5__pyx_unpickle_GlycosylationManager = {"__pyx_unpickle_GlycosylationManager", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_5__pyx_unpickle_GlycosylationManager, METH_VARARGS|METH_KEYWORDS, __pyx_doc_14glycopeptidepy_2_c_9structure_6glycan_4__pyx_unpickle_GlycosylationManager};
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_5__pyx_unpickle_GlycosylationManager(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_GlycosylationManager (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_type)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_checksum)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_GlycosylationManager", 1, 3, 3, 1); __PYX_ERR(2, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_GlycosylationManager", 1, 3, 3, 2); __PYX_ERR(2, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_unpickle_GlycosylationManager") < 0)) __PYX_ERR(2, 1, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_GlycosylationManager", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(2, 1, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("glycopeptidepy._c.structure.glycan.__pyx_unpickle_GlycosylationManager", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_4__pyx_unpickle_GlycosylationManager(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_6glycan_4__pyx_unpickle_GlycosylationManager(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_GlycosylationManager", 0);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xeaeda1b, 0x38357ea, 0x45d3e33):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0xeaeda1b, 0x38357ea, 0x45d3e33) = (_aggregate, _proxy, _total_glycosylation_size, _type_track, mapping, parent))" % __pyx_checksum)
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__3, Py_NE)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xeaeda1b, 0x38357ea, 0x45d3e33):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0xeaeda1b, 0x38357ea, 0x45d3e33) = (_aggregate, _proxy, _total_glycosylation_size, _type_track, mapping, parent))" % __pyx_checksum)
 *     __pyx_result = GlycosylationManager.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError);
    __pyx_t_4 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, -1); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_4, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0xeaeda1b, 0x38357ea, 0x45d3e33):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0xeaeda1b, 0x38357ea, 0x45d3e33) = (_aggregate, _proxy, _total_glycosylation_size, _type_track, mapping, parent))" % __pyx_checksum)             # <<<<<<<<<<<<<<
 *     __pyx_result = GlycosylationManager.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_3, __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_INCREF(__pyx_v___pyx_PickleError);
    __pyx_t_1 = __pyx_v___pyx_PickleError; __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_4 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_6, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_5);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __PYX_ERR(2, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xeaeda1b, 0x38357ea, 0x45d3e33):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0xeaeda1b, 0x38357ea, 0x45d3e33) = (_aggregate, _proxy, _total_glycosylation_size, _type_track, mapping, parent))" % __pyx_checksum)
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0xeaeda1b, 0x38357ea, 0x45d3e33) = (_aggregate, _proxy, _total_glycosylation_size, _type_track, mapping, parent))" % __pyx_checksum)
 *     __pyx_result = GlycosylationManager.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_GlycosylationManager__set_state(<GlycosylationManager> __pyx_result, __pyx_state)
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager), __pyx_n_s_new); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_4 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_5, __pyx_v___pyx_type) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v___pyx_type);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0xeaeda1b, 0x38357ea, 0x45d3e33) = (_aggregate, _proxy, _total_glycosylation_size, _type_track, mapping, parent))" % __pyx_checksum)
 *     __pyx_result = GlycosylationManager.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_GlycosylationManager__set_state(<GlycosylationManager> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_3 = (__pyx_v___pyx_state != Py_None);
  __pyx_t_2 = (__pyx_t_3 != 0);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = GlycosylationManager.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_GlycosylationManager__set_state(<GlycosylationManager> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_GlycosylationManager__set_state(GlycosylationManager __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(2, 9, __pyx_L1_error)
    __pyx_t_4 = __pyx_f_14glycopeptidepy_2_c_9structure_6glycan___pyx_unpickle_GlycosylationManager__set_state(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0xeaeda1b, 0x38357ea, 0x45d3e33) = (_aggregate, _proxy, _total_glycosylation_size, _type_track, mapping, parent))" % __pyx_checksum)
 *     __pyx_result = GlycosylationManager.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_GlycosylationManager__set_state(<GlycosylationManager> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_GlycosylationManager__set_state(<GlycosylationManager> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_GlycosylationManager__set_state(GlycosylationManager __pyx_result, tuple __pyx_state):
 *     __pyx_result._aggregate = __pyx_state[0]; __pyx_result._proxy = __pyx_state[1]; __pyx_result._total_glycosylation_size = __pyx_state[2]; __pyx_result._type_track = __pyx_state[3]; __pyx_result.mapping = __pyx_state[4]; __pyx_result.parent = __pyx_state[5]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_GlycosylationManager(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.glycan.__pyx_unpickle_GlycosylationManager", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_GlycosylationManager__set_state(<GlycosylationManager> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_GlycosylationManager__set_state(GlycosylationManager __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result._aggregate = __pyx_state[0]; __pyx_result._proxy = __pyx_state[1]; __pyx_result._total_glycosylation_size = __pyx_state[2]; __pyx_result._type_track = __pyx_state[3]; __pyx_result.mapping = __pyx_state[4]; __pyx_result.parent = __pyx_state[5]
 *     if len(__pyx_state) > 6 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_14glycopeptidepy_2_c_9structure_6glycan___pyx_unpickle_GlycosylationManager__set_state(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_GlycosylationManager__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_GlycosylationManager__set_state(GlycosylationManager __pyx_result, tuple __pyx_state):
 *     __pyx_result._aggregate = __pyx_state[0]; __pyx_result._proxy = __pyx_state[1]; __pyx_result._total_glycosylation_size = __pyx_state[2]; __pyx_result._type_track = __pyx_state[3]; __pyx_result.mapping = __pyx_state[4]; __pyx_result.parent = __pyx_state[5]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 6 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[6])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->_aggregate);
  __Pyx_DECREF(__pyx_v___pyx_result->_aggregate);
  __pyx_v___pyx_result->_aggregate = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy))))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->_proxy);
  __Pyx_DECREF(((PyObject *)__pyx_v___pyx_result->_proxy));
  __pyx_v___pyx_result->_proxy = ((struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->_total_glycosylation_size = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyDict_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "dict", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->_type_track);
  __Pyx_DECREF(__pyx_v___pyx_result->_type_track);
  __pyx_v___pyx_result->_type_track = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 4, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyDict_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "dict", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->mapping);
  __Pyx_DECREF(__pyx_v___pyx_result->mapping);
  __pyx_v___pyx_result->mapping = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 5, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->parent);
  __Pyx_DECREF(__pyx_v___pyx_result->parent);
  __pyx_v___pyx_result->parent = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_GlycosylationManager__set_state(GlycosylationManager __pyx_result, tuple __pyx_state):
 *     __pyx_result._aggregate = __pyx_state[0]; __pyx_result._proxy = __pyx_state[1]; __pyx_result._total_glycosylation_size = __pyx_state[2]; __pyx_result._type_track = __pyx_state[3]; __pyx_result.mapping = __pyx_state[4]; __pyx_result.parent = __pyx_state[5]
 *     if len(__pyx_state) > 6 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[6])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(2, 13, __pyx_L1_error)
  }
  __pyx_t_4 = PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_5 = ((__pyx_t_4 > 6) != 0);
  if (__pyx_t_5) {
  } else {
    __pyx_t_3 = __pyx_t_5;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_5 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_6 = (__pyx_t_5 != 0);
  __pyx_t_3 = __pyx_t_6;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_3) {

    /* "(tree fragment)":14
 *     __pyx_result._aggregate = __pyx_state[0]; __pyx_result._proxy = __pyx_state[1]; __pyx_result._total_glycosylation_size = __pyx_state[2]; __pyx_result._type_track = __pyx_state[3]; __pyx_result.mapping = __pyx_state[4]; __pyx_result.parent = __pyx_state[5]
 *     if len(__pyx_state) > 6 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[6])             # <<<<<<<<<<<<<<
 */
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_update); if (unlikely(!__pyx_t_8)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(2, 14, __pyx_L1_error)
    }
    __pyx_t_7 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 6, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_9 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_8))) {
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_8);
      if (likely(__pyx_t_9)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_8, function);
      }
    }
    __pyx_t_1 = (__pyx_t_9) ? __Pyx_PyObject_Call2Args(__pyx_t_8, __pyx_t_9, __pyx_t_7) : __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_t_7);
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_GlycosylationManager__set_state(GlycosylationManager __pyx_result, tuple __pyx_state):
 *     __pyx_result._aggregate = __pyx_state[0]; __pyx_result._proxy = __pyx_state[1]; __pyx_result._total_glycosylation_size = __pyx_state[2]; __pyx_result._type_track = __pyx_state[3]; __pyx_result.mapping = __pyx_state[4]; __pyx_result.parent = __pyx_state[5]
 *     if len(__pyx_state) > 6 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[6])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_GlycosylationManager__set_state(<GlycosylationManager> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_GlycosylationManager__set_state(GlycosylationManager __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result._aggregate = __pyx_state[0]; __pyx_result._proxy = __pyx_state[1]; __pyx_result._total_glycosylation_size = __pyx_state[2]; __pyx_result._type_track = __pyx_state[3]; __pyx_result.mapping = __pyx_state[4]; __pyx_result.parent = __pyx_state[5]
 *     if len(__pyx_state) > 6 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.glycan.__pyx_unpickle_GlycosylationManager__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy __pyx_vtable_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy;

static struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *__pyx_freelist_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy[10000];
static int __pyx_freecount_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy = 0;

static PyObject *__pyx_tp_new_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *p;
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy)) & ((t->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)) == 0))) {
    o = (PyObject*)__pyx_freelist_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy[--__pyx_freecount_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy];
    memset(o, 0, sizeof(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
      o = (*t->tp_alloc)(t, 0);
    } else {
      o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
    }
    if (unlikely(!o)) return 0;
  }
  p = ((struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *)o);
  p->__pyx_vtab = __pyx_vtabptr_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy;
  p->obj = ((struct __pyx_obj_5glypy_2_c_9structure_18glycan_composition__CompositionBase *)Py_None); Py_INCREF(Py_None);
  p->_serialized = ((PyObject*)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy(PyObject *o) {
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *p = (struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->obj);
  Py_CLEAR(p->_serialized);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy < 10000) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy)) & ((Py_TYPE(o)->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)) == 0))) {
    __pyx_freelist_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy[__pyx_freecount_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy++] = ((struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *p = (struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *)o;
  if (p->obj) {
    e = (*v)(((PyObject *)p->obj), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *p = (struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *)o;
  tmp = ((PyObject*)p->obj);
  p->obj = ((struct __pyx_obj_5glypy_2_c_9structure_18glycan_composition__CompositionBase *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}
static PyObject *__pyx_sq_item_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy(PyObject *o, Py_ssize_t i) {
  PyObject *r;
  PyObject *x = PyInt_FromSsize_t(i); if(!x) return 0;
  r = Py_TYPE(o)->tp_as_mapping->mp_subscript(o, x);
  Py_DECREF(x);
  return r;
}

static PyObject *__pyx_tp_richcompare_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy(PyObject *o1, PyObject *o2, int op) {
  switch (op) {
    case Py_EQ: {
      return __pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_33__eq__(o1, o2);
    }
    case Py_NE: {
      return __pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_35__ne__(o1, o2);
    }
    default: {
      return __Pyx_NewRef(Py_NotImplemented);
    }
  }
}

static PyObject *__pyx_getprop_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_obj(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_3obj_1__get__(o);
}

static int __pyx_setprop_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_obj(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_3obj_3__set__(o, v);
  }
  else {
    return __pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_3obj_5__del__(o);
  }
}

static PyObject *__pyx_getprop_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy__serialized(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_11_serialized_1__get__(o);
}

static int __pyx_setprop_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy__serialized(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_11_serialized_3__set__(o, v);
  }
  else {
    return __pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_11_serialized_5__del__(o);
  }
}

static PyMethodDef __pyx_methods_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy[] = {
  {"mass", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_3mass, METH_VARARGS|METH_KEYWORDS, __pyx_doc_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_2mass},
  {"total_composition", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_5total_composition, METH_VARARGS|METH_KEYWORDS, __pyx_doc_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_4total_composition},
  {"_getitem_fast", (PyCFunction)__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_11_getitem_fast, METH_O, __pyx_doc_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_10_getitem_fast},
  {"keys", (PyCFunction)__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_15keys, METH_NOARGS, __pyx_doc_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_14keys},
  {"values", (PyCFunction)__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_17values, METH_NOARGS, __pyx_doc_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_16values},
  {"items", (PyCFunction)__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_19items, METH_NOARGS, __pyx_doc_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_18items},
  {"clone", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_21clone, METH_VARARGS|METH_KEYWORDS, __pyx_doc_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_20clone},
  {"query", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_23query, METH_VARARGS|METH_KEYWORDS, __pyx_doc_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_22query},
  {"serialize", (PyCFunction)__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_25serialize, METH_NOARGS, __pyx_doc_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_24serialize},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_45__reduce_cython__, METH_NOARGS, __pyx_doc_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_44__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_47__setstate_cython__, METH_O, __pyx_doc_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_46__setstate_cython__},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy[] = {
  {(char *)"obj", __pyx_getprop_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_obj, __pyx_setprop_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_obj, (char *)"obj: glypy._c.structure.glycan_composition._CompositionBase", 0},
  {(char *)"_serialized", __pyx_getprop_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy__serialized, __pyx_setprop_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy__serialized, (char *)"_serialized: str", 0},
  {0, 0, 0, 0, 0}
};

static PyNumberMethods __pyx_tp_as_number_GlycanCompositionProxy = {
  __pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_37__add__, /*nb_add*/
  __pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_39__sub__, /*nb_subtract*/
  __pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_41__mul__, /*nb_multiply*/
  #if PY_MAJOR_VERSION < 3 || (CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x03050000)
  0, /*nb_divide*/
  #endif
  0, /*nb_remainder*/
  0, /*nb_divmod*/
  0, /*nb_power*/
  0, /*nb_negative*/
  0, /*nb_positive*/
  0, /*nb_absolute*/
  0, /*nb_nonzero*/
  0, /*nb_invert*/
  0, /*nb_lshift*/
  0, /*nb_rshift*/
  0, /*nb_and*/
  0, /*nb_xor*/
  0, /*nb_or*/
  #if PY_MAJOR_VERSION < 3 || (CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x03050000)
  0, /*nb_coerce*/
  #endif
  0, /*nb_int*/
  #if PY_MAJOR_VERSION < 3
  0, /*nb_long*/
  #else
  0, /*reserved*/
  #endif
  0, /*nb_float*/
  #if PY_MAJOR_VERSION < 3 || (CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x03050000)
  0, /*nb_oct*/
  #endif
  #if PY_MAJOR_VERSION < 3 || (CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x03050000)
  0, /*nb_hex*/
  #endif
  0, /*nb_inplace_add*/
  0, /*nb_inplace_subtract*/
  0, /*nb_inplace_multiply*/
  #if PY_MAJOR_VERSION < 3 || (CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x03050000)
  0, /*nb_inplace_divide*/
  #endif
  0, /*nb_inplace_remainder*/
  0, /*nb_inplace_power*/
  0, /*nb_inplace_lshift*/
  0, /*nb_inplace_rshift*/
  0, /*nb_inplace_and*/
  0, /*nb_inplace_xor*/
  0, /*nb_inplace_or*/
  0, /*nb_floor_divide*/
  0, /*nb_true_divide*/
  0, /*nb_inplace_floor_divide*/
  0, /*nb_inplace_true_divide*/
  0, /*nb_index*/
  #if PY_VERSION_HEX >= 0x03050000
  0, /*nb_matrix_multiply*/
  #endif
  #if PY_VERSION_HEX >= 0x03050000
  0, /*nb_inplace_matrix_multiply*/
  #endif
};

static PySequenceMethods __pyx_tp_as_sequence_GlycanCompositionProxy = {
  __pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_43__len__, /*sq_length*/
  0, /*sq_concat*/
  0, /*sq_repeat*/
  __pyx_sq_item_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy, /*sq_item*/
  0, /*sq_slice*/
  0, /*sq_ass_item*/
  0, /*sq_ass_slice*/
  __pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_13__contains__, /*sq_contains*/
  0, /*sq_inplace_concat*/
  0, /*sq_inplace_repeat*/
};

static PyMappingMethods __pyx_tp_as_mapping_GlycanCompositionProxy = {
  __pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_43__len__, /*mp_length*/
  __pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_9__getitem__, /*mp_subscript*/
  0, /*mp_ass_subscript*/
};

static PyTypeObject __pyx_type_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy = {
  PyVarObject_HEAD_INIT(0, 0)
  "glycopeptidepy._c.structure.glycan.GlycanCompositionProxy", /*tp_name*/
  sizeof(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_27__repr__, /*tp_repr*/
  &__pyx_tp_as_number_GlycanCompositionProxy, /*tp_as_number*/
  &__pyx_tp_as_sequence_GlycanCompositionProxy, /*tp_as_sequence*/
  &__pyx_tp_as_mapping_GlycanCompositionProxy, /*tp_as_mapping*/
  __pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_31__hash__, /*tp_hash*/
  0, /*tp_call*/
  __pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_29__str__, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  "GlycanCompositionProxy(glycan_composition)\nA mapping-like object that imitates the GlycanComposition interface in\n    a read-only fashion.\n    ", /*tp_doc*/
  __pyx_tp_traverse_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy, /*tp_traverse*/
  __pyx_tp_clear_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy, /*tp_clear*/
  __pyx_tp_richcompare_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  __pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_7__iter__, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};
static struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionWithOffsetProxy __pyx_vtable_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionWithOffsetProxy;

static PyObject *__pyx_tp_new_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionWithOffsetProxy(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionWithOffsetProxy *p;
  PyObject *o = __pyx_tp_new_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionWithOffsetProxy *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy*)__pyx_vtabptr_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionWithOffsetProxy;
  p->composition_offset = ((struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionWithOffsetProxy(PyObject *o) {
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionWithOffsetProxy *p = (struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionWithOffsetProxy *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->composition_offset);
  PyObject_GC_Track(o);
  __pyx_tp_dealloc_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy(o);
}

static int __pyx_tp_traverse_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionWithOffsetProxy(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionWithOffsetProxy *p = (struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionWithOffsetProxy *)o;
  e = __pyx_tp_traverse_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy(o, v, a); if (e) return e;
  if (p->composition_offset) {
    e = (*v)(((PyObject *)p->composition_offset), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionWithOffsetProxy(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionWithOffsetProxy *p = (struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionWithOffsetProxy *)o;
  __pyx_tp_clear_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy(o);
  tmp = ((PyObject*)p->composition_offset);
  p->composition_offset = ((struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_14glycopeptidepy_2_c_9structure_6glycan_32GlycanCompositionWithOffsetProxy_composition_offset(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_32GlycanCompositionWithOffsetProxy_18composition_offset_1__get__(o);
}

static int __pyx_setprop_14glycopeptidepy_2_c_9structure_6glycan_32GlycanCompositionWithOffsetProxy_composition_offset(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_32GlycanCompositionWithOffsetProxy_18composition_offset_3__set__(o, v);
  }
  else {
    return __pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_32GlycanCompositionWithOffsetProxy_18composition_offset_5__del__(o);
  }
}

static PyMethodDef __pyx_methods_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionWithOffsetProxy[] = {
  {"clone", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_32GlycanCompositionWithOffsetProxy_3clone, METH_VARARGS|METH_KEYWORDS, __pyx_doc_14glycopeptidepy_2_c_9structure_6glycan_32GlycanCompositionWithOffsetProxy_2clone},
  {"mass", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_32GlycanCompositionWithOffsetProxy_5mass, METH_VARARGS|METH_KEYWORDS, __pyx_doc_14glycopeptidepy_2_c_9structure_6glycan_32GlycanCompositionWithOffsetProxy_4mass},
  {"total_composition", (PyCFunction)__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_32GlycanCompositionWithOffsetProxy_7total_composition, METH_NOARGS, __pyx_doc_14glycopeptidepy_2_c_9structure_6glycan_32GlycanCompositionWithOffsetProxy_6total_composition},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_32GlycanCompositionWithOffsetProxy_9__reduce_cython__, METH_NOARGS, __pyx_doc_14glycopeptidepy_2_c_9structure_6glycan_32GlycanCompositionWithOffsetProxy_8__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_32GlycanCompositionWithOffsetProxy_11__setstate_cython__, METH_O, __pyx_doc_14glycopeptidepy_2_c_9structure_6glycan_32GlycanCompositionWithOffsetProxy_10__setstate_cython__},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionWithOffsetProxy[] = {
  {(char *)"composition_offset", __pyx_getprop_14glycopeptidepy_2_c_9structure_6glycan_32GlycanCompositionWithOffsetProxy_composition_offset, __pyx_setprop_14glycopeptidepy_2_c_9structure_6glycan_32GlycanCompositionWithOffsetProxy_composition_offset, (char *)"composition_offset: glypy.composition.ccomposition.CComposition", 0},
  {0, 0, 0, 0, 0}
};

static PyTypeObject __pyx_type_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionWithOffsetProxy = {
  PyVarObject_HEAD_INIT(0, 0)
  "glycopeptidepy._c.structure.glycan.GlycanCompositionWithOffsetProxy", /*tp_name*/
  sizeof(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionWithOffsetProxy), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionWithOffsetProxy, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY
  __pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_27__repr__, /*tp_repr*/
  #else
  0, /*tp_repr*/
  #endif
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  #if CYTHON_COMPILING_IN_PYPY
  __pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_29__str__, /*tp_str*/
  #else
  0, /*tp_str*/
  #endif
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  "GlycanCompositionWithOffsetProxy(obj, offset=None)", /*tp_doc*/
  __pyx_tp_traverse_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionWithOffsetProxy, /*tp_traverse*/
  __pyx_tp_clear_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionWithOffsetProxy, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  #if CYTHON_COMPILING_IN_PYPY
  __pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_7__iter__, /*tp_iter*/
  #else
  0, /*tp_iter*/
  #endif
  0, /*tp_iternext*/
  __pyx_methods_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionWithOffsetProxy, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionWithOffsetProxy, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_32GlycanCompositionWithOffsetProxy_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionWithOffsetProxy, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};
static struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager __pyx_vtable_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager;

static struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *__pyx_freelist_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager[10000];
static int __pyx_freecount_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager = 0;

static PyObject *__pyx_tp_new_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *p;
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager)) & ((t->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)) == 0))) {
    o = (PyObject*)__pyx_freelist_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager[--__pyx_freecount_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager];
    memset(o, 0, sizeof(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
      o = (*t->tp_alloc)(t, 0);
    } else {
      o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
    }
    if (unlikely(!o)) return 0;
  }
  p = ((struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *)o);
  p->__pyx_vtab = __pyx_vtabptr_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager;
  p->mapping = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->parent = Py_None; Py_INCREF(Py_None);
  p->_aggregate = Py_None; Py_INCREF(Py_None);
  p->_proxy = ((struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *)Py_None); Py_INCREF(Py_None);
  p->_type_track = ((PyObject*)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager(PyObject *o) {
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *p = (struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->mapping);
  Py_CLEAR(p->parent);
  Py_CLEAR(p->_aggregate);
  Py_CLEAR(p->_proxy);
  Py_CLEAR(p->_type_track);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager < 10000) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager)) & ((Py_TYPE(o)->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)) == 0))) {
    __pyx_freelist_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager[__pyx_freecount_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager++] = ((struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *p = (struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *)o;
  if (p->mapping) {
    e = (*v)(p->mapping, a); if (e) return e;
  }
  if (p->parent) {
    e = (*v)(p->parent, a); if (e) return e;
  }
  if (p->_aggregate) {
    e = (*v)(p->_aggregate, a); if (e) return e;
  }
  if (p->_proxy) {
    e = (*v)(((PyObject *)p->_proxy), a); if (e) return e;
  }
  if (p->_type_track) {
    e = (*v)(p->_type_track, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *p = (struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *)o;
  tmp = ((PyObject*)p->mapping);
  p->mapping = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->parent);
  p->parent = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_aggregate);
  p->_aggregate = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_proxy);
  p->_proxy = ((struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_type_track);
  p->_type_track = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}
static PyObject *__pyx_sq_item_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager(PyObject *o, Py_ssize_t i) {
  PyObject *r;
  PyObject *x = PyInt_FromSsize_t(i); if(!x) return 0;
  r = Py_TYPE(o)->tp_as_mapping->mp_subscript(o, x);
  Py_DECREF(x);
  return r;
}

static int __pyx_mp_ass_subscript_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager(PyObject *o, PyObject *i, PyObject *v) {
  if (v) {
    return __pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_5__setitem__(o, i, v);
  }
  else {
    return __pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_7__delitem__(o, i);
  }
}

static PyObject *__pyx_getprop_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_aggregate(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_9aggregate_1__get__(o);
}

static int __pyx_setprop_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_aggregate(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_9aggregate_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_glycan_composition(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_18glycan_composition_1__get__(o);
}

static PyObject *__pyx_getprop_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_mapping(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_7mapping_1__get__(o);
}

static int __pyx_setprop_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_mapping(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_7mapping_3__set__(o, v);
  }
  else {
    return __pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_7mapping_5__del__(o);
  }
}

static PyObject *__pyx_getprop_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_parent(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_6parent_1__get__(o);
}

static int __pyx_setprop_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_parent(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_6parent_3__set__(o, v);
  }
  else {
    return __pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_6parent_5__del__(o);
  }
}

static PyObject *__pyx_getprop_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager__aggregate(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_10_aggregate_1__get__(o);
}

static int __pyx_setprop_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager__aggregate(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_10_aggregate_3__set__(o, v);
  }
  else {
    return __pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_10_aggregate_5__del__(o);
  }
}

static PyObject *__pyx_getprop_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager__proxy(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_6_proxy_1__get__(o);
}

static int __pyx_setprop_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager__proxy(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_6_proxy_3__set__(o, v);
  }
  else {
    return __pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_6_proxy_5__del__(o);
  }
}

static PyObject *__pyx_getprop_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager__type_track(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_11_type_track_1__get__(o);
}

static int __pyx_setprop_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager__type_track(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_11_type_track_3__set__(o, v);
  }
  else {
    return __pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_11_type_track_5__del__(o);
  }
}

static PyObject *__pyx_getprop_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager__total_glycosylation_size(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_25_total_glycosylation_size_1__get__(o);
}

static int __pyx_setprop_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager__total_glycosylation_size(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_25_total_glycosylation_size_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyMethodDef __pyx_methods_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager[] = {
  {"keys", (PyCFunction)__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_11keys, METH_NOARGS, __pyx_doc_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_10keys},
  {"values", (PyCFunction)__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_13values, METH_NOARGS, __pyx_doc_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_12values},
  {"items", (PyCFunction)__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_15items, METH_NOARGS, __pyx_doc_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_14items},
  {"pop", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_17pop, METH_VARARGS|METH_KEYWORDS, __pyx_doc_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_16pop},
  {"clear", (PyCFunction)__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_19clear, METH_NOARGS, __pyx_doc_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_18clear},
  {"update", (PyCFunction)__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_21update, METH_O, __pyx_doc_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_20update},
  {"copy", (PyCFunction)__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_23copy, METH_NOARGS, __pyx_doc_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_22copy},
  {"clone", (PyCFunction)__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_25clone, METH_NOARGS, __pyx_doc_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_24clone},
  {"_patch_aggregate", (PyCFunction)__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_27_patch_aggregate, METH_NOARGS, __pyx_doc_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_26_patch_aggregate},
  {"invalidate", (PyCFunction)__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_29invalidate, METH_NOARGS, __pyx_doc_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_28invalidate},
  {"total_composition", (PyCFunction)__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_31total_composition, METH_NOARGS, __pyx_doc_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_30total_composition},
  {"mass", (PyCFunction)__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_33mass, METH_NOARGS, __pyx_doc_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_32mass},
  {"total_glycosylation_size", (PyCFunction)__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_35total_glycosylation_size, METH_NOARGS, __pyx_doc_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_34total_glycosylation_size},
  {"is_fully_specified_topologies", (PyCFunction)__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_37is_fully_specified_topologies, METH_NOARGS, __pyx_doc_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_36is_fully_specified_topologies},
  {"count_glycosylation_type", (PyCFunction)__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_39count_glycosylation_type, METH_O, __pyx_doc_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_38count_glycosylation_type},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_43__reduce_cython__, METH_NOARGS, __pyx_doc_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_42__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_45__setstate_cython__, METH_O, __pyx_doc_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_44__setstate_cython__},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager[] = {
  {(char *)"aggregate", __pyx_getprop_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_aggregate, __pyx_setprop_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_aggregate, (char *)0, 0},
  {(char *)"glycan_composition", __pyx_getprop_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_glycan_composition, 0, (char *)0, 0},
  {(char *)"mapping", __pyx_getprop_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_mapping, __pyx_setprop_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_mapping, (char *)"mapping: dict", 0},
  {(char *)"parent", __pyx_getprop_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_parent, __pyx_setprop_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_parent, (char *)"parent: object", 0},
  {(char *)"_aggregate", __pyx_getprop_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager__aggregate, __pyx_setprop_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager__aggregate, (char *)"_aggregate: object", 0},
  {(char *)"_proxy", __pyx_getprop_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager__proxy, __pyx_setprop_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager__proxy, (char *)"_proxy: glycopeptidepy._c.structure.glycan.GlycanCompositionProxy", 0},
  {(char *)"_type_track", __pyx_getprop_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager__type_track, __pyx_setprop_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager__type_track, (char *)"_type_track: dict", 0},
  {(char *)"_total_glycosylation_size", __pyx_getprop_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager__total_glycosylation_size, __pyx_setprop_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager__total_glycosylation_size, (char *)"_total_glycosylation_size: 'int'", 0},
  {0, 0, 0, 0, 0}
};

static PySequenceMethods __pyx_tp_as_sequence_GlycosylationManager = {
  __pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_9__len__, /*sq_length*/
  0, /*sq_concat*/
  0, /*sq_repeat*/
  __pyx_sq_item_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager, /*sq_item*/
  0, /*sq_slice*/
  0, /*sq_ass_item*/
  0, /*sq_ass_slice*/
  0, /*sq_contains*/
  0, /*sq_inplace_concat*/
  0, /*sq_inplace_repeat*/
};

static PyMappingMethods __pyx_tp_as_mapping_GlycosylationManager = {
  __pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_9__len__, /*mp_length*/
  __pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_3__getitem__, /*mp_subscript*/
  __pyx_mp_ass_subscript_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager, /*mp_ass_subscript*/
};

static PyTypeObject __pyx_type_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager = {
  PyVarObject_HEAD_INIT(0, 0)
  "glycopeptidepy._c.structure.glycan.GlycosylationManager", /*tp_name*/
  sizeof(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_41__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  &__pyx_tp_as_sequence_GlycosylationManager, /*tp_as_sequence*/
  &__pyx_tp_as_mapping_GlycosylationManager, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  "GlycosylationManager(parent, aggregate=None)", /*tp_doc*/
  __pyx_tp_traverse_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager, /*tp_traverse*/
  __pyx_tp_clear_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};

static PyMethodDef __pyx_methods[] = {
  {0, 0, 0, 0}
};

#if PY_MAJOR_VERSION >= 3
#if CYTHON_PEP489_MULTI_PHASE_INIT
static PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def); /*proto*/
static int __pyx_pymod_exec_glycan(PyObject* module); /*proto*/
static PyModuleDef_Slot __pyx_moduledef_slots[] = {
  {Py_mod_create, (void*)__pyx_pymod_create},
  {Py_mod_exec, (void*)__pyx_pymod_exec_glycan},
  {0, NULL}
};
#endif

static struct PyModuleDef __pyx_moduledef = {
    PyModuleDef_HEAD_INIT,
    "glycan",
    0, /* m_doc */
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    0, /* m_size */
  #else
    -1, /* m_size */
  #endif
    __pyx_methods /* m_methods */,
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    __pyx_moduledef_slots, /* m_slots */
  #else
    NULL, /* m_reload */
  #endif
    NULL, /* m_traverse */
    NULL, /* m_clear */
    NULL /* m_free */
};
#endif
#ifndef CYTHON_SMALL_CODE
#if defined(__clang__)
    #define CYTHON_SMALL_CODE
#elif defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 3))
    #define CYTHON_SMALL_CODE __attribute__((cold))
#else
    #define CYTHON_SMALL_CODE
#endif
#endif

static __Pyx_StringTabEntry __pyx_string_tab[] = {
  {&__pyx_n_s_AttributeError, __pyx_k_AttributeError, sizeof(__pyx_k_AttributeError), 0, 0, 1, 1},
  {&__pyx_n_s_GlycanCompositionProxy, __pyx_k_GlycanCompositionProxy, sizeof(__pyx_k_GlycanCompositionProxy), 0, 0, 1, 1},
  {&__pyx_n_s_GlycanCompositionWithOffsetProxy, __pyx_k_GlycanCompositionWithOffsetProxy, sizeof(__pyx_k_GlycanCompositionWithOffsetProxy), 0, 0, 1, 1},
  {&__pyx_n_s_GlycosylationManager, __pyx_k_GlycosylationManager, sizeof(__pyx_k_GlycosylationManager), 0, 0, 1, 1},
  {&__pyx_n_s_H, __pyx_k_H, sizeof(__pyx_k_H), 0, 0, 1, 1},
  {&__pyx_n_s_HashableGlycanComposition, __pyx_k_HashableGlycanComposition, sizeof(__pyx_k_HashableGlycanComposition), 0, 0, 1, 1},
  {&__pyx_kp_s_Incompatible_checksums_0x_x_vs_0, __pyx_k_Incompatible_checksums_0x_x_vs_0, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0), 0, 0, 1, 0},
  {&__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_2, __pyx_k_Incompatible_checksums_0x_x_vs_0_2, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0_2), 0, 0, 1, 0},
  {&__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_3, __pyx_k_Incompatible_checksums_0x_x_vs_0_3, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0_3), 0, 0, 1, 0},
  {&__pyx_n_s_O, __pyx_k_O, sizeof(__pyx_k_O), 0, 0, 1, 1},
  {&__pyx_n_s_PickleError, __pyx_k_PickleError, sizeof(__pyx_k_PickleError), 0, 0, 1, 1},
  {&__pyx_n_s_aggregate, __pyx_k_aggregate, sizeof(__pyx_k_aggregate), 0, 0, 1, 1},
  {&__pyx_n_s_class, __pyx_k_class, sizeof(__pyx_k_class), 0, 0, 1, 1},
  {&__pyx_n_s_clear, __pyx_k_clear, sizeof(__pyx_k_clear), 0, 0, 1, 1},
  {&__pyx_n_s_cline_in_traceback, __pyx_k_cline_in_traceback, sizeof(__pyx_k_cline_in_traceback), 0, 0, 1, 1},
  {&__pyx_n_s_clone, __pyx_k_clone, sizeof(__pyx_k_clone), 0, 0, 1, 1},
  {&__pyx_n_s_composition_offset, __pyx_k_composition_offset, sizeof(__pyx_k_composition_offset), 0, 0, 1, 1},
  {&__pyx_n_s_copy, __pyx_k_copy, sizeof(__pyx_k_copy), 0, 0, 1, 1},
  {&__pyx_n_s_count_glycosylation_type, __pyx_k_count_glycosylation_type, sizeof(__pyx_k_count_glycosylation_type), 0, 0, 1, 1},
  {&__pyx_n_s_default, __pyx_k_default, sizeof(__pyx_k_default), 0, 0, 1, 1},
  {&__pyx_n_s_dict, __pyx_k_dict, sizeof(__pyx_k_dict), 0, 0, 1, 1},
  {&__pyx_n_s_exact, __pyx_k_exact, sizeof(__pyx_k_exact), 0, 0, 1, 1},
  {&__pyx_n_s_format, __pyx_k_format, sizeof(__pyx_k_format), 0, 0, 1, 1},
  {&__pyx_n_s_from_glycan, __pyx_k_from_glycan, sizeof(__pyx_k_from_glycan), 0, 0, 1, 1},
  {&__pyx_n_s_getitem_fast, __pyx_k_getitem_fast, sizeof(__pyx_k_getitem_fast), 0, 0, 1, 1},
  {&__pyx_n_s_getstate, __pyx_k_getstate, sizeof(__pyx_k_getstate), 0, 0, 1, 1},
  {&__pyx_n_s_glycan, __pyx_k_glycan, sizeof(__pyx_k_glycan), 0, 0, 1, 1},
  {&__pyx_n_s_glycan_composition, __pyx_k_glycan_composition, sizeof(__pyx_k_glycan_composition), 0, 0, 1, 1},
  {&__pyx_n_s_glycopeptidepy__c_structure_glyc, __pyx_k_glycopeptidepy__c_structure_glyc, sizeof(__pyx_k_glycopeptidepy__c_structure_glyc), 0, 0, 1, 1},
  {&__pyx_n_s_glypy_structure_glycan_compositi, __pyx_k_glypy_structure_glycan_compositi, sizeof(__pyx_k_glypy_structure_glycan_compositi), 0, 0, 1, 1},
  {&__pyx_n_s_import, __pyx_k_import, sizeof(__pyx_k_import), 0, 0, 1, 1},
  {&__pyx_n_s_init, __pyx_k_init, sizeof(__pyx_k_init), 0, 0, 1, 1},
  {&__pyx_n_s_invalidate, __pyx_k_invalidate, sizeof(__pyx_k_invalidate), 0, 0, 1, 1},
  {&__pyx_n_s_is_composition, __pyx_k_is_composition, sizeof(__pyx_k_is_composition), 0, 0, 1, 1},
  {&__pyx_n_s_is_core, __pyx_k_is_core, sizeof(__pyx_k_is_core), 0, 0, 1, 1},
  {&__pyx_n_s_is_fully_specified_topologies, __pyx_k_is_fully_specified_topologies, sizeof(__pyx_k_is_fully_specified_topologies), 0, 0, 1, 1},
  {&__pyx_n_s_items, __pyx_k_items, sizeof(__pyx_k_items), 0, 0, 1, 1},
  {&__pyx_n_s_key, __pyx_k_key, sizeof(__pyx_k_key), 0, 0, 1, 1},
  {&__pyx_n_s_keys, __pyx_k_keys, sizeof(__pyx_k_keys), 0, 0, 1, 1},
  {&__pyx_n_s_main, __pyx_k_main, sizeof(__pyx_k_main), 0, 0, 1, 1},
  {&__pyx_n_s_mapping, __pyx_k_mapping, sizeof(__pyx_k_mapping), 0, 0, 1, 1},
  {&__pyx_n_s_mass, __pyx_k_mass, sizeof(__pyx_k_mass), 0, 0, 1, 1},
  {&__pyx_n_s_name, __pyx_k_name, sizeof(__pyx_k_name), 0, 0, 1, 1},
  {&__pyx_n_s_new, __pyx_k_new, sizeof(__pyx_k_new), 0, 0, 1, 1},
  {&__pyx_n_s_obj, __pyx_k_obj, sizeof(__pyx_k_obj), 0, 0, 1, 1},
  {&__pyx_n_s_offset, __pyx_k_offset, sizeof(__pyx_k_offset), 0, 0, 1, 1},
  {&__pyx_n_s_original, __pyx_k_original, sizeof(__pyx_k_original), 0, 0, 1, 1},
  {&__pyx_n_s_parent, __pyx_k_parent, sizeof(__pyx_k_parent), 0, 0, 1, 1},
  {&__pyx_n_s_patch_aggregate, __pyx_k_patch_aggregate, sizeof(__pyx_k_patch_aggregate), 0, 0, 1, 1},
  {&__pyx_n_s_pickle, __pyx_k_pickle, sizeof(__pyx_k_pickle), 0, 0, 1, 1},
  {&__pyx_n_s_pop, __pyx_k_pop, sizeof(__pyx_k_pop), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_PickleError, __pyx_k_pyx_PickleError, sizeof(__pyx_k_pyx_PickleError), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_checksum, __pyx_k_pyx_checksum, sizeof(__pyx_k_pyx_checksum), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_result, __pyx_k_pyx_result, sizeof(__pyx_k_pyx_result), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_state, __pyx_k_pyx_state, sizeof(__pyx_k_pyx_state), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_type, __pyx_k_pyx_type, sizeof(__pyx_k_pyx_type), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_unpickle_GlycanComposition, __pyx_k_pyx_unpickle_GlycanComposition, sizeof(__pyx_k_pyx_unpickle_GlycanComposition), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_unpickle_GlycanComposition_2, __pyx_k_pyx_unpickle_GlycanComposition_2, sizeof(__pyx_k_pyx_unpickle_GlycanComposition_2), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_unpickle_GlycosylationMana, __pyx_k_pyx_unpickle_GlycosylationMana, sizeof(__pyx_k_pyx_unpickle_GlycosylationMana), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_vtable, __pyx_k_pyx_vtable, sizeof(__pyx_k_pyx_vtable), 0, 0, 1, 1},
  {&__pyx_n_s_query, __pyx_k_query, sizeof(__pyx_k_query), 0, 0, 1, 1},
  {&__pyx_n_s_reduce, __pyx_k_reduce, sizeof(__pyx_k_reduce), 0, 0, 1, 1},
  {&__pyx_n_s_reduce_cython, __pyx_k_reduce_cython, sizeof(__pyx_k_reduce_cython), 0, 0, 1, 1},
  {&__pyx_n_s_reduce_ex, __pyx_k_reduce_ex, sizeof(__pyx_k_reduce_ex), 0, 0, 1, 1},
  {&__pyx_n_s_self, __pyx_k_self, sizeof(__pyx_k_self), 0, 0, 1, 1},
  {&__pyx_kp_s_self___class_____name___self_ma, __pyx_k_self___class_____name___self_ma, sizeof(__pyx_k_self___class_____name___self_ma), 0, 0, 1, 0},
  {&__pyx_n_s_serialize, __pyx_k_serialize, sizeof(__pyx_k_serialize), 0, 0, 1, 1},
  {&__pyx_n_s_serialized, __pyx_k_serialized, sizeof(__pyx_k_serialized), 0, 0, 1, 1},
  {&__pyx_n_s_setstate, __pyx_k_setstate, sizeof(__pyx_k_setstate), 0, 0, 1, 1},
  {&__pyx_n_s_setstate_cython, __pyx_k_setstate_cython, sizeof(__pyx_k_setstate_cython), 0, 0, 1, 1},
  {&__pyx_kp_s_stringsource, __pyx_k_stringsource, sizeof(__pyx_k_stringsource), 0, 0, 1, 0},
  {&__pyx_n_s_sum, __pyx_k_sum, sizeof(__pyx_k_sum), 0, 0, 1, 1},
  {&__pyx_n_s_test, __pyx_k_test, sizeof(__pyx_k_test), 0, 0, 1, 1},
  {&__pyx_n_s_total_composition, __pyx_k_total_composition, sizeof(__pyx_k_total_composition), 0, 0, 1, 1},
  {&__pyx_n_s_total_glycosylation_size, __pyx_k_total_glycosylation_size, sizeof(__pyx_k_total_glycosylation_size), 0, 0, 1, 1},
  {&__pyx_n_s_update, __pyx_k_update, sizeof(__pyx_k_update), 0, 0, 1, 1},
  {&__pyx_n_s_values, __pyx_k_values, sizeof(__pyx_k_values), 0, 0, 1, 1},
  {0, 0, 0, 0, 0, 0, 0}
};
static CYTHON_SMALL_CODE int __Pyx_InitCachedBuiltins(void) {
  __pyx_builtin_AttributeError = __Pyx_GetBuiltinName(__pyx_n_s_AttributeError); if (!__pyx_builtin_AttributeError) __PYX_ERR(0, 95, __pyx_L1_error)
  __pyx_builtin_sum = __Pyx_GetBuiltinName(__pyx_n_s_sum); if (!__pyx_builtin_sum) __PYX_ERR(0, 352, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x4126d5b, 0x40511ef, 0x04f04be):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0x4126d5b, 0x40511ef, 0x04f04be) = (_serialized, obj))" % __pyx_checksum)
 */
  __pyx_tuple_ = PyTuple_Pack(3, __pyx_int_68316507, __pyx_int_67441135, __pyx_int_5178558); if (unlikely(!__pyx_tuple_)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple_);
  __Pyx_GIVEREF(__pyx_tuple_);
  __pyx_tuple__2 = PyTuple_Pack(3, __pyx_int_53161881, __pyx_int_116448577, __pyx_int_54009504); if (unlikely(!__pyx_tuple__2)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2);
  __Pyx_GIVEREF(__pyx_tuple__2);
  __pyx_tuple__3 = PyTuple_Pack(3, __pyx_int_246340123, __pyx_int_58939370, __pyx_int_73219635); if (unlikely(!__pyx_tuple__3)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__3);
  __Pyx_GIVEREF(__pyx_tuple__3);

  /* "(tree fragment)":1
 * def __pyx_unpickle_GlycanCompositionProxy(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_tuple__4 = PyTuple_Pack(5, __pyx_n_s_pyx_type, __pyx_n_s_pyx_checksum, __pyx_n_s_pyx_state, __pyx_n_s_pyx_PickleError, __pyx_n_s_pyx_result); if (unlikely(!__pyx_tuple__4)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__4);
  __Pyx_GIVEREF(__pyx_tuple__4);
  __pyx_codeobj__5 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__4, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_GlycanComposition, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__5)) __PYX_ERR(2, 1, __pyx_L1_error)
  __pyx_tuple__6 = PyTuple_Pack(5, __pyx_n_s_pyx_type, __pyx_n_s_pyx_checksum, __pyx_n_s_pyx_state, __pyx_n_s_pyx_PickleError, __pyx_n_s_pyx_result); if (unlikely(!__pyx_tuple__6)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__6);
  __Pyx_GIVEREF(__pyx_tuple__6);
  __pyx_codeobj__7 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__6, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_GlycanComposition_2, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__7)) __PYX_ERR(2, 1, __pyx_L1_error)
  __pyx_tuple__8 = PyTuple_Pack(5, __pyx_n_s_pyx_type, __pyx_n_s_pyx_checksum, __pyx_n_s_pyx_state, __pyx_n_s_pyx_PickleError, __pyx_n_s_pyx_result); if (unlikely(!__pyx_tuple__8)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__8);
  __Pyx_GIVEREF(__pyx_tuple__8);
  __pyx_codeobj__9 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__8, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_GlycosylationMana, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__9)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitGlobals(void) {
  __pyx_umethod_PyDict_Type_items.type = (PyObject*)&PyDict_Type;
  __pyx_umethod_PyDict_Type_keys.type = (PyObject*)&PyDict_Type;
  __pyx_umethod_PyDict_Type_pop.type = (PyObject*)&PyDict_Type;
  __pyx_umethod_PyDict_Type_values.type = (PyObject*)&PyDict_Type;
  if (__Pyx_InitStrings(__pyx_string_tab) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_1 = PyInt_FromLong(1); if (unlikely(!__pyx_int_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_2 = PyInt_FromLong(2); if (unlikely(!__pyx_int_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_5178558 = PyInt_FromLong(5178558L); if (unlikely(!__pyx_int_5178558)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_53161881 = PyInt_FromLong(53161881L); if (unlikely(!__pyx_int_53161881)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_54009504 = PyInt_FromLong(54009504L); if (unlikely(!__pyx_int_54009504)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_58939370 = PyInt_FromLong(58939370L); if (unlikely(!__pyx_int_58939370)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_67441135 = PyInt_FromLong(67441135L); if (unlikely(!__pyx_int_67441135)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_68316507 = PyInt_FromLong(68316507L); if (unlikely(!__pyx_int_68316507)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_73219635 = PyInt_FromLong(73219635L); if (unlikely(!__pyx_int_73219635)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_116448577 = PyInt_FromLong(116448577L); if (unlikely(!__pyx_int_116448577)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_246340123 = PyInt_FromLong(246340123L); if (unlikely(!__pyx_int_246340123)) __PYX_ERR(0, 1, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_modinit_global_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_import_code(void); /*proto*/

static int __Pyx_modinit_global_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_global_init_code", 0);
  /*--- Global init code ---*/
  __pyx_v_14glycopeptidepy_2_c_9structure_6glycan_WATER_OFFSET = ((struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *)Py_None); Py_INCREF(Py_None);
  __pyx_v_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxyType = Py_None; Py_INCREF(Py_None);
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_export_code", 0);
  /*--- Variable export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_export_code", 0);
  /*--- Function export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_type_init_code(void) {
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_init_code", 0);
  /*--- Type init code ---*/
  __pyx_vtabptr_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy = &__pyx_vtable_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy;
  __pyx_vtable_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy._getitem_fast = (PyObject *(*)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *, PyObject *, int __pyx_skip_dispatch))__pyx_f_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy__getitem_fast;
  __pyx_vtable_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy.serialize = (PyObject *(*)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *, int __pyx_skip_dispatch))__pyx_f_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy_serialize;
  __pyx_vtable_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy.__pyx_fuse_0_create = (struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *(*)(struct __pyx_obj_5glypy_2_c_9structure_18glycan_composition__CompositionBase *))__pyx_fuse_0__pyx_f_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy__create;
  __pyx_vtable_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy.__pyx_fuse_1_create = (struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *(*)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *))__pyx_fuse_1__pyx_f_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy__create;
  __pyx_vtable_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy.__pyx_fuse_2_create = (struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *(*)(PyObject *))__pyx_fuse_2__pyx_f_14glycopeptidepy_2_c_9structure_6glycan_22GlycanCompositionProxy__create;
  if (PyType_Ready(&__pyx_type_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy) < 0) __PYX_ERR(0, 16, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy.tp_dictoffset && __pyx_type_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy.tp_dict, __pyx_vtabptr_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy) < 0) __PYX_ERR(0, 16, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_GlycanCompositionProxy, (PyObject *)&__pyx_type_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy) < 0) __PYX_ERR(0, 16, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy) < 0) __PYX_ERR(0, 16, __pyx_L1_error)
  __pyx_ptype_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy = &__pyx_type_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy;
  __pyx_vtabptr_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionWithOffsetProxy = &__pyx_vtable_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionWithOffsetProxy;
  __pyx_vtable_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionWithOffsetProxy.__pyx_base = *__pyx_vtabptr_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy;
  __pyx_type_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionWithOffsetProxy.tp_base = __pyx_ptype_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy;
  if (PyType_Ready(&__pyx_type_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionWithOffsetProxy) < 0) __PYX_ERR(0, 119, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionWithOffsetProxy.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionWithOffsetProxy.tp_dictoffset && __pyx_type_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionWithOffsetProxy.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionWithOffsetProxy.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionWithOffsetProxy.tp_dict, __pyx_vtabptr_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionWithOffsetProxy) < 0) __PYX_ERR(0, 119, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_GlycanCompositionWithOffsetProxy, (PyObject *)&__pyx_type_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionWithOffsetProxy) < 0) __PYX_ERR(0, 119, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionWithOffsetProxy) < 0) __PYX_ERR(0, 119, __pyx_L1_error)
  __pyx_ptype_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionWithOffsetProxy = &__pyx_type_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionWithOffsetProxy;
  __pyx_vtabptr_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager = &__pyx_vtable_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager;
  __pyx_vtable_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager.update = (PyObject *(*)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *, PyObject *, int __pyx_skip_dispatch))__pyx_f_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_update;
  __pyx_vtable_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager.clear = (PyObject *(*)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *, int __pyx_skip_dispatch))__pyx_f_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_clear;
  __pyx_vtable_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager.keys = (PyObject *(*)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *, int __pyx_skip_dispatch))__pyx_f_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_keys;
  __pyx_vtable_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager.values = (PyObject *(*)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *, int __pyx_skip_dispatch))__pyx_f_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_values;
  __pyx_vtable_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager.items = (PyObject *(*)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *, int __pyx_skip_dispatch))__pyx_f_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_items;
  __pyx_vtable_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager.pop = (PyObject *(*)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *, PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_pop *__pyx_optional_args))__pyx_f_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_pop;
  __pyx_vtable_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager.get_size = (size_t (*)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *))__pyx_f_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_get_size;
  __pyx_vtable_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager.copy = (struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *(*)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *, int __pyx_skip_dispatch))__pyx_f_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_copy;
  __pyx_vtable_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager.clone = (struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *(*)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *, int __pyx_skip_dispatch))__pyx_f_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_clone;
  __pyx_vtable_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager.invalidate = (PyObject *(*)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *, int __pyx_skip_dispatch))__pyx_f_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_invalidate;
  __pyx_vtable_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager.get_aggregate = (PyObject *(*)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *))__pyx_f_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_get_aggregate;
  __pyx_vtable_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager.set_aggregate = (int (*)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *, PyObject *))__pyx_f_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_set_aggregate;
  __pyx_vtable_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager._patch_aggregate = (PyObject *(*)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *, int __pyx_skip_dispatch))__pyx_f_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager__patch_aggregate;
  __pyx_vtable_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager.get_glycosylation_types = (PyObject *(*)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *))__pyx_f_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_get_glycosylation_types;
  __pyx_vtable_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager._update_track = (PyObject *(*)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *))__pyx_f_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager__update_track;
  __pyx_vtable_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager.total_composition = (struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *(*)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *, int __pyx_skip_dispatch))__pyx_f_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_total_composition;
  __pyx_vtable_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager.mass = (double (*)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *, int __pyx_skip_dispatch))__pyx_f_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_mass;
  __pyx_vtable_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager.is_fully_specified_topologies = (int (*)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *, int __pyx_skip_dispatch))__pyx_f_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_is_fully_specified_topologies;
  __pyx_vtable_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager.get_glycan_composition = (struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *(*)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *))__pyx_f_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_get_glycan_composition;
  __pyx_vtable_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager._make_glycan_composition_proxy = (struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxy *(*)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *))__pyx_f_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager__make_glycan_composition_proxy;
  __pyx_vtable_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager.get_total_glycosylation_size = (int (*)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *))__pyx_f_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_get_total_glycosylation_size;
  __pyx_vtable_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager.total_glycosylation_size = (int (*)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *, int __pyx_skip_dispatch))__pyx_f_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_total_glycosylation_size;
  __pyx_vtable_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager.count_glycosylation_type = (int (*)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *, PyObject *, int __pyx_skip_dispatch))__pyx_f_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager_count_glycosylation_type;
  __pyx_vtable_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager._init = (void (*)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *))__pyx_f_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager__init;
  __pyx_vtable_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager._create = (struct __pyx_obj_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager *(*)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_PeptideSequenceBase *, PyObject *))__pyx_f_14glycopeptidepy_2_c_9structure_6glycan_20GlycosylationManager__create;
  if (PyType_Ready(&__pyx_type_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager) < 0) __PYX_ERR(0, 146, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager.tp_dictoffset && __pyx_type_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager.tp_dict, __pyx_vtabptr_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager) < 0) __PYX_ERR(0, 146, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_GlycosylationManager, (PyObject *)&__pyx_type_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager) < 0) __PYX_ERR(0, 146, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager) < 0) __PYX_ERR(0, 146, __pyx_L1_error)
  __pyx_ptype_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager = &__pyx_type_14glycopeptidepy_2_c_9structure_6glycan_GlycosylationManager;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_import_code", 0);
  /*--- Type import code ---*/
  __pyx_t_1 = PyImport_ImportModule("glypy.composition.ccomposition"); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_5glypy_11composition_12ccomposition_CComposition = __Pyx_ImportType(__pyx_t_1, "glypy.composition.ccomposition", "CComposition", sizeof(struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5glypy_11composition_12ccomposition_CComposition) __PYX_ERR(3, 2, __pyx_L1_error)
  __pyx_vtabptr_5glypy_11composition_12ccomposition_CComposition = (struct __pyx_vtabstruct_5glypy_11composition_12ccomposition_CComposition*)__Pyx_GetVtable(__pyx_ptype_5glypy_11composition_12ccomposition_CComposition->tp_dict); if (unlikely(!__pyx_vtabptr_5glypy_11composition_12ccomposition_CComposition)) __PYX_ERR(3, 2, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("glypy._c.structure.glycan_composition"); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_5glypy_2_c_9structure_18glycan_composition__CompositionBase = __Pyx_ImportType(__pyx_t_1, "glypy._c.structure.glycan_composition", "_CompositionBase", sizeof(struct __pyx_obj_5glypy_2_c_9structure_18glycan_composition__CompositionBase), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5glypy_2_c_9structure_18glycan_composition__CompositionBase) __PYX_ERR(4, 3, __pyx_L1_error)
  __pyx_vtabptr_5glypy_2_c_9structure_18glycan_composition__CompositionBase = (struct __pyx_vtabstruct_5glypy_2_c_9structure_18glycan_composition__CompositionBase*)__Pyx_GetVtable(__pyx_ptype_5glypy_2_c_9structure_18glycan_composition__CompositionBase->tp_dict); if (unlikely(!__pyx_vtabptr_5glypy_2_c_9structure_18glycan_composition__CompositionBase)) __PYX_ERR(4, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("glycopeptidepy._c.structure.base"); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_14glycopeptidepy_2_c_9structure_4base_PeptideSequenceBase = __Pyx_ImportType(__pyx_t_1, "glycopeptidepy._c.structure.base", "PeptideSequenceBase", sizeof(struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_PeptideSequenceBase), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_14glycopeptidepy_2_c_9structure_4base_PeptideSequenceBase) __PYX_ERR(5, 5, __pyx_L1_error)
  __pyx_vtabptr_14glycopeptidepy_2_c_9structure_4base_PeptideSequenceBase = (struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_4base_PeptideSequenceBase*)__Pyx_GetVtable(__pyx_ptype_14glycopeptidepy_2_c_9structure_4base_PeptideSequenceBase->tp_dict); if (unlikely(!__pyx_vtabptr_14glycopeptidepy_2_c_9structure_4base_PeptideSequenceBase)) __PYX_ERR(5, 5, __pyx_L1_error)
  __pyx_ptype_14glycopeptidepy_2_c_9structure_4base_TerminalGroup = __Pyx_ImportType(__pyx_t_1, "glycopeptidepy._c.structure.base", "TerminalGroup", sizeof(struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_TerminalGroup), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_14glycopeptidepy_2_c_9structure_4base_TerminalGroup) __PYX_ERR(5, 21, __pyx_L1_error)
  __pyx_vtabptr_14glycopeptidepy_2_c_9structure_4base_TerminalGroup = (struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_4base_TerminalGroup*)__Pyx_GetVtable(__pyx_ptype_14glycopeptidepy_2_c_9structure_4base_TerminalGroup->tp_dict); if (unlikely(!__pyx_vtabptr_14glycopeptidepy_2_c_9structure_4base_TerminalGroup)) __PYX_ERR(5, 21, __pyx_L1_error)
  __pyx_ptype_14glycopeptidepy_2_c_9structure_4base_AminoAcidResidueBase = __Pyx_ImportType(__pyx_t_1, "glycopeptidepy._c.structure.base", "AminoAcidResidueBase", sizeof(struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_AminoAcidResidueBase), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_14glycopeptidepy_2_c_9structure_4base_AminoAcidResidueBase) __PYX_ERR(5, 35, __pyx_L1_error)
  __pyx_vtabptr_14glycopeptidepy_2_c_9structure_4base_AminoAcidResidueBase = (struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_4base_AminoAcidResidueBase*)__Pyx_GetVtable(__pyx_ptype_14glycopeptidepy_2_c_9structure_4base_AminoAcidResidueBase->tp_dict); if (unlikely(!__pyx_vtabptr_14glycopeptidepy_2_c_9structure_4base_AminoAcidResidueBase)) __PYX_ERR(5, 35, __pyx_L1_error)
  __pyx_ptype_14glycopeptidepy_2_c_9structure_4base_ModificationBase = __Pyx_ImportType(__pyx_t_1, "glycopeptidepy._c.structure.base", "ModificationBase", sizeof(struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_ModificationBase), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_14glycopeptidepy_2_c_9structure_4base_ModificationBase) __PYX_ERR(5, 50, __pyx_L1_error)
  __pyx_vtabptr_14glycopeptidepy_2_c_9structure_4base_ModificationBase = (struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_4base_ModificationBase*)__Pyx_GetVtable(__pyx_ptype_14glycopeptidepy_2_c_9structure_4base_ModificationBase->tp_dict); if (unlikely(!__pyx_vtabptr_14glycopeptidepy_2_c_9structure_4base_ModificationBase)) __PYX_ERR(5, 50, __pyx_L1_error)
  __pyx_ptype_14glycopeptidepy_2_c_9structure_4base_SequencePosition = __Pyx_ImportType(__pyx_t_1, "glycopeptidepy._c.structure.base", "SequencePosition", sizeof(struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_SequencePosition), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_14glycopeptidepy_2_c_9structure_4base_SequencePosition) __PYX_ERR(5, 82, __pyx_L1_error)
  __pyx_vtabptr_14glycopeptidepy_2_c_9structure_4base_SequencePosition = (struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_4base_SequencePosition*)__Pyx_GetVtable(__pyx_ptype_14glycopeptidepy_2_c_9structure_4base_SequencePosition->tp_dict); if (unlikely(!__pyx_vtabptr_14glycopeptidepy_2_c_9structure_4base_SequencePosition)) __PYX_ERR(5, 82, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7cpython_4type_type = __Pyx_ImportType(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "type", 
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyTypeObject),
  #else
  sizeof(PyHeapTypeObject),
  #endif
  __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7cpython_4type_type) __PYX_ERR(6, 9, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("glypy.utils.cenum"); if (unlikely(!__pyx_t_1)) __PYX_ERR(7, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_5glypy_5utils_5cenum_EnumValue = __Pyx_ImportType(__pyx_t_1, "glypy.utils.cenum", "EnumValue", sizeof(struct __pyx_obj_5glypy_5utils_5cenum_EnumValue), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5glypy_5utils_5cenum_EnumValue) __PYX_ERR(7, 2, __pyx_L1_error)
  __pyx_vtabptr_5glypy_5utils_5cenum_EnumValue = (struct __pyx_vtabstruct_5glypy_5utils_5cenum_EnumValue*)__Pyx_GetVtable(__pyx_ptype_5glypy_5utils_5cenum_EnumValue->tp_dict); if (unlikely(!__pyx_vtabptr_5glypy_5utils_5cenum_EnumValue)) __PYX_ERR(7, 2, __pyx_L1_error)
  __pyx_ptype_5glypy_5utils_5cenum_IntEnumValue = __Pyx_ImportType(__pyx_t_1, "glypy.utils.cenum", "IntEnumValue", sizeof(struct __pyx_obj_5glypy_5utils_5cenum_IntEnumValue), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5glypy_5utils_5cenum_IntEnumValue) __PYX_ERR(7, 14, __pyx_L1_error)
  __pyx_vtabptr_5glypy_5utils_5cenum_IntEnumValue = (struct __pyx_vtabstruct_5glypy_5utils_5cenum_IntEnumValue*)__Pyx_GetVtable(__pyx_ptype_5glypy_5utils_5cenum_IntEnumValue->tp_dict); if (unlikely(!__pyx_vtabptr_5glypy_5utils_5cenum_IntEnumValue)) __PYX_ERR(7, 14, __pyx_L1_error)
  __pyx_ptype_5glypy_5utils_5cenum_EnumMeta = __Pyx_ImportType(__pyx_t_1, "glypy.utils.cenum", "EnumMeta", sizeof(struct __pyx_obj_5glypy_5utils_5cenum_EnumMeta), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5glypy_5utils_5cenum_EnumMeta) __PYX_ERR(7, 19, __pyx_L1_error)
  __pyx_vtabptr_5glypy_5utils_5cenum_EnumMeta = (struct __pyx_vtabstruct_5glypy_5utils_5cenum_EnumMeta*)__Pyx_GetVtable(__pyx_ptype_5glypy_5utils_5cenum_EnumMeta->tp_dict); if (unlikely(!__pyx_vtabptr_5glypy_5utils_5cenum_EnumMeta)) __PYX_ERR(7, 19, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("glycopeptidepy._c.structure.modification.rule"); if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_14glycopeptidepy_2_c_9structure_12modification_4rule_ModificationRuleBase = __Pyx_ImportType(__pyx_t_1, "glycopeptidepy._c.structure.modification.rule", "ModificationRuleBase", sizeof(struct __pyx_obj_14glycopeptidepy_2_c_9structure_12modification_4rule_ModificationRuleBase), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_14glycopeptidepy_2_c_9structure_12modification_4rule_ModificationRuleBase) __PYX_ERR(8, 5, __pyx_L1_error)
  __pyx_vtabptr_14glycopeptidepy_2_c_9structure_12modification_4rule_ModificationRuleBase = (struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_12modification_4rule_ModificationRuleBase*)__Pyx_GetVtable(__pyx_ptype_14glycopeptidepy_2_c_9structure_12modification_4rule_ModificationRuleBase->tp_dict); if (unlikely(!__pyx_vtabptr_14glycopeptidepy_2_c_9structure_12modification_4rule_ModificationRuleBase)) __PYX_ERR(8, 5, __pyx_L1_error)
  __pyx_ptype_14glycopeptidepy_2_c_9structure_12modification_4rule_NeutralLossBase = __Pyx_ImportType(__pyx_t_1, "glycopeptidepy._c.structure.modification.rule", "NeutralLossBase", sizeof(struct __pyx_obj_14glycopeptidepy_2_c_9structure_12modification_4rule_NeutralLossBase), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_14glycopeptidepy_2_c_9structure_12modification_4rule_NeutralLossBase) __PYX_ERR(8, 22, __pyx_L1_error)
  __pyx_ptype_14glycopeptidepy_2_c_9structure_12modification_4rule_GlycosylationBase = __Pyx_ImportType(__pyx_t_1, "glycopeptidepy._c.structure.modification.rule", "GlycosylationBase", sizeof(struct __pyx_obj_14glycopeptidepy_2_c_9structure_12modification_4rule_GlycosylationBase), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_14glycopeptidepy_2_c_9structure_12modification_4rule_GlycosylationBase) __PYX_ERR(8, 32, __pyx_L1_error)
  __pyx_vtabptr_14glycopeptidepy_2_c_9structure_12modification_4rule_GlycosylationBase = (struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_12modification_4rule_GlycosylationBase*)__Pyx_GetVtable(__pyx_ptype_14glycopeptidepy_2_c_9structure_12modification_4rule_GlycosylationBase->tp_dict); if (unlikely(!__pyx_vtabptr_14glycopeptidepy_2_c_9structure_12modification_4rule_GlycosylationBase)) __PYX_ERR(8, 32, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("glycopeptidepy._c.structure.modification.modification"); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_14glycopeptidepy_2_c_9structure_12modification_12modification_ModificationInstanceBase = __Pyx_ImportType(__pyx_t_1, "glycopeptidepy._c.structure.modification.modification", "ModificationInstanceBase", sizeof(struct __pyx_obj_14glycopeptidepy_2_c_9structure_12modification_12modification_ModificationInstanceBase), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_14glycopeptidepy_2_c_9structure_12modification_12modification_ModificationInstanceBase) __PYX_ERR(9, 5, __pyx_L1_error)
  __pyx_vtabptr_14glycopeptidepy_2_c_9structure_12modification_12modification_ModificationInstanceBase = (struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_12modification_12modification_ModificationInstanceBase*)__Pyx_GetVtable(__pyx_ptype_14glycopeptidepy_2_c_9structure_12modification_12modification_ModificationInstanceBase->tp_dict); if (unlikely(!__pyx_vtabptr_14glycopeptidepy_2_c_9structure_12modification_12modification_ModificationInstanceBase)) __PYX_ERR(9, 5, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_variable_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_import_code", 0);
  /*--- Variable import code ---*/
  __pyx_t_1 = PyImport_ImportModule("glypy.composition.ccomposition"); if (!__pyx_t_1) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_ImportVoidPtr(__pyx_t_1, "_atom", (void **)&__pyx_vp_5glypy_11composition_12ccomposition__atom, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr(__pyx_t_1, "_formula", (void **)&__pyx_vp_5glypy_11composition_12ccomposition__formula, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr(__pyx_t_1, "_isotope_string", (void **)&__pyx_vp_5glypy_11composition_12ccomposition__isotope_string, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr(__pyx_t_1, "isotope_pattern", (void **)&__pyx_vp_5glypy_11composition_12ccomposition_isotope_pattern, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr(__pyx_t_1, "formula_pattern", (void **)&__pyx_vp_5glypy_11composition_12ccomposition_formula_pattern, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("glypy.utils.cenum"); if (!__pyx_t_1) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_ImportVoidPtr(__pyx_t_1, "QMARK", (void **)&__pyx_vp_5glypy_5utils_5cenum_QMARK, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_function_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_import_code", 0);
  /*--- Function import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}


#ifndef CYTHON_NO_PYINIT_EXPORT
#define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#elif PY_MAJOR_VERSION < 3
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" void
#else
#define __Pyx_PyMODINIT_FUNC void
#endif
#else
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" PyObject *
#else
#define __Pyx_PyMODINIT_FUNC PyObject *
#endif
#endif


#if PY_MAJOR_VERSION < 3
__Pyx_PyMODINIT_FUNC initglycan(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC initglycan(void)
#else
__Pyx_PyMODINIT_FUNC PyInit_glycan(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC PyInit_glycan(void)
#if CYTHON_PEP489_MULTI_PHASE_INIT
{
  return PyModuleDef_Init(&__pyx_moduledef);
}
static CYTHON_SMALL_CODE int __Pyx_check_single_interpreter(void) {
    #if PY_VERSION_HEX >= 0x030700A1
    static PY_INT64_T main_interpreter_id = -1;
    PY_INT64_T current_id = PyInterpreterState_GetID(PyThreadState_Get()->interp);
    if (main_interpreter_id == -1) {
        main_interpreter_id = current_id;
        return (unlikely(current_id == -1)) ? -1 : 0;
    } else if (unlikely(main_interpreter_id != current_id))
    #else
    static PyInterpreterState *main_interpreter = NULL;
    PyInterpreterState *current_interpreter = PyThreadState_Get()->interp;
    if (!main_interpreter) {
        main_interpreter = current_interpreter;
    } else if (unlikely(main_interpreter != current_interpreter))
    #endif
    {
        PyErr_SetString(
            PyExc_ImportError,
            "Interpreter change detected - this module can only be loaded into one interpreter per process.");
        return -1;
    }
    return 0;
}
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *moddict, const char* from_name, const char* to_name, int allow_none) {
    PyObject *value = PyObject_GetAttrString(spec, from_name);
    int result = 0;
    if (likely(value)) {
        if (allow_none || value != Py_None) {
            result = PyDict_SetItemString(moddict, to_name, value);
        }
        Py_DECREF(value);
    } else if (PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Clear();
    } else {
        result = -1;
    }
    return result;
}
static CYTHON_SMALL_CODE PyObject* __pyx_pymod_create(PyObject *spec, CYTHON_UNUSED PyModuleDef *def) {
    PyObject *module = NULL, *moddict, *modname;
    if (__Pyx_check_single_interpreter())
        return NULL;
    if (__pyx_m)
        return __Pyx_NewRef(__pyx_m);
    modname = PyObject_GetAttrString(spec, "name");
    if (unlikely(!modname)) goto bad;
    module = PyModule_NewObject(modname);
    Py_DECREF(modname);
    if (unlikely(!module)) goto bad;
    moddict = PyModule_GetDict(module);
    if (unlikely(!moddict)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "loader", "__loader__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "origin", "__file__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "parent", "__package__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "submodule_search_locations", "__path__", 0) < 0)) goto bad;
    return module;
bad:
    Py_XDECREF(module);
    return NULL;
}


static CYTHON_SMALL_CODE int __pyx_pymod_exec_glycan(PyObject *__pyx_pyinit_module)
#endif
#endif
{
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  if (__pyx_m) {
    if (__pyx_m == __pyx_pyinit_module) return 0;
    PyErr_SetString(PyExc_RuntimeError, "Module 'glycan' has already been imported. Re-initialisation is not supported.");
    return -1;
  }
  #elif PY_MAJOR_VERSION >= 3
  if (__pyx_m) return __Pyx_NewRef(__pyx_m);
  #endif
  #if CYTHON_REFNANNY
__Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
if (!__Pyx_RefNanny) {
  PyErr_Clear();
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
  if (!__Pyx_RefNanny)
      Py_FatalError("failed to import 'refnanny' module");
}
#endif
  __Pyx_RefNannySetupContext("__Pyx_PyMODINIT_FUNC PyInit_glycan(void)", 0);
  if (__Pyx_check_binary_version() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pxy_PyFrame_Initialize_Offsets
  __Pxy_PyFrame_Initialize_Offsets();
  #endif
  __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_unicode)) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pyx_CyFunction_USED
  if (__pyx_CyFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Coroutine_USED
  if (__pyx_Coroutine_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_AsyncGen_USED
  if (__pyx_AsyncGen_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_StopAsyncIteration_USED
  if (__pyx_StopAsyncIteration_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  /*--- Library function declarations ---*/
  /*--- Threads initialization code ---*/
  #if defined(WITH_THREAD) && PY_VERSION_HEX < 0x030700F0 && defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
  PyEval_InitThreads();
  #endif
  /*--- Module creation code ---*/
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __pyx_m = __pyx_pyinit_module;
  Py_INCREF(__pyx_m);
  #else
  #if PY_MAJOR_VERSION < 3
  __pyx_m = Py_InitModule4("glycan", __pyx_methods, 0, 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
  #else
  __pyx_m = PyModule_Create(&__pyx_moduledef);
  #endif
  if (unlikely(!__pyx_m)) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  __pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_d)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_d);
  __pyx_b = PyImport_AddModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_b)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_b);
  __pyx_cython_runtime = PyImport_AddModule((char *) "cython_runtime"); if (unlikely(!__pyx_cython_runtime)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_cython_runtime);
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitGlobals() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3 && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
  if (__Pyx_init_sys_getdefaultencoding_params() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  if (__pyx_module_is_main_glycopeptidepy___c__structure__glycan) {
    if (PyObject_SetAttr(__pyx_m, __pyx_n_s_name, __pyx_n_s_main) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  }
  #if PY_MAJOR_VERSION >= 3
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(0, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "glycopeptidepy._c.structure.glycan")) {
      if (unlikely(PyDict_SetItemString(modules, "glycopeptidepy._c.structure.glycan", __pyx_m) < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  #endif
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Global type/function init code ---*/
  (void)__Pyx_modinit_global_init_code();
  (void)__Pyx_modinit_variable_export_code();
  (void)__Pyx_modinit_function_export_code();
  if (unlikely(__Pyx_modinit_type_init_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (unlikely(__Pyx_modinit_type_import_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (unlikely(__Pyx_modinit_variable_import_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  (void)__Pyx_modinit_function_import_code();
  /*--- Execution code ---*/
  #if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
  if (__Pyx_patch_abc() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif

  /* "glycopeptidepy/_c/structure/glycan.pyx":8
 * from glypy.composition.ccomposition cimport CComposition
 * from glypy._c.structure.glycan_composition cimport _CompositionBase
 * from glypy.structure.glycan_composition import HashableGlycanComposition             # <<<<<<<<<<<<<<
 * 
 * from glycopeptidepy._c.structure.modification.modification cimport ModificationInstanceBase
 */
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 8, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_HashableGlycanComposition);
  __Pyx_GIVEREF(__pyx_n_s_HashableGlycanComposition);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_HashableGlycanComposition);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_glypy_structure_glycan_compositi, __pyx_t_1, -1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 8, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_HashableGlycanComposition); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 8, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_HashableGlycanComposition, __pyx_t_1) < 0) __PYX_ERR(0, 8, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "glycopeptidepy/_c/structure/glycan.pyx":139
 * 
 * 
 * cdef CComposition WATER_OFFSET = CComposition({"H": 2, "O": 1})             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 139, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_H, __pyx_int_2) < 0) __PYX_ERR(0, 139, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_O, __pyx_int_1) < 0) __PYX_ERR(0, 139, __pyx_L1_error)
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_5glypy_11composition_12ccomposition_CComposition), __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 139, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_XGOTREF(((PyObject *)__pyx_v_14glycopeptidepy_2_c_9structure_6glycan_WATER_OFFSET));
  __Pyx_DECREF_SET(__pyx_v_14glycopeptidepy_2_c_9structure_6glycan_WATER_OFFSET, ((struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *)__pyx_t_1));
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "glycopeptidepy/_c/structure/glycan.pyx":142
 * 
 * 
 * cdef object GlycanCompositionProxyType = None             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_INCREF(Py_None);
  __Pyx_XGOTREF(__pyx_v_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxyType);
  __Pyx_DECREF_SET(__pyx_v_14glycopeptidepy_2_c_9structure_6glycan_GlycanCompositionProxyType, Py_None);
  __Pyx_GIVEREF(Py_None);

  /* "(tree fragment)":1
 * def __pyx_unpickle_GlycanCompositionProxy(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_14glycopeptidepy_2_c_9structure_6glycan_1__pyx_unpickle_GlycanCompositionProxy, NULL, __pyx_n_s_glycopeptidepy__c_structure_glyc); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_GlycanComposition, __pyx_t_1) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":11
 *         __pyx_unpickle_GlycanCompositionProxy__set_state(<GlycanCompositionProxy> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_GlycanCompositionProxy__set_state(GlycanCompositionProxy __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result._serialized = __pyx_state[0]; __pyx_result.obj = __pyx_state[1]
 *     if len(__pyx_state) > 2 and hasattr(__pyx_result, '__dict__'):
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_14glycopeptidepy_2_c_9structure_6glycan_3__pyx_unpickle_GlycanCompositionWithOffsetProxy, NULL, __pyx_n_s_glycopeptidepy__c_structure_glyc); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_GlycanComposition_2, __pyx_t_1) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_GlycosylationManager(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_14glycopeptidepy_2_c_9structure_6glycan_5__pyx_unpickle_GlycosylationManager, NULL, __pyx_n_s_glycopeptidepy__c_structure_glyc); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_GlycosylationMana, __pyx_t_1) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "glycopeptidepy/_c/structure/glycan.pyx":1
 * cimport cython             # <<<<<<<<<<<<<<
 * from cpython.object cimport PyObject
 * from cpython.list cimport PyList_Append, PyList_Size
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test, __pyx_t_1) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  if (__pyx_m) {
    if (__pyx_d) {
      __Pyx_AddTraceback("init glycopeptidepy._c.structure.glycan", __pyx_clineno, __pyx_lineno, __pyx_filename);
    }
    Py_CLEAR(__pyx_m);
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init glycopeptidepy._c.structure.glycan");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  return (__pyx_m != NULL) ? 0 : -1;
  #elif PY_MAJOR_VERSION >= 3
  return __pyx_m;
  #else
  return;
  #endif
}

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule(modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, "RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* PyObjectGetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_getattr))
        return tp->tp_getattr(obj, PyString_AS_STRING(attr_name));
#endif
    return PyObject_GetAttr(obj, attr_name);
}
#endif

/* GetBuiltinName */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStr(__pyx_b, name);
    if (unlikely(!result)) {
        PyErr_Format(PyExc_NameError,
#if PY_MAJOR_VERSION >= 3
            "name '%U' is not defined", name);
#else
            "name '%.200s' is not defined", PyString_AS_STRING(name));
#endif
    }
    return result;
}

/* RaiseDoubleKeywords */
static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        #if PY_MAJOR_VERSION >= 3
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
        #else
        "%s() got multiple values for keyword argument '%s'", func_name,
        PyString_AsString(kw_name));
        #endif
}

/* ParseKeywords */
static int __Pyx_ParseOptionalKeywords(
    PyObject *kwds,
    PyObject **argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject *key = 0, *value = 0;
    Py_ssize_t pos = 0;
    PyObject*** name;
    PyObject*** first_kw_arg = argnames + num_pos_args;
    while (PyDict_Next(kwds, &pos, &key, &value)) {
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            values[name-argnames] = value;
            continue;
        }
        name = first_kw_arg;
        #if PY_MAJOR_VERSION < 3
        if (likely(PyString_Check(key))) {
            while (*name) {
                if ((CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**name) == PyString_GET_SIZE(key))
                        && _PyString_Eq(**name, key)) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    if ((**argname == key) || (
                            (CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**argname) == PyString_GET_SIZE(key))
                             && _PyString_Eq(**argname, key))) {
                        goto arg_passed_twice;
                    }
                    argname++;
                }
            }
        } else
        #endif
        if (likely(PyUnicode_Check(key))) {
            while (*name) {
                int cmp = (**name == key) ? 0 :
                #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                    (__Pyx_PyUnicode_GET_LENGTH(**name) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                #endif
                    PyUnicode_Compare(**name, key);
                if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                if (cmp == 0) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    int cmp = (**argname == key) ? 0 :
                    #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                        (__Pyx_PyUnicode_GET_LENGTH(**argname) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                    #endif
                        PyUnicode_Compare(**argname, key);
                    if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                    if (cmp == 0) goto arg_passed_twice;
                    argname++;
                }
            }
        } else
            goto invalid_keyword_type;
        if (kwds2) {
            if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
        } else {
            goto invalid_keyword;
        }
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
    #if PY_MAJOR_VERSION < 3
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
bad:
    return -1;
}

/* RaiseArgTupleInvalid */
static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

/* ExtTypeTest */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type) {
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    if (likely(__Pyx_TypeCheck(obj, type)))
        return 1;
    PyErr_Format(PyExc_TypeError, "Cannot convert %.200s to %.200s",
                 Py_TYPE(obj)->tp_name, type->tp_name);
    return 0;
}

/* KeywordStringCheck */
static int __Pyx_CheckKeywordStrings(
    PyObject *kwdict,
    const char* function_name,
    int kw_allowed)
{
    PyObject* key = 0;
    Py_ssize_t pos = 0;
#if CYTHON_COMPILING_IN_PYPY
    if (!kw_allowed && PyDict_Next(kwdict, &pos, &key, 0))
        goto invalid_keyword;
    return 1;
#else
    while (PyDict_Next(kwdict, &pos, &key, 0)) {
        #if PY_MAJOR_VERSION < 3
        if (unlikely(!PyString_Check(key)))
        #endif
            if (unlikely(!PyUnicode_Check(key)))
                goto invalid_keyword_type;
    }
    if ((!kw_allowed) && unlikely(key))
        goto invalid_keyword;
    return 1;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    return 0;
#endif
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
    #if PY_MAJOR_VERSION < 3
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
    return 0;
}

/* PyObjectCall */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = Py_TYPE(func)->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* GetItemInt */
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j) {
    PyObject *r;
    if (!j) return NULL;
    r = PyObject_GetItem(o, j);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyList_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyList_GET_SIZE(o)))) {
        PyObject *r = PyList_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyTuple_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyTuple_GET_SIZE(o)))) {
        PyObject *r = PyTuple_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i, int is_list,
                                                     CYTHON_NCP_UNUSED int wraparound,
                                                     CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS && CYTHON_USE_TYPE_SLOTS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyList_GET_SIZE(o);
        if ((!boundscheck) || (likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o))))) {
            PyObject *r = PyList_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    }
    else if (PyTuple_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyTuple_GET_SIZE(o);
        if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyTuple_GET_SIZE(o)))) {
            PyObject *r = PyTuple_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    } else {
        PySequenceMethods *m = Py_TYPE(o)->tp_as_sequence;
        if (likely(m && m->sq_item)) {
            if (wraparound && unlikely(i < 0) && likely(m->sq_length)) {
                Py_ssize_t l = m->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return NULL;
                    PyErr_Clear();
                }
            }
            return m->sq_item(o, i);
        }
    }
#else
    if (is_list || PySequence_Check(o)) {
        return PySequence_GetItem(o, i);
    }
#endif
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
}

/* ObjectGetItem */
#if CYTHON_USE_TYPE_SLOTS
static PyObject *__Pyx_PyObject_GetIndex(PyObject *obj, PyObject* index) {
    PyObject *runerr = NULL;
    Py_ssize_t key_value;
    PySequenceMethods *m = Py_TYPE(obj)->tp_as_sequence;
    if (unlikely(!(m && m->sq_item))) {
        PyErr_Format(PyExc_TypeError, "'%.200s' object is not subscriptable", Py_TYPE(obj)->tp_name);
        return NULL;
    }
    key_value = __Pyx_PyIndex_AsSsize_t(index);
    if (likely(key_value != -1 || !(runerr = PyErr_Occurred()))) {
        return __Pyx_GetItemInt_Fast(obj, key_value, 0, 1, 1);
    }
    if (PyErr_GivenExceptionMatches(runerr, PyExc_OverflowError)) {
        PyErr_Clear();
        PyErr_Format(PyExc_IndexError, "cannot fit '%.200s' into an index-sized integer", Py_TYPE(index)->tp_name);
    }
    return NULL;
}
static PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject* key) {
    PyMappingMethods *m = Py_TYPE(obj)->tp_as_mapping;
    if (likely(m && m->mp_subscript)) {
        return m->mp_subscript(obj, key);
    }
    return __Pyx_PyObject_GetIndex(obj, key);
}
#endif

/* PyDictVersioning */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    return likely(dict) ? __PYX_GET_DICT_VERSION(dict) : 0;
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj) {
    PyObject **dictptr = NULL;
    Py_ssize_t offset = Py_TYPE(obj)->tp_dictoffset;
    if (offset) {
#if CYTHON_COMPILING_IN_CPYTHON
        dictptr = (likely(offset > 0)) ? (PyObject **) ((char *)obj + offset) : _PyObject_GetDictPtr(obj);
#else
        dictptr = _PyObject_GetDictPtr(obj);
#endif
    }
    return (dictptr && *dictptr) ? __PYX_GET_DICT_VERSION(*dictptr) : 0;
}
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    if (unlikely(!dict) || unlikely(tp_dict_version != __PYX_GET_DICT_VERSION(dict)))
        return 0;
    return obj_dict_version == __Pyx_get_object_dict_version(obj);
}
#endif

/* PyCFunctionFastCall */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject * __Pyx_PyCFunction_FastCall(PyObject *func_obj, PyObject **args, Py_ssize_t nargs) {
    PyCFunctionObject *func = (PyCFunctionObject*)func_obj;
    PyCFunction meth = PyCFunction_GET_FUNCTION(func);
    PyObject *self = PyCFunction_GET_SELF(func);
    int flags = PyCFunction_GET_FLAGS(func);
    assert(PyCFunction_Check(func));
    assert(METH_FASTCALL == (flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)));
    assert(nargs >= 0);
    assert(nargs == 0 || args != NULL);
    /* _PyCFunction_FastCallDict() must not be called with an exception set,
       because it may clear it (directly or indirectly) and so the
       caller loses its exception */
    assert(!PyErr_Occurred());
    if ((PY_VERSION_HEX < 0x030700A0) || unlikely(flags & METH_KEYWORDS)) {
        return (*((__Pyx_PyCFunctionFastWithKeywords)(void*)meth)) (self, args, nargs, NULL);
    } else {
        return (*((__Pyx_PyCFunctionFast)(void*)meth)) (self, args, nargs);
    }
}
#endif

/* PyFunctionFastCall */
#if CYTHON_FAST_PYCALL
static PyObject* __Pyx_PyFunction_FastCallNoKw(PyCodeObject *co, PyObject **args, Py_ssize_t na,
                                               PyObject *globals) {
    PyFrameObject *f;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject **fastlocals;
    Py_ssize_t i;
    PyObject *result;
    assert(globals != NULL);
    /* XXX Perhaps we should create a specialized
       PyFrame_New() that doesn't take locals, but does
       take builtins without sanity checking them.
       */
    assert(tstate != NULL);
    f = PyFrame_New(tstate, co, globals, NULL);
    if (f == NULL) {
        return NULL;
    }
    fastlocals = __Pyx_PyFrame_GetLocalsplus(f);
    for (i = 0; i < na; i++) {
        Py_INCREF(*args);
        fastlocals[i] = *args++;
    }
    result = PyEval_EvalFrameEx(f,0);
    ++tstate->recursion_depth;
    Py_DECREF(f);
    --tstate->recursion_depth;
    return result;
}
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs) {
    PyCodeObject *co = (PyCodeObject *)PyFunction_GET_CODE(func);
    PyObject *globals = PyFunction_GET_GLOBALS(func);
    PyObject *argdefs = PyFunction_GET_DEFAULTS(func);
    PyObject *closure;
#if PY_MAJOR_VERSION >= 3
    PyObject *kwdefs;
#endif
    PyObject *kwtuple, **k;
    PyObject **d;
    Py_ssize_t nd;
    Py_ssize_t nk;
    PyObject *result;
    assert(kwargs == NULL || PyDict_Check(kwargs));
    nk = kwargs ? PyDict_Size(kwargs) : 0;
    if (Py_EnterRecursiveCall((char*)" while calling a Python object")) {
        return NULL;
    }
    if (
#if PY_MAJOR_VERSION >= 3
            co->co_kwonlyargcount == 0 &&
#endif
            likely(kwargs == NULL || nk == 0) &&
            co->co_flags == (CO_OPTIMIZED | CO_NEWLOCALS | CO_NOFREE)) {
        if (argdefs == NULL && co->co_argcount == nargs) {
            result = __Pyx_PyFunction_FastCallNoKw(co, args, nargs, globals);
            goto done;
        }
        else if (nargs == 0 && argdefs != NULL
                 && co->co_argcount == Py_SIZE(argdefs)) {
            /* function called with no arguments, but all parameters have
               a default value: use default values as arguments .*/
            args = &PyTuple_GET_ITEM(argdefs, 0);
            result =__Pyx_PyFunction_FastCallNoKw(co, args, Py_SIZE(argdefs), globals);
            goto done;
        }
    }
    if (kwargs != NULL) {
        Py_ssize_t pos, i;
        kwtuple = PyTuple_New(2 * nk);
        if (kwtuple == NULL) {
            result = NULL;
            goto done;
        }
        k = &PyTuple_GET_ITEM(kwtuple, 0);
        pos = i = 0;
        while (PyDict_Next(kwargs, &pos, &k[i], &k[i+1])) {
            Py_INCREF(k[i]);
            Py_INCREF(k[i+1]);
            i += 2;
        }
        nk = i / 2;
    }
    else {
        kwtuple = NULL;
        k = NULL;
    }
    closure = PyFunction_GET_CLOSURE(func);
#if PY_MAJOR_VERSION >= 3
    kwdefs = PyFunction_GET_KW_DEFAULTS(func);
#endif
    if (argdefs != NULL) {
        d = &PyTuple_GET_ITEM(argdefs, 0);
        nd = Py_SIZE(argdefs);
    }
    else {
        d = NULL;
        nd = 0;
    }
#if PY_MAJOR_VERSION >= 3
    result = PyEval_EvalCodeEx((PyObject*)co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, kwdefs, closure);
#else
    result = PyEval_EvalCodeEx(co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, closure);
#endif
    Py_XDECREF(kwtuple);
done:
    Py_LeaveRecursiveCall();
    return result;
}
#endif
#endif

/* PyObjectCall2Args */
static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2) {
    PyObject *args, *result = NULL;
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyFunction_FastCall(function, args, 2);
    }
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyCFunction_FastCall(function, args, 2);
    }
    #endif
    args = PyTuple_New(2);
    if (unlikely(!args)) goto done;
    Py_INCREF(arg1);
    PyTuple_SET_ITEM(args, 0, arg1);
    Py_INCREF(arg2);
    PyTuple_SET_ITEM(args, 1, arg2);
    Py_INCREF(function);
    result = __Pyx_PyObject_Call(function, args, NULL);
    Py_DECREF(args);
    Py_DECREF(function);
done:
    return result;
}

/* PyObjectCallMethO */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = PyCFunction_GET_FUNCTION(func);
    self = PyCFunction_GET_SELF(func);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallOneArg */
#if CYTHON_COMPILING_IN_CPYTHON
static PyObject* __Pyx__PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_New(1);
    if (unlikely(!args)) return NULL;
    Py_INCREF(arg);
    PyTuple_SET_ITEM(args, 0, arg);
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, &arg, 1);
    }
#endif
    if (likely(PyCFunction_Check(func))) {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_O)) {
            return __Pyx_PyObject_CallMethO(func, arg);
#if CYTHON_FAST_PYCCALL
        } else if (__Pyx_PyFastCFunction_Check(func)) {
            return __Pyx_PyCFunction_FastCall(func, &arg, 1);
#endif
        }
    }
    return __Pyx__PyObject_CallOneArg(func, arg);
}
#else
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_Pack(1, arg);
    if (unlikely(!args)) return NULL;
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
#endif

/* PyObjectCallNoArg */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, NULL, 0);
    }
#endif
#if defined(__Pyx_CyFunction_USED) && defined(NDEBUG)
    if (likely(PyCFunction_Check(func) || __Pyx_CyFunction_Check(func)))
#else
    if (likely(PyCFunction_Check(func)))
#endif
    {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_NOARGS)) {
            return __Pyx_PyObject_CallMethO(func, NULL);
        }
    }
    return __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL);
}
#endif

/* IterFinish */
static CYTHON_INLINE int __Pyx_IterFinish(void) {
#if CYTHON_FAST_THREAD_STATE
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject* exc_type = tstate->curexc_type;
    if (unlikely(exc_type)) {
        if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) {
            PyObject *exc_value, *exc_tb;
            exc_value = tstate->curexc_value;
            exc_tb = tstate->curexc_traceback;
            tstate->curexc_type = 0;
            tstate->curexc_value = 0;
            tstate->curexc_traceback = 0;
            Py_DECREF(exc_type);
            Py_XDECREF(exc_value);
            Py_XDECREF(exc_tb);
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#else
    if (unlikely(PyErr_Occurred())) {
        if (likely(PyErr_ExceptionMatches(PyExc_StopIteration))) {
            PyErr_Clear();
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#endif
}

/* PyObjectGetMethod */
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method) {
    PyObject *attr;
#if CYTHON_UNPACK_METHODS && CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_PYTYPE_LOOKUP
    PyTypeObject *tp = Py_TYPE(obj);
    PyObject *descr;
    descrgetfunc f = NULL;
    PyObject **dictptr, *dict;
    int meth_found = 0;
    assert (*method == NULL);
    if (unlikely(tp->tp_getattro != PyObject_GenericGetAttr)) {
        attr = __Pyx_PyObject_GetAttrStr(obj, name);
        goto try_unpack;
    }
    if (unlikely(tp->tp_dict == NULL) && unlikely(PyType_Ready(tp) < 0)) {
        return 0;
    }
    descr = _PyType_Lookup(tp, name);
    if (likely(descr != NULL)) {
        Py_INCREF(descr);
#if PY_MAJOR_VERSION >= 3
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || (Py_TYPE(descr) == &PyMethodDescr_Type) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr) || (Py_TYPE(descr) == &PyMethodDescr_Type)))
        #endif
#else
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr)))
        #endif
#endif
        {
            meth_found = 1;
        } else {
            f = Py_TYPE(descr)->tp_descr_get;
            if (f != NULL && PyDescr_IsData(descr)) {
                attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
                Py_DECREF(descr);
                goto try_unpack;
            }
        }
    }
    dictptr = _PyObject_GetDictPtr(obj);
    if (dictptr != NULL && (dict = *dictptr) != NULL) {
        Py_INCREF(dict);
        attr = __Pyx_PyDict_GetItemStr(dict, name);
        if (attr != NULL) {
            Py_INCREF(attr);
            Py_DECREF(dict);
            Py_XDECREF(descr);
            goto try_unpack;
        }
        Py_DECREF(dict);
    }
    if (meth_found) {
        *method = descr;
        return 1;
    }
    if (f != NULL) {
        attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
        Py_DECREF(descr);
        goto try_unpack;
    }
    if (descr != NULL) {
        *method = descr;
        return 0;
    }
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'%.50s' object has no attribute '%U'",
                 tp->tp_name, name);
#else
                 "'%.50s' object has no attribute '%.400s'",
                 tp->tp_name, PyString_AS_STRING(name));
#endif
    return 0;
#else
    attr = __Pyx_PyObject_GetAttrStr(obj, name);
    goto try_unpack;
#endif
try_unpack:
#if CYTHON_UNPACK_METHODS
    if (likely(attr) && PyMethod_Check(attr) && likely(PyMethod_GET_SELF(attr) == obj)) {
        PyObject *function = PyMethod_GET_FUNCTION(attr);
        Py_INCREF(function);
        Py_DECREF(attr);
        *method = function;
        return 1;
    }
#endif
    *method = attr;
    return 0;
}

/* PyObjectCallMethod0 */
static PyObject* __Pyx_PyObject_CallMethod0(PyObject* obj, PyObject* method_name) {
    PyObject *method = NULL, *result = NULL;
    int is_method = __Pyx_PyObject_GetMethod(obj, method_name, &method);
    if (likely(is_method)) {
        result = __Pyx_PyObject_CallOneArg(method, obj);
        Py_DECREF(method);
        return result;
    }
    if (unlikely(!method)) goto bad;
    result = __Pyx_PyObject_CallNoArg(method);
    Py_DECREF(method);
bad:
    return result;
}

/* RaiseNeedMoreValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index) {
    PyErr_Format(PyExc_ValueError,
                 "need more than %" CYTHON_FORMAT_SSIZE_T "d value%.1s to unpack",
                 index, (index == 1) ? "" : "s");
}

/* RaiseTooManyValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected) {
    PyErr_Format(PyExc_ValueError,
                 "too many values to unpack (expected %" CYTHON_FORMAT_SSIZE_T "d)", expected);
}

/* UnpackItemEndCheck */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected) {
    if (unlikely(retval)) {
        Py_DECREF(retval);
        __Pyx_RaiseTooManyValuesError(expected);
        return -1;
    }
    return __Pyx_IterFinish();
}

/* RaiseNoneIterError */
static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
}

/* UnpackTupleError */
static void __Pyx_UnpackTupleError(PyObject *t, Py_ssize_t index) {
    if (t == Py_None) {
      __Pyx_RaiseNoneNotIterableError();
    } else if (PyTuple_GET_SIZE(t) < index) {
      __Pyx_RaiseNeedMoreValuesError(PyTuple_GET_SIZE(t));
    } else {
      __Pyx_RaiseTooManyValuesError(index);
    }
}

/* UnpackTuple2 */
static CYTHON_INLINE int __Pyx_unpack_tuple2_exact(
        PyObject* tuple, PyObject** pvalue1, PyObject** pvalue2, int decref_tuple) {
    PyObject *value1 = NULL, *value2 = NULL;
#if CYTHON_COMPILING_IN_PYPY
    value1 = PySequence_ITEM(tuple, 0);  if (unlikely(!value1)) goto bad;
    value2 = PySequence_ITEM(tuple, 1);  if (unlikely(!value2)) goto bad;
#else
    value1 = PyTuple_GET_ITEM(tuple, 0);  Py_INCREF(value1);
    value2 = PyTuple_GET_ITEM(tuple, 1);  Py_INCREF(value2);
#endif
    if (decref_tuple) {
        Py_DECREF(tuple);
    }
    *pvalue1 = value1;
    *pvalue2 = value2;
    return 0;
#if CYTHON_COMPILING_IN_PYPY
bad:
    Py_XDECREF(value1);
    Py_XDECREF(value2);
    if (decref_tuple) { Py_XDECREF(tuple); }
    return -1;
#endif
}
static int __Pyx_unpack_tuple2_generic(PyObject* tuple, PyObject** pvalue1, PyObject** pvalue2,
                                       int has_known_size, int decref_tuple) {
    Py_ssize_t index;
    PyObject *value1 = NULL, *value2 = NULL, *iter = NULL;
    iternextfunc iternext;
    iter = PyObject_GetIter(tuple);
    if (unlikely(!iter)) goto bad;
    if (decref_tuple) { Py_DECREF(tuple); tuple = NULL; }
    iternext = Py_TYPE(iter)->tp_iternext;
    value1 = iternext(iter); if (unlikely(!value1)) { index = 0; goto unpacking_failed; }
    value2 = iternext(iter); if (unlikely(!value2)) { index = 1; goto unpacking_failed; }
    if (!has_known_size && unlikely(__Pyx_IternextUnpackEndCheck(iternext(iter), 2))) goto bad;
    Py_DECREF(iter);
    *pvalue1 = value1;
    *pvalue2 = value2;
    return 0;
unpacking_failed:
    if (!has_known_size && __Pyx_IterFinish() == 0)
        __Pyx_RaiseNeedMoreValuesError(index);
bad:
    Py_XDECREF(iter);
    Py_XDECREF(value1);
    Py_XDECREF(value2);
    if (decref_tuple) { Py_XDECREF(tuple); }
    return -1;
}

/* dict_iter */
static CYTHON_INLINE PyObject* __Pyx_dict_iterator(PyObject* iterable, int is_dict, PyObject* method_name,
                                                   Py_ssize_t* p_orig_length, int* p_source_is_dict) {
    is_dict = is_dict || likely(PyDict_CheckExact(iterable));
    *p_source_is_dict = is_dict;
    if (is_dict) {
#if !CYTHON_COMPILING_IN_PYPY
        *p_orig_length = PyDict_Size(iterable);
        Py_INCREF(iterable);
        return iterable;
#elif PY_MAJOR_VERSION >= 3
        static PyObject *py_items = NULL, *py_keys = NULL, *py_values = NULL;
        PyObject **pp = NULL;
        if (method_name) {
            const char *name = PyUnicode_AsUTF8(method_name);
            if (strcmp(name, "iteritems") == 0) pp = &py_items;
            else if (strcmp(name, "iterkeys") == 0) pp = &py_keys;
            else if (strcmp(name, "itervalues") == 0) pp = &py_values;
            if (pp) {
                if (!*pp) {
                    *pp = PyUnicode_FromString(name + 4);
                    if (!*pp)
                        return NULL;
                }
                method_name = *pp;
            }
        }
#endif
    }
    *p_orig_length = 0;
    if (method_name) {
        PyObject* iter;
        iterable = __Pyx_PyObject_CallMethod0(iterable, method_name);
        if (!iterable)
            return NULL;
#if !CYTHON_COMPILING_IN_PYPY
        if (PyTuple_CheckExact(iterable) || PyList_CheckExact(iterable))
            return iterable;
#endif
        iter = PyObject_GetIter(iterable);
        Py_DECREF(iterable);
        return iter;
    }
    return PyObject_GetIter(iterable);
}
static CYTHON_INLINE int __Pyx_dict_iter_next(
        PyObject* iter_obj, CYTHON_NCP_UNUSED Py_ssize_t orig_length, CYTHON_NCP_UNUSED Py_ssize_t* ppos,
        PyObject** pkey, PyObject** pvalue, PyObject** pitem, int source_is_dict) {
    PyObject* next_item;
#if !CYTHON_COMPILING_IN_PYPY
    if (source_is_dict) {
        PyObject *key, *value;
        if (unlikely(orig_length != PyDict_Size(iter_obj))) {
            PyErr_SetString(PyExc_RuntimeError, "dictionary changed size during iteration");
            return -1;
        }
        if (unlikely(!PyDict_Next(iter_obj, ppos, &key, &value))) {
            return 0;
        }
        if (pitem) {
            PyObject* tuple = PyTuple_New(2);
            if (unlikely(!tuple)) {
                return -1;
            }
            Py_INCREF(key);
            Py_INCREF(value);
            PyTuple_SET_ITEM(tuple, 0, key);
            PyTuple_SET_ITEM(tuple, 1, value);
            *pitem = tuple;
        } else {
            if (pkey) {
                Py_INCREF(key);
                *pkey = key;
            }
            if (pvalue) {
                Py_INCREF(value);
                *pvalue = value;
            }
        }
        return 1;
    } else if (PyTuple_CheckExact(iter_obj)) {
        Py_ssize_t pos = *ppos;
        if (unlikely(pos >= PyTuple_GET_SIZE(iter_obj))) return 0;
        *ppos = pos + 1;
        next_item = PyTuple_GET_ITEM(iter_obj, pos);
        Py_INCREF(next_item);
    } else if (PyList_CheckExact(iter_obj)) {
        Py_ssize_t pos = *ppos;
        if (unlikely(pos >= PyList_GET_SIZE(iter_obj))) return 0;
        *ppos = pos + 1;
        next_item = PyList_GET_ITEM(iter_obj, pos);
        Py_INCREF(next_item);
    } else
#endif
    {
        next_item = PyIter_Next(iter_obj);
        if (unlikely(!next_item)) {
            return __Pyx_IterFinish();
        }
    }
    if (pitem) {
        *pitem = next_item;
    } else if (pkey && pvalue) {
        if (__Pyx_unpack_tuple2(next_item, pkey, pvalue, source_is_dict, source_is_dict, 1))
            return -1;
    } else if (pkey) {
        *pkey = next_item;
    } else {
        *pvalue = next_item;
    }
    return 1;
}

/* MergeKeywords */
static int __Pyx_MergeKeywords(PyObject *kwdict, PyObject *source_mapping) {
    PyObject *iter, *key = NULL, *value = NULL;
    int source_is_dict, result;
    Py_ssize_t orig_length, ppos = 0;
    iter = __Pyx_dict_iterator(source_mapping, 0, __pyx_n_s_items, &orig_length, &source_is_dict);
    if (unlikely(!iter)) {
        PyObject *args;
        if (!PyErr_ExceptionMatches(PyExc_AttributeError)) goto bad;
        PyErr_Clear();
        args = PyTuple_Pack(1, source_mapping);
        if (likely(args)) {
            PyObject *fallback = PyObject_Call((PyObject*)&PyDict_Type, args, NULL);
            Py_DECREF(args);
            if (likely(fallback)) {
                iter = __Pyx_dict_iterator(fallback, 1, __pyx_n_s_items, &orig_length, &source_is_dict);
                Py_DECREF(fallback);
            }
        }
        if (unlikely(!iter)) goto bad;
    }
    while (1) {
        result = __Pyx_dict_iter_next(iter, orig_length, &ppos, &key, &value, NULL, source_is_dict);
        if (unlikely(result < 0)) goto bad;
        if (!result) break;
        if (unlikely(PyDict_Contains(kwdict, key))) {
            __Pyx_RaiseDoubleKeywordsError("function", key);
            result = -1;
        } else {
            result = PyDict_SetItem(kwdict, key, value);
        }
        Py_DECREF(key);
        Py_DECREF(value);
        if (unlikely(result < 0)) goto bad;
    }
    Py_XDECREF(iter);
    return 0;
bad:
    Py_XDECREF(iter);
    return -1;
}

/* GetTopmostException */
#if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem *
__Pyx_PyErr_GetTopmostException(PyThreadState *tstate)
{
    _PyErr_StackItem *exc_info = tstate->exc_info;
    while ((exc_info->exc_type == NULL || exc_info->exc_type == Py_None) &&
           exc_info->previous_item != NULL)
    {
        exc_info = exc_info->previous_item;
    }
    return exc_info;
}
#endif

/* SaveResetException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = __Pyx_PyErr_GetTopmostException(tstate);
    *type = exc_info->exc_type;
    *value = exc_info->exc_value;
    *tb = exc_info->exc_traceback;
    #else
    *type = tstate->exc_type;
    *value = tstate->exc_value;
    *tb = tstate->exc_traceback;
    #endif
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
}
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = type;
    exc_info->exc_value = value;
    exc_info->exc_traceback = tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = type;
    tstate->exc_value = value;
    tstate->exc_traceback = tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
#endif

/* PyErrExceptionMatches */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx_PyErr_ExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        if (__Pyx_PyErr_GivenExceptionMatches(exc_type, PyTuple_GET_ITEM(tuple, i))) return 1;
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err) {
    PyObject *exc_type = tstate->curexc_type;
    if (exc_type == err) return 1;
    if (unlikely(!exc_type)) return 0;
    if (unlikely(PyTuple_Check(err)))
        return __Pyx_PyErr_ExceptionMatchesTuple(exc_type, err);
    return __Pyx_PyErr_GivenExceptionMatches(exc_type, err);
}
#endif

/* PyErrFetchRestore */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
}
#endif

/* GetAttr */
static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *o, PyObject *n) {
#if CYTHON_USE_TYPE_SLOTS
#if PY_MAJOR_VERSION >= 3
    if (likely(PyUnicode_Check(n)))
#else
    if (likely(PyString_Check(n)))
#endif
        return __Pyx_PyObject_GetAttrStr(o, n);
#endif
    return PyObject_GetAttr(o, n);
}

/* GetAttr3 */
static PyObject *__Pyx_GetAttr3Default(PyObject *d) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (unlikely(!__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        return NULL;
    __Pyx_PyErr_Clear();
    Py_INCREF(d);
    return d;
}
static CYTHON_INLINE PyObject *__Pyx_GetAttr3(PyObject *o, PyObject *n, PyObject *d) {
    PyObject *r = __Pyx_GetAttr(o, n);
    return (likely(r)) ? r : __Pyx_GetAttr3Default(d);
}

/* GetModuleGlobalName */
#if CYTHON_USE_DICT_VERSIONS
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value)
#else
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name)
#endif
{
    PyObject *result;
#if !CYTHON_AVOID_BORROWED_REFS
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1
    result = _PyDict_GetItem_KnownHash(__pyx_d, name, ((PyASCIIObject *) name)->hash);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    } else if (unlikely(PyErr_Occurred())) {
        return NULL;
    }
#else
    result = PyDict_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
#endif
#else
    result = PyObject_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
    PyErr_Clear();
#endif
    return __Pyx_GetBuiltinName(name);
}

/* WriteUnraisableException */
static void __Pyx_WriteUnraisable(const char *name, CYTHON_UNUSED int clineno,
                                  CYTHON_UNUSED int lineno, CYTHON_UNUSED const char *filename,
                                  int full_traceback, CYTHON_UNUSED int nogil) {
    PyObject *old_exc, *old_val, *old_tb;
    PyObject *ctx;
    __Pyx_PyThreadState_declare
#ifdef WITH_THREAD
    PyGILState_STATE state;
    if (nogil)
        state = PyGILState_Ensure();
    else state = (PyGILState_STATE)0;
#endif
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&old_exc, &old_val, &old_tb);
    if (full_traceback) {
        Py_XINCREF(old_exc);
        Py_XINCREF(old_val);
        Py_XINCREF(old_tb);
        __Pyx_ErrRestore(old_exc, old_val, old_tb);
        PyErr_PrintEx(1);
    }
    #if PY_MAJOR_VERSION < 3
    ctx = PyString_FromString(name);
    #else
    ctx = PyUnicode_FromString(name);
    #endif
    __Pyx_ErrRestore(old_exc, old_val, old_tb);
    if (!ctx) {
        PyErr_WriteUnraisable(Py_None);
    } else {
        PyErr_WriteUnraisable(ctx);
        Py_DECREF(ctx);
    }
#ifdef WITH_THREAD
    if (nogil)
        PyGILState_Release(state);
#endif
}

/* DictGetItem */
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key) {
    PyObject *value;
    value = PyDict_GetItemWithError(d, key);
    if (unlikely(!value)) {
        if (!PyErr_Occurred()) {
            if (unlikely(PyTuple_Check(key))) {
                PyObject* args = PyTuple_Pack(1, key);
                if (likely(args)) {
                    PyErr_SetObject(PyExc_KeyError, args);
                    Py_DECREF(args);
                }
            } else {
                PyErr_SetObject(PyExc_KeyError, key);
            }
        }
        return NULL;
    }
    Py_INCREF(value);
    return value;
}
#endif

/* UnpackUnboundCMethod */
static int __Pyx_TryUnpackUnboundCMethod(__Pyx_CachedCFunction* target) {
    PyObject *method;
    method = __Pyx_PyObject_GetAttrStr(target->type, *target->method_name);
    if (unlikely(!method))
        return -1;
    target->method = method;
#if CYTHON_COMPILING_IN_CPYTHON
    #if PY_MAJOR_VERSION >= 3
    if (likely(__Pyx_TypeCheck(method, &PyMethodDescr_Type)))
    #endif
    {
        PyMethodDescrObject *descr = (PyMethodDescrObject*) method;
        target->func = descr->d_method->ml_meth;
        target->flag = descr->d_method->ml_flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_STACKLESS);
    }
#endif
    return 0;
}

/* CallUnboundCMethod0 */
static PyObject* __Pyx__CallUnboundCMethod0(__Pyx_CachedCFunction* cfunc, PyObject* self) {
    PyObject *args, *result = NULL;
    if (unlikely(!cfunc->method) && unlikely(__Pyx_TryUnpackUnboundCMethod(cfunc) < 0)) return NULL;
#if CYTHON_ASSUME_SAFE_MACROS
    args = PyTuple_New(1);
    if (unlikely(!args)) goto bad;
    Py_INCREF(self);
    PyTuple_SET_ITEM(args, 0, self);
#else
    args = PyTuple_Pack(1, self);
    if (unlikely(!args)) goto bad;
#endif
    result = __Pyx_PyObject_Call(cfunc->method, args, NULL);
    Py_DECREF(args);
bad:
    return result;
}

/* py_dict_keys */
static CYTHON_INLINE PyObject* __Pyx_PyDict_Keys(PyObject* d) {
    if (PY_MAJOR_VERSION >= 3)
        return __Pyx_CallUnboundCMethod0(&__pyx_umethod_PyDict_Type_keys, d);
    else
        return PyDict_Keys(d);
}

/* py_dict_values */
static CYTHON_INLINE PyObject* __Pyx_PyDict_Values(PyObject* d) {
    if (PY_MAJOR_VERSION >= 3)
        return __Pyx_CallUnboundCMethod0(&__pyx_umethod_PyDict_Type_values, d);
    else
        return PyDict_Values(d);
}

/* py_dict_items */
static CYTHON_INLINE PyObject* __Pyx_PyDict_Items(PyObject* d) {
    if (PY_MAJOR_VERSION >= 3)
        return __Pyx_CallUnboundCMethod0(&__pyx_umethod_PyDict_Type_items, d);
    else
        return PyDict_Items(d);
}

/* CallUnboundCMethod2 */
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030600B1
static CYTHON_INLINE PyObject *__Pyx_CallUnboundCMethod2(__Pyx_CachedCFunction *cfunc, PyObject *self, PyObject *arg1, PyObject *arg2) {
    if (likely(cfunc->func)) {
        PyObject *args[2] = {arg1, arg2};
        if (cfunc->flag == METH_FASTCALL) {
            #if PY_VERSION_HEX >= 0x030700A0
            return (*(__Pyx_PyCFunctionFast)(void*)(PyCFunction)cfunc->func)(self, args, 2);
            #else
            return (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)cfunc->func)(self, args, 2, NULL);
            #endif
        }
        #if PY_VERSION_HEX >= 0x030700A0
        if (cfunc->flag == (METH_FASTCALL | METH_KEYWORDS))
            return (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)cfunc->func)(self, args, 2, NULL);
        #endif
    }
    return __Pyx__CallUnboundCMethod2(cfunc, self, arg1, arg2);
}
#endif
static PyObject* __Pyx__CallUnboundCMethod2(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg1, PyObject* arg2){
    PyObject *args, *result = NULL;
    if (unlikely(!cfunc->func && !cfunc->method) && unlikely(__Pyx_TryUnpackUnboundCMethod(cfunc) < 0)) return NULL;
#if CYTHON_COMPILING_IN_CPYTHON
    if (cfunc->func && (cfunc->flag & METH_VARARGS)) {
        args = PyTuple_New(2);
        if (unlikely(!args)) goto bad;
        Py_INCREF(arg1);
        PyTuple_SET_ITEM(args, 0, arg1);
        Py_INCREF(arg2);
        PyTuple_SET_ITEM(args, 1, arg2);
        if (cfunc->flag & METH_KEYWORDS)
            result = (*(PyCFunctionWithKeywords)(void*)(PyCFunction)cfunc->func)(self, args, NULL);
        else
            result = (*cfunc->func)(self, args);
    } else {
        args = PyTuple_New(3);
        if (unlikely(!args)) goto bad;
        Py_INCREF(self);
        PyTuple_SET_ITEM(args, 0, self);
        Py_INCREF(arg1);
        PyTuple_SET_ITEM(args, 1, arg1);
        Py_INCREF(arg2);
        PyTuple_SET_ITEM(args, 2, arg2);
        result = __Pyx_PyObject_Call(cfunc->method, args, NULL);
    }
#else
    args = PyTuple_Pack(3, self, arg1, arg2);
    if (unlikely(!args)) goto bad;
    result = __Pyx_PyObject_Call(cfunc->method, args, NULL);
#endif
bad:
    Py_XDECREF(args);
    return result;
}

/* CallUnboundCMethod1 */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_CallUnboundCMethod1(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg) {
    if (likely(cfunc->func)) {
        int flag = cfunc->flag;
        if (flag == METH_O) {
            return (*(cfunc->func))(self, arg);
        } else if (PY_VERSION_HEX >= 0x030600B1 && flag == METH_FASTCALL) {
            #if PY_VERSION_HEX >= 0x030700A0
                return (*(__Pyx_PyCFunctionFast)(void*)(PyCFunction)cfunc->func)(self, &arg, 1);
            #else
                return (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)cfunc->func)(self, &arg, 1, NULL);
            #endif
        } else if (PY_VERSION_HEX >= 0x030700A0 && flag == (METH_FASTCALL | METH_KEYWORDS)) {
            return (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)cfunc->func)(self, &arg, 1, NULL);
        }
    }
    return __Pyx__CallUnboundCMethod1(cfunc, self, arg);
}
#endif
static PyObject* __Pyx__CallUnboundCMethod1(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg){
    PyObject *args, *result = NULL;
    if (unlikely(!cfunc->func && !cfunc->method) && unlikely(__Pyx_TryUnpackUnboundCMethod(cfunc) < 0)) return NULL;
#if CYTHON_COMPILING_IN_CPYTHON
    if (cfunc->func && (cfunc->flag & METH_VARARGS)) {
        args = PyTuple_New(1);
        if (unlikely(!args)) goto bad;
        Py_INCREF(arg);
        PyTuple_SET_ITEM(args, 0, arg);
        if (cfunc->flag & METH_KEYWORDS)
            result = (*(PyCFunctionWithKeywords)(void*)(PyCFunction)cfunc->func)(self, args, NULL);
        else
            result = (*cfunc->func)(self, args);
    } else {
        args = PyTuple_New(2);
        if (unlikely(!args)) goto bad;
        Py_INCREF(self);
        PyTuple_SET_ITEM(args, 0, self);
        Py_INCREF(arg);
        PyTuple_SET_ITEM(args, 1, arg);
        result = __Pyx_PyObject_Call(cfunc->method, args, NULL);
    }
#else
    args = PyTuple_Pack(2, self, arg);
    if (unlikely(!args)) goto bad;
    result = __Pyx_PyObject_Call(cfunc->method, args, NULL);
#endif
bad:
    Py_XDECREF(args);
    return result;
}

/* py_dict_pop */
static CYTHON_INLINE PyObject *__Pyx_PyDict_Pop(PyObject *d, PyObject *key, PyObject *default_value) {
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX > 0x030600B3
    if ((1)) {
        return _PyDict_Pop(d, key, default_value);
    } else
#endif
    if (default_value) {
        return __Pyx_CallUnboundCMethod2(&__pyx_umethod_PyDict_Type_pop, d, key, default_value);
    } else {
        return __Pyx_CallUnboundCMethod1(&__pyx_umethod_PyDict_Type_pop, d, key);
    }
}

/* PyObjectSetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_setattro))
        return tp->tp_setattro(obj, attr_name, value);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_setattr))
        return tp->tp_setattr(obj, PyString_AS_STRING(attr_name), value);
#endif
    return PyObject_SetAttr(obj, attr_name, value);
}
#endif

/* Import */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level) {
    PyObject *empty_list = 0;
    PyObject *module = 0;
    PyObject *global_dict = 0;
    PyObject *empty_dict = 0;
    PyObject *list;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_import;
    py_import = __Pyx_PyObject_GetAttrStr(__pyx_b, __pyx_n_s_import);
    if (!py_import)
        goto bad;
    #endif
    if (from_list)
        list = from_list;
    else {
        empty_list = PyList_New(0);
        if (!empty_list)
            goto bad;
        list = empty_list;
    }
    global_dict = PyModule_GetDict(__pyx_m);
    if (!global_dict)
        goto bad;
    empty_dict = PyDict_New();
    if (!empty_dict)
        goto bad;
    {
        #if PY_MAJOR_VERSION >= 3
        if (level == -1) {
            if ((1) && (strchr(__Pyx_MODULE_NAME, '.'))) {
                module = PyImport_ImportModuleLevelObject(
                    name, global_dict, empty_dict, list, 1);
                if (!module) {
                    if (!PyErr_ExceptionMatches(PyExc_ImportError))
                        goto bad;
                    PyErr_Clear();
                }
            }
            level = 0;
        }
        #endif
        if (!module) {
            #if PY_MAJOR_VERSION < 3
            PyObject *py_level = PyInt_FromLong(level);
            if (!py_level)
                goto bad;
            module = PyObject_CallFunctionObjArgs(py_import,
                name, global_dict, empty_dict, list, py_level, (PyObject *)NULL);
            Py_DECREF(py_level);
            #else
            module = PyImport_ImportModuleLevelObject(
                name, global_dict, empty_dict, list, level);
            #endif
        }
    }
bad:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_import);
    #endif
    Py_XDECREF(empty_list);
    Py_XDECREF(empty_dict);
    return module;
}

/* ImportFrom */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name) {
    PyObject* value = __Pyx_PyObject_GetAttrStr(module, name);
    if (unlikely(!value) && PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Format(PyExc_ImportError,
        #if PY_MAJOR_VERSION < 3
            "cannot import name %.230s", PyString_AS_STRING(name));
        #else
            "cannot import name %S", name);
        #endif
    }
    return value;
}

/* RaiseException */
#if PY_MAJOR_VERSION < 3
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb,
                        CYTHON_UNUSED PyObject *cause) {
    __Pyx_PyThreadState_declare
    Py_XINCREF(type);
    if (!value || value == Py_None)
        value = NULL;
    else
        Py_INCREF(value);
    if (!tb || tb == Py_None)
        tb = NULL;
    else {
        Py_INCREF(tb);
        if (!PyTraceBack_Check(tb)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: arg 3 must be a traceback or None");
            goto raise_error;
        }
    }
    if (PyType_Check(type)) {
#if CYTHON_COMPILING_IN_PYPY
        if (!value) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#endif
        PyErr_NormalizeException(&type, &value, &tb);
    } else {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto raise_error;
        }
        value = type;
        type = (PyObject*) Py_TYPE(type);
        Py_INCREF(type);
        if (!PyType_IsSubtype((PyTypeObject *)type, (PyTypeObject *)PyExc_BaseException)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: exception class must be a subclass of BaseException");
            goto raise_error;
        }
    }
    __Pyx_PyThreadState_assign
    __Pyx_ErrRestore(type, value, tb);
    return;
raise_error:
    Py_XDECREF(value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
    return;
}
#else
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
    if (cause) {
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
#if CYTHON_COMPILING_IN_PYPY
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#else
        PyThreadState *tstate = __Pyx_PyThreadState_Current;
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}
#endif

/* HasAttr */
static CYTHON_INLINE int __Pyx_HasAttr(PyObject *o, PyObject *n) {
    PyObject *r;
    if (unlikely(!__Pyx_PyBaseString_Check(n))) {
        PyErr_SetString(PyExc_TypeError,
                        "hasattr(): attribute name must be string");
        return -1;
    }
    r = __Pyx_GetAttr(o, n);
    if (unlikely(!r)) {
        PyErr_Clear();
        return 0;
    } else {
        Py_DECREF(r);
        return 1;
    }
}

/* PyObject_GenericGetAttrNoDict */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject *__Pyx_RaiseGenericGetAttributeError(PyTypeObject *tp, PyObject *attr_name) {
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'%.50s' object has no attribute '%U'",
                 tp->tp_name, attr_name);
#else
                 "'%.50s' object has no attribute '%.400s'",
                 tp->tp_name, PyString_AS_STRING(attr_name));
#endif
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name) {
    PyObject *descr;
    PyTypeObject *tp = Py_TYPE(obj);
    if (unlikely(!PyString_Check(attr_name))) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    assert(!tp->tp_dictoffset);
    descr = _PyType_Lookup(tp, attr_name);
    if (unlikely(!descr)) {
        return __Pyx_RaiseGenericGetAttributeError(tp, attr_name);
    }
    Py_INCREF(descr);
    #if PY_MAJOR_VERSION < 3
    if (likely(PyType_HasFeature(Py_TYPE(descr), Py_TPFLAGS_HAVE_CLASS)))
    #endif
    {
        descrgetfunc f = Py_TYPE(descr)->tp_descr_get;
        if (unlikely(f)) {
            PyObject *res = f(descr, obj, (PyObject *)tp);
            Py_DECREF(descr);
            return res;
        }
    }
    return descr;
}
#endif

/* PyObject_GenericGetAttr */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name) {
    if (unlikely(Py_TYPE(obj)->tp_dictoffset)) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    return __Pyx_PyObject_GenericGetAttrNoDict(obj, attr_name);
}
#endif

/* SetVTable */
static int __Pyx_SetVtable(PyObject *dict, void *vtable) {
#if PY_VERSION_HEX >= 0x02070000
    PyObject *ob = PyCapsule_New(vtable, 0, 0);
#else
    PyObject *ob = PyCObject_FromVoidPtr(vtable, 0);
#endif
    if (!ob)
        goto bad;
    if (PyDict_SetItem(dict, __pyx_n_s_pyx_vtable, ob) < 0)
        goto bad;
    Py_DECREF(ob);
    return 0;
bad:
    Py_XDECREF(ob);
    return -1;
}

/* PyObjectGetAttrStrNoError */
static void __Pyx_PyObject_GetAttrStr_ClearAttributeError(void) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (likely(__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        __Pyx_PyErr_Clear();
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name) {
    PyObject *result;
#if CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_TYPE_SLOTS && PY_VERSION_HEX >= 0x030700B1
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro == PyObject_GenericGetAttr)) {
        return _PyObject_GenericGetAttrWithDict(obj, attr_name, NULL, 1);
    }
#endif
    result = __Pyx_PyObject_GetAttrStr(obj, attr_name);
    if (unlikely(!result)) {
        __Pyx_PyObject_GetAttrStr_ClearAttributeError();
    }
    return result;
}

/* SetupReduce */
static int __Pyx_setup_reduce_is_named(PyObject* meth, PyObject* name) {
  int ret;
  PyObject *name_attr;
  name_attr = __Pyx_PyObject_GetAttrStr(meth, __pyx_n_s_name);
  if (likely(name_attr)) {
      ret = PyObject_RichCompareBool(name_attr, name, Py_EQ);
  } else {
      ret = -1;
  }
  if (unlikely(ret < 0)) {
      PyErr_Clear();
      ret = 0;
  }
  Py_XDECREF(name_attr);
  return ret;
}
static int __Pyx_setup_reduce(PyObject* type_obj) {
    int ret = 0;
    PyObject *object_reduce = NULL;
    PyObject *object_getstate = NULL;
    PyObject *object_reduce_ex = NULL;
    PyObject *reduce = NULL;
    PyObject *reduce_ex = NULL;
    PyObject *reduce_cython = NULL;
    PyObject *setstate = NULL;
    PyObject *setstate_cython = NULL;
    PyObject *getstate = NULL;
#if CYTHON_USE_PYTYPE_LOOKUP
    getstate = _PyType_Lookup((PyTypeObject*)type_obj, __pyx_n_s_getstate);
#else
    getstate = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_getstate);
    if (!getstate && PyErr_Occurred()) {
        goto __PYX_BAD;
    }
#endif
    if (getstate) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_getstate = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_getstate);
#else
        object_getstate = __Pyx_PyObject_GetAttrStrNoError((PyObject*)&PyBaseObject_Type, __pyx_n_s_getstate);
        if (!object_getstate && PyErr_Occurred()) {
            goto __PYX_BAD;
        }
#endif
        if (object_getstate != getstate) {
            goto __PYX_GOOD;
        }
    }
#if CYTHON_USE_PYTYPE_LOOKUP
    object_reduce_ex = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#else
    object_reduce_ex = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#endif
    reduce_ex = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce_ex); if (unlikely(!reduce_ex)) goto __PYX_BAD;
    if (reduce_ex == object_reduce_ex) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_reduce = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto __PYX_BAD;
#else
        object_reduce = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto __PYX_BAD;
#endif
        reduce = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce); if (unlikely(!reduce)) goto __PYX_BAD;
        if (reduce == object_reduce || __Pyx_setup_reduce_is_named(reduce, __pyx_n_s_reduce_cython)) {
            reduce_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_reduce_cython);
            if (likely(reduce_cython)) {
                ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce, reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
            } else if (reduce == object_reduce || PyErr_Occurred()) {
                goto __PYX_BAD;
            }
            setstate = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_setstate);
            if (!setstate) PyErr_Clear();
            if (!setstate || __Pyx_setup_reduce_is_named(setstate, __pyx_n_s_setstate_cython)) {
                setstate_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_setstate_cython);
                if (likely(setstate_cython)) {
                    ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate, setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                    ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                } else if (!setstate || PyErr_Occurred()) {
                    goto __PYX_BAD;
                }
            }
            PyType_Modified((PyTypeObject*)type_obj);
        }
    }
    goto __PYX_GOOD;
__PYX_BAD:
    if (!PyErr_Occurred())
        PyErr_Format(PyExc_RuntimeError, "Unable to initialize pickling for %s", ((PyTypeObject*)type_obj)->tp_name);
    ret = -1;
__PYX_GOOD:
#if !CYTHON_USE_PYTYPE_LOOKUP
    Py_XDECREF(object_reduce);
    Py_XDECREF(object_reduce_ex);
    Py_XDECREF(object_getstate);
    Py_XDECREF(getstate);
#endif
    Py_XDECREF(reduce);
    Py_XDECREF(reduce_ex);
    Py_XDECREF(reduce_cython);
    Py_XDECREF(setstate);
    Py_XDECREF(setstate_cython);
    return ret;
}

/* TypeImport */
#ifndef __PYX_HAVE_RT_ImportType
#define __PYX_HAVE_RT_ImportType
static PyTypeObject *__Pyx_ImportType(PyObject *module, const char *module_name, const char *class_name,
    size_t size, enum __Pyx_ImportType_CheckSize check_size)
{
    PyObject *result = 0;
    char warning[200];
    Py_ssize_t basicsize;
#ifdef Py_LIMITED_API
    PyObject *py_basicsize;
#endif
    result = PyObject_GetAttrString(module, class_name);
    if (!result)
        goto bad;
    if (!PyType_Check(result)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s.%.200s is not a type object",
            module_name, class_name);
        goto bad;
    }
#ifndef Py_LIMITED_API
    basicsize = ((PyTypeObject *)result)->tp_basicsize;
#else
    py_basicsize = PyObject_GetAttrString(result, "__basicsize__");
    if (!py_basicsize)
        goto bad;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = 0;
    if (basicsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
#endif
    if ((size_t)basicsize < size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        goto bad;
    }
    if (check_size == __Pyx_ImportType_CheckSize_Error && (size_t)basicsize != size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        goto bad;
    }
    else if (check_size == __Pyx_ImportType_CheckSize_Warn && (size_t)basicsize > size) {
        PyOS_snprintf(warning, sizeof(warning),
            "%s.%s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        if (PyErr_WarnEx(NULL, warning, 0) < 0) goto bad;
    }
    return (PyTypeObject *)result;
bad:
    Py_XDECREF(result);
    return NULL;
}
#endif

/* GetVTable */
static void* __Pyx_GetVtable(PyObject *dict) {
    void* ptr;
    PyObject *ob = PyObject_GetItem(dict, __pyx_n_s_pyx_vtable);
    if (!ob)
        goto bad;
#if PY_VERSION_HEX >= 0x02070000
    ptr = PyCapsule_GetPointer(ob, 0);
#else
    ptr = PyCObject_AsVoidPtr(ob);
#endif
    if (!ptr && !PyErr_Occurred())
        PyErr_SetString(PyExc_RuntimeError, "invalid vtable found for imported type");
    Py_DECREF(ob);
    return ptr;
bad:
    Py_XDECREF(ob);
    return NULL;
}

/* CLineInTraceback */
#ifndef CYTHON_CLINE_IN_TRACEBACK
static int __Pyx_CLineForTraceback(CYTHON_UNUSED PyThreadState *tstate, int c_line) {
    PyObject *use_cline;
    PyObject *ptype, *pvalue, *ptraceback;
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject **cython_runtime_dict;
#endif
    if (unlikely(!__pyx_cython_runtime)) {
        return c_line;
    }
    __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
#if CYTHON_COMPILING_IN_CPYTHON
    cython_runtime_dict = _PyObject_GetDictPtr(__pyx_cython_runtime);
    if (likely(cython_runtime_dict)) {
        __PYX_PY_DICT_LOOKUP_IF_MODIFIED(
            use_cline, *cython_runtime_dict,
            __Pyx_PyDict_GetItemStr(*cython_runtime_dict, __pyx_n_s_cline_in_traceback))
    } else
#endif
    {
      PyObject *use_cline_obj = __Pyx_PyObject_GetAttrStr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback);
      if (use_cline_obj) {
        use_cline = PyObject_Not(use_cline_obj) ? Py_False : Py_True;
        Py_DECREF(use_cline_obj);
      } else {
        PyErr_Clear();
        use_cline = NULL;
      }
    }
    if (!use_cline) {
        c_line = 0;
        (void) PyObject_SetAttr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback, Py_False);
    }
    else if (use_cline == Py_False || (use_cline != Py_True && PyObject_Not(use_cline) != 0)) {
        c_line = 0;
    }
    __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
    return c_line;
}
#endif

/* CodeObjectCache */
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, ((size_t)new_max) * sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}

/* AddTraceback */
#include "compile.h"
#include "frameobject.h"
#include "traceback.h"
#if PY_VERSION_HEX >= 0x030b00a6
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = NULL;
    PyObject *py_funcname = NULL;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_srcfile = NULL;
    py_srcfile = PyString_FromString(filename);
    if (!py_srcfile) goto bad;
    #endif
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        if (!py_funcname) goto bad;
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        if (!py_funcname) goto bad;
        funcname = PyUnicode_AsUTF8(py_funcname);
        if (!funcname) goto bad;
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        if (!py_funcname) goto bad;
        #endif
    }
    #if PY_MAJOR_VERSION < 3
    py_code = __Pyx_PyCode_New(
        0,
        0,
        0,
        0,
        0,
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    #else
    py_code = PyCode_NewEmpty(filename, funcname, py_line);
    #endif
    Py_XDECREF(py_funcname);  // XDECREF since it's only set on Py3 if cline
    return py_code;
bad:
    Py_XDECREF(py_funcname);
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_srcfile);
    #endif
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject *ptype, *pvalue, *ptraceback;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(tstate, c_line);
    }
    py_code = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!py_code) {
        __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) {
            /* If the code object creation fails, then we should clear the
               fetched exception references and propagate the new exception */
            Py_XDECREF(ptype);
            Py_XDECREF(pvalue);
            Py_XDECREF(ptraceback);
            goto bad;
        }
        __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
        __pyx_insert_code_object(c_line ? -c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        tstate,            /*PyThreadState *tstate,*/
        py_code,           /*PyCodeObject *code,*/
        __pyx_d,    /*PyObject *globals,*/
        0                  /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}

/* CIntFromPyVerify */
#define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(int),
                                     little, !is_unsigned);
    }
}

/* CIntFromPy */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(int) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case  1: __PYX_VERIFY_RETURN_INT(int, digit, digits[0])
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 2 * PyLong_SHIFT) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 3 * PyLong_SHIFT) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 4 * PyLong_SHIFT) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(int) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case -1: __PYX_VERIFY_RETURN_INT(int, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(int,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(int) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(int) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            int val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (int) -1;
        }
    } else {
        int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyInt_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* CIntFromPy */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(long) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (long) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case  1: __PYX_VERIFY_RETURN_INT(long, digit, digits[0])
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 2 * PyLong_SHIFT) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 3 * PyLong_SHIFT) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 4 * PyLong_SHIFT) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (long) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(long) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case -1: __PYX_VERIFY_RETURN_INT(long, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(long,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(long) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(long) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            long val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (long) -1;
        }
    } else {
        long val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyInt_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
    }
}

/* FastTypeChecks */
#if CYTHON_COMPILING_IN_CPYTHON
static int __Pyx_InBases(PyTypeObject *a, PyTypeObject *b) {
    while (a) {
        a = a->tp_base;
        if (a == b)
            return 1;
    }
    return b == &PyBaseObject_Type;
}
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (a == b) return 1;
    mro = a->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            if (PyTuple_GET_ITEM(mro, i) == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(a, b);
}
#if PY_MAJOR_VERSION == 2
static int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject* exc_type2) {
    PyObject *exception, *value, *tb;
    int res;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&exception, &value, &tb);
    res = exc_type1 ? PyObject_IsSubclass(err, exc_type1) : 0;
    if (unlikely(res == -1)) {
        PyErr_WriteUnraisable(err);
        res = 0;
    }
    if (!res) {
        res = PyObject_IsSubclass(err, exc_type2);
        if (unlikely(res == -1)) {
            PyErr_WriteUnraisable(err);
            res = 0;
        }
    }
    __Pyx_ErrRestore(exception, value, tb);
    return res;
}
#else
static CYTHON_INLINE int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject *exc_type2) {
    int res = exc_type1 ? __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type1) : 0;
    if (!res) {
        res = __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type2);
    }
    return res;
}
#endif
static int __Pyx_PyErr_GivenExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    assert(PyExceptionClass_Check(exc_type));
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        PyObject *t = PyTuple_GET_ITEM(tuple, i);
        #if PY_MAJOR_VERSION < 3
        if (likely(exc_type == t)) return 1;
        #endif
        if (likely(PyExceptionClass_Check(t))) {
            if (__Pyx_inner_PyErr_GivenExceptionMatches2(exc_type, NULL, t)) return 1;
        } else {
        }
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject* exc_type) {
    if (likely(err == exc_type)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        if (likely(PyExceptionClass_Check(exc_type))) {
            return __Pyx_inner_PyErr_GivenExceptionMatches2(err, NULL, exc_type);
        } else if (likely(PyTuple_Check(exc_type))) {
            return __Pyx_PyErr_GivenExceptionMatchesTuple(err, exc_type);
        } else {
        }
    }
    return PyErr_GivenExceptionMatches(err, exc_type);
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *exc_type1, PyObject *exc_type2) {
    assert(PyExceptionClass_Check(exc_type1));
    assert(PyExceptionClass_Check(exc_type2));
    if (likely(err == exc_type1 || err == exc_type2)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        return __Pyx_inner_PyErr_GivenExceptionMatches2(err, exc_type1, exc_type2);
    }
    return (PyErr_GivenExceptionMatches(err, exc_type1) || PyErr_GivenExceptionMatches(err, exc_type2));
}
#endif

/* CheckBinaryVersion */
static int __Pyx_check_binary_version(void) {
    char ctversion[5];
    int same=1, i, found_dot;
    const char* rt_from_call = Py_GetVersion();
    PyOS_snprintf(ctversion, 5, "%d.%d", PY_MAJOR_VERSION, PY_MINOR_VERSION);
    found_dot = 0;
    for (i = 0; i < 4; i++) {
        if (!ctversion[i]) {
            same = (rt_from_call[i] < '0' || rt_from_call[i] > '9');
            break;
        }
        if (rt_from_call[i] != ctversion[i]) {
            same = 0;
            break;
        }
    }
    if (!same) {
        char rtversion[5] = {'\0'};
        char message[200];
        for (i=0; i<4; ++i) {
            if (rt_from_call[i] == '.') {
                if (found_dot) break;
                found_dot = 1;
            } else if (rt_from_call[i] < '0' || rt_from_call[i] > '9') {
                break;
            }
            rtversion[i] = rt_from_call[i];
        }
        PyOS_snprintf(message, sizeof(message),
                      "compiletime version %s of module '%.100s' "
                      "does not match runtime version %s",
                      ctversion, __Pyx_MODULE_NAME, rtversion);
        return PyErr_WarnEx(NULL, message, 1);
    }
    return 0;
}

/* VoidPtrImport */
#ifndef __PYX_HAVE_RT_ImportVoidPtr
#define __PYX_HAVE_RT_ImportVoidPtr
static int __Pyx_ImportVoidPtr(PyObject *module, const char *name, void **p, const char *sig) {
    PyObject *d = 0;
    PyObject *cobj = 0;
    d = PyObject_GetAttrString(module, (char *)"__pyx_capi__");
    if (!d)
        goto bad;
    cobj = PyDict_GetItemString(d, name);
    if (!cobj) {
        PyErr_Format(PyExc_ImportError,
            "%.200s does not export expected C variable %.200s",
                PyModule_GetName(module), name);
        goto bad;
    }
#if PY_VERSION_HEX >= 0x02070000
    if (!PyCapsule_IsValid(cobj, sig)) {
        PyErr_Format(PyExc_TypeError,
            "C variable %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), name, sig, PyCapsule_GetName(cobj));
        goto bad;
    }
    *p = PyCapsule_GetPointer(cobj, sig);
#else
    {const char *desc, *s1, *s2;
    desc = (const char *)PyCObject_GetDesc(cobj);
    if (!desc)
        goto bad;
    s1 = desc; s2 = sig;
    while (*s1 != '\0' && *s1 == *s2) { s1++; s2++; }
    if (*s1 != *s2) {
        PyErr_Format(PyExc_TypeError,
            "C variable %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), name, sig, desc);
        goto bad;
    }
    *p = PyCObject_AsVoidPtr(cobj);}
#endif
    if (!(*p))
        goto bad;
    Py_DECREF(d);
    return 0;
bad:
    Py_XDECREF(d);
    return -1;
}
#endif

/* InitStrings */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION < 3
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        #else
        if (t->is_unicode | t->is_str) {
            if (t->intern) {
                *t->p = PyUnicode_InternFromString(t->s);
            } else if (t->encoding) {
                *t->p = PyUnicode_Decode(t->s, t->n - 1, t->encoding, NULL);
            } else {
                *t->p = PyUnicode_FromStringAndSize(t->s, t->n - 1);
            }
        } else {
            *t->p = PyBytes_FromStringAndSize(t->s, t->n - 1);
        }
        #endif
        if (!*t->p)
            return -1;
        if (PyObject_Hash(*t->p) == -1)
            return -1;
        ++t;
    }
    return 0;
}

static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    return __Pyx_PyUnicode_FromStringAndSize(c_str, (Py_ssize_t)strlen(c_str));
}
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
#if !CYTHON_PEP393_ENABLED
static const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    char* defenc_c;
    PyObject* defenc = _PyUnicode_AsDefaultEncodedString(o, NULL);
    if (!defenc) return NULL;
    defenc_c = PyBytes_AS_STRING(defenc);
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    {
        char* end = defenc_c + PyBytes_GET_SIZE(defenc);
        char* c;
        for (c = defenc_c; c < end; c++) {
            if ((unsigned char) (*c) >= 128) {
                PyUnicode_AsASCIIString(o);
                return NULL;
            }
        }
    }
#endif
    *length = PyBytes_GET_SIZE(defenc);
    return defenc_c;
}
#else
static CYTHON_INLINE const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    if (unlikely(__Pyx_PyUnicode_READY(o) == -1)) return NULL;
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    if (likely(PyUnicode_IS_ASCII(o))) {
        *length = PyUnicode_GET_LENGTH(o);
        return PyUnicode_AsUTF8(o);
    } else {
        PyUnicode_AsASCIIString(o);
        return NULL;
    }
#else
    return PyUnicode_AsUTF8AndSize(o, length);
#endif
}
#endif
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
    if (
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
            __Pyx_sys_getdefaultencoding_not_ascii &&
#endif
            PyUnicode_Check(o)) {
        return __Pyx_PyUnicode_AsStringAndSize(o, length);
    } else
#endif
#if (!CYTHON_COMPILING_IN_PYPY) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    if (PyByteArray_Check(o)) {
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
    } else
#endif
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject* x) {
    int retval;
    if (unlikely(!x)) return -1;
    retval = __Pyx_PyObject_IsTrue(x);
    Py_DECREF(x);
    return retval;
}
static PyObject* __Pyx_PyNumber_IntOrLongWrongResultType(PyObject* result, const char* type_name) {
#if PY_MAJOR_VERSION >= 3
    if (PyLong_Check(result)) {
        if (PyErr_WarnFormat(PyExc_DeprecationWarning, 1,
                "__int__ returned non-int (type %.200s).  "
                "The ability to return an instance of a strict subclass of int "
                "is deprecated, and may be removed in a future version of Python.",
                Py_TYPE(result)->tp_name)) {
            Py_DECREF(result);
            return NULL;
        }
        return result;
    }
#endif
    PyErr_Format(PyExc_TypeError,
                 "__%.4s__ returned non-%.4s (type %.200s)",
                 type_name, type_name, Py_TYPE(result)->tp_name);
    Py_DECREF(result);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_Check(x) || PyLong_Check(x)))
#else
  if (likely(PyLong_Check(x)))
#endif
    return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  #if PY_MAJOR_VERSION < 3
  if (m && m->nb_int) {
    name = "int";
    res = m->nb_int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = m->nb_long(x);
  }
  #else
  if (likely(m && m->nb_int)) {
    name = "int";
    res = m->nb_int(x);
  }
  #endif
#else
  if (!PyBytes_CheckExact(x) && !PyUnicode_CheckExact(x)) {
    res = PyNumber_Int(x);
  }
#endif
  if (likely(res)) {
#if PY_MAJOR_VERSION < 3
    if (unlikely(!PyInt_Check(res) && !PyLong_Check(res))) {
#else
    if (unlikely(!PyLong_CheckExact(res))) {
#endif
        return __Pyx_PyNumber_IntOrLongWrongResultType(res, name);
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_CheckExact(b))) {
    if (sizeof(Py_ssize_t) >= sizeof(long))
        return PyInt_AS_LONG(b);
    else
        return PyInt_AsSsize_t(b);
  }
#endif
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    const digit* digits = ((PyLongObject*)b)->ob_digit;
    const Py_ssize_t size = Py_SIZE(b);
    if (likely(__Pyx_sst_abs(size) <= 1)) {
        ival = likely(size) ? digits[0] : 0;
        if (size == -1) ival = -ival;
        return ival;
    } else {
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject* o) {
  if (sizeof(Py_hash_t) == sizeof(Py_ssize_t)) {
    return (Py_hash_t) __Pyx_PyIndex_AsSsize_t(o);
#if PY_MAJOR_VERSION < 3
  } else if (likely(PyInt_CheckExact(o))) {
    return PyInt_AS_LONG(o);
#endif
  } else {
    Py_ssize_t ival;
    PyObject *x;
    x = PyNumber_Index(o);
    if (!x) return -1;
    ival = PyInt_AsLong(x);
    Py_DECREF(x);
    return ival;
  }
}
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b) {
  return b ? __Pyx_NewRef(Py_True) : __Pyx_NewRef(Py_False);
}
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
    return PyInt_FromSize_t(ival);
}


#endif /* Py_PYTHON_H */
