/* Generated by Cython 0.29.33 */

/* BEGIN: Cython Metadata
{
    "distutils": {
        "depends": [
            "/Users/runner/hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/glycopeptidepy/_c/compat.h"
        ],
        "include_dirs": [
            "/Users/runner/hostedtoolcache/Python/3.9.14/x64/lib/python3.9/site-packages/glycopeptidepy/_c",
            "src/glycopeptidepy/_c"
        ],
        "name": "glycopeptidepy._c.structure.fragment",
        "sources": [
            "src/glycopeptidepy/_c/structure/fragment.pyx"
        ]
    },
    "module_name": "glycopeptidepy._c.structure.fragment"
}
END: Cython Metadata */

#ifndef PY_SSIZE_T_CLEAN
#define PY_SSIZE_T_CLEAN
#endif /* PY_SSIZE_T_CLEAN */
#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02060000 || (0x03000000 <= PY_VERSION_HEX && PY_VERSION_HEX < 0x03030000)
    #error Cython requires Python 2.6+ or Python 3.3+.
#else
#define CYTHON_ABI "0_29_33"
#define CYTHON_HEX_VERSION 0x001D21F0
#define CYTHON_FUTURE_DIVISION 0
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#define __PYX_COMMA ,
#ifndef HAVE_LONG_LONG
  #if PY_VERSION_HEX >= 0x02070000
    #define HAVE_LONG_LONG
  #endif
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#ifdef PYPY_VERSION
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_NOGIL 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
#elif defined(PYSTON_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_NOGIL 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_ASYNC_SLOTS
  #define CYTHON_USE_ASYNC_SLOTS 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
#elif defined(PY_NOGIL)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_NOGIL 1
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #ifndef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE 1
  #endif
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #define CYTHON_COMPILING_IN_NOGIL 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYTYPE_LOOKUP
    #define CYTHON_USE_PYTYPE_LOOKUP 0
  #elif !defined(CYTHON_USE_PYTYPE_LOOKUP)
    #define CYTHON_USE_PYTYPE_LOOKUP 1
  #endif
  #if PY_MAJOR_VERSION < 3
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 0
  #elif !defined(CYTHON_USE_PYLONG_INTERNALS)
    #define CYTHON_USE_PYLONG_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if PY_VERSION_HEX < 0x030300F0 || PY_VERSION_HEX >= 0x030B00A2
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #if PY_VERSION_HEX >= 0x030B00A4
    #undef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 0
  #elif !defined(CYTHON_FAST_THREAD_STATE)
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL (PY_VERSION_HEX < 0x030A0000)
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT (PY_VERSION_HEX >= 0x03050000)
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE (PY_VERSION_HEX >= 0x030400a1)
  #endif
  #ifndef CYTHON_USE_DICT_VERSIONS
    #define CYTHON_USE_DICT_VERSIONS (PY_VERSION_HEX >= 0x030600B1)
  #endif
  #if PY_VERSION_HEX >= 0x030B00A4
    #undef CYTHON_USE_EXC_INFO_STACK
    #define CYTHON_USE_EXC_INFO_STACK 0
  #elif !defined(CYTHON_USE_EXC_INFO_STACK)
    #define CYTHON_USE_EXC_INFO_STACK (PY_VERSION_HEX >= 0x030700A3)
  #endif
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 1
  #endif
#endif
#if !defined(CYTHON_FAST_PYCCALL)
#define CYTHON_FAST_PYCCALL  (CYTHON_FAST_PYCALL && PY_VERSION_HEX >= 0x030600B1)
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #if PY_MAJOR_VERSION < 3
    #include "longintrepr.h"
  #endif
  #undef SHIFT
  #undef BASE
  #undef MASK
  #ifdef SIZEOF_VOID_P
    enum { __pyx_check_sizeof_voidp = 1 / (int)(SIZEOF_VOID_P == sizeof(void*)) };
  #endif
#endif
#ifndef __has_attribute
  #define __has_attribute(x) 0
#endif
#ifndef __has_cpp_attribute
  #define __has_cpp_attribute(x) 0
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_MAYBE_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_MAYBE_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)
#ifdef _MSC_VER
    #ifndef _MSC_STDINT_H_
        #if _MSC_VER < 1300
           typedef unsigned char     uint8_t;
           typedef unsigned int      uint32_t;
        #else
           typedef unsigned __int8   uint8_t;
           typedef unsigned __int32  uint32_t;
        #endif
    #endif
#else
   #include <stdint.h>
#endif
#ifndef CYTHON_FALLTHROUGH
  #if defined(__cplusplus) && __cplusplus >= 201103L
    #if __has_cpp_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH [[fallthrough]]
    #elif __has_cpp_attribute(clang::fallthrough)
      #define CYTHON_FALLTHROUGH [[clang::fallthrough]]
    #elif __has_cpp_attribute(gnu::fallthrough)
      #define CYTHON_FALLTHROUGH [[gnu::fallthrough]]
    #endif
  #endif
  #ifndef CYTHON_FALLTHROUGH
    #if __has_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH __attribute__((fallthrough))
    #else
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
  #if defined(__clang__ ) && defined(__apple_build_version__)
    #if __apple_build_version__ < 7000000
      #undef  CYTHON_FALLTHROUGH
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
#endif

#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #elif defined(__GNUC__)
    #define CYTHON_INLINE __inline__
  #elif defined(_MSC_VER)
    #define CYTHON_INLINE __inline
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_INLINE inline
  #else
    #define CYTHON_INLINE
  #endif
#endif

#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x02070600 && !defined(Py_OptimizeFlag)
  #define Py_OptimizeFlag 0
#endif
#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a+k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyClass_Type
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
  #define __Pyx_DefaultClassType PyType_Type
#if PY_VERSION_HEX >= 0x030B00A1
    static CYTHON_INLINE PyCodeObject* __Pyx_PyCode_New(int a, int k, int l, int s, int f,
                                                    PyObject *code, PyObject *c, PyObject* n, PyObject *v,
                                                    PyObject *fv, PyObject *cell, PyObject* fn,
                                                    PyObject *name, int fline, PyObject *lnos) {
        PyObject *kwds=NULL, *argcount=NULL, *posonlyargcount=NULL, *kwonlyargcount=NULL;
        PyObject *nlocals=NULL, *stacksize=NULL, *flags=NULL, *replace=NULL, *call_result=NULL, *empty=NULL;
        const char *fn_cstr=NULL;
        const char *name_cstr=NULL;
        PyCodeObject* co=NULL;
        PyObject *type, *value, *traceback;
        PyErr_Fetch(&type, &value, &traceback);
        if (!(kwds=PyDict_New())) goto end;
        if (!(argcount=PyLong_FromLong(a))) goto end;
        if (PyDict_SetItemString(kwds, "co_argcount", argcount) != 0) goto end;
        if (!(posonlyargcount=PyLong_FromLong(0))) goto end;
        if (PyDict_SetItemString(kwds, "co_posonlyargcount", posonlyargcount) != 0) goto end;
        if (!(kwonlyargcount=PyLong_FromLong(k))) goto end;
        if (PyDict_SetItemString(kwds, "co_kwonlyargcount", kwonlyargcount) != 0) goto end;
        if (!(nlocals=PyLong_FromLong(l))) goto end;
        if (PyDict_SetItemString(kwds, "co_nlocals", nlocals) != 0) goto end;
        if (!(stacksize=PyLong_FromLong(s))) goto end;
        if (PyDict_SetItemString(kwds, "co_stacksize", stacksize) != 0) goto end;
        if (!(flags=PyLong_FromLong(f))) goto end;
        if (PyDict_SetItemString(kwds, "co_flags", flags) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_code", code) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_consts", c) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_names", n) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_varnames", v) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_freevars", fv) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_cellvars", cell) != 0) goto end;
        if (PyDict_SetItemString(kwds, "co_linetable", lnos) != 0) goto end;
        if (!(fn_cstr=PyUnicode_AsUTF8AndSize(fn, NULL))) goto end;
        if (!(name_cstr=PyUnicode_AsUTF8AndSize(name, NULL))) goto end;
        if (!(co = PyCode_NewEmpty(fn_cstr, name_cstr, fline))) goto end;
        if (!(replace = PyObject_GetAttrString((PyObject*)co, "replace"))) goto cleanup_code_too;
        if (!(empty = PyTuple_New(0))) goto cleanup_code_too; // unfortunately __pyx_empty_tuple isn't available here
        if (!(call_result = PyObject_Call(replace, empty, kwds))) goto cleanup_code_too;
        Py_XDECREF((PyObject*)co);
        co = (PyCodeObject*)call_result;
        call_result = NULL;
        if (0) {
            cleanup_code_too:
            Py_XDECREF((PyObject*)co);
            co = NULL;
        }
        end:
        Py_XDECREF(kwds);
        Py_XDECREF(argcount);
        Py_XDECREF(posonlyargcount);
        Py_XDECREF(kwonlyargcount);
        Py_XDECREF(nlocals);
        Py_XDECREF(stacksize);
        Py_XDECREF(replace);
        Py_XDECREF(call_result);
        Py_XDECREF(empty);
        if (type) {
            PyErr_Restore(type, value, traceback);
        }
        return co;
    }
#else
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#endif
  #define __Pyx_DefaultClassType PyType_Type
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#ifndef METH_STACKLESS
  #define METH_STACKLESS 0
#endif
#if PY_VERSION_HEX <= 0x030700A3 || !defined(METH_FASTCALL)
  #ifndef METH_FASTCALL
     #define METH_FASTCALL 0x80
  #endif
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject *const *args, Py_ssize_t nargs);
  typedef PyObject *(*__Pyx_PyCFunctionFastWithKeywords) (PyObject *self, PyObject *const *args,
                                                          Py_ssize_t nargs, PyObject *kwnames);
#else
  #define __Pyx_PyCFunctionFast _PyCFunctionFast
  #define __Pyx_PyCFunctionFastWithKeywords _PyCFunctionFastWithKeywords
#endif
#if CYTHON_FAST_PYCCALL
#define __Pyx_PyFastCFunction_Check(func)\
    ((PyCFunction_Check(func) && (METH_FASTCALL == (PyCFunction_GET_FLAGS(func) & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)))))
#else
#define __Pyx_PyFastCFunction_Check(func) 0
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030400A1
  #define PyMem_RawMalloc(n)           PyMem_Malloc(n)
  #define PyMem_RawRealloc(p, n)       PyMem_Realloc(p, n)
  #define PyMem_RawFree(p)             PyMem_Free(p)
#endif
#if CYTHON_COMPILING_IN_PYSTON
  #define __Pyx_PyCode_HasFreeVars(co)  PyCode_HasFreeVars(co)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) PyFrame_SetLineNumber(frame, lineno)
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#if !CYTHON_FAST_THREAD_STATE || PY_VERSION_HEX < 0x02070000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#elif PY_VERSION_HEX >= 0x03060000
  #define __Pyx_PyThreadState_Current _PyThreadState_UncheckedGet()
#elif PY_VERSION_HEX >= 0x03000000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#else
  #define __Pyx_PyThreadState_Current _PyThreadState_Current
#endif
#if PY_VERSION_HEX < 0x030700A2 && !defined(PyThread_tss_create) && !defined(Py_tss_NEEDS_INIT)
#include "pythread.h"
#define Py_tss_NEEDS_INIT 0
typedef int Py_tss_t;
static CYTHON_INLINE int PyThread_tss_create(Py_tss_t *key) {
  *key = PyThread_create_key();
  return 0;
}
static CYTHON_INLINE Py_tss_t * PyThread_tss_alloc(void) {
  Py_tss_t *key = (Py_tss_t *)PyObject_Malloc(sizeof(Py_tss_t));
  *key = Py_tss_NEEDS_INIT;
  return key;
}
static CYTHON_INLINE void PyThread_tss_free(Py_tss_t *key) {
  PyObject_Free(key);
}
static CYTHON_INLINE int PyThread_tss_is_created(Py_tss_t *key) {
  return *key != Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE void PyThread_tss_delete(Py_tss_t *key) {
  PyThread_delete_key(*key);
  *key = Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE int PyThread_tss_set(Py_tss_t *key, void *value) {
  return PyThread_set_key_value(*key, value);
}
static CYTHON_INLINE void * PyThread_tss_get(Py_tss_t *key) {
  return PyThread_get_key_value(*key);
}
#endif
#if CYTHON_COMPILING_IN_CPYTHON || defined(_PyDict_NewPresized)
#define __Pyx_PyDict_NewPresized(n)  ((n <= 8) ? PyDict_New() : _PyDict_NewPresized(n))
#else
#define __Pyx_PyDict_NewPresized(n)  PyDict_New()
#endif
#if PY_MAJOR_VERSION >= 3 || CYTHON_FUTURE_DIVISION
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_Divide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceDivide(x,y)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1 && CYTHON_USE_UNICODE_INTERNALS
#define __Pyx_PyDict_GetItemStr(dict, name)  _PyDict_GetItem_KnownHash(dict, name, ((PyASCIIObject *) name)->hash)
#else
#define __Pyx_PyDict_GetItemStr(dict, name)  PyDict_GetItem(dict, name)
#endif
#if PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #if PY_VERSION_HEX >= 0x030C0000
    #define __Pyx_PyUnicode_READY(op)       (0)
  #else
    #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                                0 : _PyUnicode_Ready((PyObject *)(op)))
  #endif
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         PyUnicode_KIND(u)
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, ch)
  #if PY_VERSION_HEX >= 0x030C0000
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_LENGTH(u))
  #else
    #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03090000
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : ((PyCompactUnicodeObject *)(u))->wstr_length))
    #else
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
    #endif
  #endif
#else
  #define CYTHON_PEP393_ENABLED 0
  #define PyUnicode_1BYTE_KIND  1
  #define PyUnicode_2BYTE_KIND  2
  #define PyUnicode_4BYTE_KIND  4
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((sizeof(Py_UNICODE) == 2) ? 65535 : 1114111)
  #define __Pyx_PyUnicode_KIND(u)         (sizeof(Py_UNICODE))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)PyUnicode_AS_UNICODE(u))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)(k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  (((void)(k)), ((Py_UNICODE*)d)[i] = ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_SIZE(u))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyUnicode_Contains)
  #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyByteArray_Check)
  #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Format)
  #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
#endif
#define __Pyx_PyString_FormatSafe(a, b)   ((unlikely((a) == Py_None || (PyString_Check(b) && !PyString_CheckExact(b)))) ? PyNumber_Remainder(a, b) : __Pyx_PyString_Format(a, b))
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None || (PyUnicode_Check(b) && !PyUnicode_CheckExact(b)))) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyString_Format(a, b)  PyUnicode_Format(a, b)
#else
  #define __Pyx_PyString_Format(a, b)  PyString_Format(a, b)
#endif
#if PY_MAJOR_VERSION < 3 && !defined(PyObject_ASCII)
  #define PyObject_ASCII(o)            PyObject_Repr(o)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
#ifndef PyObject_Unicode
  #define PyObject_Unicode             PyObject_Str
#endif
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyBaseString_Check(obj) PyUnicode_Check(obj)
  #define __Pyx_PyBaseString_CheckExact(obj) PyUnicode_CheckExact(obj)
#else
  #define __Pyx_PyBaseString_Check(obj) (PyString_Check(obj) || PyUnicode_Check(obj))
  #define __Pyx_PyBaseString_CheckExact(obj) (PyString_CheckExact(obj) || PyUnicode_CheckExact(obj))
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        (Py_TYPE(obj) == &PySet_Type)
#endif
#if PY_VERSION_HEX >= 0x030900A4
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_SET_REFCNT(obj, refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SET_SIZE(obj, size)
#else
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_REFCNT(obj) = (refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SIZE(obj) = (size)
#endif
#if CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PySequence_SIZE(seq)  Py_SIZE(seq)
#else
  #define __Pyx_PySequence_SIZE(seq)  PySequence_Size(seq)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
  #define PyNumber_Int                 PyNumber_Long
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_MAJOR_VERSION >= 3 && CYTHON_COMPILING_IN_PYPY
  #ifndef PyUnicode_InternFromString
    #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
  #endif
#endif
#if PY_VERSION_HEX < 0x030200A4
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   __Pyx_PyIndex_AsHash_t
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   __Pyx_PyIndex_AsSsize_t
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyMethod_New(func, self, klass) ((self) ? ((void)(klass), PyMethod_New(func, self)) : __Pyx_NewRef(func))
#else
  #define __Pyx_PyMethod_New(func, self, klass) PyMethod_New(func, self, klass)
#endif
#if CYTHON_USE_ASYNC_SLOTS
  #if PY_VERSION_HEX >= 0x030500B1
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
    #define __Pyx_PyType_AsAsync(obj) (Py_TYPE(obj)->tp_as_async)
  #else
    #define __Pyx_PyType_AsAsync(obj) ((__Pyx_PyAsyncMethodsStruct*) (Py_TYPE(obj)->tp_reserved))
  #endif
#else
  #define __Pyx_PyType_AsAsync(obj) NULL
#endif
#ifndef __Pyx_PyAsyncMethodsStruct
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
    } __Pyx_PyAsyncMethodsStruct;
#endif

#if defined(_WIN32) || defined(WIN32) || defined(MS_WINDOWS)
  #if !defined(_USE_MATH_DEFINES)
    #define _USE_MATH_DEFINES
  #endif
#endif
#include <math.h>
#ifdef NAN
#define __PYX_NAN() ((float) NAN)
#else
static CYTHON_INLINE float __PYX_NAN() {
  float value;
  memset(&value, 0xFF, sizeof(value));
  return value;
}
#endif
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif

#define __PYX_MARK_ERR_POS(f_index, lineno) \
    { __pyx_filename = __pyx_f[f_index]; (void)__pyx_filename; __pyx_lineno = lineno; (void)__pyx_lineno; __pyx_clineno = __LINE__; (void)__pyx_clineno; }
#define __PYX_ERR(f_index, lineno, Ln_error) \
    { __PYX_MARK_ERR_POS(f_index, lineno) goto Ln_error; }

#ifndef __PYX_EXTERN_C
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE__glycopeptidepy___c__structure__fragment
#define __PYX_HAVE_API__glycopeptidepy___c__structure__fragment
/* Early includes */
#include <string.h>
#include <stdio.h>
#include "pythread.h"
#include <stdlib.h>
#include "compat.h"
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#if defined(PYREX_WITHOUT_ASSERTIONS) && !defined(CYTHON_WITHOUT_ASSERTIONS)
#define CYTHON_WITHOUT_ASSERTIONS
#endif

typedef struct {PyObject **p; const char *s; const Py_ssize_t n; const char* encoding;
                const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry;

#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_UTF8 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT (PY_MAJOR_VERSION >= 3 && __PYX_DEFAULT_STRING_ENCODING_IS_UTF8)
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
static CYTHON_INLINE int __Pyx_is_valid_index(Py_ssize_t i, Py_ssize_t limit) {
    return (size_t) i < (size_t) limit;
}
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER)
    #define __Pyx_sst_abs(value) ((Py_ssize_t)_abs64(value))
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
#define __Pyx_PyByteArray_FromString(s) PyByteArray_FromStringAndSize((const char*)s, strlen((const char*)s))
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if PY_MAJOR_VERSION < 3
    #define __Pyx_PyStr_FromString        __Pyx_PyBytes_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#else
    #define __Pyx_PyStr_FromString        __Pyx_PyUnicode_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyUnicode_FromStringAndSize
#endif
#define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyObject_AsWritableString(s)    ((char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableSString(s)    ((signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableUString(s)    ((unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsSString(s)    ((const signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((const unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyStr_FromCString(s)     __Pyx_PyStr_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
static CYTHON_INLINE size_t __Pyx_Py_UNICODE_strlen(const Py_UNICODE *u) {
    const Py_UNICODE *u_end = u;
    while (*u_end++) ;
    return (size_t)(u_end - u - 1);
}
#define __Pyx_PyUnicode_FromUnicode(u)       PyUnicode_FromUnicode(u, __Pyx_Py_UNICODE_strlen(u))
#define __Pyx_PyUnicode_FromUnicodeAndLength PyUnicode_FromUnicode
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
#define __Pyx_NewRef(obj) (Py_INCREF(obj), obj)
#define __Pyx_Owned_Py_None(b) __Pyx_NewRef(Py_None)
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b);
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x);
#define __Pyx_PySequence_Tuple(obj)\
    (likely(PyTuple_CheckExact(obj)) ? __Pyx_NewRef(obj) : PySequence_Tuple(obj))
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject*);
#if CYTHON_ASSUME_SAFE_MACROS
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#else
#define __Pyx_PyNumber_Int(x) (PyInt_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Int(x))
#endif
#define __Pyx_PyNumber_Float(x) (PyFloat_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Float(x))
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
static int __Pyx_sys_getdefaultencoding_not_ascii;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    PyObject* ascii_chars_u = NULL;
    PyObject* ascii_chars_b = NULL;
    const char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    if (strcmp(default_encoding_c, "ascii") == 0) {
        __Pyx_sys_getdefaultencoding_not_ascii = 0;
    } else {
        char ascii_chars[128];
        int c;
        for (c = 0; c < 128; c++) {
            ascii_chars[c] = c;
        }
        __Pyx_sys_getdefaultencoding_not_ascii = 1;
        ascii_chars_u = PyUnicode_DecodeASCII(ascii_chars, 128, NULL);
        if (!ascii_chars_u) goto bad;
        ascii_chars_b = PyUnicode_AsEncodedString(ascii_chars_u, default_encoding_c, NULL);
        if (!ascii_chars_b || !PyBytes_Check(ascii_chars_b) || memcmp(ascii_chars, PyBytes_AS_STRING(ascii_chars_b), 128) != 0) {
            PyErr_Format(
                PyExc_ValueError,
                "This module compiled with c_string_encoding=ascii, but default encoding '%.200s' is not a superset of ascii.",
                default_encoding_c);
            goto bad;
        }
        Py_DECREF(ascii_chars_u);
        Py_DECREF(ascii_chars_b);
    }
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    Py_XDECREF(ascii_chars_u);
    Py_XDECREF(ascii_chars_b);
    return -1;
}
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT && PY_MAJOR_VERSION >= 3
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#else
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
static char* __PYX_DEFAULT_STRING_ENCODING;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) (const char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    __PYX_DEFAULT_STRING_ENCODING = (char*) malloc(strlen(default_encoding_c) + 1);
    if (!__PYX_DEFAULT_STRING_ENCODING) goto bad;
    strcpy(__PYX_DEFAULT_STRING_ENCODING, default_encoding_c);
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    return -1;
}
#endif
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */
static CYTHON_INLINE void __Pyx_pretend_to_initialize(void* ptr) { (void)ptr; }

static PyObject *__pyx_m = NULL;
static PyObject *__pyx_d;
static PyObject *__pyx_b;
static PyObject *__pyx_cython_runtime = NULL;
static PyObject *__pyx_empty_tuple;
static PyObject *__pyx_empty_bytes;
static PyObject *__pyx_empty_unicode;
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm= __FILE__;
static const char *__pyx_filename;


static const char *__pyx_f[] = {
  "src/glycopeptidepy/_c/structure/fragment.pyx",
  "src/glycopeptidepy/_c/structure/fragment.pxd",
  "stringsource",
  "ccomposition.pxd",
  "type.pxd",
  "bool.pxd",
  "complex.pxd",
  "src/glycopeptidepy/_c/count_table.pxd",
  "src/glycopeptidepy/_c/structure/base.pxd",
};

/*--- Type declarations ---*/
struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition;
struct __pyx_obj_14glycopeptidepy_2_c_11count_table_CountTableIterator;
struct __pyx_obj_14glycopeptidepy_2_c_11count_table_CountTable;
struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_PeptideSequenceBase;
struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_TerminalGroup;
struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_AminoAcidResidueBase;
struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_ModificationBase;
struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_SequencePosition;
struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase;
struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase;
struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase;
struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_SimpleFragment;
struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_StubFragment;
struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment;
struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment__NameTree;
struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct____pyx_fuse_0__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition;
struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_1_genexpr;
struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_2___pyx_fuse_1__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition;
struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_3_genexpr;
struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_4___pyx_fuse_2__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition;
struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_5_genexpr;
struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_object____tuple___to_py;
struct __pyx_opt_args_5glypy_11composition_12ccomposition_12CComposition_calc_mass;
struct __pyx_opt_args_5glypy_11composition_12ccomposition_calculate_mass;

/* "glypy/composition/ccomposition.pxd":6
 *     cdef tuple _mass_args
 *     cpdef CComposition clone(self)
 *     cpdef double calc_mass(self, int average=?, charge=?, dict mass_data=?) except -1             # <<<<<<<<<<<<<<
 *     cpdef _from_formula(self, str formula, dict mass_data)
 *     cpdef _from_dict(self, comp)
 */
struct __pyx_opt_args_5glypy_11composition_12ccomposition_12CComposition_calc_mass {
  int __pyx_n;
  int average;
  PyObject *charge;
  PyObject *mass_data;
};

/* "glypy/composition/ccomposition.pxd":31
 * 
 * 
 * cpdef double calculate_mass(CComposition composition=?, str formula=?, int average=?, charge=?, mass_data=?) except -1             # <<<<<<<<<<<<<<
 */
struct __pyx_opt_args_5glypy_11composition_12ccomposition_calculate_mass {
  int __pyx_n;
  struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *composition;
  PyObject *formula;
  int average;
  PyObject *charge;
  PyObject *mass_data;
};
struct __pyx_t_14glycopeptidepy_2_c_11count_table_count_table_bin_cell;
struct __pyx_t_14glycopeptidepy_2_c_11count_table_count_table_bin;
struct __pyx_t_14glycopeptidepy_2_c_11count_table_count_table;
struct __pyx_opt_args_14glycopeptidepy_2_c_11count_table_10CountTable_pop;
struct __pyx_opt_args_14glycopeptidepy_2_c_11count_table_10CountTable_get;

/* "glycopeptidepy/_c/count_table.pxd":3
 * from cpython cimport PyObject
 * 
 * cdef struct count_table_bin_cell:             # <<<<<<<<<<<<<<
 *     long value
 *     PyObject* key
 */
struct __pyx_t_14glycopeptidepy_2_c_11count_table_count_table_bin_cell {
  long value;
  PyObject *key;
};

/* "glycopeptidepy/_c/count_table.pxd":8
 * 
 * 
 * cdef struct count_table_bin:             # <<<<<<<<<<<<<<
 *     count_table_bin_cell* cells
 *     size_t size
 */
struct __pyx_t_14glycopeptidepy_2_c_11count_table_count_table_bin {
  struct __pyx_t_14glycopeptidepy_2_c_11count_table_count_table_bin_cell *cells;
  size_t size;
  size_t used;
};

/* "glycopeptidepy/_c/count_table.pxd":14
 * 
 * 
 * cdef struct count_table:             # <<<<<<<<<<<<<<
 *     count_table_bin* bins
 *     size_t size
 */
struct __pyx_t_14glycopeptidepy_2_c_11count_table_count_table {
  struct __pyx_t_14glycopeptidepy_2_c_11count_table_count_table_bin *bins;
  size_t size;
  size_t count;
};

/* "glycopeptidepy/_c/count_table.pxd":100
 *     cpdef clear(self)
 *     cpdef setdefault(self, key, value)
 *     cpdef long pop(self, object key, object default=?) except *             # <<<<<<<<<<<<<<
 *     cpdef get(self, object key, object default=?)
 * 
 */
struct __pyx_opt_args_14glycopeptidepy_2_c_11count_table_10CountTable_pop {
  int __pyx_n;
  PyObject *__pyx_default;
};

/* "glycopeptidepy/_c/count_table.pxd":101
 *     cpdef setdefault(self, key, value)
 *     cpdef long pop(self, object key, object default=?) except *
 *     cpdef get(self, object key, object default=?)             # <<<<<<<<<<<<<<
 * 
 *     cdef void increment(self, object key, long value)
 */
struct __pyx_opt_args_14glycopeptidepy_2_c_11count_table_10CountTable_get {
  int __pyx_n;
  PyObject *__pyx_default;
};
struct __pyx_opt_args_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment__create;
struct __pyx_opt_args_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment__create;
struct __pyx_t_14glycopeptidepy_2_c_9structure_8fragment_string_cell;

/* "glycopeptidepy/_c/structure/fragment.pxd":84
 * 
 *     @staticmethod
 *     cdef StubFragment _create(str name, double mass, IonSeriesBase kind, CComposition composition,             # <<<<<<<<<<<<<<
 *                               ChemicalShiftBase chemical_shift, bint is_glycosylated, object glycosylation,
 *                               bint is_extended, int glycosylation_size=*)
 */
struct __pyx_opt_args_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment__create {
  int __pyx_n;
  int glycosylation_size;
};

/* "glycopeptidepy/_c/structure/fragment.pxd":108
 * 
 *     @staticmethod
 *     cdef PeptideFragment _create(IonSeriesBase kind, int position, CountTable modification_dict, double mass,             # <<<<<<<<<<<<<<
 *                                  list flanking_amino_acids=*, dict glycosylation=*,
 *                                  ChemicalShiftBase chemical_shift=*, CComposition composition=?, double* delta_mass=?)
 */
struct __pyx_opt_args_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment__create {
  int __pyx_n;
  PyObject *flanking_amino_acids;
  PyObject *glycosylation;
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase *chemical_shift;
  struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *composition;
  double *delta_mass;
};

/* "glycopeptidepy/_c/structure/fragment.pyx":37
 * 
 * 
 * cdef struct string_cell:             # <<<<<<<<<<<<<<
 *     char* string
 *     size_t size
 */
struct __pyx_t_14glycopeptidepy_2_c_9structure_8fragment_string_cell {
  char *string;
  size_t size;
};

/* "glypy/composition/ccomposition.pxd":2
 * 
 * cdef class CComposition(dict):             # <<<<<<<<<<<<<<
 *     cdef object _mass
 *     cdef tuple _mass_args
 */
struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition {
  PyDictObject __pyx_base;
  struct __pyx_vtabstruct_5glypy_11composition_12ccomposition_CComposition *__pyx_vtab;
  PyObject *_mass;
  PyObject *_mass_args;
};


/* "glycopeptidepy/_c/count_table.pxd":52
 * 
 * 
 * cdef class CountTableIterator(object):             # <<<<<<<<<<<<<<
 *     cdef :
 *         count_table* table
 */
struct __pyx_obj_14glycopeptidepy_2_c_11count_table_CountTableIterator {
  PyObject_HEAD
  struct __pyx_vtabstruct_14glycopeptidepy_2_c_11count_table_CountTableIterator *__pyx_vtab;
  struct __pyx_t_14glycopeptidepy_2_c_11count_table_count_table *table;
  size_t bin_index;
  size_t cell_index;
  PyObject *next_key;
  long next_value;
};


/* "glycopeptidepy/_c/count_table.pxd":69
 * 
 * 
 * cdef class CountTable(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         count_table* table
 */
struct __pyx_obj_14glycopeptidepy_2_c_11count_table_CountTable {
  PyObject_HEAD
  struct __pyx_vtabstruct_14glycopeptidepy_2_c_11count_table_CountTable *__pyx_vtab;
  struct __pyx_t_14glycopeptidepy_2_c_11count_table_count_table *table;
};


/* "glycopeptidepy/_c/structure/base.pxd":5
 * 
 * 
 * cdef class PeptideSequenceBase(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public list sequence
 */
struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_PeptideSequenceBase {
  PyObject_HEAD
  struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_4base_PeptideSequenceBase *__pyx_vtab;
  PyObject *sequence;
  double _mass;
  PyObject *_glycosylation_manager;
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_TerminalGroup *_n_term;
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_TerminalGroup *_c_term;
  struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *_total_composition;
  struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *_peptide_composition;
};


/* "glycopeptidepy/_c/structure/base.pxd":21
 * 
 * 
 * cdef class TerminalGroup(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public CComposition base_composition
 */
struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_TerminalGroup {
  PyObject_HEAD
  struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_4base_TerminalGroup *__pyx_vtab;
  struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *base_composition;
  double mass;
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_ModificationBase *_modification;
};


/* "glycopeptidepy/_c/structure/base.pxd":35
 * 
 * 
 * cdef class AminoAcidResidueBase(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public basestring name
 */
struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_AminoAcidResidueBase {
  PyObject_HEAD
  struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_4base_AminoAcidResidueBase *__pyx_vtab;
  PyObject *name;
  PyObject *symbol;
  double mass;
  struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *composition;
  PyObject *neutral_loss;
  Py_hash_t _hash;
};


/* "glycopeptidepy/_c/structure/base.pxd":50
 * 
 * 
 * cdef class ModificationBase(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public basestring name
 */
struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_ModificationBase {
  PyObject_HEAD
  struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_4base_ModificationBase *__pyx_vtab;
  PyObject *name;
  double mass;
  struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *composition;
};


/* "glycopeptidepy/_c/structure/base.pxd":82
 * 
 * 
 * cdef class SequencePosition(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public AminoAcidResidueBase amino_acid
 */
struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_SequencePosition {
  PyObject_HEAD
  struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_4base_SequencePosition *__pyx_vtab;
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_AminoAcidResidueBase *amino_acid;
  PyObject *modifications;
};


/* "glycopeptidepy/_c/structure/fragment.pxd":6
 * 
 * 
 * cdef class ChemicalShiftBase(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public str name
 */
struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase {
  PyObject_HEAD
  struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase *__pyx_vtab;
  PyObject *name;
  struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *composition;
  double mass;
};


/* "glycopeptidepy/_c/structure/fragment.pxd":15
 * 
 * 
 * cdef class IonSeriesBase(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public str name
 */
struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase {
  PyObject_HEAD
  PyObject *name;
  int direction;
  int includes_peptide;
  double mass_shift;
  struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *composition_shift;
  Py_hash_t _hash;
  int int_code;
};


/* "glycopeptidepy/_c/structure/fragment.pxd":26
 * 
 * 
 * cdef class FragmentBase(object):             # <<<<<<<<<<<<<<
 *     """Base class for all Fragment types. Defines basic
 *     name generation and neutral loss handling functions.
 */
struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase {
  PyObject_HEAD
  struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *__pyx_vtab;
  PyObject *_name;
  Py_hash_t _hash;
  double mass;
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase *_chemical_shift;
};


/* "glycopeptidepy/_c/structure/fragment.pxd":63
 * 
 * 
 * cdef class SimpleFragment(FragmentBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public IonSeriesBase kind
 */
struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_SimpleFragment {
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase __pyx_base;
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *kind;
  int is_glycosylated;
  struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *composition;
};


/* "glycopeptidepy/_c/structure/fragment.pxd":72
 * 
 * 
 * cdef class StubFragment(FragmentBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public IonSeriesBase kind
 */
struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_StubFragment {
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase __pyx_base;
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *kind;
  int is_glycosylated;
  struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *composition;
  PyObject *glycosylation;
  int is_extended;
  int _glycosylation_size;
};


/* "glycopeptidepy/_c/structure/fragment.pxd":97
 * 
 * 
 * cdef class PeptideFragment(FragmentBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public IonSeriesBase kind
 */
struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment {
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase __pyx_base;
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *kind;
  int position;
  struct __pyx_obj_14glycopeptidepy_2_c_11count_table_CountTable *modification_dict;
  double bare_mass;
  PyObject *flanking_amino_acids;
  PyObject *glycosylation;
  struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *composition;
};


/* "glycopeptidepy/_c/structure/fragment.pxd":125
 * 
 * @cython.final
 * cdef class _NameTree(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public object name
 */
struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment__NameTree {
  PyObject_HEAD
  struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_8fragment__NameTree *__pyx_vtab;
  PyObject *name;
  PyObject *value;
  PyObject *children;
};


/* "glycopeptidepy/_c/structure/fragment.pyx":675
 * 
 * 
 * cdef basestring build_name_from_composition(mapping_types glycan_composition):             # <<<<<<<<<<<<<<
 *         cdef:
 *             basestring name, extended_key
 */
struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct____pyx_fuse_0__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition {
  PyObject_HEAD
  PyObject *__pyx_v_parts;
};


/* "glycopeptidepy/_c/structure/fragment.pyx":694
 *         if extended_key is None:
 *             parts.sort(key=pair_mass)
 *             node.name = extended_key = ''.join("%s%d" % kv for kv in parts)             # <<<<<<<<<<<<<<
 *         if extended_key:
 *             name = "%s+%s" % (name, extended_key)
 */
struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_1_genexpr {
  PyObject_HEAD
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct____pyx_fuse_0__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition *__pyx_outer_scope;
  PyObject *__pyx_v_kv;
  PyObject *__pyx_t_0;
  Py_ssize_t __pyx_t_1;
};


/* "glycopeptidepy/_c/structure/fragment.pyx":675
 * 
 * 
 * cdef basestring build_name_from_composition(mapping_types glycan_composition):             # <<<<<<<<<<<<<<
 *         cdef:
 *             basestring name, extended_key
 */
struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_2___pyx_fuse_1__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition {
  PyObject_HEAD
  PyObject *__pyx_v_parts;
};


/* "glycopeptidepy/_c/structure/fragment.pyx":694
 *         if extended_key is None:
 *             parts.sort(key=pair_mass)
 *             node.name = extended_key = ''.join("%s%d" % kv for kv in parts)             # <<<<<<<<<<<<<<
 *         if extended_key:
 *             name = "%s+%s" % (name, extended_key)
 */
struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_3_genexpr {
  PyObject_HEAD
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_2___pyx_fuse_1__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition *__pyx_outer_scope;
  PyObject *__pyx_v_kv;
  PyObject *__pyx_t_0;
  Py_ssize_t __pyx_t_1;
};


/* "glycopeptidepy/_c/structure/fragment.pyx":675
 * 
 * 
 * cdef basestring build_name_from_composition(mapping_types glycan_composition):             # <<<<<<<<<<<<<<
 *         cdef:
 *             basestring name, extended_key
 */
struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_4___pyx_fuse_2__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition {
  PyObject_HEAD
  PyObject *__pyx_v_parts;
};


/* "glycopeptidepy/_c/structure/fragment.pyx":694
 *         if extended_key is None:
 *             parts.sort(key=pair_mass)
 *             node.name = extended_key = ''.join("%s%d" % kv for kv in parts)             # <<<<<<<<<<<<<<
 *         if extended_key:
 *             name = "%s+%s" % (name, extended_key)
 */
struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_5_genexpr {
  PyObject_HEAD
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_4___pyx_fuse_2__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition *__pyx_outer_scope;
  PyObject *__pyx_v_kv;
  PyObject *__pyx_t_0;
  Py_ssize_t __pyx_t_1;
};


/* "cfunc.to_py":64
 * 
 * @cname("__Pyx_CFunc_object____tuple___to_py")
 * cdef object __Pyx_CFunc_object____tuple___to_py(object (*f)(tuple) ):             # <<<<<<<<<<<<<<
 *     def wrap(tuple x):
 *         """wrap(x: tuple)"""
 */
struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_object____tuple___to_py {
  PyObject_HEAD
  PyObject *(*__pyx_v_f)(PyObject *);
};



/* "glypy/composition/ccomposition.pxd":2
 * 
 * cdef class CComposition(dict):             # <<<<<<<<<<<<<<
 *     cdef object _mass
 *     cdef tuple _mass_args
 */

struct __pyx_vtabstruct_5glypy_11composition_12ccomposition_CComposition {
  struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *(*clone)(struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *, int __pyx_skip_dispatch);
  double (*calc_mass)(struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *, int __pyx_skip_dispatch, struct __pyx_opt_args_5glypy_11composition_12ccomposition_12CComposition_calc_mass *__pyx_optional_args);
  PyObject *(*_from_formula)(struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *, PyObject *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*_from_dict)(struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *, PyObject *, int __pyx_skip_dispatch);
  long (*getitem)(struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *, PyObject *);
  void (*setitem)(struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *, PyObject *, long);
  struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *(*add_from)(struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *, PyObject *);
  struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *(*subtract_from)(struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *, PyObject *);
  struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *(*_create)(struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *);
};
static struct __pyx_vtabstruct_5glypy_11composition_12ccomposition_CComposition *__pyx_vtabptr_5glypy_11composition_12ccomposition_CComposition;


/* "glycopeptidepy/_c/count_table.pxd":52
 * 
 * 
 * cdef class CountTableIterator(object):             # <<<<<<<<<<<<<<
 *     cdef :
 *         count_table* table
 */

struct __pyx_vtabstruct_14glycopeptidepy_2_c_11count_table_CountTableIterator {
  struct __pyx_obj_14glycopeptidepy_2_c_11count_table_CountTableIterator *(*_create)(struct __pyx_obj_14glycopeptidepy_2_c_11count_table_CountTable *);
  void (*advance)(struct __pyx_obj_14glycopeptidepy_2_c_11count_table_CountTableIterator *);
  int (*has_more)(struct __pyx_obj_14glycopeptidepy_2_c_11count_table_CountTableIterator *);
  int (*_locate_next_value)(struct __pyx_obj_14glycopeptidepy_2_c_11count_table_CountTableIterator *, PyObject **, long *);
  int (*get_next_value)(struct __pyx_obj_14glycopeptidepy_2_c_11count_table_CountTableIterator *, PyObject **, long *);
};
static struct __pyx_vtabstruct_14glycopeptidepy_2_c_11count_table_CountTableIterator *__pyx_vtabptr_14glycopeptidepy_2_c_11count_table_CountTableIterator;


/* "glycopeptidepy/_c/count_table.pxd":69
 * 
 * 
 * cdef class CountTable(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         count_table* table
 */

struct __pyx_vtabstruct_14glycopeptidepy_2_c_11count_table_CountTable {
  struct __pyx_obj_14glycopeptidepy_2_c_11count_table_CountTable *(*_create)(void);
  struct __pyx_obj_14glycopeptidepy_2_c_11count_table_CountTable *(*_create_from)(struct __pyx_obj_14glycopeptidepy_2_c_11count_table_CountTable *);
  int (*_initialize_table)(struct __pyx_obj_14glycopeptidepy_2_c_11count_table_CountTable *);
  void (*_update_from_dict)(struct __pyx_obj_14glycopeptidepy_2_c_11count_table_CountTable *, PyObject *);
  void (*_update_from_count_table)(struct __pyx_obj_14glycopeptidepy_2_c_11count_table_CountTable *, struct __pyx_obj_14glycopeptidepy_2_c_11count_table_CountTable *);
  void (*_add_from)(struct __pyx_obj_14glycopeptidepy_2_c_11count_table_CountTable *, struct __pyx_obj_14glycopeptidepy_2_c_11count_table_CountTable *);
  void (*_subtract_from)(struct __pyx_obj_14glycopeptidepy_2_c_11count_table_CountTable *, struct __pyx_obj_14glycopeptidepy_2_c_11count_table_CountTable *);
  int (*_add_from_dict)(struct __pyx_obj_14glycopeptidepy_2_c_11count_table_CountTable *, PyObject *);
  int (*_subtract_from_dict)(struct __pyx_obj_14glycopeptidepy_2_c_11count_table_CountTable *, PyObject *);
  void (*_scale_by)(struct __pyx_obj_14glycopeptidepy_2_c_11count_table_CountTable *, long);
  int (*equal_to)(struct __pyx_obj_14glycopeptidepy_2_c_11count_table_CountTable *, struct __pyx_obj_14glycopeptidepy_2_c_11count_table_CountTable *);
  PyObject *(*_to_dict)(struct __pyx_obj_14glycopeptidepy_2_c_11count_table_CountTable *, int __pyx_skip_dispatch);
  PyObject *(*update)(struct __pyx_obj_14glycopeptidepy_2_c_11count_table_CountTable *, PyObject *, int __pyx_skip_dispatch);
  struct __pyx_obj_14glycopeptidepy_2_c_11count_table_CountTable *(*copy)(struct __pyx_obj_14glycopeptidepy_2_c_11count_table_CountTable *, int __pyx_skip_dispatch);
  PyObject *(*items)(struct __pyx_obj_14glycopeptidepy_2_c_11count_table_CountTable *, int __pyx_skip_dispatch);
  PyObject *(*values)(struct __pyx_obj_14glycopeptidepy_2_c_11count_table_CountTable *, int __pyx_skip_dispatch);
  PyObject *(*keys)(struct __pyx_obj_14glycopeptidepy_2_c_11count_table_CountTable *, int __pyx_skip_dispatch);
  PyObject *(*clear)(struct __pyx_obj_14glycopeptidepy_2_c_11count_table_CountTable *, int __pyx_skip_dispatch);
  PyObject *(*setdefault)(struct __pyx_obj_14glycopeptidepy_2_c_11count_table_CountTable *, PyObject *, PyObject *, int __pyx_skip_dispatch);
  long (*pop)(struct __pyx_obj_14glycopeptidepy_2_c_11count_table_CountTable *, PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_14glycopeptidepy_2_c_11count_table_10CountTable_pop *__pyx_optional_args);
  PyObject *(*get)(struct __pyx_obj_14glycopeptidepy_2_c_11count_table_CountTable *, PyObject *, int __pyx_skip_dispatch, struct __pyx_opt_args_14glycopeptidepy_2_c_11count_table_10CountTable_get *__pyx_optional_args);
  void (*increment)(struct __pyx_obj_14glycopeptidepy_2_c_11count_table_CountTable *, PyObject *, long);
  void (*decrement)(struct __pyx_obj_14glycopeptidepy_2_c_11count_table_CountTable *, PyObject *, long);
  long (*getitem)(struct __pyx_obj_14glycopeptidepy_2_c_11count_table_CountTable *, PyObject *);
  int (*setitem)(struct __pyx_obj_14glycopeptidepy_2_c_11count_table_CountTable *, PyObject *, long);
  long (*delitem)(struct __pyx_obj_14glycopeptidepy_2_c_11count_table_CountTable *, PyObject *);
  long (*total)(struct __pyx_obj_14glycopeptidepy_2_c_11count_table_CountTable *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_14glycopeptidepy_2_c_11count_table_CountTable *__pyx_vtabptr_14glycopeptidepy_2_c_11count_table_CountTable;


/* "glycopeptidepy/_c/structure/base.pxd":5
 * 
 * 
 * cdef class PeptideSequenceBase(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public list sequence
 */

struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_4base_PeptideSequenceBase {
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_SequencePosition *(*get)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_PeptideSequenceBase *, Py_ssize_t);
  PyObject *(*_invalidate)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_PeptideSequenceBase *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_4base_PeptideSequenceBase *__pyx_vtabptr_14glycopeptidepy_2_c_9structure_4base_PeptideSequenceBase;


/* "glycopeptidepy/_c/structure/base.pxd":21
 * 
 * 
 * cdef class TerminalGroup(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public CComposition base_composition
 */

struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_4base_TerminalGroup {
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_ModificationBase *(*get_modification)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_TerminalGroup *);
  struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *(*get_composition)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_TerminalGroup *);
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_TerminalGroup *(*modify)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_TerminalGroup *, struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_ModificationBase *, int __pyx_skip_dispatch);
  void (*set_modification)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_TerminalGroup *, struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_ModificationBase *);
  int (*equal_to)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_TerminalGroup *, PyObject *);
  PyObject *(*serialize)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_TerminalGroup *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_4base_TerminalGroup *__pyx_vtabptr_14glycopeptidepy_2_c_9structure_4base_TerminalGroup;


/* "glycopeptidepy/_c/structure/base.pxd":35
 * 
 * 
 * cdef class AminoAcidResidueBase(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public basestring name
 */

struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_4base_AminoAcidResidueBase {
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_AminoAcidResidueBase *(*_create)(PyObject *, PyObject *, double, struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *);
  int (*equal_to)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_AminoAcidResidueBase *, struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_AminoAcidResidueBase *);
};
static struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_4base_AminoAcidResidueBase *__pyx_vtabptr_14glycopeptidepy_2_c_9structure_4base_AminoAcidResidueBase;


/* "glycopeptidepy/_c/structure/base.pxd":50
 * 
 * 
 * cdef class ModificationBase(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public basestring name
 */

struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_4base_ModificationBase {
  int (*is_a)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_ModificationBase *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*serialize)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_ModificationBase *, int __pyx_skip_dispatch);
  int (*is_tracked_for)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_ModificationBase *, PyObject *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_4base_ModificationBase *__pyx_vtabptr_14glycopeptidepy_2_c_9structure_4base_ModificationBase;


/* "glycopeptidepy/_c/structure/base.pxd":82
 * 
 * 
 * cdef class SequencePosition(object):             # <<<<<<<<<<<<<<
 *     cdef:
 *         public AminoAcidResidueBase amino_acid
 */

struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_4base_SequencePosition {
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_SequencePosition *(*_create)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_AminoAcidResidueBase *, PyObject *);
  double (*get_mass)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_SequencePosition *);
  int (*has_modification)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_SequencePosition *, PyObject *, int __pyx_skip_dispatch);
  int (*is_modified)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_SequencePosition *, int __pyx_skip_dispatch);
  PyObject *(*drop_modification)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_SequencePosition *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*add_modification)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_SequencePosition *, PyObject *, int __pyx_skip_dispatch);
  size_t (*get_modification_count)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_SequencePosition *);
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_ModificationBase *(*get_modification)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_SequencePosition *, size_t);
};
static struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_4base_SequencePosition *__pyx_vtabptr_14glycopeptidepy_2_c_9structure_4base_SequencePosition;


/* "glycopeptidepy/_c/structure/fragment.pyx":116
 * 
 * @cython.freelist(100)
 * cdef class ChemicalShiftBase(object):             # <<<<<<<<<<<<<<
 * 
 *     cpdef ChemicalShiftBase clone(self):
 */

struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase {
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase *(*clone)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase *__pyx_vtabptr_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase;


/* "glycopeptidepy/_c/structure/fragment.pyx":166
 * 
 * @cython.freelist(1000000)
 * cdef class FragmentBase(object):             # <<<<<<<<<<<<<<
 *     """Base class for all Fragment types. Defines basic
 *     name generation and neutral loss handling functions.
 */

struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase {
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *(*get_series)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *, int __pyx_skip_dispatch);
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase *(*get_chemical_shift)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *, int __pyx_skip_dispatch);
  PyObject *(*set_chemical_shift)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *, struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase *, int __pyx_skip_dispatch);
  PyObject *(*base_name)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *, int __pyx_skip_dispatch);
  PyObject *(*get_fragment_name)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *, int __pyx_skip_dispatch);
  void (*_update_hash_name)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *);
  PyObject *(*get_name)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *);
};
static struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *__pyx_vtabptr_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase;


/* "glycopeptidepy/_c/structure/fragment.pyx":575
 * 
 * 
 * cdef class SimpleFragment(FragmentBase):             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, name, mass, kind, composition, chemical_shift=None, is_glycosylated=False):
 */

struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_8fragment_SimpleFragment {
  struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase __pyx_base;
  PyObject *(*clone)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_SimpleFragment *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_8fragment_SimpleFragment *__pyx_vtabptr_14glycopeptidepy_2_c_9structure_8fragment_SimpleFragment;


/* "glycopeptidepy/_c/structure/fragment.pyx":700
 * 
 * 
 * cdef class StubFragment(FragmentBase):             # <<<<<<<<<<<<<<
 * 
 *     _name_cache = stub_fragment_name_cache
 */

struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_8fragment_StubFragment {
  struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase __pyx_base;
  PyObject *(*clone)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_StubFragment *, int __pyx_skip_dispatch);
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_StubFragment *(*_create)(PyObject *, double, struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *, struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *, struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase *, int, PyObject *, int, struct __pyx_opt_args_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment__create *__pyx_optional_args);
  int (*get_glycosylation_size)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_StubFragment *);
};
static struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_8fragment_StubFragment *__pyx_vtabptr_14glycopeptidepy_2_c_9structure_8fragment_StubFragment;


/* "glycopeptidepy/_c/structure/fragment.pyx":285
 * 
 * 
 * cdef class PeptideFragment(FragmentBase):             # <<<<<<<<<<<<<<
 *     '''Represents a peptide backbone fragment, such as a peptide b or y ion.
 * 
 */

struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment {
  struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase __pyx_base;
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *(*_create)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *, int, struct __pyx_obj_14glycopeptidepy_2_c_11count_table_CountTable *, double, struct __pyx_opt_args_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment__create *__pyx_optional_args);
  PyObject *(*clone)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *, int __pyx_skip_dispatch);
  void (*_update_mass_with_modifications)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *);
  int (*_is_glycosylated)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *);
  long (*get_glycosylation_size)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *);
};
static struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *__pyx_vtabptr_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment;


/* "glycopeptidepy/_c/structure/fragment.pyx":614
 * 
 * @cython.final
 * cdef class _NameTree(object):             # <<<<<<<<<<<<<<
 *     @staticmethod
 *     cdef _NameTree _create():
 */

struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_8fragment__NameTree {
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment__NameTree *(*_create)(void);
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment__NameTree *(*get)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment__NameTree *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*__pyx_fuse_0traverse)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment__NameTree *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*__pyx_fuse_1traverse)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment__NameTree *, PyObject *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_8fragment__NameTree *__pyx_vtabptr_14glycopeptidepy_2_c_9structure_8fragment__NameTree;
static struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment__NameTree *__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree__create(void);
static struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment__NameTree *__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_get(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment__NameTree *, PyObject *, int __pyx_skip_dispatch);

/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, int);
    void (*DECREF)(void*, PyObject*, int);
    void (*GOTREF)(void*, PyObject*, int);
    void (*GIVEREF)(void*, PyObject*, int);
    void* (*SetupContext)(const char*, int, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__)
#endif
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_XINCREF(r)  do { if((r) != NULL) {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) != NULL) {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) != NULL) {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) != NULL) {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyObjectGetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* GetBuiltinName.proto */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* PyCFunctionFastCall.proto */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject *__Pyx_PyCFunction_FastCall(PyObject *func, PyObject **args, Py_ssize_t nargs);
#else
#define __Pyx_PyCFunction_FastCall(func, args, nargs)  (assert(0), NULL)
#endif

/* PyFunctionFastCall.proto */
#if CYTHON_FAST_PYCALL
#define __Pyx_PyFunction_FastCall(func, args, nargs)\
    __Pyx_PyFunction_FastCallDict((func), (args), (nargs), NULL)
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs);
#else
#define __Pyx_PyFunction_FastCallDict(func, args, nargs, kwargs) _PyFunction_FastCallDict(func, args, nargs, kwargs)
#endif
#define __Pyx_BUILD_ASSERT_EXPR(cond)\
    (sizeof(char [1 - 2*!(cond)]) - 1)
#ifndef Py_MEMBER_SIZE
#define Py_MEMBER_SIZE(type, member) sizeof(((type *)0)->member)
#endif
#if CYTHON_FAST_PYCALL
  static size_t __pyx_pyframe_localsplus_offset = 0;
  #include "frameobject.h"
#if PY_VERSION_HEX >= 0x030b00a6
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
  #define __Pxy_PyFrame_Initialize_Offsets()\
    ((void)__Pyx_BUILD_ASSERT_EXPR(sizeof(PyFrameObject) == offsetof(PyFrameObject, f_localsplus) + Py_MEMBER_SIZE(PyFrameObject, f_localsplus)),\
     (void)(__pyx_pyframe_localsplus_offset = ((size_t)PyFrame_Type.tp_basicsize) - Py_MEMBER_SIZE(PyFrameObject, f_localsplus)))
  #define __Pyx_PyFrame_GetLocalsplus(frame)\
    (assert(__pyx_pyframe_localsplus_offset), (PyObject **)(((char *)(frame)) + __pyx_pyframe_localsplus_offset))
#endif // CYTHON_FAST_PYCALL
#endif

/* PyObjectCall.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* PyObjectCallMethO.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectCallOneArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

/* PyThreadStateGet.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = __Pyx_PyThreadState_Current;
#define __Pyx_PyErr_Occurred()  __pyx_tstate->curexc_type
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#define __Pyx_PyErr_Occurred()  PyErr_Occurred()
#endif

/* PyErrFetchRestore.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_Clear() __Pyx_ErrRestore(NULL, NULL, NULL)
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_PyErr_SetNone(exc) (Py_INCREF(exc), __Pyx_ErrRestore((exc), NULL, NULL))
#else
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#endif
#else
#define __Pyx_PyErr_Clear() PyErr_Clear()
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestoreInState(tstate, type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchInState(tstate, type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* WriteUnraisableException.proto */
static void __Pyx_WriteUnraisable(const char *name, int clineno,
                                  int lineno, const char *filename,
                                  int full_traceback, int nogil);

/* PyDictVersioning.proto */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
#define __PYX_DICT_VERSION_INIT  ((PY_UINT64_T) -1)
#define __PYX_GET_DICT_VERSION(dict)  (((PyDictObject*)(dict))->ma_version_tag)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)\
    (version_var) = __PYX_GET_DICT_VERSION(dict);\
    (cache_var) = (value);
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP) {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    if (likely(__PYX_GET_DICT_VERSION(DICT) == __pyx_dict_version)) {\
        (VAR) = __pyx_dict_cached_value;\
    } else {\
        (VAR) = __pyx_dict_cached_value = (LOOKUP);\
        __pyx_dict_version = __PYX_GET_DICT_VERSION(DICT);\
    }\
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj);
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj);
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version);
#else
#define __PYX_GET_DICT_VERSION(dict)  (0)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP)  (VAR) = (LOOKUP);
#endif

/* PyObjectCallNoArg.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func);
#else
#define __Pyx_PyObject_CallNoArg(func) __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL)
#endif

/* ExtTypeTest.proto */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type);

/* PyErrExceptionMatches.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_ExceptionMatches(err) __Pyx_PyErr_ExceptionMatchesInState(__pyx_tstate, err)
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err);
#else
#define __Pyx_PyErr_ExceptionMatches(err)  PyErr_ExceptionMatches(err)
#endif

/* GetAttr.proto */
static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *, PyObject *);

/* GetAttr3.proto */
static CYTHON_INLINE PyObject *__Pyx_GetAttr3(PyObject *, PyObject *, PyObject *);

/* GetModuleGlobalName.proto */
#if CYTHON_USE_DICT_VERSIONS
#define __Pyx_GetModuleGlobalName(var, name)  do {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    (var) = (likely(__pyx_dict_version == __PYX_GET_DICT_VERSION(__pyx_d))) ?\
        (likely(__pyx_dict_cached_value) ? __Pyx_NewRef(__pyx_dict_cached_value) : __Pyx_GetBuiltinName(name)) :\
        __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
} while(0)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  do {\
    PY_UINT64_T __pyx_dict_version;\
    PyObject *__pyx_dict_cached_value;\
    (var) = __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
} while(0)
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value);
#else
#define __Pyx_GetModuleGlobalName(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name);
#endif

/* IncludeStringH.proto */
#include <string.h>

/* BytesEquals.proto */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals);

/* UnicodeEquals.proto */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals);

/* StrEquals.proto */
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyString_Equals __Pyx_PyUnicode_Equals
#else
#define __Pyx_PyString_Equals __Pyx_PyBytes_Equals
#endif

/* RaiseException.proto */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* py_abs.proto */
#if CYTHON_USE_PYLONG_INTERNALS
static PyObject *__Pyx_PyLong_AbsNeg(PyObject *num);
#define __Pyx_PyNumber_Absolute(x)\
    ((likely(PyLong_CheckExact(x))) ?\
         (likely(Py_SIZE(x) >= 0) ? (Py_INCREF(x), (x)) : __Pyx_PyLong_AbsNeg(x)) :\
         PyNumber_Absolute(x))
#else
#define __Pyx_PyNumber_Absolute(x)  PyNumber_Absolute(x)
#endif

/* GetTopmostException.proto */
#if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem * __Pyx_PyErr_GetTopmostException(PyThreadState *tstate);
#endif

/* SaveResetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSave(type, value, tb)  __Pyx__ExceptionSave(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#define __Pyx_ExceptionReset(type, value, tb)  __Pyx__ExceptionReset(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
#else
#define __Pyx_ExceptionSave(type, value, tb)   PyErr_GetExcInfo(type, value, tb)
#define __Pyx_ExceptionReset(type, value, tb)  PyErr_SetExcInfo(type, value, tb)
#endif

/* GetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_GetException(type, value, tb)  __Pyx__GetException(__pyx_tstate, type, value, tb)
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* PyObjectCall2Args.proto */
static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2);

/* ArgTypeTest.proto */
#define __Pyx_ArgTypeTest(obj, type, none_allowed, name, exact)\
    ((likely((Py_TYPE(obj) == type) | (none_allowed && (obj == Py_None)))) ? 1 :\
        __Pyx__ArgTypeTest(obj, type, name, exact))
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact);

/* ListAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_PyList_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len) & likely(len > (L->allocated >> 1))) {
        Py_INCREF(x);
        PyList_SET_ITEM(list, len, x);
        __Pyx_SET_SIZE(list, len + 1);
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_PyList_Append(L,x) PyList_Append(L,x)
#endif

/* StringJoin.proto */
#if PY_MAJOR_VERSION < 3
#define __Pyx_PyString_Join __Pyx_PyBytes_Join
#define __Pyx_PyBaseString_Join(s, v) (PyUnicode_CheckExact(s) ? PyUnicode_Join(s, v) : __Pyx_PyBytes_Join(s, v))
#else
#define __Pyx_PyString_Join PyUnicode_Join
#define __Pyx_PyBaseString_Join PyUnicode_Join
#endif
#if CYTHON_COMPILING_IN_CPYTHON
    #if PY_MAJOR_VERSION < 3
    #define __Pyx_PyBytes_Join _PyString_Join
    #else
    #define __Pyx_PyBytes_Join _PyBytes_Join
    #endif
#else
static CYTHON_INLINE PyObject* __Pyx_PyBytes_Join(PyObject* sep, PyObject* values);
#endif

/* RaiseArgTupleInvalid.proto */
static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

/* RaiseDoubleKeywords.proto */
static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

/* ParseKeywords.proto */
static int __Pyx_ParseOptionalKeywords(PyObject *kwds, PyObject **argnames[],\
    PyObject *kwds2, PyObject *values[], Py_ssize_t num_pos_args,\
    const char* function_name);

/* GetItemInt.proto */
#define __Pyx_GetItemInt(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Fast(o, (Py_ssize_t)i, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL) :\
               __Pyx_GetItemInt_Generic(o, to_py_func(i))))
#define __Pyx_GetItemInt_List(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_List_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
#define __Pyx_GetItemInt_Tuple(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Tuple_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "tuple index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j);
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i,
                                                     int is_list, int wraparound, int boundscheck);

/* PyDictContains.proto */
static CYTHON_INLINE int __Pyx_PyDict_ContainsTF(PyObject* item, PyObject* dict, int eq) {
    int result = PyDict_Contains(dict, item);
    return unlikely(result < 0) ? result : (result == (eq == Py_EQ));
}

/* DictGetItem.proto */
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key);
#define __Pyx_PyObject_Dict_GetItem(obj, name)\
    (likely(PyDict_CheckExact(obj)) ?\
     __Pyx_PyDict_GetItem(obj, name) : PyObject_GetItem(obj, name))
#else
#define __Pyx_PyDict_GetItem(d, key) PyObject_GetItem(d, key)
#define __Pyx_PyObject_Dict_GetItem(obj, name)  PyObject_GetItem(obj, name)
#endif

/* SetItemInt.proto */
#define __Pyx_SetItemInt(o, i, v, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_SetItemInt_Fast(o, (Py_ssize_t)i, v, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list assignment index out of range"), -1) :\
               __Pyx_SetItemInt_Generic(o, to_py_func(i), v)))
static int __Pyx_SetItemInt_Generic(PyObject *o, PyObject *j, PyObject *v);
static CYTHON_INLINE int __Pyx_SetItemInt_Fast(PyObject *o, Py_ssize_t i, PyObject *v,
                                               int is_list, int wraparound, int boundscheck);

/* IterFinish.proto */
static CYTHON_INLINE int __Pyx_IterFinish(void);

/* PyObjectGetMethod.proto */
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method);

/* PyObjectCallMethod0.proto */
static PyObject* __Pyx_PyObject_CallMethod0(PyObject* obj, PyObject* method_name);

/* RaiseNeedMoreValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index);

/* RaiseTooManyValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected);

/* UnpackItemEndCheck.proto */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected);

/* RaiseNoneIterError.proto */
static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void);

/* UnpackTupleError.proto */
static void __Pyx_UnpackTupleError(PyObject *, Py_ssize_t index);

/* UnpackTuple2.proto */
#define __Pyx_unpack_tuple2(tuple, value1, value2, is_tuple, has_known_size, decref_tuple)\
    (likely(is_tuple || PyTuple_Check(tuple)) ?\
        (likely(has_known_size || PyTuple_GET_SIZE(tuple) == 2) ?\
            __Pyx_unpack_tuple2_exact(tuple, value1, value2, decref_tuple) :\
            (__Pyx_UnpackTupleError(tuple, 2), -1)) :\
        __Pyx_unpack_tuple2_generic(tuple, value1, value2, has_known_size, decref_tuple))
static CYTHON_INLINE int __Pyx_unpack_tuple2_exact(
    PyObject* tuple, PyObject** value1, PyObject** value2, int decref_tuple);
static int __Pyx_unpack_tuple2_generic(
    PyObject* tuple, PyObject** value1, PyObject** value2, int has_known_size, int decref_tuple);

/* dict_iter.proto */
static CYTHON_INLINE PyObject* __Pyx_dict_iterator(PyObject* dict, int is_dict, PyObject* method_name,
                                                   Py_ssize_t* p_orig_length, int* p_is_dict);
static CYTHON_INLINE int __Pyx_dict_iter_next(PyObject* dict_or_iter, Py_ssize_t orig_length, Py_ssize_t* ppos,
                                              PyObject** pkey, PyObject** pvalue, PyObject** pitem, int is_dict);

/* None.proto */
static CYTHON_INLINE void __Pyx_RaiseClosureNameError(const char *varname);

/* PySequenceContains.proto */
static CYTHON_INLINE int __Pyx_PySequence_ContainsTF(PyObject* item, PyObject* seq, int eq) {
    int result = PySequence_Contains(seq, item);
    return unlikely(result < 0) ? result : (result == (eq == Py_EQ));
}

/* Import.proto */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level);

/* ImportFrom.proto */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name);

/* HasAttr.proto */
static CYTHON_INLINE int __Pyx_HasAttr(PyObject *, PyObject *);

/* FetchCommonType.proto */
static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type);

/* CythonFunctionShared.proto */
#define __Pyx_CyFunction_USED 1
#define __Pyx_CYFUNCTION_STATICMETHOD  0x01
#define __Pyx_CYFUNCTION_CLASSMETHOD   0x02
#define __Pyx_CYFUNCTION_CCLASS        0x04
#define __Pyx_CyFunction_GetClosure(f)\
    (((__pyx_CyFunctionObject *) (f))->func_closure)
#define __Pyx_CyFunction_GetClassObj(f)\
    (((__pyx_CyFunctionObject *) (f))->func_classobj)
#define __Pyx_CyFunction_Defaults(type, f)\
    ((type *)(((__pyx_CyFunctionObject *) (f))->defaults))
#define __Pyx_CyFunction_SetDefaultsGetter(f, g)\
    ((__pyx_CyFunctionObject *) (f))->defaults_getter = (g)
typedef struct {
    PyCFunctionObject func;
#if PY_VERSION_HEX < 0x030500A0
    PyObject *func_weakreflist;
#endif
    PyObject *func_dict;
    PyObject *func_name;
    PyObject *func_qualname;
    PyObject *func_doc;
    PyObject *func_globals;
    PyObject *func_code;
    PyObject *func_closure;
    PyObject *func_classobj;
    void *defaults;
    int defaults_pyobjects;
    size_t defaults_size;  // used by FusedFunction for copying defaults
    int flags;
    PyObject *defaults_tuple;
    PyObject *defaults_kwdict;
    PyObject *(*defaults_getter)(PyObject *);
    PyObject *func_annotations;
} __pyx_CyFunctionObject;
static PyTypeObject *__pyx_CyFunctionType = 0;
#define __Pyx_CyFunction_Check(obj)  (__Pyx_TypeCheck(obj, __pyx_CyFunctionType))
static PyObject *__Pyx_CyFunction_Init(__pyx_CyFunctionObject* op, PyMethodDef *ml,
                                      int flags, PyObject* qualname,
                                      PyObject *self,
                                      PyObject *module, PyObject *globals,
                                      PyObject* code);
static CYTHON_INLINE void *__Pyx_CyFunction_InitDefaults(PyObject *m,
                                                         size_t size,
                                                         int pyobjects);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *m,
                                                            PyObject *tuple);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *m,
                                                             PyObject *dict);
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *m,
                                                              PyObject *dict);
static int __pyx_CyFunction_init(void);

/* CythonFunction.proto */
static PyObject *__Pyx_CyFunction_New(PyMethodDef *ml,
                                      int flags, PyObject* qualname,
                                      PyObject *closure,
                                      PyObject *module, PyObject *globals,
                                      PyObject* code);

/* PyObject_GenericGetAttrNoDict.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttrNoDict PyObject_GenericGetAttr
#endif

/* PyObject_GenericGetAttr.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttr PyObject_GenericGetAttr
#endif

/* SetVTable.proto */
static int __Pyx_SetVtable(PyObject *dict, void *vtable);

/* PyObjectGetAttrStrNoError.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name);

/* SetupReduce.proto */
static int __Pyx_setup_reduce(PyObject* type_obj);

/* TypeImport.proto */
#ifndef __PYX_HAVE_RT_ImportType_proto
#define __PYX_HAVE_RT_ImportType_proto
enum __Pyx_ImportType_CheckSize {
   __Pyx_ImportType_CheckSize_Error = 0,
   __Pyx_ImportType_CheckSize_Warn = 1,
   __Pyx_ImportType_CheckSize_Ignore = 2
};
static PyTypeObject *__Pyx_ImportType(PyObject* module, const char *module_name, const char *class_name, size_t size, enum __Pyx_ImportType_CheckSize check_size);
#endif

/* GetVTable.proto */
static void* __Pyx_GetVtable(PyObject *dict);

/* GetNameInClass.proto */
#define __Pyx_GetNameInClass(var, nmspace, name)  (var) = __Pyx__GetNameInClass(nmspace, name)
static PyObject *__Pyx__GetNameInClass(PyObject *nmspace, PyObject *name);

/* FusedFunction.proto */
typedef struct {
    __pyx_CyFunctionObject func;
    PyObject *__signatures__;
    PyObject *type;
    PyObject *self;
} __pyx_FusedFunctionObject;
static PyObject *__pyx_FusedFunction_New(PyMethodDef *ml, int flags,
                                         PyObject *qualname, PyObject *closure,
                                         PyObject *module, PyObject *globals,
                                         PyObject *code);
static int __pyx_FusedFunction_clear(__pyx_FusedFunctionObject *self);
static PyTypeObject *__pyx_FusedFunctionType = NULL;
static int __pyx_FusedFunction_init(void);
#define __Pyx_FusedFunction_USED

/* ClassMethod.proto */
#include "descrobject.h"
static CYTHON_UNUSED PyObject* __Pyx_Method_ClassMethod(PyObject *method);

/* CLineInTraceback.proto */
#ifdef CYTHON_CLINE_IN_TRACEBACK
#define __Pyx_CLineForTraceback(tstate, c_line)  (((CYTHON_CLINE_IN_TRACEBACK)) ? c_line : 0)
#else
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line);
#endif

/* CodeObjectCache.proto */
typedef struct {
    PyCodeObject* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

/* GCCDiagnostics.proto */
#if defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 6))
#define __Pyx_HAS_GCC_DIAGNOSTIC
#endif

/* Print.proto */
static int __Pyx_Print(PyObject*, PyObject *, int);
#if CYTHON_COMPILING_IN_PYPY || PY_MAJOR_VERSION >= 3
static PyObject* __pyx_print = 0;
static PyObject* __pyx_print_kwargs = 0;
#endif

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value);

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value);

/* CIntFromPy.proto */
static CYTHON_INLINE size_t __Pyx_PyInt_As_size_t(PyObject *);

/* PrintOne.proto */
static int __Pyx_PrintOne(PyObject* stream, PyObject *o);

/* ImportNumPyArray.proto */
static PyObject *__pyx_numpy_ndarray = NULL;
static PyObject* __Pyx_ImportNumPyArrayTypeIfAvailable(void);

/* FastTypeChecks.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_TypeCheck(obj, type) __Pyx_IsSubtype(Py_TYPE(obj), (PyTypeObject *)type)
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject *type);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2);
#else
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_PyErr_GivenExceptionMatches(err, type) PyErr_GivenExceptionMatches(err, type)
#define __Pyx_PyErr_GivenExceptionMatches2(err, type1, type2) (PyErr_GivenExceptionMatches(err, type1) || PyErr_GivenExceptionMatches(err, type2))
#endif
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)

/* SwapException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSwap(type, value, tb)  __Pyx__ExceptionSwap(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* PyObjectCallMethod1.proto */
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg);

/* CoroutineBase.proto */
typedef PyObject *(*__pyx_coroutine_body_t)(PyObject *, PyThreadState *, PyObject *);
#if CYTHON_USE_EXC_INFO_STACK
#define __Pyx_ExcInfoStruct  _PyErr_StackItem
#else
typedef struct {
    PyObject *exc_type;
    PyObject *exc_value;
    PyObject *exc_traceback;
} __Pyx_ExcInfoStruct;
#endif
typedef struct {
    PyObject_HEAD
    __pyx_coroutine_body_t body;
    PyObject *closure;
    __Pyx_ExcInfoStruct gi_exc_state;
    PyObject *gi_weakreflist;
    PyObject *classobj;
    PyObject *yieldfrom;
    PyObject *gi_name;
    PyObject *gi_qualname;
    PyObject *gi_modulename;
    PyObject *gi_code;
    PyObject *gi_frame;
    int resume_label;
    char is_running;
} __pyx_CoroutineObject;
static __pyx_CoroutineObject *__Pyx__Coroutine_New(
    PyTypeObject *type, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
    PyObject *name, PyObject *qualname, PyObject *module_name);
static __pyx_CoroutineObject *__Pyx__Coroutine_NewInit(
            __pyx_CoroutineObject *gen, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name);
static CYTHON_INLINE void __Pyx_Coroutine_ExceptionClear(__Pyx_ExcInfoStruct *self);
static int __Pyx_Coroutine_clear(PyObject *self);
static PyObject *__Pyx_Coroutine_Send(PyObject *self, PyObject *value);
static PyObject *__Pyx_Coroutine_Close(PyObject *self);
static PyObject *__Pyx_Coroutine_Throw(PyObject *gen, PyObject *args);
#if CYTHON_USE_EXC_INFO_STACK
#define __Pyx_Coroutine_SwapException(self)
#define __Pyx_Coroutine_ResetAndClearException(self)  __Pyx_Coroutine_ExceptionClear(&(self)->gi_exc_state)
#else
#define __Pyx_Coroutine_SwapException(self) {\
    __Pyx_ExceptionSwap(&(self)->gi_exc_state.exc_type, &(self)->gi_exc_state.exc_value, &(self)->gi_exc_state.exc_traceback);\
    __Pyx_Coroutine_ResetFrameBackpointer(&(self)->gi_exc_state);\
    }
#define __Pyx_Coroutine_ResetAndClearException(self) {\
    __Pyx_ExceptionReset((self)->gi_exc_state.exc_type, (self)->gi_exc_state.exc_value, (self)->gi_exc_state.exc_traceback);\
    (self)->gi_exc_state.exc_type = (self)->gi_exc_state.exc_value = (self)->gi_exc_state.exc_traceback = NULL;\
    }
#endif
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyGen_FetchStopIterationValue(pvalue)\
    __Pyx_PyGen__FetchStopIterationValue(__pyx_tstate, pvalue)
#else
#define __Pyx_PyGen_FetchStopIterationValue(pvalue)\
    __Pyx_PyGen__FetchStopIterationValue(__Pyx_PyThreadState_Current, pvalue)
#endif
static int __Pyx_PyGen__FetchStopIterationValue(PyThreadState *tstate, PyObject **pvalue);
static CYTHON_INLINE void __Pyx_Coroutine_ResetFrameBackpointer(__Pyx_ExcInfoStruct *exc_state);

/* PatchModuleWithCoroutine.proto */
static PyObject* __Pyx_Coroutine_patch_module(PyObject* module, const char* py_code);

/* PatchGeneratorABC.proto */
static int __Pyx_patch_abc(void);

/* Generator.proto */
#define __Pyx_Generator_USED
static PyTypeObject *__pyx_GeneratorType = 0;
#define __Pyx_Generator_CheckExact(obj) (Py_TYPE(obj) == __pyx_GeneratorType)
#define __Pyx_Generator_New(body, code, closure, name, qualname, module_name)\
    __Pyx__Coroutine_New(__pyx_GeneratorType, body, code, closure, name, qualname, module_name)
static PyObject *__Pyx_Generator_Next(PyObject *self);
static int __pyx_Generator_init(void);

/* CheckBinaryVersion.proto */
static int __Pyx_check_binary_version(void);

/* FunctionExport.proto */
static int __Pyx_ExportFunction(const char *name, void (*f)(void), const char *sig);

/* VoidPtrImport.proto */
static int __Pyx_ImportVoidPtr(PyObject *module, const char *name, void **p, const char *sig);

/* FunctionImport.proto */
static int __Pyx_ImportFunction(PyObject *module, const char *funcname, void (**f)(void), const char *sig);

/* InitStrings.proto */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t);

static struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase *__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_17ChemicalShiftBase_clone(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_get_series(CYTHON_UNUSED struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase *__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_get_chemical_shift(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_set_chemical_shift(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *__pyx_v_self, struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase *__pyx_v_chemical_shift, int __pyx_skip_dispatch); /* proto*/
static void __pyx_f_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase__update_hash_name(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_get_fragment_name(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_base_name(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_get_name(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *__pyx_v_self); /* proto*/
static struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment__create(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *__pyx_v_kind, int __pyx_v_position, struct __pyx_obj_14glycopeptidepy_2_c_11count_table_CountTable *__pyx_v_modification_dict, double __pyx_v_mass, struct __pyx_opt_args_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment__create *__pyx_optional_args); /* proto*/
static void __pyx_f_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment__update_mass_with_modifications(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *__pyx_v_self); /* proto*/
static struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_get_series(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_clone(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_base_name(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_get_fragment_name(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static int __pyx_f_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment__is_glycosylated(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *__pyx_v_self); /* proto*/
static long __pyx_f_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_get_glycosylation_size(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_14SimpleFragment_clone(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_SimpleFragment *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_14SimpleFragment_get_series(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_SimpleFragment *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment__NameTree *__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree__create(void); /* proto*/
static struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment__NameTree *__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_get(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment__NameTree *__pyx_v_self, PyObject *__pyx_v_key, CYTHON_UNUSED int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_fuse_0__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_traverse(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment__NameTree *__pyx_v_self, PyObject *__pyx_v_parts, CYTHON_UNUSED int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_fuse_1__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_traverse(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment__NameTree *__pyx_v_self, PyObject *__pyx_v_parts, CYTHON_UNUSED int __pyx_skip_dispatch); /* proto*/
static struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_StubFragment *__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment__create(PyObject *__pyx_v_name, double __pyx_v_mass, struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *__pyx_v_kind, struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *__pyx_v_composition, struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase *__pyx_v_chemical_shift, int __pyx_v_is_glycosylated, PyObject *__pyx_v_glycosylation, int __pyx_v_is_extended, struct __pyx_opt_args_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment__create *__pyx_optional_args); /* proto*/
static PyObject *__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_clone(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_StubFragment *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static int __pyx_f_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_get_glycosylation_size(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_StubFragment *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_base_name(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_StubFragment *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_get_series(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_StubFragment *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/

/* Module declarations from 'cython' */

/* Module declarations from 'glypy.composition.ccomposition' */
static PyTypeObject *__pyx_ptype_5glypy_11composition_12ccomposition_CComposition = 0;
static PyObject **__pyx_vp_5glypy_11composition_12ccomposition__atom = 0;
#define __pyx_v_5glypy_11composition_12ccomposition__atom (*__pyx_vp_5glypy_11composition_12ccomposition__atom)
static PyObject **__pyx_vp_5glypy_11composition_12ccomposition__formula = 0;
#define __pyx_v_5glypy_11composition_12ccomposition__formula (*__pyx_vp_5glypy_11composition_12ccomposition__formula)
static PyObject **__pyx_vp_5glypy_11composition_12ccomposition__isotope_string = 0;
#define __pyx_v_5glypy_11composition_12ccomposition__isotope_string (*__pyx_vp_5glypy_11composition_12ccomposition__isotope_string)
static PyObject **__pyx_vp_5glypy_11composition_12ccomposition_isotope_pattern = 0;
#define __pyx_v_5glypy_11composition_12ccomposition_isotope_pattern (*__pyx_vp_5glypy_11composition_12ccomposition_isotope_pattern)
static PyObject **__pyx_vp_5glypy_11composition_12ccomposition_formula_pattern = 0;
#define __pyx_v_5glypy_11composition_12ccomposition_formula_pattern (*__pyx_vp_5glypy_11composition_12ccomposition_formula_pattern)

/* Module declarations from 'cpython.version' */

/* Module declarations from '__builtin__' */

/* Module declarations from 'cpython.type' */
static PyTypeObject *__pyx_ptype_7cpython_4type_type = 0;

/* Module declarations from 'libc.string' */

/* Module declarations from 'libc.stdio' */

/* Module declarations from 'cpython.object' */

/* Module declarations from 'cpython.ref' */

/* Module declarations from 'cpython.exc' */

/* Module declarations from 'cpython.module' */

/* Module declarations from 'cpython.mem' */

/* Module declarations from 'cpython.tuple' */

/* Module declarations from 'cpython.list' */

/* Module declarations from 'cpython.sequence' */

/* Module declarations from 'cpython.mapping' */

/* Module declarations from 'cpython.iterator' */

/* Module declarations from 'cpython.number' */

/* Module declarations from 'cpython.int' */

/* Module declarations from '__builtin__' */

/* Module declarations from 'cpython.bool' */
static PyTypeObject *__pyx_ptype_7cpython_4bool_bool = 0;

/* Module declarations from 'cpython.long' */

/* Module declarations from 'cpython.float' */

/* Module declarations from '__builtin__' */

/* Module declarations from 'cpython.complex' */
static PyTypeObject *__pyx_ptype_7cpython_7complex_complex = 0;

/* Module declarations from 'cpython.string' */

/* Module declarations from 'cpython.unicode' */

/* Module declarations from 'cpython.dict' */

/* Module declarations from 'cpython.instance' */

/* Module declarations from 'cpython.function' */

/* Module declarations from 'cpython.method' */

/* Module declarations from 'cpython.weakref' */

/* Module declarations from 'cpython.getargs' */

/* Module declarations from 'cpython.pythread' */

/* Module declarations from 'cpython.pystate' */

/* Module declarations from 'cpython.cobject' */

/* Module declarations from 'cpython.oldbuffer' */

/* Module declarations from 'cpython.set' */

/* Module declarations from 'cpython.buffer' */

/* Module declarations from 'cpython.bytes' */

/* Module declarations from 'cpython.pycapsule' */

/* Module declarations from 'cpython' */

/* Module declarations from 'glycopeptidepy._c.count_table' */
static PyTypeObject *__pyx_ptype_14glycopeptidepy_2_c_11count_table_CountTableIterator = 0;
static PyTypeObject *__pyx_ptype_14glycopeptidepy_2_c_11count_table_CountTable = 0;
static PyObject *(*__pyx_f_14glycopeptidepy_2_c_11count_table_count_table_items)(struct __pyx_t_14glycopeptidepy_2_c_11count_table_count_table *); /*proto*/

/* Module declarations from 'libc.stdlib' */

/* Module declarations from 'glycopeptidepy._c.structure.base' */
static PyTypeObject *__pyx_ptype_14glycopeptidepy_2_c_9structure_4base_PeptideSequenceBase = 0;
static PyTypeObject *__pyx_ptype_14glycopeptidepy_2_c_9structure_4base_TerminalGroup = 0;
static PyTypeObject *__pyx_ptype_14glycopeptidepy_2_c_9structure_4base_AminoAcidResidueBase = 0;
static PyTypeObject *__pyx_ptype_14glycopeptidepy_2_c_9structure_4base_ModificationBase = 0;
static PyTypeObject *__pyx_ptype_14glycopeptidepy_2_c_9structure_4base_SequencePosition = 0;

/* Module declarations from 'glycopeptidepy._c.compat' */

/* Module declarations from 'glycopeptidepy._c.structure.fragment' */
static PyTypeObject *__pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase = 0;
static PyTypeObject *__pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase = 0;
static PyTypeObject *__pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase = 0;
static PyTypeObject *__pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment_SimpleFragment = 0;
static PyTypeObject *__pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment_StubFragment = 0;
static PyTypeObject *__pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment = 0;
static PyTypeObject *__pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment__NameTree = 0;
static PyTypeObject *__pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct____pyx_fuse_0__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition = 0;
static PyTypeObject *__pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_1_genexpr = 0;
static PyTypeObject *__pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_2___pyx_fuse_1__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition = 0;
static PyTypeObject *__pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_3_genexpr = 0;
static PyTypeObject *__pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_4___pyx_fuse_2__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition = 0;
static PyTypeObject *__pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_5_genexpr = 0;
static PyTypeObject *__pyx_ptype___pyx_scope_struct____Pyx_CFunc_object____tuple___to_py = 0;
static struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_ModificationBase *__pyx_v_14glycopeptidepy_2_c_9structure_8fragment__n_glycosylation = 0;
static struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_ModificationBase *__pyx_v_14glycopeptidepy_2_c_9structure_8fragment__o_glycosylation = 0;
static struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_ModificationBase *__pyx_v_14glycopeptidepy_2_c_9structure_8fragment__gag_linker_glycosylation = 0;
static struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_ModificationBase *__pyx_v_14glycopeptidepy_2_c_9structure_8fragment__modification_hexnac = 0;
static struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_ModificationBase *__pyx_v_14glycopeptidepy_2_c_9structure_8fragment__modification_xylose = 0;
static struct __pyx_t_14glycopeptidepy_2_c_9structure_8fragment_string_cell __pyx_v_14glycopeptidepy_2_c_9structure_8fragment_str_ints[0x1000];
static PyObject *__pyx_v_14glycopeptidepy_2_c_9structure_8fragment_str_int_overflow_intern_cache = 0;
static PyObject *__pyx_v_14glycopeptidepy_2_c_9structure_8fragment_ModificationCategory_glycosylation = 0;
static struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment__NameTree *__pyx_v_14glycopeptidepy_2_c_9structure_8fragment_stub_fragment_name_cache = 0;
static void __pyx_f_14glycopeptidepy_2_c_9structure_8fragment_init_str_ints(void); /*proto*/
static int __pyx_f_14glycopeptidepy_2_c_9structure_8fragment_get_str_int(int, struct __pyx_t_14glycopeptidepy_2_c_9structure_8fragment_string_cell *); /*proto*/
static PyObject *__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_pair_mass(PyObject *); /*proto*/
static PyObject *__pyx_f_14glycopeptidepy_2_c_9structure_8fragment___pyx_unpickle_ChemicalShiftBase__set_state(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase *, PyObject *); /*proto*/
static PyObject *__pyx_f_14glycopeptidepy_2_c_9structure_8fragment___pyx_unpickle_IonSeriesBase__set_state(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *, PyObject *); /*proto*/
static PyObject *__pyx_f_14glycopeptidepy_2_c_9structure_8fragment___pyx_unpickle_FragmentBase__set_state(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *, PyObject *); /*proto*/
static PyObject *__pyx_fuse_0__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition(struct __pyx_obj_14glycopeptidepy_2_c_11count_table_CountTable *); /*proto*/
static PyObject *__pyx_fuse_1__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition(PyObject *); /*proto*/
static PyObject *__pyx_fuse_2__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition(PyObject *); /*proto*/
static PyObject *__pyx_f_14glycopeptidepy_2_c_9structure_8fragment___pyx_unpickle__NameTree__set_state(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment__NameTree *, PyObject *); /*proto*/
static PyObject *__Pyx_CFunc_object____tuple___to_py(PyObject *(*)(PyObject *)); /*proto*/
#define __Pyx_MODULE_NAME "glycopeptidepy._c.structure.fragment"
extern int __pyx_module_is_main_glycopeptidepy___c__structure__fragment;
int __pyx_module_is_main_glycopeptidepy___c__structure__fragment = 0;

/* Implementation of 'glycopeptidepy._c.structure.fragment' */
static PyObject *__pyx_builtin_property;
static PyObject *__pyx_builtin_range;
static PyObject *__pyx_builtin_NotImplementedError;
static PyObject *__pyx_builtin_AttributeError;
static PyObject *__pyx_builtin_ValueError;
static PyObject *__pyx_builtin_TypeError;
static PyObject *__pyx_builtin_sum;
static const char __pyx_k_[] = "";
static const char __pyx_k_i[] = "i";
static const char __pyx_k_s[] = "s";
static const char __pyx_k_x[] = "x";
static const char __pyx_k__3[] = "()";
static const char __pyx_k__4[] = "|";
static const char __pyx_k_Xyl[] = "Xyl";
static const char __pyx_k_end[] = "end";
static const char __pyx_k_key[] = "key";
static const char __pyx_k_new[] = "__new__";
static const char __pyx_k_out[] = "out";
static const char __pyx_k_s_d[] = "%s%d";
static const char __pyx_k_s_s[] = "%s+%s";
static const char __pyx_k_sum[] = "sum";
static const char __pyx_k_args[] = "args";
static const char __pyx_k_cell[] = "cell";
static const char __pyx_k_dict[] = "__dict__";
static const char __pyx_k_file[] = "file";
static const char __pyx_k_join[] = "join";
static const char __pyx_k_kind[] = "kind";
static const char __pyx_k_list[] = "list";
static const char __pyx_k_main[] = "__main__";
static const char __pyx_k_mass[] = "mass";
static const char __pyx_k_name[] = "__name__";
static const char __pyx_k_rule[] = "rule";
static const char __pyx_k_self[] = "self";
static const char __pyx_k_send[] = "send";
static const char __pyx_k_sort[] = "sort";
static const char __pyx_k_test[] = "__test__";
static const char __pyx_k_wrap[] = "wrap";
static const char __pyx_k_class[] = "__class__";
static const char __pyx_k_clone[] = "clone";
static const char __pyx_k_close[] = "close";
static const char __pyx_k_items[] = "items";
static const char __pyx_k_numpy[] = "numpy";
static const char __pyx_k_parts[] = "parts";
static const char __pyx_k_print[] = "print";
static const char __pyx_k_range[] = "range";
static const char __pyx_k_split[] = "split";
static const char __pyx_k_strip[] = "strip";
static const char __pyx_k_throw[] = "throw";
static const char __pyx_k_HexNAc[] = "HexNAc";
static const char __pyx_k_format[] = "format";
static const char __pyx_k_import[] = "__import__";
static const char __pyx_k_kwargs[] = "kwargs";
static const char __pyx_k_name_2[] = "name";
static const char __pyx_k_object[] = "object";
static const char __pyx_k_pickle[] = "pickle";
static const char __pyx_k_reduce[] = "__reduce__";
static const char __pyx_k_result[] = "result";
static const char __pyx_k_series[] = "series";
static const char __pyx_k_update[] = "update";
static const char __pyx_k_values[] = "values";
static const char __pyx_k_genexpr[] = "genexpr";
static const char __pyx_k_peptide[] = "peptide";
static const char __pyx_k_NameTree[] = "_NameTree";
static const char __pyx_k_children[] = "children";
static const char __pyx_k_defaults[] = "defaults";
static const char __pyx_k_getstate[] = "__getstate__";
static const char __pyx_k_position[] = "position";
static const char __pyx_k_property[] = "property";
static const char __pyx_k_pyx_type[] = "__pyx_type";
static const char __pyx_k_s_name_r[] = "%s(name=%r)";
static const char __pyx_k_setstate[] = "__setstate__";
static const char __pyx_k_traverse[] = "traverse";
static const char __pyx_k_TypeError[] = "TypeError";
static const char __pyx_k_base_name[] = "base_name";
static const char __pyx_k_pyx_state[] = "__pyx_state";
static const char __pyx_k_reduce_ex[] = "__reduce_ex__";
static const char __pyx_k_ValueError[] = "ValueError";
static const char __pyx_k_get_series[] = "get_series";
static const char __pyx_k_name_cache[] = "_name_cache";
static const char __pyx_k_pyx_result[] = "__pyx_result";
static const char __pyx_k_pyx_vtable[] = "__pyx_vtable__";
static const char __pyx_k_PickleError[] = "PickleError";
static const char __pyx_k_cfunc_to_py[] = "cfunc.to_py";
static const char __pyx_k_composition[] = "composition";
static const char __pyx_k_is_extended[] = "is_extended";
static const char __pyx_k_FragmentBase[] = "FragmentBase";
static const char __pyx_k_Modification[] = "Modification";
static const char __pyx_k_StubFragment[] = "StubFragment";
static const char __pyx_k_pyx_checksum[] = "__pyx_checksum";
static const char __pyx_k_stringsource[] = "stringsource";
static const char __pyx_k_IonSeriesBase[] = "IonSeriesBase";
static const char __pyx_k_glycosylation[] = "glycosylation";
static const char __pyx_k_reduce_cython[] = "__reduce_cython__";
static const char __pyx_k_AttributeError[] = "AttributeError";
static const char __pyx_k_SimpleFragment[] = "SimpleFragment";
static const char __pyx_k_chemical_shift[] = "chemical_shift";
static const char __pyx_k_PeptideFragment[] = "PeptideFragment";
static const char __pyx_k_get_string_cell[] = "_get_string_cell";
static const char __pyx_k_is_glycosylated[] = "is_glycosylated";
static const char __pyx_k_pyx_PickleError[] = "__pyx_PickleError";
static const char __pyx_k_setstate_cython[] = "__setstate_cython__";
static const char __pyx_k_ChemicalShiftBase[] = "ChemicalShiftBase";
static const char __pyx_k_NameTree_traverse[] = "_NameTree.traverse";
static const char __pyx_k_debug_get_str_int[] = "_debug_get_str_int";
static const char __pyx_k_get_fragment_name[] = "get_fragment_name";
static const char __pyx_k_modification_dict[] = "modification_dict";
static const char __pyx_k_cline_in_traceback[] = "cline_in_traceback";
static const char __pyx_k_get_chemical_shift[] = "get_chemical_shift";
static const char __pyx_k_pyx_fuse_0traverse[] = "__pyx_fuse_0traverse";
static const char __pyx_k_set_chemical_shift[] = "set_chemical_shift";
static const char __pyx_k_NotImplementedError[] = "NotImplementedError";
static const char __pyx_k_ModificationCategory[] = "ModificationCategory";
static const char __pyx_k_flanking_amino_acids[] = "flanking_amino_acids";
static const char __pyx_k_pyx_unpickle__NameTree[] = "__pyx_unpickle__NameTree";
static const char __pyx_k_NGlycanCoreGlycosylation[] = "NGlycanCoreGlycosylation";
static const char __pyx_k_OGlycanCoreGlycosylation[] = "OGlycanCoreGlycosylation";
static const char __pyx_k_pyx_unpickle_FragmentBase[] = "__pyx_unpickle_FragmentBase";
static const char __pyx_k_pyx_unpickle_IonSeriesBase[] = "__pyx_unpickle_IonSeriesBase";
static const char __pyx_k_Composition_data_is_missing[] = "Composition data is missing";
static const char __pyx_k_No_matching_signature_found[] = "No matching signature found";
static const char __pyx_k_build_name_from_composition[] = "build_name_from_composition";
static const char __pyx_k_NameTree___pyx_fuse_0traverse[] = "_NameTree.__pyx_fuse_0traverse";
static const char __pyx_k_NameTree___pyx_fuse_1traverse[] = "_NameTree.__pyx_fuse_1traverse";
static const char __pyx_k_Pyx_CFunc_object____tuple___to[] = "__Pyx_CFunc_object____tuple___to_py.<locals>.wrap";
static const char __pyx_k_pyx_fuse_0build_name_from_comp[] = "__pyx_fuse_0build_name_from_composition.<locals>.genexpr";
static const char __pyx_k_pyx_fuse_1build_name_from_comp[] = "__pyx_fuse_1build_name_from_composition.<locals>.genexpr";
static const char __pyx_k_pyx_fuse_2build_name_from_comp[] = "__pyx_fuse_2build_name_from_composition.<locals>.genexpr";
static const char __pyx_k_pyx_unpickle_ChemicalShiftBase[] = "__pyx_unpickle_ChemicalShiftBase";
static const char __pyx_k_self___class_____name___name_se[] = "{self.__class__.__name__}(name={self.name}, mass={self.mass:.04f}, series={self.kind})";
static const char __pyx_k_src_glycopeptidepy__c_structure[] = "src/glycopeptidepy/_c/structure/fragment.pyx";
static const char __pyx_k_Expected_at_least_d_argument_s_g[] = "Expected at least %d argument%s, got %d";
static const char __pyx_k_Function_call_with_ambiguous_arg[] = "Function call with ambiguous argument types";
static const char __pyx_k_GlycosaminoglycanLinkerGlycosyla[] = "GlycosaminoglycanLinkerGlycosylation";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0[] = "Incompatible checksums (0x%x vs (0x54d98d7, 0xbef1095, 0x80a4bd4) = (composition, mass, name))";
static const char __pyx_k_PeptideFragment_self_series_self[] = "PeptideFragment({self.series}, {self.position}, {self.mass}, {modification_dict}, {self.flanking_amino_acids}, {self.chemical_shift})";
static const char __pyx_k_glycopeptidepy__c_structure_frag[] = "glycopeptidepy._c.structure.fragment";
static const char __pyx_k_glycopeptidepy_structure_modific[] = "glycopeptidepy.structure.modification";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0_2[] = "Incompatible checksums (0x%x vs (0xf30379a, 0x05a3932, 0xc0e134f) = (_hash, composition_shift, direction, includes_peptide, int_code, mass_shift, name))";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0_3[] = "Incompatible checksums (0x%x vs (0x033a8ac, 0xb3ef693, 0xc951cb2) = (_chemical_shift, _hash, _name, mass))";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0_4[] = "Incompatible checksums (0x%x vs (0xd2a53ba, 0x1a09e23, 0x7fd03ce) = (children, name, value))";
static PyObject *__pyx_kp_s_;
static PyObject *__pyx_n_s_AttributeError;
static PyObject *__pyx_n_s_ChemicalShiftBase;
static PyObject *__pyx_kp_s_Composition_data_is_missing;
static PyObject *__pyx_kp_s_Expected_at_least_d_argument_s_g;
static PyObject *__pyx_n_s_FragmentBase;
static PyObject *__pyx_kp_s_Function_call_with_ambiguous_arg;
static PyObject *__pyx_n_s_GlycosaminoglycanLinkerGlycosyla;
static PyObject *__pyx_n_s_HexNAc;
static PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0;
static PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_2;
static PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_3;
static PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_4;
static PyObject *__pyx_n_s_IonSeriesBase;
static PyObject *__pyx_n_s_Modification;
static PyObject *__pyx_n_s_ModificationCategory;
static PyObject *__pyx_n_s_NGlycanCoreGlycosylation;
static PyObject *__pyx_n_s_NameTree;
static PyObject *__pyx_n_s_NameTree___pyx_fuse_0traverse;
static PyObject *__pyx_n_s_NameTree___pyx_fuse_1traverse;
static PyObject *__pyx_n_s_NameTree_traverse;
static PyObject *__pyx_kp_s_No_matching_signature_found;
static PyObject *__pyx_n_s_NotImplementedError;
static PyObject *__pyx_n_s_OGlycanCoreGlycosylation;
static PyObject *__pyx_n_s_PeptideFragment;
static PyObject *__pyx_kp_s_PeptideFragment_self_series_self;
static PyObject *__pyx_n_s_PickleError;
static PyObject *__pyx_n_s_Pyx_CFunc_object____tuple___to;
static PyObject *__pyx_n_s_SimpleFragment;
static PyObject *__pyx_n_s_StubFragment;
static PyObject *__pyx_n_s_TypeError;
static PyObject *__pyx_n_s_ValueError;
static PyObject *__pyx_n_s_Xyl;
static PyObject *__pyx_kp_s__3;
static PyObject *__pyx_kp_s__4;
static PyObject *__pyx_n_s_args;
static PyObject *__pyx_n_s_base_name;
static PyObject *__pyx_n_s_build_name_from_composition;
static PyObject *__pyx_n_s_cell;
static PyObject *__pyx_n_s_cfunc_to_py;
static PyObject *__pyx_n_s_chemical_shift;
static PyObject *__pyx_n_s_children;
static PyObject *__pyx_n_s_class;
static PyObject *__pyx_n_s_cline_in_traceback;
static PyObject *__pyx_n_s_clone;
static PyObject *__pyx_n_s_close;
static PyObject *__pyx_n_s_composition;
static PyObject *__pyx_n_s_debug_get_str_int;
static PyObject *__pyx_n_s_defaults;
static PyObject *__pyx_n_s_dict;
static PyObject *__pyx_n_s_end;
static PyObject *__pyx_n_s_file;
static PyObject *__pyx_n_s_flanking_amino_acids;
static PyObject *__pyx_n_s_format;
static PyObject *__pyx_n_s_genexpr;
static PyObject *__pyx_n_s_get_chemical_shift;
static PyObject *__pyx_n_s_get_fragment_name;
static PyObject *__pyx_n_s_get_series;
static PyObject *__pyx_n_s_get_string_cell;
static PyObject *__pyx_n_s_getstate;
static PyObject *__pyx_n_s_glycopeptidepy__c_structure_frag;
static PyObject *__pyx_n_s_glycopeptidepy_structure_modific;
static PyObject *__pyx_n_s_glycosylation;
static PyObject *__pyx_n_s_i;
static PyObject *__pyx_n_s_import;
static PyObject *__pyx_n_s_is_extended;
static PyObject *__pyx_n_s_is_glycosylated;
static PyObject *__pyx_n_s_items;
static PyObject *__pyx_n_s_join;
static PyObject *__pyx_n_s_key;
static PyObject *__pyx_n_s_kind;
static PyObject *__pyx_n_s_kwargs;
static PyObject *__pyx_n_s_list;
static PyObject *__pyx_n_s_main;
static PyObject *__pyx_n_s_mass;
static PyObject *__pyx_n_s_modification_dict;
static PyObject *__pyx_n_s_name;
static PyObject *__pyx_n_s_name_2;
static PyObject *__pyx_n_s_name_cache;
static PyObject *__pyx_n_s_new;
static PyObject *__pyx_n_s_numpy;
static PyObject *__pyx_n_s_object;
static PyObject *__pyx_n_s_out;
static PyObject *__pyx_n_s_parts;
static PyObject *__pyx_n_s_peptide;
static PyObject *__pyx_n_s_pickle;
static PyObject *__pyx_n_s_position;
static PyObject *__pyx_n_s_print;
static PyObject *__pyx_n_s_property;
static PyObject *__pyx_n_s_pyx_PickleError;
static PyObject *__pyx_n_s_pyx_checksum;
static PyObject *__pyx_n_s_pyx_fuse_0build_name_from_comp;
static PyObject *__pyx_n_s_pyx_fuse_0traverse;
static PyObject *__pyx_n_s_pyx_fuse_1build_name_from_comp;
static PyObject *__pyx_n_s_pyx_fuse_2build_name_from_comp;
static PyObject *__pyx_n_s_pyx_result;
static PyObject *__pyx_n_s_pyx_state;
static PyObject *__pyx_n_s_pyx_type;
static PyObject *__pyx_n_s_pyx_unpickle_ChemicalShiftBase;
static PyObject *__pyx_n_s_pyx_unpickle_FragmentBase;
static PyObject *__pyx_n_s_pyx_unpickle_IonSeriesBase;
static PyObject *__pyx_n_s_pyx_unpickle__NameTree;
static PyObject *__pyx_n_s_pyx_vtable;
static PyObject *__pyx_n_s_range;
static PyObject *__pyx_n_s_reduce;
static PyObject *__pyx_n_s_reduce_cython;
static PyObject *__pyx_n_s_reduce_ex;
static PyObject *__pyx_n_s_result;
static PyObject *__pyx_n_s_rule;
static PyObject *__pyx_n_s_s;
static PyObject *__pyx_kp_s_s_d;
static PyObject *__pyx_kp_s_s_name_r;
static PyObject *__pyx_kp_s_s_s;
static PyObject *__pyx_n_s_self;
static PyObject *__pyx_kp_s_self___class_____name___name_se;
static PyObject *__pyx_n_s_send;
static PyObject *__pyx_n_s_series;
static PyObject *__pyx_n_s_set_chemical_shift;
static PyObject *__pyx_n_s_setstate;
static PyObject *__pyx_n_s_setstate_cython;
static PyObject *__pyx_n_s_sort;
static PyObject *__pyx_n_s_split;
static PyObject *__pyx_kp_s_src_glycopeptidepy__c_structure;
static PyObject *__pyx_kp_s_stringsource;
static PyObject *__pyx_n_s_strip;
static PyObject *__pyx_n_s_sum;
static PyObject *__pyx_n_s_test;
static PyObject *__pyx_n_s_throw;
static PyObject *__pyx_n_s_traverse;
static PyObject *__pyx_n_s_update;
static PyObject *__pyx_n_s_values;
static PyObject *__pyx_n_s_wrap;
static PyObject *__pyx_n_s_x;
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment__debug_get_str_int(CYTHON_UNUSED PyObject *__pyx_self, int __pyx_v_i); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_2_get_string_cell(CYTHON_UNUSED PyObject *__pyx_self, int __pyx_v_i); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_17ChemicalShiftBase_clone(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_17ChemicalShiftBase_2__str__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_17ChemicalShiftBase_4__repr__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_17ChemicalShiftBase_6is_loss(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_17ChemicalShiftBase_4name___get__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase *__pyx_v_self); /* proto */
static int __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_17ChemicalShiftBase_4name_2__set__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_17ChemicalShiftBase_4name_4__del__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_17ChemicalShiftBase_11composition___get__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase *__pyx_v_self); /* proto */
static int __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_17ChemicalShiftBase_11composition_2__set__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_17ChemicalShiftBase_11composition_4__del__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_17ChemicalShiftBase_4mass___get__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase *__pyx_v_self); /* proto */
static int __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_17ChemicalShiftBase_4mass_2__set__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_17ChemicalShiftBase_8__reduce_cython__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_17ChemicalShiftBase_10__setstate_cython__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase___eq__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *__pyx_v_self, PyObject *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_2__ne__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *__pyx_v_self, PyObject *__pyx_v_other); /* proto */
static Py_hash_t __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_4__hash__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_4name___get__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *__pyx_v_self); /* proto */
static int __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_4name_2__set__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_4name_4__del__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_9direction___get__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *__pyx_v_self); /* proto */
static int __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_9direction_2__set__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_16includes_peptide___get__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *__pyx_v_self); /* proto */
static int __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_16includes_peptide_2__set__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_10mass_shift___get__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *__pyx_v_self); /* proto */
static int __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_10mass_shift_2__set__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_17composition_shift___get__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *__pyx_v_self); /* proto */
static int __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_17composition_shift_2__set__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_17composition_shift_4__del__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_5_hash___get__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *__pyx_v_self); /* proto */
static int __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_5_hash_2__set__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_8int_code___get__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *__pyx_v_self); /* proto */
static int __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_8int_code_2__set__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_6__reduce_cython__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_8__setstate_cython__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_get_series(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *__pyx_v_self); /* proto */
static Py_hash_t __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_2__hash__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_4__eq__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *__pyx_v_self, PyObject *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_6__ne__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *__pyx_v_self, PyObject *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_6series___get__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_8clone(CYTHON_UNUSED struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_10get_chemical_shift(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_12set_chemical_shift(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *__pyx_v_self, struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase *__pyx_v_chemical_shift); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_14get_fragment_name(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_16base_name(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_4name___get__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *__pyx_v_self); /* proto */
static int __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_4name_2__set__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *__pyx_v_self, PyObject *__pyx_v_name); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_5_name___get__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *__pyx_v_self); /* proto */
static int __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_5_name_2__set__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_5_name_4__del__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_5_hash___get__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *__pyx_v_self); /* proto */
static int __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_5_hash_2__set__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_4mass___get__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *__pyx_v_self); /* proto */
static int __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_4mass_2__set__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_15_chemical_shift___get__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *__pyx_v_self); /* proto */
static int __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_15_chemical_shift_2__set__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_15_chemical_shift_4__del__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_18__reduce_cython__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_20__setstate_cython__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment___init__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *__pyx_v_self, PyObject *__pyx_v_kind, PyObject *__pyx_v_position, PyObject *__pyx_v_modification_dict, PyObject *__pyx_v_mass, PyObject *__pyx_v_flanking_amino_acids, PyObject *__pyx_v_glycosylation, PyObject *__pyx_v_chemical_shift, PyObject *__pyx_v_composition); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_2get_series(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_4clone(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_6total_composition(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_8__reduce__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_10base_name(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_12get_fragment_name(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_15is_glycosylated___get__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_18glycosylation_size___get__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_14__repr__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_4kind___get__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *__pyx_v_self); /* proto */
static int __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_4kind_2__set__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_4kind_4__del__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_8position___get__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *__pyx_v_self); /* proto */
static int __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_8position_2__set__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_17modification_dict___get__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *__pyx_v_self); /* proto */
static int __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_17modification_dict_2__set__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_17modification_dict_4__del__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_9bare_mass___get__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *__pyx_v_self); /* proto */
static int __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_9bare_mass_2__set__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_20flanking_amino_acids___get__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *__pyx_v_self); /* proto */
static int __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_20flanking_amino_acids_2__set__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_20flanking_amino_acids_4__del__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_13glycosylation___get__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *__pyx_v_self); /* proto */
static int __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_13glycosylation_2__set__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_13glycosylation_4__del__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_11composition___get__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *__pyx_v_self); /* proto */
static int __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_11composition_2__set__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_11composition_4__del__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *__pyx_v_self); /* proto */
static int __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_14SimpleFragment___init__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_SimpleFragment *__pyx_v_self, PyObject *__pyx_v_name, PyObject *__pyx_v_mass, PyObject *__pyx_v_kind, PyObject *__pyx_v_composition, PyObject *__pyx_v_chemical_shift, PyObject *__pyx_v_is_glycosylated); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_14SimpleFragment_2clone(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_SimpleFragment *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_14SimpleFragment_4__reduce__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_SimpleFragment *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_14SimpleFragment_6__repr__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_SimpleFragment *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_14SimpleFragment_8get_series(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_SimpleFragment *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_14SimpleFragment_4kind___get__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_SimpleFragment *__pyx_v_self); /* proto */
static int __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_14SimpleFragment_4kind_2__set__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_SimpleFragment *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_14SimpleFragment_4kind_4__del__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_SimpleFragment *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_14SimpleFragment_15is_glycosylated___get__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_SimpleFragment *__pyx_v_self); /* proto */
static int __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_14SimpleFragment_15is_glycosylated_2__set__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_SimpleFragment *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_14SimpleFragment_11composition___get__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_SimpleFragment *__pyx_v_self); /* proto */
static int __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_14SimpleFragment_11composition_2__set__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_SimpleFragment *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_14SimpleFragment_11composition_4__del__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_SimpleFragment *__pyx_v_self); /* proto */
static int __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree___init__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment__NameTree *__pyx_v_self, PyObject *__pyx_v_name, PyObject *__pyx_v_children); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_2__getitem__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment__NameTree *__pyx_v_self, PyObject *__pyx_v_key); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_4get(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment__NameTree *__pyx_v_self, PyObject *__pyx_v_key); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_6traverse(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment__NameTree *__pyx_v_signatures, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs, CYTHON_UNUSED PyObject *__pyx_v_defaults); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_8__pyx_fuse_0traverse(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment__NameTree *__pyx_v_self, PyObject *__pyx_v_parts); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_10__pyx_fuse_1traverse(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment__NameTree *__pyx_v_self, PyObject *__pyx_v_parts); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_4name___get__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment__NameTree *__pyx_v_self); /* proto */
static int __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_4name_2__set__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment__NameTree *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_4name_4__del__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment__NameTree *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_5value___get__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment__NameTree *__pyx_v_self); /* proto */
static int __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_5value_2__set__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment__NameTree *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_5value_4__del__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment__NameTree *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_8children___get__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment__NameTree *__pyx_v_self); /* proto */
static int __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_8children_2__set__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment__NameTree *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_8children_4__del__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment__NameTree *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_14__reduce_cython__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment__NameTree *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_16__setstate_cython__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment__NameTree *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_39__pyx_fuse_0build_name_from_composition_genexpr(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_39__pyx_fuse_1build_name_from_composition_genexpr(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_39__pyx_fuse_2build_name_from_composition_genexpr(PyObject *__pyx_self); /* proto */
static int __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment___init__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_StubFragment *__pyx_v_self, PyObject *__pyx_v_name, PyObject *__pyx_v_mass, PyObject *__pyx_v_kind, PyObject *__pyx_v_composition, PyObject *__pyx_v_chemical_shift, PyObject *__pyx_v_is_glycosylated, PyObject *__pyx_v_glycosylation, PyObject *__pyx_v_is_extended); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_2clone(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_StubFragment *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_18glycosylation_size___get__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_StubFragment *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_4base_name(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_StubFragment *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_6build_name_from_composition(CYTHON_UNUSED PyTypeObject *__pyx_v_cls, PyObject *__pyx_v_glycan_composition); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_8__reduce__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_StubFragment *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_10__repr__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_StubFragment *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_12get_series(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_StubFragment *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_4kind___get__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_StubFragment *__pyx_v_self); /* proto */
static int __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_4kind_2__set__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_StubFragment *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_4kind_4__del__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_StubFragment *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_15is_glycosylated___get__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_StubFragment *__pyx_v_self); /* proto */
static int __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_15is_glycosylated_2__set__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_StubFragment *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_11composition___get__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_StubFragment *__pyx_v_self); /* proto */
static int __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_11composition_2__set__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_StubFragment *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_11composition_4__del__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_StubFragment *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_13glycosylation___get__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_StubFragment *__pyx_v_self); /* proto */
static int __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_13glycosylation_2__set__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_StubFragment *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_13glycosylation_4__del__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_StubFragment *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_11is_extended___get__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_StubFragment *__pyx_v_self); /* proto */
static int __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_11is_extended_2__set__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_StubFragment *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_4__pyx_unpickle_ChemicalShiftBase(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_6__pyx_unpickle_IonSeriesBase(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_8__pyx_unpickle_FragmentBase(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_10__pyx_unpickle__NameTree(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_11cfunc_dot_to_py_35__Pyx_CFunc_object____tuple___to_py_wrap(PyObject *__pyx_self, PyObject *__pyx_v_x); /* proto */
static PyObject *__pyx_tp_new_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_14glycopeptidepy_2_c_9structure_8fragment_SimpleFragment(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_14glycopeptidepy_2_c_9structure_8fragment_StubFragment(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_14glycopeptidepy_2_c_9structure_8fragment__NameTree(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct____pyx_fuse_0__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_1_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_2___pyx_fuse_1__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_3_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_4___pyx_fuse_2__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_5_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new___pyx_scope_struct____Pyx_CFunc_object____tuple___to_py(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_float_1eneg_5;
static PyObject *__pyx_int_0;
static PyObject *__pyx_int_2;
static PyObject *__pyx_int_3385516;
static PyObject *__pyx_int_5912882;
static PyObject *__pyx_int_27303459;
static PyObject *__pyx_int_88971479;
static PyObject *__pyx_int_134022094;
static PyObject *__pyx_int_134892500;
static PyObject *__pyx_int_188675731;
static PyObject *__pyx_int_200216725;
static PyObject *__pyx_int_202249039;
static PyObject *__pyx_int_211098802;
static PyObject *__pyx_int_220877754;
static PyObject *__pyx_int_254818202;
static PyObject *__pyx_tuple__2;
static PyObject *__pyx_tuple__5;
static PyObject *__pyx_tuple__6;
static PyObject *__pyx_tuple__7;
static PyObject *__pyx_tuple__8;
static PyObject *__pyx_tuple__9;
static PyObject *__pyx_tuple__10;
static PyObject *__pyx_tuple__11;
static PyObject *__pyx_tuple__13;
static PyObject *__pyx_tuple__14;
static PyObject *__pyx_tuple__15;
static PyObject *__pyx_tuple__17;
static PyObject *__pyx_tuple__19;
static PyObject *__pyx_tuple__21;
static PyObject *__pyx_tuple__23;
static PyObject *__pyx_tuple__25;
static PyObject *__pyx_tuple__27;
static PyObject *__pyx_codeobj__12;
static PyObject *__pyx_codeobj__16;
static PyObject *__pyx_codeobj__18;
static PyObject *__pyx_codeobj__20;
static PyObject *__pyx_codeobj__22;
static PyObject *__pyx_codeobj__24;
static PyObject *__pyx_codeobj__26;
static PyObject *__pyx_codeobj__28;
/* Late includes */

/* "glycopeptidepy/_c/structure/fragment.pyx":47
 * cdef dict str_int_overflow_intern_cache = {}
 * 
 * cdef void init_str_ints():             # <<<<<<<<<<<<<<
 *     cdef int i
 *     cdef Py_ssize_t z
 */

static void __pyx_f_14glycopeptidepy_2_c_9structure_8fragment_init_str_ints(void) {
  int __pyx_v_i;
  Py_ssize_t __pyx_v_z;
  PyObject *__pyx_v_pa = 0;
  char *__pyx_v_a;
  char *__pyx_v_atemp;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("init_str_ints", 0);

  /* "glycopeptidepy/_c/structure/fragment.pyx":54
 *     cdef char* atemp
 * 
 *     for i in range(ARRAY_SIZE):             # <<<<<<<<<<<<<<
 *         pa = str(i)
 *         atemp = PyStr_AsUTF8AndSize(pa, &z)
 */
  for (__pyx_t_1 = 0; __pyx_t_1 < 0x1000; __pyx_t_1+=1) {
    __pyx_v_i = __pyx_t_1;

    /* "glycopeptidepy/_c/structure/fragment.pyx":55
 * 
 *     for i in range(ARRAY_SIZE):
 *         pa = str(i)             # <<<<<<<<<<<<<<
 *         atemp = PyStr_AsUTF8AndSize(pa, &z)
 *         a = <char*>malloc(sizeof(char) * z)
 */
    __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_i); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 55, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyString_Type)), __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 55, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (!(likely(PyString_CheckExact(__pyx_t_3))||((__pyx_t_3) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "str", Py_TYPE(__pyx_t_3)->tp_name), 0))) __PYX_ERR(0, 55, __pyx_L1_error)
    __Pyx_XDECREF_SET(__pyx_v_pa, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "glycopeptidepy/_c/structure/fragment.pyx":56
 *     for i in range(ARRAY_SIZE):
 *         pa = str(i)
 *         atemp = PyStr_AsUTF8AndSize(pa, &z)             # <<<<<<<<<<<<<<
 *         a = <char*>malloc(sizeof(char) * z)
 *         strcpy(a, atemp)
 */
    __pyx_v_atemp = PyStr_AsUTF8AndSize(__pyx_v_pa, (&__pyx_v_z));

    /* "glycopeptidepy/_c/structure/fragment.pyx":57
 *         pa = str(i)
 *         atemp = PyStr_AsUTF8AndSize(pa, &z)
 *         a = <char*>malloc(sizeof(char) * z)             # <<<<<<<<<<<<<<
 *         strcpy(a, atemp)
 *         a[z] = "\0"
 */
    __pyx_v_a = ((char *)malloc(((sizeof(char)) * __pyx_v_z)));

    /* "glycopeptidepy/_c/structure/fragment.pyx":58
 *         atemp = PyStr_AsUTF8AndSize(pa, &z)
 *         a = <char*>malloc(sizeof(char) * z)
 *         strcpy(a, atemp)             # <<<<<<<<<<<<<<
 *         a[z] = "\0"
 *         str_ints[i].string = a
 */
    (void)(strcpy(__pyx_v_a, __pyx_v_atemp));

    /* "glycopeptidepy/_c/structure/fragment.pyx":59
 *         a = <char*>malloc(sizeof(char) * z)
 *         strcpy(a, atemp)
 *         a[z] = "\0"             # <<<<<<<<<<<<<<
 *         str_ints[i].string = a
 *         str_ints[i].size = z
 */
    (__pyx_v_a[__pyx_v_z]) = '\x00';

    /* "glycopeptidepy/_c/structure/fragment.pyx":60
 *         strcpy(a, atemp)
 *         a[z] = "\0"
 *         str_ints[i].string = a             # <<<<<<<<<<<<<<
 *         str_ints[i].size = z
 * 
 */
    (__pyx_v_14glycopeptidepy_2_c_9structure_8fragment_str_ints[__pyx_v_i]).string = __pyx_v_a;

    /* "glycopeptidepy/_c/structure/fragment.pyx":61
 *         a[z] = "\0"
 *         str_ints[i].string = a
 *         str_ints[i].size = z             # <<<<<<<<<<<<<<
 * 
 * 
 */
    (__pyx_v_14glycopeptidepy_2_c_9structure_8fragment_str_ints[__pyx_v_i]).size = __pyx_v_z;
  }

  /* "glycopeptidepy/_c/structure/fragment.pyx":47
 * cdef dict str_int_overflow_intern_cache = {}
 * 
 * cdef void init_str_ints():             # <<<<<<<<<<<<<<
 *     cdef int i
 *     cdef Py_ssize_t z
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_WriteUnraisable("glycopeptidepy._c.structure.fragment.init_str_ints", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_pa);
  __Pyx_RefNannyFinishContext();
}

/* "glycopeptidepy/_c/structure/fragment.pyx":67
 * 
 * 
 * cdef int get_str_int(int i, string_cell* out):             # <<<<<<<<<<<<<<
 *     '''Get a `string_cell` for a given integer.
 * 
 */

static int __pyx_f_14glycopeptidepy_2_c_9structure_8fragment_get_str_int(int __pyx_v_i, struct __pyx_t_14glycopeptidepy_2_c_9structure_8fragment_string_cell *__pyx_v_out) {
  PyObject *__pyx_v_ptemp;
  PyObject *__pyx_v_py_i = 0;
  PyObject *__pyx_v_pa = 0;
  Py_ssize_t __pyx_v_z;
  char *__pyx_v_atemp;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_str_int", 0);

  /* "glycopeptidepy/_c/structure/fragment.pyx":82
 *         str pa
 *         Py_ssize_t z
 *     if i < ARRAY_SIZE:             # <<<<<<<<<<<<<<
 *         out[0] = str_ints[i]
 *         return 0
 */
  __pyx_t_1 = ((__pyx_v_i < 0x1000) != 0);
  if (__pyx_t_1) {

    /* "glycopeptidepy/_c/structure/fragment.pyx":83
 *         Py_ssize_t z
 *     if i < ARRAY_SIZE:
 *         out[0] = str_ints[i]             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
    (__pyx_v_out[0]) = (__pyx_v_14glycopeptidepy_2_c_9structure_8fragment_str_ints[__pyx_v_i]);

    /* "glycopeptidepy/_c/structure/fragment.pyx":84
 *     if i < ARRAY_SIZE:
 *         out[0] = str_ints[i]
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     py_i = PyInt_FromLong(i)
 */
    __pyx_r = 0;
    goto __pyx_L0;

    /* "glycopeptidepy/_c/structure/fragment.pyx":82
 *         str pa
 *         Py_ssize_t z
 *     if i < ARRAY_SIZE:             # <<<<<<<<<<<<<<
 *         out[0] = str_ints[i]
 *         return 0
 */
  }

  /* "glycopeptidepy/_c/structure/fragment.pyx":86
 *         return 0
 * 
 *     py_i = PyInt_FromLong(i)             # <<<<<<<<<<<<<<
 *     ptemp = PyDict_GetItem(str_int_overflow_intern_cache, py_i)
 *     if ptemp == NULL:
 */
  __pyx_t_2 = PyInt_FromLong(__pyx_v_i); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 86, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_py_i = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "glycopeptidepy/_c/structure/fragment.pyx":87
 * 
 *     py_i = PyInt_FromLong(i)
 *     ptemp = PyDict_GetItem(str_int_overflow_intern_cache, py_i)             # <<<<<<<<<<<<<<
 *     if ptemp == NULL:
 *         pa = str(py_i)
 */
  __pyx_t_2 = __pyx_v_14glycopeptidepy_2_c_9structure_8fragment_str_int_overflow_intern_cache;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_v_ptemp = PyDict_GetItem(__pyx_t_2, __pyx_v_py_i);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "glycopeptidepy/_c/structure/fragment.pyx":88
 *     py_i = PyInt_FromLong(i)
 *     ptemp = PyDict_GetItem(str_int_overflow_intern_cache, py_i)
 *     if ptemp == NULL:             # <<<<<<<<<<<<<<
 *         pa = str(py_i)
 *         Py_INCREF(pa)
 */
  __pyx_t_1 = ((__pyx_v_ptemp == NULL) != 0);
  if (__pyx_t_1) {

    /* "glycopeptidepy/_c/structure/fragment.pyx":89
 *     ptemp = PyDict_GetItem(str_int_overflow_intern_cache, py_i)
 *     if ptemp == NULL:
 *         pa = str(py_i)             # <<<<<<<<<<<<<<
 *         Py_INCREF(pa)
 *         PyDict_SetItem(str_int_overflow_intern_cache, py_i, pa)
 */
    __pyx_t_2 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyString_Type)), __pyx_v_py_i); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 89, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (!(likely(PyString_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "str", Py_TYPE(__pyx_t_2)->tp_name), 0))) __PYX_ERR(0, 89, __pyx_L1_error)
    __pyx_v_pa = ((PyObject*)__pyx_t_2);
    __pyx_t_2 = 0;

    /* "glycopeptidepy/_c/structure/fragment.pyx":90
 *     if ptemp == NULL:
 *         pa = str(py_i)
 *         Py_INCREF(pa)             # <<<<<<<<<<<<<<
 *         PyDict_SetItem(str_int_overflow_intern_cache, py_i, pa)
 *     else:
 */
    Py_INCREF(__pyx_v_pa);

    /* "glycopeptidepy/_c/structure/fragment.pyx":91
 *         pa = str(py_i)
 *         Py_INCREF(pa)
 *         PyDict_SetItem(str_int_overflow_intern_cache, py_i, pa)             # <<<<<<<<<<<<<<
 *     else:
 *         pa = <str>ptemp
 */
    __pyx_t_2 = __pyx_v_14glycopeptidepy_2_c_9structure_8fragment_str_int_overflow_intern_cache;
    __Pyx_INCREF(__pyx_t_2);
    __pyx_t_3 = PyDict_SetItem(__pyx_t_2, __pyx_v_py_i, __pyx_v_pa); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(0, 91, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "glycopeptidepy/_c/structure/fragment.pyx":88
 *     py_i = PyInt_FromLong(i)
 *     ptemp = PyDict_GetItem(str_int_overflow_intern_cache, py_i)
 *     if ptemp == NULL:             # <<<<<<<<<<<<<<
 *         pa = str(py_i)
 *         Py_INCREF(pa)
 */
    goto __pyx_L4;
  }

  /* "glycopeptidepy/_c/structure/fragment.pyx":93
 *         PyDict_SetItem(str_int_overflow_intern_cache, py_i, pa)
 *     else:
 *         pa = <str>ptemp             # <<<<<<<<<<<<<<
 *     atemp = PyStr_AsUTF8AndSize(pa, &z)
 *     out.string = atemp
 */
  /*else*/ {
    __pyx_t_2 = ((PyObject *)__pyx_v_ptemp);
    __Pyx_INCREF(__pyx_t_2);
    __pyx_v_pa = ((PyObject*)__pyx_t_2);
    __pyx_t_2 = 0;
  }
  __pyx_L4:;

  /* "glycopeptidepy/_c/structure/fragment.pyx":94
 *     else:
 *         pa = <str>ptemp
 *     atemp = PyStr_AsUTF8AndSize(pa, &z)             # <<<<<<<<<<<<<<
 *     out.string = atemp
 *     out.size = z
 */
  __pyx_v_atemp = PyStr_AsUTF8AndSize(__pyx_v_pa, (&__pyx_v_z));

  /* "glycopeptidepy/_c/structure/fragment.pyx":95
 *         pa = <str>ptemp
 *     atemp = PyStr_AsUTF8AndSize(pa, &z)
 *     out.string = atemp             # <<<<<<<<<<<<<<
 *     out.size = z
 *     return 0
 */
  __pyx_v_out->string = __pyx_v_atemp;

  /* "glycopeptidepy/_c/structure/fragment.pyx":96
 *     atemp = PyStr_AsUTF8AndSize(pa, &z)
 *     out.string = atemp
 *     out.size = z             # <<<<<<<<<<<<<<
 *     return 0
 * 
 */
  __pyx_v_out->size = __pyx_v_z;

  /* "glycopeptidepy/_c/structure/fragment.pyx":97
 *     out.string = atemp
 *     out.size = z
 *     return 0             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "glycopeptidepy/_c/structure/fragment.pyx":67
 * 
 * 
 * cdef int get_str_int(int i, string_cell* out):             # <<<<<<<<<<<<<<
 *     '''Get a `string_cell` for a given integer.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_WriteUnraisable("glycopeptidepy._c.structure.fragment.get_str_int", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_py_i);
  __Pyx_XDECREF(__pyx_v_pa);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/fragment.pyx":100
 * 
 * 
 * def _debug_get_str_int(int i):             # <<<<<<<<<<<<<<
 *     cdef:
 *         string_cell out
 */

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_1_debug_get_str_int(PyObject *__pyx_self, PyObject *__pyx_arg_i); /*proto*/
static char __pyx_doc_14glycopeptidepy_2_c_9structure_8fragment__debug_get_str_int[] = "_debug_get_str_int(int i)";
static PyMethodDef __pyx_mdef_14glycopeptidepy_2_c_9structure_8fragment_1_debug_get_str_int = {"_debug_get_str_int", (PyCFunction)__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_1_debug_get_str_int, METH_O, __pyx_doc_14glycopeptidepy_2_c_9structure_8fragment__debug_get_str_int};
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_1_debug_get_str_int(PyObject *__pyx_self, PyObject *__pyx_arg_i) {
  int __pyx_v_i;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_debug_get_str_int (wrapper)", 0);
  assert(__pyx_arg_i); {
    __pyx_v_i = __Pyx_PyInt_As_int(__pyx_arg_i); if (unlikely((__pyx_v_i == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 100, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment._debug_get_str_int", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment__debug_get_str_int(__pyx_self, ((int)__pyx_v_i));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment__debug_get_str_int(CYTHON_UNUSED PyObject *__pyx_self, int __pyx_v_i) {
  struct __pyx_t_14glycopeptidepy_2_c_9structure_8fragment_string_cell __pyx_v_out;
  int __pyx_v_result;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_debug_get_str_int", 0);

  /* "glycopeptidepy/_c/structure/fragment.pyx":104
 *         string_cell out
 *         int result
 *     result = get_str_int(i, &out)             # <<<<<<<<<<<<<<
 *     print(result)
 *     print(out.string)
 */
  __pyx_v_result = __pyx_f_14glycopeptidepy_2_c_9structure_8fragment_get_str_int(__pyx_v_i, (&__pyx_v_out));

  /* "glycopeptidepy/_c/structure/fragment.pyx":105
 *         int result
 *     result = get_str_int(i, &out)
 *     print(result)             # <<<<<<<<<<<<<<
 *     print(out.string)
 *     print(out.size)
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_result); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 105, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_PrintOne(0, __pyx_t_1) < 0) __PYX_ERR(0, 105, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "glycopeptidepy/_c/structure/fragment.pyx":106
 *     result = get_str_int(i, &out)
 *     print(result)
 *     print(out.string)             # <<<<<<<<<<<<<<
 *     print(out.size)
 * 
 */
  __pyx_t_1 = __Pyx_PyBytes_FromString(__pyx_v_out.string); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 106, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_PrintOne(0, __pyx_t_1) < 0) __PYX_ERR(0, 106, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "glycopeptidepy/_c/structure/fragment.pyx":107
 *     print(result)
 *     print(out.string)
 *     print(out.size)             # <<<<<<<<<<<<<<
 * 
 * def _get_string_cell(int i):
 */
  __pyx_t_1 = __Pyx_PyInt_FromSize_t(__pyx_v_out.size); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 107, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_PrintOne(0, __pyx_t_1) < 0) __PYX_ERR(0, 107, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "glycopeptidepy/_c/structure/fragment.pyx":100
 * 
 * 
 * def _debug_get_str_int(int i):             # <<<<<<<<<<<<<<
 *     cdef:
 *         string_cell out
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment._debug_get_str_int", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/fragment.pyx":109
 *     print(out.size)
 * 
 * def _get_string_cell(int i):             # <<<<<<<<<<<<<<
 *     cdef string_cell cell = str_ints[i]
 *     print(cell.size)
 */

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_3_get_string_cell(PyObject *__pyx_self, PyObject *__pyx_arg_i); /*proto*/
static char __pyx_doc_14glycopeptidepy_2_c_9structure_8fragment_2_get_string_cell[] = "_get_string_cell(int i)";
static PyMethodDef __pyx_mdef_14glycopeptidepy_2_c_9structure_8fragment_3_get_string_cell = {"_get_string_cell", (PyCFunction)__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_3_get_string_cell, METH_O, __pyx_doc_14glycopeptidepy_2_c_9structure_8fragment_2_get_string_cell};
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_3_get_string_cell(PyObject *__pyx_self, PyObject *__pyx_arg_i) {
  int __pyx_v_i;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_get_string_cell (wrapper)", 0);
  assert(__pyx_arg_i); {
    __pyx_v_i = __Pyx_PyInt_As_int(__pyx_arg_i); if (unlikely((__pyx_v_i == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 109, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment._get_string_cell", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_2_get_string_cell(__pyx_self, ((int)__pyx_v_i));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_2_get_string_cell(CYTHON_UNUSED PyObject *__pyx_self, int __pyx_v_i) {
  struct __pyx_t_14glycopeptidepy_2_c_9structure_8fragment_string_cell __pyx_v_cell;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_get_string_cell", 0);

  /* "glycopeptidepy/_c/structure/fragment.pyx":110
 * 
 * def _get_string_cell(int i):
 *     cdef string_cell cell = str_ints[i]             # <<<<<<<<<<<<<<
 *     print(cell.size)
 *     print(cell.string)
 */
  __pyx_v_cell = (__pyx_v_14glycopeptidepy_2_c_9structure_8fragment_str_ints[__pyx_v_i]);

  /* "glycopeptidepy/_c/structure/fragment.pyx":111
 * def _get_string_cell(int i):
 *     cdef string_cell cell = str_ints[i]
 *     print(cell.size)             # <<<<<<<<<<<<<<
 *     print(cell.string)
 * 
 */
  __pyx_t_1 = __Pyx_PyInt_FromSize_t(__pyx_v_cell.size); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 111, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_PrintOne(0, __pyx_t_1) < 0) __PYX_ERR(0, 111, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "glycopeptidepy/_c/structure/fragment.pyx":112
 *     cdef string_cell cell = str_ints[i]
 *     print(cell.size)
 *     print(cell.string)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __Pyx_PyBytes_FromString(__pyx_v_cell.string); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 112, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_PrintOne(0, __pyx_t_1) < 0) __PYX_ERR(0, 112, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "glycopeptidepy/_c/structure/fragment.pyx":109
 *     print(out.size)
 * 
 * def _get_string_cell(int i):             # <<<<<<<<<<<<<<
 *     cdef string_cell cell = str_ints[i]
 *     print(cell.size)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment._get_string_cell", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/fragment.pyx":118
 * cdef class ChemicalShiftBase(object):
 * 
 *     cpdef ChemicalShiftBase clone(self):             # <<<<<<<<<<<<<<
 *         return self.__class__(self.name, self.composition.clone())
 * 
 */

static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_17ChemicalShiftBase_1clone(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase *__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_17ChemicalShiftBase_clone(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase *__pyx_v_self, int __pyx_skip_dispatch) {
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("clone", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_clone); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 118, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_17ChemicalShiftBase_1clone)) {
        __Pyx_XDECREF(((PyObject *)__pyx_r));
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 118, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase))))) __PYX_ERR(0, 118, __pyx_L1_error)
        __pyx_r = ((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase *)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "glycopeptidepy/_c/structure/fragment.pyx":119
 * 
 *     cpdef ChemicalShiftBase clone(self):
 *         return self.__class__(self.name, self.composition.clone())             # <<<<<<<<<<<<<<
 * 
 *     def __str__(self):
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_class); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 119, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = ((PyObject *)((struct __pyx_vtabstruct_5glypy_11composition_12ccomposition_CComposition *)__pyx_v_self->composition->__pyx_vtab)->clone(__pyx_v_self->composition, 0)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 119, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_self->name, __pyx_t_3};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 119, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_v_self->name, __pyx_t_3};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 119, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 119, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_4) {
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4); __pyx_t_4 = NULL;
    }
    __Pyx_INCREF(__pyx_v_self->name);
    __Pyx_GIVEREF(__pyx_v_self->name);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_v_self->name);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_t_3);
    __pyx_t_3 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 119, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase))))) __PYX_ERR(0, 119, __pyx_L1_error)
  __pyx_r = ((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase *)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "glycopeptidepy/_c/structure/fragment.pyx":118
 * cdef class ChemicalShiftBase(object):
 * 
 *     cpdef ChemicalShiftBase clone(self):             # <<<<<<<<<<<<<<
 *         return self.__class__(self.name, self.composition.clone())
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment.ChemicalShiftBase.clone", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_17ChemicalShiftBase_1clone(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_14glycopeptidepy_2_c_9structure_8fragment_17ChemicalShiftBase_clone[] = "ChemicalShiftBase.clone(self) -> ChemicalShiftBase";
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_17ChemicalShiftBase_1clone(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("clone (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_17ChemicalShiftBase_clone(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_17ChemicalShiftBase_clone(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("clone", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_17ChemicalShiftBase_clone(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 118, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment.ChemicalShiftBase.clone", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/fragment.pyx":121
 *         return self.__class__(self.name, self.composition.clone())
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return self.name
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_17ChemicalShiftBase_3__str__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_17ChemicalShiftBase_3__str__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__str__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_17ChemicalShiftBase_2__str__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_17ChemicalShiftBase_2__str__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__str__", 0);

  /* "glycopeptidepy/_c/structure/fragment.pyx":122
 * 
 *     def __str__(self):
 *         return self.name             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->name);
  __pyx_r = __pyx_v_self->name;
  goto __pyx_L0;

  /* "glycopeptidepy/_c/structure/fragment.pyx":121
 *         return self.__class__(self.name, self.composition.clone())
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return self.name
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/fragment.pyx":124
 *         return self.name
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return "%s(name=%r)" % (self.__class__.__name__, self.name)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_17ChemicalShiftBase_5__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_17ChemicalShiftBase_5__repr__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_17ChemicalShiftBase_4__repr__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_17ChemicalShiftBase_4__repr__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "glycopeptidepy/_c/structure/fragment.pyx":125
 * 
 *     def __repr__(self):
 *         return "%s(name=%r)" % (self.__class__.__name__, self.name)             # <<<<<<<<<<<<<<
 * 
 *     def is_loss(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_class); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 125, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_name); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 125, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 125, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self->name);
  __Pyx_GIVEREF(__pyx_v_self->name);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_self->name);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyString_Format(__pyx_kp_s_s_name_r, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 125, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "glycopeptidepy/_c/structure/fragment.pyx":124
 *         return self.name
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return "%s(name=%r)" % (self.__class__.__name__, self.name)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment.ChemicalShiftBase.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/fragment.pyx":127
 *         return "%s(name=%r)" % (self.__class__.__name__, self.name)
 * 
 *     def is_loss(self):             # <<<<<<<<<<<<<<
 *         return self.mass < 0
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_17ChemicalShiftBase_7is_loss(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_14glycopeptidepy_2_c_9structure_8fragment_17ChemicalShiftBase_6is_loss[] = "ChemicalShiftBase.is_loss(self)";
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_17ChemicalShiftBase_7is_loss(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("is_loss (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_17ChemicalShiftBase_6is_loss(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_17ChemicalShiftBase_6is_loss(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("is_loss", 0);

  /* "glycopeptidepy/_c/structure/fragment.pyx":128
 * 
 *     def is_loss(self):
 *         return self.mass < 0             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong((__pyx_v_self->mass < 0.0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 128, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "glycopeptidepy/_c/structure/fragment.pyx":127
 *         return "%s(name=%r)" % (self.__class__.__name__, self.name)
 * 
 *     def is_loss(self):             # <<<<<<<<<<<<<<
 *         return self.mass < 0
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment.ChemicalShiftBase.is_loss", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/fragment.pxd":8
 * cdef class ChemicalShiftBase(object):
 *     cdef:
 *         public str name             # <<<<<<<<<<<<<<
 *         public CComposition composition
 *         public double mass
 */

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_17ChemicalShiftBase_4name_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_17ChemicalShiftBase_4name_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_17ChemicalShiftBase_4name___get__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_17ChemicalShiftBase_4name___get__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->name);
  __pyx_r = __pyx_v_self->name;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_17ChemicalShiftBase_4name_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_17ChemicalShiftBase_4name_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_17ChemicalShiftBase_4name_2__set__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_17ChemicalShiftBase_4name_2__set__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);
  if (!(likely(PyString_CheckExact(__pyx_v_value))||((__pyx_v_value) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "str", Py_TYPE(__pyx_v_value)->tp_name), 0))) __PYX_ERR(1, 8, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_value;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->name);
  __Pyx_DECREF(__pyx_v_self->name);
  __pyx_v_self->name = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment.ChemicalShiftBase.name.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_17ChemicalShiftBase_4name_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_17ChemicalShiftBase_4name_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_17ChemicalShiftBase_4name_4__del__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_17ChemicalShiftBase_4name_4__del__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->name);
  __Pyx_DECREF(__pyx_v_self->name);
  __pyx_v_self->name = ((PyObject*)Py_None);

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/fragment.pxd":9
 *     cdef:
 *         public str name
 *         public CComposition composition             # <<<<<<<<<<<<<<
 *         public double mass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_17ChemicalShiftBase_11composition_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_17ChemicalShiftBase_11composition_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_17ChemicalShiftBase_11composition___get__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_17ChemicalShiftBase_11composition___get__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->composition));
  __pyx_r = ((PyObject *)__pyx_v_self->composition);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_17ChemicalShiftBase_11composition_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_17ChemicalShiftBase_11composition_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_17ChemicalShiftBase_11composition_2__set__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_17ChemicalShiftBase_11composition_2__set__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);
  if (!(likely(((__pyx_v_value) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_value, __pyx_ptype_5glypy_11composition_12ccomposition_CComposition))))) __PYX_ERR(1, 9, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_value;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->composition);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->composition));
  __pyx_v_self->composition = ((struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment.ChemicalShiftBase.composition.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_17ChemicalShiftBase_11composition_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_17ChemicalShiftBase_11composition_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_17ChemicalShiftBase_11composition_4__del__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_17ChemicalShiftBase_11composition_4__del__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->composition);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->composition));
  __pyx_v_self->composition = ((struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *)Py_None);

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/fragment.pxd":10
 *         public str name
 *         public CComposition composition
 *         public double mass             # <<<<<<<<<<<<<<
 * 
 *     cpdef ChemicalShiftBase clone(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_17ChemicalShiftBase_4mass_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_17ChemicalShiftBase_4mass_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_17ChemicalShiftBase_4mass___get__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_17ChemicalShiftBase_4mass___get__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->mass); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 10, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment.ChemicalShiftBase.mass.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_17ChemicalShiftBase_4mass_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_17ChemicalShiftBase_4mass_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_17ChemicalShiftBase_4mass_2__set__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_17ChemicalShiftBase_4mass_2__set__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  double __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_value); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 10, __pyx_L1_error)
  __pyx_v_self->mass = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment.ChemicalShiftBase.mass.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_17ChemicalShiftBase_9__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_14glycopeptidepy_2_c_9structure_8fragment_17ChemicalShiftBase_8__reduce_cython__[] = "ChemicalShiftBase.__reduce_cython__(self)";
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_17ChemicalShiftBase_9__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_17ChemicalShiftBase_8__reduce_cython__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_17ChemicalShiftBase_8__reduce_cython__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.composition, self.mass, self.name)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->mass); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->composition));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self->composition));
  PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)__pyx_v_self->composition));
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_1);
  __Pyx_INCREF(__pyx_v_self->name);
  __Pyx_GIVEREF(__pyx_v_self->name);
  PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_v_self->name);
  __pyx_t_1 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.composition, self.mass, self.name)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_2 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v__dict = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "(tree fragment)":7
 *     state = (self.composition, self.mass, self.name)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_3 = (__pyx_v__dict != Py_None);
  __pyx_t_4 = (__pyx_t_3 != 0);
  if (__pyx_t_4) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v__dict);
    __pyx_t_1 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.composition is not None or self.name is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.composition, self.mass, self.name)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.composition is not None or self.name is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_ChemicalShiftBase, (type(self), 0x54d98d7, None), state
 */
  /*else*/ {
    __pyx_t_3 = (((PyObject *)__pyx_v_self->composition) != Py_None);
    __pyx_t_5 = (__pyx_t_3 != 0);
    if (!__pyx_t_5) {
    } else {
      __pyx_t_4 = __pyx_t_5;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_5 = (__pyx_v_self->name != ((PyObject*)Py_None));
    __pyx_t_3 = (__pyx_t_5 != 0);
    __pyx_t_4 = __pyx_t_3;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_4;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.composition is not None or self.name is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_ChemicalShiftBase, (type(self), 0x54d98d7, None), state
 *     else:
 */
  __pyx_t_4 = (__pyx_v_use_setstate != 0);
  if (__pyx_t_4) {

    /* "(tree fragment)":13
 *         use_setstate = self.composition is not None or self.name is not None
 *     if use_setstate:
 *         return __pyx_unpickle_ChemicalShiftBase, (type(self), 0x54d98d7, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_ChemicalShiftBase, (type(self), 0x54d98d7, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_pyx_unpickle_ChemicalShiftBase); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_88971479);
    __Pyx_GIVEREF(__pyx_int_88971479);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_int_88971479);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    PyTuple_SET_ITEM(__pyx_t_2, 2, Py_None);
    __pyx_t_6 = PyTuple_New(3); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_2);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_v_state);
    __pyx_t_1 = 0;
    __pyx_t_2 = 0;
    __pyx_r = __pyx_t_6;
    __pyx_t_6 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.composition is not None or self.name is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_ChemicalShiftBase, (type(self), 0x54d98d7, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_ChemicalShiftBase, (type(self), 0x54d98d7, None), state
 *     else:
 *         return __pyx_unpickle_ChemicalShiftBase, (type(self), 0x54d98d7, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_ChemicalShiftBase__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_pyx_unpickle_ChemicalShiftBase); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_88971479);
    __Pyx_GIVEREF(__pyx_int_88971479);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_int_88971479);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_v_state);
    __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_2);
    __pyx_t_6 = 0;
    __pyx_t_2 = 0;
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment.ChemicalShiftBase.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_ChemicalShiftBase, (type(self), 0x54d98d7, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_ChemicalShiftBase__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_17ChemicalShiftBase_11__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static char __pyx_doc_14glycopeptidepy_2_c_9structure_8fragment_17ChemicalShiftBase_10__setstate_cython__[] = "ChemicalShiftBase.__setstate_cython__(self, __pyx_state)";
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_17ChemicalShiftBase_11__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_17ChemicalShiftBase_10__setstate_cython__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_17ChemicalShiftBase_10__setstate_cython__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_ChemicalShiftBase, (type(self), 0x54d98d7, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_ChemicalShiftBase__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(2, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_14glycopeptidepy_2_c_9structure_8fragment___pyx_unpickle_ChemicalShiftBase__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_ChemicalShiftBase, (type(self), 0x54d98d7, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_ChemicalShiftBase__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment.ChemicalShiftBase.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/fragment.pyx":133
 * cdef class IonSeriesBase(object):
 * 
 *     def __eq__(self, other):             # <<<<<<<<<<<<<<
 *         cdef:
 *             IonSeriesBase other_tp
 */

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_1__eq__(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_1__eq__(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__eq__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase___eq__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *)__pyx_v_self), ((PyObject *)__pyx_v_other));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase___eq__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *__pyx_v_self, PyObject *__pyx_v_other) {
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *__pyx_v_other_tp = 0;
  PyObject *__pyx_v_temp = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__eq__", 0);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __Pyx_INCREF(__pyx_v_other);

  /* "glycopeptidepy/_c/structure/fragment.pyx":137
 *             IonSeriesBase other_tp
 *             object temp
 *         if not isinstance(self, IonSeriesBase):             # <<<<<<<<<<<<<<
 *             temp = self
 *             self = <IonSeriesBase>other
 */
  __pyx_t_1 = __Pyx_TypeCheck(((PyObject *)__pyx_v_self), __pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase); 
  __pyx_t_2 = ((!(__pyx_t_1 != 0)) != 0);
  if (__pyx_t_2) {

    /* "glycopeptidepy/_c/structure/fragment.pyx":138
 *             object temp
 *         if not isinstance(self, IonSeriesBase):
 *             temp = self             # <<<<<<<<<<<<<<
 *             self = <IonSeriesBase>other
 *             other = temp
 */
    __Pyx_INCREF(((PyObject *)__pyx_v_self));
    __pyx_v_temp = ((PyObject *)__pyx_v_self);

    /* "glycopeptidepy/_c/structure/fragment.pyx":139
 *         if not isinstance(self, IonSeriesBase):
 *             temp = self
 *             self = <IonSeriesBase>other             # <<<<<<<<<<<<<<
 *             other = temp
 *         if isinstance(other, IonSeriesBase):
 */
    __pyx_t_3 = __pyx_v_other;
    __Pyx_INCREF(__pyx_t_3);
    __Pyx_DECREF_SET(__pyx_v_self, ((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "glycopeptidepy/_c/structure/fragment.pyx":140
 *             temp = self
 *             self = <IonSeriesBase>other
 *             other = temp             # <<<<<<<<<<<<<<
 *         if isinstance(other, IonSeriesBase):
 *             other_tp = <IonSeriesBase>other
 */
    __Pyx_INCREF(__pyx_v_temp);
    __Pyx_DECREF_SET(__pyx_v_other, __pyx_v_temp);

    /* "glycopeptidepy/_c/structure/fragment.pyx":137
 *             IonSeriesBase other_tp
 *             object temp
 *         if not isinstance(self, IonSeriesBase):             # <<<<<<<<<<<<<<
 *             temp = self
 *             self = <IonSeriesBase>other
 */
  }

  /* "glycopeptidepy/_c/structure/fragment.pyx":141
 *             self = <IonSeriesBase>other
 *             other = temp
 *         if isinstance(other, IonSeriesBase):             # <<<<<<<<<<<<<<
 *             other_tp = <IonSeriesBase>other
 *             return self is other_tp or self.name == other_tp.name
 */
  __pyx_t_2 = __Pyx_TypeCheck(__pyx_v_other, __pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase); 
  __pyx_t_1 = (__pyx_t_2 != 0);
  if (__pyx_t_1) {

    /* "glycopeptidepy/_c/structure/fragment.pyx":142
 *             other = temp
 *         if isinstance(other, IonSeriesBase):
 *             other_tp = <IonSeriesBase>other             # <<<<<<<<<<<<<<
 *             return self is other_tp or self.name == other_tp.name
 *         else:
 */
    __pyx_t_3 = __pyx_v_other;
    __Pyx_INCREF(__pyx_t_3);
    __pyx_v_other_tp = ((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "glycopeptidepy/_c/structure/fragment.pyx":143
 *         if isinstance(other, IonSeriesBase):
 *             other_tp = <IonSeriesBase>other
 *             return self is other_tp or self.name == other_tp.name             # <<<<<<<<<<<<<<
 *         else:
 *             return self.name == other
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = (__pyx_v_self == __pyx_v_other_tp);
    if (!__pyx_t_1) {
    } else {
      __pyx_t_4 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 143, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = __pyx_t_4;
      __pyx_t_4 = 0;
      goto __pyx_L5_bool_binop_done;
    }
    __pyx_t_1 = (__Pyx_PyString_Equals(__pyx_v_self->name, __pyx_v_other_tp->name, Py_EQ)); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 143, __pyx_L1_error)
    __pyx_t_4 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 143, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = __pyx_t_4;
    __pyx_t_4 = 0;
    __pyx_L5_bool_binop_done:;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "glycopeptidepy/_c/structure/fragment.pyx":141
 *             self = <IonSeriesBase>other
 *             other = temp
 *         if isinstance(other, IonSeriesBase):             # <<<<<<<<<<<<<<
 *             other_tp = <IonSeriesBase>other
 *             return self is other_tp or self.name == other_tp.name
 */
  }

  /* "glycopeptidepy/_c/structure/fragment.pyx":145
 *             return self is other_tp or self.name == other_tp.name
 *         else:
 *             return self.name == other             # <<<<<<<<<<<<<<
 * 
 *     def __ne__(self, other):
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = PyObject_RichCompare(__pyx_v_self->name, __pyx_v_other, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 145, __pyx_L1_error)
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "glycopeptidepy/_c/structure/fragment.pyx":133
 * cdef class IonSeriesBase(object):
 * 
 *     def __eq__(self, other):             # <<<<<<<<<<<<<<
 *         cdef:
 *             IonSeriesBase other_tp
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment.IonSeriesBase.__eq__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_other_tp);
  __Pyx_XDECREF(__pyx_v_temp);
  __Pyx_XDECREF((PyObject *)__pyx_v_self);
  __Pyx_XDECREF(__pyx_v_other);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/fragment.pyx":147
 *             return self.name == other
 * 
 *     def __ne__(self, other):             # <<<<<<<<<<<<<<
 *         cdef:
 *             IonSeriesBase other_tp
 */

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_3__ne__(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_3__ne__(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__ne__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_2__ne__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *)__pyx_v_self), ((PyObject *)__pyx_v_other));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_2__ne__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *__pyx_v_self, PyObject *__pyx_v_other) {
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *__pyx_v_other_tp = 0;
  PyObject *__pyx_v_temp = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__ne__", 0);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __Pyx_INCREF(__pyx_v_other);

  /* "glycopeptidepy/_c/structure/fragment.pyx":151
 *             IonSeriesBase other_tp
 *             object temp
 *         if not isinstance(self, IonSeriesBase):             # <<<<<<<<<<<<<<
 *             temp = self
 *             self = <IonSeriesBase>other
 */
  __pyx_t_1 = __Pyx_TypeCheck(((PyObject *)__pyx_v_self), __pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase); 
  __pyx_t_2 = ((!(__pyx_t_1 != 0)) != 0);
  if (__pyx_t_2) {

    /* "glycopeptidepy/_c/structure/fragment.pyx":152
 *             object temp
 *         if not isinstance(self, IonSeriesBase):
 *             temp = self             # <<<<<<<<<<<<<<
 *             self = <IonSeriesBase>other
 *             other = temp
 */
    __Pyx_INCREF(((PyObject *)__pyx_v_self));
    __pyx_v_temp = ((PyObject *)__pyx_v_self);

    /* "glycopeptidepy/_c/structure/fragment.pyx":153
 *         if not isinstance(self, IonSeriesBase):
 *             temp = self
 *             self = <IonSeriesBase>other             # <<<<<<<<<<<<<<
 *             other = temp
 *         if isinstance(other, IonSeriesBase):
 */
    __pyx_t_3 = __pyx_v_other;
    __Pyx_INCREF(__pyx_t_3);
    __Pyx_DECREF_SET(__pyx_v_self, ((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "glycopeptidepy/_c/structure/fragment.pyx":154
 *             temp = self
 *             self = <IonSeriesBase>other
 *             other = temp             # <<<<<<<<<<<<<<
 *         if isinstance(other, IonSeriesBase):
 *             other_tp = <IonSeriesBase>other
 */
    __Pyx_INCREF(__pyx_v_temp);
    __Pyx_DECREF_SET(__pyx_v_other, __pyx_v_temp);

    /* "glycopeptidepy/_c/structure/fragment.pyx":151
 *             IonSeriesBase other_tp
 *             object temp
 *         if not isinstance(self, IonSeriesBase):             # <<<<<<<<<<<<<<
 *             temp = self
 *             self = <IonSeriesBase>other
 */
  }

  /* "glycopeptidepy/_c/structure/fragment.pyx":155
 *             self = <IonSeriesBase>other
 *             other = temp
 *         if isinstance(other, IonSeriesBase):             # <<<<<<<<<<<<<<
 *             other_tp = <IonSeriesBase>other
 *             return self.name != other_tp.name
 */
  __pyx_t_2 = __Pyx_TypeCheck(__pyx_v_other, __pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase); 
  __pyx_t_1 = (__pyx_t_2 != 0);
  if (__pyx_t_1) {

    /* "glycopeptidepy/_c/structure/fragment.pyx":156
 *             other = temp
 *         if isinstance(other, IonSeriesBase):
 *             other_tp = <IonSeriesBase>other             # <<<<<<<<<<<<<<
 *             return self.name != other_tp.name
 *         else:
 */
    __pyx_t_3 = __pyx_v_other;
    __Pyx_INCREF(__pyx_t_3);
    __pyx_v_other_tp = ((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "glycopeptidepy/_c/structure/fragment.pyx":157
 *         if isinstance(other, IonSeriesBase):
 *             other_tp = <IonSeriesBase>other
 *             return self.name != other_tp.name             # <<<<<<<<<<<<<<
 *         else:
 *             return self.name != other
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = (__Pyx_PyString_Equals(__pyx_v_self->name, __pyx_v_other_tp->name, Py_NE)); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 157, __pyx_L1_error)
    __pyx_t_3 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 157, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "glycopeptidepy/_c/structure/fragment.pyx":155
 *             self = <IonSeriesBase>other
 *             other = temp
 *         if isinstance(other, IonSeriesBase):             # <<<<<<<<<<<<<<
 *             other_tp = <IonSeriesBase>other
 *             return self.name != other_tp.name
 */
  }

  /* "glycopeptidepy/_c/structure/fragment.pyx":159
 *             return self.name != other_tp.name
 *         else:
 *             return self.name != other             # <<<<<<<<<<<<<<
 * 
 *     def __hash__(self):
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = PyObject_RichCompare(__pyx_v_self->name, __pyx_v_other, Py_NE); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 159, __pyx_L1_error)
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "glycopeptidepy/_c/structure/fragment.pyx":147
 *             return self.name == other
 * 
 *     def __ne__(self, other):             # <<<<<<<<<<<<<<
 *         cdef:
 *             IonSeriesBase other_tp
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment.IonSeriesBase.__ne__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_other_tp);
  __Pyx_XDECREF(__pyx_v_temp);
  __Pyx_XDECREF((PyObject *)__pyx_v_self);
  __Pyx_XDECREF(__pyx_v_other);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/fragment.pyx":161
 *             return self.name != other
 * 
 *     def __hash__(self):             # <<<<<<<<<<<<<<
 *         return self._hash
 * 
 */

/* Python wrapper */
static Py_hash_t __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_5__hash__(PyObject *__pyx_v_self); /*proto*/
static Py_hash_t __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_5__hash__(PyObject *__pyx_v_self) {
  Py_hash_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__hash__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_4__hash__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static Py_hash_t __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_4__hash__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *__pyx_v_self) {
  Py_hash_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__hash__", 0);

  /* "glycopeptidepy/_c/structure/fragment.pyx":162
 * 
 *     def __hash__(self):
 *         return self._hash             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_self->_hash;
  goto __pyx_L0;

  /* "glycopeptidepy/_c/structure/fragment.pyx":161
 *             return self.name != other
 * 
 *     def __hash__(self):             # <<<<<<<<<<<<<<
 *         return self._hash
 * 
 */

  /* function exit code */
  __pyx_L0:;
  if (unlikely(__pyx_r == -1) && !PyErr_Occurred()) __pyx_r = -2;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/fragment.pxd":17
 * cdef class IonSeriesBase(object):
 *     cdef:
 *         public str name             # <<<<<<<<<<<<<<
 *         public int direction
 *         public bint includes_peptide
 */

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_4name_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_4name_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_4name___get__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_4name___get__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->name);
  __pyx_r = __pyx_v_self->name;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_4name_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_4name_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_4name_2__set__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_4name_2__set__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);
  if (!(likely(PyString_CheckExact(__pyx_v_value))||((__pyx_v_value) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "str", Py_TYPE(__pyx_v_value)->tp_name), 0))) __PYX_ERR(1, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_value;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->name);
  __Pyx_DECREF(__pyx_v_self->name);
  __pyx_v_self->name = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment.IonSeriesBase.name.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_4name_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_4name_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_4name_4__del__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_4name_4__del__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->name);
  __Pyx_DECREF(__pyx_v_self->name);
  __pyx_v_self->name = ((PyObject*)Py_None);

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/fragment.pxd":18
 *     cdef:
 *         public str name
 *         public int direction             # <<<<<<<<<<<<<<
 *         public bint includes_peptide
 *         public double mass_shift
 */

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_9direction_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_9direction_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_9direction___get__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_9direction___get__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_self->direction); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 18, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment.IonSeriesBase.direction.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_9direction_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_9direction_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_9direction_2__set__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_9direction_2__set__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyInt_As_int(__pyx_v_value); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 18, __pyx_L1_error)
  __pyx_v_self->direction = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment.IonSeriesBase.direction.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/fragment.pxd":19
 *         public str name
 *         public int direction
 *         public bint includes_peptide             # <<<<<<<<<<<<<<
 *         public double mass_shift
 *         public CComposition composition_shift
 */

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_16includes_peptide_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_16includes_peptide_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_16includes_peptide___get__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_16includes_peptide___get__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->includes_peptide); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 19, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment.IonSeriesBase.includes_peptide.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_16includes_peptide_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_16includes_peptide_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_16includes_peptide_2__set__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_16includes_peptide_2__set__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_value); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 19, __pyx_L1_error)
  __pyx_v_self->includes_peptide = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment.IonSeriesBase.includes_peptide.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/fragment.pxd":20
 *         public int direction
 *         public bint includes_peptide
 *         public double mass_shift             # <<<<<<<<<<<<<<
 *         public CComposition composition_shift
 *         public Py_hash_t _hash
 */

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_10mass_shift_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_10mass_shift_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_10mass_shift___get__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_10mass_shift___get__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->mass_shift); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 20, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment.IonSeriesBase.mass_shift.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_10mass_shift_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_10mass_shift_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_10mass_shift_2__set__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_10mass_shift_2__set__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  double __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_value); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 20, __pyx_L1_error)
  __pyx_v_self->mass_shift = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment.IonSeriesBase.mass_shift.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/fragment.pxd":21
 *         public bint includes_peptide
 *         public double mass_shift
 *         public CComposition composition_shift             # <<<<<<<<<<<<<<
 *         public Py_hash_t _hash
 *         public int int_code
 */

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_17composition_shift_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_17composition_shift_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_17composition_shift___get__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_17composition_shift___get__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->composition_shift));
  __pyx_r = ((PyObject *)__pyx_v_self->composition_shift);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_17composition_shift_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_17composition_shift_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_17composition_shift_2__set__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_17composition_shift_2__set__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);
  if (!(likely(((__pyx_v_value) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_value, __pyx_ptype_5glypy_11composition_12ccomposition_CComposition))))) __PYX_ERR(1, 21, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_value;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->composition_shift);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->composition_shift));
  __pyx_v_self->composition_shift = ((struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment.IonSeriesBase.composition_shift.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_17composition_shift_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_17composition_shift_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_17composition_shift_4__del__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_17composition_shift_4__del__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->composition_shift);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->composition_shift));
  __pyx_v_self->composition_shift = ((struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *)Py_None);

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/fragment.pxd":22
 *         public double mass_shift
 *         public CComposition composition_shift
 *         public Py_hash_t _hash             # <<<<<<<<<<<<<<
 *         public int int_code
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_5_hash_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_5_hash_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_5_hash___get__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_5_hash___get__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_FromHash_t(__pyx_v_self->_hash); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 22, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment.IonSeriesBase._hash.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_5_hash_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_5_hash_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_5_hash_2__set__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_5_hash_2__set__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  Py_hash_t __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyInt_AsHash_t(__pyx_v_value); if (unlikely((__pyx_t_1 == (Py_hash_t)-1) && PyErr_Occurred())) __PYX_ERR(1, 22, __pyx_L1_error)
  __pyx_v_self->_hash = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment.IonSeriesBase._hash.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/fragment.pxd":23
 *         public CComposition composition_shift
 *         public Py_hash_t _hash
 *         public int int_code             # <<<<<<<<<<<<<<
 * 
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_8int_code_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_8int_code_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_8int_code___get__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_8int_code___get__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_self->int_code); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment.IonSeriesBase.int_code.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_8int_code_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_8int_code_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_8int_code_2__set__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_8int_code_2__set__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyInt_As_int(__pyx_v_value); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 23, __pyx_L1_error)
  __pyx_v_self->int_code = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment.IonSeriesBase.int_code.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_7__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_6__reduce_cython__[] = "IonSeriesBase.__reduce_cython__(self)";
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_7__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_6__reduce_cython__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_6__reduce_cython__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self._hash, self.composition_shift, self.direction, self.includes_peptide, self.int_code, self.mass_shift, self.name)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = __Pyx_PyInt_FromHash_t(__pyx_v_self->_hash); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_self->direction); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyBool_FromLong(__pyx_v_self->includes_peptide); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_v_self->int_code); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = PyFloat_FromDouble(__pyx_v_self->mass_shift); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = PyTuple_New(7); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_1);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->composition_shift));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self->composition_shift));
  PyTuple_SET_ITEM(__pyx_t_6, 1, ((PyObject *)__pyx_v_self->composition_shift));
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_6, 3, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_6, 4, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_6, 5, __pyx_t_5);
  __Pyx_INCREF(__pyx_v_self->name);
  __Pyx_GIVEREF(__pyx_v_self->name);
  PyTuple_SET_ITEM(__pyx_t_6, 6, __pyx_v_self->name);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_t_5 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_6);
  __pyx_t_6 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self._hash, self.composition_shift, self.direction, self.includes_peptide, self.int_code, self.mass_shift, self.name)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_6 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_v__dict = __pyx_t_6;
  __pyx_t_6 = 0;

  /* "(tree fragment)":7
 *     state = (self._hash, self.composition_shift, self.direction, self.includes_peptide, self.int_code, self.mass_shift, self.name)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_7 = (__pyx_v__dict != Py_None);
  __pyx_t_8 = (__pyx_t_7 != 0);
  if (__pyx_t_8) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_6 = PyTuple_New(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_v__dict);
    __pyx_t_5 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_5));
    __pyx_t_5 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.composition_shift is not None or self.name is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self._hash, self.composition_shift, self.direction, self.includes_peptide, self.int_code, self.mass_shift, self.name)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.composition_shift is not None or self.name is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_IonSeriesBase, (type(self), 0xf30379a, None), state
 */
  /*else*/ {
    __pyx_t_7 = (((PyObject *)__pyx_v_self->composition_shift) != Py_None);
    __pyx_t_9 = (__pyx_t_7 != 0);
    if (!__pyx_t_9) {
    } else {
      __pyx_t_8 = __pyx_t_9;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_9 = (__pyx_v_self->name != ((PyObject*)Py_None));
    __pyx_t_7 = (__pyx_t_9 != 0);
    __pyx_t_8 = __pyx_t_7;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_8;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.composition_shift is not None or self.name is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_IonSeriesBase, (type(self), 0xf30379a, None), state
 *     else:
 */
  __pyx_t_8 = (__pyx_v_use_setstate != 0);
  if (__pyx_t_8) {

    /* "(tree fragment)":13
 *         use_setstate = self.composition_shift is not None or self.name is not None
 *     if use_setstate:
 *         return __pyx_unpickle_IonSeriesBase, (type(self), 0xf30379a, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_IonSeriesBase, (type(self), 0xf30379a, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_pyx_unpickle_IonSeriesBase); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = PyTuple_New(3); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_6, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_254818202);
    __Pyx_GIVEREF(__pyx_int_254818202);
    PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_int_254818202);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    PyTuple_SET_ITEM(__pyx_t_6, 2, Py_None);
    __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_6);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_v_state);
    __pyx_t_5 = 0;
    __pyx_t_6 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.composition_shift is not None or self.name is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_IonSeriesBase, (type(self), 0xf30379a, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_IonSeriesBase, (type(self), 0xf30379a, None), state
 *     else:
 *         return __pyx_unpickle_IonSeriesBase, (type(self), 0xf30379a, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_IonSeriesBase__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_pyx_unpickle_IonSeriesBase); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = PyTuple_New(3); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_6, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_254818202);
    __Pyx_GIVEREF(__pyx_int_254818202);
    PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_int_254818202);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_v_state);
    __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_6);
    __pyx_t_4 = 0;
    __pyx_t_6 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment.IonSeriesBase.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_IonSeriesBase, (type(self), 0xf30379a, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_IonSeriesBase__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_9__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static char __pyx_doc_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_8__setstate_cython__[] = "IonSeriesBase.__setstate_cython__(self, __pyx_state)";
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_9__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_8__setstate_cython__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_8__setstate_cython__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_IonSeriesBase, (type(self), 0xf30379a, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_IonSeriesBase__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(2, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_14glycopeptidepy_2_c_9structure_8fragment___pyx_unpickle_IonSeriesBase__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_IonSeriesBase, (type(self), 0xf30379a, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_IonSeriesBase__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment.IonSeriesBase.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/fragment.pyx":183
 *     """
 * 
 *     cpdef IonSeriesBase get_series(self):             # <<<<<<<<<<<<<<
 *         '''Return the :class:`IonSeries` this fragment belongs to.
 * 
 */

static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_1get_series(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_get_series(CYTHON_UNUSED struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *__pyx_v_self, int __pyx_skip_dispatch) {
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_series", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_get_series); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 183, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_1get_series)) {
        __Pyx_XDECREF(((PyObject *)__pyx_r));
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 183, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase))))) __PYX_ERR(0, 183, __pyx_L1_error)
        __pyx_r = ((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "glycopeptidepy/_c/structure/fragment.pyx":190
 *         :class:`IonSeries`
 *         '''
 *         raise NotImplementedError()             # <<<<<<<<<<<<<<
 * 
 *     def __hash__(self):
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_builtin_NotImplementedError); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 190, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(0, 190, __pyx_L1_error)

  /* "glycopeptidepy/_c/structure/fragment.pyx":183
 *     """
 * 
 *     cpdef IonSeriesBase get_series(self):             # <<<<<<<<<<<<<<
 *         '''Return the :class:`IonSeries` this fragment belongs to.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment.FragmentBase.get_series", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_1get_series(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_get_series[] = "FragmentBase.get_series(self) -> IonSeriesBase\nReturn the :class:`IonSeries` this fragment belongs to.\n\n        Returns\n        -------\n        :class:`IonSeries`\n        ";
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_1get_series(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_series (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_get_series(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_get_series(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_series", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_get_series(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 183, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment.FragmentBase.get_series", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/fragment.pyx":192
 *         raise NotImplementedError()
 * 
 *     def __hash__(self):             # <<<<<<<<<<<<<<
 *         if self._hash == -1:
 *             if self._name is None:
 */

/* Python wrapper */
static Py_hash_t __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_3__hash__(PyObject *__pyx_v_self); /*proto*/
static Py_hash_t __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_3__hash__(PyObject *__pyx_v_self) {
  Py_hash_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__hash__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_2__hash__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static Py_hash_t __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_2__hash__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *__pyx_v_self) {
  Py_hash_t __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  Py_hash_t __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__hash__", 0);

  /* "glycopeptidepy/_c/structure/fragment.pyx":193
 * 
 *     def __hash__(self):
 *         if self._hash == -1:             # <<<<<<<<<<<<<<
 *             if self._name is None:
 *                 self._update_hash_name()
 */
  __pyx_t_1 = ((__pyx_v_self->_hash == -1L) != 0);
  if (__pyx_t_1) {

    /* "glycopeptidepy/_c/structure/fragment.pyx":194
 *     def __hash__(self):
 *         if self._hash == -1:
 *             if self._name is None:             # <<<<<<<<<<<<<<
 *                 self._update_hash_name()
 *             else:
 */
    __pyx_t_1 = (__pyx_v_self->_name == ((PyObject*)Py_None));
    __pyx_t_2 = (__pyx_t_1 != 0);
    if (__pyx_t_2) {

      /* "glycopeptidepy/_c/structure/fragment.pyx":195
 *         if self._hash == -1:
 *             if self._name is None:
 *                 self._update_hash_name()             # <<<<<<<<<<<<<<
 *             else:
 *                 self._hash = hash(self._name)
 */
      ((struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *)__pyx_v_self->__pyx_vtab)->_update_hash_name(__pyx_v_self);

      /* "glycopeptidepy/_c/structure/fragment.pyx":194
 *     def __hash__(self):
 *         if self._hash == -1:
 *             if self._name is None:             # <<<<<<<<<<<<<<
 *                 self._update_hash_name()
 *             else:
 */
      goto __pyx_L4;
    }

    /* "glycopeptidepy/_c/structure/fragment.pyx":197
 *                 self._update_hash_name()
 *             else:
 *                 self._hash = hash(self._name)             # <<<<<<<<<<<<<<
 *         return self._hash
 * 
 */
    /*else*/ {
      __pyx_t_3 = __pyx_v_self->_name;
      __Pyx_INCREF(__pyx_t_3);
      __pyx_t_4 = PyObject_Hash(__pyx_t_3); if (unlikely(__pyx_t_4 == ((Py_hash_t)-1))) __PYX_ERR(0, 197, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_v_self->_hash = __pyx_t_4;
    }
    __pyx_L4:;

    /* "glycopeptidepy/_c/structure/fragment.pyx":193
 * 
 *     def __hash__(self):
 *         if self._hash == -1:             # <<<<<<<<<<<<<<
 *             if self._name is None:
 *                 self._update_hash_name()
 */
  }

  /* "glycopeptidepy/_c/structure/fragment.pyx":198
 *             else:
 *                 self._hash = hash(self._name)
 *         return self._hash             # <<<<<<<<<<<<<<
 * 
 *     def __eq__(self, other):
 */
  __pyx_r = __pyx_v_self->_hash;
  goto __pyx_L0;

  /* "glycopeptidepy/_c/structure/fragment.pyx":192
 *         raise NotImplementedError()
 * 
 *     def __hash__(self):             # <<<<<<<<<<<<<<
 *         if self._hash == -1:
 *             if self._name is None:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment.FragmentBase.__hash__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  if (unlikely(__pyx_r == -1) && !PyErr_Occurred()) __pyx_r = -2;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/fragment.pyx":200
 *         return self._hash
 * 
 *     def __eq__(self, other):             # <<<<<<<<<<<<<<
 *         try:
 *             return self.name == other.name and abs(self.mass - other.mass) < 1e-5
 */

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_5__eq__(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_5__eq__(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__eq__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_4__eq__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *)__pyx_v_self), ((PyObject *)__pyx_v_other));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_4__eq__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__eq__", 0);

  /* "glycopeptidepy/_c/structure/fragment.pyx":201
 * 
 *     def __eq__(self, other):
 *         try:             # <<<<<<<<<<<<<<
 *             return self.name == other.name and abs(self.mass - other.mass) < 1e-5
 *         except AttributeError:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "glycopeptidepy/_c/structure/fragment.pyx":202
 *     def __eq__(self, other):
 *         try:
 *             return self.name == other.name and abs(self.mass - other.mass) < 1e-5             # <<<<<<<<<<<<<<
 *         except AttributeError:
 *             return False
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_name_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 202, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_other, __pyx_n_s_name_2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 202, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_7 = PyObject_RichCompare(__pyx_t_5, __pyx_t_6, Py_EQ); __Pyx_XGOTREF(__pyx_t_7); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 202, __pyx_L3_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_8 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely(__pyx_t_8 < 0)) __PYX_ERR(0, 202, __pyx_L3_error)
      if (__pyx_t_8) {
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      } else {
        __Pyx_INCREF(__pyx_t_7);
        __pyx_t_4 = __pyx_t_7;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        goto __pyx_L9_bool_binop_done;
      }
      __pyx_t_7 = PyFloat_FromDouble(__pyx_v_self->mass); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 202, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_other, __pyx_n_s_mass); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 202, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_5 = PyNumber_Subtract(__pyx_t_7, __pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 202, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_6 = __Pyx_PyNumber_Absolute(__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 202, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_5 = PyObject_RichCompare(__pyx_t_6, __pyx_float_1eneg_5, Py_LT); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 202, __pyx_L3_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_INCREF(__pyx_t_5);
      __pyx_t_4 = __pyx_t_5;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_L9_bool_binop_done:;
      __pyx_r = __pyx_t_4;
      __pyx_t_4 = 0;
      goto __pyx_L7_try_return;

      /* "glycopeptidepy/_c/structure/fragment.pyx":201
 * 
 *     def __eq__(self, other):
 *         try:             # <<<<<<<<<<<<<<
 *             return self.name == other.name and abs(self.mass - other.mass) < 1e-5
 *         except AttributeError:
 */
    }
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "glycopeptidepy/_c/structure/fragment.pyx":203
 *         try:
 *             return self.name == other.name and abs(self.mass - other.mass) < 1e-5
 *         except AttributeError:             # <<<<<<<<<<<<<<
 *             return False
 * 
 */
    __pyx_t_9 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_AttributeError);
    if (__pyx_t_9) {
      __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment.FragmentBase.__eq__", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_4, &__pyx_t_5, &__pyx_t_6) < 0) __PYX_ERR(0, 203, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_6);

      /* "glycopeptidepy/_c/structure/fragment.pyx":204
 *             return self.name == other.name and abs(self.mass - other.mass) < 1e-5
 *         except AttributeError:
 *             return False             # <<<<<<<<<<<<<<
 * 
 *     def __ne__(self, other):
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(Py_False);
      __pyx_r = Py_False;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      goto __pyx_L6_except_return;
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "glycopeptidepy/_c/structure/fragment.pyx":201
 * 
 *     def __eq__(self, other):
 *         try:             # <<<<<<<<<<<<<<
 *             return self.name == other.name and abs(self.mass - other.mass) < 1e-5
 *         except AttributeError:
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L7_try_return:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L0;
    __pyx_L6_except_return:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L0;
  }

  /* "glycopeptidepy/_c/structure/fragment.pyx":200
 *         return self._hash
 * 
 *     def __eq__(self, other):             # <<<<<<<<<<<<<<
 *         try:
 *             return self.name == other.name and abs(self.mass - other.mass) < 1e-5
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment.FragmentBase.__eq__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/fragment.pyx":206
 *             return False
 * 
 *     def __ne__(self, other):             # <<<<<<<<<<<<<<
 *         return not self == other
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_7__ne__(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_7__ne__(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__ne__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_6__ne__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *)__pyx_v_self), ((PyObject *)__pyx_v_other));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_6__ne__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__ne__", 0);

  /* "glycopeptidepy/_c/structure/fragment.pyx":207
 * 
 *     def __ne__(self, other):
 *         return not self == other             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyObject_RichCompare(((PyObject *)__pyx_v_self), __pyx_v_other, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 207, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 207, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyBool_FromLong((!__pyx_t_2)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 207, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "glycopeptidepy/_c/structure/fragment.pyx":206
 *             return False
 * 
 *     def __ne__(self, other):             # <<<<<<<<<<<<<<
 *         return not self == other
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment.FragmentBase.__ne__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/fragment.pyx":210
 * 
 *     @property
 *     def series(self):             # <<<<<<<<<<<<<<
 *         '''The :class:`IonSeries` this fragment is drawn from.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_6series_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_6series_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_6series___get__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_6series___get__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "glycopeptidepy/_c/structure/fragment.pyx":217
 *         :class:`IonSeries`
 *         '''
 *         return self.get_series()             # <<<<<<<<<<<<<<
 * 
 *     def clone(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *)__pyx_v_self->__pyx_vtab)->get_series(__pyx_v_self, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 217, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "glycopeptidepy/_c/structure/fragment.pyx":210
 * 
 *     @property
 *     def series(self):             # <<<<<<<<<<<<<<
 *         '''The :class:`IonSeries` this fragment is drawn from.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment.FragmentBase.series.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/fragment.pyx":219
 *         return self.get_series()
 * 
 *     def clone(self):             # <<<<<<<<<<<<<<
 *         '''Create a copy of this object
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_9clone(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_8clone[] = "FragmentBase.clone(self)\nCreate a copy of this object\n\n        Returns\n        -------\n        :class:`FragmentBase`\n        ";
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_9clone(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("clone (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_8clone(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_8clone(CYTHON_UNUSED struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("clone", 0);

  /* "glycopeptidepy/_c/structure/fragment.pyx":226
 *         :class:`FragmentBase`
 *         '''
 *         raise NotImplementedError()             # <<<<<<<<<<<<<<
 * 
 *     cpdef ChemicalShiftBase get_chemical_shift(self):
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_builtin_NotImplementedError); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 226, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(0, 226, __pyx_L1_error)

  /* "glycopeptidepy/_c/structure/fragment.pyx":219
 *         return self.get_series()
 * 
 *     def clone(self):             # <<<<<<<<<<<<<<
 *         '''Create a copy of this object
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment.FragmentBase.clone", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/fragment.pyx":228
 *         raise NotImplementedError()
 * 
 *     cpdef ChemicalShiftBase get_chemical_shift(self):             # <<<<<<<<<<<<<<
 *         '''Returns the chemical shift associated with the fragment.
 * 
 */

static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_11get_chemical_shift(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase *__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_get_chemical_shift(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *__pyx_v_self, int __pyx_skip_dispatch) {
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_chemical_shift", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_get_chemical_shift); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 228, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_11get_chemical_shift)) {
        __Pyx_XDECREF(((PyObject *)__pyx_r));
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 228, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase))))) __PYX_ERR(0, 228, __pyx_L1_error)
        __pyx_r = ((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase *)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "glycopeptidepy/_c/structure/fragment.pyx":235
 *         :class:`ChemicalShift`
 *         '''
 *         return self._chemical_shift             # <<<<<<<<<<<<<<
 * 
 *     cpdef set_chemical_shift(self, ChemicalShiftBase chemical_shift):
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __Pyx_INCREF(((PyObject *)__pyx_v_self->_chemical_shift));
  __pyx_r = __pyx_v_self->_chemical_shift;
  goto __pyx_L0;

  /* "glycopeptidepy/_c/structure/fragment.pyx":228
 *         raise NotImplementedError()
 * 
 *     cpdef ChemicalShiftBase get_chemical_shift(self):             # <<<<<<<<<<<<<<
 *         '''Returns the chemical shift associated with the fragment.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment.FragmentBase.get_chemical_shift", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_11get_chemical_shift(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_10get_chemical_shift[] = "FragmentBase.get_chemical_shift(self) -> ChemicalShiftBase\nReturns the chemical shift associated with the fragment.\n\n        Returns\n        -------\n        :class:`ChemicalShift`\n        ";
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_11get_chemical_shift(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_chemical_shift (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_10get_chemical_shift(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_10get_chemical_shift(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_chemical_shift", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_get_chemical_shift(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 228, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment.FragmentBase.get_chemical_shift", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/fragment.pyx":237
 *         return self._chemical_shift
 * 
 *     cpdef set_chemical_shift(self, ChemicalShiftBase chemical_shift):             # <<<<<<<<<<<<<<
 *         '''Sets the chemical shift associated with the fragment, updating
 *         the :attr:`mass` and :attr:`chemical_shift`.
 */

static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_13set_chemical_shift(PyObject *__pyx_v_self, PyObject *__pyx_v_chemical_shift); /*proto*/
static PyObject *__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_set_chemical_shift(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *__pyx_v_self, struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase *__pyx_v_chemical_shift, int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  Py_hash_t __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("set_chemical_shift", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_set_chemical_shift); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 237, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_13set_chemical_shift)) {
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, ((PyObject *)__pyx_v_chemical_shift)) : __Pyx_PyObject_CallOneArg(__pyx_t_3, ((PyObject *)__pyx_v_chemical_shift));
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 237, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "glycopeptidepy/_c/structure/fragment.pyx":241
 *         the :attr:`mass` and :attr:`chemical_shift`.
 *         '''
 *         if self._chemical_shift is None and chemical_shift is None:             # <<<<<<<<<<<<<<
 *             return
 *         if self._chemical_shift is not None:
 */
  __pyx_t_6 = (((PyObject *)__pyx_v_self->_chemical_shift) == Py_None);
  __pyx_t_7 = (__pyx_t_6 != 0);
  if (__pyx_t_7) {
  } else {
    __pyx_t_5 = __pyx_t_7;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_7 = (((PyObject *)__pyx_v_chemical_shift) == Py_None);
  __pyx_t_6 = (__pyx_t_7 != 0);
  __pyx_t_5 = __pyx_t_6;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_5) {

    /* "glycopeptidepy/_c/structure/fragment.pyx":242
 *         '''
 *         if self._chemical_shift is None and chemical_shift is None:
 *             return             # <<<<<<<<<<<<<<
 *         if self._chemical_shift is not None:
 *             self.mass -= self._chemical_shift.mass
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "glycopeptidepy/_c/structure/fragment.pyx":241
 *         the :attr:`mass` and :attr:`chemical_shift`.
 *         '''
 *         if self._chemical_shift is None and chemical_shift is None:             # <<<<<<<<<<<<<<
 *             return
 *         if self._chemical_shift is not None:
 */
  }

  /* "glycopeptidepy/_c/structure/fragment.pyx":243
 *         if self._chemical_shift is None and chemical_shift is None:
 *             return
 *         if self._chemical_shift is not None:             # <<<<<<<<<<<<<<
 *             self.mass -= self._chemical_shift.mass
 *         self._chemical_shift = chemical_shift
 */
  __pyx_t_5 = (((PyObject *)__pyx_v_self->_chemical_shift) != Py_None);
  __pyx_t_6 = (__pyx_t_5 != 0);
  if (__pyx_t_6) {

    /* "glycopeptidepy/_c/structure/fragment.pyx":244
 *             return
 *         if self._chemical_shift is not None:
 *             self.mass -= self._chemical_shift.mass             # <<<<<<<<<<<<<<
 *         self._chemical_shift = chemical_shift
 *         if chemical_shift is not None:
 */
    __pyx_v_self->mass = (__pyx_v_self->mass - __pyx_v_self->_chemical_shift->mass);

    /* "glycopeptidepy/_c/structure/fragment.pyx":243
 *         if self._chemical_shift is None and chemical_shift is None:
 *             return
 *         if self._chemical_shift is not None:             # <<<<<<<<<<<<<<
 *             self.mass -= self._chemical_shift.mass
 *         self._chemical_shift = chemical_shift
 */
  }

  /* "glycopeptidepy/_c/structure/fragment.pyx":245
 *         if self._chemical_shift is not None:
 *             self.mass -= self._chemical_shift.mass
 *         self._chemical_shift = chemical_shift             # <<<<<<<<<<<<<<
 *         if chemical_shift is not None:
 *             self.mass += chemical_shift.mass
 */
  __Pyx_INCREF(((PyObject *)__pyx_v_chemical_shift));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_chemical_shift));
  __Pyx_GOTREF(__pyx_v_self->_chemical_shift);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->_chemical_shift));
  __pyx_v_self->_chemical_shift = __pyx_v_chemical_shift;

  /* "glycopeptidepy/_c/structure/fragment.pyx":246
 *             self.mass -= self._chemical_shift.mass
 *         self._chemical_shift = chemical_shift
 *         if chemical_shift is not None:             # <<<<<<<<<<<<<<
 *             self.mass += chemical_shift.mass
 *         self._name = self.get_fragment_name()
 */
  __pyx_t_6 = (((PyObject *)__pyx_v_chemical_shift) != Py_None);
  __pyx_t_5 = (__pyx_t_6 != 0);
  if (__pyx_t_5) {

    /* "glycopeptidepy/_c/structure/fragment.pyx":247
 *         self._chemical_shift = chemical_shift
 *         if chemical_shift is not None:
 *             self.mass += chemical_shift.mass             # <<<<<<<<<<<<<<
 *         self._name = self.get_fragment_name()
 *         self._hash = hash(self._name)
 */
    __pyx_v_self->mass = (__pyx_v_self->mass + __pyx_v_chemical_shift->mass);

    /* "glycopeptidepy/_c/structure/fragment.pyx":246
 *             self.mass -= self._chemical_shift.mass
 *         self._chemical_shift = chemical_shift
 *         if chemical_shift is not None:             # <<<<<<<<<<<<<<
 *             self.mass += chemical_shift.mass
 *         self._name = self.get_fragment_name()
 */
  }

  /* "glycopeptidepy/_c/structure/fragment.pyx":248
 *         if chemical_shift is not None:
 *             self.mass += chemical_shift.mass
 *         self._name = self.get_fragment_name()             # <<<<<<<<<<<<<<
 *         self._hash = hash(self._name)
 * 
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *)__pyx_v_self->__pyx_vtab)->get_fragment_name(__pyx_v_self, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 248, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->_name);
  __Pyx_DECREF(__pyx_v_self->_name);
  __pyx_v_self->_name = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "glycopeptidepy/_c/structure/fragment.pyx":249
 *             self.mass += chemical_shift.mass
 *         self._name = self.get_fragment_name()
 *         self._hash = hash(self._name)             # <<<<<<<<<<<<<<
 * 
 *     cdef void _update_hash_name(self):
 */
  __pyx_t_1 = __pyx_v_self->_name;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_8 = PyObject_Hash(__pyx_t_1); if (unlikely(__pyx_t_8 == ((Py_hash_t)-1))) __PYX_ERR(0, 249, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_self->_hash = __pyx_t_8;

  /* "glycopeptidepy/_c/structure/fragment.pyx":237
 *         return self._chemical_shift
 * 
 *     cpdef set_chemical_shift(self, ChemicalShiftBase chemical_shift):             # <<<<<<<<<<<<<<
 *         '''Sets the chemical shift associated with the fragment, updating
 *         the :attr:`mass` and :attr:`chemical_shift`.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment.FragmentBase.set_chemical_shift", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_13set_chemical_shift(PyObject *__pyx_v_self, PyObject *__pyx_v_chemical_shift); /*proto*/
static char __pyx_doc_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_12set_chemical_shift[] = "FragmentBase.set_chemical_shift(self, ChemicalShiftBase chemical_shift)\nSets the chemical shift associated with the fragment, updating\n        the :attr:`mass` and :attr:`chemical_shift`.\n        ";
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_13set_chemical_shift(PyObject *__pyx_v_self, PyObject *__pyx_v_chemical_shift) {
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_chemical_shift (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_chemical_shift), __pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase, 1, "chemical_shift", 0))) __PYX_ERR(0, 237, __pyx_L1_error)
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_12set_chemical_shift(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *)__pyx_v_self), ((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase *)__pyx_v_chemical_shift));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_12set_chemical_shift(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *__pyx_v_self, struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase *__pyx_v_chemical_shift) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("set_chemical_shift", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_set_chemical_shift(__pyx_v_self, __pyx_v_chemical_shift, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 237, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment.FragmentBase.set_chemical_shift", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/fragment.pyx":251
 *         self._hash = hash(self._name)
 * 
 *     cdef void _update_hash_name(self):             # <<<<<<<<<<<<<<
 *         self._name = self.get_fragment_name()
 *         self._hash = hash(self._name)
 */

static void __pyx_f_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase__update_hash_name(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_hash_t __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_update_hash_name", 0);

  /* "glycopeptidepy/_c/structure/fragment.pyx":252
 * 
 *     cdef void _update_hash_name(self):
 *         self._name = self.get_fragment_name()             # <<<<<<<<<<<<<<
 *         self._hash = hash(self._name)
 * 
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *)__pyx_v_self->__pyx_vtab)->get_fragment_name(__pyx_v_self, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 252, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->_name);
  __Pyx_DECREF(__pyx_v_self->_name);
  __pyx_v_self->_name = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "glycopeptidepy/_c/structure/fragment.pyx":253
 *     cdef void _update_hash_name(self):
 *         self._name = self.get_fragment_name()
 *         self._hash = hash(self._name)             # <<<<<<<<<<<<<<
 * 
 *     chemical_shift = property(get_chemical_shift, set_chemical_shift)
 */
  __pyx_t_1 = __pyx_v_self->_name;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = PyObject_Hash(__pyx_t_1); if (unlikely(__pyx_t_2 == ((Py_hash_t)-1))) __PYX_ERR(0, 253, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_self->_hash = __pyx_t_2;

  /* "glycopeptidepy/_c/structure/fragment.pyx":251
 *         self._hash = hash(self._name)
 * 
 *     cdef void _update_hash_name(self):             # <<<<<<<<<<<<<<
 *         self._name = self.get_fragment_name()
 *         self._hash = hash(self._name)
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_WriteUnraisable("glycopeptidepy._c.structure.fragment.FragmentBase._update_hash_name", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "glycopeptidepy/_c/structure/fragment.pyx":257
 *     chemical_shift = property(get_chemical_shift, set_chemical_shift)
 * 
 *     cpdef str get_fragment_name(self):             # <<<<<<<<<<<<<<
 *         parts = [self._name]
 *         chemical_shift = self.get_chemical_shift()
 */

static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_15get_fragment_name(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_get_fragment_name(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *__pyx_v_self, int __pyx_skip_dispatch) {
  PyObject *__pyx_v_parts = NULL;
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase *__pyx_v_chemical_shift = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_fragment_name", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_get_fragment_name); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 257, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_15get_fragment_name)) {
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 257, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (!(likely(PyString_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "str", Py_TYPE(__pyx_t_2)->tp_name), 0))) __PYX_ERR(0, 257, __pyx_L1_error)
        __pyx_r = ((PyObject*)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "glycopeptidepy/_c/structure/fragment.pyx":258
 * 
 *     cpdef str get_fragment_name(self):
 *         parts = [self._name]             # <<<<<<<<<<<<<<
 *         chemical_shift = self.get_chemical_shift()
 *         if chemical_shift is not None:
 */
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 258, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_self->_name);
  __Pyx_GIVEREF(__pyx_v_self->_name);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_v_self->_name);
  __pyx_v_parts = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "glycopeptidepy/_c/structure/fragment.pyx":259
 *     cpdef str get_fragment_name(self):
 *         parts = [self._name]
 *         chemical_shift = self.get_chemical_shift()             # <<<<<<<<<<<<<<
 *         if chemical_shift is not None:
 *             parts.append(chemical_shift.name)
 */
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *)__pyx_v_self->__pyx_vtab)->get_chemical_shift(__pyx_v_self, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 259, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_chemical_shift = ((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "glycopeptidepy/_c/structure/fragment.pyx":260
 *         parts = [self._name]
 *         chemical_shift = self.get_chemical_shift()
 *         if chemical_shift is not None:             # <<<<<<<<<<<<<<
 *             parts.append(chemical_shift.name)
 *         return ''.join(parts)
 */
  __pyx_t_5 = (((PyObject *)__pyx_v_chemical_shift) != Py_None);
  __pyx_t_6 = (__pyx_t_5 != 0);
  if (__pyx_t_6) {

    /* "glycopeptidepy/_c/structure/fragment.pyx":261
 *         chemical_shift = self.get_chemical_shift()
 *         if chemical_shift is not None:
 *             parts.append(chemical_shift.name)             # <<<<<<<<<<<<<<
 *         return ''.join(parts)
 * 
 */
    __pyx_t_1 = __pyx_v_chemical_shift->name;
    __Pyx_INCREF(__pyx_t_1);
    __pyx_t_7 = __Pyx_PyList_Append(__pyx_v_parts, __pyx_t_1); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(0, 261, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "glycopeptidepy/_c/structure/fragment.pyx":260
 *         parts = [self._name]
 *         chemical_shift = self.get_chemical_shift()
 *         if chemical_shift is not None:             # <<<<<<<<<<<<<<
 *             parts.append(chemical_shift.name)
 *         return ''.join(parts)
 */
  }

  /* "glycopeptidepy/_c/structure/fragment.pyx":262
 *         if chemical_shift is not None:
 *             parts.append(chemical_shift.name)
 *         return ''.join(parts)             # <<<<<<<<<<<<<<
 * 
 *     cpdef str base_name(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyString_Join(__pyx_kp_s_, __pyx_v_parts); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 262, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyString_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "str", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(0, 262, __pyx_L1_error)
  __pyx_r = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "glycopeptidepy/_c/structure/fragment.pyx":257
 *     chemical_shift = property(get_chemical_shift, set_chemical_shift)
 * 
 *     cpdef str get_fragment_name(self):             # <<<<<<<<<<<<<<
 *         parts = [self._name]
 *         chemical_shift = self.get_chemical_shift()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment.FragmentBase.get_fragment_name", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_parts);
  __Pyx_XDECREF((PyObject *)__pyx_v_chemical_shift);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_15get_fragment_name(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_14get_fragment_name[] = "FragmentBase.get_fragment_name(self) -> str";
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_15get_fragment_name(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_fragment_name (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_14get_fragment_name(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_14get_fragment_name(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_fragment_name", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_get_fragment_name(__pyx_v_self, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 257, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment.FragmentBase.get_fragment_name", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/fragment.pyx":264
 *         return ''.join(parts)
 * 
 *     cpdef str base_name(self):             # <<<<<<<<<<<<<<
 *         """Simply return string like b2, y3 with no modification information."""
 *         return self._name
 */

static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_17base_name(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_base_name(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *__pyx_v_self, int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("base_name", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_base_name); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 264, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_17base_name)) {
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 264, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (!(likely(PyString_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "str", Py_TYPE(__pyx_t_2)->tp_name), 0))) __PYX_ERR(0, 264, __pyx_L1_error)
        __pyx_r = ((PyObject*)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "glycopeptidepy/_c/structure/fragment.pyx":266
 *     cpdef str base_name(self):
 *         """Simply return string like b2, y3 with no modification information."""
 *         return self._name             # <<<<<<<<<<<<<<
 * 
 *     cdef str get_name(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->_name);
  __pyx_r = __pyx_v_self->_name;
  goto __pyx_L0;

  /* "glycopeptidepy/_c/structure/fragment.pyx":264
 *         return ''.join(parts)
 * 
 *     cpdef str base_name(self):             # <<<<<<<<<<<<<<
 *         """Simply return string like b2, y3 with no modification information."""
 *         return self._name
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment.FragmentBase.base_name", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_17base_name(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_16base_name[] = "FragmentBase.base_name(self) -> str\nSimply return string like b2, y3 with no modification information.";
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_17base_name(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("base_name (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_16base_name(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_16base_name(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("base_name", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_base_name(__pyx_v_self, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 264, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment.FragmentBase.base_name", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/fragment.pyx":268
 *         return self._name
 * 
 *     cdef str get_name(self):             # <<<<<<<<<<<<<<
 *         if self._name is None:
 *             self._name = self.get_fragment_name()
 */

static PyObject *__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_get_name(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_name", 0);

  /* "glycopeptidepy/_c/structure/fragment.pyx":269
 * 
 *     cdef str get_name(self):
 *         if self._name is None:             # <<<<<<<<<<<<<<
 *             self._name = self.get_fragment_name()
 *         return self._name
 */
  __pyx_t_1 = (__pyx_v_self->_name == ((PyObject*)Py_None));
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "glycopeptidepy/_c/structure/fragment.pyx":270
 *     cdef str get_name(self):
 *         if self._name is None:
 *             self._name = self.get_fragment_name()             # <<<<<<<<<<<<<<
 *         return self._name
 * 
 */
    __pyx_t_3 = ((struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *)__pyx_v_self->__pyx_vtab)->get_fragment_name(__pyx_v_self, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 270, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_3);
    __Pyx_GOTREF(__pyx_v_self->_name);
    __Pyx_DECREF(__pyx_v_self->_name);
    __pyx_v_self->_name = ((PyObject*)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "glycopeptidepy/_c/structure/fragment.pyx":269
 * 
 *     cdef str get_name(self):
 *         if self._name is None:             # <<<<<<<<<<<<<<
 *             self._name = self.get_fragment_name()
 *         return self._name
 */
  }

  /* "glycopeptidepy/_c/structure/fragment.pyx":271
 *         if self._name is None:
 *             self._name = self.get_fragment_name()
 *         return self._name             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->_name);
  __pyx_r = __pyx_v_self->_name;
  goto __pyx_L0;

  /* "glycopeptidepy/_c/structure/fragment.pyx":268
 *         return self._name
 * 
 *     cdef str get_name(self):             # <<<<<<<<<<<<<<
 *         if self._name is None:
 *             self._name = self.get_fragment_name()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment.FragmentBase.get_name", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/fragment.pyx":274
 * 
 *     @property
 *     def name(self):             # <<<<<<<<<<<<<<
 *         return self.get_name()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_4name_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_4name_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_4name___get__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_4name___get__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "glycopeptidepy/_c/structure/fragment.pyx":275
 *     @property
 *     def name(self):
 *         return self.get_name()             # <<<<<<<<<<<<<<
 * 
 *     @name.setter
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *)__pyx_v_self->__pyx_vtab)->get_name(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 275, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "glycopeptidepy/_c/structure/fragment.pyx":274
 * 
 *     @property
 *     def name(self):             # <<<<<<<<<<<<<<
 *         return self.get_name()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment.FragmentBase.name.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/fragment.pyx":278
 * 
 *     @name.setter
 *     def name(self, name):             # <<<<<<<<<<<<<<
 *         self._name = name
 * 
 */

/* Python wrapper */
static int __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_4name_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_name); /*proto*/
static int __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_4name_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_name) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_4name_2__set__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *)__pyx_v_self), ((PyObject *)__pyx_v_name));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_4name_2__set__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *__pyx_v_self, PyObject *__pyx_v_name) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);

  /* "glycopeptidepy/_c/structure/fragment.pyx":279
 *     @name.setter
 *     def name(self, name):
 *         self._name = name             # <<<<<<<<<<<<<<
 * 
 * 
 */
  if (!(likely(PyString_CheckExact(__pyx_v_name))||((__pyx_v_name) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "str", Py_TYPE(__pyx_v_name)->tp_name), 0))) __PYX_ERR(0, 279, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_name;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->_name);
  __Pyx_DECREF(__pyx_v_self->_name);
  __pyx_v_self->_name = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "glycopeptidepy/_c/structure/fragment.pyx":278
 * 
 *     @name.setter
 *     def name(self, name):             # <<<<<<<<<<<<<<
 *         self._name = name
 * 
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment.FragmentBase.name.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/fragment.pxd":44
 * 
 *     cdef:
 *         public str _name             # <<<<<<<<<<<<<<
 *         public Py_hash_t _hash
 *         public double mass
 */

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_5_name_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_5_name_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_5_name___get__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_5_name___get__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->_name);
  __pyx_r = __pyx_v_self->_name;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_5_name_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_5_name_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_5_name_2__set__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_5_name_2__set__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);
  if (!(likely(PyString_CheckExact(__pyx_v_value))||((__pyx_v_value) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "str", Py_TYPE(__pyx_v_value)->tp_name), 0))) __PYX_ERR(1, 44, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_value;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->_name);
  __Pyx_DECREF(__pyx_v_self->_name);
  __pyx_v_self->_name = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment.FragmentBase._name.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_5_name_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_5_name_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_5_name_4__del__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_5_name_4__del__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->_name);
  __Pyx_DECREF(__pyx_v_self->_name);
  __pyx_v_self->_name = ((PyObject*)Py_None);

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/fragment.pxd":45
 *     cdef:
 *         public str _name
 *         public Py_hash_t _hash             # <<<<<<<<<<<<<<
 *         public double mass
 *         public ChemicalShiftBase _chemical_shift
 */

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_5_hash_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_5_hash_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_5_hash___get__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_5_hash___get__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_FromHash_t(__pyx_v_self->_hash); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 45, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment.FragmentBase._hash.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_5_hash_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_5_hash_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_5_hash_2__set__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_5_hash_2__set__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  Py_hash_t __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyInt_AsHash_t(__pyx_v_value); if (unlikely((__pyx_t_1 == (Py_hash_t)-1) && PyErr_Occurred())) __PYX_ERR(1, 45, __pyx_L1_error)
  __pyx_v_self->_hash = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment.FragmentBase._hash.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/fragment.pxd":46
 *         public str _name
 *         public Py_hash_t _hash
 *         public double mass             # <<<<<<<<<<<<<<
 *         public ChemicalShiftBase _chemical_shift
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_4mass_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_4mass_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_4mass___get__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_4mass___get__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->mass); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 46, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment.FragmentBase.mass.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_4mass_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_4mass_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_4mass_2__set__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_4mass_2__set__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  double __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_value); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 46, __pyx_L1_error)
  __pyx_v_self->mass = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment.FragmentBase.mass.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/fragment.pxd":47
 *         public Py_hash_t _hash
 *         public double mass
 *         public ChemicalShiftBase _chemical_shift             # <<<<<<<<<<<<<<
 * 
 *     cpdef IonSeriesBase get_series(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_15_chemical_shift_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_15_chemical_shift_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_15_chemical_shift___get__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_15_chemical_shift___get__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->_chemical_shift));
  __pyx_r = ((PyObject *)__pyx_v_self->_chemical_shift);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_15_chemical_shift_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_15_chemical_shift_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_15_chemical_shift_2__set__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_15_chemical_shift_2__set__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);
  if (!(likely(((__pyx_v_value) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_value, __pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase))))) __PYX_ERR(1, 47, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_value;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->_chemical_shift);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->_chemical_shift));
  __pyx_v_self->_chemical_shift = ((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment.FragmentBase._chemical_shift.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_15_chemical_shift_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_15_chemical_shift_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_15_chemical_shift_4__del__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_15_chemical_shift_4__del__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->_chemical_shift);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->_chemical_shift));
  __pyx_v_self->_chemical_shift = ((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase *)Py_None);

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_19__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_18__reduce_cython__[] = "FragmentBase.__reduce_cython__(self)";
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_19__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_18__reduce_cython__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_18__reduce_cython__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self._chemical_shift, self._hash, self._name, self.mass)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = __Pyx_PyInt_FromHash_t(__pyx_v_self->_hash); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_self->mass); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(4); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->_chemical_shift));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self->_chemical_shift));
  PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)__pyx_v_self->_chemical_shift));
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1);
  __Pyx_INCREF(__pyx_v_self->_name);
  __Pyx_GIVEREF(__pyx_v_self->_name);
  PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_v_self->_name);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 3, __pyx_t_2);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self._chemical_shift, self._hash, self._name, self.mass)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_3 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v__dict = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "(tree fragment)":7
 *     state = (self._chemical_shift, self._hash, self._name, self.mass)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_4 = (__pyx_v__dict != Py_None);
  __pyx_t_5 = (__pyx_t_4 != 0);
  if (__pyx_t_5) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v__dict);
    __pyx_t_2 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_2));
    __pyx_t_2 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self._chemical_shift is not None or self._name is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self._chemical_shift, self._hash, self._name, self.mass)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self._chemical_shift is not None or self._name is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_FragmentBase, (type(self), 0x033a8ac, None), state
 */
  /*else*/ {
    __pyx_t_4 = (((PyObject *)__pyx_v_self->_chemical_shift) != Py_None);
    __pyx_t_6 = (__pyx_t_4 != 0);
    if (!__pyx_t_6) {
    } else {
      __pyx_t_5 = __pyx_t_6;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_6 = (__pyx_v_self->_name != ((PyObject*)Py_None));
    __pyx_t_4 = (__pyx_t_6 != 0);
    __pyx_t_5 = __pyx_t_4;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_5;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self._chemical_shift is not None or self._name is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_FragmentBase, (type(self), 0x033a8ac, None), state
 *     else:
 */
  __pyx_t_5 = (__pyx_v_use_setstate != 0);
  if (__pyx_t_5) {

    /* "(tree fragment)":13
 *         use_setstate = self._chemical_shift is not None or self._name is not None
 *     if use_setstate:
 *         return __pyx_unpickle_FragmentBase, (type(self), 0x033a8ac, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_FragmentBase, (type(self), 0x033a8ac, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_pyx_unpickle_FragmentBase); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_3385516);
    __Pyx_GIVEREF(__pyx_int_3385516);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_int_3385516);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    PyTuple_SET_ITEM(__pyx_t_3, 2, Py_None);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_3);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state);
    __pyx_t_2 = 0;
    __pyx_t_3 = 0;
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self._chemical_shift is not None or self._name is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_FragmentBase, (type(self), 0x033a8ac, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_FragmentBase, (type(self), 0x033a8ac, None), state
 *     else:
 *         return __pyx_unpickle_FragmentBase, (type(self), 0x033a8ac, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_FragmentBase__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_pyx_unpickle_FragmentBase); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_3385516);
    __Pyx_GIVEREF(__pyx_int_3385516);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_int_3385516);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_v_state);
    __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_3);
    __pyx_t_1 = 0;
    __pyx_t_3 = 0;
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment.FragmentBase.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_FragmentBase, (type(self), 0x033a8ac, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_FragmentBase__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_21__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static char __pyx_doc_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_20__setstate_cython__[] = "FragmentBase.__setstate_cython__(self, __pyx_state)";
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_21__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_20__setstate_cython__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_20__setstate_cython__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_FragmentBase, (type(self), 0x033a8ac, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_FragmentBase__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(2, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_14glycopeptidepy_2_c_9structure_8fragment___pyx_unpickle_FragmentBase__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_FragmentBase, (type(self), 0x033a8ac, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_FragmentBase__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment.FragmentBase.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/fragment.pyx":311
 *     '''
 *     @staticmethod
 *     cdef PeptideFragment _create(IonSeriesBase kind, int position, CountTable modification_dict, double mass,             # <<<<<<<<<<<<<<
 *                                  list flanking_amino_acids=None, dict glycosylation=None,
 *                                  ChemicalShiftBase chemical_shift=None, CComposition composition=None,
 */

static struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment__create(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *__pyx_v_kind, int __pyx_v_position, struct __pyx_obj_14glycopeptidepy_2_c_11count_table_CountTable *__pyx_v_modification_dict, double __pyx_v_mass, struct __pyx_opt_args_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment__create *__pyx_optional_args) {

  /* "glycopeptidepy/_c/structure/fragment.pyx":312
 *     @staticmethod
 *     cdef PeptideFragment _create(IonSeriesBase kind, int position, CountTable modification_dict, double mass,
 *                                  list flanking_amino_acids=None, dict glycosylation=None,             # <<<<<<<<<<<<<<
 *                                  ChemicalShiftBase chemical_shift=None, CComposition composition=None,
 *                                  double* delta_mass=NULL):
 */
  PyObject *__pyx_v_flanking_amino_acids = ((PyObject*)Py_None);
  PyObject *__pyx_v_glycosylation = ((PyObject*)Py_None);

  /* "glycopeptidepy/_c/structure/fragment.pyx":313
 *     cdef PeptideFragment _create(IonSeriesBase kind, int position, CountTable modification_dict, double mass,
 *                                  list flanking_amino_acids=None, dict glycosylation=None,
 *                                  ChemicalShiftBase chemical_shift=None, CComposition composition=None,             # <<<<<<<<<<<<<<
 *                                  double* delta_mass=NULL):
 *         cdef PeptideFragment self = PeptideFragment.__new__(PeptideFragment)
 */
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase *__pyx_v_chemical_shift = ((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase *)Py_None);
  struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *__pyx_v_composition = ((struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *)Py_None);

  /* "glycopeptidepy/_c/structure/fragment.pyx":314
 *                                  list flanking_amino_acids=None, dict glycosylation=None,
 *                                  ChemicalShiftBase chemical_shift=None, CComposition composition=None,
 *                                  double* delta_mass=NULL):             # <<<<<<<<<<<<<<
 *         cdef PeptideFragment self = PeptideFragment.__new__(PeptideFragment)
 *         self.kind = kind
 */
  double *__pyx_v_delta_mass = ((double *)NULL);
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *__pyx_v_self = 0;
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_create", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_flanking_amino_acids = __pyx_optional_args->flanking_amino_acids;
      if (__pyx_optional_args->__pyx_n > 1) {
        __pyx_v_glycosylation = __pyx_optional_args->glycosylation;
        if (__pyx_optional_args->__pyx_n > 2) {
          __pyx_v_chemical_shift = __pyx_optional_args->chemical_shift;
          if (__pyx_optional_args->__pyx_n > 3) {
            __pyx_v_composition = __pyx_optional_args->composition;
            if (__pyx_optional_args->__pyx_n > 4) {
              __pyx_v_delta_mass = __pyx_optional_args->delta_mass;
            }
          }
        }
      }
    }
  }

  /* "glycopeptidepy/_c/structure/fragment.pyx":315
 *                                  ChemicalShiftBase chemical_shift=None, CComposition composition=None,
 *                                  double* delta_mass=NULL):
 *         cdef PeptideFragment self = PeptideFragment.__new__(PeptideFragment)             # <<<<<<<<<<<<<<
 *         self.kind = kind
 *         self.position = position
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment(((PyTypeObject *)__pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 315, __pyx_L1_error)
  __Pyx_GOTREF(((PyObject *)__pyx_t_1));
  __pyx_v_self = ((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "glycopeptidepy/_c/structure/fragment.pyx":316
 *                                  double* delta_mass=NULL):
 *         cdef PeptideFragment self = PeptideFragment.__new__(PeptideFragment)
 *         self.kind = kind             # <<<<<<<<<<<<<<
 *         self.position = position
 * 
 */
  __Pyx_INCREF(((PyObject *)__pyx_v_kind));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_kind));
  __Pyx_GOTREF(__pyx_v_self->kind);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->kind));
  __pyx_v_self->kind = __pyx_v_kind;

  /* "glycopeptidepy/_c/structure/fragment.pyx":317
 *         cdef PeptideFragment self = PeptideFragment.__new__(PeptideFragment)
 *         self.kind = kind
 *         self.position = position             # <<<<<<<<<<<<<<
 * 
 *         self.bare_mass = mass
 */
  __pyx_v_self->position = __pyx_v_position;

  /* "glycopeptidepy/_c/structure/fragment.pyx":319
 *         self.position = position
 * 
 *         self.bare_mass = mass             # <<<<<<<<<<<<<<
 *         self.mass = mass
 *         self.modification_dict = modification_dict
 */
  __pyx_v_self->bare_mass = __pyx_v_mass;

  /* "glycopeptidepy/_c/structure/fragment.pyx":320
 * 
 *         self.bare_mass = mass
 *         self.mass = mass             # <<<<<<<<<<<<<<
 *         self.modification_dict = modification_dict
 *         self.composition = composition
 */
  __pyx_v_self->__pyx_base.mass = __pyx_v_mass;

  /* "glycopeptidepy/_c/structure/fragment.pyx":321
 *         self.bare_mass = mass
 *         self.mass = mass
 *         self.modification_dict = modification_dict             # <<<<<<<<<<<<<<
 *         self.composition = composition
 *         self._chemical_shift = None
 */
  __Pyx_INCREF(((PyObject *)__pyx_v_modification_dict));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_modification_dict));
  __Pyx_GOTREF(__pyx_v_self->modification_dict);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->modification_dict));
  __pyx_v_self->modification_dict = __pyx_v_modification_dict;

  /* "glycopeptidepy/_c/structure/fragment.pyx":322
 *         self.mass = mass
 *         self.modification_dict = modification_dict
 *         self.composition = composition             # <<<<<<<<<<<<<<
 *         self._chemical_shift = None
 * 
 */
  __Pyx_INCREF(((PyObject *)__pyx_v_composition));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_composition));
  __Pyx_GOTREF(__pyx_v_self->composition);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->composition));
  __pyx_v_self->composition = __pyx_v_composition;

  /* "glycopeptidepy/_c/structure/fragment.pyx":323
 *         self.modification_dict = modification_dict
 *         self.composition = composition
 *         self._chemical_shift = None             # <<<<<<<<<<<<<<
 * 
 *         self.flanking_amino_acids = flanking_amino_acids
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->__pyx_base._chemical_shift);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->__pyx_base._chemical_shift));
  __pyx_v_self->__pyx_base._chemical_shift = ((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase *)Py_None);

  /* "glycopeptidepy/_c/structure/fragment.pyx":325
 *         self._chemical_shift = None
 * 
 *         self.flanking_amino_acids = flanking_amino_acids             # <<<<<<<<<<<<<<
 *         self.glycosylation = glycosylation
 *         if chemical_shift is not None:
 */
  __Pyx_INCREF(__pyx_v_flanking_amino_acids);
  __Pyx_GIVEREF(__pyx_v_flanking_amino_acids);
  __Pyx_GOTREF(__pyx_v_self->flanking_amino_acids);
  __Pyx_DECREF(__pyx_v_self->flanking_amino_acids);
  __pyx_v_self->flanking_amino_acids = __pyx_v_flanking_amino_acids;

  /* "glycopeptidepy/_c/structure/fragment.pyx":326
 * 
 *         self.flanking_amino_acids = flanking_amino_acids
 *         self.glycosylation = glycosylation             # <<<<<<<<<<<<<<
 *         if chemical_shift is not None:
 *             self.set_chemical_shift(chemical_shift)
 */
  __Pyx_INCREF(__pyx_v_glycosylation);
  __Pyx_GIVEREF(__pyx_v_glycosylation);
  __Pyx_GOTREF(__pyx_v_self->glycosylation);
  __Pyx_DECREF(__pyx_v_self->glycosylation);
  __pyx_v_self->glycosylation = __pyx_v_glycosylation;

  /* "glycopeptidepy/_c/structure/fragment.pyx":327
 *         self.flanking_amino_acids = flanking_amino_acids
 *         self.glycosylation = glycosylation
 *         if chemical_shift is not None:             # <<<<<<<<<<<<<<
 *             self.set_chemical_shift(chemical_shift)
 * 
 */
  __pyx_t_2 = (((PyObject *)__pyx_v_chemical_shift) != Py_None);
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {

    /* "glycopeptidepy/_c/structure/fragment.pyx":328
 *         self.glycosylation = glycosylation
 *         if chemical_shift is not None:
 *             self.set_chemical_shift(chemical_shift)             # <<<<<<<<<<<<<<
 * 
 *         if delta_mass == NULL:
 */
    __pyx_t_1 = ((struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.set_chemical_shift(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *)__pyx_v_self), __pyx_v_chemical_shift, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 328, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "glycopeptidepy/_c/structure/fragment.pyx":327
 *         self.flanking_amino_acids = flanking_amino_acids
 *         self.glycosylation = glycosylation
 *         if chemical_shift is not None:             # <<<<<<<<<<<<<<
 *             self.set_chemical_shift(chemical_shift)
 * 
 */
  }

  /* "glycopeptidepy/_c/structure/fragment.pyx":330
 *             self.set_chemical_shift(chemical_shift)
 * 
 *         if delta_mass == NULL:             # <<<<<<<<<<<<<<
 *             self._update_mass_with_modifications()
 *         else:
 */
  __pyx_t_3 = ((__pyx_v_delta_mass == NULL) != 0);
  if (__pyx_t_3) {

    /* "glycopeptidepy/_c/structure/fragment.pyx":331
 * 
 *         if delta_mass == NULL:
 *             self._update_mass_with_modifications()             # <<<<<<<<<<<<<<
 *         else:
 *             self.mass += delta_mass[0]
 */
    ((struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *)__pyx_v_self->__pyx_base.__pyx_vtab)->_update_mass_with_modifications(__pyx_v_self);

    /* "glycopeptidepy/_c/structure/fragment.pyx":330
 *             self.set_chemical_shift(chemical_shift)
 * 
 *         if delta_mass == NULL:             # <<<<<<<<<<<<<<
 *             self._update_mass_with_modifications()
 *         else:
 */
    goto __pyx_L4;
  }

  /* "glycopeptidepy/_c/structure/fragment.pyx":333
 *             self._update_mass_with_modifications()
 *         else:
 *             self.mass += delta_mass[0]             # <<<<<<<<<<<<<<
 *         # if delta_mass != NULL:
 *         #     assert abs((self.mass - self.bare_mass) - delta_mass[0]) < 1e-3, ((self.mass - self.bare_mass), delta_mass[0])
 */
  /*else*/ {
    __pyx_v_self->__pyx_base.mass = (__pyx_v_self->__pyx_base.mass + (__pyx_v_delta_mass[0]));
  }
  __pyx_L4:;

  /* "glycopeptidepy/_c/structure/fragment.pyx":336
 *         # if delta_mass != NULL:
 *         #     assert abs((self.mass - self.bare_mass) - delta_mass[0]) < 1e-3, ((self.mass - self.bare_mass), delta_mass[0])
 *         self._name = None             # <<<<<<<<<<<<<<
 *         self._hash = -1
 *         return self
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->__pyx_base._name);
  __Pyx_DECREF(__pyx_v_self->__pyx_base._name);
  __pyx_v_self->__pyx_base._name = ((PyObject*)Py_None);

  /* "glycopeptidepy/_c/structure/fragment.pyx":337
 *         #     assert abs((self.mass - self.bare_mass) - delta_mass[0]) < 1e-3, ((self.mass - self.bare_mass), delta_mass[0])
 *         self._name = None
 *         self._hash = -1             # <<<<<<<<<<<<<<
 *         return self
 * 
 */
  __pyx_v_self->__pyx_base._hash = -1L;

  /* "glycopeptidepy/_c/structure/fragment.pyx":338
 *         self._name = None
 *         self._hash = -1
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, kind, position, modification_dict, mass, flanking_amino_acids=None,
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __Pyx_INCREF(((PyObject *)__pyx_v_self));
  __pyx_r = __pyx_v_self;
  goto __pyx_L0;

  /* "glycopeptidepy/_c/structure/fragment.pyx":311
 *     '''
 *     @staticmethod
 *     cdef PeptideFragment _create(IonSeriesBase kind, int position, CountTable modification_dict, double mass,             # <<<<<<<<<<<<<<
 *                                  list flanking_amino_acids=None, dict glycosylation=None,
 *                                  ChemicalShiftBase chemical_shift=None, CComposition composition=None,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment.PeptideFragment._create", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_self);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/fragment.pyx":340
 *         return self
 * 
 *     def __init__(self, kind, position, modification_dict, mass, flanking_amino_acids=None,             # <<<<<<<<<<<<<<
 *                  glycosylation=None, chemical_shift=None, composition=None):
 *         self.kind = kind
 */

/* Python wrapper */
static int __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_kind = 0;
  PyObject *__pyx_v_position = 0;
  PyObject *__pyx_v_modification_dict = 0;
  PyObject *__pyx_v_mass = 0;
  PyObject *__pyx_v_flanking_amino_acids = 0;
  PyObject *__pyx_v_glycosylation = 0;
  PyObject *__pyx_v_chemical_shift = 0;
  PyObject *__pyx_v_composition = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_kind,&__pyx_n_s_position,&__pyx_n_s_modification_dict,&__pyx_n_s_mass,&__pyx_n_s_flanking_amino_acids,&__pyx_n_s_glycosylation,&__pyx_n_s_chemical_shift,&__pyx_n_s_composition,0};
    PyObject* values[8] = {0,0,0,0,0,0,0,0};
    values[4] = ((PyObject *)Py_None);

    /* "glycopeptidepy/_c/structure/fragment.pyx":341
 * 
 *     def __init__(self, kind, position, modification_dict, mass, flanking_amino_acids=None,
 *                  glycosylation=None, chemical_shift=None, composition=None):             # <<<<<<<<<<<<<<
 *         self.kind = kind
 * 
 */
    values[5] = ((PyObject *)Py_None);
    values[6] = ((PyObject *)Py_None);
    values[7] = ((PyObject *)Py_None);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_kind)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_position)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 4, 8, 1); __PYX_ERR(0, 340, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_modification_dict)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 4, 8, 2); __PYX_ERR(0, 340, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_mass)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 4, 8, 3); __PYX_ERR(0, 340, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_flanking_amino_acids);
          if (value) { values[4] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_glycosylation);
          if (value) { values[5] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_chemical_shift);
          if (value) { values[6] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_composition);
          if (value) { values[7] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 340, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_kind = values[0];
    __pyx_v_position = values[1];
    __pyx_v_modification_dict = values[2];
    __pyx_v_mass = values[3];
    __pyx_v_flanking_amino_acids = values[4];
    __pyx_v_glycosylation = values[5];
    __pyx_v_chemical_shift = values[6];
    __pyx_v_composition = values[7];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 4, 8, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 340, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment.PeptideFragment.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment___init__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *)__pyx_v_self), __pyx_v_kind, __pyx_v_position, __pyx_v_modification_dict, __pyx_v_mass, __pyx_v_flanking_amino_acids, __pyx_v_glycosylation, __pyx_v_chemical_shift, __pyx_v_composition);

  /* "glycopeptidepy/_c/structure/fragment.pyx":340
 *         return self
 * 
 *     def __init__(self, kind, position, modification_dict, mass, flanking_amino_acids=None,             # <<<<<<<<<<<<<<
 *                  glycosylation=None, chemical_shift=None, composition=None):
 *         self.kind = kind
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment___init__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *__pyx_v_self, PyObject *__pyx_v_kind, PyObject *__pyx_v_position, PyObject *__pyx_v_modification_dict, PyObject *__pyx_v_mass, PyObject *__pyx_v_flanking_amino_acids, PyObject *__pyx_v_glycosylation, PyObject *__pyx_v_chemical_shift, PyObject *__pyx_v_composition) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  double __pyx_t_2;
  int __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "glycopeptidepy/_c/structure/fragment.pyx":342
 *     def __init__(self, kind, position, modification_dict, mass, flanking_amino_acids=None,
 *                  glycosylation=None, chemical_shift=None, composition=None):
 *         self.kind = kind             # <<<<<<<<<<<<<<
 * 
 *         # The mass value is the bare backbone's mass
 */
  if (!(likely(((__pyx_v_kind) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_kind, __pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase))))) __PYX_ERR(0, 342, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_kind;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->kind);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->kind));
  __pyx_v_self->kind = ((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "glycopeptidepy/_c/structure/fragment.pyx":345
 * 
 *         # The mass value is the bare backbone's mass
 *         self.bare_mass = PyFloat_AsDouble(mass)             # <<<<<<<<<<<<<<
 *         self.modification_dict = modification_dict
 *         self.mass = self.bare_mass
 */
  __pyx_t_2 = PyFloat_AsDouble(__pyx_v_mass); if (unlikely(__pyx_t_2 == ((double)-1.0) && PyErr_Occurred())) __PYX_ERR(0, 345, __pyx_L1_error)
  __pyx_v_self->bare_mass = __pyx_t_2;

  /* "glycopeptidepy/_c/structure/fragment.pyx":346
 *         # The mass value is the bare backbone's mass
 *         self.bare_mass = PyFloat_AsDouble(mass)
 *         self.modification_dict = modification_dict             # <<<<<<<<<<<<<<
 *         self.mass = self.bare_mass
 *         self.composition = composition
 */
  if (!(likely(((__pyx_v_modification_dict) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_modification_dict, __pyx_ptype_14glycopeptidepy_2_c_11count_table_CountTable))))) __PYX_ERR(0, 346, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_modification_dict;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->modification_dict);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->modification_dict));
  __pyx_v_self->modification_dict = ((struct __pyx_obj_14glycopeptidepy_2_c_11count_table_CountTable *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "glycopeptidepy/_c/structure/fragment.pyx":347
 *         self.bare_mass = PyFloat_AsDouble(mass)
 *         self.modification_dict = modification_dict
 *         self.mass = self.bare_mass             # <<<<<<<<<<<<<<
 *         self.composition = composition
 *         self._chemical_shift = None
 */
  __pyx_t_2 = __pyx_v_self->bare_mass;
  __pyx_v_self->__pyx_base.mass = __pyx_t_2;

  /* "glycopeptidepy/_c/structure/fragment.pyx":348
 *         self.modification_dict = modification_dict
 *         self.mass = self.bare_mass
 *         self.composition = composition             # <<<<<<<<<<<<<<
 *         self._chemical_shift = None
 * 
 */
  if (!(likely(((__pyx_v_composition) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_composition, __pyx_ptype_5glypy_11composition_12ccomposition_CComposition))))) __PYX_ERR(0, 348, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_composition;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->composition);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->composition));
  __pyx_v_self->composition = ((struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "glycopeptidepy/_c/structure/fragment.pyx":349
 *         self.mass = self.bare_mass
 *         self.composition = composition
 *         self._chemical_shift = None             # <<<<<<<<<<<<<<
 * 
 *         self.flanking_amino_acids = flanking_amino_acids
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->__pyx_base._chemical_shift);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->__pyx_base._chemical_shift));
  __pyx_v_self->__pyx_base._chemical_shift = ((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase *)Py_None);

  /* "glycopeptidepy/_c/structure/fragment.pyx":351
 *         self._chemical_shift = None
 * 
 *         self.flanking_amino_acids = flanking_amino_acids             # <<<<<<<<<<<<<<
 *         self.position = position
 * 
 */
  if (!(likely(PyList_CheckExact(__pyx_v_flanking_amino_acids))||((__pyx_v_flanking_amino_acids) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_v_flanking_amino_acids)->tp_name), 0))) __PYX_ERR(0, 351, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_flanking_amino_acids;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->flanking_amino_acids);
  __Pyx_DECREF(__pyx_v_self->flanking_amino_acids);
  __pyx_v_self->flanking_amino_acids = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "glycopeptidepy/_c/structure/fragment.pyx":352
 * 
 *         self.flanking_amino_acids = flanking_amino_acids
 *         self.position = position             # <<<<<<<<<<<<<<
 * 
 *         self.glycosylation = glycosylation
 */
  __pyx_t_3 = __Pyx_PyInt_As_int(__pyx_v_position); if (unlikely((__pyx_t_3 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 352, __pyx_L1_error)
  __pyx_v_self->position = __pyx_t_3;

  /* "glycopeptidepy/_c/structure/fragment.pyx":354
 *         self.position = position
 * 
 *         self.glycosylation = glycosylation             # <<<<<<<<<<<<<<
 *         self.set_chemical_shift(chemical_shift)
 *         self._update_mass_with_modifications()
 */
  if (!(likely(PyDict_CheckExact(__pyx_v_glycosylation))||((__pyx_v_glycosylation) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "dict", Py_TYPE(__pyx_v_glycosylation)->tp_name), 0))) __PYX_ERR(0, 354, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_glycosylation;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->glycosylation);
  __Pyx_DECREF(__pyx_v_self->glycosylation);
  __pyx_v_self->glycosylation = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "glycopeptidepy/_c/structure/fragment.pyx":355
 * 
 *         self.glycosylation = glycosylation
 *         self.set_chemical_shift(chemical_shift)             # <<<<<<<<<<<<<<
 *         self._update_mass_with_modifications()
 * 
 */
  if (!(likely(((__pyx_v_chemical_shift) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_chemical_shift, __pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase))))) __PYX_ERR(0, 355, __pyx_L1_error)
  __pyx_t_1 = ((struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.set_chemical_shift(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *)__pyx_v_self), ((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase *)__pyx_v_chemical_shift), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 355, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "glycopeptidepy/_c/structure/fragment.pyx":356
 *         self.glycosylation = glycosylation
 *         self.set_chemical_shift(chemical_shift)
 *         self._update_mass_with_modifications()             # <<<<<<<<<<<<<<
 * 
 *         self._name = None
 */
  ((struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *)__pyx_v_self->__pyx_base.__pyx_vtab)->_update_mass_with_modifications(__pyx_v_self);

  /* "glycopeptidepy/_c/structure/fragment.pyx":358
 *         self._update_mass_with_modifications()
 * 
 *         self._name = None             # <<<<<<<<<<<<<<
 *         self._hash = -1
 * 
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->__pyx_base._name);
  __Pyx_DECREF(__pyx_v_self->__pyx_base._name);
  __pyx_v_self->__pyx_base._name = ((PyObject*)Py_None);

  /* "glycopeptidepy/_c/structure/fragment.pyx":359
 * 
 *         self._name = None
 *         self._hash = -1             # <<<<<<<<<<<<<<
 * 
 *     cdef void _update_mass_with_modifications(self):
 */
  __pyx_v_self->__pyx_base._hash = -1L;

  /* "glycopeptidepy/_c/structure/fragment.pyx":340
 *         return self
 * 
 *     def __init__(self, kind, position, modification_dict, mass, flanking_amino_acids=None,             # <<<<<<<<<<<<<<
 *                  glycosylation=None, chemical_shift=None, composition=None):
 *         self.kind = kind
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment.PeptideFragment.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/fragment.pyx":361
 *         self._hash = -1
 * 
 *     cdef void _update_mass_with_modifications(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             ModificationBase mod
 */

static void __pyx_f_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment__update_mass_with_modifications(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *__pyx_v_self) {
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_ModificationBase *__pyx_v_mod = 0;
  struct __pyx_obj_14glycopeptidepy_2_c_11count_table_CountTable *__pyx_v_modification_dict = 0;
  struct __pyx_obj_14glycopeptidepy_2_c_11count_table_CountTableIterator *__pyx_v_modification_iterator = 0;
  PyObject *__pyx_v_pkey;
  long __pyx_v_count;
  Py_ssize_t __pyx_v_ppos;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_update_mass_with_modifications", 0);

  /* "glycopeptidepy/_c/structure/fragment.pyx":368
 *             PyObject *pkey
 *             long count
 *             Py_ssize_t ppos = 0             # <<<<<<<<<<<<<<
 * 
 *         modification_dict = self.modification_dict
 */
  __pyx_v_ppos = 0;

  /* "glycopeptidepy/_c/structure/fragment.pyx":370
 *             Py_ssize_t ppos = 0
 * 
 *         modification_dict = self.modification_dict             # <<<<<<<<<<<<<<
 *         modification_iterator = CountTableIterator._create(modification_dict)
 * 
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_self->modification_dict);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_modification_dict = ((struct __pyx_obj_14glycopeptidepy_2_c_11count_table_CountTable *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "glycopeptidepy/_c/structure/fragment.pyx":371
 * 
 *         modification_dict = self.modification_dict
 *         modification_iterator = CountTableIterator._create(modification_dict)             # <<<<<<<<<<<<<<
 * 
 *         while modification_iterator.has_more():
 */
  __pyx_t_1 = ((PyObject *)__pyx_vtabptr_14glycopeptidepy_2_c_11count_table_CountTableIterator->_create(__pyx_v_modification_dict)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 371, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_modification_iterator = ((struct __pyx_obj_14glycopeptidepy_2_c_11count_table_CountTableIterator *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "glycopeptidepy/_c/structure/fragment.pyx":373
 *         modification_iterator = CountTableIterator._create(modification_dict)
 * 
 *         while modification_iterator.has_more():             # <<<<<<<<<<<<<<
 *             ppos = modification_iterator.get_next_value(&pkey, &count)
 *             if ppos != 0:
 */
  while (1) {
    __pyx_t_2 = (((struct __pyx_vtabstruct_14glycopeptidepy_2_c_11count_table_CountTableIterator *)__pyx_v_modification_iterator->__pyx_vtab)->has_more(__pyx_v_modification_iterator) != 0);
    if (!__pyx_t_2) break;

    /* "glycopeptidepy/_c/structure/fragment.pyx":374
 * 
 *         while modification_iterator.has_more():
 *             ppos = modification_iterator.get_next_value(&pkey, &count)             # <<<<<<<<<<<<<<
 *             if ppos != 0:
 *                 break
 */
    __pyx_v_ppos = ((struct __pyx_vtabstruct_14glycopeptidepy_2_c_11count_table_CountTableIterator *)__pyx_v_modification_iterator->__pyx_vtab)->get_next_value(__pyx_v_modification_iterator, (&__pyx_v_pkey), (&__pyx_v_count));

    /* "glycopeptidepy/_c/structure/fragment.pyx":375
 *         while modification_iterator.has_more():
 *             ppos = modification_iterator.get_next_value(&pkey, &count)
 *             if ppos != 0:             # <<<<<<<<<<<<<<
 *                 break
 *             mod = <ModificationBase>pkey
 */
    __pyx_t_2 = ((__pyx_v_ppos != 0) != 0);
    if (__pyx_t_2) {

      /* "glycopeptidepy/_c/structure/fragment.pyx":376
 *             ppos = modification_iterator.get_next_value(&pkey, &count)
 *             if ppos != 0:
 *                 break             # <<<<<<<<<<<<<<
 *             mod = <ModificationBase>pkey
 *             self.mass += mod.mass * count
 */
      goto __pyx_L4_break;

      /* "glycopeptidepy/_c/structure/fragment.pyx":375
 *         while modification_iterator.has_more():
 *             ppos = modification_iterator.get_next_value(&pkey, &count)
 *             if ppos != 0:             # <<<<<<<<<<<<<<
 *                 break
 *             mod = <ModificationBase>pkey
 */
    }

    /* "glycopeptidepy/_c/structure/fragment.pyx":377
 *             if ppos != 0:
 *                 break
 *             mod = <ModificationBase>pkey             # <<<<<<<<<<<<<<
 *             self.mass += mod.mass * count
 * 
 */
    __pyx_t_1 = ((PyObject *)__pyx_v_pkey);
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_mod, ((struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_ModificationBase *)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "glycopeptidepy/_c/structure/fragment.pyx":378
 *                 break
 *             mod = <ModificationBase>pkey
 *             self.mass += mod.mass * count             # <<<<<<<<<<<<<<
 * 
 *     cpdef IonSeriesBase get_series(self):
 */
    __pyx_v_self->__pyx_base.mass = (__pyx_v_self->__pyx_base.mass + (__pyx_v_mod->mass * __pyx_v_count));
  }
  __pyx_L4_break:;

  /* "glycopeptidepy/_c/structure/fragment.pyx":361
 *         self._hash = -1
 * 
 *     cdef void _update_mass_with_modifications(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             ModificationBase mod
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_WriteUnraisable("glycopeptidepy._c.structure.fragment.PeptideFragment._update_mass_with_modifications", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_mod);
  __Pyx_XDECREF((PyObject *)__pyx_v_modification_dict);
  __Pyx_XDECREF((PyObject *)__pyx_v_modification_iterator);
  __Pyx_RefNannyFinishContext();
}

/* "glycopeptidepy/_c/structure/fragment.pyx":380
 *             self.mass += mod.mass * count
 * 
 *     cpdef IonSeriesBase get_series(self):             # <<<<<<<<<<<<<<
 *         return self.kind
 * 
 */

static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_3get_series(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_get_series(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *__pyx_v_self, int __pyx_skip_dispatch) {
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_series", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_get_series); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 380, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_3get_series)) {
        __Pyx_XDECREF(((PyObject *)__pyx_r));
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 380, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase))))) __PYX_ERR(0, 380, __pyx_L1_error)
        __pyx_r = ((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "glycopeptidepy/_c/structure/fragment.pyx":381
 * 
 *     cpdef IonSeriesBase get_series(self):
 *         return self.kind             # <<<<<<<<<<<<<<
 * 
 *     cpdef clone(self):
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __Pyx_INCREF(((PyObject *)__pyx_v_self->kind));
  __pyx_r = __pyx_v_self->kind;
  goto __pyx_L0;

  /* "glycopeptidepy/_c/structure/fragment.pyx":380
 *             self.mass += mod.mass * count
 * 
 *     cpdef IonSeriesBase get_series(self):             # <<<<<<<<<<<<<<
 *         return self.kind
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment.PeptideFragment.get_series", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_3get_series(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_2get_series[] = "PeptideFragment.get_series(self) -> IonSeriesBase";
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_3get_series(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_series (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_2get_series(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_2get_series(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_series", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_get_series(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 380, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment.PeptideFragment.get_series", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/fragment.pyx":383
 *         return self.kind
 * 
 *     cpdef clone(self):             # <<<<<<<<<<<<<<
 *         return PeptideFragment._create(
 *             self.series, self.position, self.modification_dict.copy(),
 */

static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_5clone(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_clone(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *__pyx_v_self, int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  struct __pyx_opt_args_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment__create __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("clone", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_clone); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 383, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_5clone)) {
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 383, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "glycopeptidepy/_c/structure/fragment.pyx":384
 * 
 *     cpdef clone(self):
 *         return PeptideFragment._create(             # <<<<<<<<<<<<<<
 *             self.series, self.position, self.modification_dict.copy(),
 *             self.bare_mass, list(self.flanking_amino_acids),
 */
  __Pyx_XDECREF(__pyx_r);

  /* "glycopeptidepy/_c/structure/fragment.pyx":385
 *     cpdef clone(self):
 *         return PeptideFragment._create(
 *             self.series, self.position, self.modification_dict.copy(),             # <<<<<<<<<<<<<<
 *             self.bare_mass, list(self.flanking_amino_acids),
 *             self.glycosylation.copy() if self.glycosylation is not None else None,
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_series); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 385, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase))))) __PYX_ERR(0, 385, __pyx_L1_error)
  __pyx_t_2 = ((PyObject *)((struct __pyx_vtabstruct_14glycopeptidepy_2_c_11count_table_CountTable *)__pyx_v_self->modification_dict->__pyx_vtab)->copy(__pyx_v_self->modification_dict, 0)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 385, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "glycopeptidepy/_c/structure/fragment.pyx":386
 *         return PeptideFragment._create(
 *             self.series, self.position, self.modification_dict.copy(),
 *             self.bare_mass, list(self.flanking_amino_acids),             # <<<<<<<<<<<<<<
 *             self.glycosylation.copy() if self.glycosylation is not None else None,
 *             self._chemical_shift.clone() if self._chemical_shift is not None else None,
 */
  __pyx_t_3 = PySequence_List(__pyx_v_self->flanking_amino_acids); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 386, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);

  /* "glycopeptidepy/_c/structure/fragment.pyx":387
 *             self.series, self.position, self.modification_dict.copy(),
 *             self.bare_mass, list(self.flanking_amino_acids),
 *             self.glycosylation.copy() if self.glycosylation is not None else None,             # <<<<<<<<<<<<<<
 *             self._chemical_shift.clone() if self._chemical_shift is not None else None,
 *             self.composition.clone() if self.composition is not None else None)
 */
  __pyx_t_5 = (__pyx_v_self->glycosylation != ((PyObject*)Py_None));
  if ((__pyx_t_5 != 0)) {
    if (unlikely(__pyx_v_self->glycosylation == Py_None)) {
      PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "copy");
      __PYX_ERR(0, 387, __pyx_L1_error)
    }
    __pyx_t_6 = PyDict_Copy(__pyx_v_self->glycosylation); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 387, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_4 = __pyx_t_6;
    __pyx_t_6 = 0;
  } else {
    __Pyx_INCREF(Py_None);
    __pyx_t_4 = Py_None;
  }

  /* "glycopeptidepy/_c/structure/fragment.pyx":388
 *             self.bare_mass, list(self.flanking_amino_acids),
 *             self.glycosylation.copy() if self.glycosylation is not None else None,
 *             self._chemical_shift.clone() if self._chemical_shift is not None else None,             # <<<<<<<<<<<<<<
 *             self.composition.clone() if self.composition is not None else None)
 * 
 */
  __pyx_t_5 = (((PyObject *)__pyx_v_self->__pyx_base._chemical_shift) != Py_None);
  if ((__pyx_t_5 != 0)) {
    __pyx_t_7 = ((PyObject *)((struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase *)__pyx_v_self->__pyx_base._chemical_shift->__pyx_vtab)->clone(__pyx_v_self->__pyx_base._chemical_shift, 0)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 388, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_6 = __pyx_t_7;
    __pyx_t_7 = 0;
  } else {
    __Pyx_INCREF(Py_None);
    __pyx_t_6 = Py_None;
  }

  /* "glycopeptidepy/_c/structure/fragment.pyx":389
 *             self.glycosylation.copy() if self.glycosylation is not None else None,
 *             self._chemical_shift.clone() if self._chemical_shift is not None else None,
 *             self.composition.clone() if self.composition is not None else None)             # <<<<<<<<<<<<<<
 * 
 *     def total_composition(self):
 */
  __pyx_t_5 = (((PyObject *)__pyx_v_self->composition) != Py_None);
  if ((__pyx_t_5 != 0)) {
    __pyx_t_8 = ((PyObject *)((struct __pyx_vtabstruct_5glypy_11composition_12ccomposition_CComposition *)__pyx_v_self->composition->__pyx_vtab)->clone(__pyx_v_self->composition, 0)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 389, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_7 = __pyx_t_8;
    __pyx_t_8 = 0;
  } else {
    __Pyx_INCREF(Py_None);
    __pyx_t_7 = Py_None;
  }

  /* "glycopeptidepy/_c/structure/fragment.pyx":384
 * 
 *     cpdef clone(self):
 *         return PeptideFragment._create(             # <<<<<<<<<<<<<<
 *             self.series, self.position, self.modification_dict.copy(),
 *             self.bare_mass, list(self.flanking_amino_acids),
 */
  __pyx_t_9.__pyx_n = 4;
  __pyx_t_9.flanking_amino_acids = ((PyObject*)__pyx_t_3);
  __pyx_t_9.glycosylation = ((PyObject*)__pyx_t_4);
  __pyx_t_9.chemical_shift = ((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase *)__pyx_t_6);
  __pyx_t_9.composition = ((struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *)__pyx_t_7);
  __pyx_t_8 = ((PyObject *)__pyx_vtabptr_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment->_create(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *)__pyx_t_1), __pyx_v_self->position, ((struct __pyx_obj_14glycopeptidepy_2_c_11count_table_CountTable *)__pyx_t_2), __pyx_v_self->bare_mass, &__pyx_t_9)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 384, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_r = __pyx_t_8;
  __pyx_t_8 = 0;
  goto __pyx_L0;

  /* "glycopeptidepy/_c/structure/fragment.pyx":383
 *         return self.kind
 * 
 *     cpdef clone(self):             # <<<<<<<<<<<<<<
 *         return PeptideFragment._create(
 *             self.series, self.position, self.modification_dict.copy(),
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment.PeptideFragment.clone", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_5clone(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_4clone[] = "PeptideFragment.clone(self)";
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_5clone(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("clone (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_4clone(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_4clone(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("clone", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_clone(__pyx_v_self, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 383, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment.PeptideFragment.clone", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/fragment.pyx":391
 *             self.composition.clone() if self.composition is not None else None)
 * 
 *     def total_composition(self):             # <<<<<<<<<<<<<<
 *         if self.composition is None:
 *             raise ValueError("Composition data is missing")
 */

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_7total_composition(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_6total_composition[] = "PeptideFragment.total_composition(self)";
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_7total_composition(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("total_composition (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_6total_composition(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_6total_composition(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *__pyx_v_self) {
  PyObject *__pyx_v_composition = NULL;
  PyObject *__pyx_v_chemical_shift = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("total_composition", 0);

  /* "glycopeptidepy/_c/structure/fragment.pyx":392
 * 
 *     def total_composition(self):
 *         if self.composition is None:             # <<<<<<<<<<<<<<
 *             raise ValueError("Composition data is missing")
 *         composition = self.composition.clone()
 */
  __pyx_t_1 = (((PyObject *)__pyx_v_self->composition) == Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (unlikely(__pyx_t_2)) {

    /* "glycopeptidepy/_c/structure/fragment.pyx":393
 *     def total_composition(self):
 *         if self.composition is None:
 *             raise ValueError("Composition data is missing")             # <<<<<<<<<<<<<<
 *         composition = self.composition.clone()
 *         chemical_shift = self.chemical_shift
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__2, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 393, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 393, __pyx_L1_error)

    /* "glycopeptidepy/_c/structure/fragment.pyx":392
 * 
 *     def total_composition(self):
 *         if self.composition is None:             # <<<<<<<<<<<<<<
 *             raise ValueError("Composition data is missing")
 *         composition = self.composition.clone()
 */
  }

  /* "glycopeptidepy/_c/structure/fragment.pyx":394
 *         if self.composition is None:
 *             raise ValueError("Composition data is missing")
 *         composition = self.composition.clone()             # <<<<<<<<<<<<<<
 *         chemical_shift = self.chemical_shift
 *         if chemical_shift is not None:
 */
  __pyx_t_3 = ((PyObject *)((struct __pyx_vtabstruct_5glypy_11composition_12ccomposition_CComposition *)__pyx_v_self->composition->__pyx_vtab)->clone(__pyx_v_self->composition, 0)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 394, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v_composition = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "glycopeptidepy/_c/structure/fragment.pyx":395
 *             raise ValueError("Composition data is missing")
 *         composition = self.composition.clone()
 *         chemical_shift = self.chemical_shift             # <<<<<<<<<<<<<<
 *         if chemical_shift is not None:
 *             composition += chemical_shift.composition
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_chemical_shift); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 395, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v_chemical_shift = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "glycopeptidepy/_c/structure/fragment.pyx":396
 *         composition = self.composition.clone()
 *         chemical_shift = self.chemical_shift
 *         if chemical_shift is not None:             # <<<<<<<<<<<<<<
 *             composition += chemical_shift.composition
 *         return composition
 */
  __pyx_t_2 = (__pyx_v_chemical_shift != Py_None);
  __pyx_t_1 = (__pyx_t_2 != 0);
  if (__pyx_t_1) {

    /* "glycopeptidepy/_c/structure/fragment.pyx":397
 *         chemical_shift = self.chemical_shift
 *         if chemical_shift is not None:
 *             composition += chemical_shift.composition             # <<<<<<<<<<<<<<
 *         return composition
 * 
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_chemical_shift, __pyx_n_s_composition); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 397, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = PyNumber_InPlaceAdd(__pyx_v_composition, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 397, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF_SET(__pyx_v_composition, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "glycopeptidepy/_c/structure/fragment.pyx":396
 *         composition = self.composition.clone()
 *         chemical_shift = self.chemical_shift
 *         if chemical_shift is not None:             # <<<<<<<<<<<<<<
 *             composition += chemical_shift.composition
 *         return composition
 */
  }

  /* "glycopeptidepy/_c/structure/fragment.pyx":398
 *         if chemical_shift is not None:
 *             composition += chemical_shift.composition
 *         return composition             # <<<<<<<<<<<<<<
 * 
 *     def __reduce__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_composition);
  __pyx_r = __pyx_v_composition;
  goto __pyx_L0;

  /* "glycopeptidepy/_c/structure/fragment.pyx":391
 *             self.composition.clone() if self.composition is not None else None)
 * 
 *     def total_composition(self):             # <<<<<<<<<<<<<<
 *         if self.composition is None:
 *             raise ValueError("Composition data is missing")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment.PeptideFragment.total_composition", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_composition);
  __Pyx_XDECREF(__pyx_v_chemical_shift);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/fragment.pyx":400
 *         return composition
 * 
 *     def __reduce__(self):             # <<<<<<<<<<<<<<
 *         return self.__class__, (
 *             self.kind, self.position, self.modification_dict, self.bare_mass,
 */

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_9__reduce__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_8__reduce__[] = "PeptideFragment.__reduce__(self)";
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_9__reduce__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_8__reduce__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_8__reduce__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce__", 0);

  /* "glycopeptidepy/_c/structure/fragment.pyx":401
 * 
 *     def __reduce__(self):
 *         return self.__class__, (             # <<<<<<<<<<<<<<
 *             self.kind, self.position, self.modification_dict, self.bare_mass,
 *             self.flanking_amino_acids, self.glycosylation,
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_class); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 401, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "glycopeptidepy/_c/structure/fragment.pyx":402
 *     def __reduce__(self):
 *         return self.__class__, (
 *             self.kind, self.position, self.modification_dict, self.bare_mass,             # <<<<<<<<<<<<<<
 *             self.flanking_amino_acids, self.glycosylation,
 *             self.chemical_shift, self.composition)
 */
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_self->position); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 402, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyFloat_FromDouble(__pyx_v_self->bare_mass); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 402, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);

  /* "glycopeptidepy/_c/structure/fragment.pyx":404
 *             self.kind, self.position, self.modification_dict, self.bare_mass,
 *             self.flanking_amino_acids, self.glycosylation,
 *             self.chemical_shift, self.composition)             # <<<<<<<<<<<<<<
 * 
 *     cpdef str base_name(self):
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_chemical_shift); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 404, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);

  /* "glycopeptidepy/_c/structure/fragment.pyx":402
 *     def __reduce__(self):
 *         return self.__class__, (
 *             self.kind, self.position, self.modification_dict, self.bare_mass,             # <<<<<<<<<<<<<<
 *             self.flanking_amino_acids, self.glycosylation,
 *             self.chemical_shift, self.composition)
 */
  __pyx_t_5 = PyTuple_New(8); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 402, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->kind));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self->kind));
  PyTuple_SET_ITEM(__pyx_t_5, 0, ((PyObject *)__pyx_v_self->kind));
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_2);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->modification_dict));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self->modification_dict));
  PyTuple_SET_ITEM(__pyx_t_5, 2, ((PyObject *)__pyx_v_self->modification_dict));
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_5, 3, __pyx_t_3);
  __Pyx_INCREF(__pyx_v_self->flanking_amino_acids);
  __Pyx_GIVEREF(__pyx_v_self->flanking_amino_acids);
  PyTuple_SET_ITEM(__pyx_t_5, 4, __pyx_v_self->flanking_amino_acids);
  __Pyx_INCREF(__pyx_v_self->glycosylation);
  __Pyx_GIVEREF(__pyx_v_self->glycosylation);
  PyTuple_SET_ITEM(__pyx_t_5, 5, __pyx_v_self->glycosylation);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_5, 6, __pyx_t_4);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->composition));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self->composition));
  PyTuple_SET_ITEM(__pyx_t_5, 7, ((PyObject *)__pyx_v_self->composition));
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;

  /* "glycopeptidepy/_c/structure/fragment.pyx":401
 * 
 *     def __reduce__(self):
 *         return self.__class__, (             # <<<<<<<<<<<<<<
 *             self.kind, self.position, self.modification_dict, self.bare_mass,
 *             self.flanking_amino_acids, self.glycosylation,
 */
  __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 401, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_5);
  __pyx_t_1 = 0;
  __pyx_t_5 = 0;
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "glycopeptidepy/_c/structure/fragment.pyx":400
 *         return composition
 * 
 *     def __reduce__(self):             # <<<<<<<<<<<<<<
 *         return self.__class__, (
 *             self.kind, self.position, self.modification_dict, self.bare_mass,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment.PeptideFragment.__reduce__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/fragment.pyx":406
 *             self.chemical_shift, self.composition)
 * 
 *     cpdef str base_name(self):             # <<<<<<<<<<<<<<
 *         """Simply return string like b2, y3 with no modification information."""
 *         fragment_name = []
 */

static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_11base_name(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_base_name(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *__pyx_v_self, int __pyx_skip_dispatch) {
  PyObject *__pyx_v_fragment_name = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("base_name", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_base_name); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 406, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_11base_name)) {
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 406, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (!(likely(PyString_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "str", Py_TYPE(__pyx_t_2)->tp_name), 0))) __PYX_ERR(0, 406, __pyx_L1_error)
        __pyx_r = ((PyObject*)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "glycopeptidepy/_c/structure/fragment.pyx":408
 *     cpdef str base_name(self):
 *         """Simply return string like b2, y3 with no modification information."""
 *         fragment_name = []             # <<<<<<<<<<<<<<
 *         fragment_name.append(self.get_series().name)
 *         fragment_name.append(str(self.position))
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 408, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_fragment_name = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "glycopeptidepy/_c/structure/fragment.pyx":409
 *         """Simply return string like b2, y3 with no modification information."""
 *         fragment_name = []
 *         fragment_name.append(self.get_series().name)             # <<<<<<<<<<<<<<
 *         fragment_name.append(str(self.position))
 *         return ''.join(fragment_name)
 */
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.get_series(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *)__pyx_v_self), 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 409, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = ((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *)__pyx_t_1)->name;
  __Pyx_INCREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_5 = __Pyx_PyList_Append(__pyx_v_fragment_name, __pyx_t_2); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 409, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "glycopeptidepy/_c/structure/fragment.pyx":410
 *         fragment_name = []
 *         fragment_name.append(self.get_series().name)
 *         fragment_name.append(str(self.position))             # <<<<<<<<<<<<<<
 *         return ''.join(fragment_name)
 * 
 */
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_self->position); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 410, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyString_Type)), __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 410, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_5 = __Pyx_PyList_Append(__pyx_v_fragment_name, __pyx_t_1); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 410, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "glycopeptidepy/_c/structure/fragment.pyx":411
 *         fragment_name.append(self.get_series().name)
 *         fragment_name.append(str(self.position))
 *         return ''.join(fragment_name)             # <<<<<<<<<<<<<<
 * 
 *     cpdef str get_fragment_name(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyString_Join(__pyx_kp_s_, __pyx_v_fragment_name); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 411, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyString_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "str", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(0, 411, __pyx_L1_error)
  __pyx_r = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "glycopeptidepy/_c/structure/fragment.pyx":406
 *             self.chemical_shift, self.composition)
 * 
 *     cpdef str base_name(self):             # <<<<<<<<<<<<<<
 *         """Simply return string like b2, y3 with no modification information."""
 *         fragment_name = []
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment.PeptideFragment.base_name", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_fragment_name);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_11base_name(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_10base_name[] = "PeptideFragment.base_name(self) -> str\nSimply return string like b2, y3 with no modification information.";
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_11base_name(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("base_name (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_10base_name(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_10base_name(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("base_name", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_base_name(__pyx_v_self, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 406, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment.PeptideFragment.base_name", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/fragment.pyx":413
 *         return ''.join(fragment_name)
 * 
 *     cpdef str get_fragment_name(self):             # <<<<<<<<<<<<<<
 *         """Build a complete fragment name listing any glycosylation modifications
 *         """
 */

static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_13get_fragment_name(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_get_fragment_name(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *__pyx_v_self, int __pyx_skip_dispatch) {
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_ModificationBase *__pyx_v_mod_rule = 0;
  long __pyx_v_count;
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase *__pyx_v_chemical_shift = 0;
  struct __pyx_obj_14glycopeptidepy_2_c_11count_table_CountTable *__pyx_v_modification_dict = 0;
  struct __pyx_obj_14glycopeptidepy_2_c_11count_table_CountTableIterator *__pyx_v_modification_iterator = 0;
  PyObject *__pyx_v_name = 0;
  PyObject *__pyx_v_pkey;
  Py_ssize_t __pyx_v_index;
  Py_ssize_t __pyx_v_size_ref;
  Py_ssize_t __pyx_v_buffer_size;
  Py_ssize_t __pyx_v_ppos;
  struct __pyx_t_14glycopeptidepy_2_c_9structure_8fragment_string_cell __pyx_v_int_conv;
  char *__pyx_v_tmp_buffer;
  char __pyx_v_default_name_buffer[0x80];
  char *__pyx_v_name_buffer;
  char *__pyx_v_oversized_buffer;
  int __pyx_v_needs_free;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_fragment_name", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_get_fragment_name); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 413, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_13get_fragment_name)) {
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 413, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (!(likely(PyString_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "str", Py_TYPE(__pyx_t_2)->tp_name), 0))) __PYX_ERR(0, 413, __pyx_L1_error)
        __pyx_r = ((PyObject*)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "glycopeptidepy/_c/structure/fragment.pyx":426
 *             PyObject *pvalue
 *             Py_ssize_t index, size_ref, buffer_size
 *             Py_ssize_t ppos = 0             # <<<<<<<<<<<<<<
 * 
 *             string_cell int_conv
 */
  __pyx_v_ppos = 0;

  /* "glycopeptidepy/_c/structure/fragment.pyx":436
 *             bint needs_free
 * 
 *         name_buffer = <char*>default_name_buffer             # <<<<<<<<<<<<<<
 *         buffer_size = DEFAULT_FRAGMENT_NAME_BUFFER_SIZE
 *         needs_free = False
 */
  __pyx_v_name_buffer = ((char *)__pyx_v_default_name_buffer);

  /* "glycopeptidepy/_c/structure/fragment.pyx":437
 * 
 *         name_buffer = <char*>default_name_buffer
 *         buffer_size = DEFAULT_FRAGMENT_NAME_BUFFER_SIZE             # <<<<<<<<<<<<<<
 *         needs_free = False
 *         index = 0
 */
  __pyx_v_buffer_size = 0x80;

  /* "glycopeptidepy/_c/structure/fragment.pyx":438
 *         name_buffer = <char*>default_name_buffer
 *         buffer_size = DEFAULT_FRAGMENT_NAME_BUFFER_SIZE
 *         needs_free = False             # <<<<<<<<<<<<<<
 *         index = 0
 *         tmp_buffer = PyStr_AsUTF8AndSize(self.get_series().name, &size_ref)
 */
  __pyx_v_needs_free = 0;

  /* "glycopeptidepy/_c/structure/fragment.pyx":439
 *         buffer_size = DEFAULT_FRAGMENT_NAME_BUFFER_SIZE
 *         needs_free = False
 *         index = 0             # <<<<<<<<<<<<<<
 *         tmp_buffer = PyStr_AsUTF8AndSize(self.get_series().name, &size_ref)
 *         memcpy(&name_buffer[index], tmp_buffer, size_ref)
 */
  __pyx_v_index = 0;

  /* "glycopeptidepy/_c/structure/fragment.pyx":440
 *         needs_free = False
 *         index = 0
 *         tmp_buffer = PyStr_AsUTF8AndSize(self.get_series().name, &size_ref)             # <<<<<<<<<<<<<<
 *         memcpy(&name_buffer[index], tmp_buffer, size_ref)
 *         index += size_ref
 */
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.get_series(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *)__pyx_v_self), 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 440, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = ((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *)__pyx_t_1)->name;
  __Pyx_INCREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_tmp_buffer = PyStr_AsUTF8AndSize(((PyObject*)__pyx_t_2), (&__pyx_v_size_ref));
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "glycopeptidepy/_c/structure/fragment.pyx":441
 *         index = 0
 *         tmp_buffer = PyStr_AsUTF8AndSize(self.get_series().name, &size_ref)
 *         memcpy(&name_buffer[index], tmp_buffer, size_ref)             # <<<<<<<<<<<<<<
 *         index += size_ref
 * 
 */
  (void)(memcpy((&(__pyx_v_name_buffer[__pyx_v_index])), __pyx_v_tmp_buffer, __pyx_v_size_ref));

  /* "glycopeptidepy/_c/structure/fragment.pyx":442
 *         tmp_buffer = PyStr_AsUTF8AndSize(self.get_series().name, &size_ref)
 *         memcpy(&name_buffer[index], tmp_buffer, size_ref)
 *         index += size_ref             # <<<<<<<<<<<<<<
 * 
 *         get_str_int(self.position, &int_conv)
 */
  __pyx_v_index = (__pyx_v_index + __pyx_v_size_ref);

  /* "glycopeptidepy/_c/structure/fragment.pyx":444
 *         index += size_ref
 * 
 *         get_str_int(self.position, &int_conv)             # <<<<<<<<<<<<<<
 *         memcpy(&name_buffer[index], int_conv.string, int_conv.size)
 *         index += int_conv.size
 */
  (void)(__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_get_str_int(__pyx_v_self->position, (&__pyx_v_int_conv)));

  /* "glycopeptidepy/_c/structure/fragment.pyx":445
 * 
 *         get_str_int(self.position, &int_conv)
 *         memcpy(&name_buffer[index], int_conv.string, int_conv.size)             # <<<<<<<<<<<<<<
 *         index += int_conv.size
 * 
 */
  (void)(memcpy((&(__pyx_v_name_buffer[__pyx_v_index])), __pyx_v_int_conv.string, __pyx_v_int_conv.size));

  /* "glycopeptidepy/_c/structure/fragment.pyx":446
 *         get_str_int(self.position, &int_conv)
 *         memcpy(&name_buffer[index], int_conv.string, int_conv.size)
 *         index += int_conv.size             # <<<<<<<<<<<<<<
 * 
 *         modification_dict = self.modification_dict
 */
  __pyx_v_index = (__pyx_v_index + __pyx_v_int_conv.size);

  /* "glycopeptidepy/_c/structure/fragment.pyx":448
 *         index += int_conv.size
 * 
 *         modification_dict = self.modification_dict             # <<<<<<<<<<<<<<
 *         modification_iterator = CountTableIterator._create(modification_dict)
 *         while modification_iterator.has_more():
 */
  __pyx_t_2 = ((PyObject *)__pyx_v_self->modification_dict);
  __Pyx_INCREF(__pyx_t_2);
  __pyx_v_modification_dict = ((struct __pyx_obj_14glycopeptidepy_2_c_11count_table_CountTable *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "glycopeptidepy/_c/structure/fragment.pyx":449
 * 
 *         modification_dict = self.modification_dict
 *         modification_iterator = CountTableIterator._create(modification_dict)             # <<<<<<<<<<<<<<
 *         while modification_iterator.has_more():
 *             ppos = modification_iterator.get_next_value(&pkey, &count)
 */
  __pyx_t_2 = ((PyObject *)__pyx_vtabptr_14glycopeptidepy_2_c_11count_table_CountTableIterator->_create(__pyx_v_modification_dict)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 449, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_modification_iterator = ((struct __pyx_obj_14glycopeptidepy_2_c_11count_table_CountTableIterator *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "glycopeptidepy/_c/structure/fragment.pyx":450
 *         modification_dict = self.modification_dict
 *         modification_iterator = CountTableIterator._create(modification_dict)
 *         while modification_iterator.has_more():             # <<<<<<<<<<<<<<
 *             ppos = modification_iterator.get_next_value(&pkey, &count)
 *             if ppos != 0:
 */
  while (1) {
    __pyx_t_5 = (((struct __pyx_vtabstruct_14glycopeptidepy_2_c_11count_table_CountTableIterator *)__pyx_v_modification_iterator->__pyx_vtab)->has_more(__pyx_v_modification_iterator) != 0);
    if (!__pyx_t_5) break;

    /* "glycopeptidepy/_c/structure/fragment.pyx":451
 *         modification_iterator = CountTableIterator._create(modification_dict)
 *         while modification_iterator.has_more():
 *             ppos = modification_iterator.get_next_value(&pkey, &count)             # <<<<<<<<<<<<<<
 *             if ppos != 0:
 *                 break
 */
    __pyx_v_ppos = ((struct __pyx_vtabstruct_14glycopeptidepy_2_c_11count_table_CountTableIterator *)__pyx_v_modification_iterator->__pyx_vtab)->get_next_value(__pyx_v_modification_iterator, (&__pyx_v_pkey), (&__pyx_v_count));

    /* "glycopeptidepy/_c/structure/fragment.pyx":452
 *         while modification_iterator.has_more():
 *             ppos = modification_iterator.get_next_value(&pkey, &count)
 *             if ppos != 0:             # <<<<<<<<<<<<<<
 *                 break
 *             mod_rule = <ModificationBase>pkey
 */
    __pyx_t_5 = ((__pyx_v_ppos != 0) != 0);
    if (__pyx_t_5) {

      /* "glycopeptidepy/_c/structure/fragment.pyx":453
 *             ppos = modification_iterator.get_next_value(&pkey, &count)
 *             if ppos != 0:
 *                 break             # <<<<<<<<<<<<<<
 *             mod_rule = <ModificationBase>pkey
 *             if buffer_size - index < 30:
 */
      goto __pyx_L4_break;

      /* "glycopeptidepy/_c/structure/fragment.pyx":452
 *         while modification_iterator.has_more():
 *             ppos = modification_iterator.get_next_value(&pkey, &count)
 *             if ppos != 0:             # <<<<<<<<<<<<<<
 *                 break
 *             mod_rule = <ModificationBase>pkey
 */
    }

    /* "glycopeptidepy/_c/structure/fragment.pyx":454
 *             if ppos != 0:
 *                 break
 *             mod_rule = <ModificationBase>pkey             # <<<<<<<<<<<<<<
 *             if buffer_size - index < 30:
 *                 if needs_free:
 */
    __pyx_t_2 = ((PyObject *)__pyx_v_pkey);
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_XDECREF_SET(__pyx_v_mod_rule, ((struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_ModificationBase *)__pyx_t_2));
    __pyx_t_2 = 0;

    /* "glycopeptidepy/_c/structure/fragment.pyx":455
 *                 break
 *             mod_rule = <ModificationBase>pkey
 *             if buffer_size - index < 30:             # <<<<<<<<<<<<<<
 *                 if needs_free:
 *                     name_buffer = <char*>realloc(name_buffer, sizeof(char) * buffer_size * 2)
 */
    __pyx_t_5 = (((__pyx_v_buffer_size - __pyx_v_index) < 30) != 0);
    if (__pyx_t_5) {

      /* "glycopeptidepy/_c/structure/fragment.pyx":456
 *             mod_rule = <ModificationBase>pkey
 *             if buffer_size - index < 30:
 *                 if needs_free:             # <<<<<<<<<<<<<<
 *                     name_buffer = <char*>realloc(name_buffer, sizeof(char) * buffer_size * 2)
 *                     buffer_size *= 2
 */
      __pyx_t_5 = (__pyx_v_needs_free != 0);
      if (__pyx_t_5) {

        /* "glycopeptidepy/_c/structure/fragment.pyx":457
 *             if buffer_size - index < 30:
 *                 if needs_free:
 *                     name_buffer = <char*>realloc(name_buffer, sizeof(char) * buffer_size * 2)             # <<<<<<<<<<<<<<
 *                     buffer_size *= 2
 *                 else:
 */
        __pyx_v_name_buffer = ((char *)realloc(__pyx_v_name_buffer, (((sizeof(char)) * __pyx_v_buffer_size) * 2)));

        /* "glycopeptidepy/_c/structure/fragment.pyx":458
 *                 if needs_free:
 *                     name_buffer = <char*>realloc(name_buffer, sizeof(char) * buffer_size * 2)
 *                     buffer_size *= 2             # <<<<<<<<<<<<<<
 *                 else:
 *                     oversized_buffer = <char*>malloc(sizeof(char) * buffer_size * 2)
 */
        __pyx_v_buffer_size = (__pyx_v_buffer_size * 2);

        /* "glycopeptidepy/_c/structure/fragment.pyx":456
 *             mod_rule = <ModificationBase>pkey
 *             if buffer_size - index < 30:
 *                 if needs_free:             # <<<<<<<<<<<<<<
 *                     name_buffer = <char*>realloc(name_buffer, sizeof(char) * buffer_size * 2)
 *                     buffer_size *= 2
 */
        goto __pyx_L7;
      }

      /* "glycopeptidepy/_c/structure/fragment.pyx":460
 *                     buffer_size *= 2
 *                 else:
 *                     oversized_buffer = <char*>malloc(sizeof(char) * buffer_size * 2)             # <<<<<<<<<<<<<<
 *                     memcpy(oversized_buffer, name_buffer, index)
 *                     name_buffer = oversized_buffer
 */
      /*else*/ {
        __pyx_v_oversized_buffer = ((char *)malloc((((sizeof(char)) * __pyx_v_buffer_size) * 2)));

        /* "glycopeptidepy/_c/structure/fragment.pyx":461
 *                 else:
 *                     oversized_buffer = <char*>malloc(sizeof(char) * buffer_size * 2)
 *                     memcpy(oversized_buffer, name_buffer, index)             # <<<<<<<<<<<<<<
 *                     name_buffer = oversized_buffer
 *                     needs_free = True
 */
        (void)(memcpy(__pyx_v_oversized_buffer, __pyx_v_name_buffer, __pyx_v_index));

        /* "glycopeptidepy/_c/structure/fragment.pyx":462
 *                     oversized_buffer = <char*>malloc(sizeof(char) * buffer_size * 2)
 *                     memcpy(oversized_buffer, name_buffer, index)
 *                     name_buffer = oversized_buffer             # <<<<<<<<<<<<<<
 *                     needs_free = True
 *                     buffer_size *= 2
 */
        __pyx_v_name_buffer = __pyx_v_oversized_buffer;

        /* "glycopeptidepy/_c/structure/fragment.pyx":463
 *                     memcpy(oversized_buffer, name_buffer, index)
 *                     name_buffer = oversized_buffer
 *                     needs_free = True             # <<<<<<<<<<<<<<
 *                     buffer_size *= 2
 *             if mod_rule.is_a(ModificationCategory_glycosylation):
 */
        __pyx_v_needs_free = 1;

        /* "glycopeptidepy/_c/structure/fragment.pyx":464
 *                     name_buffer = oversized_buffer
 *                     needs_free = True
 *                     buffer_size *= 2             # <<<<<<<<<<<<<<
 *             if mod_rule.is_a(ModificationCategory_glycosylation):
 *                 if count > 1:
 */
        __pyx_v_buffer_size = (__pyx_v_buffer_size * 2);
      }
      __pyx_L7:;

      /* "glycopeptidepy/_c/structure/fragment.pyx":455
 *                 break
 *             mod_rule = <ModificationBase>pkey
 *             if buffer_size - index < 30:             # <<<<<<<<<<<<<<
 *                 if needs_free:
 *                     name_buffer = <char*>realloc(name_buffer, sizeof(char) * buffer_size * 2)
 */
    }

    /* "glycopeptidepy/_c/structure/fragment.pyx":465
 *                     needs_free = True
 *                     buffer_size *= 2
 *             if mod_rule.is_a(ModificationCategory_glycosylation):             # <<<<<<<<<<<<<<
 *                 if count > 1:
 *                     name_buffer[index] = "+"
 */
    __pyx_t_2 = __pyx_v_14glycopeptidepy_2_c_9structure_8fragment_ModificationCategory_glycosylation;
    __Pyx_INCREF(__pyx_t_2);
    __pyx_t_5 = (((struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_4base_ModificationBase *)__pyx_v_mod_rule->__pyx_vtab)->is_a(__pyx_v_mod_rule, __pyx_t_2, 0) != 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (__pyx_t_5) {

      /* "glycopeptidepy/_c/structure/fragment.pyx":466
 *                     buffer_size *= 2
 *             if mod_rule.is_a(ModificationCategory_glycosylation):
 *                 if count > 1:             # <<<<<<<<<<<<<<
 *                     name_buffer[index] = "+"
 *                     index += 1
 */
      __pyx_t_5 = ((__pyx_v_count > 1) != 0);
      if (__pyx_t_5) {

        /* "glycopeptidepy/_c/structure/fragment.pyx":467
 *             if mod_rule.is_a(ModificationCategory_glycosylation):
 *                 if count > 1:
 *                     name_buffer[index] = "+"             # <<<<<<<<<<<<<<
 *                     index += 1
 *                     int_conv = str_ints[count]
 */
        (__pyx_v_name_buffer[__pyx_v_index]) = '+';

        /* "glycopeptidepy/_c/structure/fragment.pyx":468
 *                 if count > 1:
 *                     name_buffer[index] = "+"
 *                     index += 1             # <<<<<<<<<<<<<<
 *                     int_conv = str_ints[count]
 *                     memcpy(&name_buffer[index], int_conv.string, int_conv.size)
 */
        __pyx_v_index = (__pyx_v_index + 1);

        /* "glycopeptidepy/_c/structure/fragment.pyx":469
 *                     name_buffer[index] = "+"
 *                     index += 1
 *                     int_conv = str_ints[count]             # <<<<<<<<<<<<<<
 *                     memcpy(&name_buffer[index], int_conv.string, int_conv.size)
 *                     index += int_conv.size
 */
        __pyx_v_int_conv = (__pyx_v_14glycopeptidepy_2_c_9structure_8fragment_str_ints[__pyx_v_count]);

        /* "glycopeptidepy/_c/structure/fragment.pyx":470
 *                     index += 1
 *                     int_conv = str_ints[count]
 *                     memcpy(&name_buffer[index], int_conv.string, int_conv.size)             # <<<<<<<<<<<<<<
 *                     index += int_conv.size
 *                     tmp_buffer = PyStr_AsUTF8AndSize(mod_rule.name, &size_ref)
 */
        (void)(memcpy((&(__pyx_v_name_buffer[__pyx_v_index])), __pyx_v_int_conv.string, __pyx_v_int_conv.size));

        /* "glycopeptidepy/_c/structure/fragment.pyx":471
 *                     int_conv = str_ints[count]
 *                     memcpy(&name_buffer[index], int_conv.string, int_conv.size)
 *                     index += int_conv.size             # <<<<<<<<<<<<<<
 *                     tmp_buffer = PyStr_AsUTF8AndSize(mod_rule.name, &size_ref)
 *                     if size_ref + index > buffer_size:
 */
        __pyx_v_index = (__pyx_v_index + __pyx_v_int_conv.size);

        /* "glycopeptidepy/_c/structure/fragment.pyx":472
 *                     memcpy(&name_buffer[index], int_conv.string, int_conv.size)
 *                     index += int_conv.size
 *                     tmp_buffer = PyStr_AsUTF8AndSize(mod_rule.name, &size_ref)             # <<<<<<<<<<<<<<
 *                     if size_ref + index > buffer_size:
 *                         if needs_free:
 */
        if (!(likely(PyString_CheckExact(__pyx_v_mod_rule->name))||((__pyx_v_mod_rule->name) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "str", Py_TYPE(__pyx_v_mod_rule->name)->tp_name), 0))) __PYX_ERR(0, 472, __pyx_L1_error)
        __pyx_t_2 = __pyx_v_mod_rule->name;
        __Pyx_INCREF(__pyx_t_2);
        __pyx_v_tmp_buffer = PyStr_AsUTF8AndSize(((PyObject*)__pyx_t_2), (&__pyx_v_size_ref));
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

        /* "glycopeptidepy/_c/structure/fragment.pyx":473
 *                     index += int_conv.size
 *                     tmp_buffer = PyStr_AsUTF8AndSize(mod_rule.name, &size_ref)
 *                     if size_ref + index > buffer_size:             # <<<<<<<<<<<<<<
 *                         if needs_free:
 *                             name_buffer = <char*>realloc(name_buffer, sizeof(char) * (buffer_size + size_ref) * 2)
 */
        __pyx_t_5 = (((__pyx_v_size_ref + __pyx_v_index) > __pyx_v_buffer_size) != 0);
        if (__pyx_t_5) {

          /* "glycopeptidepy/_c/structure/fragment.pyx":474
 *                     tmp_buffer = PyStr_AsUTF8AndSize(mod_rule.name, &size_ref)
 *                     if size_ref + index > buffer_size:
 *                         if needs_free:             # <<<<<<<<<<<<<<
 *                             name_buffer = <char*>realloc(name_buffer, sizeof(char) * (buffer_size + size_ref) * 2)
 *                             buffer_size = (buffer_size + size_ref) * 2
 */
          __pyx_t_5 = (__pyx_v_needs_free != 0);
          if (__pyx_t_5) {

            /* "glycopeptidepy/_c/structure/fragment.pyx":475
 *                     if size_ref + index > buffer_size:
 *                         if needs_free:
 *                             name_buffer = <char*>realloc(name_buffer, sizeof(char) * (buffer_size + size_ref) * 2)             # <<<<<<<<<<<<<<
 *                             buffer_size = (buffer_size + size_ref) * 2
 *                         else:
 */
            __pyx_v_name_buffer = ((char *)realloc(__pyx_v_name_buffer, (((sizeof(char)) * (__pyx_v_buffer_size + __pyx_v_size_ref)) * 2)));

            /* "glycopeptidepy/_c/structure/fragment.pyx":476
 *                         if needs_free:
 *                             name_buffer = <char*>realloc(name_buffer, sizeof(char) * (buffer_size + size_ref) * 2)
 *                             buffer_size = (buffer_size + size_ref) * 2             # <<<<<<<<<<<<<<
 *                         else:
 *                             oversized_buffer = <char*>malloc(sizeof(char) * (buffer_size + size_ref) * 2)
 */
            __pyx_v_buffer_size = ((__pyx_v_buffer_size + __pyx_v_size_ref) * 2);

            /* "glycopeptidepy/_c/structure/fragment.pyx":474
 *                     tmp_buffer = PyStr_AsUTF8AndSize(mod_rule.name, &size_ref)
 *                     if size_ref + index > buffer_size:
 *                         if needs_free:             # <<<<<<<<<<<<<<
 *                             name_buffer = <char*>realloc(name_buffer, sizeof(char) * (buffer_size + size_ref) * 2)
 *                             buffer_size = (buffer_size + size_ref) * 2
 */
            goto __pyx_L11;
          }

          /* "glycopeptidepy/_c/structure/fragment.pyx":478
 *                             buffer_size = (buffer_size + size_ref) * 2
 *                         else:
 *                             oversized_buffer = <char*>malloc(sizeof(char) * (buffer_size + size_ref) * 2)             # <<<<<<<<<<<<<<
 *                             memcpy(oversized_buffer, name_buffer, index)
 *                             name_buffer = oversized_buffer
 */
          /*else*/ {
            __pyx_v_oversized_buffer = ((char *)malloc((((sizeof(char)) * (__pyx_v_buffer_size + __pyx_v_size_ref)) * 2)));

            /* "glycopeptidepy/_c/structure/fragment.pyx":479
 *                         else:
 *                             oversized_buffer = <char*>malloc(sizeof(char) * (buffer_size + size_ref) * 2)
 *                             memcpy(oversized_buffer, name_buffer, index)             # <<<<<<<<<<<<<<
 *                             name_buffer = oversized_buffer
 *                             needs_free = True
 */
            (void)(memcpy(__pyx_v_oversized_buffer, __pyx_v_name_buffer, __pyx_v_index));

            /* "glycopeptidepy/_c/structure/fragment.pyx":480
 *                             oversized_buffer = <char*>malloc(sizeof(char) * (buffer_size + size_ref) * 2)
 *                             memcpy(oversized_buffer, name_buffer, index)
 *                             name_buffer = oversized_buffer             # <<<<<<<<<<<<<<
 *                             needs_free = True
 *                             buffer_size = (buffer_size + size_ref) * 2
 */
            __pyx_v_name_buffer = __pyx_v_oversized_buffer;

            /* "glycopeptidepy/_c/structure/fragment.pyx":481
 *                             memcpy(oversized_buffer, name_buffer, index)
 *                             name_buffer = oversized_buffer
 *                             needs_free = True             # <<<<<<<<<<<<<<
 *                             buffer_size = (buffer_size + size_ref) * 2
 *                     memcpy(&name_buffer[index], tmp_buffer, size_ref)
 */
            __pyx_v_needs_free = 1;

            /* "glycopeptidepy/_c/structure/fragment.pyx":482
 *                             name_buffer = oversized_buffer
 *                             needs_free = True
 *                             buffer_size = (buffer_size + size_ref) * 2             # <<<<<<<<<<<<<<
 *                     memcpy(&name_buffer[index], tmp_buffer, size_ref)
 *                     index += size_ref
 */
            __pyx_v_buffer_size = ((__pyx_v_buffer_size + __pyx_v_size_ref) * 2);
          }
          __pyx_L11:;

          /* "glycopeptidepy/_c/structure/fragment.pyx":473
 *                     index += int_conv.size
 *                     tmp_buffer = PyStr_AsUTF8AndSize(mod_rule.name, &size_ref)
 *                     if size_ref + index > buffer_size:             # <<<<<<<<<<<<<<
 *                         if needs_free:
 *                             name_buffer = <char*>realloc(name_buffer, sizeof(char) * (buffer_size + size_ref) * 2)
 */
        }

        /* "glycopeptidepy/_c/structure/fragment.pyx":483
 *                             needs_free = True
 *                             buffer_size = (buffer_size + size_ref) * 2
 *                     memcpy(&name_buffer[index], tmp_buffer, size_ref)             # <<<<<<<<<<<<<<
 *                     index += size_ref
 *                 elif count == 1:
 */
        (void)(memcpy((&(__pyx_v_name_buffer[__pyx_v_index])), __pyx_v_tmp_buffer, __pyx_v_size_ref));

        /* "glycopeptidepy/_c/structure/fragment.pyx":484
 *                             buffer_size = (buffer_size + size_ref) * 2
 *                     memcpy(&name_buffer[index], tmp_buffer, size_ref)
 *                     index += size_ref             # <<<<<<<<<<<<<<
 *                 elif count == 1:
 *                     name_buffer[index] = "+"
 */
        __pyx_v_index = (__pyx_v_index + __pyx_v_size_ref);

        /* "glycopeptidepy/_c/structure/fragment.pyx":466
 *                     buffer_size *= 2
 *             if mod_rule.is_a(ModificationCategory_glycosylation):
 *                 if count > 1:             # <<<<<<<<<<<<<<
 *                     name_buffer[index] = "+"
 *                     index += 1
 */
        goto __pyx_L9;
      }

      /* "glycopeptidepy/_c/structure/fragment.pyx":485
 *                     memcpy(&name_buffer[index], tmp_buffer, size_ref)
 *                     index += size_ref
 *                 elif count == 1:             # <<<<<<<<<<<<<<
 *                     name_buffer[index] = "+"
 *                     index += 1
 */
      __pyx_t_5 = ((__pyx_v_count == 1) != 0);
      if (__pyx_t_5) {

        /* "glycopeptidepy/_c/structure/fragment.pyx":486
 *                     index += size_ref
 *                 elif count == 1:
 *                     name_buffer[index] = "+"             # <<<<<<<<<<<<<<
 *                     index += 1
 *                     tmp_buffer = PyStr_AsUTF8AndSize(mod_rule.name, &size_ref)
 */
        (__pyx_v_name_buffer[__pyx_v_index]) = '+';

        /* "glycopeptidepy/_c/structure/fragment.pyx":487
 *                 elif count == 1:
 *                     name_buffer[index] = "+"
 *                     index += 1             # <<<<<<<<<<<<<<
 *                     tmp_buffer = PyStr_AsUTF8AndSize(mod_rule.name, &size_ref)
 *                     if size_ref + index > buffer_size:
 */
        __pyx_v_index = (__pyx_v_index + 1);

        /* "glycopeptidepy/_c/structure/fragment.pyx":488
 *                     name_buffer[index] = "+"
 *                     index += 1
 *                     tmp_buffer = PyStr_AsUTF8AndSize(mod_rule.name, &size_ref)             # <<<<<<<<<<<<<<
 *                     if size_ref + index > buffer_size:
 *                         if needs_free:
 */
        if (!(likely(PyString_CheckExact(__pyx_v_mod_rule->name))||((__pyx_v_mod_rule->name) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "str", Py_TYPE(__pyx_v_mod_rule->name)->tp_name), 0))) __PYX_ERR(0, 488, __pyx_L1_error)
        __pyx_t_2 = __pyx_v_mod_rule->name;
        __Pyx_INCREF(__pyx_t_2);
        __pyx_v_tmp_buffer = PyStr_AsUTF8AndSize(((PyObject*)__pyx_t_2), (&__pyx_v_size_ref));
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

        /* "glycopeptidepy/_c/structure/fragment.pyx":489
 *                     index += 1
 *                     tmp_buffer = PyStr_AsUTF8AndSize(mod_rule.name, &size_ref)
 *                     if size_ref + index > buffer_size:             # <<<<<<<<<<<<<<
 *                         if needs_free:
 *                             name_buffer = <char*>realloc(name_buffer, sizeof(char) * (buffer_size + size_ref) * 2)
 */
        __pyx_t_5 = (((__pyx_v_size_ref + __pyx_v_index) > __pyx_v_buffer_size) != 0);
        if (__pyx_t_5) {

          /* "glycopeptidepy/_c/structure/fragment.pyx":490
 *                     tmp_buffer = PyStr_AsUTF8AndSize(mod_rule.name, &size_ref)
 *                     if size_ref + index > buffer_size:
 *                         if needs_free:             # <<<<<<<<<<<<<<
 *                             name_buffer = <char*>realloc(name_buffer, sizeof(char) * (buffer_size + size_ref) * 2)
 *                             buffer_size = (buffer_size + size_ref) * 2
 */
          __pyx_t_5 = (__pyx_v_needs_free != 0);
          if (__pyx_t_5) {

            /* "glycopeptidepy/_c/structure/fragment.pyx":491
 *                     if size_ref + index > buffer_size:
 *                         if needs_free:
 *                             name_buffer = <char*>realloc(name_buffer, sizeof(char) * (buffer_size + size_ref) * 2)             # <<<<<<<<<<<<<<
 *                             buffer_size = (buffer_size + size_ref) * 2
 *                         else:
 */
            __pyx_v_name_buffer = ((char *)realloc(__pyx_v_name_buffer, (((sizeof(char)) * (__pyx_v_buffer_size + __pyx_v_size_ref)) * 2)));

            /* "glycopeptidepy/_c/structure/fragment.pyx":492
 *                         if needs_free:
 *                             name_buffer = <char*>realloc(name_buffer, sizeof(char) * (buffer_size + size_ref) * 2)
 *                             buffer_size = (buffer_size + size_ref) * 2             # <<<<<<<<<<<<<<
 *                         else:
 *                             oversized_buffer = <char*>malloc(sizeof(char) * (buffer_size + size_ref) * 2)
 */
            __pyx_v_buffer_size = ((__pyx_v_buffer_size + __pyx_v_size_ref) * 2);

            /* "glycopeptidepy/_c/structure/fragment.pyx":490
 *                     tmp_buffer = PyStr_AsUTF8AndSize(mod_rule.name, &size_ref)
 *                     if size_ref + index > buffer_size:
 *                         if needs_free:             # <<<<<<<<<<<<<<
 *                             name_buffer = <char*>realloc(name_buffer, sizeof(char) * (buffer_size + size_ref) * 2)
 *                             buffer_size = (buffer_size + size_ref) * 2
 */
            goto __pyx_L13;
          }

          /* "glycopeptidepy/_c/structure/fragment.pyx":494
 *                             buffer_size = (buffer_size + size_ref) * 2
 *                         else:
 *                             oversized_buffer = <char*>malloc(sizeof(char) * (buffer_size + size_ref) * 2)             # <<<<<<<<<<<<<<
 *                             memcpy(oversized_buffer, name_buffer, index)
 *                             name_buffer = oversized_buffer
 */
          /*else*/ {
            __pyx_v_oversized_buffer = ((char *)malloc((((sizeof(char)) * (__pyx_v_buffer_size + __pyx_v_size_ref)) * 2)));

            /* "glycopeptidepy/_c/structure/fragment.pyx":495
 *                         else:
 *                             oversized_buffer = <char*>malloc(sizeof(char) * (buffer_size + size_ref) * 2)
 *                             memcpy(oversized_buffer, name_buffer, index)             # <<<<<<<<<<<<<<
 *                             name_buffer = oversized_buffer
 *                             needs_free = True
 */
            (void)(memcpy(__pyx_v_oversized_buffer, __pyx_v_name_buffer, __pyx_v_index));

            /* "glycopeptidepy/_c/structure/fragment.pyx":496
 *                             oversized_buffer = <char*>malloc(sizeof(char) * (buffer_size + size_ref) * 2)
 *                             memcpy(oversized_buffer, name_buffer, index)
 *                             name_buffer = oversized_buffer             # <<<<<<<<<<<<<<
 *                             needs_free = True
 *                             buffer_size = (buffer_size + size_ref) * 2
 */
            __pyx_v_name_buffer = __pyx_v_oversized_buffer;

            /* "glycopeptidepy/_c/structure/fragment.pyx":497
 *                             memcpy(oversized_buffer, name_buffer, index)
 *                             name_buffer = oversized_buffer
 *                             needs_free = True             # <<<<<<<<<<<<<<
 *                             buffer_size = (buffer_size + size_ref) * 2
 *                     memcpy(&name_buffer[index], tmp_buffer, size_ref)
 */
            __pyx_v_needs_free = 1;

            /* "glycopeptidepy/_c/structure/fragment.pyx":498
 *                             name_buffer = oversized_buffer
 *                             needs_free = True
 *                             buffer_size = (buffer_size + size_ref) * 2             # <<<<<<<<<<<<<<
 *                     memcpy(&name_buffer[index], tmp_buffer, size_ref)
 *                     index += size_ref
 */
            __pyx_v_buffer_size = ((__pyx_v_buffer_size + __pyx_v_size_ref) * 2);
          }
          __pyx_L13:;

          /* "glycopeptidepy/_c/structure/fragment.pyx":489
 *                     index += 1
 *                     tmp_buffer = PyStr_AsUTF8AndSize(mod_rule.name, &size_ref)
 *                     if size_ref + index > buffer_size:             # <<<<<<<<<<<<<<
 *                         if needs_free:
 *                             name_buffer = <char*>realloc(name_buffer, sizeof(char) * (buffer_size + size_ref) * 2)
 */
        }

        /* "glycopeptidepy/_c/structure/fragment.pyx":499
 *                             needs_free = True
 *                             buffer_size = (buffer_size + size_ref) * 2
 *                     memcpy(&name_buffer[index], tmp_buffer, size_ref)             # <<<<<<<<<<<<<<
 *                     index += size_ref
 * 
 */
        (void)(memcpy((&(__pyx_v_name_buffer[__pyx_v_index])), __pyx_v_tmp_buffer, __pyx_v_size_ref));

        /* "glycopeptidepy/_c/structure/fragment.pyx":500
 *                             buffer_size = (buffer_size + size_ref) * 2
 *                     memcpy(&name_buffer[index], tmp_buffer, size_ref)
 *                     index += size_ref             # <<<<<<<<<<<<<<
 * 
 *         chemical_shift = self.get_chemical_shift()
 */
        __pyx_v_index = (__pyx_v_index + __pyx_v_size_ref);

        /* "glycopeptidepy/_c/structure/fragment.pyx":485
 *                     memcpy(&name_buffer[index], tmp_buffer, size_ref)
 *                     index += size_ref
 *                 elif count == 1:             # <<<<<<<<<<<<<<
 *                     name_buffer[index] = "+"
 *                     index += 1
 */
      }
      __pyx_L9:;

      /* "glycopeptidepy/_c/structure/fragment.pyx":465
 *                     needs_free = True
 *                     buffer_size *= 2
 *             if mod_rule.is_a(ModificationCategory_glycosylation):             # <<<<<<<<<<<<<<
 *                 if count > 1:
 *                     name_buffer[index] = "+"
 */
    }
  }
  __pyx_L4_break:;

  /* "glycopeptidepy/_c/structure/fragment.pyx":502
 *                     index += size_ref
 * 
 *         chemical_shift = self.get_chemical_shift()             # <<<<<<<<<<<<<<
 *         if chemical_shift is not None:
 *             tmp_buffer = PyStr_AsUTF8AndSize(chemical_shift.name, &size_ref)
 */
  __pyx_t_2 = ((PyObject *)((struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.get_chemical_shift(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *)__pyx_v_self), 0)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 502, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_chemical_shift = ((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "glycopeptidepy/_c/structure/fragment.pyx":503
 * 
 *         chemical_shift = self.get_chemical_shift()
 *         if chemical_shift is not None:             # <<<<<<<<<<<<<<
 *             tmp_buffer = PyStr_AsUTF8AndSize(chemical_shift.name, &size_ref)
 *             if size_ref + index > buffer_size:
 */
  __pyx_t_5 = (((PyObject *)__pyx_v_chemical_shift) != Py_None);
  __pyx_t_6 = (__pyx_t_5 != 0);
  if (__pyx_t_6) {

    /* "glycopeptidepy/_c/structure/fragment.pyx":504
 *         chemical_shift = self.get_chemical_shift()
 *         if chemical_shift is not None:
 *             tmp_buffer = PyStr_AsUTF8AndSize(chemical_shift.name, &size_ref)             # <<<<<<<<<<<<<<
 *             if size_ref + index > buffer_size:
 *                 if needs_free:
 */
    __pyx_t_2 = __pyx_v_chemical_shift->name;
    __Pyx_INCREF(__pyx_t_2);
    __pyx_v_tmp_buffer = PyStr_AsUTF8AndSize(((PyObject*)__pyx_t_2), (&__pyx_v_size_ref));
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "glycopeptidepy/_c/structure/fragment.pyx":505
 *         if chemical_shift is not None:
 *             tmp_buffer = PyStr_AsUTF8AndSize(chemical_shift.name, &size_ref)
 *             if size_ref + index > buffer_size:             # <<<<<<<<<<<<<<
 *                 if needs_free:
 *                     name_buffer = <char*>realloc(name_buffer, sizeof(char) * (buffer_size + size_ref))
 */
    __pyx_t_6 = (((__pyx_v_size_ref + __pyx_v_index) > __pyx_v_buffer_size) != 0);
    if (__pyx_t_6) {

      /* "glycopeptidepy/_c/structure/fragment.pyx":506
 *             tmp_buffer = PyStr_AsUTF8AndSize(chemical_shift.name, &size_ref)
 *             if size_ref + index > buffer_size:
 *                 if needs_free:             # <<<<<<<<<<<<<<
 *                     name_buffer = <char*>realloc(name_buffer, sizeof(char) * (buffer_size + size_ref))
 *                     buffer_size = (buffer_size + size_ref)
 */
      __pyx_t_6 = (__pyx_v_needs_free != 0);
      if (__pyx_t_6) {

        /* "glycopeptidepy/_c/structure/fragment.pyx":507
 *             if size_ref + index > buffer_size:
 *                 if needs_free:
 *                     name_buffer = <char*>realloc(name_buffer, sizeof(char) * (buffer_size + size_ref))             # <<<<<<<<<<<<<<
 *                     buffer_size = (buffer_size + size_ref)
 *                 else:
 */
        __pyx_v_name_buffer = ((char *)realloc(__pyx_v_name_buffer, ((sizeof(char)) * (__pyx_v_buffer_size + __pyx_v_size_ref))));

        /* "glycopeptidepy/_c/structure/fragment.pyx":508
 *                 if needs_free:
 *                     name_buffer = <char*>realloc(name_buffer, sizeof(char) * (buffer_size + size_ref))
 *                     buffer_size = (buffer_size + size_ref)             # <<<<<<<<<<<<<<
 *                 else:
 *                     oversized_buffer = <char*>malloc(sizeof(char) * (buffer_size + size_ref))
 */
        __pyx_v_buffer_size = (__pyx_v_buffer_size + __pyx_v_size_ref);

        /* "glycopeptidepy/_c/structure/fragment.pyx":506
 *             tmp_buffer = PyStr_AsUTF8AndSize(chemical_shift.name, &size_ref)
 *             if size_ref + index > buffer_size:
 *                 if needs_free:             # <<<<<<<<<<<<<<
 *                     name_buffer = <char*>realloc(name_buffer, sizeof(char) * (buffer_size + size_ref))
 *                     buffer_size = (buffer_size + size_ref)
 */
        goto __pyx_L16;
      }

      /* "glycopeptidepy/_c/structure/fragment.pyx":510
 *                     buffer_size = (buffer_size + size_ref)
 *                 else:
 *                     oversized_buffer = <char*>malloc(sizeof(char) * (buffer_size + size_ref))             # <<<<<<<<<<<<<<
 *                     memcpy(oversized_buffer, name_buffer, index)
 *                     name_buffer = oversized_buffer
 */
      /*else*/ {
        __pyx_v_oversized_buffer = ((char *)malloc(((sizeof(char)) * (__pyx_v_buffer_size + __pyx_v_size_ref))));

        /* "glycopeptidepy/_c/structure/fragment.pyx":511
 *                 else:
 *                     oversized_buffer = <char*>malloc(sizeof(char) * (buffer_size + size_ref))
 *                     memcpy(oversized_buffer, name_buffer, index)             # <<<<<<<<<<<<<<
 *                     name_buffer = oversized_buffer
 *                     needs_free = True
 */
        (void)(memcpy(__pyx_v_oversized_buffer, __pyx_v_name_buffer, __pyx_v_index));

        /* "glycopeptidepy/_c/structure/fragment.pyx":512
 *                     oversized_buffer = <char*>malloc(sizeof(char) * (buffer_size + size_ref))
 *                     memcpy(oversized_buffer, name_buffer, index)
 *                     name_buffer = oversized_buffer             # <<<<<<<<<<<<<<
 *                     needs_free = True
 *                     buffer_size = (buffer_size + size_ref)
 */
        __pyx_v_name_buffer = __pyx_v_oversized_buffer;

        /* "glycopeptidepy/_c/structure/fragment.pyx":513
 *                     memcpy(oversized_buffer, name_buffer, index)
 *                     name_buffer = oversized_buffer
 *                     needs_free = True             # <<<<<<<<<<<<<<
 *                     buffer_size = (buffer_size + size_ref)
 *             memcpy(&name_buffer[index], tmp_buffer, size_ref)
 */
        __pyx_v_needs_free = 1;

        /* "glycopeptidepy/_c/structure/fragment.pyx":514
 *                     name_buffer = oversized_buffer
 *                     needs_free = True
 *                     buffer_size = (buffer_size + size_ref)             # <<<<<<<<<<<<<<
 *             memcpy(&name_buffer[index], tmp_buffer, size_ref)
 *             index += size_ref
 */
        __pyx_v_buffer_size = (__pyx_v_buffer_size + __pyx_v_size_ref);
      }
      __pyx_L16:;

      /* "glycopeptidepy/_c/structure/fragment.pyx":505
 *         if chemical_shift is not None:
 *             tmp_buffer = PyStr_AsUTF8AndSize(chemical_shift.name, &size_ref)
 *             if size_ref + index > buffer_size:             # <<<<<<<<<<<<<<
 *                 if needs_free:
 *                     name_buffer = <char*>realloc(name_buffer, sizeof(char) * (buffer_size + size_ref))
 */
    }

    /* "glycopeptidepy/_c/structure/fragment.pyx":515
 *                     needs_free = True
 *                     buffer_size = (buffer_size + size_ref)
 *             memcpy(&name_buffer[index], tmp_buffer, size_ref)             # <<<<<<<<<<<<<<
 *             index += size_ref
 *         name = PyStr_FromStringAndSize(&name_buffer[0], index)
 */
    (void)(memcpy((&(__pyx_v_name_buffer[__pyx_v_index])), __pyx_v_tmp_buffer, __pyx_v_size_ref));

    /* "glycopeptidepy/_c/structure/fragment.pyx":516
 *                     buffer_size = (buffer_size + size_ref)
 *             memcpy(&name_buffer[index], tmp_buffer, size_ref)
 *             index += size_ref             # <<<<<<<<<<<<<<
 *         name = PyStr_FromStringAndSize(&name_buffer[0], index)
 *         if needs_free:
 */
    __pyx_v_index = (__pyx_v_index + __pyx_v_size_ref);

    /* "glycopeptidepy/_c/structure/fragment.pyx":503
 * 
 *         chemical_shift = self.get_chemical_shift()
 *         if chemical_shift is not None:             # <<<<<<<<<<<<<<
 *             tmp_buffer = PyStr_AsUTF8AndSize(chemical_shift.name, &size_ref)
 *             if size_ref + index > buffer_size:
 */
  }

  /* "glycopeptidepy/_c/structure/fragment.pyx":517
 *             memcpy(&name_buffer[index], tmp_buffer, size_ref)
 *             index += size_ref
 *         name = PyStr_FromStringAndSize(&name_buffer[0], index)             # <<<<<<<<<<<<<<
 *         if needs_free:
 *             free(name_buffer)
 */
  __pyx_t_2 = PyStr_FromStringAndSize((&(__pyx_v_name_buffer[0])), __pyx_v_index); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 517, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_name = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "glycopeptidepy/_c/structure/fragment.pyx":518
 *             index += size_ref
 *         name = PyStr_FromStringAndSize(&name_buffer[0], index)
 *         if needs_free:             # <<<<<<<<<<<<<<
 *             free(name_buffer)
 *         return name
 */
  __pyx_t_6 = (__pyx_v_needs_free != 0);
  if (__pyx_t_6) {

    /* "glycopeptidepy/_c/structure/fragment.pyx":519
 *         name = PyStr_FromStringAndSize(&name_buffer[0], index)
 *         if needs_free:
 *             free(name_buffer)             # <<<<<<<<<<<<<<
 *         return name
 * 
 */
    free(__pyx_v_name_buffer);

    /* "glycopeptidepy/_c/structure/fragment.pyx":518
 *             index += size_ref
 *         name = PyStr_FromStringAndSize(&name_buffer[0], index)
 *         if needs_free:             # <<<<<<<<<<<<<<
 *             free(name_buffer)
 *         return name
 */
  }

  /* "glycopeptidepy/_c/structure/fragment.pyx":520
 *         if needs_free:
 *             free(name_buffer)
 *         return name             # <<<<<<<<<<<<<<
 * 
 *     cdef bint _is_glycosylated(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_name);
  __pyx_r = __pyx_v_name;
  goto __pyx_L0;

  /* "glycopeptidepy/_c/structure/fragment.pyx":413
 *         return ''.join(fragment_name)
 * 
 *     cpdef str get_fragment_name(self):             # <<<<<<<<<<<<<<
 *         """Build a complete fragment name listing any glycosylation modifications
 *         """
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment.PeptideFragment.get_fragment_name", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_mod_rule);
  __Pyx_XDECREF((PyObject *)__pyx_v_chemical_shift);
  __Pyx_XDECREF((PyObject *)__pyx_v_modification_dict);
  __Pyx_XDECREF((PyObject *)__pyx_v_modification_iterator);
  __Pyx_XDECREF(__pyx_v_name);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_13get_fragment_name(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_12get_fragment_name[] = "PeptideFragment.get_fragment_name(self) -> str\nBuild a complete fragment name listing any glycosylation modifications\n        ";
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_13get_fragment_name(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_fragment_name (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_12get_fragment_name(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_12get_fragment_name(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_fragment_name", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_get_fragment_name(__pyx_v_self, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 413, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment.PeptideFragment.get_fragment_name", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/fragment.pyx":522
 *         return name
 * 
 *     cdef bint _is_glycosylated(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             PyObject *pkey
 */

static int __pyx_f_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment__is_glycosylated(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *__pyx_v_self) {
  PyObject *__pyx_v_pkey;
  long __pyx_v_value;
  Py_ssize_t __pyx_v_ppos;
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_ModificationBase *__pyx_v_mod = 0;
  struct __pyx_obj_14glycopeptidepy_2_c_11count_table_CountTable *__pyx_v_modification_dict = 0;
  struct __pyx_obj_14glycopeptidepy_2_c_11count_table_CountTableIterator *__pyx_v_modification_iterator = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_is_glycosylated", 0);

  /* "glycopeptidepy/_c/structure/fragment.pyx":526
 *             PyObject *pkey
 *             long value
 *             Py_ssize_t ppos = 0             # <<<<<<<<<<<<<<
 *             ModificationBase mod
 *             CountTable modification_dict
 */
  __pyx_v_ppos = 0;

  /* "glycopeptidepy/_c/structure/fragment.pyx":530
 *             CountTable modification_dict
 *             CountTableIterator modification_iterator
 *         if self.glycosylation:             # <<<<<<<<<<<<<<
 *             return True
 *         else:
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_self->glycosylation); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 530, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "glycopeptidepy/_c/structure/fragment.pyx":531
 *             CountTableIterator modification_iterator
 *         if self.glycosylation:
 *             return True             # <<<<<<<<<<<<<<
 *         else:
 *             modification_dict = self.modification_dict
 */
    __pyx_r = 1;
    goto __pyx_L0;

    /* "glycopeptidepy/_c/structure/fragment.pyx":530
 *             CountTable modification_dict
 *             CountTableIterator modification_iterator
 *         if self.glycosylation:             # <<<<<<<<<<<<<<
 *             return True
 *         else:
 */
  }

  /* "glycopeptidepy/_c/structure/fragment.pyx":533
 *             return True
 *         else:
 *             modification_dict = self.modification_dict             # <<<<<<<<<<<<<<
 *             modification_iterator = CountTableIterator._create(modification_dict)
 *             while modification_iterator.has_more():
 */
  /*else*/ {
    __pyx_t_2 = ((PyObject *)__pyx_v_self->modification_dict);
    __Pyx_INCREF(__pyx_t_2);
    __pyx_v_modification_dict = ((struct __pyx_obj_14glycopeptidepy_2_c_11count_table_CountTable *)__pyx_t_2);
    __pyx_t_2 = 0;

    /* "glycopeptidepy/_c/structure/fragment.pyx":534
 *         else:
 *             modification_dict = self.modification_dict
 *             modification_iterator = CountTableIterator._create(modification_dict)             # <<<<<<<<<<<<<<
 *             while modification_iterator.has_more():
 *                 ppos = modification_iterator.get_next_value(&pkey, &value)
 */
    __pyx_t_2 = ((PyObject *)__pyx_vtabptr_14glycopeptidepy_2_c_11count_table_CountTableIterator->_create(__pyx_v_modification_dict)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 534, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_v_modification_iterator = ((struct __pyx_obj_14glycopeptidepy_2_c_11count_table_CountTableIterator *)__pyx_t_2);
    __pyx_t_2 = 0;

    /* "glycopeptidepy/_c/structure/fragment.pyx":535
 *             modification_dict = self.modification_dict
 *             modification_iterator = CountTableIterator._create(modification_dict)
 *             while modification_iterator.has_more():             # <<<<<<<<<<<<<<
 *                 ppos = modification_iterator.get_next_value(&pkey, &value)
 *                 if ppos != 0:
 */
    while (1) {
      __pyx_t_1 = (((struct __pyx_vtabstruct_14glycopeptidepy_2_c_11count_table_CountTableIterator *)__pyx_v_modification_iterator->__pyx_vtab)->has_more(__pyx_v_modification_iterator) != 0);
      if (!__pyx_t_1) break;

      /* "glycopeptidepy/_c/structure/fragment.pyx":536
 *             modification_iterator = CountTableIterator._create(modification_dict)
 *             while modification_iterator.has_more():
 *                 ppos = modification_iterator.get_next_value(&pkey, &value)             # <<<<<<<<<<<<<<
 *                 if ppos != 0:
 *                     break
 */
      __pyx_v_ppos = ((struct __pyx_vtabstruct_14glycopeptidepy_2_c_11count_table_CountTableIterator *)__pyx_v_modification_iterator->__pyx_vtab)->get_next_value(__pyx_v_modification_iterator, (&__pyx_v_pkey), (&__pyx_v_value));

      /* "glycopeptidepy/_c/structure/fragment.pyx":537
 *             while modification_iterator.has_more():
 *                 ppos = modification_iterator.get_next_value(&pkey, &value)
 *                 if ppos != 0:             # <<<<<<<<<<<<<<
 *                     break
 *                 mod = <ModificationBase>pkey
 */
      __pyx_t_1 = ((__pyx_v_ppos != 0) != 0);
      if (__pyx_t_1) {

        /* "glycopeptidepy/_c/structure/fragment.pyx":538
 *                 ppos = modification_iterator.get_next_value(&pkey, &value)
 *                 if ppos != 0:
 *                     break             # <<<<<<<<<<<<<<
 *                 mod = <ModificationBase>pkey
 *                 if mod.is_a(ModificationCategory_glycosylation):
 */
        goto __pyx_L5_break;

        /* "glycopeptidepy/_c/structure/fragment.pyx":537
 *             while modification_iterator.has_more():
 *                 ppos = modification_iterator.get_next_value(&pkey, &value)
 *                 if ppos != 0:             # <<<<<<<<<<<<<<
 *                     break
 *                 mod = <ModificationBase>pkey
 */
      }

      /* "glycopeptidepy/_c/structure/fragment.pyx":539
 *                 if ppos != 0:
 *                     break
 *                 mod = <ModificationBase>pkey             # <<<<<<<<<<<<<<
 *                 if mod.is_a(ModificationCategory_glycosylation):
 *                     return True
 */
      __pyx_t_2 = ((PyObject *)__pyx_v_pkey);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_XDECREF_SET(__pyx_v_mod, ((struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_ModificationBase *)__pyx_t_2));
      __pyx_t_2 = 0;

      /* "glycopeptidepy/_c/structure/fragment.pyx":540
 *                     break
 *                 mod = <ModificationBase>pkey
 *                 if mod.is_a(ModificationCategory_glycosylation):             # <<<<<<<<<<<<<<
 *                     return True
 *         return False
 */
      __pyx_t_2 = __pyx_v_14glycopeptidepy_2_c_9structure_8fragment_ModificationCategory_glycosylation;
      __Pyx_INCREF(__pyx_t_2);
      __pyx_t_1 = (((struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_4base_ModificationBase *)__pyx_v_mod->__pyx_vtab)->is_a(__pyx_v_mod, __pyx_t_2, 0) != 0);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (__pyx_t_1) {

        /* "glycopeptidepy/_c/structure/fragment.pyx":541
 *                 mod = <ModificationBase>pkey
 *                 if mod.is_a(ModificationCategory_glycosylation):
 *                     return True             # <<<<<<<<<<<<<<
 *         return False
 * 
 */
        __pyx_r = 1;
        goto __pyx_L0;

        /* "glycopeptidepy/_c/structure/fragment.pyx":540
 *                     break
 *                 mod = <ModificationBase>pkey
 *                 if mod.is_a(ModificationCategory_glycosylation):             # <<<<<<<<<<<<<<
 *                     return True
 *         return False
 */
      }
    }
    __pyx_L5_break:;
  }

  /* "glycopeptidepy/_c/structure/fragment.pyx":542
 *                 if mod.is_a(ModificationCategory_glycosylation):
 *                     return True
 *         return False             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "glycopeptidepy/_c/structure/fragment.pyx":522
 *         return name
 * 
 *     cdef bint _is_glycosylated(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             PyObject *pkey
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_WriteUnraisable("glycopeptidepy._c.structure.fragment.PeptideFragment._is_glycosylated", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_mod);
  __Pyx_XDECREF((PyObject *)__pyx_v_modification_dict);
  __Pyx_XDECREF((PyObject *)__pyx_v_modification_iterator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/fragment.pyx":545
 * 
 *     @property
 *     def is_glycosylated(self):             # <<<<<<<<<<<<<<
 *         return self._is_glycosylated()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_15is_glycosylated_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_15is_glycosylated_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_15is_glycosylated___get__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_15is_glycosylated___get__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "glycopeptidepy/_c/structure/fragment.pyx":546
 *     @property
 *     def is_glycosylated(self):
 *         return self._is_glycosylated()             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(((struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *)__pyx_v_self->__pyx_base.__pyx_vtab)->_is_glycosylated(__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 546, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "glycopeptidepy/_c/structure/fragment.pyx":545
 * 
 *     @property
 *     def is_glycosylated(self):             # <<<<<<<<<<<<<<
 *         return self._is_glycosylated()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment.PeptideFragment.is_glycosylated.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/fragment.pyx":549
 * 
 *     @property
 *     def glycosylation_size(self):             # <<<<<<<<<<<<<<
 *         if self.glycosylation is not None:
 *             raise NotImplementedError()
 */

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_18glycosylation_size_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_18glycosylation_size_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_18glycosylation_size___get__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_18glycosylation_size___get__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *__pyx_v_self) {
  PyObject *__pyx_v_size = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  struct __pyx_opt_args_14glycopeptidepy_2_c_11count_table_10CountTable_get __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "glycopeptidepy/_c/structure/fragment.pyx":550
 *     @property
 *     def glycosylation_size(self):
 *         if self.glycosylation is not None:             # <<<<<<<<<<<<<<
 *             raise NotImplementedError()
 *         size = 0
 */
  __pyx_t_1 = (__pyx_v_self->glycosylation != ((PyObject*)Py_None));
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (unlikely(__pyx_t_2)) {

    /* "glycopeptidepy/_c/structure/fragment.pyx":551
 *     def glycosylation_size(self):
 *         if self.glycosylation is not None:
 *             raise NotImplementedError()             # <<<<<<<<<<<<<<
 *         size = 0
 *         size += self.modification_dict.get(_modification_hexnac, 0)
 */
    __pyx_t_3 = __Pyx_PyObject_CallNoArg(__pyx_builtin_NotImplementedError); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 551, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 551, __pyx_L1_error)

    /* "glycopeptidepy/_c/structure/fragment.pyx":550
 *     @property
 *     def glycosylation_size(self):
 *         if self.glycosylation is not None:             # <<<<<<<<<<<<<<
 *             raise NotImplementedError()
 *         size = 0
 */
  }

  /* "glycopeptidepy/_c/structure/fragment.pyx":552
 *         if self.glycosylation is not None:
 *             raise NotImplementedError()
 *         size = 0             # <<<<<<<<<<<<<<
 *         size += self.modification_dict.get(_modification_hexnac, 0)
 *         size += self.modification_dict.get(_modification_xylose, 0)
 */
  __Pyx_INCREF(__pyx_int_0);
  __pyx_v_size = __pyx_int_0;

  /* "glycopeptidepy/_c/structure/fragment.pyx":553
 *             raise NotImplementedError()
 *         size = 0
 *         size += self.modification_dict.get(_modification_hexnac, 0)             # <<<<<<<<<<<<<<
 *         size += self.modification_dict.get(_modification_xylose, 0)
 *         return size
 */
  __pyx_t_3 = ((PyObject *)__pyx_v_14glycopeptidepy_2_c_9structure_8fragment__modification_hexnac);
  __Pyx_INCREF(__pyx_t_3);
  __pyx_t_5.__pyx_n = 1;
  __pyx_t_5.__pyx_default = __pyx_int_0;
  __pyx_t_4 = ((struct __pyx_vtabstruct_14glycopeptidepy_2_c_11count_table_CountTable *)__pyx_v_self->modification_dict->__pyx_vtab)->get(__pyx_v_self->modification_dict, __pyx_t_3, 0, &__pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 553, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_size, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 553, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF_SET(__pyx_v_size, __pyx_t_3);
  __pyx_t_3 = 0;

  /* "glycopeptidepy/_c/structure/fragment.pyx":554
 *         size = 0
 *         size += self.modification_dict.get(_modification_hexnac, 0)
 *         size += self.modification_dict.get(_modification_xylose, 0)             # <<<<<<<<<<<<<<
 *         return size
 * 
 */
  __pyx_t_3 = ((PyObject *)__pyx_v_14glycopeptidepy_2_c_9structure_8fragment__modification_xylose);
  __Pyx_INCREF(__pyx_t_3);
  __pyx_t_5.__pyx_n = 1;
  __pyx_t_5.__pyx_default = __pyx_int_0;
  __pyx_t_4 = ((struct __pyx_vtabstruct_14glycopeptidepy_2_c_11count_table_CountTable *)__pyx_v_self->modification_dict->__pyx_vtab)->get(__pyx_v_self->modification_dict, __pyx_t_3, 0, &__pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 554, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_size, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 554, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF_SET(__pyx_v_size, __pyx_t_3);
  __pyx_t_3 = 0;

  /* "glycopeptidepy/_c/structure/fragment.pyx":555
 *         size += self.modification_dict.get(_modification_hexnac, 0)
 *         size += self.modification_dict.get(_modification_xylose, 0)
 *         return size             # <<<<<<<<<<<<<<
 * 
 *     cdef long get_glycosylation_size(self) except -1:
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_size);
  __pyx_r = __pyx_v_size;
  goto __pyx_L0;

  /* "glycopeptidepy/_c/structure/fragment.pyx":549
 * 
 *     @property
 *     def glycosylation_size(self):             # <<<<<<<<<<<<<<
 *         if self.glycosylation is not None:
 *             raise NotImplementedError()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment.PeptideFragment.glycosylation_size.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_size);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/fragment.pyx":557
 *         return size
 * 
 *     cdef long get_glycosylation_size(self) except -1:             # <<<<<<<<<<<<<<
 *         cdef:
 *             long size
 */

static long __pyx_f_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_get_glycosylation_size(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *__pyx_v_self) {
  long __pyx_v_size;
  long __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  long __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_glycosylation_size", 0);

  /* "glycopeptidepy/_c/structure/fragment.pyx":560
 *         cdef:
 *             long size
 *         if self.glycosylation is not None:             # <<<<<<<<<<<<<<
 *             PyErr_SetString(NotImplementedError, "Method not implemented for glycan collections yet")
 *             return -1
 */
  __pyx_t_1 = (__pyx_v_self->glycosylation != ((PyObject*)Py_None));
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "glycopeptidepy/_c/structure/fragment.pyx":561
 *             long size
 *         if self.glycosylation is not None:
 *             PyErr_SetString(NotImplementedError, "Method not implemented for glycan collections yet")             # <<<<<<<<<<<<<<
 *             return -1
 *         size = 0
 */
    PyErr_SetString(__pyx_builtin_NotImplementedError, ((char *)"Method not implemented for glycan collections yet"));

    /* "glycopeptidepy/_c/structure/fragment.pyx":562
 *         if self.glycosylation is not None:
 *             PyErr_SetString(NotImplementedError, "Method not implemented for glycan collections yet")
 *             return -1             # <<<<<<<<<<<<<<
 *         size = 0
 *         size += self.modification_dict.getitem(_modification_hexnac)
 */
    __pyx_r = -1L;
    goto __pyx_L0;

    /* "glycopeptidepy/_c/structure/fragment.pyx":560
 *         cdef:
 *             long size
 *         if self.glycosylation is not None:             # <<<<<<<<<<<<<<
 *             PyErr_SetString(NotImplementedError, "Method not implemented for glycan collections yet")
 *             return -1
 */
  }

  /* "glycopeptidepy/_c/structure/fragment.pyx":563
 *             PyErr_SetString(NotImplementedError, "Method not implemented for glycan collections yet")
 *             return -1
 *         size = 0             # <<<<<<<<<<<<<<
 *         size += self.modification_dict.getitem(_modification_hexnac)
 *         size += self.modification_dict.getitem(_modification_xylose)
 */
  __pyx_v_size = 0;

  /* "glycopeptidepy/_c/structure/fragment.pyx":564
 *             return -1
 *         size = 0
 *         size += self.modification_dict.getitem(_modification_hexnac)             # <<<<<<<<<<<<<<
 *         size += self.modification_dict.getitem(_modification_xylose)
 *         return size
 */
  __pyx_t_3 = ((PyObject *)__pyx_v_14glycopeptidepy_2_c_9structure_8fragment__modification_hexnac);
  __Pyx_INCREF(__pyx_t_3);
  __pyx_t_4 = ((struct __pyx_vtabstruct_14glycopeptidepy_2_c_11count_table_CountTable *)__pyx_v_self->modification_dict->__pyx_vtab)->getitem(__pyx_v_self->modification_dict, __pyx_t_3); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 564, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_size = (__pyx_v_size + __pyx_t_4);

  /* "glycopeptidepy/_c/structure/fragment.pyx":565
 *         size = 0
 *         size += self.modification_dict.getitem(_modification_hexnac)
 *         size += self.modification_dict.getitem(_modification_xylose)             # <<<<<<<<<<<<<<
 *         return size
 * 
 */
  __pyx_t_3 = ((PyObject *)__pyx_v_14glycopeptidepy_2_c_9structure_8fragment__modification_xylose);
  __Pyx_INCREF(__pyx_t_3);
  __pyx_t_4 = ((struct __pyx_vtabstruct_14glycopeptidepy_2_c_11count_table_CountTable *)__pyx_v_self->modification_dict->__pyx_vtab)->getitem(__pyx_v_self->modification_dict, __pyx_t_3); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 565, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_size = (__pyx_v_size + __pyx_t_4);

  /* "glycopeptidepy/_c/structure/fragment.pyx":566
 *         size += self.modification_dict.getitem(_modification_hexnac)
 *         size += self.modification_dict.getitem(_modification_xylose)
 *         return size             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
  __pyx_r = __pyx_v_size;
  goto __pyx_L0;

  /* "glycopeptidepy/_c/structure/fragment.pyx":557
 *         return size
 * 
 *     cdef long get_glycosylation_size(self) except -1:             # <<<<<<<<<<<<<<
 *         cdef:
 *             long size
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment.PeptideFragment.get_glycosylation_size", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1L;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/fragment.pyx":568
 *         return size
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         template = ("PeptideFragment({self.series}, {self.position}, "
 *                     "{self.mass}, {modification_dict}, {self.flanking_amino_acids},"
 */

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_15__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_15__repr__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_14__repr__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_14__repr__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *__pyx_v_self) {
  PyObject *__pyx_v_template = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "glycopeptidepy/_c/structure/fragment.pyx":569
 * 
 *     def __repr__(self):
 *         template = ("PeptideFragment({self.series}, {self.position}, "             # <<<<<<<<<<<<<<
 *                     "{self.mass}, {modification_dict}, {self.flanking_amino_acids},"
 *                     " {self.chemical_shift})")
 */
  __Pyx_INCREF(__pyx_kp_s_PeptideFragment_self_series_self);
  __pyx_v_template = __pyx_kp_s_PeptideFragment_self_series_self;

  /* "glycopeptidepy/_c/structure/fragment.pyx":572
 *                     "{self.mass}, {modification_dict}, {self.flanking_amino_acids},"
 *                     " {self.chemical_shift})")
 *         return template.format(self=self, modification_dict=dict(self.modification_dict))             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_template, __pyx_n_s_format); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 572, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 572, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_self, ((PyObject *)__pyx_v_self)) < 0) __PYX_ERR(0, 572, __pyx_L1_error)
  __pyx_t_3 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyDict_Type)), ((PyObject *)__pyx_v_self->modification_dict)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 572, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_modification_dict, __pyx_t_3) < 0) __PYX_ERR(0, 572, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_empty_tuple, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 572, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "glycopeptidepy/_c/structure/fragment.pyx":568
 *         return size
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         template = ("PeptideFragment({self.series}, {self.position}, "
 *                     "{self.mass}, {modification_dict}, {self.flanking_amino_acids},"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment.PeptideFragment.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_template);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/fragment.pxd":99
 * cdef class PeptideFragment(FragmentBase):
 *     cdef:
 *         public IonSeriesBase kind             # <<<<<<<<<<<<<<
 *         public int position
 *         public CountTable modification_dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_4kind_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_4kind_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_4kind___get__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_4kind___get__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->kind));
  __pyx_r = ((PyObject *)__pyx_v_self->kind);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_4kind_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_4kind_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_4kind_2__set__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_4kind_2__set__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);
  if (!(likely(((__pyx_v_value) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_value, __pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase))))) __PYX_ERR(1, 99, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_value;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->kind);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->kind));
  __pyx_v_self->kind = ((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment.PeptideFragment.kind.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_4kind_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_4kind_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_4kind_4__del__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_4kind_4__del__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->kind);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->kind));
  __pyx_v_self->kind = ((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *)Py_None);

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/fragment.pxd":100
 *     cdef:
 *         public IonSeriesBase kind
 *         public int position             # <<<<<<<<<<<<<<
 *         public CountTable modification_dict
 *         public double bare_mass
 */

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_8position_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_8position_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_8position___get__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_8position___get__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_self->position); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 100, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment.PeptideFragment.position.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_8position_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_8position_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_8position_2__set__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_8position_2__set__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyInt_As_int(__pyx_v_value); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 100, __pyx_L1_error)
  __pyx_v_self->position = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment.PeptideFragment.position.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/fragment.pxd":101
 *         public IonSeriesBase kind
 *         public int position
 *         public CountTable modification_dict             # <<<<<<<<<<<<<<
 *         public double bare_mass
 *         public list flanking_amino_acids
 */

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_17modification_dict_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_17modification_dict_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_17modification_dict___get__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_17modification_dict___get__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->modification_dict));
  __pyx_r = ((PyObject *)__pyx_v_self->modification_dict);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_17modification_dict_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_17modification_dict_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_17modification_dict_2__set__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_17modification_dict_2__set__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);
  if (!(likely(((__pyx_v_value) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_value, __pyx_ptype_14glycopeptidepy_2_c_11count_table_CountTable))))) __PYX_ERR(1, 101, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_value;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->modification_dict);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->modification_dict));
  __pyx_v_self->modification_dict = ((struct __pyx_obj_14glycopeptidepy_2_c_11count_table_CountTable *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment.PeptideFragment.modification_dict.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_17modification_dict_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_17modification_dict_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_17modification_dict_4__del__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_17modification_dict_4__del__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->modification_dict);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->modification_dict));
  __pyx_v_self->modification_dict = ((struct __pyx_obj_14glycopeptidepy_2_c_11count_table_CountTable *)Py_None);

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/fragment.pxd":102
 *         public int position
 *         public CountTable modification_dict
 *         public double bare_mass             # <<<<<<<<<<<<<<
 *         public list flanking_amino_acids
 *         public dict glycosylation
 */

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_9bare_mass_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_9bare_mass_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_9bare_mass___get__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_9bare_mass___get__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->bare_mass); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 102, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment.PeptideFragment.bare_mass.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_9bare_mass_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_9bare_mass_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_9bare_mass_2__set__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_9bare_mass_2__set__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  double __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_value); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 102, __pyx_L1_error)
  __pyx_v_self->bare_mass = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment.PeptideFragment.bare_mass.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/fragment.pxd":103
 *         public CountTable modification_dict
 *         public double bare_mass
 *         public list flanking_amino_acids             # <<<<<<<<<<<<<<
 *         public dict glycosylation
 *         public CComposition composition
 */

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_20flanking_amino_acids_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_20flanking_amino_acids_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_20flanking_amino_acids___get__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_20flanking_amino_acids___get__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->flanking_amino_acids);
  __pyx_r = __pyx_v_self->flanking_amino_acids;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_20flanking_amino_acids_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_20flanking_amino_acids_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_20flanking_amino_acids_2__set__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_20flanking_amino_acids_2__set__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);
  if (!(likely(PyList_CheckExact(__pyx_v_value))||((__pyx_v_value) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "list", Py_TYPE(__pyx_v_value)->tp_name), 0))) __PYX_ERR(1, 103, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_value;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->flanking_amino_acids);
  __Pyx_DECREF(__pyx_v_self->flanking_amino_acids);
  __pyx_v_self->flanking_amino_acids = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment.PeptideFragment.flanking_amino_acids.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_20flanking_amino_acids_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_20flanking_amino_acids_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_20flanking_amino_acids_4__del__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_20flanking_amino_acids_4__del__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->flanking_amino_acids);
  __Pyx_DECREF(__pyx_v_self->flanking_amino_acids);
  __pyx_v_self->flanking_amino_acids = ((PyObject*)Py_None);

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/fragment.pxd":104
 *         public double bare_mass
 *         public list flanking_amino_acids
 *         public dict glycosylation             # <<<<<<<<<<<<<<
 *         public CComposition composition
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_13glycosylation_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_13glycosylation_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_13glycosylation___get__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_13glycosylation___get__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->glycosylation);
  __pyx_r = __pyx_v_self->glycosylation;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_13glycosylation_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_13glycosylation_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_13glycosylation_2__set__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_13glycosylation_2__set__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);
  if (!(likely(PyDict_CheckExact(__pyx_v_value))||((__pyx_v_value) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "dict", Py_TYPE(__pyx_v_value)->tp_name), 0))) __PYX_ERR(1, 104, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_value;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->glycosylation);
  __Pyx_DECREF(__pyx_v_self->glycosylation);
  __pyx_v_self->glycosylation = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment.PeptideFragment.glycosylation.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_13glycosylation_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_13glycosylation_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_13glycosylation_4__del__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_13glycosylation_4__del__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->glycosylation);
  __Pyx_DECREF(__pyx_v_self->glycosylation);
  __pyx_v_self->glycosylation = ((PyObject*)Py_None);

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/fragment.pxd":105
 *         public list flanking_amino_acids
 *         public dict glycosylation
 *         public CComposition composition             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_11composition_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_11composition_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_11composition___get__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_11composition___get__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->composition));
  __pyx_r = ((PyObject *)__pyx_v_self->composition);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_11composition_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_11composition_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_11composition_2__set__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_11composition_2__set__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);
  if (!(likely(((__pyx_v_value) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_value, __pyx_ptype_5glypy_11composition_12ccomposition_CComposition))))) __PYX_ERR(1, 105, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_value;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->composition);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->composition));
  __pyx_v_self->composition = ((struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment.PeptideFragment.composition.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_11composition_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_11composition_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_11composition_4__del__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_11composition_4__del__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->composition);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->composition));
  __pyx_v_self->composition = ((struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *)Py_None);

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/fragment.pyx":577
 * cdef class SimpleFragment(FragmentBase):
 * 
 *     def __init__(self, name, mass, kind, composition, chemical_shift=None, is_glycosylated=False):             # <<<<<<<<<<<<<<
 *         self._name = name
 *         self._hash = hash(self._name)
 */

/* Python wrapper */
static int __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_14SimpleFragment_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_14SimpleFragment_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_name = 0;
  PyObject *__pyx_v_mass = 0;
  PyObject *__pyx_v_kind = 0;
  PyObject *__pyx_v_composition = 0;
  PyObject *__pyx_v_chemical_shift = 0;
  PyObject *__pyx_v_is_glycosylated = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_name_2,&__pyx_n_s_mass,&__pyx_n_s_kind,&__pyx_n_s_composition,&__pyx_n_s_chemical_shift,&__pyx_n_s_is_glycosylated,0};
    PyObject* values[6] = {0,0,0,0,0,0};
    values[4] = ((PyObject *)Py_None);
    values[5] = ((PyObject *)Py_False);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_name_2)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_mass)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 4, 6, 1); __PYX_ERR(0, 577, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_kind)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 4, 6, 2); __PYX_ERR(0, 577, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_composition)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 4, 6, 3); __PYX_ERR(0, 577, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_chemical_shift);
          if (value) { values[4] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_is_glycosylated);
          if (value) { values[5] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 577, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_name = values[0];
    __pyx_v_mass = values[1];
    __pyx_v_kind = values[2];
    __pyx_v_composition = values[3];
    __pyx_v_chemical_shift = values[4];
    __pyx_v_is_glycosylated = values[5];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 4, 6, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 577, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment.SimpleFragment.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_14SimpleFragment___init__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_SimpleFragment *)__pyx_v_self), __pyx_v_name, __pyx_v_mass, __pyx_v_kind, __pyx_v_composition, __pyx_v_chemical_shift, __pyx_v_is_glycosylated);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_14SimpleFragment___init__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_SimpleFragment *__pyx_v_self, PyObject *__pyx_v_name, PyObject *__pyx_v_mass, PyObject *__pyx_v_kind, PyObject *__pyx_v_composition, PyObject *__pyx_v_chemical_shift, PyObject *__pyx_v_is_glycosylated) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_hash_t __pyx_t_2;
  double __pyx_t_3;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "glycopeptidepy/_c/structure/fragment.pyx":578
 * 
 *     def __init__(self, name, mass, kind, composition, chemical_shift=None, is_glycosylated=False):
 *         self._name = name             # <<<<<<<<<<<<<<
 *         self._hash = hash(self._name)
 *         self._chemical_shift = None
 */
  if (!(likely(PyString_CheckExact(__pyx_v_name))||((__pyx_v_name) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "str", Py_TYPE(__pyx_v_name)->tp_name), 0))) __PYX_ERR(0, 578, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_name;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->__pyx_base._name);
  __Pyx_DECREF(__pyx_v_self->__pyx_base._name);
  __pyx_v_self->__pyx_base._name = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "glycopeptidepy/_c/structure/fragment.pyx":579
 *     def __init__(self, name, mass, kind, composition, chemical_shift=None, is_glycosylated=False):
 *         self._name = name
 *         self._hash = hash(self._name)             # <<<<<<<<<<<<<<
 *         self._chemical_shift = None
 * 
 */
  __pyx_t_1 = __pyx_v_self->__pyx_base._name;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = PyObject_Hash(__pyx_t_1); if (unlikely(__pyx_t_2 == ((Py_hash_t)-1))) __PYX_ERR(0, 579, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_self->__pyx_base._hash = __pyx_t_2;

  /* "glycopeptidepy/_c/structure/fragment.pyx":580
 *         self._name = name
 *         self._hash = hash(self._name)
 *         self._chemical_shift = None             # <<<<<<<<<<<<<<
 * 
 *         self.mass = mass
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->__pyx_base._chemical_shift);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->__pyx_base._chemical_shift));
  __pyx_v_self->__pyx_base._chemical_shift = ((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase *)Py_None);

  /* "glycopeptidepy/_c/structure/fragment.pyx":582
 *         self._chemical_shift = None
 * 
 *         self.mass = mass             # <<<<<<<<<<<<<<
 *         self.kind = kind
 *         self.composition = composition
 */
  __pyx_t_3 = __pyx_PyFloat_AsDouble(__pyx_v_mass); if (unlikely((__pyx_t_3 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 582, __pyx_L1_error)
  __pyx_v_self->__pyx_base.mass = __pyx_t_3;

  /* "glycopeptidepy/_c/structure/fragment.pyx":583
 * 
 *         self.mass = mass
 *         self.kind = kind             # <<<<<<<<<<<<<<
 *         self.composition = composition
 * 
 */
  if (!(likely(((__pyx_v_kind) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_kind, __pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase))))) __PYX_ERR(0, 583, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_kind;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->kind);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->kind));
  __pyx_v_self->kind = ((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "glycopeptidepy/_c/structure/fragment.pyx":584
 *         self.mass = mass
 *         self.kind = kind
 *         self.composition = composition             # <<<<<<<<<<<<<<
 * 
 *         self.set_chemical_shift(chemical_shift)
 */
  if (!(likely(((__pyx_v_composition) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_composition, __pyx_ptype_5glypy_11composition_12ccomposition_CComposition))))) __PYX_ERR(0, 584, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_composition;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->composition);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->composition));
  __pyx_v_self->composition = ((struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "glycopeptidepy/_c/structure/fragment.pyx":586
 *         self.composition = composition
 * 
 *         self.set_chemical_shift(chemical_shift)             # <<<<<<<<<<<<<<
 *         self.is_glycosylated = is_glycosylated
 * 
 */
  if (!(likely(((__pyx_v_chemical_shift) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_chemical_shift, __pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase))))) __PYX_ERR(0, 586, __pyx_L1_error)
  __pyx_t_1 = ((struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_8fragment_SimpleFragment *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.set_chemical_shift(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *)__pyx_v_self), ((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase *)__pyx_v_chemical_shift), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 586, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "glycopeptidepy/_c/structure/fragment.pyx":587
 * 
 *         self.set_chemical_shift(chemical_shift)
 *         self.is_glycosylated = is_glycosylated             # <<<<<<<<<<<<<<
 * 
 *     cpdef clone(self):
 */
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_is_glycosylated); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 587, __pyx_L1_error)
  __pyx_v_self->is_glycosylated = __pyx_t_4;

  /* "glycopeptidepy/_c/structure/fragment.pyx":577
 * cdef class SimpleFragment(FragmentBase):
 * 
 *     def __init__(self, name, mass, kind, composition, chemical_shift=None, is_glycosylated=False):             # <<<<<<<<<<<<<<
 *         self._name = name
 *         self._hash = hash(self._name)
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment.SimpleFragment.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/fragment.pyx":589
 *         self.is_glycosylated = is_glycosylated
 * 
 *     cpdef clone(self):             # <<<<<<<<<<<<<<
 *         corrected_mass = self.mass
 *         if self._chemical_shift is not None:
 */

static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_14SimpleFragment_3clone(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_14SimpleFragment_clone(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_SimpleFragment *__pyx_v_self, int __pyx_skip_dispatch) {
  PyObject *__pyx_v_corrected_mass = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("clone", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_clone); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 589, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_14SimpleFragment_3clone)) {
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 589, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "glycopeptidepy/_c/structure/fragment.pyx":590
 * 
 *     cpdef clone(self):
 *         corrected_mass = self.mass             # <<<<<<<<<<<<<<
 *         if self._chemical_shift is not None:
 *             corrected_mass -= self.chemical_shift.mass
 */
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->__pyx_base.mass); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 590, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_corrected_mass = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "glycopeptidepy/_c/structure/fragment.pyx":591
 *     cpdef clone(self):
 *         corrected_mass = self.mass
 *         if self._chemical_shift is not None:             # <<<<<<<<<<<<<<
 *             corrected_mass -= self.chemical_shift.mass
 *         return self.__class__(self.name, corrected_mass, self.kind,
 */
  __pyx_t_5 = (((PyObject *)__pyx_v_self->__pyx_base._chemical_shift) != Py_None);
  __pyx_t_6 = (__pyx_t_5 != 0);
  if (__pyx_t_6) {

    /* "glycopeptidepy/_c/structure/fragment.pyx":592
 *         corrected_mass = self.mass
 *         if self._chemical_shift is not None:
 *             corrected_mass -= self.chemical_shift.mass             # <<<<<<<<<<<<<<
 *         return self.__class__(self.name, corrected_mass, self.kind,
 *                               self.composition,
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_chemical_shift); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 592, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_mass); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 592, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = PyNumber_InPlaceSubtract(__pyx_v_corrected_mass, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 592, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF_SET(__pyx_v_corrected_mass, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "glycopeptidepy/_c/structure/fragment.pyx":591
 *     cpdef clone(self):
 *         corrected_mass = self.mass
 *         if self._chemical_shift is not None:             # <<<<<<<<<<<<<<
 *             corrected_mass -= self.chemical_shift.mass
 *         return self.__class__(self.name, corrected_mass, self.kind,
 */
  }

  /* "glycopeptidepy/_c/structure/fragment.pyx":593
 *         if self._chemical_shift is not None:
 *             corrected_mass -= self.chemical_shift.mass
 *         return self.__class__(self.name, corrected_mass, self.kind,             # <<<<<<<<<<<<<<
 *                               self.composition,
 *                               self._chemical_shift.clone() if self._chemical_shift is not None else None,
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_class); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 593, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_name_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 593, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);

  /* "glycopeptidepy/_c/structure/fragment.pyx":595
 *         return self.__class__(self.name, corrected_mass, self.kind,
 *                               self.composition,
 *                               self._chemical_shift.clone() if self._chemical_shift is not None else None,             # <<<<<<<<<<<<<<
 *                               self.is_glycosylated)
 * 
 */
  __pyx_t_6 = (((PyObject *)__pyx_v_self->__pyx_base._chemical_shift) != Py_None);
  if ((__pyx_t_6 != 0)) {
    __pyx_t_7 = ((PyObject *)((struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase *)__pyx_v_self->__pyx_base._chemical_shift->__pyx_vtab)->clone(__pyx_v_self->__pyx_base._chemical_shift, 0)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 595, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_4 = __pyx_t_7;
    __pyx_t_7 = 0;
  } else {
    __Pyx_INCREF(Py_None);
    __pyx_t_4 = Py_None;
  }

  /* "glycopeptidepy/_c/structure/fragment.pyx":596
 *                               self.composition,
 *                               self._chemical_shift.clone() if self._chemical_shift is not None else None,
 *                               self.is_glycosylated)             # <<<<<<<<<<<<<<
 * 
 *     def __reduce__(self):
 */
  __pyx_t_7 = __Pyx_PyBool_FromLong(__pyx_v_self->is_glycosylated); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 596, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = NULL;
  __pyx_t_9 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_8)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_8);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_9 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[7] = {__pyx_t_8, __pyx_t_3, __pyx_v_corrected_mass, ((PyObject *)__pyx_v_self->kind), ((PyObject *)__pyx_v_self->composition), __pyx_t_4, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_9, 6+__pyx_t_9); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 593, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[7] = {__pyx_t_8, __pyx_t_3, __pyx_v_corrected_mass, ((PyObject *)__pyx_v_self->kind), ((PyObject *)__pyx_v_self->composition), __pyx_t_4, __pyx_t_7};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_9, 6+__pyx_t_9); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 593, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else
  #endif
  {
    __pyx_t_10 = PyTuple_New(6+__pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 593, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    if (__pyx_t_8) {
      __Pyx_GIVEREF(__pyx_t_8); PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_8); __pyx_t_8 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_10, 0+__pyx_t_9, __pyx_t_3);
    __Pyx_INCREF(__pyx_v_corrected_mass);
    __Pyx_GIVEREF(__pyx_v_corrected_mass);
    PyTuple_SET_ITEM(__pyx_t_10, 1+__pyx_t_9, __pyx_v_corrected_mass);
    __Pyx_INCREF(((PyObject *)__pyx_v_self->kind));
    __Pyx_GIVEREF(((PyObject *)__pyx_v_self->kind));
    PyTuple_SET_ITEM(__pyx_t_10, 2+__pyx_t_9, ((PyObject *)__pyx_v_self->kind));
    __Pyx_INCREF(((PyObject *)__pyx_v_self->composition));
    __Pyx_GIVEREF(((PyObject *)__pyx_v_self->composition));
    PyTuple_SET_ITEM(__pyx_t_10, 3+__pyx_t_9, ((PyObject *)__pyx_v_self->composition));
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_10, 4+__pyx_t_9, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_7);
    PyTuple_SET_ITEM(__pyx_t_10, 5+__pyx_t_9, __pyx_t_7);
    __pyx_t_3 = 0;
    __pyx_t_4 = 0;
    __pyx_t_7 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_10, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 593, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "glycopeptidepy/_c/structure/fragment.pyx":589
 *         self.is_glycosylated = is_glycosylated
 * 
 *     cpdef clone(self):             # <<<<<<<<<<<<<<
 *         corrected_mass = self.mass
 *         if self._chemical_shift is not None:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment.SimpleFragment.clone", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_corrected_mass);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_14SimpleFragment_3clone(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_14glycopeptidepy_2_c_9structure_8fragment_14SimpleFragment_2clone[] = "SimpleFragment.clone(self)";
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_14SimpleFragment_3clone(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("clone (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_14SimpleFragment_2clone(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_SimpleFragment *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_14SimpleFragment_2clone(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_SimpleFragment *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("clone", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_14glycopeptidepy_2_c_9structure_8fragment_14SimpleFragment_clone(__pyx_v_self, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 589, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment.SimpleFragment.clone", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/fragment.pyx":598
 *                               self.is_glycosylated)
 * 
 *     def __reduce__(self):             # <<<<<<<<<<<<<<
 *         corrected_mass = self.mass
 *         if self._chemical_shift is not None:
 */

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_14SimpleFragment_5__reduce__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_14glycopeptidepy_2_c_9structure_8fragment_14SimpleFragment_4__reduce__[] = "SimpleFragment.__reduce__(self)";
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_14SimpleFragment_5__reduce__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_14SimpleFragment_4__reduce__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_SimpleFragment *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_14SimpleFragment_4__reduce__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_SimpleFragment *__pyx_v_self) {
  PyObject *__pyx_v_corrected_mass = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce__", 0);

  /* "glycopeptidepy/_c/structure/fragment.pyx":599
 * 
 *     def __reduce__(self):
 *         corrected_mass = self.mass             # <<<<<<<<<<<<<<
 *         if self._chemical_shift is not None:
 *             corrected_mass -= self.chemical_shift.mass
 */
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->__pyx_base.mass); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 599, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_corrected_mass = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "glycopeptidepy/_c/structure/fragment.pyx":600
 *     def __reduce__(self):
 *         corrected_mass = self.mass
 *         if self._chemical_shift is not None:             # <<<<<<<<<<<<<<
 *             corrected_mass -= self.chemical_shift.mass
 *         return self.__class__, (self.name, corrected_mass, self.kind, self.composition,
 */
  __pyx_t_2 = (((PyObject *)__pyx_v_self->__pyx_base._chemical_shift) != Py_None);
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {

    /* "glycopeptidepy/_c/structure/fragment.pyx":601
 *         corrected_mass = self.mass
 *         if self._chemical_shift is not None:
 *             corrected_mass -= self.chemical_shift.mass             # <<<<<<<<<<<<<<
 *         return self.__class__, (self.name, corrected_mass, self.kind, self.composition,
 *                                 self.chemical_shift, self.is_glycosylated)
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_chemical_shift); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 601, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_mass); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 601, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = PyNumber_InPlaceSubtract(__pyx_v_corrected_mass, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 601, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF_SET(__pyx_v_corrected_mass, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "glycopeptidepy/_c/structure/fragment.pyx":600
 *     def __reduce__(self):
 *         corrected_mass = self.mass
 *         if self._chemical_shift is not None:             # <<<<<<<<<<<<<<
 *             corrected_mass -= self.chemical_shift.mass
 *         return self.__class__, (self.name, corrected_mass, self.kind, self.composition,
 */
  }

  /* "glycopeptidepy/_c/structure/fragment.pyx":602
 *         if self._chemical_shift is not None:
 *             corrected_mass -= self.chemical_shift.mass
 *         return self.__class__, (self.name, corrected_mass, self.kind, self.composition,             # <<<<<<<<<<<<<<
 *                                 self.chemical_shift, self.is_glycosylated)
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_class); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 602, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_name_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 602, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);

  /* "glycopeptidepy/_c/structure/fragment.pyx":603
 *             corrected_mass -= self.chemical_shift.mass
 *         return self.__class__, (self.name, corrected_mass, self.kind, self.composition,
 *                                 self.chemical_shift, self.is_glycosylated)             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_chemical_shift); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 603, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_PyBool_FromLong(__pyx_v_self->is_glycosylated); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 603, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);

  /* "glycopeptidepy/_c/structure/fragment.pyx":602
 *         if self._chemical_shift is not None:
 *             corrected_mass -= self.chemical_shift.mass
 *         return self.__class__, (self.name, corrected_mass, self.kind, self.composition,             # <<<<<<<<<<<<<<
 *                                 self.chemical_shift, self.is_glycosylated)
 * 
 */
  __pyx_t_7 = PyTuple_New(6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 602, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_4);
  __Pyx_INCREF(__pyx_v_corrected_mass);
  __Pyx_GIVEREF(__pyx_v_corrected_mass);
  PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_v_corrected_mass);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->kind));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self->kind));
  PyTuple_SET_ITEM(__pyx_t_7, 2, ((PyObject *)__pyx_v_self->kind));
  __Pyx_INCREF(((PyObject *)__pyx_v_self->composition));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self->composition));
  PyTuple_SET_ITEM(__pyx_t_7, 3, ((PyObject *)__pyx_v_self->composition));
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_7, 4, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_7, 5, __pyx_t_6);
  __pyx_t_4 = 0;
  __pyx_t_5 = 0;
  __pyx_t_6 = 0;
  __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 602, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_7);
  __pyx_t_1 = 0;
  __pyx_t_7 = 0;
  __pyx_r = __pyx_t_6;
  __pyx_t_6 = 0;
  goto __pyx_L0;

  /* "glycopeptidepy/_c/structure/fragment.pyx":598
 *                               self.is_glycosylated)
 * 
 *     def __reduce__(self):             # <<<<<<<<<<<<<<
 *         corrected_mass = self.mass
 *         if self._chemical_shift is not None:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment.SimpleFragment.__reduce__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_corrected_mass);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/fragment.pyx":605
 *                                 self.chemical_shift, self.is_glycosylated)
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return ("{self.__class__.__name__}(name={self.name}, "
 *                 "mass={self.mass:.04f}, series={self.kind})").format(self=self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_14SimpleFragment_7__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_14SimpleFragment_7__repr__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_14SimpleFragment_6__repr__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_SimpleFragment *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_14SimpleFragment_6__repr__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_SimpleFragment *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "glycopeptidepy/_c/structure/fragment.pyx":606
 * 
 *     def __repr__(self):
 *         return ("{self.__class__.__name__}(name={self.name}, "             # <<<<<<<<<<<<<<
 *                 "mass={self.mass:.04f}, series={self.kind})").format(self=self)
 * 
 */
  __Pyx_XDECREF(__pyx_r);

  /* "glycopeptidepy/_c/structure/fragment.pyx":607
 *     def __repr__(self):
 *         return ("{self.__class__.__name__}(name={self.name}, "
 *                 "mass={self.mass:.04f}, series={self.kind})").format(self=self)             # <<<<<<<<<<<<<<
 * 
 *     cpdef IonSeriesBase get_series(self):
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_kp_s_self___class_____name___name_se, __pyx_n_s_format); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 607, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 607, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_self, ((PyObject *)__pyx_v_self)) < 0) __PYX_ERR(0, 607, __pyx_L1_error)
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_empty_tuple, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 607, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "glycopeptidepy/_c/structure/fragment.pyx":605
 *                                 self.chemical_shift, self.is_glycosylated)
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return ("{self.__class__.__name__}(name={self.name}, "
 *                 "mass={self.mass:.04f}, series={self.kind})").format(self=self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment.SimpleFragment.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/fragment.pyx":609
 *                 "mass={self.mass:.04f}, series={self.kind})").format(self=self)
 * 
 *     cpdef IonSeriesBase get_series(self):             # <<<<<<<<<<<<<<
 *         return self.kind
 * 
 */

static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_14SimpleFragment_9get_series(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_14SimpleFragment_get_series(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_SimpleFragment *__pyx_v_self, int __pyx_skip_dispatch) {
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_series", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_get_series); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 609, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_14SimpleFragment_9get_series)) {
        __Pyx_XDECREF(((PyObject *)__pyx_r));
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 609, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase))))) __PYX_ERR(0, 609, __pyx_L1_error)
        __pyx_r = ((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "glycopeptidepy/_c/structure/fragment.pyx":610
 * 
 *     cpdef IonSeriesBase get_series(self):
 *         return self.kind             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __Pyx_INCREF(((PyObject *)__pyx_v_self->kind));
  __pyx_r = __pyx_v_self->kind;
  goto __pyx_L0;

  /* "glycopeptidepy/_c/structure/fragment.pyx":609
 *                 "mass={self.mass:.04f}, series={self.kind})").format(self=self)
 * 
 *     cpdef IonSeriesBase get_series(self):             # <<<<<<<<<<<<<<
 *         return self.kind
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment.SimpleFragment.get_series", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_14SimpleFragment_9get_series(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_14glycopeptidepy_2_c_9structure_8fragment_14SimpleFragment_8get_series[] = "SimpleFragment.get_series(self) -> IonSeriesBase";
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_14SimpleFragment_9get_series(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_series (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_14SimpleFragment_8get_series(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_SimpleFragment *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_14SimpleFragment_8get_series(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_SimpleFragment *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_series", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_14SimpleFragment_get_series(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 609, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment.SimpleFragment.get_series", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/fragment.pxd":65
 * cdef class SimpleFragment(FragmentBase):
 *     cdef:
 *         public IonSeriesBase kind             # <<<<<<<<<<<<<<
 *         public bint is_glycosylated
 *         public CComposition composition
 */

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_14SimpleFragment_4kind_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_14SimpleFragment_4kind_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_14SimpleFragment_4kind___get__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_SimpleFragment *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_14SimpleFragment_4kind___get__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_SimpleFragment *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->kind));
  __pyx_r = ((PyObject *)__pyx_v_self->kind);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_14SimpleFragment_4kind_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_14SimpleFragment_4kind_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_14SimpleFragment_4kind_2__set__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_SimpleFragment *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_14SimpleFragment_4kind_2__set__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_SimpleFragment *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);
  if (!(likely(((__pyx_v_value) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_value, __pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase))))) __PYX_ERR(1, 65, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_value;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->kind);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->kind));
  __pyx_v_self->kind = ((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment.SimpleFragment.kind.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_14SimpleFragment_4kind_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_14SimpleFragment_4kind_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_14SimpleFragment_4kind_4__del__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_SimpleFragment *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_14SimpleFragment_4kind_4__del__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_SimpleFragment *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->kind);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->kind));
  __pyx_v_self->kind = ((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *)Py_None);

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/fragment.pxd":66
 *     cdef:
 *         public IonSeriesBase kind
 *         public bint is_glycosylated             # <<<<<<<<<<<<<<
 *         public CComposition composition
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_14SimpleFragment_15is_glycosylated_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_14SimpleFragment_15is_glycosylated_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_14SimpleFragment_15is_glycosylated___get__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_SimpleFragment *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_14SimpleFragment_15is_glycosylated___get__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_SimpleFragment *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->is_glycosylated); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 66, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment.SimpleFragment.is_glycosylated.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_14SimpleFragment_15is_glycosylated_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_14SimpleFragment_15is_glycosylated_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_14SimpleFragment_15is_glycosylated_2__set__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_SimpleFragment *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_14SimpleFragment_15is_glycosylated_2__set__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_SimpleFragment *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_value); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 66, __pyx_L1_error)
  __pyx_v_self->is_glycosylated = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment.SimpleFragment.is_glycosylated.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/fragment.pxd":67
 *         public IonSeriesBase kind
 *         public bint is_glycosylated
 *         public CComposition composition             # <<<<<<<<<<<<<<
 * 
 *     cpdef clone(self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_14SimpleFragment_11composition_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_14SimpleFragment_11composition_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_14SimpleFragment_11composition___get__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_SimpleFragment *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_14SimpleFragment_11composition___get__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_SimpleFragment *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->composition));
  __pyx_r = ((PyObject *)__pyx_v_self->composition);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_14SimpleFragment_11composition_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_14SimpleFragment_11composition_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_14SimpleFragment_11composition_2__set__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_SimpleFragment *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_14SimpleFragment_11composition_2__set__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_SimpleFragment *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);
  if (!(likely(((__pyx_v_value) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_value, __pyx_ptype_5glypy_11composition_12ccomposition_CComposition))))) __PYX_ERR(1, 67, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_value;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->composition);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->composition));
  __pyx_v_self->composition = ((struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment.SimpleFragment.composition.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_14SimpleFragment_11composition_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_14SimpleFragment_11composition_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_14SimpleFragment_11composition_4__del__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_SimpleFragment *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_14SimpleFragment_11composition_4__del__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_SimpleFragment *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->composition);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->composition));
  __pyx_v_self->composition = ((struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *)Py_None);

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/fragment.pyx":616
 * cdef class _NameTree(object):
 *     @staticmethod
 *     cdef _NameTree _create():             # <<<<<<<<<<<<<<
 *         cdef _NameTree self = _NameTree.__new__(_NameTree)
 *         self.name = None
 */

static struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment__NameTree *__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree__create(void) {
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment__NameTree *__pyx_v_self = 0;
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment__NameTree *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_create", 0);

  /* "glycopeptidepy/_c/structure/fragment.pyx":617
 *     @staticmethod
 *     cdef _NameTree _create():
 *         cdef _NameTree self = _NameTree.__new__(_NameTree)             # <<<<<<<<<<<<<<
 *         self.name = None
 *         self.value = None
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_14glycopeptidepy_2_c_9structure_8fragment__NameTree(((PyTypeObject *)__pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment__NameTree), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 617, __pyx_L1_error)
  __Pyx_GOTREF(((PyObject *)__pyx_t_1));
  __pyx_v_self = ((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment__NameTree *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "glycopeptidepy/_c/structure/fragment.pyx":618
 *     cdef _NameTree _create():
 *         cdef _NameTree self = _NameTree.__new__(_NameTree)
 *         self.name = None             # <<<<<<<<<<<<<<
 *         self.value = None
 *         self.children = dict()
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->name);
  __Pyx_DECREF(__pyx_v_self->name);
  __pyx_v_self->name = Py_None;

  /* "glycopeptidepy/_c/structure/fragment.pyx":619
 *         cdef _NameTree self = _NameTree.__new__(_NameTree)
 *         self.name = None
 *         self.value = None             # <<<<<<<<<<<<<<
 *         self.children = dict()
 *         return self
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->value);
  __Pyx_DECREF(__pyx_v_self->value);
  __pyx_v_self->value = Py_None;

  /* "glycopeptidepy/_c/structure/fragment.pyx":620
 *         self.name = None
 *         self.value = None
 *         self.children = dict()             # <<<<<<<<<<<<<<
 *         return self
 * 
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 620, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->children);
  __Pyx_DECREF(__pyx_v_self->children);
  __pyx_v_self->children = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "glycopeptidepy/_c/structure/fragment.pyx":621
 *         self.value = None
 *         self.children = dict()
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, name=None, children=None):
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __Pyx_INCREF(((PyObject *)__pyx_v_self));
  __pyx_r = __pyx_v_self;
  goto __pyx_L0;

  /* "glycopeptidepy/_c/structure/fragment.pyx":616
 * cdef class _NameTree(object):
 *     @staticmethod
 *     cdef _NameTree _create():             # <<<<<<<<<<<<<<
 *         cdef _NameTree self = _NameTree.__new__(_NameTree)
 *         self.name = None
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment._NameTree._create", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_self);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/fragment.pyx":623
 *         return self
 * 
 *     def __init__(self, name=None, children=None):             # <<<<<<<<<<<<<<
 *         if children is None:
 *             children = dict()
 */

/* Python wrapper */
static int __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_name = 0;
  PyObject *__pyx_v_children = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_name_2,&__pyx_n_s_children,0};
    PyObject* values[2] = {0,0};
    values[0] = ((PyObject *)Py_None);
    values[1] = ((PyObject *)Py_None);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_name_2);
          if (value) { values[0] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_children);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 623, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_name = values[0];
    __pyx_v_children = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 0, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 623, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment._NameTree.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree___init__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment__NameTree *)__pyx_v_self), __pyx_v_name, __pyx_v_children);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree___init__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment__NameTree *__pyx_v_self, PyObject *__pyx_v_name, PyObject *__pyx_v_children) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);
  __Pyx_INCREF(__pyx_v_children);

  /* "glycopeptidepy/_c/structure/fragment.pyx":624
 * 
 *     def __init__(self, name=None, children=None):
 *         if children is None:             # <<<<<<<<<<<<<<
 *             children = dict()
 *         self.name = name
 */
  __pyx_t_1 = (__pyx_v_children == Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "glycopeptidepy/_c/structure/fragment.pyx":625
 *     def __init__(self, name=None, children=None):
 *         if children is None:
 *             children = dict()             # <<<<<<<<<<<<<<
 *         self.name = name
 *         self.value = None
 */
    __pyx_t_3 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 625, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF_SET(__pyx_v_children, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "glycopeptidepy/_c/structure/fragment.pyx":624
 * 
 *     def __init__(self, name=None, children=None):
 *         if children is None:             # <<<<<<<<<<<<<<
 *             children = dict()
 *         self.name = name
 */
  }

  /* "glycopeptidepy/_c/structure/fragment.pyx":626
 *         if children is None:
 *             children = dict()
 *         self.name = name             # <<<<<<<<<<<<<<
 *         self.value = None
 *         self.children = children
 */
  __Pyx_INCREF(__pyx_v_name);
  __Pyx_GIVEREF(__pyx_v_name);
  __Pyx_GOTREF(__pyx_v_self->name);
  __Pyx_DECREF(__pyx_v_self->name);
  __pyx_v_self->name = __pyx_v_name;

  /* "glycopeptidepy/_c/structure/fragment.pyx":627
 *             children = dict()
 *         self.name = name
 *         self.value = None             # <<<<<<<<<<<<<<
 *         self.children = children
 * 
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->value);
  __Pyx_DECREF(__pyx_v_self->value);
  __pyx_v_self->value = Py_None;

  /* "glycopeptidepy/_c/structure/fragment.pyx":628
 *         self.name = name
 *         self.value = None
 *         self.children = children             # <<<<<<<<<<<<<<
 * 
 *     def __getitem__(self, key):
 */
  if (!(likely(PyDict_CheckExact(__pyx_v_children))||((__pyx_v_children) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "dict", Py_TYPE(__pyx_v_children)->tp_name), 0))) __PYX_ERR(0, 628, __pyx_L1_error)
  __pyx_t_3 = __pyx_v_children;
  __Pyx_INCREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __Pyx_GOTREF(__pyx_v_self->children);
  __Pyx_DECREF(__pyx_v_self->children);
  __pyx_v_self->children = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "glycopeptidepy/_c/structure/fragment.pyx":623
 *         return self
 * 
 *     def __init__(self, name=None, children=None):             # <<<<<<<<<<<<<<
 *         if children is None:
 *             children = dict()
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment._NameTree.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_children);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/fragment.pyx":630
 *         self.children = children
 * 
 *     def __getitem__(self, key):             # <<<<<<<<<<<<<<
 *         return self.get(key)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_3__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_key); /*proto*/
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_3__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_key) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getitem__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_2__getitem__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment__NameTree *)__pyx_v_self), ((PyObject *)__pyx_v_key));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_2__getitem__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment__NameTree *__pyx_v_self, PyObject *__pyx_v_key) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__getitem__", 0);

  /* "glycopeptidepy/_c/structure/fragment.pyx":631
 * 
 *     def __getitem__(self, key):
 *         return self.get(key)             # <<<<<<<<<<<<<<
 * 
 *     cpdef _NameTree get(self, key):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_get(__pyx_v_self, __pyx_v_key, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 631, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "glycopeptidepy/_c/structure/fragment.pyx":630
 *         self.children = children
 * 
 *     def __getitem__(self, key):             # <<<<<<<<<<<<<<
 *         return self.get(key)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment._NameTree.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/fragment.pyx":633
 *         return self.get(key)
 * 
 *     cpdef _NameTree get(self, key):             # <<<<<<<<<<<<<<
 *         cdef:
 *             PyObject* presult
 */

static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_5get(PyObject *__pyx_v_self, PyObject *__pyx_v_key); /*proto*/
static struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment__NameTree *__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_get(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment__NameTree *__pyx_v_self, PyObject *__pyx_v_key, CYTHON_UNUSED int __pyx_skip_dispatch) {
  PyObject *__pyx_v_presult;
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment__NameTree *__pyx_v_result = 0;
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment__NameTree *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get", 0);

  /* "glycopeptidepy/_c/structure/fragment.pyx":637
 *             PyObject* presult
 *             _NameTree result
 *         presult = PyDict_GetItem(self.children, key)             # <<<<<<<<<<<<<<
 *         if presult == NULL:
 *             result = _NameTree._create()
 */
  __pyx_t_1 = __pyx_v_self->children;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_presult = PyDict_GetItem(__pyx_t_1, __pyx_v_key);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "glycopeptidepy/_c/structure/fragment.pyx":638
 *             _NameTree result
 *         presult = PyDict_GetItem(self.children, key)
 *         if presult == NULL:             # <<<<<<<<<<<<<<
 *             result = _NameTree._create()
 *             PyDict_SetItem(self.children, key, result)
 */
  __pyx_t_2 = ((__pyx_v_presult == NULL) != 0);
  if (__pyx_t_2) {

    /* "glycopeptidepy/_c/structure/fragment.pyx":639
 *         presult = PyDict_GetItem(self.children, key)
 *         if presult == NULL:
 *             result = _NameTree._create()             # <<<<<<<<<<<<<<
 *             PyDict_SetItem(self.children, key, result)
 *             return result
 */
    __pyx_t_1 = ((PyObject *)__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree__create()); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 639, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_result = ((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment__NameTree *)__pyx_t_1);
    __pyx_t_1 = 0;

    /* "glycopeptidepy/_c/structure/fragment.pyx":640
 *         if presult == NULL:
 *             result = _NameTree._create()
 *             PyDict_SetItem(self.children, key, result)             # <<<<<<<<<<<<<<
 *             return result
 *         else:
 */
    __pyx_t_1 = __pyx_v_self->children;
    __Pyx_INCREF(__pyx_t_1);
    __pyx_t_3 = PyDict_SetItem(__pyx_t_1, __pyx_v_key, ((PyObject *)__pyx_v_result)); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(0, 640, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "glycopeptidepy/_c/structure/fragment.pyx":641
 *             result = _NameTree._create()
 *             PyDict_SetItem(self.children, key, result)
 *             return result             # <<<<<<<<<<<<<<
 *         else:
 *             result = <_NameTree>presult
 */
    __Pyx_XDECREF(((PyObject *)__pyx_r));
    __Pyx_INCREF(((PyObject *)__pyx_v_result));
    __pyx_r = __pyx_v_result;
    goto __pyx_L0;

    /* "glycopeptidepy/_c/structure/fragment.pyx":638
 *             _NameTree result
 *         presult = PyDict_GetItem(self.children, key)
 *         if presult == NULL:             # <<<<<<<<<<<<<<
 *             result = _NameTree._create()
 *             PyDict_SetItem(self.children, key, result)
 */
  }

  /* "glycopeptidepy/_c/structure/fragment.pyx":643
 *             return result
 *         else:
 *             result = <_NameTree>presult             # <<<<<<<<<<<<<<
 *             return result
 * 
 */
  /*else*/ {
    __pyx_t_1 = ((PyObject *)__pyx_v_presult);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v_result = ((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment__NameTree *)__pyx_t_1);
    __pyx_t_1 = 0;

    /* "glycopeptidepy/_c/structure/fragment.pyx":644
 *         else:
 *             result = <_NameTree>presult
 *             return result             # <<<<<<<<<<<<<<
 * 
 *     cpdef traverse(self, list_or_iterable parts):
 */
    __Pyx_XDECREF(((PyObject *)__pyx_r));
    __Pyx_INCREF(((PyObject *)__pyx_v_result));
    __pyx_r = __pyx_v_result;
    goto __pyx_L0;
  }

  /* "glycopeptidepy/_c/structure/fragment.pyx":633
 *         return self.get(key)
 * 
 *     cpdef _NameTree get(self, key):             # <<<<<<<<<<<<<<
 *         cdef:
 *             PyObject* presult
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment._NameTree.get", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_result);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_5get(PyObject *__pyx_v_self, PyObject *__pyx_v_key); /*proto*/
static char __pyx_doc_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_4get[] = "_NameTree.get(self, key) -> _NameTree";
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_5get(PyObject *__pyx_v_self, PyObject *__pyx_v_key) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_4get(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment__NameTree *)__pyx_v_self), ((PyObject *)__pyx_v_key));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_4get(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment__NameTree *__pyx_v_self, PyObject *__pyx_v_key) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_get(__pyx_v_self, __pyx_v_key, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 633, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment._NameTree.get", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/fragment.pyx":646
 *             return result
 * 
 *     cpdef traverse(self, list_or_iterable parts):             # <<<<<<<<<<<<<<
 *         cdef:
 *             _NameTree node
 */

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_7traverse(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_6traverse[] = "_NameTree.traverse(signatures, args, kwargs, defaults)";
static PyMethodDef __pyx_mdef_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_7traverse = {"traverse", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_7traverse, METH_VARARGS|METH_KEYWORDS, __pyx_doc_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_6traverse};
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_7traverse(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_args = 0;
  PyObject *__pyx_v_kwargs = 0;
  CYTHON_UNUSED PyObject *__pyx_v_defaults = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fused_cpdef (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_args,&__pyx_n_s_kwargs,&__pyx_n_s_defaults,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_args)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_kwargs)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fused_cpdef", 1, 3, 3, 1); __PYX_ERR(0, 646, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_defaults)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_fused_cpdef", 1, 3, 3, 2); __PYX_ERR(0, 646, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_fused_cpdef") < 0)) __PYX_ERR(0, 646, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_args = values[0];
    __pyx_v_kwargs = values[1];
    __pyx_v_defaults = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_fused_cpdef", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 646, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment._NameTree.__pyx_fused_cpdef", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_6traverse(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment__NameTree *)__pyx_v_signatures), __pyx_v_args, __pyx_v_kwargs, __pyx_v_defaults);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_6traverse(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment__NameTree *__pyx_v_signatures, PyObject *__pyx_v_args, PyObject *__pyx_v_kwargs, CYTHON_UNUSED PyObject *__pyx_v_defaults) {
  PyObject *__pyx_v_dest_sig = NULL;
  Py_ssize_t __pyx_v_i;
  CYTHON_UNUSED PyTypeObject *__pyx_v_ndarray = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_v_itemsize;
  PyObject *__pyx_v_arg = NULL;
  PyObject *__pyx_v_candidates = NULL;
  PyObject *__pyx_v_sig = NULL;
  int __pyx_v_match_found;
  PyObject *__pyx_v_src_sig = NULL;
  PyObject *__pyx_v_dst_type = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  Py_ssize_t __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  Py_ssize_t __pyx_t_13;
  Py_ssize_t __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  int __pyx_t_16;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("traverse", 0);
  __Pyx_INCREF(__pyx_v_kwargs);
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 646, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  PyList_SET_ITEM(__pyx_t_1, 0, Py_None);
  __pyx_v_dest_sig = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_3 = (__pyx_v_kwargs != Py_None);
  __pyx_t_4 = (__pyx_t_3 != 0);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_kwargs); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 646, __pyx_L1_error)
  __pyx_t_3 = ((!__pyx_t_4) != 0);
  __pyx_t_2 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {
    __Pyx_INCREF(Py_None);
    __Pyx_DECREF_SET(__pyx_v_kwargs, Py_None);
  }
  __pyx_t_1 = ((PyObject *)__Pyx_ImportNumPyArrayTypeIfAvailable()); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 646, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_ndarray = ((PyTypeObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_v_itemsize = -1L;
  if (unlikely(__pyx_v_args == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(0, 646, __pyx_L1_error)
  }
  __pyx_t_5 = PyTuple_GET_SIZE(((PyObject*)__pyx_v_args)); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(0, 646, __pyx_L1_error)
  __pyx_t_2 = ((1 < __pyx_t_5) != 0);
  if (__pyx_t_2) {
    if (unlikely(__pyx_v_args == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 646, __pyx_L1_error)
    }
    __pyx_t_1 = __Pyx_GetItemInt_Tuple(((PyObject*)__pyx_v_args), 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 646, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_arg = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L6;
  }
  __pyx_t_3 = (__pyx_v_kwargs != Py_None);
  __pyx_t_4 = (__pyx_t_3 != 0);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L7_bool_binop_done;
  }
  if (unlikely(__pyx_v_kwargs == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(0, 646, __pyx_L1_error)
  }
  __pyx_t_4 = (__Pyx_PyDict_ContainsTF(__pyx_n_s_parts, ((PyObject*)__pyx_v_kwargs), Py_EQ)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 646, __pyx_L1_error)
  __pyx_t_3 = (__pyx_t_4 != 0);
  __pyx_t_2 = __pyx_t_3;
  __pyx_L7_bool_binop_done:;
  if (__pyx_t_2) {
    if (unlikely(__pyx_v_kwargs == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 646, __pyx_L1_error)
    }
    __pyx_t_1 = __Pyx_PyDict_GetItem(((PyObject*)__pyx_v_kwargs), __pyx_n_s_parts); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 646, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_arg = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L6;
  }
  /*else*/ {
    if (unlikely(__pyx_v_args == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
      __PYX_ERR(0, 646, __pyx_L1_error)
    }
    __pyx_t_5 = PyTuple_GET_SIZE(((PyObject*)__pyx_v_args)); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(0, 646, __pyx_L1_error)
    __pyx_t_1 = PyInt_FromSsize_t(__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 646, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_6 = PyTuple_New(3); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 646, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_INCREF(__pyx_int_2);
    __Pyx_GIVEREF(__pyx_int_2);
    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_int_2);
    __Pyx_INCREF(__pyx_n_s_s);
    __Pyx_GIVEREF(__pyx_n_s_s);
    PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_n_s_s);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Expected_at_least_d_argument_s_g, __pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 646, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 646, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __PYX_ERR(0, 646, __pyx_L1_error)
  }
  __pyx_L6:;
  while (1) {
    __pyx_t_2 = PyList_Check(__pyx_v_arg); 
    __pyx_t_3 = (__pyx_t_2 != 0);
    if (__pyx_t_3) {
      if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, __pyx_n_s_list, long, 1, __Pyx_PyInt_From_long, 1, 0, 1) < 0)) __PYX_ERR(0, 646, __pyx_L1_error)
      goto __pyx_L10_break;
    }
    if (unlikely(__Pyx_SetItemInt(__pyx_v_dest_sig, 0, __pyx_n_s_object, long, 1, __Pyx_PyInt_From_long, 1, 0, 1) < 0)) __PYX_ERR(0, 646, __pyx_L1_error)
    goto __pyx_L10_break;
  }
  __pyx_L10_break:;
  __pyx_t_6 = PyList_New(0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 646, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_v_candidates = ((PyObject*)__pyx_t_6);
  __pyx_t_6 = 0;
  __pyx_t_5 = 0;
  if (unlikely(((PyObject *)__pyx_v_signatures) == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(0, 646, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_dict_iterator(((PyObject*)__pyx_v_signatures), 1, ((PyObject *)NULL), (&__pyx_t_7), (&__pyx_t_8)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 646, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __pyx_t_6 = __pyx_t_1;
  __pyx_t_1 = 0;
  while (1) {
    __pyx_t_9 = __Pyx_dict_iter_next(__pyx_t_6, __pyx_t_7, &__pyx_t_5, &__pyx_t_1, NULL, NULL, __pyx_t_8);
    if (unlikely(__pyx_t_9 == 0)) break;
    if (unlikely(__pyx_t_9 == -1)) __PYX_ERR(0, 646, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_sig, __pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_v_match_found = 0;
    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_v_sig, __pyx_n_s_strip); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 646, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __pyx_t_12 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_11))) {
      __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_11);
      if (likely(__pyx_t_12)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
        __Pyx_INCREF(__pyx_t_12);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_11, function);
      }
    }
    __pyx_t_10 = (__pyx_t_12) ? __Pyx_PyObject_Call2Args(__pyx_t_11, __pyx_t_12, __pyx_kp_s__3) : __Pyx_PyObject_CallOneArg(__pyx_t_11, __pyx_kp_s__3);
    __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
    if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 646, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_n_s_split); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 646, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __pyx_t_10 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_11))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_11);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_11, function);
      }
    }
    __pyx_t_1 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_11, __pyx_t_10, __pyx_kp_s__4) : __Pyx_PyObject_CallOneArg(__pyx_t_11, __pyx_kp_s__4);
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 646, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_XDECREF_SET(__pyx_v_src_sig, __pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_t_13 = PyList_GET_SIZE(__pyx_v_dest_sig); if (unlikely(__pyx_t_13 == ((Py_ssize_t)-1))) __PYX_ERR(0, 646, __pyx_L1_error)
    __pyx_t_14 = __pyx_t_13;
    for (__pyx_t_15 = 0; __pyx_t_15 < __pyx_t_14; __pyx_t_15+=1) {
      __pyx_v_i = __pyx_t_15;
      __pyx_t_1 = __Pyx_GetItemInt_List(__pyx_v_dest_sig, __pyx_v_i, Py_ssize_t, 1, PyInt_FromSsize_t, 1, 1, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 646, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_XDECREF_SET(__pyx_v_dst_type, __pyx_t_1);
      __pyx_t_1 = 0;
      __pyx_t_3 = (__pyx_v_dst_type != Py_None);
      __pyx_t_2 = (__pyx_t_3 != 0);
      if (__pyx_t_2) {
        __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_src_sig, __pyx_v_i, Py_ssize_t, 1, PyInt_FromSsize_t, 0, 1, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 646, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_11 = PyObject_RichCompare(__pyx_t_1, __pyx_v_dst_type, Py_EQ); __Pyx_XGOTREF(__pyx_t_11); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 646, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_11); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 646, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        if (__pyx_t_2) {
          __pyx_v_match_found = 1;
          goto __pyx_L17;
        }
        /*else*/ {
          __pyx_v_match_found = 0;
          goto __pyx_L15_break;
        }
        __pyx_L17:;
      }
    }
    __pyx_L15_break:;
    __pyx_t_2 = (__pyx_v_match_found != 0);
    if (__pyx_t_2) {
      __pyx_t_16 = __Pyx_PyList_Append(__pyx_v_candidates, __pyx_v_sig); if (unlikely(__pyx_t_16 == ((int)-1))) __PYX_ERR(0, 646, __pyx_L1_error)
    }
  }
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_2 = (PyList_GET_SIZE(__pyx_v_candidates) != 0);
  __pyx_t_3 = ((!__pyx_t_2) != 0);
  if (__pyx_t_3) {
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__5, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 646, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __PYX_ERR(0, 646, __pyx_L1_error)
  }
  __pyx_t_7 = PyList_GET_SIZE(__pyx_v_candidates); if (unlikely(__pyx_t_7 == ((Py_ssize_t)-1))) __PYX_ERR(0, 646, __pyx_L1_error)
  __pyx_t_3 = ((__pyx_t_7 > 1) != 0);
  if (__pyx_t_3) {
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__6, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 646, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __PYX_ERR(0, 646, __pyx_L1_error)
  }
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    if (unlikely(((PyObject *)__pyx_v_signatures) == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 646, __pyx_L1_error)
    }
    __pyx_t_6 = __Pyx_GetItemInt_List(__pyx_v_candidates, 0, long, 1, __Pyx_PyInt_From_long, 1, 0, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 646, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_11 = __Pyx_PyDict_GetItem(((PyObject*)__pyx_v_signatures), __pyx_t_6); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 646, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_r = __pyx_t_11;
    __pyx_t_11 = 0;
    goto __pyx_L0;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment._NameTree.__pyx_fused_cpdef", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_dest_sig);
  __Pyx_XDECREF(__pyx_v_ndarray);
  __Pyx_XDECREF(__pyx_v_arg);
  __Pyx_XDECREF(__pyx_v_candidates);
  __Pyx_XDECREF(__pyx_v_sig);
  __Pyx_XDECREF(__pyx_v_src_sig);
  __Pyx_XDECREF(__pyx_v_dst_type);
  __Pyx_XDECREF(__pyx_v_kwargs);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_9__pyx_fuse_0traverse(PyObject *__pyx_v_self, PyObject *__pyx_v_parts); /*proto*/
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_7traverse(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_fuse_0__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_traverse(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment__NameTree *__pyx_v_self, PyObject *__pyx_v_parts, CYTHON_UNUSED int __pyx_skip_dispatch) {
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment__NameTree *__pyx_v_node = 0;
  size_t __pyx_v_i;
  size_t __pyx_v_n;
  PyObject *__pyx_v_kv = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  size_t __pyx_t_1;
  size_t __pyx_t_2;
  size_t __pyx_t_3;
  PyObject *__pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_fuse_0traverse", 0);

  /* "glycopeptidepy/_c/structure/fragment.pyx":652
 *             object parts_fast
 * 
 *         node = self             # <<<<<<<<<<<<<<
 *         if list_or_iterable is list:
 *             n = PyList_GET_SIZE(parts)
 */
  __Pyx_INCREF(((PyObject *)__pyx_v_self));
  __pyx_v_node = __pyx_v_self;

  /* "glycopeptidepy/_c/structure/fragment.pyx":654
 *         node = self
 *         if list_or_iterable is list:
 *             n = PyList_GET_SIZE(parts)             # <<<<<<<<<<<<<<
 *             for i in range(n):
 *                 kv = <object>PyList_GET_ITEM(parts, i)
 */
  __pyx_v_n = PyList_GET_SIZE(__pyx_v_parts);

  /* "glycopeptidepy/_c/structure/fragment.pyx":655
 *         if list_or_iterable is list:
 *             n = PyList_GET_SIZE(parts)
 *             for i in range(n):             # <<<<<<<<<<<<<<
 *                 kv = <object>PyList_GET_ITEM(parts, i)
 *                 node = node.get(kv)
 */
  __pyx_t_1 = __pyx_v_n;
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;

    /* "glycopeptidepy/_c/structure/fragment.pyx":656
 *             n = PyList_GET_SIZE(parts)
 *             for i in range(n):
 *                 kv = <object>PyList_GET_ITEM(parts, i)             # <<<<<<<<<<<<<<
 *                 node = node.get(kv)
 *         else:
 */
    __pyx_t_4 = PyList_GET_ITEM(__pyx_v_parts, __pyx_v_i);
    __pyx_t_5 = ((PyObject *)__pyx_t_4);
    __Pyx_INCREF(__pyx_t_5);
    __Pyx_XDECREF_SET(__pyx_v_kv, __pyx_t_5);
    __pyx_t_5 = 0;

    /* "glycopeptidepy/_c/structure/fragment.pyx":657
 *             for i in range(n):
 *                 kv = <object>PyList_GET_ITEM(parts, i)
 *                 node = node.get(kv)             # <<<<<<<<<<<<<<
 *         else:
 *             parts_fast = PySequence_Fast(parts, "Expected a sequence")
 */
    __pyx_t_5 = ((PyObject *)__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_get(__pyx_v_node, __pyx_v_kv, 0)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 657, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF_SET(__pyx_v_node, ((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment__NameTree *)__pyx_t_5));
    __pyx_t_5 = 0;
  }

  /* "glycopeptidepy/_c/structure/fragment.pyx":664
 *                 kv = <object>PySequence_Fast_GET_ITEM(parts_fast, i)
 *                 node = node.get(kv)
 *         return node             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_node));
  __pyx_r = ((PyObject *)__pyx_v_node);
  goto __pyx_L0;

  /* "glycopeptidepy/_c/structure/fragment.pyx":646
 *             return result
 * 
 *     cpdef traverse(self, list_or_iterable parts):             # <<<<<<<<<<<<<<
 *         cdef:
 *             _NameTree node
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment._NameTree.traverse", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_node);
  __Pyx_XDECREF(__pyx_v_kv);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_9__pyx_fuse_0traverse(PyObject *__pyx_v_self, PyObject *__pyx_v_parts); /*proto*/
static PyMethodDef __pyx_fuse_0__pyx_mdef_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_9__pyx_fuse_0traverse = {"__pyx_fuse_0traverse", (PyCFunction)__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_9__pyx_fuse_0traverse, METH_O, __pyx_doc_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_6traverse};
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_9__pyx_fuse_0traverse(PyObject *__pyx_v_self, PyObject *__pyx_v_parts) {
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fuse_0traverse (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_parts), (&PyList_Type), 1, "parts", 1))) __PYX_ERR(0, 646, __pyx_L1_error)
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_8__pyx_fuse_0traverse(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment__NameTree *)__pyx_v_self), ((PyObject*)__pyx_v_parts));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_8__pyx_fuse_0traverse(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment__NameTree *__pyx_v_self, PyObject *__pyx_v_parts) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_fuse_0traverse", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_fuse_0__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_traverse(__pyx_v_self, __pyx_v_parts, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 646, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment._NameTree.__pyx_fuse_0traverse", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_11__pyx_fuse_1traverse(PyObject *__pyx_v_self, PyObject *__pyx_v_parts); /*proto*/
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_7traverse(PyObject *__pyx_v_signatures, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_fuse_1__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_traverse(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment__NameTree *__pyx_v_self, PyObject *__pyx_v_parts, CYTHON_UNUSED int __pyx_skip_dispatch) {
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment__NameTree *__pyx_v_node = 0;
  size_t __pyx_v_i;
  size_t __pyx_v_n;
  PyObject *__pyx_v_parts_fast = 0;
  PyObject *__pyx_v_kv = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  size_t __pyx_t_2;
  size_t __pyx_t_3;
  size_t __pyx_t_4;
  PyObject *__pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_fuse_1traverse", 0);

  /* "glycopeptidepy/_c/structure/fragment.pyx":652
 *             object parts_fast
 * 
 *         node = self             # <<<<<<<<<<<<<<
 *         if list_or_iterable is list:
 *             n = PyList_GET_SIZE(parts)
 */
  __Pyx_INCREF(((PyObject *)__pyx_v_self));
  __pyx_v_node = __pyx_v_self;

  /* "glycopeptidepy/_c/structure/fragment.pyx":659
 *                 node = node.get(kv)
 *         else:
 *             parts_fast = PySequence_Fast(parts, "Expected a sequence")             # <<<<<<<<<<<<<<
 *             n = PySequence_Fast_GET_SIZE(parts_fast)
 *             for i in range(n):
 */
  __pyx_t_1 = PySequence_Fast(__pyx_v_parts, ((char *)"Expected a sequence")); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 659, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_parts_fast = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "glycopeptidepy/_c/structure/fragment.pyx":660
 *         else:
 *             parts_fast = PySequence_Fast(parts, "Expected a sequence")
 *             n = PySequence_Fast_GET_SIZE(parts_fast)             # <<<<<<<<<<<<<<
 *             for i in range(n):
 *                 kv = <object>PySequence_Fast_GET_ITEM(parts_fast, i)
 */
  __pyx_v_n = PySequence_Fast_GET_SIZE(__pyx_v_parts_fast);

  /* "glycopeptidepy/_c/structure/fragment.pyx":661
 *             parts_fast = PySequence_Fast(parts, "Expected a sequence")
 *             n = PySequence_Fast_GET_SIZE(parts_fast)
 *             for i in range(n):             # <<<<<<<<<<<<<<
 *                 kv = <object>PySequence_Fast_GET_ITEM(parts_fast, i)
 *                 node = node.get(kv)
 */
  __pyx_t_2 = __pyx_v_n;
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "glycopeptidepy/_c/structure/fragment.pyx":662
 *             n = PySequence_Fast_GET_SIZE(parts_fast)
 *             for i in range(n):
 *                 kv = <object>PySequence_Fast_GET_ITEM(parts_fast, i)             # <<<<<<<<<<<<<<
 *                 node = node.get(kv)
 *         return node
 */
    __pyx_t_5 = PySequence_Fast_GET_ITEM(__pyx_v_parts_fast, __pyx_v_i);
    __pyx_t_1 = ((PyObject *)__pyx_t_5);
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_kv, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "glycopeptidepy/_c/structure/fragment.pyx":663
 *             for i in range(n):
 *                 kv = <object>PySequence_Fast_GET_ITEM(parts_fast, i)
 *                 node = node.get(kv)             # <<<<<<<<<<<<<<
 *         return node
 * 
 */
    __pyx_t_1 = ((PyObject *)__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_get(__pyx_v_node, __pyx_v_kv, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 663, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF_SET(__pyx_v_node, ((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment__NameTree *)__pyx_t_1));
    __pyx_t_1 = 0;
  }

  /* "glycopeptidepy/_c/structure/fragment.pyx":664
 *                 kv = <object>PySequence_Fast_GET_ITEM(parts_fast, i)
 *                 node = node.get(kv)
 *         return node             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_node));
  __pyx_r = ((PyObject *)__pyx_v_node);
  goto __pyx_L0;

  /* "glycopeptidepy/_c/structure/fragment.pyx":646
 *             return result
 * 
 *     cpdef traverse(self, list_or_iterable parts):             # <<<<<<<<<<<<<<
 *         cdef:
 *             _NameTree node
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment._NameTree.traverse", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_node);
  __Pyx_XDECREF(__pyx_v_parts_fast);
  __Pyx_XDECREF(__pyx_v_kv);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_11__pyx_fuse_1traverse(PyObject *__pyx_v_self, PyObject *__pyx_v_parts); /*proto*/
static PyMethodDef __pyx_fuse_1__pyx_mdef_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_11__pyx_fuse_1traverse = {"__pyx_fuse_1traverse", (PyCFunction)__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_11__pyx_fuse_1traverse, METH_O, __pyx_doc_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_6traverse};
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_11__pyx_fuse_1traverse(PyObject *__pyx_v_self, PyObject *__pyx_v_parts) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_fuse_1traverse (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_10__pyx_fuse_1traverse(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment__NameTree *)__pyx_v_self), ((PyObject *)__pyx_v_parts));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_10__pyx_fuse_1traverse(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment__NameTree *__pyx_v_self, PyObject *__pyx_v_parts) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_fuse_1traverse", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_fuse_1__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_traverse(__pyx_v_self, __pyx_v_parts, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 646, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment._NameTree.__pyx_fuse_1traverse", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/fragment.pxd":127
 * cdef class _NameTree(object):
 *     cdef:
 *         public object name             # <<<<<<<<<<<<<<
 *         public object value
 *         public dict children
 */

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_4name_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_4name_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_4name___get__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment__NameTree *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_4name___get__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment__NameTree *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->name);
  __pyx_r = __pyx_v_self->name;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_4name_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_4name_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_4name_2__set__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment__NameTree *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_4name_2__set__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment__NameTree *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 0);
  __Pyx_INCREF(__pyx_v_value);
  __Pyx_GIVEREF(__pyx_v_value);
  __Pyx_GOTREF(__pyx_v_self->name);
  __Pyx_DECREF(__pyx_v_self->name);
  __pyx_v_self->name = __pyx_v_value;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_4name_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_4name_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_4name_4__del__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment__NameTree *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_4name_4__del__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment__NameTree *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->name);
  __Pyx_DECREF(__pyx_v_self->name);
  __pyx_v_self->name = Py_None;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/fragment.pxd":128
 *     cdef:
 *         public object name
 *         public object value             # <<<<<<<<<<<<<<
 *         public dict children
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_5value_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_5value_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_5value___get__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment__NameTree *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_5value___get__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment__NameTree *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->value);
  __pyx_r = __pyx_v_self->value;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_5value_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_5value_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_5value_2__set__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment__NameTree *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_5value_2__set__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment__NameTree *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 0);
  __Pyx_INCREF(__pyx_v_value);
  __Pyx_GIVEREF(__pyx_v_value);
  __Pyx_GOTREF(__pyx_v_self->value);
  __Pyx_DECREF(__pyx_v_self->value);
  __pyx_v_self->value = __pyx_v_value;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_5value_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_5value_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_5value_4__del__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment__NameTree *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_5value_4__del__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment__NameTree *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->value);
  __Pyx_DECREF(__pyx_v_self->value);
  __pyx_v_self->value = Py_None;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/fragment.pxd":129
 *         public object name
 *         public object value
 *         public dict children             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_8children_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_8children_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_8children___get__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment__NameTree *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_8children___get__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment__NameTree *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->children);
  __pyx_r = __pyx_v_self->children;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_8children_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_8children_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_8children_2__set__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment__NameTree *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_8children_2__set__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment__NameTree *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);
  if (!(likely(PyDict_CheckExact(__pyx_v_value))||((__pyx_v_value) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "dict", Py_TYPE(__pyx_v_value)->tp_name), 0))) __PYX_ERR(1, 129, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_value;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->children);
  __Pyx_DECREF(__pyx_v_self->children);
  __pyx_v_self->children = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment._NameTree.children.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_8children_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_8children_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_8children_4__del__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment__NameTree *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_8children_4__del__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment__NameTree *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->children);
  __Pyx_DECREF(__pyx_v_self->children);
  __pyx_v_self->children = ((PyObject*)Py_None);

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_15__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_14__reduce_cython__[] = "_NameTree.__reduce_cython__(self)";
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_15__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_14__reduce_cython__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment__NameTree *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_14__reduce_cython__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment__NameTree *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.children, self.name, self.value)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_self->children);
  __Pyx_GIVEREF(__pyx_v_self->children);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_self->children);
  __Pyx_INCREF(__pyx_v_self->name);
  __Pyx_GIVEREF(__pyx_v_self->name);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_self->name);
  __Pyx_INCREF(__pyx_v_self->value);
  __Pyx_GIVEREF(__pyx_v_self->value);
  PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_self->value);
  __pyx_v_state = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.children, self.name, self.value)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = (self.children, self.name, self.value)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict);
    __pyx_t_4 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_4));
    __pyx_t_4 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.children is not None or self.name is not None or self.value is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.children, self.name, self.value)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.children is not None or self.name is not None or self.value is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle__NameTree, (type(self), 0xd2a53ba, None), state
 */
  /*else*/ {
    __pyx_t_2 = (__pyx_v_self->children != ((PyObject*)Py_None));
    __pyx_t_5 = (__pyx_t_2 != 0);
    if (!__pyx_t_5) {
    } else {
      __pyx_t_3 = __pyx_t_5;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_5 = (__pyx_v_self->name != Py_None);
    __pyx_t_2 = (__pyx_t_5 != 0);
    if (!__pyx_t_2) {
    } else {
      __pyx_t_3 = __pyx_t_2;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_2 = (__pyx_v_self->value != Py_None);
    __pyx_t_5 = (__pyx_t_2 != 0);
    __pyx_t_3 = __pyx_t_5;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_3;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.children is not None or self.name is not None or self.value is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle__NameTree, (type(self), 0xd2a53ba, None), state
 *     else:
 */
  __pyx_t_3 = (__pyx_v_use_setstate != 0);
  if (__pyx_t_3) {

    /* "(tree fragment)":13
 *         use_setstate = self.children is not None or self.name is not None or self.value is not None
 *     if use_setstate:
 *         return __pyx_unpickle__NameTree, (type(self), 0xd2a53ba, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle__NameTree, (type(self), 0xd2a53ba, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_pyx_unpickle__NameTree); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_220877754);
    __Pyx_GIVEREF(__pyx_int_220877754);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_220877754);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None);
    __pyx_t_6 = PyTuple_New(3); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_1);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_v_state);
    __pyx_t_4 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_6;
    __pyx_t_6 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.children is not None or self.name is not None or self.value is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle__NameTree, (type(self), 0xd2a53ba, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle__NameTree, (type(self), 0xd2a53ba, None), state
 *     else:
 *         return __pyx_unpickle__NameTree, (type(self), 0xd2a53ba, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle__NameTree__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_pyx_unpickle__NameTree); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_220877754);
    __Pyx_GIVEREF(__pyx_int_220877754);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_220877754);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state);
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1);
    __pyx_t_6 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment._NameTree.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle__NameTree, (type(self), 0xd2a53ba, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle__NameTree__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_17__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static char __pyx_doc_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_16__setstate_cython__[] = "_NameTree.__setstate_cython__(self, __pyx_state)";
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_17__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_16__setstate_cython__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment__NameTree *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_16__setstate_cython__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment__NameTree *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle__NameTree, (type(self), 0xd2a53ba, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle__NameTree__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(2, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_14glycopeptidepy_2_c_9structure_8fragment___pyx_unpickle__NameTree__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle__NameTree, (type(self), 0xd2a53ba, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle__NameTree__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment._NameTree.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/fragment.pyx":670
 * 
 * 
 * cdef object pair_mass(tuple x):             # <<<<<<<<<<<<<<
 *     tmp = <object>PyTuple_GetItem(x, 0)
 *     return tmp.mass()
 */

static PyObject *__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_pair_mass(PyObject *__pyx_v_x) {
  PyObject *__pyx_v_tmp = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("pair_mass", 0);

  /* "glycopeptidepy/_c/structure/fragment.pyx":671
 * 
 * cdef object pair_mass(tuple x):
 *     tmp = <object>PyTuple_GetItem(x, 0)             # <<<<<<<<<<<<<<
 *     return tmp.mass()
 * 
 */
  __pyx_t_1 = PyTuple_GetItem(__pyx_v_x, 0); if (unlikely(__pyx_t_1 == ((PyObject *)NULL))) __PYX_ERR(0, 671, __pyx_L1_error)
  __pyx_t_2 = ((PyObject *)__pyx_t_1);
  __Pyx_INCREF(__pyx_t_2);
  __pyx_v_tmp = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "glycopeptidepy/_c/structure/fragment.pyx":672
 * cdef object pair_mass(tuple x):
 *     tmp = <object>PyTuple_GetItem(x, 0)
 *     return tmp.mass()             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_tmp, __pyx_n_s_mass); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 672, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 672, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "glycopeptidepy/_c/structure/fragment.pyx":670
 * 
 * 
 * cdef object pair_mass(tuple x):             # <<<<<<<<<<<<<<
 *     tmp = <object>PyTuple_GetItem(x, 0)
 *     return tmp.mass()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment.pair_mass", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_tmp);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/fragment.pyx":675
 * 
 * 
 * cdef basestring build_name_from_composition(mapping_types glycan_composition):             # <<<<<<<<<<<<<<
 *         cdef:
 *             basestring name, extended_key
 */
static PyObject *__pyx_gb_14glycopeptidepy_2_c_9structure_8fragment_39__pyx_fuse_0build_name_from_composition_2generator(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "glycopeptidepy/_c/structure/fragment.pyx":694
 *         if extended_key is None:
 *             parts.sort(key=pair_mass)
 *             node.name = extended_key = ''.join("%s%d" % kv for kv in parts)             # <<<<<<<<<<<<<<
 *         if extended_key:
 *             name = "%s+%s" % (name, extended_key)
 */

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_39__pyx_fuse_0build_name_from_composition_genexpr(PyObject *__pyx_self) {
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_1_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_1_genexpr *)__pyx_tp_new_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_1_genexpr(__pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_1_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_1_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 694, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct____pyx_fuse_0__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition *) __pyx_self;
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_outer_scope));
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_14glycopeptidepy_2_c_9structure_8fragment_39__pyx_fuse_0build_name_from_composition_2generator, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_pyx_fuse_0build_name_from_comp, __pyx_n_s_glycopeptidepy__c_structure_frag); if (unlikely(!gen)) __PYX_ERR(0, 694, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment.__pyx_fuse_0build_name_from_composition.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_14glycopeptidepy_2_c_9structure_8fragment_39__pyx_fuse_0build_name_from_composition_2generator(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_1_genexpr *__pyx_cur_scope = ((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_1_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L6_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 694, __pyx_L1_error)
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_parts)) { __Pyx_RaiseClosureNameError("parts"); __PYX_ERR(0, 694, __pyx_L1_error) }
  if (unlikely(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_parts == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(0, 694, __pyx_L1_error)
  }
  __pyx_t_1 = __pyx_cur_scope->__pyx_outer_scope->__pyx_v_parts; __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
  for (;;) {
    if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_1)) break;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_3 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_3); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 694, __pyx_L1_error)
    #else
    __pyx_t_3 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 694, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_kv);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_kv, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_3);
    __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyString_FormatSafe(__pyx_kp_s_s_d, __pyx_cur_scope->__pyx_v_kv); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 694, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    __Pyx_XGIVEREF(__pyx_t_1);
    __pyx_cur_scope->__pyx_t_0 = __pyx_t_1;
    __pyx_cur_scope->__pyx_t_1 = __pyx_t_2;
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
    /* return from generator, yielding value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L6_resume_from_yield:;
    __pyx_t_1 = __pyx_cur_scope->__pyx_t_0;
    __pyx_cur_scope->__pyx_t_0 = 0;
    __Pyx_XGOTREF(__pyx_t_1);
    __pyx_t_2 = __pyx_cur_scope->__pyx_t_1;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 694, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/fragment.pyx":675
 * 
 * 
 * cdef basestring build_name_from_composition(mapping_types glycan_composition):             # <<<<<<<<<<<<<<
 *         cdef:
 *             basestring name, extended_key
 */

static PyObject *__pyx_fuse_0__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition(struct __pyx_obj_14glycopeptidepy_2_c_11count_table_CountTable *__pyx_v_glycan_composition) {
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct____pyx_fuse_0__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition *__pyx_cur_scope;
  PyObject *__pyx_v_name = 0;
  PyObject *__pyx_v_extended_key = 0;
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment__NameTree *__pyx_v_root = 0;
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment__NameTree *__pyx_v_node = 0;
  PyObject *__pyx_gb_14glycopeptidepy_2_c_9structure_8fragment_39__pyx_fuse_0build_name_from_composition_2generator = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_fuse_0build_name_from_composition", 0);
  __pyx_cur_scope = (struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct____pyx_fuse_0__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition *)__pyx_tp_new_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct____pyx_fuse_0__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition(__pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct____pyx_fuse_0__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct____pyx_fuse_0__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 675, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }

  /* "glycopeptidepy/_c/structure/fragment.pyx":680
 *             _NameTree root, node
 *             list parts
 *         name = 'peptide'             # <<<<<<<<<<<<<<
 *         root = stub_fragment_name_cache
 *         if mapping_types is dict:
 */
  __Pyx_INCREF(__pyx_n_s_peptide);
  __pyx_v_name = ((PyObject*)__pyx_n_s_peptide);

  /* "glycopeptidepy/_c/structure/fragment.pyx":681
 *             list parts
 *         name = 'peptide'
 *         root = stub_fragment_name_cache             # <<<<<<<<<<<<<<
 *         if mapping_types is dict:
 *             parts = PyDict_Items(glycan_composition)
 */
  __Pyx_INCREF(((PyObject *)__pyx_v_14glycopeptidepy_2_c_9structure_8fragment_stub_fragment_name_cache));
  __pyx_v_root = __pyx_v_14glycopeptidepy_2_c_9structure_8fragment_stub_fragment_name_cache;

  /* "glycopeptidepy/_c/structure/fragment.pyx":685
 *             parts = PyDict_Items(glycan_composition)
 *         elif mapping_types is CountTable:
 *             parts = count_table_items(glycan_composition.table)             # <<<<<<<<<<<<<<
 *         elif mapping_types is object:
 *             parts = list(glycan_composition.items())
 */
  __pyx_t_1 = __pyx_f_14glycopeptidepy_2_c_11count_table_count_table_items(__pyx_v_glycan_composition->table); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 685, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_cur_scope->__pyx_v_parts = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "glycopeptidepy/_c/structure/fragment.pyx":690
 *         # traverse the tree to find the node referring to
 *         # this combination of monosaccharides and counts
 *         node = root.traverse(parts)             # <<<<<<<<<<<<<<
 *         extended_key = node.name
 *         if extended_key is None:
 */
  __pyx_t_1 = __pyx_cur_scope->__pyx_v_parts;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = ((struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_8fragment__NameTree *)__pyx_v_root->__pyx_vtab)->__pyx_fuse_0traverse(__pyx_v_root, ((PyObject*)__pyx_t_1), 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 690, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment__NameTree))))) __PYX_ERR(0, 690, __pyx_L1_error)
  __pyx_v_node = ((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment__NameTree *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "glycopeptidepy/_c/structure/fragment.pyx":691
 *         # this combination of monosaccharides and counts
 *         node = root.traverse(parts)
 *         extended_key = node.name             # <<<<<<<<<<<<<<
 *         if extended_key is None:
 *             parts.sort(key=pair_mass)
 */
  if (!(likely(__Pyx_PyBaseString_CheckExact(__pyx_v_node->name))||((__pyx_v_node->name) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", (PY_MAJOR_VERSION < 3 ? "basestring" : "str"), Py_TYPE(__pyx_v_node->name)->tp_name), 0))) __PYX_ERR(0, 691, __pyx_L1_error)
  __pyx_t_2 = __pyx_v_node->name;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_v_extended_key = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "glycopeptidepy/_c/structure/fragment.pyx":692
 *         node = root.traverse(parts)
 *         extended_key = node.name
 *         if extended_key is None:             # <<<<<<<<<<<<<<
 *             parts.sort(key=pair_mass)
 *             node.name = extended_key = ''.join("%s%d" % kv for kv in parts)
 */
  __pyx_t_3 = (__pyx_v_extended_key == ((PyObject*)Py_None));
  __pyx_t_4 = (__pyx_t_3 != 0);
  if (__pyx_t_4) {

    /* "glycopeptidepy/_c/structure/fragment.pyx":693
 *         extended_key = node.name
 *         if extended_key is None:
 *             parts.sort(key=pair_mass)             # <<<<<<<<<<<<<<
 *             node.name = extended_key = ''.join("%s%d" % kv for kv in parts)
 *         if extended_key:
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_parts, __pyx_n_s_sort); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 693, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 693, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = __Pyx_CFunc_object____tuple___to_py(__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_pair_mass); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 693, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_key, __pyx_t_5) < 0) __PYX_ERR(0, 693, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_empty_tuple, __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 693, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "glycopeptidepy/_c/structure/fragment.pyx":694
 *         if extended_key is None:
 *             parts.sort(key=pair_mass)
 *             node.name = extended_key = ''.join("%s%d" % kv for kv in parts)             # <<<<<<<<<<<<<<
 *         if extended_key:
 *             name = "%s+%s" % (name, extended_key)
 */
    __pyx_t_5 = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_39__pyx_fuse_0build_name_from_composition_genexpr(((PyObject*)__pyx_cur_scope)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 694, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = __Pyx_PyString_Join(__pyx_kp_s_, __pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 694, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    __Pyx_GOTREF(__pyx_v_node->name);
    __Pyx_DECREF(__pyx_v_node->name);
    __pyx_v_node->name = __pyx_t_1;
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_DECREF_SET(__pyx_v_extended_key, ((PyObject*)__pyx_t_1));
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "glycopeptidepy/_c/structure/fragment.pyx":692
 *         node = root.traverse(parts)
 *         extended_key = node.name
 *         if extended_key is None:             # <<<<<<<<<<<<<<
 *             parts.sort(key=pair_mass)
 *             node.name = extended_key = ''.join("%s%d" % kv for kv in parts)
 */
  }

  /* "glycopeptidepy/_c/structure/fragment.pyx":695
 *             parts.sort(key=pair_mass)
 *             node.name = extended_key = ''.join("%s%d" % kv for kv in parts)
 *         if extended_key:             # <<<<<<<<<<<<<<
 *             name = "%s+%s" % (name, extended_key)
 *         return name
 */
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_extended_key); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 695, __pyx_L1_error)
  if (__pyx_t_4) {

    /* "glycopeptidepy/_c/structure/fragment.pyx":696
 *             node.name = extended_key = ''.join("%s%d" % kv for kv in parts)
 *         if extended_key:
 *             name = "%s+%s" % (name, extended_key)             # <<<<<<<<<<<<<<
 *         return name
 * 
 */
    __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 696, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v_name);
    __Pyx_GIVEREF(__pyx_v_name);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_name);
    __Pyx_INCREF(__pyx_v_extended_key);
    __Pyx_GIVEREF(__pyx_v_extended_key);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_extended_key);
    __pyx_t_5 = __Pyx_PyString_Format(__pyx_kp_s_s_s, __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 696, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_name, ((PyObject*)__pyx_t_5));
    __pyx_t_5 = 0;

    /* "glycopeptidepy/_c/structure/fragment.pyx":695
 *             parts.sort(key=pair_mass)
 *             node.name = extended_key = ''.join("%s%d" % kv for kv in parts)
 *         if extended_key:             # <<<<<<<<<<<<<<
 *             name = "%s+%s" % (name, extended_key)
 *         return name
 */
  }

  /* "glycopeptidepy/_c/structure/fragment.pyx":697
 *         if extended_key:
 *             name = "%s+%s" % (name, extended_key)
 *         return name             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_name);
  __pyx_r = __pyx_v_name;
  goto __pyx_L0;

  /* "glycopeptidepy/_c/structure/fragment.pyx":675
 * 
 * 
 * cdef basestring build_name_from_composition(mapping_types glycan_composition):             # <<<<<<<<<<<<<<
 *         cdef:
 *             basestring name, extended_key
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment.build_name_from_composition", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_name);
  __Pyx_XDECREF(__pyx_v_extended_key);
  __Pyx_XDECREF((PyObject *)__pyx_v_root);
  __Pyx_XDECREF((PyObject *)__pyx_v_node);
  __Pyx_XDECREF(__pyx_gb_14glycopeptidepy_2_c_9structure_8fragment_39__pyx_fuse_0build_name_from_composition_2generator);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_14glycopeptidepy_2_c_9structure_8fragment_39__pyx_fuse_1build_name_from_composition_2generator1(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "glycopeptidepy/_c/structure/fragment.pyx":694
 *         if extended_key is None:
 *             parts.sort(key=pair_mass)
 *             node.name = extended_key = ''.join("%s%d" % kv for kv in parts)             # <<<<<<<<<<<<<<
 *         if extended_key:
 *             name = "%s+%s" % (name, extended_key)
 */

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_39__pyx_fuse_1build_name_from_composition_genexpr(PyObject *__pyx_self) {
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_3_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_3_genexpr *)__pyx_tp_new_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_3_genexpr(__pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_3_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_3_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 694, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_2___pyx_fuse_1__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition *) __pyx_self;
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_outer_scope));
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_14glycopeptidepy_2_c_9structure_8fragment_39__pyx_fuse_1build_name_from_composition_2generator1, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_pyx_fuse_1build_name_from_comp, __pyx_n_s_glycopeptidepy__c_structure_frag); if (unlikely(!gen)) __PYX_ERR(0, 694, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment.__pyx_fuse_1build_name_from_composition.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_14glycopeptidepy_2_c_9structure_8fragment_39__pyx_fuse_1build_name_from_composition_2generator1(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_3_genexpr *__pyx_cur_scope = ((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_3_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L6_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 694, __pyx_L1_error)
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_parts)) { __Pyx_RaiseClosureNameError("parts"); __PYX_ERR(0, 694, __pyx_L1_error) }
  if (unlikely(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_parts == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(0, 694, __pyx_L1_error)
  }
  __pyx_t_1 = __pyx_cur_scope->__pyx_outer_scope->__pyx_v_parts; __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
  for (;;) {
    if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_1)) break;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_3 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_3); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 694, __pyx_L1_error)
    #else
    __pyx_t_3 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 694, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_kv);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_kv, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_3);
    __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyString_FormatSafe(__pyx_kp_s_s_d, __pyx_cur_scope->__pyx_v_kv); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 694, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    __Pyx_XGIVEREF(__pyx_t_1);
    __pyx_cur_scope->__pyx_t_0 = __pyx_t_1;
    __pyx_cur_scope->__pyx_t_1 = __pyx_t_2;
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
    /* return from generator, yielding value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L6_resume_from_yield:;
    __pyx_t_1 = __pyx_cur_scope->__pyx_t_0;
    __pyx_cur_scope->__pyx_t_0 = 0;
    __Pyx_XGOTREF(__pyx_t_1);
    __pyx_t_2 = __pyx_cur_scope->__pyx_t_1;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 694, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/fragment.pyx":675
 * 
 * 
 * cdef basestring build_name_from_composition(mapping_types glycan_composition):             # <<<<<<<<<<<<<<
 *         cdef:
 *             basestring name, extended_key
 */

static PyObject *__pyx_fuse_1__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition(PyObject *__pyx_v_glycan_composition) {
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_2___pyx_fuse_1__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition *__pyx_cur_scope;
  PyObject *__pyx_v_name = 0;
  PyObject *__pyx_v_extended_key = 0;
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment__NameTree *__pyx_v_root = 0;
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment__NameTree *__pyx_v_node = 0;
  PyObject *__pyx_gb_14glycopeptidepy_2_c_9structure_8fragment_39__pyx_fuse_1build_name_from_composition_2generator1 = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_fuse_1build_name_from_composition", 0);
  __pyx_cur_scope = (struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_2___pyx_fuse_1__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition *)__pyx_tp_new_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_2___pyx_fuse_1__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition(__pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_2___pyx_fuse_1__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_2___pyx_fuse_1__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 675, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }

  /* "glycopeptidepy/_c/structure/fragment.pyx":680
 *             _NameTree root, node
 *             list parts
 *         name = 'peptide'             # <<<<<<<<<<<<<<
 *         root = stub_fragment_name_cache
 *         if mapping_types is dict:
 */
  __Pyx_INCREF(__pyx_n_s_peptide);
  __pyx_v_name = ((PyObject*)__pyx_n_s_peptide);

  /* "glycopeptidepy/_c/structure/fragment.pyx":681
 *             list parts
 *         name = 'peptide'
 *         root = stub_fragment_name_cache             # <<<<<<<<<<<<<<
 *         if mapping_types is dict:
 *             parts = PyDict_Items(glycan_composition)
 */
  __Pyx_INCREF(((PyObject *)__pyx_v_14glycopeptidepy_2_c_9structure_8fragment_stub_fragment_name_cache));
  __pyx_v_root = __pyx_v_14glycopeptidepy_2_c_9structure_8fragment_stub_fragment_name_cache;

  /* "glycopeptidepy/_c/structure/fragment.pyx":683
 *         root = stub_fragment_name_cache
 *         if mapping_types is dict:
 *             parts = PyDict_Items(glycan_composition)             # <<<<<<<<<<<<<<
 *         elif mapping_types is CountTable:
 *             parts = count_table_items(glycan_composition.table)
 */
  __pyx_t_1 = PyDict_Items(__pyx_v_glycan_composition); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 683, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_cur_scope->__pyx_v_parts = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "glycopeptidepy/_c/structure/fragment.pyx":690
 *         # traverse the tree to find the node referring to
 *         # this combination of monosaccharides and counts
 *         node = root.traverse(parts)             # <<<<<<<<<<<<<<
 *         extended_key = node.name
 *         if extended_key is None:
 */
  __pyx_t_1 = __pyx_cur_scope->__pyx_v_parts;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = ((struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_8fragment__NameTree *)__pyx_v_root->__pyx_vtab)->__pyx_fuse_0traverse(__pyx_v_root, ((PyObject*)__pyx_t_1), 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 690, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment__NameTree))))) __PYX_ERR(0, 690, __pyx_L1_error)
  __pyx_v_node = ((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment__NameTree *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "glycopeptidepy/_c/structure/fragment.pyx":691
 *         # this combination of monosaccharides and counts
 *         node = root.traverse(parts)
 *         extended_key = node.name             # <<<<<<<<<<<<<<
 *         if extended_key is None:
 *             parts.sort(key=pair_mass)
 */
  if (!(likely(__Pyx_PyBaseString_CheckExact(__pyx_v_node->name))||((__pyx_v_node->name) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", (PY_MAJOR_VERSION < 3 ? "basestring" : "str"), Py_TYPE(__pyx_v_node->name)->tp_name), 0))) __PYX_ERR(0, 691, __pyx_L1_error)
  __pyx_t_2 = __pyx_v_node->name;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_v_extended_key = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "glycopeptidepy/_c/structure/fragment.pyx":692
 *         node = root.traverse(parts)
 *         extended_key = node.name
 *         if extended_key is None:             # <<<<<<<<<<<<<<
 *             parts.sort(key=pair_mass)
 *             node.name = extended_key = ''.join("%s%d" % kv for kv in parts)
 */
  __pyx_t_3 = (__pyx_v_extended_key == ((PyObject*)Py_None));
  __pyx_t_4 = (__pyx_t_3 != 0);
  if (__pyx_t_4) {

    /* "glycopeptidepy/_c/structure/fragment.pyx":693
 *         extended_key = node.name
 *         if extended_key is None:
 *             parts.sort(key=pair_mass)             # <<<<<<<<<<<<<<
 *             node.name = extended_key = ''.join("%s%d" % kv for kv in parts)
 *         if extended_key:
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_parts, __pyx_n_s_sort); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 693, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 693, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = __Pyx_CFunc_object____tuple___to_py(__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_pair_mass); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 693, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_key, __pyx_t_5) < 0) __PYX_ERR(0, 693, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_empty_tuple, __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 693, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "glycopeptidepy/_c/structure/fragment.pyx":694
 *         if extended_key is None:
 *             parts.sort(key=pair_mass)
 *             node.name = extended_key = ''.join("%s%d" % kv for kv in parts)             # <<<<<<<<<<<<<<
 *         if extended_key:
 *             name = "%s+%s" % (name, extended_key)
 */
    __pyx_t_5 = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_39__pyx_fuse_1build_name_from_composition_genexpr(((PyObject*)__pyx_cur_scope)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 694, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = __Pyx_PyString_Join(__pyx_kp_s_, __pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 694, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    __Pyx_GOTREF(__pyx_v_node->name);
    __Pyx_DECREF(__pyx_v_node->name);
    __pyx_v_node->name = __pyx_t_1;
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_DECREF_SET(__pyx_v_extended_key, ((PyObject*)__pyx_t_1));
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "glycopeptidepy/_c/structure/fragment.pyx":692
 *         node = root.traverse(parts)
 *         extended_key = node.name
 *         if extended_key is None:             # <<<<<<<<<<<<<<
 *             parts.sort(key=pair_mass)
 *             node.name = extended_key = ''.join("%s%d" % kv for kv in parts)
 */
  }

  /* "glycopeptidepy/_c/structure/fragment.pyx":695
 *             parts.sort(key=pair_mass)
 *             node.name = extended_key = ''.join("%s%d" % kv for kv in parts)
 *         if extended_key:             # <<<<<<<<<<<<<<
 *             name = "%s+%s" % (name, extended_key)
 *         return name
 */
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_extended_key); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 695, __pyx_L1_error)
  if (__pyx_t_4) {

    /* "glycopeptidepy/_c/structure/fragment.pyx":696
 *             node.name = extended_key = ''.join("%s%d" % kv for kv in parts)
 *         if extended_key:
 *             name = "%s+%s" % (name, extended_key)             # <<<<<<<<<<<<<<
 *         return name
 * 
 */
    __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 696, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v_name);
    __Pyx_GIVEREF(__pyx_v_name);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_name);
    __Pyx_INCREF(__pyx_v_extended_key);
    __Pyx_GIVEREF(__pyx_v_extended_key);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_extended_key);
    __pyx_t_5 = __Pyx_PyString_Format(__pyx_kp_s_s_s, __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 696, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_name, ((PyObject*)__pyx_t_5));
    __pyx_t_5 = 0;

    /* "glycopeptidepy/_c/structure/fragment.pyx":695
 *             parts.sort(key=pair_mass)
 *             node.name = extended_key = ''.join("%s%d" % kv for kv in parts)
 *         if extended_key:             # <<<<<<<<<<<<<<
 *             name = "%s+%s" % (name, extended_key)
 *         return name
 */
  }

  /* "glycopeptidepy/_c/structure/fragment.pyx":697
 *         if extended_key:
 *             name = "%s+%s" % (name, extended_key)
 *         return name             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_name);
  __pyx_r = __pyx_v_name;
  goto __pyx_L0;

  /* "glycopeptidepy/_c/structure/fragment.pyx":675
 * 
 * 
 * cdef basestring build_name_from_composition(mapping_types glycan_composition):             # <<<<<<<<<<<<<<
 *         cdef:
 *             basestring name, extended_key
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment.build_name_from_composition", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_name);
  __Pyx_XDECREF(__pyx_v_extended_key);
  __Pyx_XDECREF((PyObject *)__pyx_v_root);
  __Pyx_XDECREF((PyObject *)__pyx_v_node);
  __Pyx_XDECREF(__pyx_gb_14glycopeptidepy_2_c_9structure_8fragment_39__pyx_fuse_1build_name_from_composition_2generator1);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_14glycopeptidepy_2_c_9structure_8fragment_39__pyx_fuse_2build_name_from_composition_2generator2(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "glycopeptidepy/_c/structure/fragment.pyx":694
 *         if extended_key is None:
 *             parts.sort(key=pair_mass)
 *             node.name = extended_key = ''.join("%s%d" % kv for kv in parts)             # <<<<<<<<<<<<<<
 *         if extended_key:
 *             name = "%s+%s" % (name, extended_key)
 */

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_39__pyx_fuse_2build_name_from_composition_genexpr(PyObject *__pyx_self) {
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_5_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_5_genexpr *)__pyx_tp_new_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_5_genexpr(__pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_5_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_5_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 694, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_4___pyx_fuse_2__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition *) __pyx_self;
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_outer_scope));
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_14glycopeptidepy_2_c_9structure_8fragment_39__pyx_fuse_2build_name_from_composition_2generator2, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_pyx_fuse_2build_name_from_comp, __pyx_n_s_glycopeptidepy__c_structure_frag); if (unlikely(!gen)) __PYX_ERR(0, 694, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment.__pyx_fuse_2build_name_from_composition.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_14glycopeptidepy_2_c_9structure_8fragment_39__pyx_fuse_2build_name_from_composition_2generator2(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_5_genexpr *__pyx_cur_scope = ((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_5_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L6_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 694, __pyx_L1_error)
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_parts)) { __Pyx_RaiseClosureNameError("parts"); __PYX_ERR(0, 694, __pyx_L1_error) }
  if (unlikely(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_parts == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(0, 694, __pyx_L1_error)
  }
  __pyx_t_1 = __pyx_cur_scope->__pyx_outer_scope->__pyx_v_parts; __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
  for (;;) {
    if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_1)) break;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_3 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_3); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 694, __pyx_L1_error)
    #else
    __pyx_t_3 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 694, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_kv);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_kv, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_3);
    __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyString_FormatSafe(__pyx_kp_s_s_d, __pyx_cur_scope->__pyx_v_kv); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 694, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    __Pyx_XGIVEREF(__pyx_t_1);
    __pyx_cur_scope->__pyx_t_0 = __pyx_t_1;
    __pyx_cur_scope->__pyx_t_1 = __pyx_t_2;
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
    /* return from generator, yielding value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L6_resume_from_yield:;
    __pyx_t_1 = __pyx_cur_scope->__pyx_t_0;
    __pyx_cur_scope->__pyx_t_0 = 0;
    __Pyx_XGOTREF(__pyx_t_1);
    __pyx_t_2 = __pyx_cur_scope->__pyx_t_1;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 694, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/fragment.pyx":675
 * 
 * 
 * cdef basestring build_name_from_composition(mapping_types glycan_composition):             # <<<<<<<<<<<<<<
 *         cdef:
 *             basestring name, extended_key
 */

static PyObject *__pyx_fuse_2__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition(PyObject *__pyx_v_glycan_composition) {
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_4___pyx_fuse_2__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition *__pyx_cur_scope;
  PyObject *__pyx_v_name = 0;
  PyObject *__pyx_v_extended_key = 0;
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment__NameTree *__pyx_v_root = 0;
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment__NameTree *__pyx_v_node = 0;
  PyObject *__pyx_gb_14glycopeptidepy_2_c_9structure_8fragment_39__pyx_fuse_2build_name_from_composition_2generator2 = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_fuse_2build_name_from_composition", 0);
  __pyx_cur_scope = (struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_4___pyx_fuse_2__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition *)__pyx_tp_new_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_4___pyx_fuse_2__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition(__pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_4___pyx_fuse_2__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_4___pyx_fuse_2__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 675, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }

  /* "glycopeptidepy/_c/structure/fragment.pyx":680
 *             _NameTree root, node
 *             list parts
 *         name = 'peptide'             # <<<<<<<<<<<<<<
 *         root = stub_fragment_name_cache
 *         if mapping_types is dict:
 */
  __Pyx_INCREF(__pyx_n_s_peptide);
  __pyx_v_name = ((PyObject*)__pyx_n_s_peptide);

  /* "glycopeptidepy/_c/structure/fragment.pyx":681
 *             list parts
 *         name = 'peptide'
 *         root = stub_fragment_name_cache             # <<<<<<<<<<<<<<
 *         if mapping_types is dict:
 *             parts = PyDict_Items(glycan_composition)
 */
  __Pyx_INCREF(((PyObject *)__pyx_v_14glycopeptidepy_2_c_9structure_8fragment_stub_fragment_name_cache));
  __pyx_v_root = __pyx_v_14glycopeptidepy_2_c_9structure_8fragment_stub_fragment_name_cache;

  /* "glycopeptidepy/_c/structure/fragment.pyx":687
 *             parts = count_table_items(glycan_composition.table)
 *         elif mapping_types is object:
 *             parts = list(glycan_composition.items())             # <<<<<<<<<<<<<<
 *         # traverse the tree to find the node referring to
 *         # this combination of monosaccharides and counts
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_glycan_composition, __pyx_n_s_items); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 687, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 687, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PySequence_List(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 687, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_cur_scope->__pyx_v_parts = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "glycopeptidepy/_c/structure/fragment.pyx":690
 *         # traverse the tree to find the node referring to
 *         # this combination of monosaccharides and counts
 *         node = root.traverse(parts)             # <<<<<<<<<<<<<<
 *         extended_key = node.name
 *         if extended_key is None:
 */
  __pyx_t_2 = __pyx_cur_scope->__pyx_v_parts;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_1 = ((struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_8fragment__NameTree *)__pyx_v_root->__pyx_vtab)->__pyx_fuse_0traverse(__pyx_v_root, ((PyObject*)__pyx_t_2), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 690, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment__NameTree))))) __PYX_ERR(0, 690, __pyx_L1_error)
  __pyx_v_node = ((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment__NameTree *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "glycopeptidepy/_c/structure/fragment.pyx":691
 *         # this combination of monosaccharides and counts
 *         node = root.traverse(parts)
 *         extended_key = node.name             # <<<<<<<<<<<<<<
 *         if extended_key is None:
 *             parts.sort(key=pair_mass)
 */
  if (!(likely(__Pyx_PyBaseString_CheckExact(__pyx_v_node->name))||((__pyx_v_node->name) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", (PY_MAJOR_VERSION < 3 ? "basestring" : "str"), Py_TYPE(__pyx_v_node->name)->tp_name), 0))) __PYX_ERR(0, 691, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_node->name;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_extended_key = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "glycopeptidepy/_c/structure/fragment.pyx":692
 *         node = root.traverse(parts)
 *         extended_key = node.name
 *         if extended_key is None:             # <<<<<<<<<<<<<<
 *             parts.sort(key=pair_mass)
 *             node.name = extended_key = ''.join("%s%d" % kv for kv in parts)
 */
  __pyx_t_4 = (__pyx_v_extended_key == ((PyObject*)Py_None));
  __pyx_t_5 = (__pyx_t_4 != 0);
  if (__pyx_t_5) {

    /* "glycopeptidepy/_c/structure/fragment.pyx":693
 *         extended_key = node.name
 *         if extended_key is None:
 *             parts.sort(key=pair_mass)             # <<<<<<<<<<<<<<
 *             node.name = extended_key = ''.join("%s%d" % kv for kv in parts)
 *         if extended_key:
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_parts, __pyx_n_s_sort); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 693, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 693, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_CFunc_object____tuple___to_py(__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_pair_mass); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 693, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_key, __pyx_t_3) < 0) __PYX_ERR(0, 693, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_empty_tuple, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 693, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "glycopeptidepy/_c/structure/fragment.pyx":694
 *         if extended_key is None:
 *             parts.sort(key=pair_mass)
 *             node.name = extended_key = ''.join("%s%d" % kv for kv in parts)             # <<<<<<<<<<<<<<
 *         if extended_key:
 *             name = "%s+%s" % (name, extended_key)
 */
    __pyx_t_3 = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_39__pyx_fuse_2build_name_from_composition_genexpr(((PyObject*)__pyx_cur_scope)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 694, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = __Pyx_PyString_Join(__pyx_kp_s_, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 694, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    __Pyx_GOTREF(__pyx_v_node->name);
    __Pyx_DECREF(__pyx_v_node->name);
    __pyx_v_node->name = __pyx_t_2;
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_DECREF_SET(__pyx_v_extended_key, ((PyObject*)__pyx_t_2));
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "glycopeptidepy/_c/structure/fragment.pyx":692
 *         node = root.traverse(parts)
 *         extended_key = node.name
 *         if extended_key is None:             # <<<<<<<<<<<<<<
 *             parts.sort(key=pair_mass)
 *             node.name = extended_key = ''.join("%s%d" % kv for kv in parts)
 */
  }

  /* "glycopeptidepy/_c/structure/fragment.pyx":695
 *             parts.sort(key=pair_mass)
 *             node.name = extended_key = ''.join("%s%d" % kv for kv in parts)
 *         if extended_key:             # <<<<<<<<<<<<<<
 *             name = "%s+%s" % (name, extended_key)
 *         return name
 */
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_v_extended_key); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 695, __pyx_L1_error)
  if (__pyx_t_5) {

    /* "glycopeptidepy/_c/structure/fragment.pyx":696
 *             node.name = extended_key = ''.join("%s%d" % kv for kv in parts)
 *         if extended_key:
 *             name = "%s+%s" % (name, extended_key)             # <<<<<<<<<<<<<<
 *         return name
 * 
 */
    __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 696, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_v_name);
    __Pyx_GIVEREF(__pyx_v_name);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_name);
    __Pyx_INCREF(__pyx_v_extended_key);
    __Pyx_GIVEREF(__pyx_v_extended_key);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_extended_key);
    __pyx_t_3 = __Pyx_PyString_Format(__pyx_kp_s_s_s, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 696, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF_SET(__pyx_v_name, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "glycopeptidepy/_c/structure/fragment.pyx":695
 *             parts.sort(key=pair_mass)
 *             node.name = extended_key = ''.join("%s%d" % kv for kv in parts)
 *         if extended_key:             # <<<<<<<<<<<<<<
 *             name = "%s+%s" % (name, extended_key)
 *         return name
 */
  }

  /* "glycopeptidepy/_c/structure/fragment.pyx":697
 *         if extended_key:
 *             name = "%s+%s" % (name, extended_key)
 *         return name             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_name);
  __pyx_r = __pyx_v_name;
  goto __pyx_L0;

  /* "glycopeptidepy/_c/structure/fragment.pyx":675
 * 
 * 
 * cdef basestring build_name_from_composition(mapping_types glycan_composition):             # <<<<<<<<<<<<<<
 *         cdef:
 *             basestring name, extended_key
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment.build_name_from_composition", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_name);
  __Pyx_XDECREF(__pyx_v_extended_key);
  __Pyx_XDECREF((PyObject *)__pyx_v_root);
  __Pyx_XDECREF((PyObject *)__pyx_v_node);
  __Pyx_XDECREF(__pyx_gb_14glycopeptidepy_2_c_9structure_8fragment_39__pyx_fuse_2build_name_from_composition_2generator2);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/fragment.pyx":705
 * 
 *     @staticmethod
 *     cdef StubFragment _create(str name, double mass, IonSeriesBase kind, CComposition composition,             # <<<<<<<<<<<<<<
 *                               ChemicalShiftBase chemical_shift, bint is_glycosylated, object glycosylation,
 *                               bint is_extended, int glycosylation_size=-1):
 */

static struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_StubFragment *__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment__create(PyObject *__pyx_v_name, double __pyx_v_mass, struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *__pyx_v_kind, struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *__pyx_v_composition, struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase *__pyx_v_chemical_shift, int __pyx_v_is_glycosylated, PyObject *__pyx_v_glycosylation, int __pyx_v_is_extended, struct __pyx_opt_args_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment__create *__pyx_optional_args) {
  int __pyx_v_glycosylation_size = ((int)-1);
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_StubFragment *__pyx_v_self = 0;
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_StubFragment *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_hash_t __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_create", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_glycosylation_size = __pyx_optional_args->glycosylation_size;
    }
  }

  /* "glycopeptidepy/_c/structure/fragment.pyx":711
 *             StubFragment self
 * 
 *         self = StubFragment.__new__(StubFragment)             # <<<<<<<<<<<<<<
 *         self._name = name
 *         self._hash = hash(self._name)
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_14glycopeptidepy_2_c_9structure_8fragment_StubFragment(((PyTypeObject *)__pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment_StubFragment), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 711, __pyx_L1_error)
  __Pyx_GOTREF(((PyObject *)__pyx_t_1));
  __pyx_v_self = ((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_StubFragment *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "glycopeptidepy/_c/structure/fragment.pyx":712
 * 
 *         self = StubFragment.__new__(StubFragment)
 *         self._name = name             # <<<<<<<<<<<<<<
 *         self._hash = hash(self._name)
 *         self._chemical_shift = None
 */
  __Pyx_INCREF(__pyx_v_name);
  __Pyx_GIVEREF(__pyx_v_name);
  __Pyx_GOTREF(__pyx_v_self->__pyx_base._name);
  __Pyx_DECREF(__pyx_v_self->__pyx_base._name);
  __pyx_v_self->__pyx_base._name = __pyx_v_name;

  /* "glycopeptidepy/_c/structure/fragment.pyx":713
 *         self = StubFragment.__new__(StubFragment)
 *         self._name = name
 *         self._hash = hash(self._name)             # <<<<<<<<<<<<<<
 *         self._chemical_shift = None
 *         self._glycosylation_size = glycosylation_size
 */
  __pyx_t_1 = __pyx_v_self->__pyx_base._name;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = PyObject_Hash(__pyx_t_1); if (unlikely(__pyx_t_2 == ((Py_hash_t)-1))) __PYX_ERR(0, 713, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_self->__pyx_base._hash = __pyx_t_2;

  /* "glycopeptidepy/_c/structure/fragment.pyx":714
 *         self._name = name
 *         self._hash = hash(self._name)
 *         self._chemical_shift = None             # <<<<<<<<<<<<<<
 *         self._glycosylation_size = glycosylation_size
 * 
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->__pyx_base._chemical_shift);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->__pyx_base._chemical_shift));
  __pyx_v_self->__pyx_base._chemical_shift = ((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase *)Py_None);

  /* "glycopeptidepy/_c/structure/fragment.pyx":715
 *         self._hash = hash(self._name)
 *         self._chemical_shift = None
 *         self._glycosylation_size = glycosylation_size             # <<<<<<<<<<<<<<
 * 
 *         self.mass = mass
 */
  __pyx_v_self->_glycosylation_size = __pyx_v_glycosylation_size;

  /* "glycopeptidepy/_c/structure/fragment.pyx":717
 *         self._glycosylation_size = glycosylation_size
 * 
 *         self.mass = mass             # <<<<<<<<<<<<<<
 *         self.kind = kind
 *         self.composition = composition
 */
  __pyx_v_self->__pyx_base.mass = __pyx_v_mass;

  /* "glycopeptidepy/_c/structure/fragment.pyx":718
 * 
 *         self.mass = mass
 *         self.kind = kind             # <<<<<<<<<<<<<<
 *         self.composition = composition
 *         if chemical_shift is not None:
 */
  __Pyx_INCREF(((PyObject *)__pyx_v_kind));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_kind));
  __Pyx_GOTREF(__pyx_v_self->kind);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->kind));
  __pyx_v_self->kind = __pyx_v_kind;

  /* "glycopeptidepy/_c/structure/fragment.pyx":719
 *         self.mass = mass
 *         self.kind = kind
 *         self.composition = composition             # <<<<<<<<<<<<<<
 *         if chemical_shift is not None:
 *             StubFragment.set_chemical_shift(self, chemical_shift)
 */
  __Pyx_INCREF(((PyObject *)__pyx_v_composition));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_composition));
  __Pyx_GOTREF(__pyx_v_self->composition);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->composition));
  __pyx_v_self->composition = __pyx_v_composition;

  /* "glycopeptidepy/_c/structure/fragment.pyx":720
 *         self.kind = kind
 *         self.composition = composition
 *         if chemical_shift is not None:             # <<<<<<<<<<<<<<
 *             StubFragment.set_chemical_shift(self, chemical_shift)
 * 
 */
  __pyx_t_3 = (((PyObject *)__pyx_v_chemical_shift) != Py_None);
  __pyx_t_4 = (__pyx_t_3 != 0);
  if (__pyx_t_4) {

    /* "glycopeptidepy/_c/structure/fragment.pyx":721
 *         self.composition = composition
 *         if chemical_shift is not None:
 *             StubFragment.set_chemical_shift(self, chemical_shift)             # <<<<<<<<<<<<<<
 * 
 *         self.is_glycosylated = is_glycosylated
 */
    __pyx_t_1 = __pyx_vtabptr_14glycopeptidepy_2_c_9structure_8fragment_StubFragment->__pyx_base.set_chemical_shift(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *)__pyx_v_self), __pyx_v_chemical_shift, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 721, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "glycopeptidepy/_c/structure/fragment.pyx":720
 *         self.kind = kind
 *         self.composition = composition
 *         if chemical_shift is not None:             # <<<<<<<<<<<<<<
 *             StubFragment.set_chemical_shift(self, chemical_shift)
 * 
 */
  }

  /* "glycopeptidepy/_c/structure/fragment.pyx":723
 *             StubFragment.set_chemical_shift(self, chemical_shift)
 * 
 *         self.is_glycosylated = is_glycosylated             # <<<<<<<<<<<<<<
 *         self.glycosylation = glycosylation
 *         self.is_extended = is_extended
 */
  __pyx_v_self->is_glycosylated = __pyx_v_is_glycosylated;

  /* "glycopeptidepy/_c/structure/fragment.pyx":724
 * 
 *         self.is_glycosylated = is_glycosylated
 *         self.glycosylation = glycosylation             # <<<<<<<<<<<<<<
 *         self.is_extended = is_extended
 * 
 */
  __Pyx_INCREF(__pyx_v_glycosylation);
  __Pyx_GIVEREF(__pyx_v_glycosylation);
  __Pyx_GOTREF(__pyx_v_self->glycosylation);
  __Pyx_DECREF(__pyx_v_self->glycosylation);
  __pyx_v_self->glycosylation = __pyx_v_glycosylation;

  /* "glycopeptidepy/_c/structure/fragment.pyx":725
 *         self.is_glycosylated = is_glycosylated
 *         self.glycosylation = glycosylation
 *         self.is_extended = is_extended             # <<<<<<<<<<<<<<
 * 
 *         return self
 */
  __pyx_v_self->is_extended = __pyx_v_is_extended;

  /* "glycopeptidepy/_c/structure/fragment.pyx":727
 *         self.is_extended = is_extended
 * 
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, name, mass, kind, composition, chemical_shift=None, is_glycosylated=False,
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __Pyx_INCREF(((PyObject *)__pyx_v_self));
  __pyx_r = __pyx_v_self;
  goto __pyx_L0;

  /* "glycopeptidepy/_c/structure/fragment.pyx":705
 * 
 *     @staticmethod
 *     cdef StubFragment _create(str name, double mass, IonSeriesBase kind, CComposition composition,             # <<<<<<<<<<<<<<
 *                               ChemicalShiftBase chemical_shift, bint is_glycosylated, object glycosylation,
 *                               bint is_extended, int glycosylation_size=-1):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment.StubFragment._create", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_self);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/fragment.pyx":729
 *         return self
 * 
 *     def __init__(self, name, mass, kind, composition, chemical_shift=None, is_glycosylated=False,             # <<<<<<<<<<<<<<
 *                  glycosylation=None, is_extended=False):
 *         self._name = name
 */

/* Python wrapper */
static int __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_name = 0;
  PyObject *__pyx_v_mass = 0;
  PyObject *__pyx_v_kind = 0;
  PyObject *__pyx_v_composition = 0;
  PyObject *__pyx_v_chemical_shift = 0;
  PyObject *__pyx_v_is_glycosylated = 0;
  PyObject *__pyx_v_glycosylation = 0;
  PyObject *__pyx_v_is_extended = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_name_2,&__pyx_n_s_mass,&__pyx_n_s_kind,&__pyx_n_s_composition,&__pyx_n_s_chemical_shift,&__pyx_n_s_is_glycosylated,&__pyx_n_s_glycosylation,&__pyx_n_s_is_extended,0};
    PyObject* values[8] = {0,0,0,0,0,0,0,0};
    values[4] = ((PyObject *)Py_None);
    values[5] = ((PyObject *)Py_False);

    /* "glycopeptidepy/_c/structure/fragment.pyx":730
 * 
 *     def __init__(self, name, mass, kind, composition, chemical_shift=None, is_glycosylated=False,
 *                  glycosylation=None, is_extended=False):             # <<<<<<<<<<<<<<
 *         self._name = name
 *         self._hash = hash(self._name)
 */
    values[6] = ((PyObject *)Py_None);
    values[7] = ((PyObject *)Py_False);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_name_2)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_mass)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 4, 8, 1); __PYX_ERR(0, 729, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_kind)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 4, 8, 2); __PYX_ERR(0, 729, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_composition)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 0, 4, 8, 3); __PYX_ERR(0, 729, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_chemical_shift);
          if (value) { values[4] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_is_glycosylated);
          if (value) { values[5] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_glycosylation);
          if (value) { values[6] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_is_extended);
          if (value) { values[7] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 729, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_name = values[0];
    __pyx_v_mass = values[1];
    __pyx_v_kind = values[2];
    __pyx_v_composition = values[3];
    __pyx_v_chemical_shift = values[4];
    __pyx_v_is_glycosylated = values[5];
    __pyx_v_glycosylation = values[6];
    __pyx_v_is_extended = values[7];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 4, 8, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 729, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment.StubFragment.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment___init__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_StubFragment *)__pyx_v_self), __pyx_v_name, __pyx_v_mass, __pyx_v_kind, __pyx_v_composition, __pyx_v_chemical_shift, __pyx_v_is_glycosylated, __pyx_v_glycosylation, __pyx_v_is_extended);

  /* "glycopeptidepy/_c/structure/fragment.pyx":729
 *         return self
 * 
 *     def __init__(self, name, mass, kind, composition, chemical_shift=None, is_glycosylated=False,             # <<<<<<<<<<<<<<
 *                  glycosylation=None, is_extended=False):
 *         self._name = name
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment___init__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_StubFragment *__pyx_v_self, PyObject *__pyx_v_name, PyObject *__pyx_v_mass, PyObject *__pyx_v_kind, PyObject *__pyx_v_composition, PyObject *__pyx_v_chemical_shift, PyObject *__pyx_v_is_glycosylated, PyObject *__pyx_v_glycosylation, PyObject *__pyx_v_is_extended) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_hash_t __pyx_t_2;
  double __pyx_t_3;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "glycopeptidepy/_c/structure/fragment.pyx":731
 *     def __init__(self, name, mass, kind, composition, chemical_shift=None, is_glycosylated=False,
 *                  glycosylation=None, is_extended=False):
 *         self._name = name             # <<<<<<<<<<<<<<
 *         self._hash = hash(self._name)
 *         self._chemical_shift = None
 */
  if (!(likely(PyString_CheckExact(__pyx_v_name))||((__pyx_v_name) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "str", Py_TYPE(__pyx_v_name)->tp_name), 0))) __PYX_ERR(0, 731, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_name;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->__pyx_base._name);
  __Pyx_DECREF(__pyx_v_self->__pyx_base._name);
  __pyx_v_self->__pyx_base._name = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "glycopeptidepy/_c/structure/fragment.pyx":732
 *                  glycosylation=None, is_extended=False):
 *         self._name = name
 *         self._hash = hash(self._name)             # <<<<<<<<<<<<<<
 *         self._chemical_shift = None
 * 
 */
  __pyx_t_1 = __pyx_v_self->__pyx_base._name;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = PyObject_Hash(__pyx_t_1); if (unlikely(__pyx_t_2 == ((Py_hash_t)-1))) __PYX_ERR(0, 732, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_self->__pyx_base._hash = __pyx_t_2;

  /* "glycopeptidepy/_c/structure/fragment.pyx":733
 *         self._name = name
 *         self._hash = hash(self._name)
 *         self._chemical_shift = None             # <<<<<<<<<<<<<<
 * 
 *         self.mass = mass
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->__pyx_base._chemical_shift);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->__pyx_base._chemical_shift));
  __pyx_v_self->__pyx_base._chemical_shift = ((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase *)Py_None);

  /* "glycopeptidepy/_c/structure/fragment.pyx":735
 *         self._chemical_shift = None
 * 
 *         self.mass = mass             # <<<<<<<<<<<<<<
 *         self.kind = kind
 *         self.composition = composition
 */
  __pyx_t_3 = __pyx_PyFloat_AsDouble(__pyx_v_mass); if (unlikely((__pyx_t_3 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 735, __pyx_L1_error)
  __pyx_v_self->__pyx_base.mass = __pyx_t_3;

  /* "glycopeptidepy/_c/structure/fragment.pyx":736
 * 
 *         self.mass = mass
 *         self.kind = kind             # <<<<<<<<<<<<<<
 *         self.composition = composition
 * 
 */
  if (!(likely(((__pyx_v_kind) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_kind, __pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase))))) __PYX_ERR(0, 736, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_kind;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->kind);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->kind));
  __pyx_v_self->kind = ((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "glycopeptidepy/_c/structure/fragment.pyx":737
 *         self.mass = mass
 *         self.kind = kind
 *         self.composition = composition             # <<<<<<<<<<<<<<
 * 
 *         self.set_chemical_shift(chemical_shift)
 */
  if (!(likely(((__pyx_v_composition) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_composition, __pyx_ptype_5glypy_11composition_12ccomposition_CComposition))))) __PYX_ERR(0, 737, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_composition;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->composition);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->composition));
  __pyx_v_self->composition = ((struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "glycopeptidepy/_c/structure/fragment.pyx":739
 *         self.composition = composition
 * 
 *         self.set_chemical_shift(chemical_shift)             # <<<<<<<<<<<<<<
 *         self.is_glycosylated = is_glycosylated
 * 
 */
  if (!(likely(((__pyx_v_chemical_shift) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_chemical_shift, __pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase))))) __PYX_ERR(0, 739, __pyx_L1_error)
  __pyx_t_1 = ((struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_8fragment_StubFragment *)__pyx_v_self->__pyx_base.__pyx_vtab)->__pyx_base.set_chemical_shift(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *)__pyx_v_self), ((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase *)__pyx_v_chemical_shift), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 739, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "glycopeptidepy/_c/structure/fragment.pyx":740
 * 
 *         self.set_chemical_shift(chemical_shift)
 *         self.is_glycosylated = is_glycosylated             # <<<<<<<<<<<<<<
 * 
 *         self.glycosylation = glycosylation
 */
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_is_glycosylated); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 740, __pyx_L1_error)
  __pyx_v_self->is_glycosylated = __pyx_t_4;

  /* "glycopeptidepy/_c/structure/fragment.pyx":742
 *         self.is_glycosylated = is_glycosylated
 * 
 *         self.glycosylation = glycosylation             # <<<<<<<<<<<<<<
 *         self.is_extended = is_extended
 *         self._glycosylation_size = -1
 */
  __Pyx_INCREF(__pyx_v_glycosylation);
  __Pyx_GIVEREF(__pyx_v_glycosylation);
  __Pyx_GOTREF(__pyx_v_self->glycosylation);
  __Pyx_DECREF(__pyx_v_self->glycosylation);
  __pyx_v_self->glycosylation = __pyx_v_glycosylation;

  /* "glycopeptidepy/_c/structure/fragment.pyx":743
 * 
 *         self.glycosylation = glycosylation
 *         self.is_extended = is_extended             # <<<<<<<<<<<<<<
 *         self._glycosylation_size = -1
 * 
 */
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_is_extended); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 743, __pyx_L1_error)
  __pyx_v_self->is_extended = __pyx_t_4;

  /* "glycopeptidepy/_c/structure/fragment.pyx":744
 *         self.glycosylation = glycosylation
 *         self.is_extended = is_extended
 *         self._glycosylation_size = -1             # <<<<<<<<<<<<<<
 * 
 *     cpdef clone(self):
 */
  __pyx_v_self->_glycosylation_size = -1;

  /* "glycopeptidepy/_c/structure/fragment.pyx":729
 *         return self
 * 
 *     def __init__(self, name, mass, kind, composition, chemical_shift=None, is_glycosylated=False,             # <<<<<<<<<<<<<<
 *                  glycosylation=None, is_extended=False):
 *         self._name = name
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment.StubFragment.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/fragment.pyx":746
 *         self._glycosylation_size = -1
 * 
 *     cpdef clone(self):             # <<<<<<<<<<<<<<
 *         cdef StubFragment dup = StubFragment._create(
 *             self.name, self.mass, self.kind, self.composition, self._chemical_shift,
 */

static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_3clone(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_clone(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_StubFragment *__pyx_v_self, int __pyx_skip_dispatch) {
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_StubFragment *__pyx_v_dup = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  struct __pyx_opt_args_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment__create __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("clone", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_clone); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 746, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_3clone)) {
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 746, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "glycopeptidepy/_c/structure/fragment.pyx":748
 *     cpdef clone(self):
 *         cdef StubFragment dup = StubFragment._create(
 *             self.name, self.mass, self.kind, self.composition, self._chemical_shift,             # <<<<<<<<<<<<<<
 *             self.is_glycosylated, self.glycosylation, self.is_extended,
 *             self._glycosylation_size)
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_name_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 748, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyString_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "str", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(0, 748, __pyx_L1_error)
  __pyx_t_2 = ((PyObject *)__pyx_v_self->kind);
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = ((PyObject *)__pyx_v_self->composition);
  __Pyx_INCREF(__pyx_t_3);
  __pyx_t_4 = ((PyObject *)__pyx_v_self->__pyx_base._chemical_shift);
  __Pyx_INCREF(__pyx_t_4);

  /* "glycopeptidepy/_c/structure/fragment.pyx":749
 *         cdef StubFragment dup = StubFragment._create(
 *             self.name, self.mass, self.kind, self.composition, self._chemical_shift,
 *             self.is_glycosylated, self.glycosylation, self.is_extended,             # <<<<<<<<<<<<<<
 *             self._glycosylation_size)
 *         return dup
 */
  __pyx_t_5 = __pyx_v_self->glycosylation;
  __Pyx_INCREF(__pyx_t_5);

  /* "glycopeptidepy/_c/structure/fragment.pyx":747
 * 
 *     cpdef clone(self):
 *         cdef StubFragment dup = StubFragment._create(             # <<<<<<<<<<<<<<
 *             self.name, self.mass, self.kind, self.composition, self._chemical_shift,
 *             self.is_glycosylated, self.glycosylation, self.is_extended,
 */
  __pyx_t_7.__pyx_n = 1;
  __pyx_t_7.glycosylation_size = __pyx_v_self->_glycosylation_size;
  __pyx_t_6 = ((PyObject *)__pyx_vtabptr_14glycopeptidepy_2_c_9structure_8fragment_StubFragment->_create(((PyObject*)__pyx_t_1), __pyx_v_self->__pyx_base.mass, ((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *)__pyx_t_2), ((struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *)__pyx_t_3), ((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase *)__pyx_t_4), __pyx_v_self->is_glycosylated, __pyx_t_5, __pyx_v_self->is_extended, &__pyx_t_7)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 747, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_v_dup = ((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_StubFragment *)__pyx_t_6);
  __pyx_t_6 = 0;

  /* "glycopeptidepy/_c/structure/fragment.pyx":751
 *             self.is_glycosylated, self.glycosylation, self.is_extended,
 *             self._glycosylation_size)
 *         return dup             # <<<<<<<<<<<<<<
 * 
 *     cdef int get_glycosylation_size(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_dup));
  __pyx_r = ((PyObject *)__pyx_v_dup);
  goto __pyx_L0;

  /* "glycopeptidepy/_c/structure/fragment.pyx":746
 *         self._glycosylation_size = -1
 * 
 *     cpdef clone(self):             # <<<<<<<<<<<<<<
 *         cdef StubFragment dup = StubFragment._create(
 *             self.name, self.mass, self.kind, self.composition, self._chemical_shift,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment.StubFragment.clone", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_dup);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_3clone(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_2clone[] = "StubFragment.clone(self)";
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_3clone(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("clone (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_2clone(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_StubFragment *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_2clone(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_StubFragment *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("clone", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_clone(__pyx_v_self, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 746, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment.StubFragment.clone", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/fragment.pyx":753
 *         return dup
 * 
 *     cdef int get_glycosylation_size(self):             # <<<<<<<<<<<<<<
 *         if self._glycosylation_size == -1:
 *             self._glycosylation_size = sum(self.glycosylation.values())
 */

static int __pyx_f_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_get_glycosylation_size(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_StubFragment *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_glycosylation_size", 0);

  /* "glycopeptidepy/_c/structure/fragment.pyx":754
 * 
 *     cdef int get_glycosylation_size(self):
 *         if self._glycosylation_size == -1:             # <<<<<<<<<<<<<<
 *             self._glycosylation_size = sum(self.glycosylation.values())
 *         return self._glycosylation_size
 */
  __pyx_t_1 = ((__pyx_v_self->_glycosylation_size == -1L) != 0);
  if (__pyx_t_1) {

    /* "glycopeptidepy/_c/structure/fragment.pyx":755
 *     cdef int get_glycosylation_size(self):
 *         if self._glycosylation_size == -1:
 *             self._glycosylation_size = sum(self.glycosylation.values())             # <<<<<<<<<<<<<<
 *         return self._glycosylation_size
 * 
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->glycosylation, __pyx_n_s_values); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 755, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 755, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_sum, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 755, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_5 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 755, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_self->_glycosylation_size = __pyx_t_5;

    /* "glycopeptidepy/_c/structure/fragment.pyx":754
 * 
 *     cdef int get_glycosylation_size(self):
 *         if self._glycosylation_size == -1:             # <<<<<<<<<<<<<<
 *             self._glycosylation_size = sum(self.glycosylation.values())
 *         return self._glycosylation_size
 */
  }

  /* "glycopeptidepy/_c/structure/fragment.pyx":756
 *         if self._glycosylation_size == -1:
 *             self._glycosylation_size = sum(self.glycosylation.values())
 *         return self._glycosylation_size             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __pyx_r = __pyx_v_self->_glycosylation_size;
  goto __pyx_L0;

  /* "glycopeptidepy/_c/structure/fragment.pyx":753
 *         return dup
 * 
 *     cdef int get_glycosylation_size(self):             # <<<<<<<<<<<<<<
 *         if self._glycosylation_size == -1:
 *             self._glycosylation_size = sum(self.glycosylation.values())
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_WriteUnraisable("glycopeptidepy._c.structure.fragment.StubFragment.get_glycosylation_size", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/fragment.pyx":759
 * 
 *     @property
 *     def glycosylation_size(self):             # <<<<<<<<<<<<<<
 *         return self.get_glycosylation_size()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_18glycosylation_size_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_18glycosylation_size_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_18glycosylation_size___get__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_StubFragment *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_18glycosylation_size___get__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_StubFragment *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "glycopeptidepy/_c/structure/fragment.pyx":760
 *     @property
 *     def glycosylation_size(self):
 *         return self.get_glycosylation_size()             # <<<<<<<<<<<<<<
 * 
 *     cpdef str base_name(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(((struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_8fragment_StubFragment *)__pyx_v_self->__pyx_base.__pyx_vtab)->get_glycosylation_size(__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 760, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "glycopeptidepy/_c/structure/fragment.pyx":759
 * 
 *     @property
 *     def glycosylation_size(self):             # <<<<<<<<<<<<<<
 *         return self.get_glycosylation_size()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment.StubFragment.glycosylation_size.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/fragment.pyx":762
 *         return self.get_glycosylation_size()
 * 
 *     cpdef str base_name(self):             # <<<<<<<<<<<<<<
 *         """Simply return the base fragment's name, omitting any chemical modification,
 *         e.g. peptide+HexNAc1"""
 */

static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_5base_name(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_base_name(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_StubFragment *__pyx_v_self, int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("base_name", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_base_name); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 762, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_5base_name)) {
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 762, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (!(likely(PyString_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "str", Py_TYPE(__pyx_t_2)->tp_name), 0))) __PYX_ERR(0, 762, __pyx_L1_error)
        __pyx_r = ((PyObject*)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "glycopeptidepy/_c/structure/fragment.pyx":769
 *         # so just return it, otherwise we need to recompute the name from scratch using
 *         # :attr:`glycosylation` and :meth:`build_name_from_composition`
 *         if self._chemical_shift is None:             # <<<<<<<<<<<<<<
 *             return self._name
 *         else:
 */
  __pyx_t_5 = (((PyObject *)__pyx_v_self->__pyx_base._chemical_shift) == Py_None);
  __pyx_t_6 = (__pyx_t_5 != 0);
  if (__pyx_t_6) {

    /* "glycopeptidepy/_c/structure/fragment.pyx":770
 *         # :attr:`glycosylation` and :meth:`build_name_from_composition`
 *         if self._chemical_shift is None:
 *             return self._name             # <<<<<<<<<<<<<<
 *         else:
 *             return self.build_name_from_composition(self.glycosylation)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_self->__pyx_base._name);
    __pyx_r = __pyx_v_self->__pyx_base._name;
    goto __pyx_L0;

    /* "glycopeptidepy/_c/structure/fragment.pyx":769
 *         # so just return it, otherwise we need to recompute the name from scratch using
 *         # :attr:`glycosylation` and :meth:`build_name_from_composition`
 *         if self._chemical_shift is None:             # <<<<<<<<<<<<<<
 *             return self._name
 *         else:
 */
  }

  /* "glycopeptidepy/_c/structure/fragment.pyx":772
 *             return self._name
 *         else:
 *             return self.build_name_from_composition(self.glycosylation)             # <<<<<<<<<<<<<<
 * 
 *     @classmethod
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_build_name_from_composition); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 772, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_3, __pyx_v_self->glycosylation) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_self->glycosylation);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 772, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (!(likely(PyString_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "str", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(0, 772, __pyx_L1_error)
    __pyx_r = ((PyObject*)__pyx_t_1);
    __pyx_t_1 = 0;
    goto __pyx_L0;
  }

  /* "glycopeptidepy/_c/structure/fragment.pyx":762
 *         return self.get_glycosylation_size()
 * 
 *     cpdef str base_name(self):             # <<<<<<<<<<<<<<
 *         """Simply return the base fragment's name, omitting any chemical modification,
 *         e.g. peptide+HexNAc1"""
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment.StubFragment.base_name", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_5base_name(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_4base_name[] = "StubFragment.base_name(self) -> str\nSimply return the base fragment's name, omitting any chemical modification,\n        e.g. peptide+HexNAc1";
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_5base_name(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("base_name (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_4base_name(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_StubFragment *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_4base_name(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_StubFragment *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("base_name", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_base_name(__pyx_v_self, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 762, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment.StubFragment.base_name", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/fragment.pyx":775
 * 
 *     @classmethod
 *     def build_name_from_composition(cls, glycan_composition):             # <<<<<<<<<<<<<<
 *         return build_name_from_composition(glycan_composition)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_7build_name_from_composition(PyObject *__pyx_v_cls, PyObject *__pyx_v_glycan_composition); /*proto*/
static char __pyx_doc_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_6build_name_from_composition[] = "StubFragment.build_name_from_composition(type cls, glycan_composition)";
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_7build_name_from_composition(PyObject *__pyx_v_cls, PyObject *__pyx_v_glycan_composition) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("build_name_from_composition (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_6build_name_from_composition(((PyTypeObject*)__pyx_v_cls), ((PyObject *)__pyx_v_glycan_composition));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_6build_name_from_composition(CYTHON_UNUSED PyTypeObject *__pyx_v_cls, PyObject *__pyx_v_glycan_composition) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("build_name_from_composition", 0);

  /* "glycopeptidepy/_c/structure/fragment.pyx":776
 *     @classmethod
 *     def build_name_from_composition(cls, glycan_composition):
 *         return build_name_from_composition(glycan_composition)             # <<<<<<<<<<<<<<
 * 
 *     def __reduce__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_fuse_2__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition(__pyx_v_glycan_composition); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 776, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "glycopeptidepy/_c/structure/fragment.pyx":775
 * 
 *     @classmethod
 *     def build_name_from_composition(cls, glycan_composition):             # <<<<<<<<<<<<<<
 *         return build_name_from_composition(glycan_composition)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment.StubFragment.build_name_from_composition", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/fragment.pyx":778
 *         return build_name_from_composition(glycan_composition)
 * 
 *     def __reduce__(self):             # <<<<<<<<<<<<<<
 *         corrected_mass = self.mass
 *         if self._chemical_shift is not None:
 */

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_9__reduce__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_8__reduce__[] = "StubFragment.__reduce__(self)";
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_9__reduce__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_8__reduce__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_StubFragment *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_8__reduce__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_StubFragment *__pyx_v_self) {
  PyObject *__pyx_v_corrected_mass = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce__", 0);

  /* "glycopeptidepy/_c/structure/fragment.pyx":779
 * 
 *     def __reduce__(self):
 *         corrected_mass = self.mass             # <<<<<<<<<<<<<<
 *         if self._chemical_shift is not None:
 *             corrected_mass -= self.chemical_shift.mass
 */
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->__pyx_base.mass); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 779, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_corrected_mass = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "glycopeptidepy/_c/structure/fragment.pyx":780
 *     def __reduce__(self):
 *         corrected_mass = self.mass
 *         if self._chemical_shift is not None:             # <<<<<<<<<<<<<<
 *             corrected_mass -= self.chemical_shift.mass
 *         return self.__class__, (self.name, corrected_mass, self.kind, self.composition,
 */
  __pyx_t_2 = (((PyObject *)__pyx_v_self->__pyx_base._chemical_shift) != Py_None);
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {

    /* "glycopeptidepy/_c/structure/fragment.pyx":781
 *         corrected_mass = self.mass
 *         if self._chemical_shift is not None:
 *             corrected_mass -= self.chemical_shift.mass             # <<<<<<<<<<<<<<
 *         return self.__class__, (self.name, corrected_mass, self.kind, self.composition,
 *                                 self.chemical_shift, self.is_glycosylated, self.glycosylation,
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_chemical_shift); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 781, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_mass); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 781, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = PyNumber_InPlaceSubtract(__pyx_v_corrected_mass, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 781, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF_SET(__pyx_v_corrected_mass, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "glycopeptidepy/_c/structure/fragment.pyx":780
 *     def __reduce__(self):
 *         corrected_mass = self.mass
 *         if self._chemical_shift is not None:             # <<<<<<<<<<<<<<
 *             corrected_mass -= self.chemical_shift.mass
 *         return self.__class__, (self.name, corrected_mass, self.kind, self.composition,
 */
  }

  /* "glycopeptidepy/_c/structure/fragment.pyx":782
 *         if self._chemical_shift is not None:
 *             corrected_mass -= self.chemical_shift.mass
 *         return self.__class__, (self.name, corrected_mass, self.kind, self.composition,             # <<<<<<<<<<<<<<
 *                                 self.chemical_shift, self.is_glycosylated, self.glycosylation,
 *                                 self.is_extended)
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_class); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 782, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_name_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 782, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);

  /* "glycopeptidepy/_c/structure/fragment.pyx":783
 *             corrected_mass -= self.chemical_shift.mass
 *         return self.__class__, (self.name, corrected_mass, self.kind, self.composition,
 *                                 self.chemical_shift, self.is_glycosylated, self.glycosylation,             # <<<<<<<<<<<<<<
 *                                 self.is_extended)
 * 
 */
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_chemical_shift); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 783, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_PyBool_FromLong(__pyx_v_self->is_glycosylated); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 783, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);

  /* "glycopeptidepy/_c/structure/fragment.pyx":784
 *         return self.__class__, (self.name, corrected_mass, self.kind, self.composition,
 *                                 self.chemical_shift, self.is_glycosylated, self.glycosylation,
 *                                 self.is_extended)             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
  __pyx_t_7 = __Pyx_PyBool_FromLong(__pyx_v_self->is_extended); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 784, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);

  /* "glycopeptidepy/_c/structure/fragment.pyx":782
 *         if self._chemical_shift is not None:
 *             corrected_mass -= self.chemical_shift.mass
 *         return self.__class__, (self.name, corrected_mass, self.kind, self.composition,             # <<<<<<<<<<<<<<
 *                                 self.chemical_shift, self.is_glycosylated, self.glycosylation,
 *                                 self.is_extended)
 */
  __pyx_t_8 = PyTuple_New(8); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 782, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_4);
  __Pyx_INCREF(__pyx_v_corrected_mass);
  __Pyx_GIVEREF(__pyx_v_corrected_mass);
  PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_v_corrected_mass);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->kind));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self->kind));
  PyTuple_SET_ITEM(__pyx_t_8, 2, ((PyObject *)__pyx_v_self->kind));
  __Pyx_INCREF(((PyObject *)__pyx_v_self->composition));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self->composition));
  PyTuple_SET_ITEM(__pyx_t_8, 3, ((PyObject *)__pyx_v_self->composition));
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_8, 4, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_8, 5, __pyx_t_6);
  __Pyx_INCREF(__pyx_v_self->glycosylation);
  __Pyx_GIVEREF(__pyx_v_self->glycosylation);
  PyTuple_SET_ITEM(__pyx_t_8, 6, __pyx_v_self->glycosylation);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_8, 7, __pyx_t_7);
  __pyx_t_4 = 0;
  __pyx_t_5 = 0;
  __pyx_t_6 = 0;
  __pyx_t_7 = 0;
  __pyx_t_7 = PyTuple_New(2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 782, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_8);
  PyTuple_SET_ITEM(__pyx_t_7, 1, __pyx_t_8);
  __pyx_t_1 = 0;
  __pyx_t_8 = 0;
  __pyx_r = __pyx_t_7;
  __pyx_t_7 = 0;
  goto __pyx_L0;

  /* "glycopeptidepy/_c/structure/fragment.pyx":778
 *         return build_name_from_composition(glycan_composition)
 * 
 *     def __reduce__(self):             # <<<<<<<<<<<<<<
 *         corrected_mass = self.mass
 *         if self._chemical_shift is not None:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment.StubFragment.__reduce__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_corrected_mass);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/fragment.pyx":786
 *                                 self.is_extended)
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return ("{self.__class__.__name__}(name={self.name}, "
 *                 "mass={self.mass:.04f}, series={self.kind})").format(self=self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_11__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_11__repr__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_10__repr__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_StubFragment *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_10__repr__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_StubFragment *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "glycopeptidepy/_c/structure/fragment.pyx":787
 * 
 *     def __repr__(self):
 *         return ("{self.__class__.__name__}(name={self.name}, "             # <<<<<<<<<<<<<<
 *                 "mass={self.mass:.04f}, series={self.kind})").format(self=self)
 * 
 */
  __Pyx_XDECREF(__pyx_r);

  /* "glycopeptidepy/_c/structure/fragment.pyx":788
 *     def __repr__(self):
 *         return ("{self.__class__.__name__}(name={self.name}, "
 *                 "mass={self.mass:.04f}, series={self.kind})").format(self=self)             # <<<<<<<<<<<<<<
 * 
 *     cpdef IonSeriesBase get_series(self):
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_kp_s_self___class_____name___name_se, __pyx_n_s_format); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 788, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 788, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_self, ((PyObject *)__pyx_v_self)) < 0) __PYX_ERR(0, 788, __pyx_L1_error)
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_empty_tuple, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 788, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "glycopeptidepy/_c/structure/fragment.pyx":786
 *                                 self.is_extended)
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return ("{self.__class__.__name__}(name={self.name}, "
 *                 "mass={self.mass:.04f}, series={self.kind})").format(self=self)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment.StubFragment.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/fragment.pyx":790
 *                 "mass={self.mass:.04f}, series={self.kind})").format(self=self)
 * 
 *     cpdef IonSeriesBase get_series(self):             # <<<<<<<<<<<<<<
 *         return self.kind
 */

static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_13get_series(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_get_series(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_StubFragment *__pyx_v_self, int __pyx_skip_dispatch) {
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_series", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_get_series); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 790, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_13get_series)) {
        __Pyx_XDECREF(((PyObject *)__pyx_r));
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 790, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase))))) __PYX_ERR(0, 790, __pyx_L1_error)
        __pyx_r = ((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "glycopeptidepy/_c/structure/fragment.pyx":791
 * 
 *     cpdef IonSeriesBase get_series(self):
 *         return self.kind             # <<<<<<<<<<<<<<
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __Pyx_INCREF(((PyObject *)__pyx_v_self->kind));
  __pyx_r = __pyx_v_self->kind;
  goto __pyx_L0;

  /* "glycopeptidepy/_c/structure/fragment.pyx":790
 *                 "mass={self.mass:.04f}, series={self.kind})").format(self=self)
 * 
 *     cpdef IonSeriesBase get_series(self):             # <<<<<<<<<<<<<<
 *         return self.kind
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment.StubFragment.get_series", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_13get_series(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_12get_series[] = "StubFragment.get_series(self) -> IonSeriesBase";
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_13get_series(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_series (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_12get_series(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_StubFragment *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_12get_series(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_StubFragment *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_series", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_get_series(__pyx_v_self, 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 790, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment.StubFragment.get_series", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/fragment.pxd":74
 * cdef class StubFragment(FragmentBase):
 *     cdef:
 *         public IonSeriesBase kind             # <<<<<<<<<<<<<<
 *         public bint is_glycosylated
 *         public CComposition composition
 */

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_4kind_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_4kind_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_4kind___get__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_StubFragment *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_4kind___get__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_StubFragment *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->kind));
  __pyx_r = ((PyObject *)__pyx_v_self->kind);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_4kind_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_4kind_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_4kind_2__set__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_StubFragment *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_4kind_2__set__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_StubFragment *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);
  if (!(likely(((__pyx_v_value) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_value, __pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase))))) __PYX_ERR(1, 74, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_value;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->kind);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->kind));
  __pyx_v_self->kind = ((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment.StubFragment.kind.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_4kind_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_4kind_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_4kind_4__del__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_StubFragment *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_4kind_4__del__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_StubFragment *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->kind);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->kind));
  __pyx_v_self->kind = ((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *)Py_None);

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/fragment.pxd":75
 *     cdef:
 *         public IonSeriesBase kind
 *         public bint is_glycosylated             # <<<<<<<<<<<<<<
 *         public CComposition composition
 *         public object glycosylation
 */

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_15is_glycosylated_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_15is_glycosylated_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_15is_glycosylated___get__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_StubFragment *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_15is_glycosylated___get__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_StubFragment *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->is_glycosylated); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 75, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment.StubFragment.is_glycosylated.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_15is_glycosylated_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_15is_glycosylated_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_15is_glycosylated_2__set__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_StubFragment *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_15is_glycosylated_2__set__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_StubFragment *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_value); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 75, __pyx_L1_error)
  __pyx_v_self->is_glycosylated = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment.StubFragment.is_glycosylated.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/fragment.pxd":76
 *         public IonSeriesBase kind
 *         public bint is_glycosylated
 *         public CComposition composition             # <<<<<<<<<<<<<<
 *         public object glycosylation
 *         public bint is_extended
 */

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_11composition_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_11composition_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_11composition___get__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_StubFragment *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_11composition___get__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_StubFragment *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->composition));
  __pyx_r = ((PyObject *)__pyx_v_self->composition);
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_11composition_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_11composition_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_11composition_2__set__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_StubFragment *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_11composition_2__set__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_StubFragment *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);
  if (!(likely(((__pyx_v_value) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_value, __pyx_ptype_5glypy_11composition_12ccomposition_CComposition))))) __PYX_ERR(1, 76, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_value;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->composition);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->composition));
  __pyx_v_self->composition = ((struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment.StubFragment.composition.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_11composition_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_11composition_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_11composition_4__del__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_StubFragment *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_11composition_4__del__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_StubFragment *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->composition);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->composition));
  __pyx_v_self->composition = ((struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *)Py_None);

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/fragment.pxd":77
 *         public bint is_glycosylated
 *         public CComposition composition
 *         public object glycosylation             # <<<<<<<<<<<<<<
 *         public bint is_extended
 *         int _glycosylation_size
 */

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_13glycosylation_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_13glycosylation_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_13glycosylation___get__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_StubFragment *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_13glycosylation___get__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_StubFragment *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->glycosylation);
  __pyx_r = __pyx_v_self->glycosylation;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_13glycosylation_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_13glycosylation_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_13glycosylation_2__set__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_StubFragment *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_13glycosylation_2__set__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_StubFragment *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 0);
  __Pyx_INCREF(__pyx_v_value);
  __Pyx_GIVEREF(__pyx_v_value);
  __Pyx_GOTREF(__pyx_v_self->glycosylation);
  __Pyx_DECREF(__pyx_v_self->glycosylation);
  __pyx_v_self->glycosylation = __pyx_v_value;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_13glycosylation_5__del__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_13glycosylation_5__del__(PyObject *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_13glycosylation_4__del__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_StubFragment *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_13glycosylation_4__del__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_StubFragment *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__del__", 0);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->glycosylation);
  __Pyx_DECREF(__pyx_v_self->glycosylation);
  __pyx_v_self->glycosylation = Py_None;

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "glycopeptidepy/_c/structure/fragment.pxd":78
 *         public CComposition composition
 *         public object glycosylation
 *         public bint is_extended             # <<<<<<<<<<<<<<
 *         int _glycosylation_size
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_11is_extended_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_11is_extended_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_11is_extended___get__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_StubFragment *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_11is_extended___get__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_StubFragment *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->is_extended); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 78, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment.StubFragment.is_extended.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static int __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_11is_extended_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_11is_extended_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_11is_extended_2__set__(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_StubFragment *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_11is_extended_2__set__(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_StubFragment *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 0);
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_value); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 78, __pyx_L1_error)
  __pyx_v_self->is_extended = __pyx_t_1;

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment.StubFragment.is_extended.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_ChemicalShiftBase(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_5__pyx_unpickle_ChemicalShiftBase(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_14glycopeptidepy_2_c_9structure_8fragment_4__pyx_unpickle_ChemicalShiftBase[] = "__pyx_unpickle_ChemicalShiftBase(__pyx_type, long __pyx_checksum, __pyx_state)";
static PyMethodDef __pyx_mdef_14glycopeptidepy_2_c_9structure_8fragment_5__pyx_unpickle_ChemicalShiftBase = {"__pyx_unpickle_ChemicalShiftBase", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_5__pyx_unpickle_ChemicalShiftBase, METH_VARARGS|METH_KEYWORDS, __pyx_doc_14glycopeptidepy_2_c_9structure_8fragment_4__pyx_unpickle_ChemicalShiftBase};
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_5__pyx_unpickle_ChemicalShiftBase(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_ChemicalShiftBase (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_type)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_checksum)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_ChemicalShiftBase", 1, 3, 3, 1); __PYX_ERR(2, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_ChemicalShiftBase", 1, 3, 3, 2); __PYX_ERR(2, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_unpickle_ChemicalShiftBase") < 0)) __PYX_ERR(2, 1, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_ChemicalShiftBase", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(2, 1, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment.__pyx_unpickle_ChemicalShiftBase", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_4__pyx_unpickle_ChemicalShiftBase(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_4__pyx_unpickle_ChemicalShiftBase(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_ChemicalShiftBase", 0);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x54d98d7, 0xbef1095, 0x80a4bd4):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0x54d98d7, 0xbef1095, 0x80a4bd4) = (composition, mass, name))" % __pyx_checksum)
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__7, Py_NE)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x54d98d7, 0xbef1095, 0x80a4bd4):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0x54d98d7, 0xbef1095, 0x80a4bd4) = (composition, mass, name))" % __pyx_checksum)
 *     __pyx_result = ChemicalShiftBase.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError);
    __pyx_t_4 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, -1); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_4, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0x54d98d7, 0xbef1095, 0x80a4bd4):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0x54d98d7, 0xbef1095, 0x80a4bd4) = (composition, mass, name))" % __pyx_checksum)             # <<<<<<<<<<<<<<
 *     __pyx_result = ChemicalShiftBase.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0, __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_INCREF(__pyx_v___pyx_PickleError);
    __pyx_t_1 = __pyx_v___pyx_PickleError; __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_4 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_6, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_5);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __PYX_ERR(2, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x54d98d7, 0xbef1095, 0x80a4bd4):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0x54d98d7, 0xbef1095, 0x80a4bd4) = (composition, mass, name))" % __pyx_checksum)
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0x54d98d7, 0xbef1095, 0x80a4bd4) = (composition, mass, name))" % __pyx_checksum)
 *     __pyx_result = ChemicalShiftBase.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_ChemicalShiftBase__set_state(<ChemicalShiftBase> __pyx_result, __pyx_state)
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase), __pyx_n_s_new); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_4 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_5, __pyx_v___pyx_type) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v___pyx_type);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0x54d98d7, 0xbef1095, 0x80a4bd4) = (composition, mass, name))" % __pyx_checksum)
 *     __pyx_result = ChemicalShiftBase.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_ChemicalShiftBase__set_state(<ChemicalShiftBase> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_3 = (__pyx_v___pyx_state != Py_None);
  __pyx_t_2 = (__pyx_t_3 != 0);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = ChemicalShiftBase.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_ChemicalShiftBase__set_state(<ChemicalShiftBase> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_ChemicalShiftBase__set_state(ChemicalShiftBase __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(2, 9, __pyx_L1_error)
    __pyx_t_4 = __pyx_f_14glycopeptidepy_2_c_9structure_8fragment___pyx_unpickle_ChemicalShiftBase__set_state(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0x54d98d7, 0xbef1095, 0x80a4bd4) = (composition, mass, name))" % __pyx_checksum)
 *     __pyx_result = ChemicalShiftBase.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_ChemicalShiftBase__set_state(<ChemicalShiftBase> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_ChemicalShiftBase__set_state(<ChemicalShiftBase> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_ChemicalShiftBase__set_state(ChemicalShiftBase __pyx_result, tuple __pyx_state):
 *     __pyx_result.composition = __pyx_state[0]; __pyx_result.mass = __pyx_state[1]; __pyx_result.name = __pyx_state[2]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_ChemicalShiftBase(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment.__pyx_unpickle_ChemicalShiftBase", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_ChemicalShiftBase__set_state(<ChemicalShiftBase> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_ChemicalShiftBase__set_state(ChemicalShiftBase __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.composition = __pyx_state[0]; __pyx_result.mass = __pyx_state[1]; __pyx_result.name = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_14glycopeptidepy_2_c_9structure_8fragment___pyx_unpickle_ChemicalShiftBase__set_state(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  double __pyx_t_2;
  int __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_ChemicalShiftBase__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_ChemicalShiftBase__set_state(ChemicalShiftBase __pyx_result, tuple __pyx_state):
 *     __pyx_result.composition = __pyx_state[0]; __pyx_result.mass = __pyx_state[1]; __pyx_result.name = __pyx_state[2]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[3])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5glypy_11composition_12ccomposition_CComposition))))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->composition);
  __Pyx_DECREF(((PyObject *)__pyx_v___pyx_result->composition));
  __pyx_v___pyx_result->composition = ((struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_2 == (double)-1) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->mass = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyString_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "str", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->name);
  __Pyx_DECREF(__pyx_v___pyx_result->name);
  __pyx_v___pyx_result->name = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_ChemicalShiftBase__set_state(ChemicalShiftBase __pyx_result, tuple __pyx_state):
 *     __pyx_result.composition = __pyx_state[0]; __pyx_result.mass = __pyx_state[1]; __pyx_result.name = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[3])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(2, 13, __pyx_L1_error)
  }
  __pyx_t_4 = PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_5 = ((__pyx_t_4 > 3) != 0);
  if (__pyx_t_5) {
  } else {
    __pyx_t_3 = __pyx_t_5;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_5 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_6 = (__pyx_t_5 != 0);
  __pyx_t_3 = __pyx_t_6;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_3) {

    /* "(tree fragment)":14
 *     __pyx_result.composition = __pyx_state[0]; __pyx_result.mass = __pyx_state[1]; __pyx_result.name = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[3])             # <<<<<<<<<<<<<<
 */
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_update); if (unlikely(!__pyx_t_8)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(2, 14, __pyx_L1_error)
    }
    __pyx_t_7 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_9 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_8))) {
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_8);
      if (likely(__pyx_t_9)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_8, function);
      }
    }
    __pyx_t_1 = (__pyx_t_9) ? __Pyx_PyObject_Call2Args(__pyx_t_8, __pyx_t_9, __pyx_t_7) : __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_t_7);
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_ChemicalShiftBase__set_state(ChemicalShiftBase __pyx_result, tuple __pyx_state):
 *     __pyx_result.composition = __pyx_state[0]; __pyx_result.mass = __pyx_state[1]; __pyx_result.name = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[3])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_ChemicalShiftBase__set_state(<ChemicalShiftBase> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_ChemicalShiftBase__set_state(ChemicalShiftBase __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.composition = __pyx_state[0]; __pyx_result.mass = __pyx_state[1]; __pyx_result.name = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment.__pyx_unpickle_ChemicalShiftBase__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_IonSeriesBase(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_7__pyx_unpickle_IonSeriesBase(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_14glycopeptidepy_2_c_9structure_8fragment_6__pyx_unpickle_IonSeriesBase[] = "__pyx_unpickle_IonSeriesBase(__pyx_type, long __pyx_checksum, __pyx_state)";
static PyMethodDef __pyx_mdef_14glycopeptidepy_2_c_9structure_8fragment_7__pyx_unpickle_IonSeriesBase = {"__pyx_unpickle_IonSeriesBase", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_7__pyx_unpickle_IonSeriesBase, METH_VARARGS|METH_KEYWORDS, __pyx_doc_14glycopeptidepy_2_c_9structure_8fragment_6__pyx_unpickle_IonSeriesBase};
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_7__pyx_unpickle_IonSeriesBase(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_IonSeriesBase (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_type)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_checksum)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_IonSeriesBase", 1, 3, 3, 1); __PYX_ERR(2, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_IonSeriesBase", 1, 3, 3, 2); __PYX_ERR(2, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_unpickle_IonSeriesBase") < 0)) __PYX_ERR(2, 1, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_IonSeriesBase", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(2, 1, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment.__pyx_unpickle_IonSeriesBase", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_6__pyx_unpickle_IonSeriesBase(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_6__pyx_unpickle_IonSeriesBase(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_IonSeriesBase", 0);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xf30379a, 0x05a3932, 0xc0e134f):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0xf30379a, 0x05a3932, 0xc0e134f) = (_hash, composition_shift, direction, includes_peptide, int_code, mass_shift, name))" % __pyx_checksum)
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__8, Py_NE)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xf30379a, 0x05a3932, 0xc0e134f):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0xf30379a, 0x05a3932, 0xc0e134f) = (_hash, composition_shift, direction, includes_peptide, int_code, mass_shift, name))" % __pyx_checksum)
 *     __pyx_result = IonSeriesBase.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError);
    __pyx_t_4 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, -1); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_4, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0xf30379a, 0x05a3932, 0xc0e134f):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0xf30379a, 0x05a3932, 0xc0e134f) = (_hash, composition_shift, direction, includes_peptide, int_code, mass_shift, name))" % __pyx_checksum)             # <<<<<<<<<<<<<<
 *     __pyx_result = IonSeriesBase.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_2, __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_INCREF(__pyx_v___pyx_PickleError);
    __pyx_t_1 = __pyx_v___pyx_PickleError; __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_4 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_6, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_5);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __PYX_ERR(2, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xf30379a, 0x05a3932, 0xc0e134f):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0xf30379a, 0x05a3932, 0xc0e134f) = (_hash, composition_shift, direction, includes_peptide, int_code, mass_shift, name))" % __pyx_checksum)
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0xf30379a, 0x05a3932, 0xc0e134f) = (_hash, composition_shift, direction, includes_peptide, int_code, mass_shift, name))" % __pyx_checksum)
 *     __pyx_result = IonSeriesBase.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_IonSeriesBase__set_state(<IonSeriesBase> __pyx_result, __pyx_state)
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase), __pyx_n_s_new); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_4 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_5, __pyx_v___pyx_type) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v___pyx_type);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0xf30379a, 0x05a3932, 0xc0e134f) = (_hash, composition_shift, direction, includes_peptide, int_code, mass_shift, name))" % __pyx_checksum)
 *     __pyx_result = IonSeriesBase.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_IonSeriesBase__set_state(<IonSeriesBase> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_3 = (__pyx_v___pyx_state != Py_None);
  __pyx_t_2 = (__pyx_t_3 != 0);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = IonSeriesBase.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_IonSeriesBase__set_state(<IonSeriesBase> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_IonSeriesBase__set_state(IonSeriesBase __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(2, 9, __pyx_L1_error)
    __pyx_t_4 = __pyx_f_14glycopeptidepy_2_c_9structure_8fragment___pyx_unpickle_IonSeriesBase__set_state(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0xf30379a, 0x05a3932, 0xc0e134f) = (_hash, composition_shift, direction, includes_peptide, int_code, mass_shift, name))" % __pyx_checksum)
 *     __pyx_result = IonSeriesBase.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_IonSeriesBase__set_state(<IonSeriesBase> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_IonSeriesBase__set_state(<IonSeriesBase> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_IonSeriesBase__set_state(IonSeriesBase __pyx_result, tuple __pyx_state):
 *     __pyx_result._hash = __pyx_state[0]; __pyx_result.composition_shift = __pyx_state[1]; __pyx_result.direction = __pyx_state[2]; __pyx_result.includes_peptide = __pyx_state[3]; __pyx_result.int_code = __pyx_state[4]; __pyx_result.mass_shift = __pyx_state[5]; __pyx_result.name = __pyx_state[6]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_IonSeriesBase(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment.__pyx_unpickle_IonSeriesBase", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_IonSeriesBase__set_state(<IonSeriesBase> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_IonSeriesBase__set_state(IonSeriesBase __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result._hash = __pyx_state[0]; __pyx_result.composition_shift = __pyx_state[1]; __pyx_result.direction = __pyx_state[2]; __pyx_result.includes_peptide = __pyx_state[3]; __pyx_result.int_code = __pyx_state[4]; __pyx_result.mass_shift = __pyx_state[5]; __pyx_result.name = __pyx_state[6]
 *     if len(__pyx_state) > 7 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_14glycopeptidepy_2_c_9structure_8fragment___pyx_unpickle_IonSeriesBase__set_state(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_hash_t __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  double __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  int __pyx_t_7;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_IonSeriesBase__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_IonSeriesBase__set_state(IonSeriesBase __pyx_result, tuple __pyx_state):
 *     __pyx_result._hash = __pyx_state[0]; __pyx_result.composition_shift = __pyx_state[1]; __pyx_result.direction = __pyx_state[2]; __pyx_result.includes_peptide = __pyx_state[3]; __pyx_result.int_code = __pyx_state[4]; __pyx_result.mass_shift = __pyx_state[5]; __pyx_result.name = __pyx_state[6]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 7 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[7])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_AsHash_t(__pyx_t_1); if (unlikely((__pyx_t_2 == (Py_hash_t)-1) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->_hash = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5glypy_11composition_12ccomposition_CComposition))))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->composition_shift);
  __Pyx_DECREF(((PyObject *)__pyx_v___pyx_result->composition_shift));
  __pyx_v___pyx_result->composition_shift = ((struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_3 == (int)-1) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->direction = __pyx_t_3;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->includes_peptide = __pyx_t_4;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 4, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_3 == (int)-1) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->int_code = __pyx_t_3;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 5, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_5 == (double)-1) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->mass_shift = __pyx_t_5;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 6, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyString_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "str", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->name);
  __Pyx_DECREF(__pyx_v___pyx_result->name);
  __pyx_v___pyx_result->name = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_IonSeriesBase__set_state(IonSeriesBase __pyx_result, tuple __pyx_state):
 *     __pyx_result._hash = __pyx_state[0]; __pyx_result.composition_shift = __pyx_state[1]; __pyx_result.direction = __pyx_state[2]; __pyx_result.includes_peptide = __pyx_state[3]; __pyx_result.int_code = __pyx_state[4]; __pyx_result.mass_shift = __pyx_state[5]; __pyx_result.name = __pyx_state[6]
 *     if len(__pyx_state) > 7 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[7])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(2, 13, __pyx_L1_error)
  }
  __pyx_t_6 = PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_6 == ((Py_ssize_t)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_7 = ((__pyx_t_6 > 7) != 0);
  if (__pyx_t_7) {
  } else {
    __pyx_t_4 = __pyx_t_7;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_7 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_8 = (__pyx_t_7 != 0);
  __pyx_t_4 = __pyx_t_8;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_4) {

    /* "(tree fragment)":14
 *     __pyx_result._hash = __pyx_state[0]; __pyx_result.composition_shift = __pyx_state[1]; __pyx_result.direction = __pyx_state[2]; __pyx_result.includes_peptide = __pyx_state[3]; __pyx_result.int_code = __pyx_state[4]; __pyx_result.mass_shift = __pyx_state[5]; __pyx_result.name = __pyx_state[6]
 *     if len(__pyx_state) > 7 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[7])             # <<<<<<<<<<<<<<
 */
    __pyx_t_9 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_9)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_update); if (unlikely(!__pyx_t_10)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(2, 14, __pyx_L1_error)
    }
    __pyx_t_9 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 7, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_9)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_11 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_10))) {
      __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_10);
      if (likely(__pyx_t_11)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
        __Pyx_INCREF(__pyx_t_11);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_10, function);
      }
    }
    __pyx_t_1 = (__pyx_t_11) ? __Pyx_PyObject_Call2Args(__pyx_t_10, __pyx_t_11, __pyx_t_9) : __Pyx_PyObject_CallOneArg(__pyx_t_10, __pyx_t_9);
    __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_IonSeriesBase__set_state(IonSeriesBase __pyx_result, tuple __pyx_state):
 *     __pyx_result._hash = __pyx_state[0]; __pyx_result.composition_shift = __pyx_state[1]; __pyx_result.direction = __pyx_state[2]; __pyx_result.includes_peptide = __pyx_state[3]; __pyx_result.int_code = __pyx_state[4]; __pyx_result.mass_shift = __pyx_state[5]; __pyx_result.name = __pyx_state[6]
 *     if len(__pyx_state) > 7 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[7])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_IonSeriesBase__set_state(<IonSeriesBase> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_IonSeriesBase__set_state(IonSeriesBase __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result._hash = __pyx_state[0]; __pyx_result.composition_shift = __pyx_state[1]; __pyx_result.direction = __pyx_state[2]; __pyx_result.includes_peptide = __pyx_state[3]; __pyx_result.int_code = __pyx_state[4]; __pyx_result.mass_shift = __pyx_state[5]; __pyx_result.name = __pyx_state[6]
 *     if len(__pyx_state) > 7 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment.__pyx_unpickle_IonSeriesBase__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_FragmentBase(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_9__pyx_unpickle_FragmentBase(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_14glycopeptidepy_2_c_9structure_8fragment_8__pyx_unpickle_FragmentBase[] = "__pyx_unpickle_FragmentBase(__pyx_type, long __pyx_checksum, __pyx_state)";
static PyMethodDef __pyx_mdef_14glycopeptidepy_2_c_9structure_8fragment_9__pyx_unpickle_FragmentBase = {"__pyx_unpickle_FragmentBase", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_9__pyx_unpickle_FragmentBase, METH_VARARGS|METH_KEYWORDS, __pyx_doc_14glycopeptidepy_2_c_9structure_8fragment_8__pyx_unpickle_FragmentBase};
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_9__pyx_unpickle_FragmentBase(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_FragmentBase (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_type)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_checksum)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_FragmentBase", 1, 3, 3, 1); __PYX_ERR(2, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_FragmentBase", 1, 3, 3, 2); __PYX_ERR(2, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_unpickle_FragmentBase") < 0)) __PYX_ERR(2, 1, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_FragmentBase", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(2, 1, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment.__pyx_unpickle_FragmentBase", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_8__pyx_unpickle_FragmentBase(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_8__pyx_unpickle_FragmentBase(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_FragmentBase", 0);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x033a8ac, 0xb3ef693, 0xc951cb2):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0x033a8ac, 0xb3ef693, 0xc951cb2) = (_chemical_shift, _hash, _name, mass))" % __pyx_checksum)
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__9, Py_NE)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x033a8ac, 0xb3ef693, 0xc951cb2):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0x033a8ac, 0xb3ef693, 0xc951cb2) = (_chemical_shift, _hash, _name, mass))" % __pyx_checksum)
 *     __pyx_result = FragmentBase.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError);
    __pyx_t_4 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, -1); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_4, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0x033a8ac, 0xb3ef693, 0xc951cb2):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0x033a8ac, 0xb3ef693, 0xc951cb2) = (_chemical_shift, _hash, _name, mass))" % __pyx_checksum)             # <<<<<<<<<<<<<<
 *     __pyx_result = FragmentBase.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_3, __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_INCREF(__pyx_v___pyx_PickleError);
    __pyx_t_1 = __pyx_v___pyx_PickleError; __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_4 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_6, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_5);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __PYX_ERR(2, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x033a8ac, 0xb3ef693, 0xc951cb2):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0x033a8ac, 0xb3ef693, 0xc951cb2) = (_chemical_shift, _hash, _name, mass))" % __pyx_checksum)
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0x033a8ac, 0xb3ef693, 0xc951cb2) = (_chemical_shift, _hash, _name, mass))" % __pyx_checksum)
 *     __pyx_result = FragmentBase.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_FragmentBase__set_state(<FragmentBase> __pyx_result, __pyx_state)
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase), __pyx_n_s_new); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_4 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_5, __pyx_v___pyx_type) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v___pyx_type);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0x033a8ac, 0xb3ef693, 0xc951cb2) = (_chemical_shift, _hash, _name, mass))" % __pyx_checksum)
 *     __pyx_result = FragmentBase.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_FragmentBase__set_state(<FragmentBase> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_3 = (__pyx_v___pyx_state != Py_None);
  __pyx_t_2 = (__pyx_t_3 != 0);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = FragmentBase.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_FragmentBase__set_state(<FragmentBase> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_FragmentBase__set_state(FragmentBase __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(2, 9, __pyx_L1_error)
    __pyx_t_4 = __pyx_f_14glycopeptidepy_2_c_9structure_8fragment___pyx_unpickle_FragmentBase__set_state(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0x033a8ac, 0xb3ef693, 0xc951cb2) = (_chemical_shift, _hash, _name, mass))" % __pyx_checksum)
 *     __pyx_result = FragmentBase.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_FragmentBase__set_state(<FragmentBase> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_FragmentBase__set_state(<FragmentBase> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_FragmentBase__set_state(FragmentBase __pyx_result, tuple __pyx_state):
 *     __pyx_result._chemical_shift = __pyx_state[0]; __pyx_result._hash = __pyx_state[1]; __pyx_result._name = __pyx_state[2]; __pyx_result.mass = __pyx_state[3]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_FragmentBase(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment.__pyx_unpickle_FragmentBase", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_FragmentBase__set_state(<FragmentBase> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_FragmentBase__set_state(FragmentBase __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result._chemical_shift = __pyx_state[0]; __pyx_result._hash = __pyx_state[1]; __pyx_result._name = __pyx_state[2]; __pyx_result.mass = __pyx_state[3]
 *     if len(__pyx_state) > 4 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_14glycopeptidepy_2_c_9structure_8fragment___pyx_unpickle_FragmentBase__set_state(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_hash_t __pyx_t_2;
  double __pyx_t_3;
  int __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_FragmentBase__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_FragmentBase__set_state(FragmentBase __pyx_result, tuple __pyx_state):
 *     __pyx_result._chemical_shift = __pyx_state[0]; __pyx_result._hash = __pyx_state[1]; __pyx_result._name = __pyx_state[2]; __pyx_result.mass = __pyx_state[3]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 4 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[4])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase))))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->_chemical_shift);
  __Pyx_DECREF(((PyObject *)__pyx_v___pyx_result->_chemical_shift));
  __pyx_v___pyx_result->_chemical_shift = ((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_AsHash_t(__pyx_t_1); if (unlikely((__pyx_t_2 == (Py_hash_t)-1) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->_hash = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyString_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "str", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->_name);
  __Pyx_DECREF(__pyx_v___pyx_result->_name);
  __pyx_v___pyx_result->_name = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_3 == (double)-1) && PyErr_Occurred())) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->mass = __pyx_t_3;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_FragmentBase__set_state(FragmentBase __pyx_result, tuple __pyx_state):
 *     __pyx_result._chemical_shift = __pyx_state[0]; __pyx_result._hash = __pyx_state[1]; __pyx_result._name = __pyx_state[2]; __pyx_result.mass = __pyx_state[3]
 *     if len(__pyx_state) > 4 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[4])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(2, 13, __pyx_L1_error)
  }
  __pyx_t_5 = PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_6 = ((__pyx_t_5 > 4) != 0);
  if (__pyx_t_6) {
  } else {
    __pyx_t_4 = __pyx_t_6;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_6 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_7 = (__pyx_t_6 != 0);
  __pyx_t_4 = __pyx_t_7;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_4) {

    /* "(tree fragment)":14
 *     __pyx_result._chemical_shift = __pyx_state[0]; __pyx_result._hash = __pyx_state[1]; __pyx_result._name = __pyx_state[2]; __pyx_result.mass = __pyx_state[3]
 *     if len(__pyx_state) > 4 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[4])             # <<<<<<<<<<<<<<
 */
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_8)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_update); if (unlikely(!__pyx_t_9)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(2, 14, __pyx_L1_error)
    }
    __pyx_t_8 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 4, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_8)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_10 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_9))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_9);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_9, function);
      }
    }
    __pyx_t_1 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_9, __pyx_t_10, __pyx_t_8) : __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_t_8);
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_FragmentBase__set_state(FragmentBase __pyx_result, tuple __pyx_state):
 *     __pyx_result._chemical_shift = __pyx_state[0]; __pyx_result._hash = __pyx_state[1]; __pyx_result._name = __pyx_state[2]; __pyx_result.mass = __pyx_state[3]
 *     if len(__pyx_state) > 4 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[4])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_FragmentBase__set_state(<FragmentBase> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_FragmentBase__set_state(FragmentBase __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result._chemical_shift = __pyx_state[0]; __pyx_result._hash = __pyx_state[1]; __pyx_result._name = __pyx_state[2]; __pyx_result.mass = __pyx_state[3]
 *     if len(__pyx_state) > 4 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment.__pyx_unpickle_FragmentBase__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle__NameTree(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_11__pyx_unpickle__NameTree(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_14glycopeptidepy_2_c_9structure_8fragment_10__pyx_unpickle__NameTree[] = "__pyx_unpickle__NameTree(__pyx_type, long __pyx_checksum, __pyx_state)";
static PyMethodDef __pyx_mdef_14glycopeptidepy_2_c_9structure_8fragment_11__pyx_unpickle__NameTree = {"__pyx_unpickle__NameTree", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_11__pyx_unpickle__NameTree, METH_VARARGS|METH_KEYWORDS, __pyx_doc_14glycopeptidepy_2_c_9structure_8fragment_10__pyx_unpickle__NameTree};
static PyObject *__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_11__pyx_unpickle__NameTree(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle__NameTree (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_type)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_checksum)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle__NameTree", 1, 3, 3, 1); __PYX_ERR(2, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle__NameTree", 1, 3, 3, 2); __PYX_ERR(2, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_unpickle__NameTree") < 0)) __PYX_ERR(2, 1, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(2, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle__NameTree", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(2, 1, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment.__pyx_unpickle__NameTree", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_10__pyx_unpickle__NameTree(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_14glycopeptidepy_2_c_9structure_8fragment_10__pyx_unpickle__NameTree(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle__NameTree", 0);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xd2a53ba, 0x1a09e23, 0x7fd03ce):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0xd2a53ba, 0x1a09e23, 0x7fd03ce) = (children, name, value))" % __pyx_checksum)
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__10, Py_NE)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xd2a53ba, 0x1a09e23, 0x7fd03ce):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0xd2a53ba, 0x1a09e23, 0x7fd03ce) = (children, name, value))" % __pyx_checksum)
 *     __pyx_result = _NameTree.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError);
    __pyx_t_4 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, -1); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_4, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0xd2a53ba, 0x1a09e23, 0x7fd03ce):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0xd2a53ba, 0x1a09e23, 0x7fd03ce) = (children, name, value))" % __pyx_checksum)             # <<<<<<<<<<<<<<
 *     __pyx_result = _NameTree.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_4, __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_INCREF(__pyx_v___pyx_PickleError);
    __pyx_t_1 = __pyx_v___pyx_PickleError; __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_4 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_6, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_5);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __PYX_ERR(2, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xd2a53ba, 0x1a09e23, 0x7fd03ce):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0xd2a53ba, 0x1a09e23, 0x7fd03ce) = (children, name, value))" % __pyx_checksum)
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0xd2a53ba, 0x1a09e23, 0x7fd03ce) = (children, name, value))" % __pyx_checksum)
 *     __pyx_result = _NameTree.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle__NameTree__set_state(<_NameTree> __pyx_result, __pyx_state)
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment__NameTree), __pyx_n_s_new); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_4 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_5, __pyx_v___pyx_type) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v___pyx_type);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0xd2a53ba, 0x1a09e23, 0x7fd03ce) = (children, name, value))" % __pyx_checksum)
 *     __pyx_result = _NameTree.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle__NameTree__set_state(<_NameTree> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_3 = (__pyx_v___pyx_state != Py_None);
  __pyx_t_2 = (__pyx_t_3 != 0);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = _NameTree.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle__NameTree__set_state(<_NameTree> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle__NameTree__set_state(_NameTree __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(2, 9, __pyx_L1_error)
    __pyx_t_4 = __pyx_f_14glycopeptidepy_2_c_9structure_8fragment___pyx_unpickle__NameTree__set_state(((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment__NameTree *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0xd2a53ba, 0x1a09e23, 0x7fd03ce) = (children, name, value))" % __pyx_checksum)
 *     __pyx_result = _NameTree.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle__NameTree__set_state(<_NameTree> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle__NameTree__set_state(<_NameTree> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle__NameTree__set_state(_NameTree __pyx_result, tuple __pyx_state):
 *     __pyx_result.children = __pyx_state[0]; __pyx_result.name = __pyx_state[1]; __pyx_result.value = __pyx_state[2]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle__NameTree(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment.__pyx_unpickle__NameTree", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle__NameTree__set_state(<_NameTree> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle__NameTree__set_state(_NameTree __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.children = __pyx_state[0]; __pyx_result.name = __pyx_state[1]; __pyx_result.value = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_14glycopeptidepy_2_c_9structure_8fragment___pyx_unpickle__NameTree__set_state(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment__NameTree *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle__NameTree__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle__NameTree__set_state(_NameTree __pyx_result, tuple __pyx_state):
 *     __pyx_result.children = __pyx_state[0]; __pyx_result.name = __pyx_state[1]; __pyx_result.value = __pyx_state[2]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[3])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyDict_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||((void)PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "dict", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->children);
  __Pyx_DECREF(__pyx_v___pyx_result->children);
  __pyx_v___pyx_result->children = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->name);
  __Pyx_DECREF(__pyx_v___pyx_result->name);
  __pyx_v___pyx_result->name = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(2, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->value);
  __Pyx_DECREF(__pyx_v___pyx_result->value);
  __pyx_v___pyx_result->value = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle__NameTree__set_state(_NameTree __pyx_result, tuple __pyx_state):
 *     __pyx_result.children = __pyx_state[0]; __pyx_result.name = __pyx_state[1]; __pyx_result.value = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[3])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(2, 13, __pyx_L1_error)
  }
  __pyx_t_3 = PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_4 = ((__pyx_t_3 > 3) != 0);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(2, 13, __pyx_L1_error)
  __pyx_t_5 = (__pyx_t_4 != 0);
  __pyx_t_2 = __pyx_t_5;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result.children = __pyx_state[0]; __pyx_result.name = __pyx_state[1]; __pyx_result.value = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[3])             # <<<<<<<<<<<<<<
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_update); if (unlikely(!__pyx_t_7)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(2, 14, __pyx_L1_error)
    }
    __pyx_t_6 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_8 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
      }
    }
    __pyx_t_1 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_8, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_6);
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle__NameTree__set_state(_NameTree __pyx_result, tuple __pyx_state):
 *     __pyx_result.children = __pyx_state[0]; __pyx_result.name = __pyx_state[1]; __pyx_result.value = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[3])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle__NameTree__set_state(<_NameTree> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle__NameTree__set_state(_NameTree __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.children = __pyx_state[0]; __pyx_result.name = __pyx_state[1]; __pyx_result.value = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("glycopeptidepy._c.structure.fragment.__pyx_unpickle__NameTree__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cfunc.to_py":65
 * @cname("__Pyx_CFunc_object____tuple___to_py")
 * cdef object __Pyx_CFunc_object____tuple___to_py(object (*f)(tuple) ):
 *     def wrap(tuple x):             # <<<<<<<<<<<<<<
 *         """wrap(x: tuple)"""
 *         return f(x)
 */

/* Python wrapper */
static PyObject *__pyx_pw_11cfunc_dot_to_py_35__Pyx_CFunc_object____tuple___to_py_1wrap(PyObject *__pyx_self, PyObject *__pyx_v_x); /*proto*/
static char __pyx_doc_11cfunc_dot_to_py_35__Pyx_CFunc_object____tuple___to_py_wrap[] = "wrap(x: tuple)";
static PyMethodDef __pyx_mdef_11cfunc_dot_to_py_35__Pyx_CFunc_object____tuple___to_py_1wrap = {"wrap", (PyCFunction)__pyx_pw_11cfunc_dot_to_py_35__Pyx_CFunc_object____tuple___to_py_1wrap, METH_O, __pyx_doc_11cfunc_dot_to_py_35__Pyx_CFunc_object____tuple___to_py_wrap};
static PyObject *__pyx_pw_11cfunc_dot_to_py_35__Pyx_CFunc_object____tuple___to_py_1wrap(PyObject *__pyx_self, PyObject *__pyx_v_x) {
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("wrap (wrapper)", 0);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_x), (&PyTuple_Type), 1, "x", 1))) __PYX_ERR(2, 65, __pyx_L1_error)
  __pyx_r = __pyx_pf_11cfunc_dot_to_py_35__Pyx_CFunc_object____tuple___to_py_wrap(__pyx_self, ((PyObject*)__pyx_v_x));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_11cfunc_dot_to_py_35__Pyx_CFunc_object____tuple___to_py_wrap(PyObject *__pyx_self, PyObject *__pyx_v_x) {
  struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_object____tuple___to_py *__pyx_cur_scope;
  struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_object____tuple___to_py *__pyx_outer_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("wrap", 0);
  __pyx_outer_scope = (struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_object____tuple___to_py *) __Pyx_CyFunction_GetClosure(__pyx_self);
  __pyx_cur_scope = __pyx_outer_scope;

  /* "cfunc.to_py":67
 *     def wrap(tuple x):
 *         """wrap(x: tuple)"""
 *         return f(x)             # <<<<<<<<<<<<<<
 *     return wrap
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_cur_scope->__pyx_v_f(__pyx_v_x); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 67, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "cfunc.to_py":65
 * @cname("__Pyx_CFunc_object____tuple___to_py")
 * cdef object __Pyx_CFunc_object____tuple___to_py(object (*f)(tuple) ):
 *     def wrap(tuple x):             # <<<<<<<<<<<<<<
 *         """wrap(x: tuple)"""
 *         return f(x)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cfunc.to_py.__Pyx_CFunc_object____tuple___to_py.wrap", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cfunc.to_py":64
 * 
 * @cname("__Pyx_CFunc_object____tuple___to_py")
 * cdef object __Pyx_CFunc_object____tuple___to_py(object (*f)(tuple) ):             # <<<<<<<<<<<<<<
 *     def wrap(tuple x):
 *         """wrap(x: tuple)"""
 */

static PyObject *__Pyx_CFunc_object____tuple___to_py(PyObject *(*__pyx_v_f)(PyObject *)) {
  struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_object____tuple___to_py *__pyx_cur_scope;
  PyObject *__pyx_v_wrap = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_CFunc_object____tuple___to_py", 0);
  __pyx_cur_scope = (struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_object____tuple___to_py *)__pyx_tp_new___pyx_scope_struct____Pyx_CFunc_object____tuple___to_py(__pyx_ptype___pyx_scope_struct____Pyx_CFunc_object____tuple___to_py, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_object____tuple___to_py *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(2, 64, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_f = __pyx_v_f;

  /* "cfunc.to_py":65
 * @cname("__Pyx_CFunc_object____tuple___to_py")
 * cdef object __Pyx_CFunc_object____tuple___to_py(object (*f)(tuple) ):
 *     def wrap(tuple x):             # <<<<<<<<<<<<<<
 *         """wrap(x: tuple)"""
 *         return f(x)
 */
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_11cfunc_dot_to_py_35__Pyx_CFunc_object____tuple___to_py_1wrap, 0, __pyx_n_s_Pyx_CFunc_object____tuple___to, ((PyObject*)__pyx_cur_scope), __pyx_n_s_cfunc_to_py, __pyx_d, ((PyObject *)__pyx_codeobj__12)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 65, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_wrap = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "cfunc.to_py":68
 *         """wrap(x: tuple)"""
 *         return f(x)
 *     return wrap             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_wrap);
  __pyx_r = __pyx_v_wrap;
  goto __pyx_L0;

  /* "cfunc.to_py":64
 * 
 * @cname("__Pyx_CFunc_object____tuple___to_py")
 * cdef object __Pyx_CFunc_object____tuple___to_py(object (*f)(tuple) ):             # <<<<<<<<<<<<<<
 *     def wrap(tuple x):
 *         """wrap(x: tuple)"""
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cfunc.to_py.__Pyx_CFunc_object____tuple___to_py", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_wrap);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase __pyx_vtable_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase;

static struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase *__pyx_freelist_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase[100];
static int __pyx_freecount_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase = 0;

static PyObject *__pyx_tp_new_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase *p;
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase)) & ((t->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)) == 0))) {
    o = (PyObject*)__pyx_freelist_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase[--__pyx_freecount_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase];
    memset(o, 0, sizeof(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
      o = (*t->tp_alloc)(t, 0);
    } else {
      o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
    }
    if (unlikely(!o)) return 0;
  }
  p = ((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase *)o);
  p->__pyx_vtab = __pyx_vtabptr_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase;
  p->name = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->composition = ((struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase(PyObject *o) {
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase *p = (struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->name);
  Py_CLEAR(p->composition);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase < 100) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase)) & ((Py_TYPE(o)->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)) == 0))) {
    __pyx_freelist_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase[__pyx_freecount_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase++] = ((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase *p = (struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase *)o;
  if (p->composition) {
    e = (*v)(((PyObject *)p->composition), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase *p = (struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase *)o;
  tmp = ((PyObject*)p->composition);
  p->composition = ((struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_14glycopeptidepy_2_c_9structure_8fragment_17ChemicalShiftBase_name(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_17ChemicalShiftBase_4name_1__get__(o);
}

static int __pyx_setprop_14glycopeptidepy_2_c_9structure_8fragment_17ChemicalShiftBase_name(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_17ChemicalShiftBase_4name_3__set__(o, v);
  }
  else {
    return __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_17ChemicalShiftBase_4name_5__del__(o);
  }
}

static PyObject *__pyx_getprop_14glycopeptidepy_2_c_9structure_8fragment_17ChemicalShiftBase_composition(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_17ChemicalShiftBase_11composition_1__get__(o);
}

static int __pyx_setprop_14glycopeptidepy_2_c_9structure_8fragment_17ChemicalShiftBase_composition(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_17ChemicalShiftBase_11composition_3__set__(o, v);
  }
  else {
    return __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_17ChemicalShiftBase_11composition_5__del__(o);
  }
}

static PyObject *__pyx_getprop_14glycopeptidepy_2_c_9structure_8fragment_17ChemicalShiftBase_mass(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_17ChemicalShiftBase_4mass_1__get__(o);
}

static int __pyx_setprop_14glycopeptidepy_2_c_9structure_8fragment_17ChemicalShiftBase_mass(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_17ChemicalShiftBase_4mass_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyMethodDef __pyx_methods_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase[] = {
  {"clone", (PyCFunction)__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_17ChemicalShiftBase_1clone, METH_NOARGS, __pyx_doc_14glycopeptidepy_2_c_9structure_8fragment_17ChemicalShiftBase_clone},
  {"is_loss", (PyCFunction)__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_17ChemicalShiftBase_7is_loss, METH_NOARGS, __pyx_doc_14glycopeptidepy_2_c_9structure_8fragment_17ChemicalShiftBase_6is_loss},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_17ChemicalShiftBase_9__reduce_cython__, METH_NOARGS, __pyx_doc_14glycopeptidepy_2_c_9structure_8fragment_17ChemicalShiftBase_8__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_17ChemicalShiftBase_11__setstate_cython__, METH_O, __pyx_doc_14glycopeptidepy_2_c_9structure_8fragment_17ChemicalShiftBase_10__setstate_cython__},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase[] = {
  {(char *)"name", __pyx_getprop_14glycopeptidepy_2_c_9structure_8fragment_17ChemicalShiftBase_name, __pyx_setprop_14glycopeptidepy_2_c_9structure_8fragment_17ChemicalShiftBase_name, (char *)"name: str", 0},
  {(char *)"composition", __pyx_getprop_14glycopeptidepy_2_c_9structure_8fragment_17ChemicalShiftBase_composition, __pyx_setprop_14glycopeptidepy_2_c_9structure_8fragment_17ChemicalShiftBase_composition, (char *)"composition: glypy.composition.ccomposition.CComposition", 0},
  {(char *)"mass", __pyx_getprop_14glycopeptidepy_2_c_9structure_8fragment_17ChemicalShiftBase_mass, __pyx_setprop_14glycopeptidepy_2_c_9structure_8fragment_17ChemicalShiftBase_mass, (char *)"mass: 'double'", 0},
  {0, 0, 0, 0, 0}
};

static PyTypeObject __pyx_type_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase = {
  PyVarObject_HEAD_INIT(0, 0)
  "glycopeptidepy._c.structure.fragment.ChemicalShiftBase", /*tp_name*/
  sizeof(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_17ChemicalShiftBase_5__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_17ChemicalShiftBase_3__str__, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase, /*tp_traverse*/
  __pyx_tp_clear_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};

static PyObject *__pyx_tp_new_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *)o);
  p->name = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->composition_shift = ((struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase(PyObject *o) {
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *p = (struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->name);
  Py_CLEAR(p->composition_shift);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *p = (struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *)o;
  if (p->composition_shift) {
    e = (*v)(((PyObject *)p->composition_shift), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *p = (struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *)o;
  tmp = ((PyObject*)p->composition_shift);
  p->composition_shift = ((struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_tp_richcompare_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase(PyObject *o1, PyObject *o2, int op) {
  switch (op) {
    case Py_EQ: {
      return __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_1__eq__(o1, o2);
    }
    case Py_NE: {
      return __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_3__ne__(o1, o2);
    }
    default: {
      return __Pyx_NewRef(Py_NotImplemented);
    }
  }
}

static PyObject *__pyx_getprop_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_name(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_4name_1__get__(o);
}

static int __pyx_setprop_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_name(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_4name_3__set__(o, v);
  }
  else {
    return __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_4name_5__del__(o);
  }
}

static PyObject *__pyx_getprop_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_direction(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_9direction_1__get__(o);
}

static int __pyx_setprop_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_direction(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_9direction_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_includes_peptide(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_16includes_peptide_1__get__(o);
}

static int __pyx_setprop_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_includes_peptide(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_16includes_peptide_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_mass_shift(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_10mass_shift_1__get__(o);
}

static int __pyx_setprop_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_mass_shift(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_10mass_shift_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_composition_shift(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_17composition_shift_1__get__(o);
}

static int __pyx_setprop_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_composition_shift(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_17composition_shift_3__set__(o, v);
  }
  else {
    return __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_17composition_shift_5__del__(o);
  }
}

static PyObject *__pyx_getprop_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase__hash(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_5_hash_1__get__(o);
}

static int __pyx_setprop_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase__hash(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_5_hash_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_int_code(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_8int_code_1__get__(o);
}

static int __pyx_setprop_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_int_code(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_8int_code_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyMethodDef __pyx_methods_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase[] = {
  {"__reduce_cython__", (PyCFunction)__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_7__reduce_cython__, METH_NOARGS, __pyx_doc_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_6__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_9__setstate_cython__, METH_O, __pyx_doc_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_8__setstate_cython__},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase[] = {
  {(char *)"name", __pyx_getprop_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_name, __pyx_setprop_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_name, (char *)"name: str", 0},
  {(char *)"direction", __pyx_getprop_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_direction, __pyx_setprop_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_direction, (char *)"direction: 'int'", 0},
  {(char *)"includes_peptide", __pyx_getprop_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_includes_peptide, __pyx_setprop_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_includes_peptide, (char *)"includes_peptide: 'bool'", 0},
  {(char *)"mass_shift", __pyx_getprop_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_mass_shift, __pyx_setprop_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_mass_shift, (char *)"mass_shift: 'double'", 0},
  {(char *)"composition_shift", __pyx_getprop_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_composition_shift, __pyx_setprop_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_composition_shift, (char *)"composition_shift: glypy.composition.ccomposition.CComposition", 0},
  {(char *)"_hash", __pyx_getprop_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase__hash, __pyx_setprop_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase__hash, (char *)"_hash: 'Py_hash_t'", 0},
  {(char *)"int_code", __pyx_getprop_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_int_code, __pyx_setprop_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_int_code, (char *)"int_code: 'int'", 0},
  {0, 0, 0, 0, 0}
};

static PyTypeObject __pyx_type_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase = {
  PyVarObject_HEAD_INIT(0, 0)
  "glycopeptidepy._c.structure.fragment.IonSeriesBase", /*tp_name*/
  sizeof(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_13IonSeriesBase_5__hash__, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase, /*tp_traverse*/
  __pyx_tp_clear_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase, /*tp_clear*/
  __pyx_tp_richcompare_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};
static struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase __pyx_vtable_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase;

static struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *__pyx_freelist_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase[1000000];
static int __pyx_freecount_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase = 0;

static PyObject *__pyx_tp_new_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *p;
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase)) & ((t->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)) == 0))) {
    o = (PyObject*)__pyx_freelist_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase[--__pyx_freecount_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase];
    memset(o, 0, sizeof(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
      o = (*t->tp_alloc)(t, 0);
    } else {
      o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
    }
    if (unlikely(!o)) return 0;
  }
  p = ((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *)o);
  p->__pyx_vtab = __pyx_vtabptr_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase;
  p->_name = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->_chemical_shift = ((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase *)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase(PyObject *o) {
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *p = (struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->_name);
  Py_CLEAR(p->_chemical_shift);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase < 1000000) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase)) & ((Py_TYPE(o)->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)) == 0))) {
    __pyx_freelist_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase[__pyx_freecount_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase++] = ((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *p = (struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *)o;
  if (p->_chemical_shift) {
    e = (*v)(((PyObject *)p->_chemical_shift), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *p = (struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *)o;
  tmp = ((PyObject*)p->_chemical_shift);
  p->_chemical_shift = ((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_tp_richcompare_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase(PyObject *o1, PyObject *o2, int op) {
  switch (op) {
    case Py_EQ: {
      return __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_5__eq__(o1, o2);
    }
    case Py_NE: {
      return __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_7__ne__(o1, o2);
    }
    default: {
      return __Pyx_NewRef(Py_NotImplemented);
    }
  }
}

static PyObject *__pyx_getprop_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_series(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_6series_1__get__(o);
}

static PyObject *__pyx_getprop_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_name(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_4name_1__get__(o);
}

static int __pyx_setprop_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_name(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_4name_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase__name(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_5_name_1__get__(o);
}

static int __pyx_setprop_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase__name(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_5_name_3__set__(o, v);
  }
  else {
    return __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_5_name_5__del__(o);
  }
}

static PyObject *__pyx_getprop_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase__hash(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_5_hash_1__get__(o);
}

static int __pyx_setprop_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase__hash(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_5_hash_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_mass(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_4mass_1__get__(o);
}

static int __pyx_setprop_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_mass(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_4mass_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase__chemical_shift(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_15_chemical_shift_1__get__(o);
}

static int __pyx_setprop_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase__chemical_shift(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_15_chemical_shift_3__set__(o, v);
  }
  else {
    return __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_15_chemical_shift_5__del__(o);
  }
}

static PyMethodDef __pyx_methods_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase[] = {
  {"get_series", (PyCFunction)__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_1get_series, METH_NOARGS, __pyx_doc_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_get_series},
  {"clone", (PyCFunction)__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_9clone, METH_NOARGS, __pyx_doc_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_8clone},
  {"get_chemical_shift", (PyCFunction)__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_11get_chemical_shift, METH_NOARGS, __pyx_doc_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_10get_chemical_shift},
  {"set_chemical_shift", (PyCFunction)__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_13set_chemical_shift, METH_O, __pyx_doc_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_12set_chemical_shift},
  {"get_fragment_name", (PyCFunction)__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_15get_fragment_name, METH_NOARGS, __pyx_doc_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_14get_fragment_name},
  {"base_name", (PyCFunction)__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_17base_name, METH_NOARGS, __pyx_doc_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_16base_name},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_19__reduce_cython__, METH_NOARGS, __pyx_doc_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_18__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_21__setstate_cython__, METH_O, __pyx_doc_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_20__setstate_cython__},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase[] = {
  {(char *)"series", __pyx_getprop_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_series, 0, (char *)"The :class:`IonSeries` this fragment is drawn from.\n\n        Returns\n        -------\n        :class:`IonSeries`\n        ", 0},
  {(char *)"name", __pyx_getprop_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_name, __pyx_setprop_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_name, (char *)0, 0},
  {(char *)"_name", __pyx_getprop_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase__name, __pyx_setprop_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase__name, (char *)"_name: str", 0},
  {(char *)"_hash", __pyx_getprop_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase__hash, __pyx_setprop_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase__hash, (char *)"_hash: 'Py_hash_t'", 0},
  {(char *)"mass", __pyx_getprop_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_mass, __pyx_setprop_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_mass, (char *)"mass: 'double'", 0},
  {(char *)"_chemical_shift", __pyx_getprop_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase__chemical_shift, __pyx_setprop_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase__chemical_shift, (char *)"_chemical_shift: glycopeptidepy._c.structure.fragment.ChemicalShiftBase", 0},
  {0, 0, 0, 0, 0}
};

static PyTypeObject __pyx_type_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase = {
  PyVarObject_HEAD_INIT(0, 0)
  "glycopeptidepy._c.structure.fragment.FragmentBase", /*tp_name*/
  sizeof(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_3__hash__, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  "Base class for all Fragment types. Defines basic\n    name generation and neutral loss handling functions.\n\n    Attributes\n    ----------\n    chemical_shift : ChemicalShift\n        The ChemicalShift associated with this fragment, or None.\n        If a ChemicalShift, its composition and mass are subtracted\n        from this object's composition and mass attributes.\n    name: str\n        The human readable description of this fragment\n    series: IonSeries\n        The ion ladder this fragment is derived from\n\n    ", /*tp_doc*/
  __pyx_tp_traverse_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase, /*tp_traverse*/
  __pyx_tp_clear_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase, /*tp_clear*/
  __pyx_tp_richcompare_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};
static struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_8fragment_SimpleFragment __pyx_vtable_14glycopeptidepy_2_c_9structure_8fragment_SimpleFragment;

static PyObject *__pyx_tp_new_14glycopeptidepy_2_c_9structure_8fragment_SimpleFragment(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_SimpleFragment *p;
  PyObject *o = __pyx_tp_new_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_SimpleFragment *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase*)__pyx_vtabptr_14glycopeptidepy_2_c_9structure_8fragment_SimpleFragment;
  p->kind = ((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *)Py_None); Py_INCREF(Py_None);
  p->composition = ((struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_14glycopeptidepy_2_c_9structure_8fragment_SimpleFragment(PyObject *o) {
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_SimpleFragment *p = (struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_SimpleFragment *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->kind);
  Py_CLEAR(p->composition);
  PyObject_GC_Track(o);
  __pyx_tp_dealloc_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase(o);
}

static int __pyx_tp_traverse_14glycopeptidepy_2_c_9structure_8fragment_SimpleFragment(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_SimpleFragment *p = (struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_SimpleFragment *)o;
  e = __pyx_tp_traverse_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase(o, v, a); if (e) return e;
  if (p->kind) {
    e = (*v)(((PyObject *)p->kind), a); if (e) return e;
  }
  if (p->composition) {
    e = (*v)(((PyObject *)p->composition), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_14glycopeptidepy_2_c_9structure_8fragment_SimpleFragment(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_SimpleFragment *p = (struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_SimpleFragment *)o;
  __pyx_tp_clear_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase(o);
  tmp = ((PyObject*)p->kind);
  p->kind = ((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->composition);
  p->composition = ((struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_14glycopeptidepy_2_c_9structure_8fragment_14SimpleFragment_kind(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_14SimpleFragment_4kind_1__get__(o);
}

static int __pyx_setprop_14glycopeptidepy_2_c_9structure_8fragment_14SimpleFragment_kind(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_14SimpleFragment_4kind_3__set__(o, v);
  }
  else {
    return __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_14SimpleFragment_4kind_5__del__(o);
  }
}

static PyObject *__pyx_getprop_14glycopeptidepy_2_c_9structure_8fragment_14SimpleFragment_is_glycosylated(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_14SimpleFragment_15is_glycosylated_1__get__(o);
}

static int __pyx_setprop_14glycopeptidepy_2_c_9structure_8fragment_14SimpleFragment_is_glycosylated(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_14SimpleFragment_15is_glycosylated_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_14glycopeptidepy_2_c_9structure_8fragment_14SimpleFragment_composition(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_14SimpleFragment_11composition_1__get__(o);
}

static int __pyx_setprop_14glycopeptidepy_2_c_9structure_8fragment_14SimpleFragment_composition(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_14SimpleFragment_11composition_3__set__(o, v);
  }
  else {
    return __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_14SimpleFragment_11composition_5__del__(o);
  }
}

static PyMethodDef __pyx_methods_14glycopeptidepy_2_c_9structure_8fragment_SimpleFragment[] = {
  {"clone", (PyCFunction)__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_14SimpleFragment_3clone, METH_NOARGS, __pyx_doc_14glycopeptidepy_2_c_9structure_8fragment_14SimpleFragment_2clone},
  {"__reduce__", (PyCFunction)__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_14SimpleFragment_5__reduce__, METH_NOARGS, __pyx_doc_14glycopeptidepy_2_c_9structure_8fragment_14SimpleFragment_4__reduce__},
  {"get_series", (PyCFunction)__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_14SimpleFragment_9get_series, METH_NOARGS, __pyx_doc_14glycopeptidepy_2_c_9structure_8fragment_14SimpleFragment_8get_series},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_14glycopeptidepy_2_c_9structure_8fragment_SimpleFragment[] = {
  {(char *)"kind", __pyx_getprop_14glycopeptidepy_2_c_9structure_8fragment_14SimpleFragment_kind, __pyx_setprop_14glycopeptidepy_2_c_9structure_8fragment_14SimpleFragment_kind, (char *)"kind: glycopeptidepy._c.structure.fragment.IonSeriesBase", 0},
  {(char *)"is_glycosylated", __pyx_getprop_14glycopeptidepy_2_c_9structure_8fragment_14SimpleFragment_is_glycosylated, __pyx_setprop_14glycopeptidepy_2_c_9structure_8fragment_14SimpleFragment_is_glycosylated, (char *)"is_glycosylated: 'bool'", 0},
  {(char *)"composition", __pyx_getprop_14glycopeptidepy_2_c_9structure_8fragment_14SimpleFragment_composition, __pyx_setprop_14glycopeptidepy_2_c_9structure_8fragment_14SimpleFragment_composition, (char *)"composition: glypy.composition.ccomposition.CComposition", 0},
  {0, 0, 0, 0, 0}
};

static PyTypeObject __pyx_type_14glycopeptidepy_2_c_9structure_8fragment_SimpleFragment = {
  PyVarObject_HEAD_INIT(0, 0)
  "glycopeptidepy._c.structure.fragment.SimpleFragment", /*tp_name*/
  sizeof(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_SimpleFragment), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_14glycopeptidepy_2_c_9structure_8fragment_SimpleFragment, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_14SimpleFragment_7__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  "SimpleFragment(name, mass, kind, composition, chemical_shift=None, is_glycosylated=False)", /*tp_doc*/
  __pyx_tp_traverse_14glycopeptidepy_2_c_9structure_8fragment_SimpleFragment, /*tp_traverse*/
  __pyx_tp_clear_14glycopeptidepy_2_c_9structure_8fragment_SimpleFragment, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_14glycopeptidepy_2_c_9structure_8fragment_SimpleFragment, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_14glycopeptidepy_2_c_9structure_8fragment_SimpleFragment, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_14SimpleFragment_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_14glycopeptidepy_2_c_9structure_8fragment_SimpleFragment, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};
static struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_8fragment_StubFragment __pyx_vtable_14glycopeptidepy_2_c_9structure_8fragment_StubFragment;

static PyObject *__pyx_tp_new_14glycopeptidepy_2_c_9structure_8fragment_StubFragment(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_StubFragment *p;
  PyObject *o = __pyx_tp_new_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_StubFragment *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase*)__pyx_vtabptr_14glycopeptidepy_2_c_9structure_8fragment_StubFragment;
  p->kind = ((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *)Py_None); Py_INCREF(Py_None);
  p->composition = ((struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *)Py_None); Py_INCREF(Py_None);
  p->glycosylation = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_14glycopeptidepy_2_c_9structure_8fragment_StubFragment(PyObject *o) {
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_StubFragment *p = (struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_StubFragment *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->kind);
  Py_CLEAR(p->composition);
  Py_CLEAR(p->glycosylation);
  PyObject_GC_Track(o);
  __pyx_tp_dealloc_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase(o);
}

static int __pyx_tp_traverse_14glycopeptidepy_2_c_9structure_8fragment_StubFragment(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_StubFragment *p = (struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_StubFragment *)o;
  e = __pyx_tp_traverse_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase(o, v, a); if (e) return e;
  if (p->kind) {
    e = (*v)(((PyObject *)p->kind), a); if (e) return e;
  }
  if (p->composition) {
    e = (*v)(((PyObject *)p->composition), a); if (e) return e;
  }
  if (p->glycosylation) {
    e = (*v)(p->glycosylation, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_14glycopeptidepy_2_c_9structure_8fragment_StubFragment(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_StubFragment *p = (struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_StubFragment *)o;
  __pyx_tp_clear_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase(o);
  tmp = ((PyObject*)p->kind);
  p->kind = ((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->composition);
  p->composition = ((struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->glycosylation);
  p->glycosylation = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_glycosylation_size(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_18glycosylation_size_1__get__(o);
}

static PyObject *__pyx_getprop_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_kind(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_4kind_1__get__(o);
}

static int __pyx_setprop_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_kind(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_4kind_3__set__(o, v);
  }
  else {
    return __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_4kind_5__del__(o);
  }
}

static PyObject *__pyx_getprop_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_is_glycosylated(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_15is_glycosylated_1__get__(o);
}

static int __pyx_setprop_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_is_glycosylated(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_15is_glycosylated_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_composition(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_11composition_1__get__(o);
}

static int __pyx_setprop_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_composition(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_11composition_3__set__(o, v);
  }
  else {
    return __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_11composition_5__del__(o);
  }
}

static PyObject *__pyx_getprop_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_glycosylation(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_13glycosylation_1__get__(o);
}

static int __pyx_setprop_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_glycosylation(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_13glycosylation_3__set__(o, v);
  }
  else {
    return __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_13glycosylation_5__del__(o);
  }
}

static PyObject *__pyx_getprop_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_is_extended(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_11is_extended_1__get__(o);
}

static int __pyx_setprop_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_is_extended(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_11is_extended_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyMethodDef __pyx_methods_14glycopeptidepy_2_c_9structure_8fragment_StubFragment[] = {
  {"clone", (PyCFunction)__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_3clone, METH_NOARGS, __pyx_doc_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_2clone},
  {"base_name", (PyCFunction)__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_5base_name, METH_NOARGS, __pyx_doc_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_4base_name},
  {"build_name_from_composition", (PyCFunction)__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_7build_name_from_composition, METH_O, __pyx_doc_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_6build_name_from_composition},
  {"__reduce__", (PyCFunction)__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_9__reduce__, METH_NOARGS, __pyx_doc_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_8__reduce__},
  {"get_series", (PyCFunction)__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_13get_series, METH_NOARGS, __pyx_doc_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_12get_series},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_14glycopeptidepy_2_c_9structure_8fragment_StubFragment[] = {
  {(char *)"glycosylation_size", __pyx_getprop_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_glycosylation_size, 0, (char *)0, 0},
  {(char *)"kind", __pyx_getprop_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_kind, __pyx_setprop_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_kind, (char *)"kind: glycopeptidepy._c.structure.fragment.IonSeriesBase", 0},
  {(char *)"is_glycosylated", __pyx_getprop_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_is_glycosylated, __pyx_setprop_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_is_glycosylated, (char *)"is_glycosylated: 'bool'", 0},
  {(char *)"composition", __pyx_getprop_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_composition, __pyx_setprop_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_composition, (char *)"composition: glypy.composition.ccomposition.CComposition", 0},
  {(char *)"glycosylation", __pyx_getprop_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_glycosylation, __pyx_setprop_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_glycosylation, (char *)"glycosylation: object", 0},
  {(char *)"is_extended", __pyx_getprop_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_is_extended, __pyx_setprop_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_is_extended, (char *)"is_extended: 'bool'", 0},
  {0, 0, 0, 0, 0}
};

static PyTypeObject __pyx_type_14glycopeptidepy_2_c_9structure_8fragment_StubFragment = {
  PyVarObject_HEAD_INIT(0, 0)
  "glycopeptidepy._c.structure.fragment.StubFragment", /*tp_name*/
  sizeof(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_StubFragment), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_14glycopeptidepy_2_c_9structure_8fragment_StubFragment, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_11__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  "StubFragment(name, mass, kind, composition, chemical_shift=None, is_glycosylated=False, glycosylation=None, is_extended=False)", /*tp_doc*/
  __pyx_tp_traverse_14glycopeptidepy_2_c_9structure_8fragment_StubFragment, /*tp_traverse*/
  __pyx_tp_clear_14glycopeptidepy_2_c_9structure_8fragment_StubFragment, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_14glycopeptidepy_2_c_9structure_8fragment_StubFragment, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_14glycopeptidepy_2_c_9structure_8fragment_StubFragment, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_14glycopeptidepy_2_c_9structure_8fragment_StubFragment, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};
static struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment __pyx_vtable_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment;

static PyObject *__pyx_tp_new_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *p;
  PyObject *o = __pyx_tp_new_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase*)__pyx_vtabptr_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment;
  p->kind = ((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *)Py_None); Py_INCREF(Py_None);
  p->modification_dict = ((struct __pyx_obj_14glycopeptidepy_2_c_11count_table_CountTable *)Py_None); Py_INCREF(Py_None);
  p->flanking_amino_acids = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->glycosylation = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->composition = ((struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment(PyObject *o) {
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *p = (struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->kind);
  Py_CLEAR(p->modification_dict);
  Py_CLEAR(p->flanking_amino_acids);
  Py_CLEAR(p->glycosylation);
  Py_CLEAR(p->composition);
  PyObject_GC_Track(o);
  __pyx_tp_dealloc_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase(o);
}

static int __pyx_tp_traverse_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *p = (struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *)o;
  e = __pyx_tp_traverse_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase(o, v, a); if (e) return e;
  if (p->kind) {
    e = (*v)(((PyObject *)p->kind), a); if (e) return e;
  }
  if (p->modification_dict) {
    e = (*v)(((PyObject *)p->modification_dict), a); if (e) return e;
  }
  if (p->flanking_amino_acids) {
    e = (*v)(p->flanking_amino_acids, a); if (e) return e;
  }
  if (p->glycosylation) {
    e = (*v)(p->glycosylation, a); if (e) return e;
  }
  if (p->composition) {
    e = (*v)(((PyObject *)p->composition), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *p = (struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *)o;
  __pyx_tp_clear_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase(o);
  tmp = ((PyObject*)p->kind);
  p->kind = ((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->modification_dict);
  p->modification_dict = ((struct __pyx_obj_14glycopeptidepy_2_c_11count_table_CountTable *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->flanking_amino_acids);
  p->flanking_amino_acids = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->glycosylation);
  p->glycosylation = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->composition);
  p->composition = ((struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_is_glycosylated(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_15is_glycosylated_1__get__(o);
}

static PyObject *__pyx_getprop_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_glycosylation_size(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_18glycosylation_size_1__get__(o);
}

static PyObject *__pyx_getprop_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_kind(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_4kind_1__get__(o);
}

static int __pyx_setprop_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_kind(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_4kind_3__set__(o, v);
  }
  else {
    return __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_4kind_5__del__(o);
  }
}

static PyObject *__pyx_getprop_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_position(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_8position_1__get__(o);
}

static int __pyx_setprop_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_position(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_8position_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_modification_dict(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_17modification_dict_1__get__(o);
}

static int __pyx_setprop_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_modification_dict(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_17modification_dict_3__set__(o, v);
  }
  else {
    return __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_17modification_dict_5__del__(o);
  }
}

static PyObject *__pyx_getprop_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_bare_mass(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_9bare_mass_1__get__(o);
}

static int __pyx_setprop_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_bare_mass(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_9bare_mass_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_flanking_amino_acids(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_20flanking_amino_acids_1__get__(o);
}

static int __pyx_setprop_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_flanking_amino_acids(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_20flanking_amino_acids_3__set__(o, v);
  }
  else {
    return __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_20flanking_amino_acids_5__del__(o);
  }
}

static PyObject *__pyx_getprop_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_glycosylation(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_13glycosylation_1__get__(o);
}

static int __pyx_setprop_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_glycosylation(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_13glycosylation_3__set__(o, v);
  }
  else {
    return __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_13glycosylation_5__del__(o);
  }
}

static PyObject *__pyx_getprop_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_composition(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_11composition_1__get__(o);
}

static int __pyx_setprop_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_composition(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_11composition_3__set__(o, v);
  }
  else {
    return __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_11composition_5__del__(o);
  }
}

static PyMethodDef __pyx_methods_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment[] = {
  {"get_series", (PyCFunction)__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_3get_series, METH_NOARGS, __pyx_doc_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_2get_series},
  {"clone", (PyCFunction)__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_5clone, METH_NOARGS, __pyx_doc_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_4clone},
  {"total_composition", (PyCFunction)__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_7total_composition, METH_NOARGS, __pyx_doc_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_6total_composition},
  {"__reduce__", (PyCFunction)__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_9__reduce__, METH_NOARGS, __pyx_doc_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_8__reduce__},
  {"base_name", (PyCFunction)__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_11base_name, METH_NOARGS, __pyx_doc_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_10base_name},
  {"get_fragment_name", (PyCFunction)__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_13get_fragment_name, METH_NOARGS, __pyx_doc_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_12get_fragment_name},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment[] = {
  {(char *)"is_glycosylated", __pyx_getprop_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_is_glycosylated, 0, (char *)0, 0},
  {(char *)"glycosylation_size", __pyx_getprop_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_glycosylation_size, 0, (char *)0, 0},
  {(char *)"kind", __pyx_getprop_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_kind, __pyx_setprop_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_kind, (char *)"kind: glycopeptidepy._c.structure.fragment.IonSeriesBase", 0},
  {(char *)"position", __pyx_getprop_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_position, __pyx_setprop_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_position, (char *)"position: 'int'", 0},
  {(char *)"modification_dict", __pyx_getprop_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_modification_dict, __pyx_setprop_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_modification_dict, (char *)"modification_dict: glycopeptidepy._c.count_table.CountTable", 0},
  {(char *)"bare_mass", __pyx_getprop_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_bare_mass, __pyx_setprop_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_bare_mass, (char *)"bare_mass: 'double'", 0},
  {(char *)"flanking_amino_acids", __pyx_getprop_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_flanking_amino_acids, __pyx_setprop_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_flanking_amino_acids, (char *)"flanking_amino_acids: list", 0},
  {(char *)"glycosylation", __pyx_getprop_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_glycosylation, __pyx_setprop_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_glycosylation, (char *)"glycosylation: dict", 0},
  {(char *)"composition", __pyx_getprop_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_composition, __pyx_setprop_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_composition, (char *)"composition: glypy.composition.ccomposition.CComposition", 0},
  {0, 0, 0, 0, 0}
};

static PyTypeObject __pyx_type_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment = {
  PyVarObject_HEAD_INIT(0, 0)
  "glycopeptidepy._c.structure.fragment.PeptideFragment", /*tp_name*/
  sizeof(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_15__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  "PeptideFragment(kind, position, modification_dict, mass, flanking_amino_acids=None, glycosylation=None, chemical_shift=None, composition=None)\nRepresents a peptide backbone fragment, such as a peptide b or y ion.\n\n    Attributes\n    ----------\n    series: :class:`IonSeries`\n        The ion series this fragment is from\n    bare_mass: float\n        The unmodified mass of this fragment\n    mass: float\n        The modified, correct mass of this fragment\n    position: int\n        The peptide backbone position index this fragment came from, relative to the\n        starting terminal\n    modification_dict: :class:`~.ModificationIndex`\n        A count of the modifications on this fragment\n    flanking_amino_acids: list\n        The adjacent amino acids of the amide bond that broke to form this fragment\n    glycosylation: dict\n        A position to glycan mapping for glycan structures\n    chemical_shift: :class:`ChemicalShift`\n        An associated chemical shift\n    composition: :class:`~.Composition`\n        The elemental composition of this fragment\n    ", /*tp_doc*/
  __pyx_tp_traverse_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment, /*tp_traverse*/
  __pyx_tp_clear_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};
static struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_8fragment__NameTree __pyx_vtable_14glycopeptidepy_2_c_9structure_8fragment__NameTree;

static PyObject *__pyx_tp_new_14glycopeptidepy_2_c_9structure_8fragment__NameTree(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment__NameTree *p;
  PyObject *o;
  o = (*t->tp_alloc)(t, 0);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment__NameTree *)o);
  p->__pyx_vtab = __pyx_vtabptr_14glycopeptidepy_2_c_9structure_8fragment__NameTree;
  p->name = Py_None; Py_INCREF(Py_None);
  p->value = Py_None; Py_INCREF(Py_None);
  p->children = ((PyObject*)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_14glycopeptidepy_2_c_9structure_8fragment__NameTree(PyObject *o) {
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment__NameTree *p = (struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment__NameTree *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->name);
  Py_CLEAR(p->value);
  Py_CLEAR(p->children);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_14glycopeptidepy_2_c_9structure_8fragment__NameTree(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment__NameTree *p = (struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment__NameTree *)o;
  if (p->name) {
    e = (*v)(p->name, a); if (e) return e;
  }
  if (p->value) {
    e = (*v)(p->value, a); if (e) return e;
  }
  if (p->children) {
    e = (*v)(p->children, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_14glycopeptidepy_2_c_9structure_8fragment__NameTree(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment__NameTree *p = (struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment__NameTree *)o;
  tmp = ((PyObject*)p->name);
  p->name = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->value);
  p->value = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->children);
  p->children = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}
static PyObject *__pyx_sq_item_14glycopeptidepy_2_c_9structure_8fragment__NameTree(PyObject *o, Py_ssize_t i) {
  PyObject *r;
  PyObject *x = PyInt_FromSsize_t(i); if(!x) return 0;
  r = Py_TYPE(o)->tp_as_mapping->mp_subscript(o, x);
  Py_DECREF(x);
  return r;
}

static PyObject *__pyx_getprop_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_name(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_4name_1__get__(o);
}

static int __pyx_setprop_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_name(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_4name_3__set__(o, v);
  }
  else {
    return __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_4name_5__del__(o);
  }
}

static PyObject *__pyx_getprop_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_value(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_5value_1__get__(o);
}

static int __pyx_setprop_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_value(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_5value_3__set__(o, v);
  }
  else {
    return __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_5value_5__del__(o);
  }
}

static PyObject *__pyx_getprop_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_children(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_8children_1__get__(o);
}

static int __pyx_setprop_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_children(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_8children_3__set__(o, v);
  }
  else {
    return __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_8children_5__del__(o);
  }
}

static PyMethodDef __pyx_methods_14glycopeptidepy_2_c_9structure_8fragment__NameTree[] = {
  {"get", (PyCFunction)__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_5get, METH_O, __pyx_doc_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_4get},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_15__reduce_cython__, METH_NOARGS, __pyx_doc_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_14__reduce_cython__},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_17__setstate_cython__, METH_O, __pyx_doc_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_16__setstate_cython__},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_14glycopeptidepy_2_c_9structure_8fragment__NameTree[] = {
  {(char *)"name", __pyx_getprop_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_name, __pyx_setprop_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_name, (char *)"name: object", 0},
  {(char *)"value", __pyx_getprop_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_value, __pyx_setprop_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_value, (char *)"value: object", 0},
  {(char *)"children", __pyx_getprop_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_children, __pyx_setprop_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_children, (char *)"children: dict", 0},
  {0, 0, 0, 0, 0}
};

static PySequenceMethods __pyx_tp_as_sequence__NameTree = {
  0, /*sq_length*/
  0, /*sq_concat*/
  0, /*sq_repeat*/
  __pyx_sq_item_14glycopeptidepy_2_c_9structure_8fragment__NameTree, /*sq_item*/
  0, /*sq_slice*/
  0, /*sq_ass_item*/
  0, /*sq_ass_slice*/
  0, /*sq_contains*/
  0, /*sq_inplace_concat*/
  0, /*sq_inplace_repeat*/
};

static PyMappingMethods __pyx_tp_as_mapping__NameTree = {
  0, /*mp_length*/
  __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_3__getitem__, /*mp_subscript*/
  0, /*mp_ass_subscript*/
};

static PyTypeObject __pyx_type_14glycopeptidepy_2_c_9structure_8fragment__NameTree = {
  PyVarObject_HEAD_INIT(0, 0)
  "glycopeptidepy._c.structure.fragment._NameTree", /*tp_name*/
  sizeof(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment__NameTree), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_14glycopeptidepy_2_c_9structure_8fragment__NameTree, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  &__pyx_tp_as_sequence__NameTree, /*tp_as_sequence*/
  &__pyx_tp_as_mapping__NameTree, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  "_NameTree(name=None, children=None)", /*tp_doc*/
  __pyx_tp_traverse_14glycopeptidepy_2_c_9structure_8fragment__NameTree, /*tp_traverse*/
  __pyx_tp_clear_14glycopeptidepy_2_c_9structure_8fragment__NameTree, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_14glycopeptidepy_2_c_9structure_8fragment__NameTree, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_14glycopeptidepy_2_c_9structure_8fragment__NameTree, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_14glycopeptidepy_2_c_9structure_8fragment__NameTree, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};

static struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct____pyx_fuse_0__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition *__pyx_freelist_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct____pyx_fuse_0__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition[8];
static int __pyx_freecount_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct____pyx_fuse_0__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition = 0;

static PyObject *__pyx_tp_new_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct____pyx_fuse_0__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct____pyx_fuse_0__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct____pyx_fuse_0__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition)))) {
    o = (PyObject*)__pyx_freelist_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct____pyx_fuse_0__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition[--__pyx_freecount_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct____pyx_fuse_0__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition];
    memset(o, 0, sizeof(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct____pyx_fuse_0__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct____pyx_fuse_0__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition(PyObject *o) {
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct____pyx_fuse_0__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition *p = (struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct____pyx_fuse_0__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_parts);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct____pyx_fuse_0__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct____pyx_fuse_0__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition)))) {
    __pyx_freelist_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct____pyx_fuse_0__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition[__pyx_freecount_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct____pyx_fuse_0__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition++] = ((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct____pyx_fuse_0__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct____pyx_fuse_0__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct____pyx_fuse_0__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition *p = (struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct____pyx_fuse_0__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition *)o;
  if (p->__pyx_v_parts) {
    e = (*v)(p->__pyx_v_parts, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct____pyx_fuse_0__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct____pyx_fuse_0__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition *p = (struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct____pyx_fuse_0__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition *)o;
  tmp = ((PyObject*)p->__pyx_v_parts);
  p->__pyx_v_parts = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct____pyx_fuse_0__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition = {
  PyVarObject_HEAD_INIT(0, 0)
  "glycopeptidepy._c.structure.fragment.__pyx_scope_struct____pyx_fuse_0__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition", /*tp_name*/
  sizeof(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct____pyx_fuse_0__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct____pyx_fuse_0__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct____pyx_fuse_0__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition, /*tp_traverse*/
  __pyx_tp_clear_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct____pyx_fuse_0__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct____pyx_fuse_0__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};

static struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_1_genexpr *__pyx_freelist_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_1_genexpr[8];
static int __pyx_freecount_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_1_genexpr = 0;

static PyObject *__pyx_tp_new_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_1_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_1_genexpr > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_1_genexpr)))) {
    o = (PyObject*)__pyx_freelist_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_1_genexpr[--__pyx_freecount_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_1_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_1_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_1_genexpr(PyObject *o) {
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_1_genexpr *p = (struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_1_genexpr *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_v_kv);
  Py_CLEAR(p->__pyx_t_0);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_1_genexpr < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_1_genexpr)))) {
    __pyx_freelist_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_1_genexpr[__pyx_freecount_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_1_genexpr++] = ((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_1_genexpr *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_1_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_1_genexpr *p = (struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_1_genexpr *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_v_kv) {
    e = (*v)(p->__pyx_v_kv, a); if (e) return e;
  }
  if (p->__pyx_t_0) {
    e = (*v)(p->__pyx_t_0, a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_1_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "glycopeptidepy._c.structure.fragment.__pyx_scope_struct_1_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_1_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_1_genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_1_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_1_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};

static struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_2___pyx_fuse_1__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition *__pyx_freelist_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_2___pyx_fuse_1__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition[8];
static int __pyx_freecount_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_2___pyx_fuse_1__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition = 0;

static PyObject *__pyx_tp_new_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_2___pyx_fuse_1__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_2___pyx_fuse_1__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_2___pyx_fuse_1__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition)))) {
    o = (PyObject*)__pyx_freelist_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_2___pyx_fuse_1__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition[--__pyx_freecount_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_2___pyx_fuse_1__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition];
    memset(o, 0, sizeof(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_2___pyx_fuse_1__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_2___pyx_fuse_1__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition(PyObject *o) {
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_2___pyx_fuse_1__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition *p = (struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_2___pyx_fuse_1__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_parts);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_2___pyx_fuse_1__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_2___pyx_fuse_1__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition)))) {
    __pyx_freelist_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_2___pyx_fuse_1__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition[__pyx_freecount_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_2___pyx_fuse_1__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition++] = ((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_2___pyx_fuse_1__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_2___pyx_fuse_1__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_2___pyx_fuse_1__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition *p = (struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_2___pyx_fuse_1__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition *)o;
  if (p->__pyx_v_parts) {
    e = (*v)(p->__pyx_v_parts, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_2___pyx_fuse_1__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_2___pyx_fuse_1__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition *p = (struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_2___pyx_fuse_1__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition *)o;
  tmp = ((PyObject*)p->__pyx_v_parts);
  p->__pyx_v_parts = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_2___pyx_fuse_1__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition = {
  PyVarObject_HEAD_INIT(0, 0)
  "glycopeptidepy._c.structure.fragment.__pyx_scope_struct_2___pyx_fuse_1__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition", /*tp_name*/
  sizeof(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_2___pyx_fuse_1__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_2___pyx_fuse_1__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_2___pyx_fuse_1__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition, /*tp_traverse*/
  __pyx_tp_clear_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_2___pyx_fuse_1__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_2___pyx_fuse_1__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};

static struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_3_genexpr *__pyx_freelist_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_3_genexpr[8];
static int __pyx_freecount_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_3_genexpr = 0;

static PyObject *__pyx_tp_new_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_3_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_3_genexpr > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_3_genexpr)))) {
    o = (PyObject*)__pyx_freelist_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_3_genexpr[--__pyx_freecount_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_3_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_3_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_3_genexpr(PyObject *o) {
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_3_genexpr *p = (struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_3_genexpr *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_v_kv);
  Py_CLEAR(p->__pyx_t_0);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_3_genexpr < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_3_genexpr)))) {
    __pyx_freelist_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_3_genexpr[__pyx_freecount_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_3_genexpr++] = ((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_3_genexpr *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_3_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_3_genexpr *p = (struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_3_genexpr *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_v_kv) {
    e = (*v)(p->__pyx_v_kv, a); if (e) return e;
  }
  if (p->__pyx_t_0) {
    e = (*v)(p->__pyx_t_0, a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_3_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "glycopeptidepy._c.structure.fragment.__pyx_scope_struct_3_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_3_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_3_genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_3_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_3_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};

static struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_4___pyx_fuse_2__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition *__pyx_freelist_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_4___pyx_fuse_2__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition[8];
static int __pyx_freecount_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_4___pyx_fuse_2__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition = 0;

static PyObject *__pyx_tp_new_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_4___pyx_fuse_2__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_4___pyx_fuse_2__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_4___pyx_fuse_2__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition)))) {
    o = (PyObject*)__pyx_freelist_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_4___pyx_fuse_2__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition[--__pyx_freecount_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_4___pyx_fuse_2__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition];
    memset(o, 0, sizeof(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_4___pyx_fuse_2__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_4___pyx_fuse_2__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition(PyObject *o) {
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_4___pyx_fuse_2__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition *p = (struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_4___pyx_fuse_2__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_parts);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_4___pyx_fuse_2__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_4___pyx_fuse_2__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition)))) {
    __pyx_freelist_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_4___pyx_fuse_2__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition[__pyx_freecount_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_4___pyx_fuse_2__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition++] = ((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_4___pyx_fuse_2__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_4___pyx_fuse_2__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_4___pyx_fuse_2__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition *p = (struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_4___pyx_fuse_2__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition *)o;
  if (p->__pyx_v_parts) {
    e = (*v)(p->__pyx_v_parts, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_4___pyx_fuse_2__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_4___pyx_fuse_2__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition *p = (struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_4___pyx_fuse_2__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition *)o;
  tmp = ((PyObject*)p->__pyx_v_parts);
  p->__pyx_v_parts = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_4___pyx_fuse_2__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition = {
  PyVarObject_HEAD_INIT(0, 0)
  "glycopeptidepy._c.structure.fragment.__pyx_scope_struct_4___pyx_fuse_2__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition", /*tp_name*/
  sizeof(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_4___pyx_fuse_2__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_4___pyx_fuse_2__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_4___pyx_fuse_2__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition, /*tp_traverse*/
  __pyx_tp_clear_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_4___pyx_fuse_2__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_4___pyx_fuse_2__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};

static struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_5_genexpr *__pyx_freelist_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_5_genexpr[8];
static int __pyx_freecount_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_5_genexpr = 0;

static PyObject *__pyx_tp_new_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_5_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_5_genexpr > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_5_genexpr)))) {
    o = (PyObject*)__pyx_freelist_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_5_genexpr[--__pyx_freecount_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_5_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_5_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_5_genexpr(PyObject *o) {
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_5_genexpr *p = (struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_5_genexpr *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_v_kv);
  Py_CLEAR(p->__pyx_t_0);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_5_genexpr < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_5_genexpr)))) {
    __pyx_freelist_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_5_genexpr[__pyx_freecount_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_5_genexpr++] = ((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_5_genexpr *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_5_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_5_genexpr *p = (struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_5_genexpr *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_v_kv) {
    e = (*v)(p->__pyx_v_kv, a); if (e) return e;
  }
  if (p->__pyx_t_0) {
    e = (*v)(p->__pyx_t_0, a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_5_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "glycopeptidepy._c.structure.fragment.__pyx_scope_struct_5_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_5_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_5_genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_5_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_5_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};

static struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_object____tuple___to_py *__pyx_freelist___pyx_scope_struct____Pyx_CFunc_object____tuple___to_py[8];
static int __pyx_freecount___pyx_scope_struct____Pyx_CFunc_object____tuple___to_py = 0;

static PyObject *__pyx_tp_new___pyx_scope_struct____Pyx_CFunc_object____tuple___to_py(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount___pyx_scope_struct____Pyx_CFunc_object____tuple___to_py > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_object____tuple___to_py)))) {
    o = (PyObject*)__pyx_freelist___pyx_scope_struct____Pyx_CFunc_object____tuple___to_py[--__pyx_freecount___pyx_scope_struct____Pyx_CFunc_object____tuple___to_py];
    memset(o, 0, sizeof(struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_object____tuple___to_py));
    (void) PyObject_INIT(o, t);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc___pyx_scope_struct____Pyx_CFunc_object____tuple___to_py(PyObject *o) {
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount___pyx_scope_struct____Pyx_CFunc_object____tuple___to_py < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_object____tuple___to_py)))) {
    __pyx_freelist___pyx_scope_struct____Pyx_CFunc_object____tuple___to_py[__pyx_freecount___pyx_scope_struct____Pyx_CFunc_object____tuple___to_py++] = ((struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_object____tuple___to_py *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static PyTypeObject __pyx_scope_struct____Pyx_CFunc_object____tuple___to_py = {
  PyVarObject_HEAD_INIT(0, 0)
  "glycopeptidepy._c.structure.fragment.__pyx_scope_struct____Pyx_CFunc_object____tuple___to_py", /*tp_name*/
  sizeof(struct __pyx_obj___pyx_scope_struct____Pyx_CFunc_object____tuple___to_py), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc___pyx_scope_struct____Pyx_CFunc_object____tuple___to_py, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new___pyx_scope_struct____Pyx_CFunc_object____tuple___to_py, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
  0, /*tp_pypy_flags*/
  #endif
};

static PyMethodDef __pyx_methods[] = {
  {0, 0, 0, 0}
};

#if PY_MAJOR_VERSION >= 3
#if CYTHON_PEP489_MULTI_PHASE_INIT
static PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def); /*proto*/
static int __pyx_pymod_exec_fragment(PyObject* module); /*proto*/
static PyModuleDef_Slot __pyx_moduledef_slots[] = {
  {Py_mod_create, (void*)__pyx_pymod_create},
  {Py_mod_exec, (void*)__pyx_pymod_exec_fragment},
  {0, NULL}
};
#endif

static struct PyModuleDef __pyx_moduledef = {
    PyModuleDef_HEAD_INIT,
    "fragment",
    0, /* m_doc */
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    0, /* m_size */
  #else
    -1, /* m_size */
  #endif
    __pyx_methods /* m_methods */,
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    __pyx_moduledef_slots, /* m_slots */
  #else
    NULL, /* m_reload */
  #endif
    NULL, /* m_traverse */
    NULL, /* m_clear */
    NULL /* m_free */
};
#endif
#ifndef CYTHON_SMALL_CODE
#if defined(__clang__)
    #define CYTHON_SMALL_CODE
#elif defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 3))
    #define CYTHON_SMALL_CODE __attribute__((cold))
#else
    #define CYTHON_SMALL_CODE
#endif
#endif

static __Pyx_StringTabEntry __pyx_string_tab[] = {
  {&__pyx_kp_s_, __pyx_k_, sizeof(__pyx_k_), 0, 0, 1, 0},
  {&__pyx_n_s_AttributeError, __pyx_k_AttributeError, sizeof(__pyx_k_AttributeError), 0, 0, 1, 1},
  {&__pyx_n_s_ChemicalShiftBase, __pyx_k_ChemicalShiftBase, sizeof(__pyx_k_ChemicalShiftBase), 0, 0, 1, 1},
  {&__pyx_kp_s_Composition_data_is_missing, __pyx_k_Composition_data_is_missing, sizeof(__pyx_k_Composition_data_is_missing), 0, 0, 1, 0},
  {&__pyx_kp_s_Expected_at_least_d_argument_s_g, __pyx_k_Expected_at_least_d_argument_s_g, sizeof(__pyx_k_Expected_at_least_d_argument_s_g), 0, 0, 1, 0},
  {&__pyx_n_s_FragmentBase, __pyx_k_FragmentBase, sizeof(__pyx_k_FragmentBase), 0, 0, 1, 1},
  {&__pyx_kp_s_Function_call_with_ambiguous_arg, __pyx_k_Function_call_with_ambiguous_arg, sizeof(__pyx_k_Function_call_with_ambiguous_arg), 0, 0, 1, 0},
  {&__pyx_n_s_GlycosaminoglycanLinkerGlycosyla, __pyx_k_GlycosaminoglycanLinkerGlycosyla, sizeof(__pyx_k_GlycosaminoglycanLinkerGlycosyla), 0, 0, 1, 1},
  {&__pyx_n_s_HexNAc, __pyx_k_HexNAc, sizeof(__pyx_k_HexNAc), 0, 0, 1, 1},
  {&__pyx_kp_s_Incompatible_checksums_0x_x_vs_0, __pyx_k_Incompatible_checksums_0x_x_vs_0, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0), 0, 0, 1, 0},
  {&__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_2, __pyx_k_Incompatible_checksums_0x_x_vs_0_2, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0_2), 0, 0, 1, 0},
  {&__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_3, __pyx_k_Incompatible_checksums_0x_x_vs_0_3, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0_3), 0, 0, 1, 0},
  {&__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_4, __pyx_k_Incompatible_checksums_0x_x_vs_0_4, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0_4), 0, 0, 1, 0},
  {&__pyx_n_s_IonSeriesBase, __pyx_k_IonSeriesBase, sizeof(__pyx_k_IonSeriesBase), 0, 0, 1, 1},
  {&__pyx_n_s_Modification, __pyx_k_Modification, sizeof(__pyx_k_Modification), 0, 0, 1, 1},
  {&__pyx_n_s_ModificationCategory, __pyx_k_ModificationCategory, sizeof(__pyx_k_ModificationCategory), 0, 0, 1, 1},
  {&__pyx_n_s_NGlycanCoreGlycosylation, __pyx_k_NGlycanCoreGlycosylation, sizeof(__pyx_k_NGlycanCoreGlycosylation), 0, 0, 1, 1},
  {&__pyx_n_s_NameTree, __pyx_k_NameTree, sizeof(__pyx_k_NameTree), 0, 0, 1, 1},
  {&__pyx_n_s_NameTree___pyx_fuse_0traverse, __pyx_k_NameTree___pyx_fuse_0traverse, sizeof(__pyx_k_NameTree___pyx_fuse_0traverse), 0, 0, 1, 1},
  {&__pyx_n_s_NameTree___pyx_fuse_1traverse, __pyx_k_NameTree___pyx_fuse_1traverse, sizeof(__pyx_k_NameTree___pyx_fuse_1traverse), 0, 0, 1, 1},
  {&__pyx_n_s_NameTree_traverse, __pyx_k_NameTree_traverse, sizeof(__pyx_k_NameTree_traverse), 0, 0, 1, 1},
  {&__pyx_kp_s_No_matching_signature_found, __pyx_k_No_matching_signature_found, sizeof(__pyx_k_No_matching_signature_found), 0, 0, 1, 0},
  {&__pyx_n_s_NotImplementedError, __pyx_k_NotImplementedError, sizeof(__pyx_k_NotImplementedError), 0, 0, 1, 1},
  {&__pyx_n_s_OGlycanCoreGlycosylation, __pyx_k_OGlycanCoreGlycosylation, sizeof(__pyx_k_OGlycanCoreGlycosylation), 0, 0, 1, 1},
  {&__pyx_n_s_PeptideFragment, __pyx_k_PeptideFragment, sizeof(__pyx_k_PeptideFragment), 0, 0, 1, 1},
  {&__pyx_kp_s_PeptideFragment_self_series_self, __pyx_k_PeptideFragment_self_series_self, sizeof(__pyx_k_PeptideFragment_self_series_self), 0, 0, 1, 0},
  {&__pyx_n_s_PickleError, __pyx_k_PickleError, sizeof(__pyx_k_PickleError), 0, 0, 1, 1},
  {&__pyx_n_s_Pyx_CFunc_object____tuple___to, __pyx_k_Pyx_CFunc_object____tuple___to, sizeof(__pyx_k_Pyx_CFunc_object____tuple___to), 0, 0, 1, 1},
  {&__pyx_n_s_SimpleFragment, __pyx_k_SimpleFragment, sizeof(__pyx_k_SimpleFragment), 0, 0, 1, 1},
  {&__pyx_n_s_StubFragment, __pyx_k_StubFragment, sizeof(__pyx_k_StubFragment), 0, 0, 1, 1},
  {&__pyx_n_s_TypeError, __pyx_k_TypeError, sizeof(__pyx_k_TypeError), 0, 0, 1, 1},
  {&__pyx_n_s_ValueError, __pyx_k_ValueError, sizeof(__pyx_k_ValueError), 0, 0, 1, 1},
  {&__pyx_n_s_Xyl, __pyx_k_Xyl, sizeof(__pyx_k_Xyl), 0, 0, 1, 1},
  {&__pyx_kp_s__3, __pyx_k__3, sizeof(__pyx_k__3), 0, 0, 1, 0},
  {&__pyx_kp_s__4, __pyx_k__4, sizeof(__pyx_k__4), 0, 0, 1, 0},
  {&__pyx_n_s_args, __pyx_k_args, sizeof(__pyx_k_args), 0, 0, 1, 1},
  {&__pyx_n_s_base_name, __pyx_k_base_name, sizeof(__pyx_k_base_name), 0, 0, 1, 1},
  {&__pyx_n_s_build_name_from_composition, __pyx_k_build_name_from_composition, sizeof(__pyx_k_build_name_from_composition), 0, 0, 1, 1},
  {&__pyx_n_s_cell, __pyx_k_cell, sizeof(__pyx_k_cell), 0, 0, 1, 1},
  {&__pyx_n_s_cfunc_to_py, __pyx_k_cfunc_to_py, sizeof(__pyx_k_cfunc_to_py), 0, 0, 1, 1},
  {&__pyx_n_s_chemical_shift, __pyx_k_chemical_shift, sizeof(__pyx_k_chemical_shift), 0, 0, 1, 1},
  {&__pyx_n_s_children, __pyx_k_children, sizeof(__pyx_k_children), 0, 0, 1, 1},
  {&__pyx_n_s_class, __pyx_k_class, sizeof(__pyx_k_class), 0, 0, 1, 1},
  {&__pyx_n_s_cline_in_traceback, __pyx_k_cline_in_traceback, sizeof(__pyx_k_cline_in_traceback), 0, 0, 1, 1},
  {&__pyx_n_s_clone, __pyx_k_clone, sizeof(__pyx_k_clone), 0, 0, 1, 1},
  {&__pyx_n_s_close, __pyx_k_close, sizeof(__pyx_k_close), 0, 0, 1, 1},
  {&__pyx_n_s_composition, __pyx_k_composition, sizeof(__pyx_k_composition), 0, 0, 1, 1},
  {&__pyx_n_s_debug_get_str_int, __pyx_k_debug_get_str_int, sizeof(__pyx_k_debug_get_str_int), 0, 0, 1, 1},
  {&__pyx_n_s_defaults, __pyx_k_defaults, sizeof(__pyx_k_defaults), 0, 0, 1, 1},
  {&__pyx_n_s_dict, __pyx_k_dict, sizeof(__pyx_k_dict), 0, 0, 1, 1},
  {&__pyx_n_s_end, __pyx_k_end, sizeof(__pyx_k_end), 0, 0, 1, 1},
  {&__pyx_n_s_file, __pyx_k_file, sizeof(__pyx_k_file), 0, 0, 1, 1},
  {&__pyx_n_s_flanking_amino_acids, __pyx_k_flanking_amino_acids, sizeof(__pyx_k_flanking_amino_acids), 0, 0, 1, 1},
  {&__pyx_n_s_format, __pyx_k_format, sizeof(__pyx_k_format), 0, 0, 1, 1},
  {&__pyx_n_s_genexpr, __pyx_k_genexpr, sizeof(__pyx_k_genexpr), 0, 0, 1, 1},
  {&__pyx_n_s_get_chemical_shift, __pyx_k_get_chemical_shift, sizeof(__pyx_k_get_chemical_shift), 0, 0, 1, 1},
  {&__pyx_n_s_get_fragment_name, __pyx_k_get_fragment_name, sizeof(__pyx_k_get_fragment_name), 0, 0, 1, 1},
  {&__pyx_n_s_get_series, __pyx_k_get_series, sizeof(__pyx_k_get_series), 0, 0, 1, 1},
  {&__pyx_n_s_get_string_cell, __pyx_k_get_string_cell, sizeof(__pyx_k_get_string_cell), 0, 0, 1, 1},
  {&__pyx_n_s_getstate, __pyx_k_getstate, sizeof(__pyx_k_getstate), 0, 0, 1, 1},
  {&__pyx_n_s_glycopeptidepy__c_structure_frag, __pyx_k_glycopeptidepy__c_structure_frag, sizeof(__pyx_k_glycopeptidepy__c_structure_frag), 0, 0, 1, 1},
  {&__pyx_n_s_glycopeptidepy_structure_modific, __pyx_k_glycopeptidepy_structure_modific, sizeof(__pyx_k_glycopeptidepy_structure_modific), 0, 0, 1, 1},
  {&__pyx_n_s_glycosylation, __pyx_k_glycosylation, sizeof(__pyx_k_glycosylation), 0, 0, 1, 1},
  {&__pyx_n_s_i, __pyx_k_i, sizeof(__pyx_k_i), 0, 0, 1, 1},
  {&__pyx_n_s_import, __pyx_k_import, sizeof(__pyx_k_import), 0, 0, 1, 1},
  {&__pyx_n_s_is_extended, __pyx_k_is_extended, sizeof(__pyx_k_is_extended), 0, 0, 1, 1},
  {&__pyx_n_s_is_glycosylated, __pyx_k_is_glycosylated, sizeof(__pyx_k_is_glycosylated), 0, 0, 1, 1},
  {&__pyx_n_s_items, __pyx_k_items, sizeof(__pyx_k_items), 0, 0, 1, 1},
  {&__pyx_n_s_join, __pyx_k_join, sizeof(__pyx_k_join), 0, 0, 1, 1},
  {&__pyx_n_s_key, __pyx_k_key, sizeof(__pyx_k_key), 0, 0, 1, 1},
  {&__pyx_n_s_kind, __pyx_k_kind, sizeof(__pyx_k_kind), 0, 0, 1, 1},
  {&__pyx_n_s_kwargs, __pyx_k_kwargs, sizeof(__pyx_k_kwargs), 0, 0, 1, 1},
  {&__pyx_n_s_list, __pyx_k_list, sizeof(__pyx_k_list), 0, 0, 1, 1},
  {&__pyx_n_s_main, __pyx_k_main, sizeof(__pyx_k_main), 0, 0, 1, 1},
  {&__pyx_n_s_mass, __pyx_k_mass, sizeof(__pyx_k_mass), 0, 0, 1, 1},
  {&__pyx_n_s_modification_dict, __pyx_k_modification_dict, sizeof(__pyx_k_modification_dict), 0, 0, 1, 1},
  {&__pyx_n_s_name, __pyx_k_name, sizeof(__pyx_k_name), 0, 0, 1, 1},
  {&__pyx_n_s_name_2, __pyx_k_name_2, sizeof(__pyx_k_name_2), 0, 0, 1, 1},
  {&__pyx_n_s_name_cache, __pyx_k_name_cache, sizeof(__pyx_k_name_cache), 0, 0, 1, 1},
  {&__pyx_n_s_new, __pyx_k_new, sizeof(__pyx_k_new), 0, 0, 1, 1},
  {&__pyx_n_s_numpy, __pyx_k_numpy, sizeof(__pyx_k_numpy), 0, 0, 1, 1},
  {&__pyx_n_s_object, __pyx_k_object, sizeof(__pyx_k_object), 0, 0, 1, 1},
  {&__pyx_n_s_out, __pyx_k_out, sizeof(__pyx_k_out), 0, 0, 1, 1},
  {&__pyx_n_s_parts, __pyx_k_parts, sizeof(__pyx_k_parts), 0, 0, 1, 1},
  {&__pyx_n_s_peptide, __pyx_k_peptide, sizeof(__pyx_k_peptide), 0, 0, 1, 1},
  {&__pyx_n_s_pickle, __pyx_k_pickle, sizeof(__pyx_k_pickle), 0, 0, 1, 1},
  {&__pyx_n_s_position, __pyx_k_position, sizeof(__pyx_k_position), 0, 0, 1, 1},
  {&__pyx_n_s_print, __pyx_k_print, sizeof(__pyx_k_print), 0, 0, 1, 1},
  {&__pyx_n_s_property, __pyx_k_property, sizeof(__pyx_k_property), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_PickleError, __pyx_k_pyx_PickleError, sizeof(__pyx_k_pyx_PickleError), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_checksum, __pyx_k_pyx_checksum, sizeof(__pyx_k_pyx_checksum), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_fuse_0build_name_from_comp, __pyx_k_pyx_fuse_0build_name_from_comp, sizeof(__pyx_k_pyx_fuse_0build_name_from_comp), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_fuse_0traverse, __pyx_k_pyx_fuse_0traverse, sizeof(__pyx_k_pyx_fuse_0traverse), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_fuse_1build_name_from_comp, __pyx_k_pyx_fuse_1build_name_from_comp, sizeof(__pyx_k_pyx_fuse_1build_name_from_comp), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_fuse_2build_name_from_comp, __pyx_k_pyx_fuse_2build_name_from_comp, sizeof(__pyx_k_pyx_fuse_2build_name_from_comp), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_result, __pyx_k_pyx_result, sizeof(__pyx_k_pyx_result), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_state, __pyx_k_pyx_state, sizeof(__pyx_k_pyx_state), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_type, __pyx_k_pyx_type, sizeof(__pyx_k_pyx_type), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_unpickle_ChemicalShiftBase, __pyx_k_pyx_unpickle_ChemicalShiftBase, sizeof(__pyx_k_pyx_unpickle_ChemicalShiftBase), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_unpickle_FragmentBase, __pyx_k_pyx_unpickle_FragmentBase, sizeof(__pyx_k_pyx_unpickle_FragmentBase), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_unpickle_IonSeriesBase, __pyx_k_pyx_unpickle_IonSeriesBase, sizeof(__pyx_k_pyx_unpickle_IonSeriesBase), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_unpickle__NameTree, __pyx_k_pyx_unpickle__NameTree, sizeof(__pyx_k_pyx_unpickle__NameTree), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_vtable, __pyx_k_pyx_vtable, sizeof(__pyx_k_pyx_vtable), 0, 0, 1, 1},
  {&__pyx_n_s_range, __pyx_k_range, sizeof(__pyx_k_range), 0, 0, 1, 1},
  {&__pyx_n_s_reduce, __pyx_k_reduce, sizeof(__pyx_k_reduce), 0, 0, 1, 1},
  {&__pyx_n_s_reduce_cython, __pyx_k_reduce_cython, sizeof(__pyx_k_reduce_cython), 0, 0, 1, 1},
  {&__pyx_n_s_reduce_ex, __pyx_k_reduce_ex, sizeof(__pyx_k_reduce_ex), 0, 0, 1, 1},
  {&__pyx_n_s_result, __pyx_k_result, sizeof(__pyx_k_result), 0, 0, 1, 1},
  {&__pyx_n_s_rule, __pyx_k_rule, sizeof(__pyx_k_rule), 0, 0, 1, 1},
  {&__pyx_n_s_s, __pyx_k_s, sizeof(__pyx_k_s), 0, 0, 1, 1},
  {&__pyx_kp_s_s_d, __pyx_k_s_d, sizeof(__pyx_k_s_d), 0, 0, 1, 0},
  {&__pyx_kp_s_s_name_r, __pyx_k_s_name_r, sizeof(__pyx_k_s_name_r), 0, 0, 1, 0},
  {&__pyx_kp_s_s_s, __pyx_k_s_s, sizeof(__pyx_k_s_s), 0, 0, 1, 0},
  {&__pyx_n_s_self, __pyx_k_self, sizeof(__pyx_k_self), 0, 0, 1, 1},
  {&__pyx_kp_s_self___class_____name___name_se, __pyx_k_self___class_____name___name_se, sizeof(__pyx_k_self___class_____name___name_se), 0, 0, 1, 0},
  {&__pyx_n_s_send, __pyx_k_send, sizeof(__pyx_k_send), 0, 0, 1, 1},
  {&__pyx_n_s_series, __pyx_k_series, sizeof(__pyx_k_series), 0, 0, 1, 1},
  {&__pyx_n_s_set_chemical_shift, __pyx_k_set_chemical_shift, sizeof(__pyx_k_set_chemical_shift), 0, 0, 1, 1},
  {&__pyx_n_s_setstate, __pyx_k_setstate, sizeof(__pyx_k_setstate), 0, 0, 1, 1},
  {&__pyx_n_s_setstate_cython, __pyx_k_setstate_cython, sizeof(__pyx_k_setstate_cython), 0, 0, 1, 1},
  {&__pyx_n_s_sort, __pyx_k_sort, sizeof(__pyx_k_sort), 0, 0, 1, 1},
  {&__pyx_n_s_split, __pyx_k_split, sizeof(__pyx_k_split), 0, 0, 1, 1},
  {&__pyx_kp_s_src_glycopeptidepy__c_structure, __pyx_k_src_glycopeptidepy__c_structure, sizeof(__pyx_k_src_glycopeptidepy__c_structure), 0, 0, 1, 0},
  {&__pyx_kp_s_stringsource, __pyx_k_stringsource, sizeof(__pyx_k_stringsource), 0, 0, 1, 0},
  {&__pyx_n_s_strip, __pyx_k_strip, sizeof(__pyx_k_strip), 0, 0, 1, 1},
  {&__pyx_n_s_sum, __pyx_k_sum, sizeof(__pyx_k_sum), 0, 0, 1, 1},
  {&__pyx_n_s_test, __pyx_k_test, sizeof(__pyx_k_test), 0, 0, 1, 1},
  {&__pyx_n_s_throw, __pyx_k_throw, sizeof(__pyx_k_throw), 0, 0, 1, 1},
  {&__pyx_n_s_traverse, __pyx_k_traverse, sizeof(__pyx_k_traverse), 0, 0, 1, 1},
  {&__pyx_n_s_update, __pyx_k_update, sizeof(__pyx_k_update), 0, 0, 1, 1},
  {&__pyx_n_s_values, __pyx_k_values, sizeof(__pyx_k_values), 0, 0, 1, 1},
  {&__pyx_n_s_wrap, __pyx_k_wrap, sizeof(__pyx_k_wrap), 0, 0, 1, 1},
  {&__pyx_n_s_x, __pyx_k_x, sizeof(__pyx_k_x), 0, 0, 1, 1},
  {0, 0, 0, 0, 0, 0, 0}
};
static CYTHON_SMALL_CODE int __Pyx_InitCachedBuiltins(void) {
  __pyx_builtin_property = __Pyx_GetBuiltinName(__pyx_n_s_property); if (!__pyx_builtin_property) __PYX_ERR(0, 255, __pyx_L1_error)
  __pyx_builtin_range = __Pyx_GetBuiltinName(__pyx_n_s_range); if (!__pyx_builtin_range) __PYX_ERR(0, 54, __pyx_L1_error)
  __pyx_builtin_NotImplementedError = __Pyx_GetBuiltinName(__pyx_n_s_NotImplementedError); if (!__pyx_builtin_NotImplementedError) __PYX_ERR(0, 190, __pyx_L1_error)
  __pyx_builtin_AttributeError = __Pyx_GetBuiltinName(__pyx_n_s_AttributeError); if (!__pyx_builtin_AttributeError) __PYX_ERR(0, 203, __pyx_L1_error)
  __pyx_builtin_ValueError = __Pyx_GetBuiltinName(__pyx_n_s_ValueError); if (!__pyx_builtin_ValueError) __PYX_ERR(0, 393, __pyx_L1_error)
  __pyx_builtin_TypeError = __Pyx_GetBuiltinName(__pyx_n_s_TypeError); if (!__pyx_builtin_TypeError) __PYX_ERR(0, 646, __pyx_L1_error)
  __pyx_builtin_sum = __Pyx_GetBuiltinName(__pyx_n_s_sum); if (!__pyx_builtin_sum) __PYX_ERR(0, 755, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "glycopeptidepy/_c/structure/fragment.pyx":393
 *     def total_composition(self):
 *         if self.composition is None:
 *             raise ValueError("Composition data is missing")             # <<<<<<<<<<<<<<
 *         composition = self.composition.clone()
 *         chemical_shift = self.chemical_shift
 */
  __pyx_tuple__2 = PyTuple_Pack(1, __pyx_kp_s_Composition_data_is_missing); if (unlikely(!__pyx_tuple__2)) __PYX_ERR(0, 393, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2);
  __Pyx_GIVEREF(__pyx_tuple__2);

  /* "glycopeptidepy/_c/structure/fragment.pyx":646
 *             return result
 * 
 *     cpdef traverse(self, list_or_iterable parts):             # <<<<<<<<<<<<<<
 *         cdef:
 *             _NameTree node
 */
  __pyx_tuple__5 = PyTuple_Pack(1, __pyx_kp_s_No_matching_signature_found); if (unlikely(!__pyx_tuple__5)) __PYX_ERR(0, 646, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__5);
  __Pyx_GIVEREF(__pyx_tuple__5);
  __pyx_tuple__6 = PyTuple_Pack(1, __pyx_kp_s_Function_call_with_ambiguous_arg); if (unlikely(!__pyx_tuple__6)) __PYX_ERR(0, 646, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__6);
  __Pyx_GIVEREF(__pyx_tuple__6);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x54d98d7, 0xbef1095, 0x80a4bd4):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (0x%x vs (0x54d98d7, 0xbef1095, 0x80a4bd4) = (composition, mass, name))" % __pyx_checksum)
 */
  __pyx_tuple__7 = PyTuple_Pack(3, __pyx_int_88971479, __pyx_int_200216725, __pyx_int_134892500); if (unlikely(!__pyx_tuple__7)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__7);
  __Pyx_GIVEREF(__pyx_tuple__7);
  __pyx_tuple__8 = PyTuple_Pack(3, __pyx_int_254818202, __pyx_int_5912882, __pyx_int_202249039); if (unlikely(!__pyx_tuple__8)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__8);
  __Pyx_GIVEREF(__pyx_tuple__8);
  __pyx_tuple__9 = PyTuple_Pack(3, __pyx_int_3385516, __pyx_int_188675731, __pyx_int_211098802); if (unlikely(!__pyx_tuple__9)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__9);
  __Pyx_GIVEREF(__pyx_tuple__9);
  __pyx_tuple__10 = PyTuple_Pack(3, __pyx_int_220877754, __pyx_int_27303459, __pyx_int_134022094); if (unlikely(!__pyx_tuple__10)) __PYX_ERR(2, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__10);
  __Pyx_GIVEREF(__pyx_tuple__10);

  /* "cfunc.to_py":65
 * @cname("__Pyx_CFunc_object____tuple___to_py")
 * cdef object __Pyx_CFunc_object____tuple___to_py(object (*f)(tuple) ):
 *     def wrap(tuple x):             # <<<<<<<<<<<<<<
 *         """wrap(x: tuple)"""
 *         return f(x)
 */
  __pyx_tuple__11 = PyTuple_Pack(1, __pyx_n_s_x); if (unlikely(!__pyx_tuple__11)) __PYX_ERR(2, 65, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__11);
  __Pyx_GIVEREF(__pyx_tuple__11);
  __pyx_codeobj__12 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__11, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_wrap, 65, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__12)) __PYX_ERR(2, 65, __pyx_L1_error)

  /* "glycopeptidepy/_c/structure/fragment.pyx":33
 * cdef ModificationBase _o_glycosylation = OGlycanCoreGlycosylation()
 * cdef ModificationBase _gag_linker_glycosylation = GlycosaminoglycanLinkerGlycosylation()
 * cdef ModificationBase _modification_hexnac = Modification("HexNAc").rule             # <<<<<<<<<<<<<<
 * cdef ModificationBase _modification_xylose = Modification("Xyl").rule
 * 
 */
  __pyx_tuple__13 = PyTuple_Pack(1, __pyx_n_s_HexNAc); if (unlikely(!__pyx_tuple__13)) __PYX_ERR(0, 33, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__13);
  __Pyx_GIVEREF(__pyx_tuple__13);

  /* "glycopeptidepy/_c/structure/fragment.pyx":34
 * cdef ModificationBase _gag_linker_glycosylation = GlycosaminoglycanLinkerGlycosylation()
 * cdef ModificationBase _modification_hexnac = Modification("HexNAc").rule
 * cdef ModificationBase _modification_xylose = Modification("Xyl").rule             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_tuple__14 = PyTuple_Pack(1, __pyx_n_s_Xyl); if (unlikely(!__pyx_tuple__14)) __PYX_ERR(0, 34, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__14);
  __Pyx_GIVEREF(__pyx_tuple__14);

  /* "glycopeptidepy/_c/structure/fragment.pyx":100
 * 
 * 
 * def _debug_get_str_int(int i):             # <<<<<<<<<<<<<<
 *     cdef:
 *         string_cell out
 */
  __pyx_tuple__15 = PyTuple_Pack(4, __pyx_n_s_i, __pyx_n_s_i, __pyx_n_s_out, __pyx_n_s_result); if (unlikely(!__pyx_tuple__15)) __PYX_ERR(0, 100, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__15);
  __Pyx_GIVEREF(__pyx_tuple__15);
  __pyx_codeobj__16 = (PyObject*)__Pyx_PyCode_New(1, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__15, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_glycopeptidepy__c_structure, __pyx_n_s_debug_get_str_int, 100, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__16)) __PYX_ERR(0, 100, __pyx_L1_error)

  /* "glycopeptidepy/_c/structure/fragment.pyx":109
 *     print(out.size)
 * 
 * def _get_string_cell(int i):             # <<<<<<<<<<<<<<
 *     cdef string_cell cell = str_ints[i]
 *     print(cell.size)
 */
  __pyx_tuple__17 = PyTuple_Pack(3, __pyx_n_s_i, __pyx_n_s_i, __pyx_n_s_cell); if (unlikely(!__pyx_tuple__17)) __PYX_ERR(0, 109, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__17);
  __Pyx_GIVEREF(__pyx_tuple__17);
  __pyx_codeobj__18 = (PyObject*)__Pyx_PyCode_New(1, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__17, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_glycopeptidepy__c_structure, __pyx_n_s_get_string_cell, 109, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__18)) __PYX_ERR(0, 109, __pyx_L1_error)

  /* "glycopeptidepy/_c/structure/fragment.pyx":646
 *             return result
 * 
 *     cpdef traverse(self, list_or_iterable parts):             # <<<<<<<<<<<<<<
 *         cdef:
 *             _NameTree node
 */
  __pyx_tuple__19 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_parts); if (unlikely(!__pyx_tuple__19)) __PYX_ERR(0, 646, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__19);
  __Pyx_GIVEREF(__pyx_tuple__19);
  __pyx_codeobj__20 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__19, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_glycopeptidepy__c_structure, __pyx_n_s_pyx_fuse_0traverse, 646, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__20)) __PYX_ERR(0, 646, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __pyx_unpickle_ChemicalShiftBase(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_tuple__21 = PyTuple_Pack(5, __pyx_n_s_pyx_type, __pyx_n_s_pyx_checksum, __pyx_n_s_pyx_state, __pyx_n_s_pyx_PickleError, __pyx_n_s_pyx_result); if (unlikely(!__pyx_tuple__21)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__21);
  __Pyx_GIVEREF(__pyx_tuple__21);
  __pyx_codeobj__22 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__21, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_ChemicalShiftBase, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__22)) __PYX_ERR(2, 1, __pyx_L1_error)
  __pyx_tuple__23 = PyTuple_Pack(5, __pyx_n_s_pyx_type, __pyx_n_s_pyx_checksum, __pyx_n_s_pyx_state, __pyx_n_s_pyx_PickleError, __pyx_n_s_pyx_result); if (unlikely(!__pyx_tuple__23)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__23);
  __Pyx_GIVEREF(__pyx_tuple__23);
  __pyx_codeobj__24 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__23, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_IonSeriesBase, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__24)) __PYX_ERR(2, 1, __pyx_L1_error)
  __pyx_tuple__25 = PyTuple_Pack(5, __pyx_n_s_pyx_type, __pyx_n_s_pyx_checksum, __pyx_n_s_pyx_state, __pyx_n_s_pyx_PickleError, __pyx_n_s_pyx_result); if (unlikely(!__pyx_tuple__25)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__25);
  __Pyx_GIVEREF(__pyx_tuple__25);
  __pyx_codeobj__26 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__25, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_FragmentBase, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__26)) __PYX_ERR(2, 1, __pyx_L1_error)
  __pyx_tuple__27 = PyTuple_Pack(5, __pyx_n_s_pyx_type, __pyx_n_s_pyx_checksum, __pyx_n_s_pyx_state, __pyx_n_s_pyx_PickleError, __pyx_n_s_pyx_result); if (unlikely(!__pyx_tuple__27)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__27);
  __Pyx_GIVEREF(__pyx_tuple__27);
  __pyx_codeobj__28 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__27, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle__NameTree, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__28)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitGlobals(void) {
  if (__Pyx_InitStrings(__pyx_string_tab) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_float_1eneg_5 = PyFloat_FromDouble(1e-5); if (unlikely(!__pyx_float_1eneg_5)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_0 = PyInt_FromLong(0); if (unlikely(!__pyx_int_0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_2 = PyInt_FromLong(2); if (unlikely(!__pyx_int_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_3385516 = PyInt_FromLong(3385516L); if (unlikely(!__pyx_int_3385516)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_5912882 = PyInt_FromLong(5912882L); if (unlikely(!__pyx_int_5912882)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_27303459 = PyInt_FromLong(27303459L); if (unlikely(!__pyx_int_27303459)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_88971479 = PyInt_FromLong(88971479L); if (unlikely(!__pyx_int_88971479)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_134022094 = PyInt_FromLong(134022094L); if (unlikely(!__pyx_int_134022094)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_134892500 = PyInt_FromLong(134892500L); if (unlikely(!__pyx_int_134892500)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_188675731 = PyInt_FromLong(188675731L); if (unlikely(!__pyx_int_188675731)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_200216725 = PyInt_FromLong(200216725L); if (unlikely(!__pyx_int_200216725)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_202249039 = PyInt_FromLong(202249039L); if (unlikely(!__pyx_int_202249039)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_211098802 = PyInt_FromLong(211098802L); if (unlikely(!__pyx_int_211098802)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_220877754 = PyInt_FromLong(220877754L); if (unlikely(!__pyx_int_220877754)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_254818202 = PyInt_FromLong(254818202L); if (unlikely(!__pyx_int_254818202)) __PYX_ERR(0, 1, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_modinit_global_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_import_code(void); /*proto*/

static int __Pyx_modinit_global_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_global_init_code", 0);
  /*--- Global init code ---*/
  __pyx_v_14glycopeptidepy_2_c_9structure_8fragment__n_glycosylation = ((struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_ModificationBase *)Py_None); Py_INCREF(Py_None);
  __pyx_v_14glycopeptidepy_2_c_9structure_8fragment__o_glycosylation = ((struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_ModificationBase *)Py_None); Py_INCREF(Py_None);
  __pyx_v_14glycopeptidepy_2_c_9structure_8fragment__gag_linker_glycosylation = ((struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_ModificationBase *)Py_None); Py_INCREF(Py_None);
  __pyx_v_14glycopeptidepy_2_c_9structure_8fragment__modification_hexnac = ((struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_ModificationBase *)Py_None); Py_INCREF(Py_None);
  __pyx_v_14glycopeptidepy_2_c_9structure_8fragment__modification_xylose = ((struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_ModificationBase *)Py_None); Py_INCREF(Py_None);
  __pyx_v_14glycopeptidepy_2_c_9structure_8fragment_str_int_overflow_intern_cache = ((PyObject*)Py_None); Py_INCREF(Py_None);
  __pyx_v_14glycopeptidepy_2_c_9structure_8fragment_ModificationCategory_glycosylation = Py_None; Py_INCREF(Py_None);
  __pyx_v_14glycopeptidepy_2_c_9structure_8fragment_stub_fragment_name_cache = ((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment__NameTree *)Py_None); Py_INCREF(Py_None);
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_export_code", 0);
  /*--- Variable export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_export_code(void) {
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_export_code", 0);
  /*--- Function export code ---*/
  if (__Pyx_ExportFunction("__pyx_fuse_0build_name_from_composition", (void (*)(void))__pyx_fuse_0__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition, "PyObject *(struct __pyx_obj_14glycopeptidepy_2_c_11count_table_CountTable *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("__pyx_fuse_1build_name_from_composition", (void (*)(void))__pyx_fuse_1__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition, "PyObject *(PyObject *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("__pyx_fuse_2build_name_from_composition", (void (*)(void))__pyx_fuse_2__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition, "PyObject *(PyObject *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_init_code(void) {
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_init_code", 0);
  /*--- Type init code ---*/
  __pyx_vtabptr_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase = &__pyx_vtable_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase;
  __pyx_vtable_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase.clone = (struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase *(*)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase *, int __pyx_skip_dispatch))__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_17ChemicalShiftBase_clone;
  if (PyType_Ready(&__pyx_type_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase) < 0) __PYX_ERR(0, 116, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase.tp_dictoffset && __pyx_type_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase.tp_dict, __pyx_vtabptr_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase) < 0) __PYX_ERR(0, 116, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_ChemicalShiftBase, (PyObject *)&__pyx_type_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase) < 0) __PYX_ERR(0, 116, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase) < 0) __PYX_ERR(0, 116, __pyx_L1_error)
  __pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase = &__pyx_type_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase;
  if (PyType_Ready(&__pyx_type_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase) < 0) __PYX_ERR(0, 131, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase.tp_dictoffset && __pyx_type_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_IonSeriesBase, (PyObject *)&__pyx_type_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase) < 0) __PYX_ERR(0, 131, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase) < 0) __PYX_ERR(0, 131, __pyx_L1_error)
  __pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase = &__pyx_type_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase;
  __pyx_vtabptr_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase = &__pyx_vtable_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase;
  __pyx_vtable_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase.get_series = (struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *(*)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *, int __pyx_skip_dispatch))__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_get_series;
  __pyx_vtable_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase.get_chemical_shift = (struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase *(*)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *, int __pyx_skip_dispatch))__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_get_chemical_shift;
  __pyx_vtable_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase.set_chemical_shift = (PyObject *(*)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *, struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase *, int __pyx_skip_dispatch))__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_set_chemical_shift;
  __pyx_vtable_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase.base_name = (PyObject *(*)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *, int __pyx_skip_dispatch))__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_base_name;
  __pyx_vtable_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase.get_fragment_name = (PyObject *(*)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *, int __pyx_skip_dispatch))__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_get_fragment_name;
  __pyx_vtable_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase._update_hash_name = (void (*)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *))__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase__update_hash_name;
  __pyx_vtable_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase.get_name = (PyObject *(*)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *))__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_12FragmentBase_get_name;
  if (PyType_Ready(&__pyx_type_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase) < 0) __PYX_ERR(0, 166, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase.tp_dictoffset && __pyx_type_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase.tp_dict, __pyx_vtabptr_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase) < 0) __PYX_ERR(0, 166, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_FragmentBase, (PyObject *)&__pyx_type_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase) < 0) __PYX_ERR(0, 166, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase) < 0) __PYX_ERR(0, 166, __pyx_L1_error)
  __pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase = &__pyx_type_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase;
  __pyx_vtabptr_14glycopeptidepy_2_c_9structure_8fragment_SimpleFragment = &__pyx_vtable_14glycopeptidepy_2_c_9structure_8fragment_SimpleFragment;
  __pyx_vtable_14glycopeptidepy_2_c_9structure_8fragment_SimpleFragment.__pyx_base = *__pyx_vtabptr_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase;
  __pyx_vtable_14glycopeptidepy_2_c_9structure_8fragment_SimpleFragment.__pyx_base.get_series = (struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *(*)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *, int __pyx_skip_dispatch))__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_14SimpleFragment_get_series;
  __pyx_vtable_14glycopeptidepy_2_c_9structure_8fragment_SimpleFragment.clone = (PyObject *(*)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_SimpleFragment *, int __pyx_skip_dispatch))__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_14SimpleFragment_clone;
  __pyx_type_14glycopeptidepy_2_c_9structure_8fragment_SimpleFragment.tp_base = __pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase;
  if (PyType_Ready(&__pyx_type_14glycopeptidepy_2_c_9structure_8fragment_SimpleFragment) < 0) __PYX_ERR(0, 575, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_14glycopeptidepy_2_c_9structure_8fragment_SimpleFragment.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_14glycopeptidepy_2_c_9structure_8fragment_SimpleFragment.tp_dictoffset && __pyx_type_14glycopeptidepy_2_c_9structure_8fragment_SimpleFragment.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_14glycopeptidepy_2_c_9structure_8fragment_SimpleFragment.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_14glycopeptidepy_2_c_9structure_8fragment_SimpleFragment.tp_dict, __pyx_vtabptr_14glycopeptidepy_2_c_9structure_8fragment_SimpleFragment) < 0) __PYX_ERR(0, 575, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_SimpleFragment, (PyObject *)&__pyx_type_14glycopeptidepy_2_c_9structure_8fragment_SimpleFragment) < 0) __PYX_ERR(0, 575, __pyx_L1_error)
  __pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment_SimpleFragment = &__pyx_type_14glycopeptidepy_2_c_9structure_8fragment_SimpleFragment;
  __pyx_vtabptr_14glycopeptidepy_2_c_9structure_8fragment_StubFragment = &__pyx_vtable_14glycopeptidepy_2_c_9structure_8fragment_StubFragment;
  __pyx_vtable_14glycopeptidepy_2_c_9structure_8fragment_StubFragment.__pyx_base = *__pyx_vtabptr_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase;
  __pyx_vtable_14glycopeptidepy_2_c_9structure_8fragment_StubFragment.__pyx_base.get_series = (struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *(*)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *, int __pyx_skip_dispatch))__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_get_series;
  __pyx_vtable_14glycopeptidepy_2_c_9structure_8fragment_StubFragment.__pyx_base.base_name = (PyObject *(*)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *, int __pyx_skip_dispatch))__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_base_name;
  __pyx_vtable_14glycopeptidepy_2_c_9structure_8fragment_StubFragment.clone = (PyObject *(*)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_StubFragment *, int __pyx_skip_dispatch))__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_clone;
  __pyx_vtable_14glycopeptidepy_2_c_9structure_8fragment_StubFragment._create = (struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_StubFragment *(*)(PyObject *, double, struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *, struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition *, struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_ChemicalShiftBase *, int, PyObject *, int, struct __pyx_opt_args_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment__create *__pyx_optional_args))__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment__create;
  __pyx_vtable_14glycopeptidepy_2_c_9structure_8fragment_StubFragment.get_glycosylation_size = (int (*)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_StubFragment *))__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_12StubFragment_get_glycosylation_size;
  __pyx_type_14glycopeptidepy_2_c_9structure_8fragment_StubFragment.tp_base = __pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase;
  if (PyType_Ready(&__pyx_type_14glycopeptidepy_2_c_9structure_8fragment_StubFragment) < 0) __PYX_ERR(0, 700, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_14glycopeptidepy_2_c_9structure_8fragment_StubFragment.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_14glycopeptidepy_2_c_9structure_8fragment_StubFragment.tp_dictoffset && __pyx_type_14glycopeptidepy_2_c_9structure_8fragment_StubFragment.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_14glycopeptidepy_2_c_9structure_8fragment_StubFragment.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_14glycopeptidepy_2_c_9structure_8fragment_StubFragment.tp_dict, __pyx_vtabptr_14glycopeptidepy_2_c_9structure_8fragment_StubFragment) < 0) __PYX_ERR(0, 700, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_StubFragment, (PyObject *)&__pyx_type_14glycopeptidepy_2_c_9structure_8fragment_StubFragment) < 0) __PYX_ERR(0, 700, __pyx_L1_error)
  __pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment_StubFragment = &__pyx_type_14glycopeptidepy_2_c_9structure_8fragment_StubFragment;
  __pyx_vtabptr_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment = &__pyx_vtable_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment;
  __pyx_vtable_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment.__pyx_base = *__pyx_vtabptr_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase;
  __pyx_vtable_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment.__pyx_base.get_series = (struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *(*)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *, int __pyx_skip_dispatch))__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_get_series;
  __pyx_vtable_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment.__pyx_base.base_name = (PyObject *(*)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *, int __pyx_skip_dispatch))__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_base_name;
  __pyx_vtable_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment.__pyx_base.get_fragment_name = (PyObject *(*)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase *, int __pyx_skip_dispatch))__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_get_fragment_name;
  __pyx_vtable_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment._create = (struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *(*)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_IonSeriesBase *, int, struct __pyx_obj_14glycopeptidepy_2_c_11count_table_CountTable *, double, struct __pyx_opt_args_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment__create *__pyx_optional_args))__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment__create;
  __pyx_vtable_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment.clone = (PyObject *(*)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *, int __pyx_skip_dispatch))__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_clone;
  __pyx_vtable_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment._update_mass_with_modifications = (void (*)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *))__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment__update_mass_with_modifications;
  __pyx_vtable_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment._is_glycosylated = (int (*)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *))__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment__is_glycosylated;
  __pyx_vtable_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment.get_glycosylation_size = (long (*)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment *))__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_15PeptideFragment_get_glycosylation_size;
  __pyx_type_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment.tp_base = __pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase;
  if (PyType_Ready(&__pyx_type_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment) < 0) __PYX_ERR(0, 285, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment.tp_dictoffset && __pyx_type_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment.tp_dict, __pyx_vtabptr_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment) < 0) __PYX_ERR(0, 285, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_PeptideFragment, (PyObject *)&__pyx_type_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment) < 0) __PYX_ERR(0, 285, __pyx_L1_error)
  __pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment = &__pyx_type_14glycopeptidepy_2_c_9structure_8fragment_PeptideFragment;
  __pyx_vtabptr_14glycopeptidepy_2_c_9structure_8fragment__NameTree = &__pyx_vtable_14glycopeptidepy_2_c_9structure_8fragment__NameTree;
  __pyx_vtable_14glycopeptidepy_2_c_9structure_8fragment__NameTree._create = (struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment__NameTree *(*)(void))__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree__create;
  __pyx_vtable_14glycopeptidepy_2_c_9structure_8fragment__NameTree.get = (struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment__NameTree *(*)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment__NameTree *, PyObject *, int __pyx_skip_dispatch))__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_get;
  __pyx_vtable_14glycopeptidepy_2_c_9structure_8fragment__NameTree.__pyx_fuse_0traverse = (PyObject *(*)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment__NameTree *, PyObject *, int __pyx_skip_dispatch))__pyx_fuse_0__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_traverse;
  __pyx_vtable_14glycopeptidepy_2_c_9structure_8fragment__NameTree.__pyx_fuse_1traverse = (PyObject *(*)(struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment__NameTree *, PyObject *, int __pyx_skip_dispatch))__pyx_fuse_1__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_traverse;
  if (PyType_Ready(&__pyx_type_14glycopeptidepy_2_c_9structure_8fragment__NameTree) < 0) __PYX_ERR(0, 614, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_14glycopeptidepy_2_c_9structure_8fragment__NameTree.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_14glycopeptidepy_2_c_9structure_8fragment__NameTree.tp_dictoffset && __pyx_type_14glycopeptidepy_2_c_9structure_8fragment__NameTree.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_14glycopeptidepy_2_c_9structure_8fragment__NameTree.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  if (__Pyx_SetVtable(__pyx_type_14glycopeptidepy_2_c_9structure_8fragment__NameTree.tp_dict, __pyx_vtabptr_14glycopeptidepy_2_c_9structure_8fragment__NameTree) < 0) __PYX_ERR(0, 614, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_NameTree, (PyObject *)&__pyx_type_14glycopeptidepy_2_c_9structure_8fragment__NameTree) < 0) __PYX_ERR(0, 614, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_14glycopeptidepy_2_c_9structure_8fragment__NameTree) < 0) __PYX_ERR(0, 614, __pyx_L1_error)
  __pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment__NameTree = &__pyx_type_14glycopeptidepy_2_c_9structure_8fragment__NameTree;
  if (PyType_Ready(&__pyx_type_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct____pyx_fuse_0__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition) < 0) __PYX_ERR(0, 675, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct____pyx_fuse_0__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct____pyx_fuse_0__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition.tp_dictoffset && __pyx_type_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct____pyx_fuse_0__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct____pyx_fuse_0__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct____pyx_fuse_0__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition = &__pyx_type_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct____pyx_fuse_0__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition;
  if (PyType_Ready(&__pyx_type_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_1_genexpr) < 0) __PYX_ERR(0, 694, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_1_genexpr.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_1_genexpr.tp_dictoffset && __pyx_type_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_1_genexpr.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_1_genexpr.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_1_genexpr = &__pyx_type_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_1_genexpr;
  if (PyType_Ready(&__pyx_type_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_2___pyx_fuse_1__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition) < 0) __PYX_ERR(0, 675, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_2___pyx_fuse_1__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_2___pyx_fuse_1__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition.tp_dictoffset && __pyx_type_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_2___pyx_fuse_1__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_2___pyx_fuse_1__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_2___pyx_fuse_1__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition = &__pyx_type_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_2___pyx_fuse_1__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition;
  if (PyType_Ready(&__pyx_type_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_3_genexpr) < 0) __PYX_ERR(0, 694, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_3_genexpr.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_3_genexpr.tp_dictoffset && __pyx_type_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_3_genexpr.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_3_genexpr.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_3_genexpr = &__pyx_type_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_3_genexpr;
  if (PyType_Ready(&__pyx_type_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_4___pyx_fuse_2__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition) < 0) __PYX_ERR(0, 675, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_4___pyx_fuse_2__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_4___pyx_fuse_2__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition.tp_dictoffset && __pyx_type_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_4___pyx_fuse_2__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_4___pyx_fuse_2__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_4___pyx_fuse_2__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition = &__pyx_type_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_4___pyx_fuse_2__pyx_f_14glycopeptidepy_2_c_9structure_8fragment_build_name_from_composition;
  if (PyType_Ready(&__pyx_type_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_5_genexpr) < 0) __PYX_ERR(0, 694, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_5_genexpr.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_5_genexpr.tp_dictoffset && __pyx_type_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_5_genexpr.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_5_genexpr.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_5_genexpr = &__pyx_type_14glycopeptidepy_2_c_9structure_8fragment___pyx_scope_struct_5_genexpr;
  if (PyType_Ready(&__pyx_scope_struct____Pyx_CFunc_object____tuple___to_py) < 0) __PYX_ERR(2, 64, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_scope_struct____Pyx_CFunc_object____tuple___to_py.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_scope_struct____Pyx_CFunc_object____tuple___to_py.tp_dictoffset && __pyx_scope_struct____Pyx_CFunc_object____tuple___to_py.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_scope_struct____Pyx_CFunc_object____tuple___to_py.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype___pyx_scope_struct____Pyx_CFunc_object____tuple___to_py = &__pyx_scope_struct____Pyx_CFunc_object____tuple___to_py;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_import_code", 0);
  /*--- Type import code ---*/
  __pyx_t_1 = PyImport_ImportModule("glypy.composition.ccomposition"); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_5glypy_11composition_12ccomposition_CComposition = __Pyx_ImportType(__pyx_t_1, "glypy.composition.ccomposition", "CComposition", sizeof(struct __pyx_obj_5glypy_11composition_12ccomposition_CComposition), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5glypy_11composition_12ccomposition_CComposition) __PYX_ERR(3, 2, __pyx_L1_error)
  __pyx_vtabptr_5glypy_11composition_12ccomposition_CComposition = (struct __pyx_vtabstruct_5glypy_11composition_12ccomposition_CComposition*)__Pyx_GetVtable(__pyx_ptype_5glypy_11composition_12ccomposition_CComposition->tp_dict); if (unlikely(!__pyx_vtabptr_5glypy_11composition_12ccomposition_CComposition)) __PYX_ERR(3, 2, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7cpython_4type_type = __Pyx_ImportType(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "type", 
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyTypeObject),
  #else
  sizeof(PyHeapTypeObject),
  #endif
  __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7cpython_4type_type) __PYX_ERR(4, 9, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 8, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7cpython_4bool_bool = __Pyx_ImportType(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "bool", sizeof(PyBoolObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7cpython_4bool_bool) __PYX_ERR(5, 8, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 15, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7cpython_7complex_complex = __Pyx_ImportType(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "complex", sizeof(PyComplexObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7cpython_7complex_complex) __PYX_ERR(6, 15, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("glycopeptidepy._c.count_table"); if (unlikely(!__pyx_t_1)) __PYX_ERR(7, 52, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_14glycopeptidepy_2_c_11count_table_CountTableIterator = __Pyx_ImportType(__pyx_t_1, "glycopeptidepy._c.count_table", "CountTableIterator", sizeof(struct __pyx_obj_14glycopeptidepy_2_c_11count_table_CountTableIterator), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_14glycopeptidepy_2_c_11count_table_CountTableIterator) __PYX_ERR(7, 52, __pyx_L1_error)
  __pyx_vtabptr_14glycopeptidepy_2_c_11count_table_CountTableIterator = (struct __pyx_vtabstruct_14glycopeptidepy_2_c_11count_table_CountTableIterator*)__Pyx_GetVtable(__pyx_ptype_14glycopeptidepy_2_c_11count_table_CountTableIterator->tp_dict); if (unlikely(!__pyx_vtabptr_14glycopeptidepy_2_c_11count_table_CountTableIterator)) __PYX_ERR(7, 52, __pyx_L1_error)
  __pyx_ptype_14glycopeptidepy_2_c_11count_table_CountTable = __Pyx_ImportType(__pyx_t_1, "glycopeptidepy._c.count_table", "CountTable", sizeof(struct __pyx_obj_14glycopeptidepy_2_c_11count_table_CountTable), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_14glycopeptidepy_2_c_11count_table_CountTable) __PYX_ERR(7, 69, __pyx_L1_error)
  __pyx_vtabptr_14glycopeptidepy_2_c_11count_table_CountTable = (struct __pyx_vtabstruct_14glycopeptidepy_2_c_11count_table_CountTable*)__Pyx_GetVtable(__pyx_ptype_14glycopeptidepy_2_c_11count_table_CountTable->tp_dict); if (unlikely(!__pyx_vtabptr_14glycopeptidepy_2_c_11count_table_CountTable)) __PYX_ERR(7, 69, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("glycopeptidepy._c.structure.base"); if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_14glycopeptidepy_2_c_9structure_4base_PeptideSequenceBase = __Pyx_ImportType(__pyx_t_1, "glycopeptidepy._c.structure.base", "PeptideSequenceBase", sizeof(struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_PeptideSequenceBase), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_14glycopeptidepy_2_c_9structure_4base_PeptideSequenceBase) __PYX_ERR(8, 5, __pyx_L1_error)
  __pyx_vtabptr_14glycopeptidepy_2_c_9structure_4base_PeptideSequenceBase = (struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_4base_PeptideSequenceBase*)__Pyx_GetVtable(__pyx_ptype_14glycopeptidepy_2_c_9structure_4base_PeptideSequenceBase->tp_dict); if (unlikely(!__pyx_vtabptr_14glycopeptidepy_2_c_9structure_4base_PeptideSequenceBase)) __PYX_ERR(8, 5, __pyx_L1_error)
  __pyx_ptype_14glycopeptidepy_2_c_9structure_4base_TerminalGroup = __Pyx_ImportType(__pyx_t_1, "glycopeptidepy._c.structure.base", "TerminalGroup", sizeof(struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_TerminalGroup), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_14glycopeptidepy_2_c_9structure_4base_TerminalGroup) __PYX_ERR(8, 21, __pyx_L1_error)
  __pyx_vtabptr_14glycopeptidepy_2_c_9structure_4base_TerminalGroup = (struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_4base_TerminalGroup*)__Pyx_GetVtable(__pyx_ptype_14glycopeptidepy_2_c_9structure_4base_TerminalGroup->tp_dict); if (unlikely(!__pyx_vtabptr_14glycopeptidepy_2_c_9structure_4base_TerminalGroup)) __PYX_ERR(8, 21, __pyx_L1_error)
  __pyx_ptype_14glycopeptidepy_2_c_9structure_4base_AminoAcidResidueBase = __Pyx_ImportType(__pyx_t_1, "glycopeptidepy._c.structure.base", "AminoAcidResidueBase", sizeof(struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_AminoAcidResidueBase), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_14glycopeptidepy_2_c_9structure_4base_AminoAcidResidueBase) __PYX_ERR(8, 35, __pyx_L1_error)
  __pyx_vtabptr_14glycopeptidepy_2_c_9structure_4base_AminoAcidResidueBase = (struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_4base_AminoAcidResidueBase*)__Pyx_GetVtable(__pyx_ptype_14glycopeptidepy_2_c_9structure_4base_AminoAcidResidueBase->tp_dict); if (unlikely(!__pyx_vtabptr_14glycopeptidepy_2_c_9structure_4base_AminoAcidResidueBase)) __PYX_ERR(8, 35, __pyx_L1_error)
  __pyx_ptype_14glycopeptidepy_2_c_9structure_4base_ModificationBase = __Pyx_ImportType(__pyx_t_1, "glycopeptidepy._c.structure.base", "ModificationBase", sizeof(struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_ModificationBase), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_14glycopeptidepy_2_c_9structure_4base_ModificationBase) __PYX_ERR(8, 50, __pyx_L1_error)
  __pyx_vtabptr_14glycopeptidepy_2_c_9structure_4base_ModificationBase = (struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_4base_ModificationBase*)__Pyx_GetVtable(__pyx_ptype_14glycopeptidepy_2_c_9structure_4base_ModificationBase->tp_dict); if (unlikely(!__pyx_vtabptr_14glycopeptidepy_2_c_9structure_4base_ModificationBase)) __PYX_ERR(8, 50, __pyx_L1_error)
  __pyx_ptype_14glycopeptidepy_2_c_9structure_4base_SequencePosition = __Pyx_ImportType(__pyx_t_1, "glycopeptidepy._c.structure.base", "SequencePosition", sizeof(struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_SequencePosition), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_14glycopeptidepy_2_c_9structure_4base_SequencePosition) __PYX_ERR(8, 82, __pyx_L1_error)
  __pyx_vtabptr_14glycopeptidepy_2_c_9structure_4base_SequencePosition = (struct __pyx_vtabstruct_14glycopeptidepy_2_c_9structure_4base_SequencePosition*)__Pyx_GetVtable(__pyx_ptype_14glycopeptidepy_2_c_9structure_4base_SequencePosition->tp_dict); if (unlikely(!__pyx_vtabptr_14glycopeptidepy_2_c_9structure_4base_SequencePosition)) __PYX_ERR(8, 82, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_variable_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_import_code", 0);
  /*--- Variable import code ---*/
  __pyx_t_1 = PyImport_ImportModule("glypy.composition.ccomposition"); if (!__pyx_t_1) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_ImportVoidPtr(__pyx_t_1, "_atom", (void **)&__pyx_vp_5glypy_11composition_12ccomposition__atom, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr(__pyx_t_1, "_formula", (void **)&__pyx_vp_5glypy_11composition_12ccomposition__formula, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr(__pyx_t_1, "_isotope_string", (void **)&__pyx_vp_5glypy_11composition_12ccomposition__isotope_string, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr(__pyx_t_1, "isotope_pattern", (void **)&__pyx_vp_5glypy_11composition_12ccomposition_isotope_pattern, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportVoidPtr(__pyx_t_1, "formula_pattern", (void **)&__pyx_vp_5glypy_11composition_12ccomposition_formula_pattern, "PyObject *") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_function_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_import_code", 0);
  /*--- Function import code ---*/
  __pyx_t_1 = PyImport_ImportModule("glycopeptidepy._c.count_table"); if (!__pyx_t_1) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_ImportFunction(__pyx_t_1, "count_table_items", (void (**)(void))&__pyx_f_14glycopeptidepy_2_c_11count_table_count_table_items, "PyObject *(struct __pyx_t_14glycopeptidepy_2_c_11count_table_count_table *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}


#ifndef CYTHON_NO_PYINIT_EXPORT
#define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#elif PY_MAJOR_VERSION < 3
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" void
#else
#define __Pyx_PyMODINIT_FUNC void
#endif
#else
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" PyObject *
#else
#define __Pyx_PyMODINIT_FUNC PyObject *
#endif
#endif


#if PY_MAJOR_VERSION < 3
__Pyx_PyMODINIT_FUNC initfragment(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC initfragment(void)
#else
__Pyx_PyMODINIT_FUNC PyInit_fragment(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC PyInit_fragment(void)
#if CYTHON_PEP489_MULTI_PHASE_INIT
{
  return PyModuleDef_Init(&__pyx_moduledef);
}
static CYTHON_SMALL_CODE int __Pyx_check_single_interpreter(void) {
    #if PY_VERSION_HEX >= 0x030700A1
    static PY_INT64_T main_interpreter_id = -1;
    PY_INT64_T current_id = PyInterpreterState_GetID(PyThreadState_Get()->interp);
    if (main_interpreter_id == -1) {
        main_interpreter_id = current_id;
        return (unlikely(current_id == -1)) ? -1 : 0;
    } else if (unlikely(main_interpreter_id != current_id))
    #else
    static PyInterpreterState *main_interpreter = NULL;
    PyInterpreterState *current_interpreter = PyThreadState_Get()->interp;
    if (!main_interpreter) {
        main_interpreter = current_interpreter;
    } else if (unlikely(main_interpreter != current_interpreter))
    #endif
    {
        PyErr_SetString(
            PyExc_ImportError,
            "Interpreter change detected - this module can only be loaded into one interpreter per process.");
        return -1;
    }
    return 0;
}
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *moddict, const char* from_name, const char* to_name, int allow_none) {
    PyObject *value = PyObject_GetAttrString(spec, from_name);
    int result = 0;
    if (likely(value)) {
        if (allow_none || value != Py_None) {
            result = PyDict_SetItemString(moddict, to_name, value);
        }
        Py_DECREF(value);
    } else if (PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Clear();
    } else {
        result = -1;
    }
    return result;
}
static CYTHON_SMALL_CODE PyObject* __pyx_pymod_create(PyObject *spec, CYTHON_UNUSED PyModuleDef *def) {
    PyObject *module = NULL, *moddict, *modname;
    if (__Pyx_check_single_interpreter())
        return NULL;
    if (__pyx_m)
        return __Pyx_NewRef(__pyx_m);
    modname = PyObject_GetAttrString(spec, "name");
    if (unlikely(!modname)) goto bad;
    module = PyModule_NewObject(modname);
    Py_DECREF(modname);
    if (unlikely(!module)) goto bad;
    moddict = PyModule_GetDict(module);
    if (unlikely(!moddict)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "loader", "__loader__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "origin", "__file__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "parent", "__package__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "submodule_search_locations", "__path__", 0) < 0)) goto bad;
    return module;
bad:
    Py_XDECREF(module);
    return NULL;
}


static CYTHON_SMALL_CODE int __pyx_pymod_exec_fragment(PyObject *__pyx_pyinit_module)
#endif
#endif
{
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  if (__pyx_m) {
    if (__pyx_m == __pyx_pyinit_module) return 0;
    PyErr_SetString(PyExc_RuntimeError, "Module 'fragment' has already been imported. Re-initialisation is not supported.");
    return -1;
  }
  #elif PY_MAJOR_VERSION >= 3
  if (__pyx_m) return __Pyx_NewRef(__pyx_m);
  #endif
  #if CYTHON_REFNANNY
__Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
if (!__Pyx_RefNanny) {
  PyErr_Clear();
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
  if (!__Pyx_RefNanny)
      Py_FatalError("failed to import 'refnanny' module");
}
#endif
  __Pyx_RefNannySetupContext("__Pyx_PyMODINIT_FUNC PyInit_fragment(void)", 0);
  if (__Pyx_check_binary_version() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pxy_PyFrame_Initialize_Offsets
  __Pxy_PyFrame_Initialize_Offsets();
  #endif
  __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_unicode)) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pyx_CyFunction_USED
  if (__pyx_CyFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Coroutine_USED
  if (__pyx_Coroutine_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_AsyncGen_USED
  if (__pyx_AsyncGen_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_StopAsyncIteration_USED
  if (__pyx_StopAsyncIteration_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  /*--- Library function declarations ---*/
  /*--- Threads initialization code ---*/
  #if defined(WITH_THREAD) && PY_VERSION_HEX < 0x030700F0 && defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
  PyEval_InitThreads();
  #endif
  /*--- Module creation code ---*/
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __pyx_m = __pyx_pyinit_module;
  Py_INCREF(__pyx_m);
  #else
  #if PY_MAJOR_VERSION < 3
  __pyx_m = Py_InitModule4("fragment", __pyx_methods, 0, 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
  #else
  __pyx_m = PyModule_Create(&__pyx_moduledef);
  #endif
  if (unlikely(!__pyx_m)) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  __pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_d)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_d);
  __pyx_b = PyImport_AddModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_b)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_b);
  __pyx_cython_runtime = PyImport_AddModule((char *) "cython_runtime"); if (unlikely(!__pyx_cython_runtime)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_cython_runtime);
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitGlobals() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3 && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
  if (__Pyx_init_sys_getdefaultencoding_params() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  if (__pyx_module_is_main_glycopeptidepy___c__structure__fragment) {
    if (PyObject_SetAttr(__pyx_m, __pyx_n_s_name, __pyx_n_s_main) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  }
  #if PY_MAJOR_VERSION >= 3
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(0, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "glycopeptidepy._c.structure.fragment")) {
      if (unlikely(PyDict_SetItemString(modules, "glycopeptidepy._c.structure.fragment", __pyx_m) < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  #endif
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Global type/function init code ---*/
  (void)__Pyx_modinit_global_init_code();
  (void)__Pyx_modinit_variable_export_code();
  if (unlikely(__Pyx_modinit_function_export_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (unlikely(__Pyx_modinit_type_init_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (unlikely(__Pyx_modinit_type_import_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (unlikely(__Pyx_modinit_variable_import_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (unlikely(__Pyx_modinit_function_import_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Execution code ---*/
  #if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
  if (__Pyx_patch_abc() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif

  /* "glycopeptidepy/_c/structure/fragment.pyx":27
 * 
 * from glycopeptidepy.structure.modification import (
 *     Modification, NGlycanCoreGlycosylation, OGlycanCoreGlycosylation,             # <<<<<<<<<<<<<<
 *     GlycosaminoglycanLinkerGlycosylation, ModificationCategory)
 * 
 */
  __pyx_t_1 = PyList_New(5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_Modification);
  __Pyx_GIVEREF(__pyx_n_s_Modification);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_Modification);
  __Pyx_INCREF(__pyx_n_s_NGlycanCoreGlycosylation);
  __Pyx_GIVEREF(__pyx_n_s_NGlycanCoreGlycosylation);
  PyList_SET_ITEM(__pyx_t_1, 1, __pyx_n_s_NGlycanCoreGlycosylation);
  __Pyx_INCREF(__pyx_n_s_OGlycanCoreGlycosylation);
  __Pyx_GIVEREF(__pyx_n_s_OGlycanCoreGlycosylation);
  PyList_SET_ITEM(__pyx_t_1, 2, __pyx_n_s_OGlycanCoreGlycosylation);
  __Pyx_INCREF(__pyx_n_s_GlycosaminoglycanLinkerGlycosyla);
  __Pyx_GIVEREF(__pyx_n_s_GlycosaminoglycanLinkerGlycosyla);
  PyList_SET_ITEM(__pyx_t_1, 3, __pyx_n_s_GlycosaminoglycanLinkerGlycosyla);
  __Pyx_INCREF(__pyx_n_s_ModificationCategory);
  __Pyx_GIVEREF(__pyx_n_s_ModificationCategory);
  PyList_SET_ITEM(__pyx_t_1, 4, __pyx_n_s_ModificationCategory);

  /* "glycopeptidepy/_c/structure/fragment.pyx":26
 * from glycopeptidepy._c.count_table cimport CountTable, CountTableIterator, count_table_items
 * 
 * from glycopeptidepy.structure.modification import (             # <<<<<<<<<<<<<<
 *     Modification, NGlycanCoreGlycosylation, OGlycanCoreGlycosylation,
 *     GlycosaminoglycanLinkerGlycosylation, ModificationCategory)
 */
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_glycopeptidepy_structure_modific, __pyx_t_1, -1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_Modification); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_Modification, __pyx_t_1) < 0) __PYX_ERR(0, 27, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_NGlycanCoreGlycosylation); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_NGlycanCoreGlycosylation, __pyx_t_1) < 0) __PYX_ERR(0, 27, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_OGlycanCoreGlycosylation); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_OGlycanCoreGlycosylation, __pyx_t_1) < 0) __PYX_ERR(0, 27, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_GlycosaminoglycanLinkerGlycosyla); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_GlycosaminoglycanLinkerGlycosyla, __pyx_t_1) < 0) __PYX_ERR(0, 28, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_ModificationCategory); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ModificationCategory, __pyx_t_1) < 0) __PYX_ERR(0, 28, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "glycopeptidepy/_c/structure/fragment.pyx":30
 *     GlycosaminoglycanLinkerGlycosylation, ModificationCategory)
 * 
 * cdef ModificationBase _n_glycosylation = NGlycanCoreGlycosylation()             # <<<<<<<<<<<<<<
 * cdef ModificationBase _o_glycosylation = OGlycanCoreGlycosylation()
 * cdef ModificationBase _gag_linker_glycosylation = GlycosaminoglycanLinkerGlycosylation()
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_NGlycanCoreGlycosylation); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_14glycopeptidepy_2_c_9structure_4base_ModificationBase))))) __PYX_ERR(0, 30, __pyx_L1_error)
  __Pyx_XGOTREF(((PyObject *)__pyx_v_14glycopeptidepy_2_c_9structure_8fragment__n_glycosylation));
  __Pyx_DECREF_SET(__pyx_v_14glycopeptidepy_2_c_9structure_8fragment__n_glycosylation, ((struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_ModificationBase *)__pyx_t_1));
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "glycopeptidepy/_c/structure/fragment.pyx":31
 * 
 * cdef ModificationBase _n_glycosylation = NGlycanCoreGlycosylation()
 * cdef ModificationBase _o_glycosylation = OGlycanCoreGlycosylation()             # <<<<<<<<<<<<<<
 * cdef ModificationBase _gag_linker_glycosylation = GlycosaminoglycanLinkerGlycosylation()
 * cdef ModificationBase _modification_hexnac = Modification("HexNAc").rule
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_OGlycanCoreGlycosylation); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_CallNoArg(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_14glycopeptidepy_2_c_9structure_4base_ModificationBase))))) __PYX_ERR(0, 31, __pyx_L1_error)
  __Pyx_XGOTREF(((PyObject *)__pyx_v_14glycopeptidepy_2_c_9structure_8fragment__o_glycosylation));
  __Pyx_DECREF_SET(__pyx_v_14glycopeptidepy_2_c_9structure_8fragment__o_glycosylation, ((struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_ModificationBase *)__pyx_t_2));
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "glycopeptidepy/_c/structure/fragment.pyx":32
 * cdef ModificationBase _n_glycosylation = NGlycanCoreGlycosylation()
 * cdef ModificationBase _o_glycosylation = OGlycanCoreGlycosylation()
 * cdef ModificationBase _gag_linker_glycosylation = GlycosaminoglycanLinkerGlycosylation()             # <<<<<<<<<<<<<<
 * cdef ModificationBase _modification_hexnac = Modification("HexNAc").rule
 * cdef ModificationBase _modification_xylose = Modification("Xyl").rule
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_GlycosaminoglycanLinkerGlycosyla); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 32, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 32, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_14glycopeptidepy_2_c_9structure_4base_ModificationBase))))) __PYX_ERR(0, 32, __pyx_L1_error)
  __Pyx_XGOTREF(((PyObject *)__pyx_v_14glycopeptidepy_2_c_9structure_8fragment__gag_linker_glycosylation));
  __Pyx_DECREF_SET(__pyx_v_14glycopeptidepy_2_c_9structure_8fragment__gag_linker_glycosylation, ((struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_ModificationBase *)__pyx_t_1));
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "glycopeptidepy/_c/structure/fragment.pyx":33
 * cdef ModificationBase _o_glycosylation = OGlycanCoreGlycosylation()
 * cdef ModificationBase _gag_linker_glycosylation = GlycosaminoglycanLinkerGlycosylation()
 * cdef ModificationBase _modification_hexnac = Modification("HexNAc").rule             # <<<<<<<<<<<<<<
 * cdef ModificationBase _modification_xylose = Modification("Xyl").rule
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_Modification); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 33, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__13, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 33, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_rule); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 33, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_14glycopeptidepy_2_c_9structure_4base_ModificationBase))))) __PYX_ERR(0, 33, __pyx_L1_error)
  __Pyx_XGOTREF(((PyObject *)__pyx_v_14glycopeptidepy_2_c_9structure_8fragment__modification_hexnac));
  __Pyx_DECREF_SET(__pyx_v_14glycopeptidepy_2_c_9structure_8fragment__modification_hexnac, ((struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_ModificationBase *)__pyx_t_1));
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "glycopeptidepy/_c/structure/fragment.pyx":34
 * cdef ModificationBase _gag_linker_glycosylation = GlycosaminoglycanLinkerGlycosylation()
 * cdef ModificationBase _modification_hexnac = Modification("HexNAc").rule
 * cdef ModificationBase _modification_xylose = Modification("Xyl").rule             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_Modification); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 34, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__14, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 34, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_rule); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 34, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_14glycopeptidepy_2_c_9structure_4base_ModificationBase))))) __PYX_ERR(0, 34, __pyx_L1_error)
  __Pyx_XGOTREF(((PyObject *)__pyx_v_14glycopeptidepy_2_c_9structure_8fragment__modification_xylose));
  __Pyx_DECREF_SET(__pyx_v_14glycopeptidepy_2_c_9structure_8fragment__modification_xylose, ((struct __pyx_obj_14glycopeptidepy_2_c_9structure_4base_ModificationBase *)__pyx_t_1));
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "glycopeptidepy/_c/structure/fragment.pyx":45
 * 
 * cdef string_cell[ARRAY_SIZE] str_ints
 * cdef dict str_int_overflow_intern_cache = {}             # <<<<<<<<<<<<<<
 * 
 * cdef void init_str_ints():
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 45, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_XGOTREF(__pyx_v_14glycopeptidepy_2_c_9structure_8fragment_str_int_overflow_intern_cache);
  __Pyx_DECREF_SET(__pyx_v_14glycopeptidepy_2_c_9structure_8fragment_str_int_overflow_intern_cache, ((PyObject*)__pyx_t_1));
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "glycopeptidepy/_c/structure/fragment.pyx":64
 * 
 * 
 * init_str_ints()             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_f_14glycopeptidepy_2_c_9structure_8fragment_init_str_ints();

  /* "glycopeptidepy/_c/structure/fragment.pyx":100
 * 
 * 
 * def _debug_get_str_int(int i):             # <<<<<<<<<<<<<<
 *     cdef:
 *         string_cell out
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_14glycopeptidepy_2_c_9structure_8fragment_1_debug_get_str_int, NULL, __pyx_n_s_glycopeptidepy__c_structure_frag); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 100, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_debug_get_str_int, __pyx_t_1) < 0) __PYX_ERR(0, 100, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "glycopeptidepy/_c/structure/fragment.pyx":109
 *     print(out.size)
 * 
 * def _get_string_cell(int i):             # <<<<<<<<<<<<<<
 *     cdef string_cell cell = str_ints[i]
 *     print(cell.size)
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_14glycopeptidepy_2_c_9structure_8fragment_3_get_string_cell, NULL, __pyx_n_s_glycopeptidepy__c_structure_frag); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 109, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_get_string_cell, __pyx_t_1) < 0) __PYX_ERR(0, 109, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "glycopeptidepy/_c/structure/fragment.pyx":255
 *         self._hash = hash(self._name)
 * 
 *     chemical_shift = property(get_chemical_shift, set_chemical_shift)             # <<<<<<<<<<<<<<
 * 
 *     cpdef str get_fragment_name(self):
 */
  __Pyx_GetNameInClass(__pyx_t_1, (PyObject *)__pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase, __pyx_n_s_get_chemical_shift); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 255, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GetNameInClass(__pyx_t_2, (PyObject *)__pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase, __pyx_n_s_set_chemical_shift); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 255, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 255, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_2);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_property, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 255, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (PyDict_SetItem((PyObject *)__pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase->tp_dict, __pyx_n_s_chemical_shift, __pyx_t_2) < 0) __PYX_ERR(0, 255, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment_FragmentBase);

  /* "glycopeptidepy/_c/structure/fragment.pyx":282
 * 
 * 
 * cdef object ModificationCategory_glycosylation = ModificationCategory.glycosylation             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_ModificationCategory); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 282, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_glycosylation); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 282, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_XGOTREF(__pyx_v_14glycopeptidepy_2_c_9structure_8fragment_ModificationCategory_glycosylation);
  __Pyx_DECREF_SET(__pyx_v_14glycopeptidepy_2_c_9structure_8fragment_ModificationCategory_glycosylation, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_t_3 = 0;

  /* "glycopeptidepy/_c/structure/fragment.pyx":646
 *             return result
 * 
 *     cpdef traverse(self, list_or_iterable parts):             # <<<<<<<<<<<<<<
 *         cdef:
 *             _NameTree node
 */
  __pyx_t_3 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 646, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __pyx_FusedFunction_New(&__pyx_fuse_0__pyx_mdef_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_9__pyx_fuse_0traverse, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_NameTree___pyx_fuse_0traverse, NULL, __pyx_n_s_glycopeptidepy__c_structure_frag, __pyx_d, ((PyObject *)__pyx_codeobj__20)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 646, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_2, __pyx_empty_tuple);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_list, __pyx_t_2) < 0) __PYX_ERR(0, 646, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __pyx_FusedFunction_New(&__pyx_fuse_1__pyx_mdef_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_11__pyx_fuse_1traverse, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_NameTree___pyx_fuse_1traverse, NULL, __pyx_n_s_glycopeptidepy__c_structure_frag, __pyx_d, ((PyObject *)__pyx_codeobj__20)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 646, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_2, __pyx_empty_tuple);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_object, __pyx_t_2) < 0) __PYX_ERR(0, 646, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __pyx_FusedFunction_New(&__pyx_mdef_14glycopeptidepy_2_c_9structure_8fragment_9_NameTree_7traverse, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_NameTree_traverse, NULL, __pyx_n_s_glycopeptidepy__c_structure_frag, __pyx_d, ((PyObject *)__pyx_codeobj__20)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 646, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_2, __pyx_empty_tuple);
  ((__pyx_FusedFunctionObject *) __pyx_t_2)->__signatures__ = __pyx_t_3;
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_t_3 = 0;
  if (PyDict_SetItem((PyObject *)__pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment__NameTree->tp_dict, __pyx_n_s_traverse, __pyx_t_2) < 0) __PYX_ERR(0, 646, __pyx_L1_error)
  PyType_Modified(__pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment__NameTree);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "glycopeptidepy/_c/structure/fragment.pyx":667
 * 
 * 
 * cdef _NameTree stub_fragment_name_cache = _NameTree()             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment__NameTree)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 667, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_XGOTREF(((PyObject *)__pyx_v_14glycopeptidepy_2_c_9structure_8fragment_stub_fragment_name_cache));
  __Pyx_DECREF_SET(__pyx_v_14glycopeptidepy_2_c_9structure_8fragment_stub_fragment_name_cache, ((struct __pyx_obj_14glycopeptidepy_2_c_9structure_8fragment__NameTree *)__pyx_t_2));
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "glycopeptidepy/_c/structure/fragment.pyx":702
 * cdef class StubFragment(FragmentBase):
 * 
 *     _name_cache = stub_fragment_name_cache             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */
  if (PyDict_SetItem((PyObject *)__pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment_StubFragment->tp_dict, __pyx_n_s_name_cache, ((PyObject *)__pyx_v_14glycopeptidepy_2_c_9structure_8fragment_stub_fragment_name_cache)) < 0) __PYX_ERR(0, 702, __pyx_L1_error)
  PyType_Modified(__pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment_StubFragment);

  /* "glycopeptidepy/_c/structure/fragment.pyx":775
 * 
 *     @classmethod
 *     def build_name_from_composition(cls, glycan_composition):             # <<<<<<<<<<<<<<
 *         return build_name_from_composition(glycan_composition)
 * 
 */
  __Pyx_GetNameInClass(__pyx_t_2, (PyObject *)__pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment_StubFragment, __pyx_n_s_build_name_from_composition); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 775, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "glycopeptidepy/_c/structure/fragment.pyx":774
 *             return self.build_name_from_composition(self.glycosylation)
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def build_name_from_composition(cls, glycan_composition):
 *         return build_name_from_composition(glycan_composition)
 */
  __pyx_t_3 = __Pyx_Method_ClassMethod(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 774, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem((PyObject *)__pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment_StubFragment->tp_dict, __pyx_n_s_build_name_from_composition, __pyx_t_3) < 0) __PYX_ERR(0, 775, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_14glycopeptidepy_2_c_9structure_8fragment_StubFragment);

  /* "(tree fragment)":1
 * def __pyx_unpickle_ChemicalShiftBase(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_3 = PyCFunction_NewEx(&__pyx_mdef_14glycopeptidepy_2_c_9structure_8fragment_5__pyx_unpickle_ChemicalShiftBase, NULL, __pyx_n_s_glycopeptidepy__c_structure_frag); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_ChemicalShiftBase, __pyx_t_3) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "(tree fragment)":11
 *         __pyx_unpickle_ChemicalShiftBase__set_state(<ChemicalShiftBase> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_ChemicalShiftBase__set_state(ChemicalShiftBase __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.composition = __pyx_state[0]; __pyx_result.mass = __pyx_state[1]; __pyx_result.name = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 */
  __pyx_t_3 = PyCFunction_NewEx(&__pyx_mdef_14glycopeptidepy_2_c_9structure_8fragment_7__pyx_unpickle_IonSeriesBase, NULL, __pyx_n_s_glycopeptidepy__c_structure_frag); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_IonSeriesBase, __pyx_t_3) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_FragmentBase(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_3 = PyCFunction_NewEx(&__pyx_mdef_14glycopeptidepy_2_c_9structure_8fragment_9__pyx_unpickle_FragmentBase, NULL, __pyx_n_s_glycopeptidepy__c_structure_frag); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_FragmentBase, __pyx_t_3) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "(tree fragment)":11
 *         __pyx_unpickle_FragmentBase__set_state(<FragmentBase> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_FragmentBase__set_state(FragmentBase __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result._chemical_shift = __pyx_state[0]; __pyx_result._hash = __pyx_state[1]; __pyx_result._name = __pyx_state[2]; __pyx_result.mass = __pyx_state[3]
 *     if len(__pyx_state) > 4 and hasattr(__pyx_result, '__dict__'):
 */
  __pyx_t_3 = PyCFunction_NewEx(&__pyx_mdef_14glycopeptidepy_2_c_9structure_8fragment_11__pyx_unpickle__NameTree, NULL, __pyx_n_s_glycopeptidepy__c_structure_frag); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle__NameTree, __pyx_t_3) < 0) __PYX_ERR(2, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "glycopeptidepy/_c/structure/fragment.pyx":1
 * # cython: embedsignature=True             # <<<<<<<<<<<<<<
 * 
 * from libc.stdlib cimport malloc, free, realloc
 */
  __pyx_t_3 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test, __pyx_t_3) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "cfunc.to_py":64
 * 
 * @cname("__Pyx_CFunc_object____tuple___to_py")
 * cdef object __Pyx_CFunc_object____tuple___to_py(object (*f)(tuple) ):             # <<<<<<<<<<<<<<
 *     def wrap(tuple x):
 *         """wrap(x: tuple)"""
 */

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  if (__pyx_m) {
    if (__pyx_d) {
      __Pyx_AddTraceback("init glycopeptidepy._c.structure.fragment", __pyx_clineno, __pyx_lineno, __pyx_filename);
    }
    Py_CLEAR(__pyx_m);
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init glycopeptidepy._c.structure.fragment");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  return (__pyx_m != NULL) ? 0 : -1;
  #elif PY_MAJOR_VERSION >= 3
  return __pyx_m;
  #else
  return;
  #endif
}

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule(modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, "RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* PyObjectGetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_getattr))
        return tp->tp_getattr(obj, PyString_AS_STRING(attr_name));
#endif
    return PyObject_GetAttr(obj, attr_name);
}
#endif

/* GetBuiltinName */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStr(__pyx_b, name);
    if (unlikely(!result)) {
        PyErr_Format(PyExc_NameError,
#if PY_MAJOR_VERSION >= 3
            "name '%U' is not defined", name);
#else
            "name '%.200s' is not defined", PyString_AS_STRING(name));
#endif
    }
    return result;
}

/* PyCFunctionFastCall */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject * __Pyx_PyCFunction_FastCall(PyObject *func_obj, PyObject **args, Py_ssize_t nargs) {
    PyCFunctionObject *func = (PyCFunctionObject*)func_obj;
    PyCFunction meth = PyCFunction_GET_FUNCTION(func);
    PyObject *self = PyCFunction_GET_SELF(func);
    int flags = PyCFunction_GET_FLAGS(func);
    assert(PyCFunction_Check(func));
    assert(METH_FASTCALL == (flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)));
    assert(nargs >= 0);
    assert(nargs == 0 || args != NULL);
    /* _PyCFunction_FastCallDict() must not be called with an exception set,
       because it may clear it (directly or indirectly) and so the
       caller loses its exception */
    assert(!PyErr_Occurred());
    if ((PY_VERSION_HEX < 0x030700A0) || unlikely(flags & METH_KEYWORDS)) {
        return (*((__Pyx_PyCFunctionFastWithKeywords)(void*)meth)) (self, args, nargs, NULL);
    } else {
        return (*((__Pyx_PyCFunctionFast)(void*)meth)) (self, args, nargs);
    }
}
#endif

/* PyFunctionFastCall */
#if CYTHON_FAST_PYCALL
static PyObject* __Pyx_PyFunction_FastCallNoKw(PyCodeObject *co, PyObject **args, Py_ssize_t na,
                                               PyObject *globals) {
    PyFrameObject *f;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject **fastlocals;
    Py_ssize_t i;
    PyObject *result;
    assert(globals != NULL);
    /* XXX Perhaps we should create a specialized
       PyFrame_New() that doesn't take locals, but does
       take builtins without sanity checking them.
       */
    assert(tstate != NULL);
    f = PyFrame_New(tstate, co, globals, NULL);
    if (f == NULL) {
        return NULL;
    }
    fastlocals = __Pyx_PyFrame_GetLocalsplus(f);
    for (i = 0; i < na; i++) {
        Py_INCREF(*args);
        fastlocals[i] = *args++;
    }
    result = PyEval_EvalFrameEx(f,0);
    ++tstate->recursion_depth;
    Py_DECREF(f);
    --tstate->recursion_depth;
    return result;
}
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs) {
    PyCodeObject *co = (PyCodeObject *)PyFunction_GET_CODE(func);
    PyObject *globals = PyFunction_GET_GLOBALS(func);
    PyObject *argdefs = PyFunction_GET_DEFAULTS(func);
    PyObject *closure;
#if PY_MAJOR_VERSION >= 3
    PyObject *kwdefs;
#endif
    PyObject *kwtuple, **k;
    PyObject **d;
    Py_ssize_t nd;
    Py_ssize_t nk;
    PyObject *result;
    assert(kwargs == NULL || PyDict_Check(kwargs));
    nk = kwargs ? PyDict_Size(kwargs) : 0;
    if (Py_EnterRecursiveCall((char*)" while calling a Python object")) {
        return NULL;
    }
    if (
#if PY_MAJOR_VERSION >= 3
            co->co_kwonlyargcount == 0 &&
#endif
            likely(kwargs == NULL || nk == 0) &&
            co->co_flags == (CO_OPTIMIZED | CO_NEWLOCALS | CO_NOFREE)) {
        if (argdefs == NULL && co->co_argcount == nargs) {
            result = __Pyx_PyFunction_FastCallNoKw(co, args, nargs, globals);
            goto done;
        }
        else if (nargs == 0 && argdefs != NULL
                 && co->co_argcount == Py_SIZE(argdefs)) {
            /* function called with no arguments, but all parameters have
               a default value: use default values as arguments .*/
            args = &PyTuple_GET_ITEM(argdefs, 0);
            result =__Pyx_PyFunction_FastCallNoKw(co, args, Py_SIZE(argdefs), globals);
            goto done;
        }
    }
    if (kwargs != NULL) {
        Py_ssize_t pos, i;
        kwtuple = PyTuple_New(2 * nk);
        if (kwtuple == NULL) {
            result = NULL;
            goto done;
        }
        k = &PyTuple_GET_ITEM(kwtuple, 0);
        pos = i = 0;
        while (PyDict_Next(kwargs, &pos, &k[i], &k[i+1])) {
            Py_INCREF(k[i]);
            Py_INCREF(k[i+1]);
            i += 2;
        }
        nk = i / 2;
    }
    else {
        kwtuple = NULL;
        k = NULL;
    }
    closure = PyFunction_GET_CLOSURE(func);
#if PY_MAJOR_VERSION >= 3
    kwdefs = PyFunction_GET_KW_DEFAULTS(func);
#endif
    if (argdefs != NULL) {
        d = &PyTuple_GET_ITEM(argdefs, 0);
        nd = Py_SIZE(argdefs);
    }
    else {
        d = NULL;
        nd = 0;
    }
#if PY_MAJOR_VERSION >= 3
    result = PyEval_EvalCodeEx((PyObject*)co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, kwdefs, closure);
#else
    result = PyEval_EvalCodeEx(co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, closure);
#endif
    Py_XDECREF(kwtuple);
done:
    Py_LeaveRecursiveCall();
    return result;
}
#endif
#endif

/* PyObjectCall */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = Py_TYPE(func)->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallMethO */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = PyCFunction_GET_FUNCTION(func);
    self = PyCFunction_GET_SELF(func);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallOneArg */
#if CYTHON_COMPILING_IN_CPYTHON
static PyObject* __Pyx__PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_New(1);
    if (unlikely(!args)) return NULL;
    Py_INCREF(arg);
    PyTuple_SET_ITEM(args, 0, arg);
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, &arg, 1);
    }
#endif
    if (likely(PyCFunction_Check(func))) {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_O)) {
            return __Pyx_PyObject_CallMethO(func, arg);
#if CYTHON_FAST_PYCCALL
        } else if (__Pyx_PyFastCFunction_Check(func)) {
            return __Pyx_PyCFunction_FastCall(func, &arg, 1);
#endif
        }
    }
    return __Pyx__PyObject_CallOneArg(func, arg);
}
#else
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_Pack(1, arg);
    if (unlikely(!args)) return NULL;
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
#endif

/* PyErrFetchRestore */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
}
#endif

/* WriteUnraisableException */
static void __Pyx_WriteUnraisable(const char *name, CYTHON_UNUSED int clineno,
                                  CYTHON_UNUSED int lineno, CYTHON_UNUSED const char *filename,
                                  int full_traceback, CYTHON_UNUSED int nogil) {
    PyObject *old_exc, *old_val, *old_tb;
    PyObject *ctx;
    __Pyx_PyThreadState_declare
#ifdef WITH_THREAD
    PyGILState_STATE state;
    if (nogil)
        state = PyGILState_Ensure();
    else state = (PyGILState_STATE)0;
#endif
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&old_exc, &old_val, &old_tb);
    if (full_traceback) {
        Py_XINCREF(old_exc);
        Py_XINCREF(old_val);
        Py_XINCREF(old_tb);
        __Pyx_ErrRestore(old_exc, old_val, old_tb);
        PyErr_PrintEx(1);
    }
    #if PY_MAJOR_VERSION < 3
    ctx = PyString_FromString(name);
    #else
    ctx = PyUnicode_FromString(name);
    #endif
    __Pyx_ErrRestore(old_exc, old_val, old_tb);
    if (!ctx) {
        PyErr_WriteUnraisable(Py_None);
    } else {
        PyErr_WriteUnraisable(ctx);
        Py_DECREF(ctx);
    }
#ifdef WITH_THREAD
    if (nogil)
        PyGILState_Release(state);
#endif
}

/* PyDictVersioning */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    return likely(dict) ? __PYX_GET_DICT_VERSION(dict) : 0;
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj) {
    PyObject **dictptr = NULL;
    Py_ssize_t offset = Py_TYPE(obj)->tp_dictoffset;
    if (offset) {
#if CYTHON_COMPILING_IN_CPYTHON
        dictptr = (likely(offset > 0)) ? (PyObject **) ((char *)obj + offset) : _PyObject_GetDictPtr(obj);
#else
        dictptr = _PyObject_GetDictPtr(obj);
#endif
    }
    return (dictptr && *dictptr) ? __PYX_GET_DICT_VERSION(*dictptr) : 0;
}
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    if (unlikely(!dict) || unlikely(tp_dict_version != __PYX_GET_DICT_VERSION(dict)))
        return 0;
    return obj_dict_version == __Pyx_get_object_dict_version(obj);
}
#endif

/* PyObjectCallNoArg */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, NULL, 0);
    }
#endif
#if defined(__Pyx_CyFunction_USED) && defined(NDEBUG)
    if (likely(PyCFunction_Check(func) || __Pyx_CyFunction_Check(func)))
#else
    if (likely(PyCFunction_Check(func)))
#endif
    {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_NOARGS)) {
            return __Pyx_PyObject_CallMethO(func, NULL);
        }
    }
    return __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL);
}
#endif

/* ExtTypeTest */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type) {
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    if (likely(__Pyx_TypeCheck(obj, type)))
        return 1;
    PyErr_Format(PyExc_TypeError, "Cannot convert %.200s to %.200s",
                 Py_TYPE(obj)->tp_name, type->tp_name);
    return 0;
}

/* PyErrExceptionMatches */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx_PyErr_ExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        if (__Pyx_PyErr_GivenExceptionMatches(exc_type, PyTuple_GET_ITEM(tuple, i))) return 1;
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err) {
    PyObject *exc_type = tstate->curexc_type;
    if (exc_type == err) return 1;
    if (unlikely(!exc_type)) return 0;
    if (unlikely(PyTuple_Check(err)))
        return __Pyx_PyErr_ExceptionMatchesTuple(exc_type, err);
    return __Pyx_PyErr_GivenExceptionMatches(exc_type, err);
}
#endif

/* GetAttr */
static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *o, PyObject *n) {
#if CYTHON_USE_TYPE_SLOTS
#if PY_MAJOR_VERSION >= 3
    if (likely(PyUnicode_Check(n)))
#else
    if (likely(PyString_Check(n)))
#endif
        return __Pyx_PyObject_GetAttrStr(o, n);
#endif
    return PyObject_GetAttr(o, n);
}

/* GetAttr3 */
static PyObject *__Pyx_GetAttr3Default(PyObject *d) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (unlikely(!__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        return NULL;
    __Pyx_PyErr_Clear();
    Py_INCREF(d);
    return d;
}
static CYTHON_INLINE PyObject *__Pyx_GetAttr3(PyObject *o, PyObject *n, PyObject *d) {
    PyObject *r = __Pyx_GetAttr(o, n);
    return (likely(r)) ? r : __Pyx_GetAttr3Default(d);
}

/* GetModuleGlobalName */
#if CYTHON_USE_DICT_VERSIONS
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value)
#else
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name)
#endif
{
    PyObject *result;
#if !CYTHON_AVOID_BORROWED_REFS
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1
    result = _PyDict_GetItem_KnownHash(__pyx_d, name, ((PyASCIIObject *) name)->hash);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    } else if (unlikely(PyErr_Occurred())) {
        return NULL;
    }
#else
    result = PyDict_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
#endif
#else
    result = PyObject_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
    PyErr_Clear();
#endif
    return __Pyx_GetBuiltinName(name);
}

/* BytesEquals */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY
    return PyObject_RichCompareBool(s1, s2, equals);
#else
    if (s1 == s2) {
        return (equals == Py_EQ);
    } else if (PyBytes_CheckExact(s1) & PyBytes_CheckExact(s2)) {
        const char *ps1, *ps2;
        Py_ssize_t length = PyBytes_GET_SIZE(s1);
        if (length != PyBytes_GET_SIZE(s2))
            return (equals == Py_NE);
        ps1 = PyBytes_AS_STRING(s1);
        ps2 = PyBytes_AS_STRING(s2);
        if (ps1[0] != ps2[0]) {
            return (equals == Py_NE);
        } else if (length == 1) {
            return (equals == Py_EQ);
        } else {
            int result;
#if CYTHON_USE_UNICODE_INTERNALS && (PY_VERSION_HEX < 0x030B0000)
            Py_hash_t hash1, hash2;
            hash1 = ((PyBytesObject*)s1)->ob_shash;
            hash2 = ((PyBytesObject*)s2)->ob_shash;
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                return (equals == Py_NE);
            }
#endif
            result = memcmp(ps1, ps2, (size_t)length);
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & PyBytes_CheckExact(s2)) {
        return (equals == Py_NE);
    } else if ((s2 == Py_None) & PyBytes_CheckExact(s1)) {
        return (equals == Py_NE);
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
#endif
}

/* UnicodeEquals */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY
    return PyObject_RichCompareBool(s1, s2, equals);
#else
#if PY_MAJOR_VERSION < 3
    PyObject* owned_ref = NULL;
#endif
    int s1_is_unicode, s2_is_unicode;
    if (s1 == s2) {
        goto return_eq;
    }
    s1_is_unicode = PyUnicode_CheckExact(s1);
    s2_is_unicode = PyUnicode_CheckExact(s2);
#if PY_MAJOR_VERSION < 3
    if ((s1_is_unicode & (!s2_is_unicode)) && PyString_CheckExact(s2)) {
        owned_ref = PyUnicode_FromObject(s2);
        if (unlikely(!owned_ref))
            return -1;
        s2 = owned_ref;
        s2_is_unicode = 1;
    } else if ((s2_is_unicode & (!s1_is_unicode)) && PyString_CheckExact(s1)) {
        owned_ref = PyUnicode_FromObject(s1);
        if (unlikely(!owned_ref))
            return -1;
        s1 = owned_ref;
        s1_is_unicode = 1;
    } else if (((!s2_is_unicode) & (!s1_is_unicode))) {
        return __Pyx_PyBytes_Equals(s1, s2, equals);
    }
#endif
    if (s1_is_unicode & s2_is_unicode) {
        Py_ssize_t length;
        int kind;
        void *data1, *data2;
        if (unlikely(__Pyx_PyUnicode_READY(s1) < 0) || unlikely(__Pyx_PyUnicode_READY(s2) < 0))
            return -1;
        length = __Pyx_PyUnicode_GET_LENGTH(s1);
        if (length != __Pyx_PyUnicode_GET_LENGTH(s2)) {
            goto return_ne;
        }
#if CYTHON_USE_UNICODE_INTERNALS
        {
            Py_hash_t hash1, hash2;
        #if CYTHON_PEP393_ENABLED
            hash1 = ((PyASCIIObject*)s1)->hash;
            hash2 = ((PyASCIIObject*)s2)->hash;
        #else
            hash1 = ((PyUnicodeObject*)s1)->hash;
            hash2 = ((PyUnicodeObject*)s2)->hash;
        #endif
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                goto return_ne;
            }
        }
#endif
        kind = __Pyx_PyUnicode_KIND(s1);
        if (kind != __Pyx_PyUnicode_KIND(s2)) {
            goto return_ne;
        }
        data1 = __Pyx_PyUnicode_DATA(s1);
        data2 = __Pyx_PyUnicode_DATA(s2);
        if (__Pyx_PyUnicode_READ(kind, data1, 0) != __Pyx_PyUnicode_READ(kind, data2, 0)) {
            goto return_ne;
        } else if (length == 1) {
            goto return_eq;
        } else {
            int result = memcmp(data1, data2, (size_t)(length * kind));
            #if PY_MAJOR_VERSION < 3
            Py_XDECREF(owned_ref);
            #endif
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & s2_is_unicode) {
        goto return_ne;
    } else if ((s2 == Py_None) & s1_is_unicode) {
        goto return_ne;
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        #if PY_MAJOR_VERSION < 3
        Py_XDECREF(owned_ref);
        #endif
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
return_eq:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(owned_ref);
    #endif
    return (equals == Py_EQ);
return_ne:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(owned_ref);
    #endif
    return (equals == Py_NE);
#endif
}

/* RaiseException */
#if PY_MAJOR_VERSION < 3
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb,
                        CYTHON_UNUSED PyObject *cause) {
    __Pyx_PyThreadState_declare
    Py_XINCREF(type);
    if (!value || value == Py_None)
        value = NULL;
    else
        Py_INCREF(value);
    if (!tb || tb == Py_None)
        tb = NULL;
    else {
        Py_INCREF(tb);
        if (!PyTraceBack_Check(tb)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: arg 3 must be a traceback or None");
            goto raise_error;
        }
    }
    if (PyType_Check(type)) {
#if CYTHON_COMPILING_IN_PYPY
        if (!value) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#endif
        PyErr_NormalizeException(&type, &value, &tb);
    } else {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto raise_error;
        }
        value = type;
        type = (PyObject*) Py_TYPE(type);
        Py_INCREF(type);
        if (!PyType_IsSubtype((PyTypeObject *)type, (PyTypeObject *)PyExc_BaseException)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: exception class must be a subclass of BaseException");
            goto raise_error;
        }
    }
    __Pyx_PyThreadState_assign
    __Pyx_ErrRestore(type, value, tb);
    return;
raise_error:
    Py_XDECREF(value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
    return;
}
#else
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
    if (cause) {
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
#if CYTHON_COMPILING_IN_PYPY
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#else
        PyThreadState *tstate = __Pyx_PyThreadState_Current;
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}
#endif

/* py_abs */
#if CYTHON_USE_PYLONG_INTERNALS
static PyObject *__Pyx_PyLong_AbsNeg(PyObject *n) {
    if (likely(Py_SIZE(n) == -1)) {
        return PyLong_FromLong(((PyLongObject*)n)->ob_digit[0]);
    }
#if CYTHON_COMPILING_IN_CPYTHON
    {
        PyObject *copy = _PyLong_Copy((PyLongObject*)n);
        if (likely(copy)) {
            __Pyx_SET_SIZE(copy, -Py_SIZE(copy));
        }
        return copy;
    }
#else
    return PyNumber_Negative(n);
#endif
}
#endif

/* GetTopmostException */
#if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem *
__Pyx_PyErr_GetTopmostException(PyThreadState *tstate)
{
    _PyErr_StackItem *exc_info = tstate->exc_info;
    while ((exc_info->exc_type == NULL || exc_info->exc_type == Py_None) &&
           exc_info->previous_item != NULL)
    {
        exc_info = exc_info->previous_item;
    }
    return exc_info;
}
#endif

/* SaveResetException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = __Pyx_PyErr_GetTopmostException(tstate);
    *type = exc_info->exc_type;
    *value = exc_info->exc_value;
    *tb = exc_info->exc_traceback;
    #else
    *type = tstate->exc_type;
    *value = tstate->exc_value;
    *tb = tstate->exc_traceback;
    #endif
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
}
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = type;
    exc_info->exc_value = value;
    exc_info->exc_traceback = tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = type;
    tstate->exc_value = value;
    tstate->exc_traceback = tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
#endif

/* GetException */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb)
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb)
#endif
{
    PyObject *local_type, *local_value, *local_tb;
#if CYTHON_FAST_THREAD_STATE
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    local_type = tstate->curexc_type;
    local_value = tstate->curexc_value;
    local_tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
#else
    PyErr_Fetch(&local_type, &local_value, &local_tb);
#endif
    PyErr_NormalizeException(&local_type, &local_value, &local_tb);
#if CYTHON_FAST_THREAD_STATE
    if (unlikely(tstate->curexc_type))
#else
    if (unlikely(PyErr_Occurred()))
#endif
        goto bad;
    #if PY_MAJOR_VERSION >= 3
    if (local_tb) {
        if (unlikely(PyException_SetTraceback(local_value, local_tb) < 0))
            goto bad;
    }
    #endif
    Py_XINCREF(local_tb);
    Py_XINCREF(local_type);
    Py_XINCREF(local_value);
    *type = local_type;
    *value = local_value;
    *tb = local_tb;
#if CYTHON_FAST_THREAD_STATE
    #if CYTHON_USE_EXC_INFO_STACK
    {
        _PyErr_StackItem *exc_info = tstate->exc_info;
        tmp_type = exc_info->exc_type;
        tmp_value = exc_info->exc_value;
        tmp_tb = exc_info->exc_traceback;
        exc_info->exc_type = local_type;
        exc_info->exc_value = local_value;
        exc_info->exc_traceback = local_tb;
    }
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = local_type;
    tstate->exc_value = local_value;
    tstate->exc_traceback = local_tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#else
    PyErr_SetExcInfo(local_type, local_value, local_tb);
#endif
    return 0;
bad:
    *type = 0;
    *value = 0;
    *tb = 0;
    Py_XDECREF(local_type);
    Py_XDECREF(local_value);
    Py_XDECREF(local_tb);
    return -1;
}

/* PyObjectCall2Args */
static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2) {
    PyObject *args, *result = NULL;
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyFunction_FastCall(function, args, 2);
    }
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyCFunction_FastCall(function, args, 2);
    }
    #endif
    args = PyTuple_New(2);
    if (unlikely(!args)) goto done;
    Py_INCREF(arg1);
    PyTuple_SET_ITEM(args, 0, arg1);
    Py_INCREF(arg2);
    PyTuple_SET_ITEM(args, 1, arg2);
    Py_INCREF(function);
    result = __Pyx_PyObject_Call(function, args, NULL);
    Py_DECREF(args);
    Py_DECREF(function);
done:
    return result;
}

/* ArgTypeTest */
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact)
{
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    else if (exact) {
        #if PY_MAJOR_VERSION == 2
        if ((type == &PyBaseString_Type) && likely(__Pyx_PyBaseString_CheckExact(obj))) return 1;
        #endif
    }
    else {
        if (likely(__Pyx_TypeCheck(obj, type))) return 1;
    }
    PyErr_Format(PyExc_TypeError,
        "Argument '%.200s' has incorrect type (expected %.200s, got %.200s)",
        name, type->tp_name, Py_TYPE(obj)->tp_name);
    return 0;
}

/* StringJoin */
#if !CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyBytes_Join(PyObject* sep, PyObject* values) {
    return PyObject_CallMethodObjArgs(sep, __pyx_n_s_join, values, NULL);
}
#endif

/* RaiseArgTupleInvalid */
static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

/* RaiseDoubleKeywords */
static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        #if PY_MAJOR_VERSION >= 3
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
        #else
        "%s() got multiple values for keyword argument '%s'", func_name,
        PyString_AsString(kw_name));
        #endif
}

/* ParseKeywords */
static int __Pyx_ParseOptionalKeywords(
    PyObject *kwds,
    PyObject **argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject *key = 0, *value = 0;
    Py_ssize_t pos = 0;
    PyObject*** name;
    PyObject*** first_kw_arg = argnames + num_pos_args;
    while (PyDict_Next(kwds, &pos, &key, &value)) {
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            values[name-argnames] = value;
            continue;
        }
        name = first_kw_arg;
        #if PY_MAJOR_VERSION < 3
        if (likely(PyString_Check(key))) {
            while (*name) {
                if ((CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**name) == PyString_GET_SIZE(key))
                        && _PyString_Eq(**name, key)) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    if ((**argname == key) || (
                            (CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**argname) == PyString_GET_SIZE(key))
                             && _PyString_Eq(**argname, key))) {
                        goto arg_passed_twice;
                    }
                    argname++;
                }
            }
        } else
        #endif
        if (likely(PyUnicode_Check(key))) {
            while (*name) {
                int cmp = (**name == key) ? 0 :
                #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                    (__Pyx_PyUnicode_GET_LENGTH(**name) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                #endif
                    PyUnicode_Compare(**name, key);
                if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                if (cmp == 0) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    int cmp = (**argname == key) ? 0 :
                    #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                        (__Pyx_PyUnicode_GET_LENGTH(**argname) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                    #endif
                        PyUnicode_Compare(**argname, key);
                    if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                    if (cmp == 0) goto arg_passed_twice;
                    argname++;
                }
            }
        } else
            goto invalid_keyword_type;
        if (kwds2) {
            if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
        } else {
            goto invalid_keyword;
        }
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
    #if PY_MAJOR_VERSION < 3
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
bad:
    return -1;
}

/* GetItemInt */
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j) {
    PyObject *r;
    if (!j) return NULL;
    r = PyObject_GetItem(o, j);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyList_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyList_GET_SIZE(o)))) {
        PyObject *r = PyList_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyTuple_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyTuple_GET_SIZE(o)))) {
        PyObject *r = PyTuple_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i, int is_list,
                                                     CYTHON_NCP_UNUSED int wraparound,
                                                     CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS && CYTHON_USE_TYPE_SLOTS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyList_GET_SIZE(o);
        if ((!boundscheck) || (likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o))))) {
            PyObject *r = PyList_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    }
    else if (PyTuple_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyTuple_GET_SIZE(o);
        if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyTuple_GET_SIZE(o)))) {
            PyObject *r = PyTuple_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    } else {
        PySequenceMethods *m = Py_TYPE(o)->tp_as_sequence;
        if (likely(m && m->sq_item)) {
            if (wraparound && unlikely(i < 0) && likely(m->sq_length)) {
                Py_ssize_t l = m->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return NULL;
                    PyErr_Clear();
                }
            }
            return m->sq_item(o, i);
        }
    }
#else
    if (is_list || PySequence_Check(o)) {
        return PySequence_GetItem(o, i);
    }
#endif
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
}

/* DictGetItem */
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key) {
    PyObject *value;
    value = PyDict_GetItemWithError(d, key);
    if (unlikely(!value)) {
        if (!PyErr_Occurred()) {
            if (unlikely(PyTuple_Check(key))) {
                PyObject* args = PyTuple_Pack(1, key);
                if (likely(args)) {
                    PyErr_SetObject(PyExc_KeyError, args);
                    Py_DECREF(args);
                }
            } else {
                PyErr_SetObject(PyExc_KeyError, key);
            }
        }
        return NULL;
    }
    Py_INCREF(value);
    return value;
}
#endif

/* SetItemInt */
static int __Pyx_SetItemInt_Generic(PyObject *o, PyObject *j, PyObject *v) {
    int r;
    if (!j) return -1;
    r = PyObject_SetItem(o, j, v);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE int __Pyx_SetItemInt_Fast(PyObject *o, Py_ssize_t i, PyObject *v, int is_list,
                                               CYTHON_NCP_UNUSED int wraparound, CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS && CYTHON_USE_TYPE_SLOTS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = (!wraparound) ? i : ((likely(i >= 0)) ? i : i + PyList_GET_SIZE(o));
        if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o)))) {
            PyObject* old = PyList_GET_ITEM(o, n);
            Py_INCREF(v);
            PyList_SET_ITEM(o, n, v);
            Py_DECREF(old);
            return 1;
        }
    } else {
        PySequenceMethods *m = Py_TYPE(o)->tp_as_sequence;
        if (likely(m && m->sq_ass_item)) {
            if (wraparound && unlikely(i < 0) && likely(m->sq_length)) {
                Py_ssize_t l = m->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return -1;
                    PyErr_Clear();
                }
            }
            return m->sq_ass_item(o, i, v);
        }
    }
#else
#if CYTHON_COMPILING_IN_PYPY
    if (is_list || (PySequence_Check(o) && !PyDict_Check(o)))
#else
    if (is_list || PySequence_Check(o))
#endif
    {
        return PySequence_SetItem(o, i, v);
    }
#endif
    return __Pyx_SetItemInt_Generic(o, PyInt_FromSsize_t(i), v);
}

/* IterFinish */
static CYTHON_INLINE int __Pyx_IterFinish(void) {
#if CYTHON_FAST_THREAD_STATE
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject* exc_type = tstate->curexc_type;
    if (unlikely(exc_type)) {
        if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) {
            PyObject *exc_value, *exc_tb;
            exc_value = tstate->curexc_value;
            exc_tb = tstate->curexc_traceback;
            tstate->curexc_type = 0;
            tstate->curexc_value = 0;
            tstate->curexc_traceback = 0;
            Py_DECREF(exc_type);
            Py_XDECREF(exc_value);
            Py_XDECREF(exc_tb);
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#else
    if (unlikely(PyErr_Occurred())) {
        if (likely(PyErr_ExceptionMatches(PyExc_StopIteration))) {
            PyErr_Clear();
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#endif
}

/* PyObjectGetMethod */
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method) {
    PyObject *attr;
#if CYTHON_UNPACK_METHODS && CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_PYTYPE_LOOKUP
    PyTypeObject *tp = Py_TYPE(obj);
    PyObject *descr;
    descrgetfunc f = NULL;
    PyObject **dictptr, *dict;
    int meth_found = 0;
    assert (*method == NULL);
    if (unlikely(tp->tp_getattro != PyObject_GenericGetAttr)) {
        attr = __Pyx_PyObject_GetAttrStr(obj, name);
        goto try_unpack;
    }
    if (unlikely(tp->tp_dict == NULL) && unlikely(PyType_Ready(tp) < 0)) {
        return 0;
    }
    descr = _PyType_Lookup(tp, name);
    if (likely(descr != NULL)) {
        Py_INCREF(descr);
#if PY_MAJOR_VERSION >= 3
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || (Py_TYPE(descr) == &PyMethodDescr_Type) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr) || (Py_TYPE(descr) == &PyMethodDescr_Type)))
        #endif
#else
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr)))
        #endif
#endif
        {
            meth_found = 1;
        } else {
            f = Py_TYPE(descr)->tp_descr_get;
            if (f != NULL && PyDescr_IsData(descr)) {
                attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
                Py_DECREF(descr);
                goto try_unpack;
            }
        }
    }
    dictptr = _PyObject_GetDictPtr(obj);
    if (dictptr != NULL && (dict = *dictptr) != NULL) {
        Py_INCREF(dict);
        attr = __Pyx_PyDict_GetItemStr(dict, name);
        if (attr != NULL) {
            Py_INCREF(attr);
            Py_DECREF(dict);
            Py_XDECREF(descr);
            goto try_unpack;
        }
        Py_DECREF(dict);
    }
    if (meth_found) {
        *method = descr;
        return 1;
    }
    if (f != NULL) {
        attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
        Py_DECREF(descr);
        goto try_unpack;
    }
    if (descr != NULL) {
        *method = descr;
        return 0;
    }
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'%.50s' object has no attribute '%U'",
                 tp->tp_name, name);
#else
                 "'%.50s' object has no attribute '%.400s'",
                 tp->tp_name, PyString_AS_STRING(name));
#endif
    return 0;
#else
    attr = __Pyx_PyObject_GetAttrStr(obj, name);
    goto try_unpack;
#endif
try_unpack:
#if CYTHON_UNPACK_METHODS
    if (likely(attr) && PyMethod_Check(attr) && likely(PyMethod_GET_SELF(attr) == obj)) {
        PyObject *function = PyMethod_GET_FUNCTION(attr);
        Py_INCREF(function);
        Py_DECREF(attr);
        *method = function;
        return 1;
    }
#endif
    *method = attr;
    return 0;
}

/* PyObjectCallMethod0 */
static PyObject* __Pyx_PyObject_CallMethod0(PyObject* obj, PyObject* method_name) {
    PyObject *method = NULL, *result = NULL;
    int is_method = __Pyx_PyObject_GetMethod(obj, method_name, &method);
    if (likely(is_method)) {
        result = __Pyx_PyObject_CallOneArg(method, obj);
        Py_DECREF(method);
        return result;
    }
    if (unlikely(!method)) goto bad;
    result = __Pyx_PyObject_CallNoArg(method);
    Py_DECREF(method);
bad:
    return result;
}

/* RaiseNeedMoreValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index) {
    PyErr_Format(PyExc_ValueError,
                 "need more than %" CYTHON_FORMAT_SSIZE_T "d value%.1s to unpack",
                 index, (index == 1) ? "" : "s");
}

/* RaiseTooManyValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected) {
    PyErr_Format(PyExc_ValueError,
                 "too many values to unpack (expected %" CYTHON_FORMAT_SSIZE_T "d)", expected);
}

/* UnpackItemEndCheck */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected) {
    if (unlikely(retval)) {
        Py_DECREF(retval);
        __Pyx_RaiseTooManyValuesError(expected);
        return -1;
    }
    return __Pyx_IterFinish();
}

/* RaiseNoneIterError */
static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
}

/* UnpackTupleError */
static void __Pyx_UnpackTupleError(PyObject *t, Py_ssize_t index) {
    if (t == Py_None) {
      __Pyx_RaiseNoneNotIterableError();
    } else if (PyTuple_GET_SIZE(t) < index) {
      __Pyx_RaiseNeedMoreValuesError(PyTuple_GET_SIZE(t));
    } else {
      __Pyx_RaiseTooManyValuesError(index);
    }
}

/* UnpackTuple2 */
static CYTHON_INLINE int __Pyx_unpack_tuple2_exact(
        PyObject* tuple, PyObject** pvalue1, PyObject** pvalue2, int decref_tuple) {
    PyObject *value1 = NULL, *value2 = NULL;
#if CYTHON_COMPILING_IN_PYPY
    value1 = PySequence_ITEM(tuple, 0);  if (unlikely(!value1)) goto bad;
    value2 = PySequence_ITEM(tuple, 1);  if (unlikely(!value2)) goto bad;
#else
    value1 = PyTuple_GET_ITEM(tuple, 0);  Py_INCREF(value1);
    value2 = PyTuple_GET_ITEM(tuple, 1);  Py_INCREF(value2);
#endif
    if (decref_tuple) {
        Py_DECREF(tuple);
    }
    *pvalue1 = value1;
    *pvalue2 = value2;
    return 0;
#if CYTHON_COMPILING_IN_PYPY
bad:
    Py_XDECREF(value1);
    Py_XDECREF(value2);
    if (decref_tuple) { Py_XDECREF(tuple); }
    return -1;
#endif
}
static int __Pyx_unpack_tuple2_generic(PyObject* tuple, PyObject** pvalue1, PyObject** pvalue2,
                                       int has_known_size, int decref_tuple) {
    Py_ssize_t index;
    PyObject *value1 = NULL, *value2 = NULL, *iter = NULL;
    iternextfunc iternext;
    iter = PyObject_GetIter(tuple);
    if (unlikely(!iter)) goto bad;
    if (decref_tuple) { Py_DECREF(tuple); tuple = NULL; }
    iternext = Py_TYPE(iter)->tp_iternext;
    value1 = iternext(iter); if (unlikely(!value1)) { index = 0; goto unpacking_failed; }
    value2 = iternext(iter); if (unlikely(!value2)) { index = 1; goto unpacking_failed; }
    if (!has_known_size && unlikely(__Pyx_IternextUnpackEndCheck(iternext(iter), 2))) goto bad;
    Py_DECREF(iter);
    *pvalue1 = value1;
    *pvalue2 = value2;
    return 0;
unpacking_failed:
    if (!has_known_size && __Pyx_IterFinish() == 0)
        __Pyx_RaiseNeedMoreValuesError(index);
bad:
    Py_XDECREF(iter);
    Py_XDECREF(value1);
    Py_XDECREF(value2);
    if (decref_tuple) { Py_XDECREF(tuple); }
    return -1;
}

/* dict_iter */
static CYTHON_INLINE PyObject* __Pyx_dict_iterator(PyObject* iterable, int is_dict, PyObject* method_name,
                                                   Py_ssize_t* p_orig_length, int* p_source_is_dict) {
    is_dict = is_dict || likely(PyDict_CheckExact(iterable));
    *p_source_is_dict = is_dict;
    if (is_dict) {
#if !CYTHON_COMPILING_IN_PYPY
        *p_orig_length = PyDict_Size(iterable);
        Py_INCREF(iterable);
        return iterable;
#elif PY_MAJOR_VERSION >= 3
        static PyObject *py_items = NULL, *py_keys = NULL, *py_values = NULL;
        PyObject **pp = NULL;
        if (method_name) {
            const char *name = PyUnicode_AsUTF8(method_name);
            if (strcmp(name, "iteritems") == 0) pp = &py_items;
            else if (strcmp(name, "iterkeys") == 0) pp = &py_keys;
            else if (strcmp(name, "itervalues") == 0) pp = &py_values;
            if (pp) {
                if (!*pp) {
                    *pp = PyUnicode_FromString(name + 4);
                    if (!*pp)
                        return NULL;
                }
                method_name = *pp;
            }
        }
#endif
    }
    *p_orig_length = 0;
    if (method_name) {
        PyObject* iter;
        iterable = __Pyx_PyObject_CallMethod0(iterable, method_name);
        if (!iterable)
            return NULL;
#if !CYTHON_COMPILING_IN_PYPY
        if (PyTuple_CheckExact(iterable) || PyList_CheckExact(iterable))
            return iterable;
#endif
        iter = PyObject_GetIter(iterable);
        Py_DECREF(iterable);
        return iter;
    }
    return PyObject_GetIter(iterable);
}
static CYTHON_INLINE int __Pyx_dict_iter_next(
        PyObject* iter_obj, CYTHON_NCP_UNUSED Py_ssize_t orig_length, CYTHON_NCP_UNUSED Py_ssize_t* ppos,
        PyObject** pkey, PyObject** pvalue, PyObject** pitem, int source_is_dict) {
    PyObject* next_item;
#if !CYTHON_COMPILING_IN_PYPY
    if (source_is_dict) {
        PyObject *key, *value;
        if (unlikely(orig_length != PyDict_Size(iter_obj))) {
            PyErr_SetString(PyExc_RuntimeError, "dictionary changed size during iteration");
            return -1;
        }
        if (unlikely(!PyDict_Next(iter_obj, ppos, &key, &value))) {
            return 0;
        }
        if (pitem) {
            PyObject* tuple = PyTuple_New(2);
            if (unlikely(!tuple)) {
                return -1;
            }
            Py_INCREF(key);
            Py_INCREF(value);
            PyTuple_SET_ITEM(tuple, 0, key);
            PyTuple_SET_ITEM(tuple, 1, value);
            *pitem = tuple;
        } else {
            if (pkey) {
                Py_INCREF(key);
                *pkey = key;
            }
            if (pvalue) {
                Py_INCREF(value);
                *pvalue = value;
            }
        }
        return 1;
    } else if (PyTuple_CheckExact(iter_obj)) {
        Py_ssize_t pos = *ppos;
        if (unlikely(pos >= PyTuple_GET_SIZE(iter_obj))) return 0;
        *ppos = pos + 1;
        next_item = PyTuple_GET_ITEM(iter_obj, pos);
        Py_INCREF(next_item);
    } else if (PyList_CheckExact(iter_obj)) {
        Py_ssize_t pos = *ppos;
        if (unlikely(pos >= PyList_GET_SIZE(iter_obj))) return 0;
        *ppos = pos + 1;
        next_item = PyList_GET_ITEM(iter_obj, pos);
        Py_INCREF(next_item);
    } else
#endif
    {
        next_item = PyIter_Next(iter_obj);
        if (unlikely(!next_item)) {
            return __Pyx_IterFinish();
        }
    }
    if (pitem) {
        *pitem = next_item;
    } else if (pkey && pvalue) {
        if (__Pyx_unpack_tuple2(next_item, pkey, pvalue, source_is_dict, source_is_dict, 1))
            return -1;
    } else if (pkey) {
        *pkey = next_item;
    } else {
        *pvalue = next_item;
    }
    return 1;
}

/* None */
static CYTHON_INLINE void __Pyx_RaiseClosureNameError(const char *varname) {
    PyErr_Format(PyExc_NameError, "free variable '%s' referenced before assignment in enclosing scope", varname);
}

/* Import */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level) {
    PyObject *empty_list = 0;
    PyObject *module = 0;
    PyObject *global_dict = 0;
    PyObject *empty_dict = 0;
    PyObject *list;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_import;
    py_import = __Pyx_PyObject_GetAttrStr(__pyx_b, __pyx_n_s_import);
    if (!py_import)
        goto bad;
    #endif
    if (from_list)
        list = from_list;
    else {
        empty_list = PyList_New(0);
        if (!empty_list)
            goto bad;
        list = empty_list;
    }
    global_dict = PyModule_GetDict(__pyx_m);
    if (!global_dict)
        goto bad;
    empty_dict = PyDict_New();
    if (!empty_dict)
        goto bad;
    {
        #if PY_MAJOR_VERSION >= 3
        if (level == -1) {
            if ((1) && (strchr(__Pyx_MODULE_NAME, '.'))) {
                module = PyImport_ImportModuleLevelObject(
                    name, global_dict, empty_dict, list, 1);
                if (!module) {
                    if (!PyErr_ExceptionMatches(PyExc_ImportError))
                        goto bad;
                    PyErr_Clear();
                }
            }
            level = 0;
        }
        #endif
        if (!module) {
            #if PY_MAJOR_VERSION < 3
            PyObject *py_level = PyInt_FromLong(level);
            if (!py_level)
                goto bad;
            module = PyObject_CallFunctionObjArgs(py_import,
                name, global_dict, empty_dict, list, py_level, (PyObject *)NULL);
            Py_DECREF(py_level);
            #else
            module = PyImport_ImportModuleLevelObject(
                name, global_dict, empty_dict, list, level);
            #endif
        }
    }
bad:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_import);
    #endif
    Py_XDECREF(empty_list);
    Py_XDECREF(empty_dict);
    return module;
}

/* ImportFrom */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name) {
    PyObject* value = __Pyx_PyObject_GetAttrStr(module, name);
    if (unlikely(!value) && PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Format(PyExc_ImportError,
        #if PY_MAJOR_VERSION < 3
            "cannot import name %.230s", PyString_AS_STRING(name));
        #else
            "cannot import name %S", name);
        #endif
    }
    return value;
}

/* HasAttr */
static CYTHON_INLINE int __Pyx_HasAttr(PyObject *o, PyObject *n) {
    PyObject *r;
    if (unlikely(!__Pyx_PyBaseString_Check(n))) {
        PyErr_SetString(PyExc_TypeError,
                        "hasattr(): attribute name must be string");
        return -1;
    }
    r = __Pyx_GetAttr(o, n);
    if (unlikely(!r)) {
        PyErr_Clear();
        return 0;
    } else {
        Py_DECREF(r);
        return 1;
    }
}

/* FetchCommonType */
static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type) {
    PyObject* fake_module;
    PyTypeObject* cached_type = NULL;
    fake_module = PyImport_AddModule((char*) "_cython_" CYTHON_ABI);
    if (!fake_module) return NULL;
    Py_INCREF(fake_module);
    cached_type = (PyTypeObject*) PyObject_GetAttrString(fake_module, type->tp_name);
    if (cached_type) {
        if (!PyType_Check((PyObject*)cached_type)) {
            PyErr_Format(PyExc_TypeError,
                "Shared Cython type %.200s is not a type object",
                type->tp_name);
            goto bad;
        }
        if (cached_type->tp_basicsize != type->tp_basicsize) {
            PyErr_Format(PyExc_TypeError,
                "Shared Cython type %.200s has the wrong size, try recompiling",
                type->tp_name);
            goto bad;
        }
    } else {
        if (!PyErr_ExceptionMatches(PyExc_AttributeError)) goto bad;
        PyErr_Clear();
        if (PyType_Ready(type) < 0) goto bad;
        if (PyObject_SetAttrString(fake_module, type->tp_name, (PyObject*) type) < 0)
            goto bad;
        Py_INCREF(type);
        cached_type = type;
    }
done:
    Py_DECREF(fake_module);
    return cached_type;
bad:
    Py_XDECREF(cached_type);
    cached_type = NULL;
    goto done;
}

/* CythonFunctionShared */
#include <structmember.h>
static PyObject *
__Pyx_CyFunction_get_doc(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *closure)
{
    if (unlikely(op->func_doc == NULL)) {
        if (op->func.m_ml->ml_doc) {
#if PY_MAJOR_VERSION >= 3
            op->func_doc = PyUnicode_FromString(op->func.m_ml->ml_doc);
#else
            op->func_doc = PyString_FromString(op->func.m_ml->ml_doc);
#endif
            if (unlikely(op->func_doc == NULL))
                return NULL;
        } else {
            Py_INCREF(Py_None);
            return Py_None;
        }
    }
    Py_INCREF(op->func_doc);
    return op->func_doc;
}
static int
__Pyx_CyFunction_set_doc(__pyx_CyFunctionObject *op, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp = op->func_doc;
    if (value == NULL) {
        value = Py_None;
    }
    Py_INCREF(value);
    op->func_doc = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_name(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    if (unlikely(op->func_name == NULL)) {
#if PY_MAJOR_VERSION >= 3
        op->func_name = PyUnicode_InternFromString(op->func.m_ml->ml_name);
#else
        op->func_name = PyString_InternFromString(op->func.m_ml->ml_name);
#endif
        if (unlikely(op->func_name == NULL))
            return NULL;
    }
    Py_INCREF(op->func_name);
    return op->func_name;
}
static int
__Pyx_CyFunction_set_name(__pyx_CyFunctionObject *op, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__name__ must be set to a string object");
        return -1;
    }
    tmp = op->func_name;
    Py_INCREF(value);
    op->func_name = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_qualname(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    Py_INCREF(op->func_qualname);
    return op->func_qualname;
}
static int
__Pyx_CyFunction_set_qualname(__pyx_CyFunctionObject *op, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__qualname__ must be set to a string object");
        return -1;
    }
    tmp = op->func_qualname;
    Py_INCREF(value);
    op->func_qualname = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_self(__pyx_CyFunctionObject *m, CYTHON_UNUSED void *closure)
{
    PyObject *self;
    self = m->func_closure;
    if (self == NULL)
        self = Py_None;
    Py_INCREF(self);
    return self;
}
static PyObject *
__Pyx_CyFunction_get_dict(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    if (unlikely(op->func_dict == NULL)) {
        op->func_dict = PyDict_New();
        if (unlikely(op->func_dict == NULL))
            return NULL;
    }
    Py_INCREF(op->func_dict);
    return op->func_dict;
}
static int
__Pyx_CyFunction_set_dict(__pyx_CyFunctionObject *op, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
    if (unlikely(value == NULL)) {
        PyErr_SetString(PyExc_TypeError,
               "function's dictionary may not be deleted");
        return -1;
    }
    if (unlikely(!PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
               "setting function's dictionary to a non-dict");
        return -1;
    }
    tmp = op->func_dict;
    Py_INCREF(value);
    op->func_dict = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_globals(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    Py_INCREF(op->func_globals);
    return op->func_globals;
}
static PyObject *
__Pyx_CyFunction_get_closure(CYTHON_UNUSED __pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    Py_INCREF(Py_None);
    return Py_None;
}
static PyObject *
__Pyx_CyFunction_get_code(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    PyObject* result = (op->func_code) ? op->func_code : Py_None;
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_init_defaults(__pyx_CyFunctionObject *op) {
    int result = 0;
    PyObject *res = op->defaults_getter((PyObject *) op);
    if (unlikely(!res))
        return -1;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    op->defaults_tuple = PyTuple_GET_ITEM(res, 0);
    Py_INCREF(op->defaults_tuple);
    op->defaults_kwdict = PyTuple_GET_ITEM(res, 1);
    Py_INCREF(op->defaults_kwdict);
    #else
    op->defaults_tuple = PySequence_ITEM(res, 0);
    if (unlikely(!op->defaults_tuple)) result = -1;
    else {
        op->defaults_kwdict = PySequence_ITEM(res, 1);
        if (unlikely(!op->defaults_kwdict)) result = -1;
    }
    #endif
    Py_DECREF(res);
    return result;
}
static int
__Pyx_CyFunction_set_defaults(__pyx_CyFunctionObject *op, PyObject* value, CYTHON_UNUSED void *context) {
    PyObject* tmp;
    if (!value) {
        value = Py_None;
    } else if (value != Py_None && !PyTuple_Check(value)) {
        PyErr_SetString(PyExc_TypeError,
                        "__defaults__ must be set to a tuple object");
        return -1;
    }
    Py_INCREF(value);
    tmp = op->defaults_tuple;
    op->defaults_tuple = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_defaults(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context) {
    PyObject* result = op->defaults_tuple;
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (__Pyx_CyFunction_init_defaults(op) < 0) return NULL;
            result = op->defaults_tuple;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_set_kwdefaults(__pyx_CyFunctionObject *op, PyObject* value, CYTHON_UNUSED void *context) {
    PyObject* tmp;
    if (!value) {
        value = Py_None;
    } else if (value != Py_None && !PyDict_Check(value)) {
        PyErr_SetString(PyExc_TypeError,
                        "__kwdefaults__ must be set to a dict object");
        return -1;
    }
    Py_INCREF(value);
    tmp = op->defaults_kwdict;
    op->defaults_kwdict = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_kwdefaults(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context) {
    PyObject* result = op->defaults_kwdict;
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (__Pyx_CyFunction_init_defaults(op) < 0) return NULL;
            result = op->defaults_kwdict;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_set_annotations(__pyx_CyFunctionObject *op, PyObject* value, CYTHON_UNUSED void *context) {
    PyObject* tmp;
    if (!value || value == Py_None) {
        value = NULL;
    } else if (!PyDict_Check(value)) {
        PyErr_SetString(PyExc_TypeError,
                        "__annotations__ must be set to a dict object");
        return -1;
    }
    Py_XINCREF(value);
    tmp = op->func_annotations;
    op->func_annotations = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_annotations(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context) {
    PyObject* result = op->func_annotations;
    if (unlikely(!result)) {
        result = PyDict_New();
        if (unlikely(!result)) return NULL;
        op->func_annotations = result;
    }
    Py_INCREF(result);
    return result;
}
static PyGetSetDef __pyx_CyFunction_getsets[] = {
    {(char *) "func_doc", (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {(char *) "__doc__",  (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {(char *) "func_name", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {(char *) "__name__", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {(char *) "__qualname__", (getter)__Pyx_CyFunction_get_qualname, (setter)__Pyx_CyFunction_set_qualname, 0, 0},
    {(char *) "__self__", (getter)__Pyx_CyFunction_get_self, 0, 0, 0},
    {(char *) "func_dict", (getter)__Pyx_CyFunction_get_dict, (setter)__Pyx_CyFunction_set_dict, 0, 0},
    {(char *) "__dict__", (getter)__Pyx_CyFunction_get_dict, (setter)__Pyx_CyFunction_set_dict, 0, 0},
    {(char *) "func_globals", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {(char *) "__globals__", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {(char *) "func_closure", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {(char *) "__closure__", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {(char *) "func_code", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {(char *) "__code__", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {(char *) "func_defaults", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {(char *) "__defaults__", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {(char *) "__kwdefaults__", (getter)__Pyx_CyFunction_get_kwdefaults, (setter)__Pyx_CyFunction_set_kwdefaults, 0, 0},
    {(char *) "__annotations__", (getter)__Pyx_CyFunction_get_annotations, (setter)__Pyx_CyFunction_set_annotations, 0, 0},
    {0, 0, 0, 0, 0}
};
static PyMemberDef __pyx_CyFunction_members[] = {
    {(char *) "__module__", T_OBJECT, offsetof(PyCFunctionObject, m_module), PY_WRITE_RESTRICTED, 0},
    {0, 0, 0,  0, 0}
};
static PyObject *
__Pyx_CyFunction_reduce(__pyx_CyFunctionObject *m, CYTHON_UNUSED PyObject *args)
{
#if PY_MAJOR_VERSION >= 3
    Py_INCREF(m->func_qualname);
    return m->func_qualname;
#else
    return PyString_FromString(m->func.m_ml->ml_name);
#endif
}
static PyMethodDef __pyx_CyFunction_methods[] = {
    {"__reduce__", (PyCFunction)__Pyx_CyFunction_reduce, METH_VARARGS, 0},
    {0, 0, 0, 0}
};
#if PY_VERSION_HEX < 0x030500A0
#define __Pyx_CyFunction_weakreflist(cyfunc) ((cyfunc)->func_weakreflist)
#else
#define __Pyx_CyFunction_weakreflist(cyfunc) ((cyfunc)->func.m_weakreflist)
#endif
static PyObject *__Pyx_CyFunction_Init(__pyx_CyFunctionObject *op, PyMethodDef *ml, int flags, PyObject* qualname,
                                       PyObject *closure, PyObject *module, PyObject* globals, PyObject* code) {
    if (unlikely(op == NULL))
        return NULL;
    op->flags = flags;
    __Pyx_CyFunction_weakreflist(op) = NULL;
    op->func.m_ml = ml;
    op->func.m_self = (PyObject *) op;
    Py_XINCREF(closure);
    op->func_closure = closure;
    Py_XINCREF(module);
    op->func.m_module = module;
    op->func_dict = NULL;
    op->func_name = NULL;
    Py_INCREF(qualname);
    op->func_qualname = qualname;
    op->func_doc = NULL;
    op->func_classobj = NULL;
    op->func_globals = globals;
    Py_INCREF(op->func_globals);
    Py_XINCREF(code);
    op->func_code = code;
    op->defaults_pyobjects = 0;
    op->defaults_size = 0;
    op->defaults = NULL;
    op->defaults_tuple = NULL;
    op->defaults_kwdict = NULL;
    op->defaults_getter = NULL;
    op->func_annotations = NULL;
    return (PyObject *) op;
}
static int
__Pyx_CyFunction_clear(__pyx_CyFunctionObject *m)
{
    Py_CLEAR(m->func_closure);
    Py_CLEAR(m->func.m_module);
    Py_CLEAR(m->func_dict);
    Py_CLEAR(m->func_name);
    Py_CLEAR(m->func_qualname);
    Py_CLEAR(m->func_doc);
    Py_CLEAR(m->func_globals);
    Py_CLEAR(m->func_code);
    Py_CLEAR(m->func_classobj);
    Py_CLEAR(m->defaults_tuple);
    Py_CLEAR(m->defaults_kwdict);
    Py_CLEAR(m->func_annotations);
    if (m->defaults) {
        PyObject **pydefaults = __Pyx_CyFunction_Defaults(PyObject *, m);
        int i;
        for (i = 0; i < m->defaults_pyobjects; i++)
            Py_XDECREF(pydefaults[i]);
        PyObject_Free(m->defaults);
        m->defaults = NULL;
    }
    return 0;
}
static void __Pyx__CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    if (__Pyx_CyFunction_weakreflist(m) != NULL)
        PyObject_ClearWeakRefs((PyObject *) m);
    __Pyx_CyFunction_clear(m);
    PyObject_GC_Del(m);
}
static void __Pyx_CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    PyObject_GC_UnTrack(m);
    __Pyx__CyFunction_dealloc(m);
}
static int __Pyx_CyFunction_traverse(__pyx_CyFunctionObject *m, visitproc visit, void *arg)
{
    Py_VISIT(m->func_closure);
    Py_VISIT(m->func.m_module);
    Py_VISIT(m->func_dict);
    Py_VISIT(m->func_name);
    Py_VISIT(m->func_qualname);
    Py_VISIT(m->func_doc);
    Py_VISIT(m->func_globals);
    Py_VISIT(m->func_code);
    Py_VISIT(m->func_classobj);
    Py_VISIT(m->defaults_tuple);
    Py_VISIT(m->defaults_kwdict);
    if (m->defaults) {
        PyObject **pydefaults = __Pyx_CyFunction_Defaults(PyObject *, m);
        int i;
        for (i = 0; i < m->defaults_pyobjects; i++)
            Py_VISIT(pydefaults[i]);
    }
    return 0;
}
static PyObject *__Pyx_CyFunction_descr_get(PyObject *func, PyObject *obj, PyObject *type)
{
#if PY_MAJOR_VERSION < 3
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    if (m->flags & __Pyx_CYFUNCTION_STATICMETHOD) {
        Py_INCREF(func);
        return func;
    }
    if (m->flags & __Pyx_CYFUNCTION_CLASSMETHOD) {
        if (type == NULL)
            type = (PyObject *)(Py_TYPE(obj));
        return __Pyx_PyMethod_New(func, type, (PyObject *)(Py_TYPE(type)));
    }
    if (obj == Py_None)
        obj = NULL;
#endif
    return __Pyx_PyMethod_New(func, obj, type);
}
static PyObject*
__Pyx_CyFunction_repr(__pyx_CyFunctionObject *op)
{
#if PY_MAJOR_VERSION >= 3
    return PyUnicode_FromFormat("<cyfunction %U at %p>",
                                op->func_qualname, (void *)op);
#else
    return PyString_FromFormat("<cyfunction %s at %p>",
                               PyString_AsString(op->func_qualname), (void *)op);
#endif
}
static PyObject * __Pyx_CyFunction_CallMethod(PyObject *func, PyObject *self, PyObject *arg, PyObject *kw) {
    PyCFunctionObject* f = (PyCFunctionObject*)func;
    PyCFunction meth = f->m_ml->ml_meth;
    Py_ssize_t size;
    switch (f->m_ml->ml_flags & (METH_VARARGS | METH_KEYWORDS | METH_NOARGS | METH_O)) {
    case METH_VARARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0))
            return (*meth)(self, arg);
        break;
    case METH_VARARGS | METH_KEYWORDS:
        return (*(PyCFunctionWithKeywords)(void*)meth)(self, arg, kw);
    case METH_NOARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
            size = PyTuple_GET_SIZE(arg);
            if (likely(size == 0))
                return (*meth)(self, NULL);
            PyErr_Format(PyExc_TypeError,
                "%.200s() takes no arguments (%" CYTHON_FORMAT_SSIZE_T "d given)",
                f->m_ml->ml_name, size);
            return NULL;
        }
        break;
    case METH_O:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
            size = PyTuple_GET_SIZE(arg);
            if (likely(size == 1)) {
                PyObject *result, *arg0;
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                arg0 = PyTuple_GET_ITEM(arg, 0);
                #else
                arg0 = PySequence_ITEM(arg, 0); if (unlikely(!arg0)) return NULL;
                #endif
                result = (*meth)(self, arg0);
                #if !(CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
                Py_DECREF(arg0);
                #endif
                return result;
            }
            PyErr_Format(PyExc_TypeError,
                "%.200s() takes exactly one argument (%" CYTHON_FORMAT_SSIZE_T "d given)",
                f->m_ml->ml_name, size);
            return NULL;
        }
        break;
    default:
        PyErr_SetString(PyExc_SystemError, "Bad call flags in "
                        "__Pyx_CyFunction_Call. METH_OLDARGS is no "
                        "longer supported!");
        return NULL;
    }
    PyErr_Format(PyExc_TypeError, "%.200s() takes no keyword arguments",
                 f->m_ml->ml_name);
    return NULL;
}
static CYTHON_INLINE PyObject *__Pyx_CyFunction_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    return __Pyx_CyFunction_CallMethod(func, ((PyCFunctionObject*)func)->m_self, arg, kw);
}
static PyObject *__Pyx_CyFunction_CallAsMethod(PyObject *func, PyObject *args, PyObject *kw) {
    PyObject *result;
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *) func;
    if ((cyfunc->flags & __Pyx_CYFUNCTION_CCLASS) && !(cyfunc->flags & __Pyx_CYFUNCTION_STATICMETHOD)) {
        Py_ssize_t argc;
        PyObject *new_args;
        PyObject *self;
        argc = PyTuple_GET_SIZE(args);
        new_args = PyTuple_GetSlice(args, 1, argc);
        if (unlikely(!new_args))
            return NULL;
        self = PyTuple_GetItem(args, 0);
        if (unlikely(!self)) {
            Py_DECREF(new_args);
#if PY_MAJOR_VERSION > 2
            PyErr_Format(PyExc_TypeError,
                         "unbound method %.200S() needs an argument",
                         cyfunc->func_qualname);
#else
            PyErr_SetString(PyExc_TypeError,
                            "unbound method needs an argument");
#endif
            return NULL;
        }
        result = __Pyx_CyFunction_CallMethod(func, self, new_args, kw);
        Py_DECREF(new_args);
    } else {
        result = __Pyx_CyFunction_Call(func, args, kw);
    }
    return result;
}
static PyTypeObject __pyx_CyFunctionType_type = {
    PyVarObject_HEAD_INIT(0, 0)
    "cython_function_or_method",
    sizeof(__pyx_CyFunctionObject),
    0,
    (destructor) __Pyx_CyFunction_dealloc,
    0,
    0,
    0,
#if PY_MAJOR_VERSION < 3
    0,
#else
    0,
#endif
    (reprfunc) __Pyx_CyFunction_repr,
    0,
    0,
    0,
    0,
    __Pyx_CyFunction_CallAsMethod,
    0,
    0,
    0,
    0,
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC,
    0,
    (traverseproc) __Pyx_CyFunction_traverse,
    (inquiry) __Pyx_CyFunction_clear,
    0,
#if PY_VERSION_HEX < 0x030500A0
    offsetof(__pyx_CyFunctionObject, func_weakreflist),
#else
    offsetof(PyCFunctionObject, m_weakreflist),
#endif
    0,
    0,
    __pyx_CyFunction_methods,
    __pyx_CyFunction_members,
    __pyx_CyFunction_getsets,
    0,
    0,
    __Pyx_CyFunction_descr_get,
    0,
    offsetof(__pyx_CyFunctionObject, func_dict),
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
#if PY_VERSION_HEX >= 0x030400a1
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
    0,
#endif
#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
    0,
#endif
};
static int __pyx_CyFunction_init(void) {
    __pyx_CyFunctionType = __Pyx_FetchCommonType(&__pyx_CyFunctionType_type);
    if (unlikely(__pyx_CyFunctionType == NULL)) {
        return -1;
    }
    return 0;
}
static CYTHON_INLINE void *__Pyx_CyFunction_InitDefaults(PyObject *func, size_t size, int pyobjects) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults = PyObject_Malloc(size);
    if (unlikely(!m->defaults))
        return PyErr_NoMemory();
    memset(m->defaults, 0, size);
    m->defaults_pyobjects = pyobjects;
    m->defaults_size = size;
    return m->defaults;
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *func, PyObject *tuple) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_tuple = tuple;
    Py_INCREF(tuple);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_kwdict = dict;
    Py_INCREF(dict);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->func_annotations = dict;
    Py_INCREF(dict);
}

/* CythonFunction */
static PyObject *__Pyx_CyFunction_New(PyMethodDef *ml, int flags, PyObject* qualname,
                                      PyObject *closure, PyObject *module, PyObject* globals, PyObject* code) {
    PyObject *op = __Pyx_CyFunction_Init(
        PyObject_GC_New(__pyx_CyFunctionObject, __pyx_CyFunctionType),
        ml, flags, qualname, closure, module, globals, code
    );
    if (likely(op)) {
        PyObject_GC_Track(op);
    }
    return op;
}

/* PyObject_GenericGetAttrNoDict */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject *__Pyx_RaiseGenericGetAttributeError(PyTypeObject *tp, PyObject *attr_name) {
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'%.50s' object has no attribute '%U'",
                 tp->tp_name, attr_name);
#else
                 "'%.50s' object has no attribute '%.400s'",
                 tp->tp_name, PyString_AS_STRING(attr_name));
#endif
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name) {
    PyObject *descr;
    PyTypeObject *tp = Py_TYPE(obj);
    if (unlikely(!PyString_Check(attr_name))) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    assert(!tp->tp_dictoffset);
    descr = _PyType_Lookup(tp, attr_name);
    if (unlikely(!descr)) {
        return __Pyx_RaiseGenericGetAttributeError(tp, attr_name);
    }
    Py_INCREF(descr);
    #if PY_MAJOR_VERSION < 3
    if (likely(PyType_HasFeature(Py_TYPE(descr), Py_TPFLAGS_HAVE_CLASS)))
    #endif
    {
        descrgetfunc f = Py_TYPE(descr)->tp_descr_get;
        if (unlikely(f)) {
            PyObject *res = f(descr, obj, (PyObject *)tp);
            Py_DECREF(descr);
            return res;
        }
    }
    return descr;
}
#endif

/* PyObject_GenericGetAttr */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name) {
    if (unlikely(Py_TYPE(obj)->tp_dictoffset)) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    return __Pyx_PyObject_GenericGetAttrNoDict(obj, attr_name);
}
#endif

/* SetVTable */
static int __Pyx_SetVtable(PyObject *dict, void *vtable) {
#if PY_VERSION_HEX >= 0x02070000
    PyObject *ob = PyCapsule_New(vtable, 0, 0);
#else
    PyObject *ob = PyCObject_FromVoidPtr(vtable, 0);
#endif
    if (!ob)
        goto bad;
    if (PyDict_SetItem(dict, __pyx_n_s_pyx_vtable, ob) < 0)
        goto bad;
    Py_DECREF(ob);
    return 0;
bad:
    Py_XDECREF(ob);
    return -1;
}

/* PyObjectGetAttrStrNoError */
static void __Pyx_PyObject_GetAttrStr_ClearAttributeError(void) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (likely(__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        __Pyx_PyErr_Clear();
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name) {
    PyObject *result;
#if CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_TYPE_SLOTS && PY_VERSION_HEX >= 0x030700B1
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro == PyObject_GenericGetAttr)) {
        return _PyObject_GenericGetAttrWithDict(obj, attr_name, NULL, 1);
    }
#endif
    result = __Pyx_PyObject_GetAttrStr(obj, attr_name);
    if (unlikely(!result)) {
        __Pyx_PyObject_GetAttrStr_ClearAttributeError();
    }
    return result;
}

/* SetupReduce */
static int __Pyx_setup_reduce_is_named(PyObject* meth, PyObject* name) {
  int ret;
  PyObject *name_attr;
  name_attr = __Pyx_PyObject_GetAttrStr(meth, __pyx_n_s_name);
  if (likely(name_attr)) {
      ret = PyObject_RichCompareBool(name_attr, name, Py_EQ);
  } else {
      ret = -1;
  }
  if (unlikely(ret < 0)) {
      PyErr_Clear();
      ret = 0;
  }
  Py_XDECREF(name_attr);
  return ret;
}
static int __Pyx_setup_reduce(PyObject* type_obj) {
    int ret = 0;
    PyObject *object_reduce = NULL;
    PyObject *object_getstate = NULL;
    PyObject *object_reduce_ex = NULL;
    PyObject *reduce = NULL;
    PyObject *reduce_ex = NULL;
    PyObject *reduce_cython = NULL;
    PyObject *setstate = NULL;
    PyObject *setstate_cython = NULL;
    PyObject *getstate = NULL;
#if CYTHON_USE_PYTYPE_LOOKUP
    getstate = _PyType_Lookup((PyTypeObject*)type_obj, __pyx_n_s_getstate);
#else
    getstate = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_getstate);
    if (!getstate && PyErr_Occurred()) {
        goto __PYX_BAD;
    }
#endif
    if (getstate) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_getstate = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_getstate);
#else
        object_getstate = __Pyx_PyObject_GetAttrStrNoError((PyObject*)&PyBaseObject_Type, __pyx_n_s_getstate);
        if (!object_getstate && PyErr_Occurred()) {
            goto __PYX_BAD;
        }
#endif
        if (object_getstate != getstate) {
            goto __PYX_GOOD;
        }
    }
#if CYTHON_USE_PYTYPE_LOOKUP
    object_reduce_ex = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#else
    object_reduce_ex = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#endif
    reduce_ex = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce_ex); if (unlikely(!reduce_ex)) goto __PYX_BAD;
    if (reduce_ex == object_reduce_ex) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_reduce = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto __PYX_BAD;
#else
        object_reduce = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto __PYX_BAD;
#endif
        reduce = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce); if (unlikely(!reduce)) goto __PYX_BAD;
        if (reduce == object_reduce || __Pyx_setup_reduce_is_named(reduce, __pyx_n_s_reduce_cython)) {
            reduce_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_reduce_cython);
            if (likely(reduce_cython)) {
                ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce, reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
            } else if (reduce == object_reduce || PyErr_Occurred()) {
                goto __PYX_BAD;
            }
            setstate = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_setstate);
            if (!setstate) PyErr_Clear();
            if (!setstate || __Pyx_setup_reduce_is_named(setstate, __pyx_n_s_setstate_cython)) {
                setstate_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_setstate_cython);
                if (likely(setstate_cython)) {
                    ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate, setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                    ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                } else if (!setstate || PyErr_Occurred()) {
                    goto __PYX_BAD;
                }
            }
            PyType_Modified((PyTypeObject*)type_obj);
        }
    }
    goto __PYX_GOOD;
__PYX_BAD:
    if (!PyErr_Occurred())
        PyErr_Format(PyExc_RuntimeError, "Unable to initialize pickling for %s", ((PyTypeObject*)type_obj)->tp_name);
    ret = -1;
__PYX_GOOD:
#if !CYTHON_USE_PYTYPE_LOOKUP
    Py_XDECREF(object_reduce);
    Py_XDECREF(object_reduce_ex);
    Py_XDECREF(object_getstate);
    Py_XDECREF(getstate);
#endif
    Py_XDECREF(reduce);
    Py_XDECREF(reduce_ex);
    Py_XDECREF(reduce_cython);
    Py_XDECREF(setstate);
    Py_XDECREF(setstate_cython);
    return ret;
}

/* TypeImport */
#ifndef __PYX_HAVE_RT_ImportType
#define __PYX_HAVE_RT_ImportType
static PyTypeObject *__Pyx_ImportType(PyObject *module, const char *module_name, const char *class_name,
    size_t size, enum __Pyx_ImportType_CheckSize check_size)
{
    PyObject *result = 0;
    char warning[200];
    Py_ssize_t basicsize;
#ifdef Py_LIMITED_API
    PyObject *py_basicsize;
#endif
    result = PyObject_GetAttrString(module, class_name);
    if (!result)
        goto bad;
    if (!PyType_Check(result)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s.%.200s is not a type object",
            module_name, class_name);
        goto bad;
    }
#ifndef Py_LIMITED_API
    basicsize = ((PyTypeObject *)result)->tp_basicsize;
#else
    py_basicsize = PyObject_GetAttrString(result, "__basicsize__");
    if (!py_basicsize)
        goto bad;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = 0;
    if (basicsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
#endif
    if ((size_t)basicsize < size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        goto bad;
    }
    if (check_size == __Pyx_ImportType_CheckSize_Error && (size_t)basicsize != size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        goto bad;
    }
    else if (check_size == __Pyx_ImportType_CheckSize_Warn && (size_t)basicsize > size) {
        PyOS_snprintf(warning, sizeof(warning),
            "%s.%s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        if (PyErr_WarnEx(NULL, warning, 0) < 0) goto bad;
    }
    return (PyTypeObject *)result;
bad:
    Py_XDECREF(result);
    return NULL;
}
#endif

/* GetVTable */
static void* __Pyx_GetVtable(PyObject *dict) {
    void* ptr;
    PyObject *ob = PyObject_GetItem(dict, __pyx_n_s_pyx_vtable);
    if (!ob)
        goto bad;
#if PY_VERSION_HEX >= 0x02070000
    ptr = PyCapsule_GetPointer(ob, 0);
#else
    ptr = PyCObject_AsVoidPtr(ob);
#endif
    if (!ptr && !PyErr_Occurred())
        PyErr_SetString(PyExc_RuntimeError, "invalid vtable found for imported type");
    Py_DECREF(ob);
    return ptr;
bad:
    Py_XDECREF(ob);
    return NULL;
}

/* GetNameInClass */
static PyObject *__Pyx_GetGlobalNameAfterAttributeLookup(PyObject *name) {
    PyObject *result;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (unlikely(!__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        return NULL;
    __Pyx_PyErr_Clear();
    __Pyx_GetModuleGlobalNameUncached(result, name);
    return result;
}
static PyObject *__Pyx__GetNameInClass(PyObject *nmspace, PyObject *name) {
    PyObject *result;
    result = __Pyx_PyObject_GetAttrStr(nmspace, name);
    if (!result) {
        result = __Pyx_GetGlobalNameAfterAttributeLookup(name);
    }
    return result;
}

/* FusedFunction */
static PyObject *
__pyx_FusedFunction_New(PyMethodDef *ml, int flags,
                        PyObject *qualname, PyObject *closure,
                        PyObject *module, PyObject *globals,
                        PyObject *code)
{
    PyObject *op = __Pyx_CyFunction_Init(
        PyObject_GC_New(__pyx_CyFunctionObject, __pyx_FusedFunctionType),
        ml, flags, qualname, closure, module, globals, code
    );
    if (likely(op)) {
        __pyx_FusedFunctionObject *fusedfunc = (__pyx_FusedFunctionObject *) op;
        fusedfunc->__signatures__ = NULL;
        fusedfunc->type = NULL;
        fusedfunc->self = NULL;
        PyObject_GC_Track(op);
    }
    return op;
}
static void
__pyx_FusedFunction_dealloc(__pyx_FusedFunctionObject *self)
{
    PyObject_GC_UnTrack(self);
    Py_CLEAR(self->self);
    Py_CLEAR(self->type);
    Py_CLEAR(self->__signatures__);
    __Pyx__CyFunction_dealloc((__pyx_CyFunctionObject *) self);
}
static int
__pyx_FusedFunction_traverse(__pyx_FusedFunctionObject *self,
                             visitproc visit,
                             void *arg)
{
    Py_VISIT(self->self);
    Py_VISIT(self->type);
    Py_VISIT(self->__signatures__);
    return __Pyx_CyFunction_traverse((__pyx_CyFunctionObject *) self, visit, arg);
}
static int
__pyx_FusedFunction_clear(__pyx_FusedFunctionObject *self)
{
    Py_CLEAR(self->self);
    Py_CLEAR(self->type);
    Py_CLEAR(self->__signatures__);
    return __Pyx_CyFunction_clear((__pyx_CyFunctionObject *) self);
}
static PyObject *
__pyx_FusedFunction_descr_get(PyObject *self, PyObject *obj, PyObject *type)
{
    __pyx_FusedFunctionObject *func, *meth;
    func = (__pyx_FusedFunctionObject *) self;
    if (func->self || func->func.flags & __Pyx_CYFUNCTION_STATICMETHOD) {
        Py_INCREF(self);
        return self;
    }
    if (obj == Py_None)
        obj = NULL;
    meth = (__pyx_FusedFunctionObject *) __pyx_FusedFunction_New(
                    ((PyCFunctionObject *) func)->m_ml,
                    ((__pyx_CyFunctionObject *) func)->flags,
                    ((__pyx_CyFunctionObject *) func)->func_qualname,
                    ((__pyx_CyFunctionObject *) func)->func_closure,
                    ((PyCFunctionObject *) func)->m_module,
                    ((__pyx_CyFunctionObject *) func)->func_globals,
                    ((__pyx_CyFunctionObject *) func)->func_code);
    if (!meth)
        return NULL;
    if (func->func.defaults) {
        PyObject **pydefaults;
        int i;
        if (!__Pyx_CyFunction_InitDefaults((PyObject*)meth,
                                      func->func.defaults_size,
                                      func->func.defaults_pyobjects)) {
            Py_XDECREF((PyObject*)meth);
            return NULL;
        }
        memcpy(meth->func.defaults, func->func.defaults, func->func.defaults_size);
        pydefaults = __Pyx_CyFunction_Defaults(PyObject *, meth);
        for (i = 0; i < meth->func.defaults_pyobjects; i++)
            Py_XINCREF(pydefaults[i]);
    }
    Py_XINCREF(func->func.func_classobj);
    meth->func.func_classobj = func->func.func_classobj;
    Py_XINCREF(func->__signatures__);
    meth->__signatures__ = func->__signatures__;
    Py_XINCREF(type);
    meth->type = type;
    Py_XINCREF(func->func.defaults_tuple);
    meth->func.defaults_tuple = func->func.defaults_tuple;
    if (func->func.flags & __Pyx_CYFUNCTION_CLASSMETHOD)
        obj = type;
    Py_XINCREF(obj);
    meth->self = obj;
    return (PyObject *) meth;
}
static PyObject *
_obj_to_str(PyObject *obj)
{
    if (PyType_Check(obj))
        return PyObject_GetAttr(obj, __pyx_n_s_name);
    else
        return PyObject_Str(obj);
}
static PyObject *
__pyx_FusedFunction_getitem(__pyx_FusedFunctionObject *self, PyObject *idx)
{
    PyObject *signature = NULL;
    PyObject *unbound_result_func;
    PyObject *result_func = NULL;
    if (self->__signatures__ == NULL) {
        PyErr_SetString(PyExc_TypeError, "Function is not fused");
        return NULL;
    }
    if (PyTuple_Check(idx)) {
        PyObject *list = PyList_New(0);
        Py_ssize_t n = PyTuple_GET_SIZE(idx);
        PyObject *sep = NULL;
        int i;
        if (unlikely(!list))
            return NULL;
        for (i = 0; i < n; i++) {
            int ret;
            PyObject *string;
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            PyObject *item = PyTuple_GET_ITEM(idx, i);
#else
            PyObject *item = PySequence_ITEM(idx, i);  if (unlikely(!item)) goto __pyx_err;
#endif
            string = _obj_to_str(item);
#if !(CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
            Py_DECREF(item);
#endif
            if (unlikely(!string)) goto __pyx_err;
            ret = PyList_Append(list, string);
            Py_DECREF(string);
            if (unlikely(ret < 0)) goto __pyx_err;
        }
        sep = PyUnicode_FromString("|");
        if (likely(sep))
            signature = PyUnicode_Join(sep, list);
__pyx_err:
;
        Py_DECREF(list);
        Py_XDECREF(sep);
    } else {
        signature = _obj_to_str(idx);
    }
    if (!signature)
        return NULL;
    unbound_result_func = PyObject_GetItem(self->__signatures__, signature);
    if (unbound_result_func) {
        if (self->self || self->type) {
            __pyx_FusedFunctionObject *unbound = (__pyx_FusedFunctionObject *) unbound_result_func;
            Py_CLEAR(unbound->func.func_classobj);
            Py_XINCREF(self->func.func_classobj);
            unbound->func.func_classobj = self->func.func_classobj;
            result_func = __pyx_FusedFunction_descr_get(unbound_result_func,
                                                        self->self, self->type);
        } else {
            result_func = unbound_result_func;
            Py_INCREF(result_func);
        }
    }
    Py_DECREF(signature);
    Py_XDECREF(unbound_result_func);
    return result_func;
}
static PyObject *
__pyx_FusedFunction_callfunction(PyObject *func, PyObject *args, PyObject *kw)
{
     __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *) func;
    int static_specialized = (cyfunc->flags & __Pyx_CYFUNCTION_STATICMETHOD &&
                              !((__pyx_FusedFunctionObject *) func)->__signatures__);
    if (cyfunc->flags & __Pyx_CYFUNCTION_CCLASS && !static_specialized) {
        return __Pyx_CyFunction_CallAsMethod(func, args, kw);
    } else {
        return __Pyx_CyFunction_Call(func, args, kw);
    }
}
static PyObject *
__pyx_FusedFunction_call(PyObject *func, PyObject *args, PyObject *kw)
{
    __pyx_FusedFunctionObject *binding_func = (__pyx_FusedFunctionObject *) func;
    Py_ssize_t argc = PyTuple_GET_SIZE(args);
    PyObject *new_args = NULL;
    __pyx_FusedFunctionObject *new_func = NULL;
    PyObject *result = NULL;
    PyObject *self = NULL;
    int is_staticmethod = binding_func->func.flags & __Pyx_CYFUNCTION_STATICMETHOD;
    int is_classmethod = binding_func->func.flags & __Pyx_CYFUNCTION_CLASSMETHOD;
    if (binding_func->self) {
        Py_ssize_t i;
        new_args = PyTuple_New(argc + 1);
        if (!new_args)
            return NULL;
        self = binding_func->self;
#if !(CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
        Py_INCREF(self);
#endif
        Py_INCREF(self);
        PyTuple_SET_ITEM(new_args, 0, self);
        for (i = 0; i < argc; i++) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            PyObject *item = PyTuple_GET_ITEM(args, i);
            Py_INCREF(item);
#else
            PyObject *item = PySequence_ITEM(args, i);  if (unlikely(!item)) goto bad;
#endif
            PyTuple_SET_ITEM(new_args, i + 1, item);
        }
        args = new_args;
    } else if (binding_func->type) {
        if (argc < 1) {
            PyErr_SetString(PyExc_TypeError, "Need at least one argument, 0 given.");
            return NULL;
        }
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        self = PyTuple_GET_ITEM(args, 0);
#else
        self = PySequence_ITEM(args, 0);  if (unlikely(!self)) return NULL;
#endif
    }
    if (self && !is_classmethod && !is_staticmethod) {
        int is_instance = PyObject_IsInstance(self, binding_func->type);
        if (unlikely(!is_instance)) {
            PyErr_Format(PyExc_TypeError,
                         "First argument should be of type %.200s, got %.200s.",
                         ((PyTypeObject *) binding_func->type)->tp_name,
                         Py_TYPE(self)->tp_name);
            goto bad;
        } else if (unlikely(is_instance == -1)) {
            goto bad;
        }
    }
#if !(CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
    Py_XDECREF(self);
    self = NULL;
#endif
    if (binding_func->__signatures__) {
        PyObject *tup;
        if (is_staticmethod && binding_func->func.flags & __Pyx_CYFUNCTION_CCLASS) {
            tup = PyTuple_Pack(3, args,
                               kw == NULL ? Py_None : kw,
                               binding_func->func.defaults_tuple);
            if (unlikely(!tup)) goto bad;
            new_func = (__pyx_FusedFunctionObject *) __Pyx_CyFunction_CallMethod(
                func, binding_func->__signatures__, tup, NULL);
        } else {
            tup = PyTuple_Pack(4, binding_func->__signatures__, args,
                               kw == NULL ? Py_None : kw,
                               binding_func->func.defaults_tuple);
            if (unlikely(!tup)) goto bad;
            new_func = (__pyx_FusedFunctionObject *) __pyx_FusedFunction_callfunction(func, tup, NULL);
        }
        Py_DECREF(tup);
        if (unlikely(!new_func))
            goto bad;
        Py_XINCREF(binding_func->func.func_classobj);
        Py_CLEAR(new_func->func.func_classobj);
        new_func->func.func_classobj = binding_func->func.func_classobj;
        func = (PyObject *) new_func;
    }
    result = __pyx_FusedFunction_callfunction(func, args, kw);
bad:
#if !(CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
    Py_XDECREF(self);
#endif
    Py_XDECREF(new_args);
    Py_XDECREF((PyObject *) new_func);
    return result;
}
static PyMemberDef __pyx_FusedFunction_members[] = {
    {(char *) "__signatures__",
     T_OBJECT,
     offsetof(__pyx_FusedFunctionObject, __signatures__),
     READONLY,
     0},
    {0, 0, 0, 0, 0},
};
static PyMappingMethods __pyx_FusedFunction_mapping_methods = {
    0,
    (binaryfunc) __pyx_FusedFunction_getitem,
    0,
};
static PyTypeObject __pyx_FusedFunctionType_type = {
    PyVarObject_HEAD_INIT(0, 0)
    "fused_cython_function",
    sizeof(__pyx_FusedFunctionObject),
    0,
    (destructor) __pyx_FusedFunction_dealloc,
    0,
    0,
    0,
#if PY_MAJOR_VERSION < 3
    0,
#else
    0,
#endif
    0,
    0,
    0,
    &__pyx_FusedFunction_mapping_methods,
    0,
    (ternaryfunc) __pyx_FusedFunction_call,
    0,
    0,
    0,
    0,
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_BASETYPE,
    0,
    (traverseproc) __pyx_FusedFunction_traverse,
    (inquiry) __pyx_FusedFunction_clear,
    0,
    0,
    0,
    0,
    0,
    __pyx_FusedFunction_members,
    __pyx_CyFunction_getsets,
    &__pyx_CyFunctionType_type,
    0,
    __pyx_FusedFunction_descr_get,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
#if PY_VERSION_HEX >= 0x030400a1
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
    0,
#endif
#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
    0,
#endif
};
static int __pyx_FusedFunction_init(void) {
    __pyx_FusedFunctionType_type.tp_base = __pyx_CyFunctionType;
    __pyx_FusedFunctionType = __Pyx_FetchCommonType(&__pyx_FusedFunctionType_type);
    if (__pyx_FusedFunctionType == NULL) {
        return -1;
    }
    return 0;
}

/* ClassMethod */
static PyObject* __Pyx_Method_ClassMethod(PyObject *method) {
#if CYTHON_COMPILING_IN_PYPY && PYPY_VERSION_NUM <= 0x05080000
    if (PyObject_TypeCheck(method, &PyWrapperDescr_Type)) {
        return PyClassMethod_New(method);
    }
#else
#if CYTHON_COMPILING_IN_PYSTON || CYTHON_COMPILING_IN_PYPY
    if (PyMethodDescr_Check(method))
#else
    #if PY_MAJOR_VERSION == 2
    static PyTypeObject *methoddescr_type = NULL;
    if (methoddescr_type == NULL) {
       PyObject *meth = PyObject_GetAttrString((PyObject*)&PyList_Type, "append");
       if (!meth) return NULL;
       methoddescr_type = Py_TYPE(meth);
       Py_DECREF(meth);
    }
    #else
    PyTypeObject *methoddescr_type = &PyMethodDescr_Type;
    #endif
    if (__Pyx_TypeCheck(method, methoddescr_type))
#endif
    {
        PyMethodDescrObject *descr = (PyMethodDescrObject *)method;
        #if PY_VERSION_HEX < 0x03020000
        PyTypeObject *d_type = descr->d_type;
        #else
        PyTypeObject *d_type = descr->d_common.d_type;
        #endif
        return PyDescr_NewClassMethod(d_type, descr->d_method);
    }
#endif
    else if (PyMethod_Check(method)) {
        return PyClassMethod_New(PyMethod_GET_FUNCTION(method));
    }
    else {
        return PyClassMethod_New(method);
    }
}

/* CLineInTraceback */
#ifndef CYTHON_CLINE_IN_TRACEBACK
static int __Pyx_CLineForTraceback(CYTHON_UNUSED PyThreadState *tstate, int c_line) {
    PyObject *use_cline;
    PyObject *ptype, *pvalue, *ptraceback;
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject **cython_runtime_dict;
#endif
    if (unlikely(!__pyx_cython_runtime)) {
        return c_line;
    }
    __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
#if CYTHON_COMPILING_IN_CPYTHON
    cython_runtime_dict = _PyObject_GetDictPtr(__pyx_cython_runtime);
    if (likely(cython_runtime_dict)) {
        __PYX_PY_DICT_LOOKUP_IF_MODIFIED(
            use_cline, *cython_runtime_dict,
            __Pyx_PyDict_GetItemStr(*cython_runtime_dict, __pyx_n_s_cline_in_traceback))
    } else
#endif
    {
      PyObject *use_cline_obj = __Pyx_PyObject_GetAttrStr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback);
      if (use_cline_obj) {
        use_cline = PyObject_Not(use_cline_obj) ? Py_False : Py_True;
        Py_DECREF(use_cline_obj);
      } else {
        PyErr_Clear();
        use_cline = NULL;
      }
    }
    if (!use_cline) {
        c_line = 0;
        (void) PyObject_SetAttr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback, Py_False);
    }
    else if (use_cline == Py_False || (use_cline != Py_True && PyObject_Not(use_cline) != 0)) {
        c_line = 0;
    }
    __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
    return c_line;
}
#endif

/* CodeObjectCache */
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, ((size_t)new_max) * sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}

/* AddTraceback */
#include "compile.h"
#include "frameobject.h"
#include "traceback.h"
#if PY_VERSION_HEX >= 0x030b00a6
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = NULL;
    PyObject *py_funcname = NULL;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_srcfile = NULL;
    py_srcfile = PyString_FromString(filename);
    if (!py_srcfile) goto bad;
    #endif
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        if (!py_funcname) goto bad;
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        if (!py_funcname) goto bad;
        funcname = PyUnicode_AsUTF8(py_funcname);
        if (!funcname) goto bad;
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        if (!py_funcname) goto bad;
        #endif
    }
    #if PY_MAJOR_VERSION < 3
    py_code = __Pyx_PyCode_New(
        0,
        0,
        0,
        0,
        0,
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    #else
    py_code = PyCode_NewEmpty(filename, funcname, py_line);
    #endif
    Py_XDECREF(py_funcname);  // XDECREF since it's only set on Py3 if cline
    return py_code;
bad:
    Py_XDECREF(py_funcname);
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_srcfile);
    #endif
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject *ptype, *pvalue, *ptraceback;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(tstate, c_line);
    }
    py_code = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!py_code) {
        __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) {
            /* If the code object creation fails, then we should clear the
               fetched exception references and propagate the new exception */
            Py_XDECREF(ptype);
            Py_XDECREF(pvalue);
            Py_XDECREF(ptraceback);
            goto bad;
        }
        __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
        __pyx_insert_code_object(c_line ? -c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        tstate,            /*PyThreadState *tstate,*/
        py_code,           /*PyCodeObject *code,*/
        __pyx_d,    /*PyObject *globals,*/
        0                  /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}

/* CIntFromPyVerify */
#define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

/* Print */
#if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION < 3
static PyObject *__Pyx_GetStdout(void) {
    PyObject *f = PySys_GetObject((char *)"stdout");
    if (!f) {
        PyErr_SetString(PyExc_RuntimeError, "lost sys.stdout");
    }
    return f;
}
static int __Pyx_Print(PyObject* f, PyObject *arg_tuple, int newline) {
    int i;
    if (!f) {
        if (!(f = __Pyx_GetStdout()))
            return -1;
    }
    Py_INCREF(f);
    for (i=0; i < PyTuple_GET_SIZE(arg_tuple); i++) {
        PyObject* v;
        if (PyFile_SoftSpace(f, 1)) {
            if (PyFile_WriteString(" ", f) < 0)
                goto error;
        }
        v = PyTuple_GET_ITEM(arg_tuple, i);
        if (PyFile_WriteObject(v, f, Py_PRINT_RAW) < 0)
            goto error;
        if (PyString_Check(v)) {
            char *s = PyString_AsString(v);
            Py_ssize_t len = PyString_Size(v);
            if (len > 0) {
                switch (s[len-1]) {
                    case ' ': break;
                    case '\f': case '\r': case '\n': case '\t': case '\v':
                        PyFile_SoftSpace(f, 0);
                        break;
                    default:  break;
                }
            }
        }
    }
    if (newline) {
        if (PyFile_WriteString("\n", f) < 0)
            goto error;
        PyFile_SoftSpace(f, 0);
    }
    Py_DECREF(f);
    return 0;
error:
    Py_DECREF(f);
    return -1;
}
#else
static int __Pyx_Print(PyObject* stream, PyObject *arg_tuple, int newline) {
    PyObject* kwargs = 0;
    PyObject* result = 0;
    PyObject* end_string;
    if (unlikely(!__pyx_print)) {
        __pyx_print = PyObject_GetAttr(__pyx_b, __pyx_n_s_print);
        if (!__pyx_print)
            return -1;
    }
    if (stream) {
        kwargs = PyDict_New();
        if (unlikely(!kwargs))
            return -1;
        if (unlikely(PyDict_SetItem(kwargs, __pyx_n_s_file, stream) < 0))
            goto bad;
        if (!newline) {
            end_string = PyUnicode_FromStringAndSize(" ", 1);
            if (unlikely(!end_string))
                goto bad;
            if (PyDict_SetItem(kwargs, __pyx_n_s_end, end_string) < 0) {
                Py_DECREF(end_string);
                goto bad;
            }
            Py_DECREF(end_string);
        }
    } else if (!newline) {
        if (unlikely(!__pyx_print_kwargs)) {
            __pyx_print_kwargs = PyDict_New();
            if (unlikely(!__pyx_print_kwargs))
                return -1;
            end_string = PyUnicode_FromStringAndSize(" ", 1);
            if (unlikely(!end_string))
                return -1;
            if (PyDict_SetItem(__pyx_print_kwargs, __pyx_n_s_end, end_string) < 0) {
                Py_DECREF(end_string);
                return -1;
            }
            Py_DECREF(end_string);
        }
        kwargs = __pyx_print_kwargs;
    }
    result = PyObject_Call(__pyx_print, arg_tuple, kwargs);
    if (unlikely(kwargs) && (kwargs != __pyx_print_kwargs))
        Py_DECREF(kwargs);
    if (!result)
        return -1;
    Py_DECREF(result);
    return 0;
bad:
    if (kwargs != __pyx_print_kwargs)
        Py_XDECREF(kwargs);
    return -1;
}
#endif

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(int),
                                     little, !is_unsigned);
    }
}

/* CIntFromPy */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(int) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case  1: __PYX_VERIFY_RETURN_INT(int, digit, digits[0])
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 2 * PyLong_SHIFT) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 3 * PyLong_SHIFT) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 4 * PyLong_SHIFT) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(int) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case -1: __PYX_VERIFY_RETURN_INT(int, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(int,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(int) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(int) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            int val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (int) -1;
        }
    } else {
        int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyInt_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* CIntFromPy */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(long) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (long) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case  1: __PYX_VERIFY_RETURN_INT(long, digit, digits[0])
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 2 * PyLong_SHIFT) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 3 * PyLong_SHIFT) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 4 * PyLong_SHIFT) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (long) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(long) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case -1: __PYX_VERIFY_RETURN_INT(long, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(long,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(long) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(long) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            long val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (long) -1;
        }
    } else {
        long val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyInt_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
    }
}

/* CIntFromPy */
static CYTHON_INLINE size_t __Pyx_PyInt_As_size_t(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const size_t neg_one = (size_t) -1, const_zero = (size_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(size_t) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(size_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (size_t) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (size_t) 0;
                case  1: __PYX_VERIFY_RETURN_INT(size_t, digit, digits[0])
                case 2:
                    if (8 * sizeof(size_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 2 * PyLong_SHIFT) {
                            return (size_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(size_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 3 * PyLong_SHIFT) {
                            return (size_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(size_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 4 * PyLong_SHIFT) {
                            return (size_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (size_t) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(size_t) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(size_t) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (size_t) 0;
                case -1: __PYX_VERIFY_RETURN_INT(size_t, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(size_t,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(size_t) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(size_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                            return (size_t) ((((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(size_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                            return (size_t) ((((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 4 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(size_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 4 * PyLong_SHIFT) {
                            return (size_t) ((((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(size_t) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(size_t) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            size_t val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (size_t) -1;
        }
    } else {
        size_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (size_t) -1;
        val = __Pyx_PyInt_As_size_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to size_t");
    return (size_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to size_t");
    return (size_t) -1;
}

/* PrintOne */
#if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION < 3
static int __Pyx_PrintOne(PyObject* f, PyObject *o) {
    if (!f) {
        if (!(f = __Pyx_GetStdout()))
            return -1;
    }
    Py_INCREF(f);
    if (PyFile_SoftSpace(f, 0)) {
        if (PyFile_WriteString(" ", f) < 0)
            goto error;
    }
    if (PyFile_WriteObject(o, f, Py_PRINT_RAW) < 0)
        goto error;
    if (PyFile_WriteString("\n", f) < 0)
        goto error;
    Py_DECREF(f);
    return 0;
error:
    Py_DECREF(f);
    return -1;
    /* the line below is just to avoid C compiler
     * warnings about unused functions */
    return __Pyx_Print(f, NULL, 0);
}
#else
static int __Pyx_PrintOne(PyObject* stream, PyObject *o) {
    int res;
    PyObject* arg_tuple = PyTuple_Pack(1, o);
    if (unlikely(!arg_tuple))
        return -1;
    res = __Pyx_Print(stream, arg_tuple, 1);
    Py_DECREF(arg_tuple);
    return res;
}
#endif

/* ImportNumPyArray */
static PyObject* __Pyx__ImportNumPyArray(void) {
    PyObject *numpy_module, *ndarray_object = NULL;
    numpy_module = __Pyx_Import(__pyx_n_s_numpy, NULL, 0);
    if (likely(numpy_module)) {
        ndarray_object = PyObject_GetAttrString(numpy_module, "ndarray");
        Py_DECREF(numpy_module);
    }
    if (unlikely(!ndarray_object)) {
        PyErr_Clear();
    }
    if (unlikely(!ndarray_object || !PyObject_TypeCheck(ndarray_object, &PyType_Type))) {
        Py_XDECREF(ndarray_object);
        Py_INCREF(Py_None);
        ndarray_object = Py_None;
    }
    return ndarray_object;
}
static CYTHON_INLINE PyObject* __Pyx_ImportNumPyArrayTypeIfAvailable(void) {
    if (unlikely(!__pyx_numpy_ndarray)) {
        __pyx_numpy_ndarray = __Pyx__ImportNumPyArray();
    }
    Py_INCREF(__pyx_numpy_ndarray);
    return __pyx_numpy_ndarray;
}

/* FastTypeChecks */
#if CYTHON_COMPILING_IN_CPYTHON
static int __Pyx_InBases(PyTypeObject *a, PyTypeObject *b) {
    while (a) {
        a = a->tp_base;
        if (a == b)
            return 1;
    }
    return b == &PyBaseObject_Type;
}
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (a == b) return 1;
    mro = a->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            if (PyTuple_GET_ITEM(mro, i) == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(a, b);
}
#if PY_MAJOR_VERSION == 2
static int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject* exc_type2) {
    PyObject *exception, *value, *tb;
    int res;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&exception, &value, &tb);
    res = exc_type1 ? PyObject_IsSubclass(err, exc_type1) : 0;
    if (unlikely(res == -1)) {
        PyErr_WriteUnraisable(err);
        res = 0;
    }
    if (!res) {
        res = PyObject_IsSubclass(err, exc_type2);
        if (unlikely(res == -1)) {
            PyErr_WriteUnraisable(err);
            res = 0;
        }
    }
    __Pyx_ErrRestore(exception, value, tb);
    return res;
}
#else
static CYTHON_INLINE int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject *exc_type2) {
    int res = exc_type1 ? __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type1) : 0;
    if (!res) {
        res = __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type2);
    }
    return res;
}
#endif
static int __Pyx_PyErr_GivenExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    assert(PyExceptionClass_Check(exc_type));
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        PyObject *t = PyTuple_GET_ITEM(tuple, i);
        #if PY_MAJOR_VERSION < 3
        if (likely(exc_type == t)) return 1;
        #endif
        if (likely(PyExceptionClass_Check(t))) {
            if (__Pyx_inner_PyErr_GivenExceptionMatches2(exc_type, NULL, t)) return 1;
        } else {
        }
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject* exc_type) {
    if (likely(err == exc_type)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        if (likely(PyExceptionClass_Check(exc_type))) {
            return __Pyx_inner_PyErr_GivenExceptionMatches2(err, NULL, exc_type);
        } else if (likely(PyTuple_Check(exc_type))) {
            return __Pyx_PyErr_GivenExceptionMatchesTuple(err, exc_type);
        } else {
        }
    }
    return PyErr_GivenExceptionMatches(err, exc_type);
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *exc_type1, PyObject *exc_type2) {
    assert(PyExceptionClass_Check(exc_type1));
    assert(PyExceptionClass_Check(exc_type2));
    if (likely(err == exc_type1 || err == exc_type2)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        return __Pyx_inner_PyErr_GivenExceptionMatches2(err, exc_type1, exc_type2);
    }
    return (PyErr_GivenExceptionMatches(err, exc_type1) || PyErr_GivenExceptionMatches(err, exc_type2));
}
#endif

/* SwapException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = *type;
    exc_info->exc_value = *value;
    exc_info->exc_traceback = *tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = *type;
    tstate->exc_value = *value;
    tstate->exc_traceback = *tb;
    #endif
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    PyErr_GetExcInfo(&tmp_type, &tmp_value, &tmp_tb);
    PyErr_SetExcInfo(*type, *value, *tb);
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#endif

/* PyObjectCallMethod1 */
static PyObject* __Pyx__PyObject_CallMethod1(PyObject* method, PyObject* arg) {
    PyObject *result = __Pyx_PyObject_CallOneArg(method, arg);
    Py_DECREF(method);
    return result;
}
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg) {
    PyObject *method = NULL, *result;
    int is_method = __Pyx_PyObject_GetMethod(obj, method_name, &method);
    if (likely(is_method)) {
        result = __Pyx_PyObject_Call2Args(method, obj, arg);
        Py_DECREF(method);
        return result;
    }
    if (unlikely(!method)) return NULL;
    return __Pyx__PyObject_CallMethod1(method, arg);
}

/* CoroutineBase */
#include <structmember.h>
#include <frameobject.h>
#if PY_VERSION_HEX >= 0x030b00a6
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
#define __Pyx_Coroutine_Undelegate(gen) Py_CLEAR((gen)->yieldfrom)
static int __Pyx_PyGen__FetchStopIterationValue(CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject **pvalue) {
    PyObject *et, *ev, *tb;
    PyObject *value = NULL;
    __Pyx_ErrFetch(&et, &ev, &tb);
    if (!et) {
        Py_XDECREF(tb);
        Py_XDECREF(ev);
        Py_INCREF(Py_None);
        *pvalue = Py_None;
        return 0;
    }
    if (likely(et == PyExc_StopIteration)) {
        if (!ev) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#if PY_VERSION_HEX >= 0x030300A0
        else if (Py_TYPE(ev) == (PyTypeObject*)PyExc_StopIteration) {
            value = ((PyStopIterationObject *)ev)->value;
            Py_INCREF(value);
            Py_DECREF(ev);
        }
#endif
        else if (unlikely(PyTuple_Check(ev))) {
            if (PyTuple_GET_SIZE(ev) >= 1) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                value = PyTuple_GET_ITEM(ev, 0);
                Py_INCREF(value);
#else
                value = PySequence_ITEM(ev, 0);
#endif
            } else {
                Py_INCREF(Py_None);
                value = Py_None;
            }
            Py_DECREF(ev);
        }
        else if (!__Pyx_TypeCheck(ev, (PyTypeObject*)PyExc_StopIteration)) {
            value = ev;
        }
        if (likely(value)) {
            Py_XDECREF(tb);
            Py_DECREF(et);
            *pvalue = value;
            return 0;
        }
    } else if (!__Pyx_PyErr_GivenExceptionMatches(et, PyExc_StopIteration)) {
        __Pyx_ErrRestore(et, ev, tb);
        return -1;
    }
    PyErr_NormalizeException(&et, &ev, &tb);
    if (unlikely(!PyObject_TypeCheck(ev, (PyTypeObject*)PyExc_StopIteration))) {
        __Pyx_ErrRestore(et, ev, tb);
        return -1;
    }
    Py_XDECREF(tb);
    Py_DECREF(et);
#if PY_VERSION_HEX >= 0x030300A0
    value = ((PyStopIterationObject *)ev)->value;
    Py_INCREF(value);
    Py_DECREF(ev);
#else
    {
        PyObject* args = __Pyx_PyObject_GetAttrStr(ev, __pyx_n_s_args);
        Py_DECREF(ev);
        if (likely(args)) {
            value = PySequence_GetItem(args, 0);
            Py_DECREF(args);
        }
        if (unlikely(!value)) {
            __Pyx_ErrRestore(NULL, NULL, NULL);
            Py_INCREF(Py_None);
            value = Py_None;
        }
    }
#endif
    *pvalue = value;
    return 0;
}
static CYTHON_INLINE
void __Pyx_Coroutine_ExceptionClear(__Pyx_ExcInfoStruct *exc_state) {
    PyObject *t, *v, *tb;
    t = exc_state->exc_type;
    v = exc_state->exc_value;
    tb = exc_state->exc_traceback;
    exc_state->exc_type = NULL;
    exc_state->exc_value = NULL;
    exc_state->exc_traceback = NULL;
    Py_XDECREF(t);
    Py_XDECREF(v);
    Py_XDECREF(tb);
}
#define __Pyx_Coroutine_AlreadyRunningError(gen)  (__Pyx__Coroutine_AlreadyRunningError(gen), (PyObject*)NULL)
static void __Pyx__Coroutine_AlreadyRunningError(CYTHON_UNUSED __pyx_CoroutineObject *gen) {
    const char *msg;
    if ((0)) {
    #ifdef __Pyx_Coroutine_USED
    } else if (__Pyx_Coroutine_Check((PyObject*)gen)) {
        msg = "coroutine already executing";
    #endif
    #ifdef __Pyx_AsyncGen_USED
    } else if (__Pyx_AsyncGen_CheckExact((PyObject*)gen)) {
        msg = "async generator already executing";
    #endif
    } else {
        msg = "generator already executing";
    }
    PyErr_SetString(PyExc_ValueError, msg);
}
#define __Pyx_Coroutine_NotStartedError(gen)  (__Pyx__Coroutine_NotStartedError(gen), (PyObject*)NULL)
static void __Pyx__Coroutine_NotStartedError(CYTHON_UNUSED PyObject *gen) {
    const char *msg;
    if ((0)) {
    #ifdef __Pyx_Coroutine_USED
    } else if (__Pyx_Coroutine_Check(gen)) {
        msg = "can't send non-None value to a just-started coroutine";
    #endif
    #ifdef __Pyx_AsyncGen_USED
    } else if (__Pyx_AsyncGen_CheckExact(gen)) {
        msg = "can't send non-None value to a just-started async generator";
    #endif
    } else {
        msg = "can't send non-None value to a just-started generator";
    }
    PyErr_SetString(PyExc_TypeError, msg);
}
#define __Pyx_Coroutine_AlreadyTerminatedError(gen, value, closing)  (__Pyx__Coroutine_AlreadyTerminatedError(gen, value, closing), (PyObject*)NULL)
static void __Pyx__Coroutine_AlreadyTerminatedError(CYTHON_UNUSED PyObject *gen, PyObject *value, CYTHON_UNUSED int closing) {
    #ifdef __Pyx_Coroutine_USED
    if (!closing && __Pyx_Coroutine_Check(gen)) {
        PyErr_SetString(PyExc_RuntimeError, "cannot reuse already awaited coroutine");
    } else
    #endif
    if (value) {
        #ifdef __Pyx_AsyncGen_USED
        if (__Pyx_AsyncGen_CheckExact(gen))
            PyErr_SetNone(__Pyx_PyExc_StopAsyncIteration);
        else
        #endif
        PyErr_SetNone(PyExc_StopIteration);
    }
}
static
PyObject *__Pyx_Coroutine_SendEx(__pyx_CoroutineObject *self, PyObject *value, int closing) {
    __Pyx_PyThreadState_declare
    PyThreadState *tstate;
    __Pyx_ExcInfoStruct *exc_state;
    PyObject *retval;
    assert(!self->is_running);
    if (unlikely(self->resume_label == 0)) {
        if (unlikely(value && value != Py_None)) {
            return __Pyx_Coroutine_NotStartedError((PyObject*)self);
        }
    }
    if (unlikely(self->resume_label == -1)) {
        return __Pyx_Coroutine_AlreadyTerminatedError((PyObject*)self, value, closing);
    }
#if CYTHON_FAST_THREAD_STATE
    __Pyx_PyThreadState_assign
    tstate = __pyx_tstate;
#else
    tstate = __Pyx_PyThreadState_Current;
#endif
    exc_state = &self->gi_exc_state;
    if (exc_state->exc_type) {
        #if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_PYSTON
        #else
        if (exc_state->exc_traceback) {
            PyTracebackObject *tb = (PyTracebackObject *) exc_state->exc_traceback;
            PyFrameObject *f = tb->tb_frame;
            assert(f->f_back == NULL);
            #if PY_VERSION_HEX >= 0x030B00A1
            f->f_back = PyThreadState_GetFrame(tstate);
            #else
            Py_XINCREF(tstate->frame);
            f->f_back = tstate->frame;
            #endif
        }
        #endif
    }
#if CYTHON_USE_EXC_INFO_STACK
    exc_state->previous_item = tstate->exc_info;
    tstate->exc_info = exc_state;
#else
    if (exc_state->exc_type) {
        __Pyx_ExceptionSwap(&exc_state->exc_type, &exc_state->exc_value, &exc_state->exc_traceback);
    } else {
        __Pyx_Coroutine_ExceptionClear(exc_state);
        __Pyx_ExceptionSave(&exc_state->exc_type, &exc_state->exc_value, &exc_state->exc_traceback);
    }
#endif
    self->is_running = 1;
    retval = self->body((PyObject *) self, tstate, value);
    self->is_running = 0;
#if CYTHON_USE_EXC_INFO_STACK
    exc_state = &self->gi_exc_state;
    tstate->exc_info = exc_state->previous_item;
    exc_state->previous_item = NULL;
    __Pyx_Coroutine_ResetFrameBackpointer(exc_state);
#endif
    return retval;
}
static CYTHON_INLINE void __Pyx_Coroutine_ResetFrameBackpointer(__Pyx_ExcInfoStruct *exc_state) {
    PyObject *exc_tb = exc_state->exc_traceback;
    if (likely(exc_tb)) {
#if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_PYSTON
#else
        PyTracebackObject *tb = (PyTracebackObject *) exc_tb;
        PyFrameObject *f = tb->tb_frame;
        Py_CLEAR(f->f_back);
#endif
    }
}
static CYTHON_INLINE
PyObject *__Pyx_Coroutine_MethodReturn(CYTHON_UNUSED PyObject* gen, PyObject *retval) {
    if (unlikely(!retval)) {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        if (!__Pyx_PyErr_Occurred()) {
            PyObject *exc = PyExc_StopIteration;
            #ifdef __Pyx_AsyncGen_USED
            if (__Pyx_AsyncGen_CheckExact(gen))
                exc = __Pyx_PyExc_StopAsyncIteration;
            #endif
            __Pyx_PyErr_SetNone(exc);
        }
    }
    return retval;
}
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03030000 && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
static CYTHON_INLINE
PyObject *__Pyx_PyGen_Send(PyGenObject *gen, PyObject *arg) {
#if PY_VERSION_HEX <= 0x030A00A1
    return _PyGen_Send(gen, arg);
#else
    PyObject *result;
    if (PyIter_Send((PyObject*)gen, arg ? arg : Py_None, &result) == PYGEN_RETURN) {
        if (PyAsyncGen_CheckExact(gen)) {
            assert(result == Py_None);
            PyErr_SetNone(PyExc_StopAsyncIteration);
        }
        else if (result == Py_None) {
            PyErr_SetNone(PyExc_StopIteration);
        }
        else {
            _PyGen_SetStopIterationValue(result);
        }
        Py_CLEAR(result);
    }
    return result;
#endif
}
#endif
static CYTHON_INLINE
PyObject *__Pyx_Coroutine_FinishDelegation(__pyx_CoroutineObject *gen) {
    PyObject *ret;
    PyObject *val = NULL;
    __Pyx_Coroutine_Undelegate(gen);
    __Pyx_PyGen__FetchStopIterationValue(__Pyx_PyThreadState_Current, &val);
    ret = __Pyx_Coroutine_SendEx(gen, val, 0);
    Py_XDECREF(val);
    return ret;
}
static PyObject *__Pyx_Coroutine_Send(PyObject *self, PyObject *value) {
    PyObject *retval;
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject*) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        PyObject *ret;
        gen->is_running = 1;
        #ifdef __Pyx_Generator_USED
        if (__Pyx_Generator_CheckExact(yf)) {
            ret = __Pyx_Coroutine_Send(yf, value);
        } else
        #endif
        #ifdef __Pyx_Coroutine_USED
        if (__Pyx_Coroutine_Check(yf)) {
            ret = __Pyx_Coroutine_Send(yf, value);
        } else
        #endif
        #ifdef __Pyx_AsyncGen_USED
        if (__pyx_PyAsyncGenASend_CheckExact(yf)) {
            ret = __Pyx_async_gen_asend_send(yf, value);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03030000 && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
        if (PyGen_CheckExact(yf)) {
            ret = __Pyx_PyGen_Send((PyGenObject*)yf, value == Py_None ? NULL : value);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03050000 && defined(PyCoro_CheckExact) && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
        if (PyCoro_CheckExact(yf)) {
            ret = __Pyx_PyGen_Send((PyGenObject*)yf, value == Py_None ? NULL : value);
        } else
        #endif
        {
            if (value == Py_None)
                ret = Py_TYPE(yf)->tp_iternext(yf);
            else
                ret = __Pyx_PyObject_CallMethod1(yf, __pyx_n_s_send, value);
        }
        gen->is_running = 0;
        if (likely(ret)) {
            return ret;
        }
        retval = __Pyx_Coroutine_FinishDelegation(gen);
    } else {
        retval = __Pyx_Coroutine_SendEx(gen, value, 0);
    }
    return __Pyx_Coroutine_MethodReturn(self, retval);
}
static int __Pyx_Coroutine_CloseIter(__pyx_CoroutineObject *gen, PyObject *yf) {
    PyObject *retval = NULL;
    int err = 0;
    #ifdef __Pyx_Generator_USED
    if (__Pyx_Generator_CheckExact(yf)) {
        retval = __Pyx_Coroutine_Close(yf);
        if (!retval)
            return -1;
    } else
    #endif
    #ifdef __Pyx_Coroutine_USED
    if (__Pyx_Coroutine_Check(yf)) {
        retval = __Pyx_Coroutine_Close(yf);
        if (!retval)
            return -1;
    } else
    if (__Pyx_CoroutineAwait_CheckExact(yf)) {
        retval = __Pyx_CoroutineAwait_Close((__pyx_CoroutineAwaitObject*)yf, NULL);
        if (!retval)
            return -1;
    } else
    #endif
    #ifdef __Pyx_AsyncGen_USED
    if (__pyx_PyAsyncGenASend_CheckExact(yf)) {
        retval = __Pyx_async_gen_asend_close(yf, NULL);
    } else
    if (__pyx_PyAsyncGenAThrow_CheckExact(yf)) {
        retval = __Pyx_async_gen_athrow_close(yf, NULL);
    } else
    #endif
    {
        PyObject *meth;
        gen->is_running = 1;
        meth = __Pyx_PyObject_GetAttrStr(yf, __pyx_n_s_close);
        if (unlikely(!meth)) {
            if (!PyErr_ExceptionMatches(PyExc_AttributeError)) {
                PyErr_WriteUnraisable(yf);
            }
            PyErr_Clear();
        } else {
            retval = PyObject_CallFunction(meth, NULL);
            Py_DECREF(meth);
            if (!retval)
                err = -1;
        }
        gen->is_running = 0;
    }
    Py_XDECREF(retval);
    return err;
}
static PyObject *__Pyx_Generator_Next(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject*) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        PyObject *ret;
        gen->is_running = 1;
        #ifdef __Pyx_Generator_USED
        if (__Pyx_Generator_CheckExact(yf)) {
            ret = __Pyx_Generator_Next(yf);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03030000 && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
        if (PyGen_CheckExact(yf)) {
            ret = __Pyx_PyGen_Send((PyGenObject*)yf, NULL);
        } else
        #endif
        #ifdef __Pyx_Coroutine_USED
        if (__Pyx_Coroutine_Check(yf)) {
            ret = __Pyx_Coroutine_Send(yf, Py_None);
        } else
        #endif
            ret = Py_TYPE(yf)->tp_iternext(yf);
        gen->is_running = 0;
        if (likely(ret)) {
            return ret;
        }
        return __Pyx_Coroutine_FinishDelegation(gen);
    }
    return __Pyx_Coroutine_SendEx(gen, Py_None, 0);
}
static PyObject *__Pyx_Coroutine_Close_Method(PyObject *self, CYTHON_UNUSED PyObject *arg) {
    return __Pyx_Coroutine_Close(self);
}
static PyObject *__Pyx_Coroutine_Close(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject *retval, *raised_exception;
    PyObject *yf = gen->yieldfrom;
    int err = 0;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        Py_INCREF(yf);
        err = __Pyx_Coroutine_CloseIter(gen, yf);
        __Pyx_Coroutine_Undelegate(gen);
        Py_DECREF(yf);
    }
    if (err == 0)
        PyErr_SetNone(PyExc_GeneratorExit);
    retval = __Pyx_Coroutine_SendEx(gen, NULL, 1);
    if (unlikely(retval)) {
        const char *msg;
        Py_DECREF(retval);
        if ((0)) {
        #ifdef __Pyx_Coroutine_USED
        } else if (__Pyx_Coroutine_Check(self)) {
            msg = "coroutine ignored GeneratorExit";
        #endif
        #ifdef __Pyx_AsyncGen_USED
        } else if (__Pyx_AsyncGen_CheckExact(self)) {
#if PY_VERSION_HEX < 0x03060000
            msg = "async generator ignored GeneratorExit - might require Python 3.6+ finalisation (PEP 525)";
#else
            msg = "async generator ignored GeneratorExit";
#endif
        #endif
        } else {
            msg = "generator ignored GeneratorExit";
        }
        PyErr_SetString(PyExc_RuntimeError, msg);
        return NULL;
    }
    raised_exception = PyErr_Occurred();
    if (likely(!raised_exception || __Pyx_PyErr_GivenExceptionMatches2(raised_exception, PyExc_GeneratorExit, PyExc_StopIteration))) {
        if (raised_exception) PyErr_Clear();
        Py_INCREF(Py_None);
        return Py_None;
    }
    return NULL;
}
static PyObject *__Pyx__Coroutine_Throw(PyObject *self, PyObject *typ, PyObject *val, PyObject *tb,
                                        PyObject *args, int close_on_genexit) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        PyObject *ret;
        Py_INCREF(yf);
        if (__Pyx_PyErr_GivenExceptionMatches(typ, PyExc_GeneratorExit) && close_on_genexit) {
            int err = __Pyx_Coroutine_CloseIter(gen, yf);
            Py_DECREF(yf);
            __Pyx_Coroutine_Undelegate(gen);
            if (err < 0)
                return __Pyx_Coroutine_MethodReturn(self, __Pyx_Coroutine_SendEx(gen, NULL, 0));
            goto throw_here;
        }
        gen->is_running = 1;
        if (0
        #ifdef __Pyx_Generator_USED
            || __Pyx_Generator_CheckExact(yf)
        #endif
        #ifdef __Pyx_Coroutine_USED
            || __Pyx_Coroutine_Check(yf)
        #endif
            ) {
            ret = __Pyx__Coroutine_Throw(yf, typ, val, tb, args, close_on_genexit);
        #ifdef __Pyx_Coroutine_USED
        } else if (__Pyx_CoroutineAwait_CheckExact(yf)) {
            ret = __Pyx__Coroutine_Throw(((__pyx_CoroutineAwaitObject*)yf)->coroutine, typ, val, tb, args, close_on_genexit);
        #endif
        } else {
            PyObject *meth = __Pyx_PyObject_GetAttrStr(yf, __pyx_n_s_throw);
            if (unlikely(!meth)) {
                Py_DECREF(yf);
                if (!PyErr_ExceptionMatches(PyExc_AttributeError)) {
                    gen->is_running = 0;
                    return NULL;
                }
                PyErr_Clear();
                __Pyx_Coroutine_Undelegate(gen);
                gen->is_running = 0;
                goto throw_here;
            }
            if (likely(args)) {
                ret = PyObject_CallObject(meth, args);
            } else {
                ret = PyObject_CallFunctionObjArgs(meth, typ, val, tb, NULL);
            }
            Py_DECREF(meth);
        }
        gen->is_running = 0;
        Py_DECREF(yf);
        if (!ret) {
            ret = __Pyx_Coroutine_FinishDelegation(gen);
        }
        return __Pyx_Coroutine_MethodReturn(self, ret);
    }
throw_here:
    __Pyx_Raise(typ, val, tb, NULL);
    return __Pyx_Coroutine_MethodReturn(self, __Pyx_Coroutine_SendEx(gen, NULL, 0));
}
static PyObject *__Pyx_Coroutine_Throw(PyObject *self, PyObject *args) {
    PyObject *typ;
    PyObject *val = NULL;
    PyObject *tb = NULL;
    if (!PyArg_UnpackTuple(args, (char *)"throw", 1, 3, &typ, &val, &tb))
        return NULL;
    return __Pyx__Coroutine_Throw(self, typ, val, tb, args, 1);
}
static CYTHON_INLINE int __Pyx_Coroutine_traverse_excstate(__Pyx_ExcInfoStruct *exc_state, visitproc visit, void *arg) {
    Py_VISIT(exc_state->exc_type);
    Py_VISIT(exc_state->exc_value);
    Py_VISIT(exc_state->exc_traceback);
    return 0;
}
static int __Pyx_Coroutine_traverse(__pyx_CoroutineObject *gen, visitproc visit, void *arg) {
    Py_VISIT(gen->closure);
    Py_VISIT(gen->classobj);
    Py_VISIT(gen->yieldfrom);
    return __Pyx_Coroutine_traverse_excstate(&gen->gi_exc_state, visit, arg);
}
static int __Pyx_Coroutine_clear(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    Py_CLEAR(gen->closure);
    Py_CLEAR(gen->classobj);
    Py_CLEAR(gen->yieldfrom);
    __Pyx_Coroutine_ExceptionClear(&gen->gi_exc_state);
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        Py_CLEAR(((__pyx_PyAsyncGenObject*)gen)->ag_finalizer);
    }
#endif
    Py_CLEAR(gen->gi_code);
    Py_CLEAR(gen->gi_frame);
    Py_CLEAR(gen->gi_name);
    Py_CLEAR(gen->gi_qualname);
    Py_CLEAR(gen->gi_modulename);
    return 0;
}
static void __Pyx_Coroutine_dealloc(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject_GC_UnTrack(gen);
    if (gen->gi_weakreflist != NULL)
        PyObject_ClearWeakRefs(self);
    if (gen->resume_label >= 0) {
        PyObject_GC_Track(self);
#if PY_VERSION_HEX >= 0x030400a1 && CYTHON_USE_TP_FINALIZE
        if (PyObject_CallFinalizerFromDealloc(self))
#else
        Py_TYPE(gen)->tp_del(self);
        if (Py_REFCNT(self) > 0)
#endif
        {
            return;
        }
        PyObject_GC_UnTrack(self);
    }
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        /* We have to handle this case for asynchronous generators
           right here, because this code has to be between UNTRACK
           and GC_Del. */
        Py_CLEAR(((__pyx_PyAsyncGenObject*)self)->ag_finalizer);
    }
#endif
    __Pyx_Coroutine_clear(self);
    PyObject_GC_Del(gen);
}
static void __Pyx_Coroutine_del(PyObject *self) {
    PyObject *error_type, *error_value, *error_traceback;
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    __Pyx_PyThreadState_declare
    if (gen->resume_label < 0) {
        return;
    }
#if !CYTHON_USE_TP_FINALIZE
    assert(self->ob_refcnt == 0);
    __Pyx_SET_REFCNT(self, 1);
#endif
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&error_type, &error_value, &error_traceback);
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        __pyx_PyAsyncGenObject *agen = (__pyx_PyAsyncGenObject*)self;
        PyObject *finalizer = agen->ag_finalizer;
        if (finalizer && !agen->ag_closed) {
            PyObject *res = __Pyx_PyObject_CallOneArg(finalizer, self);
            if (unlikely(!res)) {
                PyErr_WriteUnraisable(self);
            } else {
                Py_DECREF(res);
            }
            __Pyx_ErrRestore(error_type, error_value, error_traceback);
            return;
        }
    }
#endif
    if (unlikely(gen->resume_label == 0 && !error_value)) {
#ifdef __Pyx_Coroutine_USED
#ifdef __Pyx_Generator_USED
    if (!__Pyx_Generator_CheckExact(self))
#endif
        {
        PyObject_GC_UnTrack(self);
#if PY_MAJOR_VERSION >= 3  || defined(PyErr_WarnFormat)
        if (unlikely(PyErr_WarnFormat(PyExc_RuntimeWarning, 1, "coroutine '%.50S' was never awaited", gen->gi_qualname) < 0))
            PyErr_WriteUnraisable(self);
#else
        {PyObject *msg;
        char *cmsg;
        #if CYTHON_COMPILING_IN_PYPY
        msg = NULL;
        cmsg = (char*) "coroutine was never awaited";
        #else
        char *cname;
        PyObject *qualname;
        qualname = gen->gi_qualname;
        cname = PyString_AS_STRING(qualname);
        msg = PyString_FromFormat("coroutine '%.50s' was never awaited", cname);
        if (unlikely(!msg)) {
            PyErr_Clear();
            cmsg = (char*) "coroutine was never awaited";
        } else {
            cmsg = PyString_AS_STRING(msg);
        }
        #endif
        if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning, cmsg, 1) < 0))
            PyErr_WriteUnraisable(self);
        Py_XDECREF(msg);}
#endif
        PyObject_GC_Track(self);
        }
#endif
    } else {
        PyObject *res = __Pyx_Coroutine_Close(self);
        if (unlikely(!res)) {
            if (PyErr_Occurred())
                PyErr_WriteUnraisable(self);
        } else {
            Py_DECREF(res);
        }
    }
    __Pyx_ErrRestore(error_type, error_value, error_traceback);
#if !CYTHON_USE_TP_FINALIZE
    assert(Py_REFCNT(self) > 0);
    if (--self->ob_refcnt == 0) {
        return;
    }
    {
        Py_ssize_t refcnt = Py_REFCNT(self);
        _Py_NewReference(self);
        __Pyx_SET_REFCNT(self, refcnt);
    }
#if CYTHON_COMPILING_IN_CPYTHON
    assert(PyType_IS_GC(Py_TYPE(self)) &&
           _Py_AS_GC(self)->gc.gc_refs != _PyGC_REFS_UNTRACKED);
    _Py_DEC_REFTOTAL;
#endif
#ifdef COUNT_ALLOCS
    --Py_TYPE(self)->tp_frees;
    --Py_TYPE(self)->tp_allocs;
#endif
#endif
}
static PyObject *
__Pyx_Coroutine_get_name(__pyx_CoroutineObject *self, CYTHON_UNUSED void *context)
{
    PyObject *name = self->gi_name;
    if (unlikely(!name)) name = Py_None;
    Py_INCREF(name);
    return name;
}
static int
__Pyx_Coroutine_set_name(__pyx_CoroutineObject *self, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__name__ must be set to a string object");
        return -1;
    }
    tmp = self->gi_name;
    Py_INCREF(value);
    self->gi_name = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_Coroutine_get_qualname(__pyx_CoroutineObject *self, CYTHON_UNUSED void *context)
{
    PyObject *name = self->gi_qualname;
    if (unlikely(!name)) name = Py_None;
    Py_INCREF(name);
    return name;
}
static int
__Pyx_Coroutine_set_qualname(__pyx_CoroutineObject *self, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__qualname__ must be set to a string object");
        return -1;
    }
    tmp = self->gi_qualname;
    Py_INCREF(value);
    self->gi_qualname = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_Coroutine_get_frame(__pyx_CoroutineObject *self, CYTHON_UNUSED void *context)
{
    PyObject *frame = self->gi_frame;
    if (!frame) {
        if (unlikely(!self->gi_code)) {
            Py_RETURN_NONE;
        }
        frame = (PyObject *) PyFrame_New(
            PyThreadState_Get(),            /*PyThreadState *tstate,*/
            (PyCodeObject*) self->gi_code,  /*PyCodeObject *code,*/
            __pyx_d,                 /*PyObject *globals,*/
            0                               /*PyObject *locals*/
        );
        if (unlikely(!frame))
            return NULL;
        self->gi_frame = frame;
    }
    Py_INCREF(frame);
    return frame;
}
static __pyx_CoroutineObject *__Pyx__Coroutine_New(
            PyTypeObject* type, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name) {
    __pyx_CoroutineObject *gen = PyObject_GC_New(__pyx_CoroutineObject, type);
    if (unlikely(!gen))
        return NULL;
    return __Pyx__Coroutine_NewInit(gen, body, code, closure, name, qualname, module_name);
}
static __pyx_CoroutineObject *__Pyx__Coroutine_NewInit(
            __pyx_CoroutineObject *gen, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name) {
    gen->body = body;
    gen->closure = closure;
    Py_XINCREF(closure);
    gen->is_running = 0;
    gen->resume_label = 0;
    gen->classobj = NULL;
    gen->yieldfrom = NULL;
    gen->gi_exc_state.exc_type = NULL;
    gen->gi_exc_state.exc_value = NULL;
    gen->gi_exc_state.exc_traceback = NULL;
#if CYTHON_USE_EXC_INFO_STACK
    gen->gi_exc_state.previous_item = NULL;
#endif
    gen->gi_weakreflist = NULL;
    Py_XINCREF(qualname);
    gen->gi_qualname = qualname;
    Py_XINCREF(name);
    gen->gi_name = name;
    Py_XINCREF(module_name);
    gen->gi_modulename = module_name;
    Py_XINCREF(code);
    gen->gi_code = code;
    gen->gi_frame = NULL;
    PyObject_GC_Track(gen);
    return gen;
}

/* PatchModuleWithCoroutine */
static PyObject* __Pyx_Coroutine_patch_module(PyObject* module, const char* py_code) {
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
    int result;
    PyObject *globals, *result_obj;
    globals = PyDict_New();  if (unlikely(!globals)) goto ignore;
    result = PyDict_SetItemString(globals, "_cython_coroutine_type",
    #ifdef __Pyx_Coroutine_USED
        (PyObject*)__pyx_CoroutineType);
    #else
        Py_None);
    #endif
    if (unlikely(result < 0)) goto ignore;
    result = PyDict_SetItemString(globals, "_cython_generator_type",
    #ifdef __Pyx_Generator_USED
        (PyObject*)__pyx_GeneratorType);
    #else
        Py_None);
    #endif
    if (unlikely(result < 0)) goto ignore;
    if (unlikely(PyDict_SetItemString(globals, "_module", module) < 0)) goto ignore;
    if (unlikely(PyDict_SetItemString(globals, "__builtins__", __pyx_b) < 0)) goto ignore;
    result_obj = PyRun_String(py_code, Py_file_input, globals, globals);
    if (unlikely(!result_obj)) goto ignore;
    Py_DECREF(result_obj);
    Py_DECREF(globals);
    return module;
ignore:
    Py_XDECREF(globals);
    PyErr_WriteUnraisable(module);
    if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning, "Cython module failed to patch module with custom type", 1) < 0)) {
        Py_DECREF(module);
        module = NULL;
    }
#else
    py_code++;
#endif
    return module;
}

/* PatchGeneratorABC */
#ifndef CYTHON_REGISTER_ABCS
#define CYTHON_REGISTER_ABCS 1
#endif
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
static PyObject* __Pyx_patch_abc_module(PyObject *module);
static PyObject* __Pyx_patch_abc_module(PyObject *module) {
    module = __Pyx_Coroutine_patch_module(
        module, ""
"if _cython_generator_type is not None:\n"
"    try: Generator = _module.Generator\n"
"    except AttributeError: pass\n"
"    else: Generator.register(_cython_generator_type)\n"
"if _cython_coroutine_type is not None:\n"
"    try: Coroutine = _module.Coroutine\n"
"    except AttributeError: pass\n"
"    else: Coroutine.register(_cython_coroutine_type)\n"
    );
    return module;
}
#endif
static int __Pyx_patch_abc(void) {
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
    static int abc_patched = 0;
    if (CYTHON_REGISTER_ABCS && !abc_patched) {
        PyObject *module;
        module = PyImport_ImportModule((PY_MAJOR_VERSION >= 3) ? "collections.abc" : "collections");
        if (!module) {
            PyErr_WriteUnraisable(NULL);
            if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning,
                    ((PY_MAJOR_VERSION >= 3) ?
                        "Cython module failed to register with collections.abc module" :
                        "Cython module failed to register with collections module"), 1) < 0)) {
                return -1;
            }
        } else {
            module = __Pyx_patch_abc_module(module);
            abc_patched = 1;
            if (unlikely(!module))
                return -1;
            Py_DECREF(module);
        }
        module = PyImport_ImportModule("backports_abc");
        if (module) {
            module = __Pyx_patch_abc_module(module);
            Py_XDECREF(module);
        }
        if (!module) {
            PyErr_Clear();
        }
    }
#else
    if ((0)) __Pyx_Coroutine_patch_module(NULL, NULL);
#endif
    return 0;
}

/* Generator */
static PyMethodDef __pyx_Generator_methods[] = {
    {"send", (PyCFunction) __Pyx_Coroutine_Send, METH_O,
     (char*) PyDoc_STR("send(arg) -> send 'arg' into generator,\nreturn next yielded value or raise StopIteration.")},
    {"throw", (PyCFunction) __Pyx_Coroutine_Throw, METH_VARARGS,
     (char*) PyDoc_STR("throw(typ[,val[,tb]]) -> raise exception in generator,\nreturn next yielded value or raise StopIteration.")},
    {"close", (PyCFunction) __Pyx_Coroutine_Close_Method, METH_NOARGS,
     (char*) PyDoc_STR("close() -> raise GeneratorExit inside generator.")},
    {0, 0, 0, 0}
};
static PyMemberDef __pyx_Generator_memberlist[] = {
    {(char *) "gi_running", T_BOOL, offsetof(__pyx_CoroutineObject, is_running), READONLY, NULL},
    {(char*) "gi_yieldfrom", T_OBJECT, offsetof(__pyx_CoroutineObject, yieldfrom), READONLY,
     (char*) PyDoc_STR("object being iterated by 'yield from', or None")},
    {(char*) "gi_code", T_OBJECT, offsetof(__pyx_CoroutineObject, gi_code), READONLY, NULL},
    {0, 0, 0, 0, 0}
};
static PyGetSetDef __pyx_Generator_getsets[] = {
    {(char *) "__name__", (getter)__Pyx_Coroutine_get_name, (setter)__Pyx_Coroutine_set_name,
     (char*) PyDoc_STR("name of the generator"), 0},
    {(char *) "__qualname__", (getter)__Pyx_Coroutine_get_qualname, (setter)__Pyx_Coroutine_set_qualname,
     (char*) PyDoc_STR("qualified name of the generator"), 0},
    {(char *) "gi_frame", (getter)__Pyx_Coroutine_get_frame, NULL,
     (char*) PyDoc_STR("Frame of the generator"), 0},
    {0, 0, 0, 0, 0}
};
static PyTypeObject __pyx_GeneratorType_type = {
    PyVarObject_HEAD_INIT(0, 0)
    "generator",
    sizeof(__pyx_CoroutineObject),
    0,
    (destructor) __Pyx_Coroutine_dealloc,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_HAVE_FINALIZE,
    0,
    (traverseproc) __Pyx_Coroutine_traverse,
    0,
    0,
    offsetof(__pyx_CoroutineObject, gi_weakreflist),
    0,
    (iternextfunc) __Pyx_Generator_Next,
    __pyx_Generator_methods,
    __pyx_Generator_memberlist,
    __pyx_Generator_getsets,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
#if CYTHON_USE_TP_FINALIZE
    0,
#else
    __Pyx_Coroutine_del,
#endif
    0,
#if CYTHON_USE_TP_FINALIZE
    __Pyx_Coroutine_del,
#elif PY_VERSION_HEX >= 0x030400a1
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
    0,
#endif
#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000
    0,
#endif
};
static int __pyx_Generator_init(void) {
    __pyx_GeneratorType_type.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
    __pyx_GeneratorType_type.tp_iter = PyObject_SelfIter;
    __pyx_GeneratorType = __Pyx_FetchCommonType(&__pyx_GeneratorType_type);
    if (unlikely(!__pyx_GeneratorType)) {
        return -1;
    }
    return 0;
}

/* CheckBinaryVersion */
static int __Pyx_check_binary_version(void) {
    char ctversion[5];
    int same=1, i, found_dot;
    const char* rt_from_call = Py_GetVersion();
    PyOS_snprintf(ctversion, 5, "%d.%d", PY_MAJOR_VERSION, PY_MINOR_VERSION);
    found_dot = 0;
    for (i = 0; i < 4; i++) {
        if (!ctversion[i]) {
            same = (rt_from_call[i] < '0' || rt_from_call[i] > '9');
            break;
        }
        if (rt_from_call[i] != ctversion[i]) {
            same = 0;
            break;
        }
    }
    if (!same) {
        char rtversion[5] = {'\0'};
        char message[200];
        for (i=0; i<4; ++i) {
            if (rt_from_call[i] == '.') {
                if (found_dot) break;
                found_dot = 1;
            } else if (rt_from_call[i] < '0' || rt_from_call[i] > '9') {
                break;
            }
            rtversion[i] = rt_from_call[i];
        }
        PyOS_snprintf(message, sizeof(message),
                      "compiletime version %s of module '%.100s' "
                      "does not match runtime version %s",
                      ctversion, __Pyx_MODULE_NAME, rtversion);
        return PyErr_WarnEx(NULL, message, 1);
    }
    return 0;
}

/* FunctionExport */
static int __Pyx_ExportFunction(const char *name, void (*f)(void), const char *sig) {
    PyObject *d = 0;
    PyObject *cobj = 0;
    union {
        void (*fp)(void);
        void *p;
    } tmp;
    d = PyObject_GetAttrString(__pyx_m, (char *)"__pyx_capi__");
    if (!d) {
        PyErr_Clear();
        d = PyDict_New();
        if (!d)
            goto bad;
        Py_INCREF(d);
        if (PyModule_AddObject(__pyx_m, (char *)"__pyx_capi__", d) < 0)
            goto bad;
    }
    tmp.fp = f;
#if PY_VERSION_HEX >= 0x02070000
    cobj = PyCapsule_New(tmp.p, sig, 0);
#else
    cobj = PyCObject_FromVoidPtrAndDesc(tmp.p, (void *)sig, 0);
#endif
    if (!cobj)
        goto bad;
    if (PyDict_SetItemString(d, name, cobj) < 0)
        goto bad;
    Py_DECREF(cobj);
    Py_DECREF(d);
    return 0;
bad:
    Py_XDECREF(cobj);
    Py_XDECREF(d);
    return -1;
}

/* VoidPtrImport */
#ifndef __PYX_HAVE_RT_ImportVoidPtr
#define __PYX_HAVE_RT_ImportVoidPtr
static int __Pyx_ImportVoidPtr(PyObject *module, const char *name, void **p, const char *sig) {
    PyObject *d = 0;
    PyObject *cobj = 0;
    d = PyObject_GetAttrString(module, (char *)"__pyx_capi__");
    if (!d)
        goto bad;
    cobj = PyDict_GetItemString(d, name);
    if (!cobj) {
        PyErr_Format(PyExc_ImportError,
            "%.200s does not export expected C variable %.200s",
                PyModule_GetName(module), name);
        goto bad;
    }
#if PY_VERSION_HEX >= 0x02070000
    if (!PyCapsule_IsValid(cobj, sig)) {
        PyErr_Format(PyExc_TypeError,
            "C variable %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), name, sig, PyCapsule_GetName(cobj));
        goto bad;
    }
    *p = PyCapsule_GetPointer(cobj, sig);
#else
    {const char *desc, *s1, *s2;
    desc = (const char *)PyCObject_GetDesc(cobj);
    if (!desc)
        goto bad;
    s1 = desc; s2 = sig;
    while (*s1 != '\0' && *s1 == *s2) { s1++; s2++; }
    if (*s1 != *s2) {
        PyErr_Format(PyExc_TypeError,
            "C variable %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), name, sig, desc);
        goto bad;
    }
    *p = PyCObject_AsVoidPtr(cobj);}
#endif
    if (!(*p))
        goto bad;
    Py_DECREF(d);
    return 0;
bad:
    Py_XDECREF(d);
    return -1;
}
#endif

/* FunctionImport */
#ifndef __PYX_HAVE_RT_ImportFunction
#define __PYX_HAVE_RT_ImportFunction
static int __Pyx_ImportFunction(PyObject *module, const char *funcname, void (**f)(void), const char *sig) {
    PyObject *d = 0;
    PyObject *cobj = 0;
    union {
        void (*fp)(void);
        void *p;
    } tmp;
    d = PyObject_GetAttrString(module, (char *)"__pyx_capi__");
    if (!d)
        goto bad;
    cobj = PyDict_GetItemString(d, funcname);
    if (!cobj) {
        PyErr_Format(PyExc_ImportError,
            "%.200s does not export expected C function %.200s",
                PyModule_GetName(module), funcname);
        goto bad;
    }
#if PY_VERSION_HEX >= 0x02070000
    if (!PyCapsule_IsValid(cobj, sig)) {
        PyErr_Format(PyExc_TypeError,
            "C function %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), funcname, sig, PyCapsule_GetName(cobj));
        goto bad;
    }
    tmp.p = PyCapsule_GetPointer(cobj, sig);
#else
    {const char *desc, *s1, *s2;
    desc = (const char *)PyCObject_GetDesc(cobj);
    if (!desc)
        goto bad;
    s1 = desc; s2 = sig;
    while (*s1 != '\0' && *s1 == *s2) { s1++; s2++; }
    if (*s1 != *s2) {
        PyErr_Format(PyExc_TypeError,
            "C function %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), funcname, sig, desc);
        goto bad;
    }
    tmp.p = PyCObject_AsVoidPtr(cobj);}
#endif
    *f = tmp.fp;
    if (!(*f))
        goto bad;
    Py_DECREF(d);
    return 0;
bad:
    Py_XDECREF(d);
    return -1;
}
#endif

/* InitStrings */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION < 3
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        #else
        if (t->is_unicode | t->is_str) {
            if (t->intern) {
                *t->p = PyUnicode_InternFromString(t->s);
            } else if (t->encoding) {
                *t->p = PyUnicode_Decode(t->s, t->n - 1, t->encoding, NULL);
            } else {
                *t->p = PyUnicode_FromStringAndSize(t->s, t->n - 1);
            }
        } else {
            *t->p = PyBytes_FromStringAndSize(t->s, t->n - 1);
        }
        #endif
        if (!*t->p)
            return -1;
        if (PyObject_Hash(*t->p) == -1)
            return -1;
        ++t;
    }
    return 0;
}

static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    return __Pyx_PyUnicode_FromStringAndSize(c_str, (Py_ssize_t)strlen(c_str));
}
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
#if !CYTHON_PEP393_ENABLED
static const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    char* defenc_c;
    PyObject* defenc = _PyUnicode_AsDefaultEncodedString(o, NULL);
    if (!defenc) return NULL;
    defenc_c = PyBytes_AS_STRING(defenc);
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    {
        char* end = defenc_c + PyBytes_GET_SIZE(defenc);
        char* c;
        for (c = defenc_c; c < end; c++) {
            if ((unsigned char) (*c) >= 128) {
                PyUnicode_AsASCIIString(o);
                return NULL;
            }
        }
    }
#endif
    *length = PyBytes_GET_SIZE(defenc);
    return defenc_c;
}
#else
static CYTHON_INLINE const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    if (unlikely(__Pyx_PyUnicode_READY(o) == -1)) return NULL;
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    if (likely(PyUnicode_IS_ASCII(o))) {
        *length = PyUnicode_GET_LENGTH(o);
        return PyUnicode_AsUTF8(o);
    } else {
        PyUnicode_AsASCIIString(o);
        return NULL;
    }
#else
    return PyUnicode_AsUTF8AndSize(o, length);
#endif
}
#endif
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
    if (
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
            __Pyx_sys_getdefaultencoding_not_ascii &&
#endif
            PyUnicode_Check(o)) {
        return __Pyx_PyUnicode_AsStringAndSize(o, length);
    } else
#endif
#if (!CYTHON_COMPILING_IN_PYPY) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    if (PyByteArray_Check(o)) {
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
    } else
#endif
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject* x) {
    int retval;
    if (unlikely(!x)) return -1;
    retval = __Pyx_PyObject_IsTrue(x);
    Py_DECREF(x);
    return retval;
}
static PyObject* __Pyx_PyNumber_IntOrLongWrongResultType(PyObject* result, const char* type_name) {
#if PY_MAJOR_VERSION >= 3
    if (PyLong_Check(result)) {
        if (PyErr_WarnFormat(PyExc_DeprecationWarning, 1,
                "__int__ returned non-int (type %.200s).  "
                "The ability to return an instance of a strict subclass of int "
                "is deprecated, and may be removed in a future version of Python.",
                Py_TYPE(result)->tp_name)) {
            Py_DECREF(result);
            return NULL;
        }
        return result;
    }
#endif
    PyErr_Format(PyExc_TypeError,
                 "__%.4s__ returned non-%.4s (type %.200s)",
                 type_name, type_name, Py_TYPE(result)->tp_name);
    Py_DECREF(result);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_Check(x) || PyLong_Check(x)))
#else
  if (likely(PyLong_Check(x)))
#endif
    return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  #if PY_MAJOR_VERSION < 3
  if (m && m->nb_int) {
    name = "int";
    res = m->nb_int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = m->nb_long(x);
  }
  #else
  if (likely(m && m->nb_int)) {
    name = "int";
    res = m->nb_int(x);
  }
  #endif
#else
  if (!PyBytes_CheckExact(x) && !PyUnicode_CheckExact(x)) {
    res = PyNumber_Int(x);
  }
#endif
  if (likely(res)) {
#if PY_MAJOR_VERSION < 3
    if (unlikely(!PyInt_Check(res) && !PyLong_Check(res))) {
#else
    if (unlikely(!PyLong_CheckExact(res))) {
#endif
        return __Pyx_PyNumber_IntOrLongWrongResultType(res, name);
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_CheckExact(b))) {
    if (sizeof(Py_ssize_t) >= sizeof(long))
        return PyInt_AS_LONG(b);
    else
        return PyInt_AsSsize_t(b);
  }
#endif
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    const digit* digits = ((PyLongObject*)b)->ob_digit;
    const Py_ssize_t size = Py_SIZE(b);
    if (likely(__Pyx_sst_abs(size) <= 1)) {
        ival = likely(size) ? digits[0] : 0;
        if (size == -1) ival = -ival;
        return ival;
    } else {
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject* o) {
  if (sizeof(Py_hash_t) == sizeof(Py_ssize_t)) {
    return (Py_hash_t) __Pyx_PyIndex_AsSsize_t(o);
#if PY_MAJOR_VERSION < 3
  } else if (likely(PyInt_CheckExact(o))) {
    return PyInt_AS_LONG(o);
#endif
  } else {
    Py_ssize_t ival;
    PyObject *x;
    x = PyNumber_Index(o);
    if (!x) return -1;
    ival = PyInt_AsLong(x);
    Py_DECREF(x);
    return ival;
  }
}
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b) {
  return b ? __Pyx_NewRef(Py_True) : __Pyx_NewRef(Py_False);
}
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
    return PyInt_FromSize_t(ival);
}


#endif /* Py_PYTHON_H */
